#!jinja|pyobjects

import json
import operator


def get_disk_type(disks):
    '''
    Uses disk_rotationrates to assign ssd or rpm depending on
    rotationrate. Then a sum of each disktype being used.

    Parameters
    ----------
    disks : dict
        disk dict from salt
    Returns
    -------
    string
        name of the common disk type for this host
    '''
    disk_rpms = {}

    for disk in disks.values():
        if disk['rotationrate'] == -1:
            disk_type = "UNKNOWN"
        elif disk['rotationrate'] == 0:
            disk_type = "ssd"
        elif disk['rotationrate'] < 8000:
            disk_type = "7200"
        elif disk['rotationrate'] < 12500:
            disk_type = "10000"
        else:
            disk_type = "15000"
        if disk_type not in disk_rpms:
            disk_rpms[disk_type] = 0
        disk_rpms[disk_type] = disk_rpms[disk_type] + 1

    use_disk = "none"
    max_count = 0
    for rpm in disk_rpms:
        if disk_rpms[rpm] > max_count:
            max_count = disk_rpms[rpm]
            use_disk = rpm

    return use_disk


def get_dmi_info(dmi_name, cpu_model, disk_type):
    '''
    The dmi_name will change accordingly using ProductName and
    CpuModel depending on criteria

    Parameters
    ----------
    dmi_name : string
        dmi_name from salt
    cpu_model : string
        cpu_model from salt

    Returns
    -------
    string
        name updated dmi_name where applicable
    '''
    if dmi_name.startswith("X6"):
        dmi_name = 'X6'
    elif dmi_name.startswith("X7"):
        dmi_name = 'X7'
    elif dmi_name.startswith("X9") or dmi_name.startswith("SYS-2UDP-") or dmi_name.startswith("SYS-2027TR-"):
        if disk_type == 'ssd':
            if operator.contains(cpu_model, 'E5-2640'):
                dmi_name = "X9v1"
            else:
                dmi_name = "X9v2"
        else:
            dmi_name = "X9"
    elif dmi_name.startswith("SYS-1018R-WC0R"):
        if operator.contains(cpu_model, 'v4'):
            dmi_name = "X10v2"
        else:
            dmi_name = "X10"
    elif operator.contains(dmi_name, 'R430'):
        if operator.contains(cpu_model, 'v4'):
            dmi_name = "R430v2"
        else:
            dmi_name = "R430"
    elif dmi_name.startswith("SYS-1019P-WTR"):
        if operator.contains(cpu_model, '6136'):
            dmi_name = "X11v1"
        else:
            dmi_name = "X11v2"
    elif operator.contains(dmi_name, 'R440'):
        if operator.contains(cpu_model, '6240'):
            dmi_name = "R440v2"
        else:
            dmi_name = "R440v1"
    elif operator.contains(dmi_name, 'R650'):
        if operator.contains(cpu_model, '6338N'):
            dmi_name = "R650v1"
    return dmi_name.replace(' ', '_')


def save_cap_class(capacity_class, grain='capacity_class', filename='/usr/local/llnw/capclass'):
    '''
    Sets the capacaity class file and grain

    Parameters
    ----------
    capacity_class : string
        generted capacity_class to be written to file and gain

    Returns
    -------

    '''
    __salt__['grains.set'](grain, val=capacity_class)

    File.managed(
        filename,
        contents=capacity_class,
        mode='644',
        makedirs=True
    )


def get_capacity_class():
    '''
    Calls salt grains to get variable data

    Parameters
    ----------

    Returns
    -------
    string
        name updated dmi_name where applicable
    '''
    dmi_name = '{{ grains['productname'] }}'
    cpu_model = '{{ grains['cpu_model'] }}'
    disk_rotationrates = json.loads('''{{ grains['disks']|json }}''')
    disk_type = get_disk_type(disk_rotationrates)
    if disk_type == 'UNKNOWN':
        capacity_class = 'UNKNOWN'
    else:
        capacity_class = '{}_{}'.format(
            get_dmi_info(dmi_name, cpu_model, disk_type),
            disk_type
    )

    return capacity_class


def main():
    capacity_class = get_capacity_class()
    save_cap_class(capacity_class)


main()
