#!/usr/bin/env python

# created by @dvachaiev
# This script is intended to merge 2 EQ Agent checkpoint files into single one

import sys
import json


def merge_files(files):
    """
    >>> merge_files([{'path1': {'name1': {'endpos': 11, 'processing_timestamp': 111.11}}}, {'path2': {'name1': {'endpos': 22, 'processing_timestamp': 222.22}}}])
    {'path2': {'name1': {'processing_timestamp': 222.22, 'endpos': 22}}, 'path1': {'name1': {'processing_timestamp': 111.11, 'endpos': 11}}}
    >>> merge_files([{'path1': {'name1': {'endpos': 11, 'processing_timestamp': 111.11}}}, {'path1': {'name2': {'endpos': 22, 'processing_timestamp': 222.22}}}])
    {'path1': {'name2': {'processing_timestamp': 222.22, 'endpos': 22}, 'name1': {'processing_timestamp': 111.11, 'endpos': 11}}}
    >>> merge_files([{'path1': {'name1': {'endpos': 11, 'processing_timestamp': 111.11}}}, {'path1': {'name1': {'endpos': 22, 'processing_timestamp': 222.22}}}])
    {'path1': {'name1': {'processing_timestamp': 222.22, 'endpos': 22}}}
    >>> merge_files([{'path1': {'name1': {'endpos': 33, 'processing_timestamp': 111.11}}}, {'path1': {'name1': {'endpos': 22, 'processing_timestamp': 222.22}}}])
    {'path1': {'name1': {'processing_timestamp': 111.11, 'endpos': 33}}}
    """
    files = list(files)
    res = files[0]

    for chkpt_files in files[1:]:
        for path, files_info in chkpt_files.items():
            if path not in res:
                res[path] = files_info
                continue

            res_files_info = res[path]
            for name, info in files_info.items():
                if (name not in res_files_info or
                    info["endpos"] > res_files_info[name]["endpos"]):
                    res_files_info[name] = info

    return res


def merge_timebins(batch1, batch2):
    """
    >>> merge_timebins([], [])
    []
    >>> merge_timebins([[None, 13]], [[None, 42]])
    [(None, 42)]
    >>> merge_timebins([[None, 42]], [[None, 13]])
    [(None, 42)]
    >>> merge_timebins([["equal", 42], ["smaller", 13], ["bigger", 42]], [["equal", 42], ["smaller", 42], ["bigger", 13]])
    [('smaller', 42), ('equal', 42), ('bigger', 42)]
    >>> merge_timebins([["fst", 42]], [["sec", 42]])
    [('sec', 42), ('fst', 42)]
    """
    res = dict(batch1)

    for tag, timebin in batch2:
        cur_tb = res.get(tag, 0)
        if timebin > cur_tb:
            res[tag] = timebin

    return res.items()


def merge_batches(batches):
    """
    >>> merge_batches([{"10": [[None, 10]], "20": [[None, 20], ["fst", 20]]}, {'20': [['sec', 22], [None, 22], ['trd', 22]]}, {'30': [[None, 30]]}])
    {'10': [[None, 10]], '30': [[None, 30]], '20': [('sec', 22), ('fst', 20), (None, 22), ('trd', 22)]}
    """
    batches = list(batches)
    res = batches[0]
    for batch in batches[1:]:
        for timebin_duration, submitted_list in batch.items():
            if timebin_duration not in res:
                res[timebin_duration] = submitted_list
                continue

            res[timebin_duration] = merge_timebins(res[timebin_duration], submitted_list)
    return res


def merge_checkpoints(checkpoints):
    return {"files": merge_files(ch["files"] for ch in checkpoints),
        "submitted_batch": merge_batches(ch["submitted_batch"] for ch in checkpoints)
    }


def main():
    if len(sys.argv) < 3:
        print >> sys.stderr, "Usage: {} <checkpoint1> <checkpoint2> ...".format(sys.argv[0])
        print >> sys.stderr, "\n\tTo run tests: python -m doctest -v {}".format(sys.argv[0])
        sys.exit(2)

    checkpoints = [json.load(open(path)) for path in sys.argv[1:]]

    print json.dumps(merge_checkpoints(checkpoints))


if __name__ == "__main__":
    main()

