#!/bin/bash
#
# EP Gridconfig Delay Collector
#
# This script is used to calulate the delay for Edgeprism config
# files that we served from the Gridconfig system and sent into
# tsdb for storage. 

usage(){
    echo "usage: $0 [options]"
    echo "Options:"
    echo "  -d     dryrun output instead of sending to TSDB  "
    echo "  -i     Run immediately instead of waiting for splay"
    echo "Example:"
    echo "  $0 -d -i"
    exit 1
}

while getopts "di" opt; do
  case ${opt} in
      d)
        dryrun=1
        ;;
      i)
        immediate=1
        ;;
      *)
        usage
        ;;
    esac
done

if [ -z "${immediate}" ]; then
  box_number=$(uname -n | grep -Eo '[0-9]+' | head -n1)
  sleep $((box_number % 60))
fi

# Script Vars
#####################################################################
GRID_HASH_URL="http://grconf2.vo.llnwd.net/sli/gridconfig/manifest/EdgePrism/"

# Squid config base path
SQUID_CONF_DIR="/etc/sq/"

# CURL vars
CURL_TIMEOUT=10
CURL_RETRIES=2

# TSDB vars
TSDB_HOST="tsdb-data.lldns.net"
TSDB_PORT="4242"
TSDB_DATA=""

# Hostname - "dns201.pop.llnw.net"
FQDN=$(uname -n)

# Pop - "lax"
POP=$(echo "${FQDN}" | cut -d'.' -f2 | grep -Eo '[a-z]+')

# Files to process
COLLECTION_FILES='auto-cdn_rewrite-v341.conf auto-pg_rewrites.cfg auto-sw_cdn_rewrite.conf'


# Gridconfigclient vertion - "gridconfigclient-2.0.3.106"
GRIDCLIENT_VERSION=$(/usr/local/gridconfig/bin/GridConfigClient.pl -v | awk '{print tolower($1"-"$3)}')

# Get the time and the time aligned to the rearest minute
# This is needed for grafana and tsdb to graph and calulate the 
# data for all cds servers at the same minute.
now=$(date +%s)
now_aligned=$(( now - now % 60 ))
#####################################################################

getFileHash() {
    local filename="$1"
    hash=$(openssl md5 "${SQUID_CONF_DIR}${filename}" | cut -d' ' -f2)
    echo "$hash"
}

getDelay() {
    local filename=$1
    local delay=0
    local url="${GRID_HASH_URL}${filename}.hash"

    # Get the md5 of the local file on disk
    local local_hash=$(getFileHash "${filename}")

    # Get the hash manifest from the GRID_HASH_URL server
    local hashes=$(curl --compressed --user-agent "Gridconfig Delay check from ${FQDN}" -s --retry-max-time ${CURL_TIMEOUT} --connect-timeout ${CURL_TIMEOUT} \
                --max-time ${CURL_TIMEOUT} --no-keepalive --retry ${CURL_RETRIES} --retry-delay 1 --url "${url}" --ipv4)

    # Get the hash that happened before currest disk hash
    local last_hash_line=$(echo "${hashes}" | grep -m1 -B1 "${local_hash}" | head -1)
    local last_hash=$(echo "${last_hash_line}" | awk '{print $2}')
    local last_hash_time=$(echo "${last_hash_line}" | awk '{print $3}')
    
    # If no hash return -1
    if [ -z "$last_hash" ] ; then
        delay=-1
    # If the hashes dont match calculate the delay
    elif [[ "${local_hash}" != "${last_hash}" ]]; then
        delay=$((now - last_hash_time))
    # Assume that the hashs are upto date
    else
        delay=0
    fi

    echo "${delay}"
}

# Build tsdb data lines
for filename in $COLLECTION_FILES ; do
    delay=$(getDelay "$filename")
    if [ "${delay}" -ge 0 ] ; then 
        TSDB_DATA+="put app.gridconfig.client.delay ${now_aligned} ${delay} version=${GRIDCLIENT_VERSION} fqdn=${FQDN} pop=${POP} filename=${filename}\n"
    fi
done

# Print TSDB_DATA in dry or send to TSDB
if [ -n "${dryrun}" ]; then
    echo -ne "${TSDB_DATA}" 
else
    # nc does one retry after a failed attempt. Since the job runs every 60s we're setting this 
    # timeout to 29 to allow 2 seconds for the lock to clear. SRE-5065
    echo -ne "${TSDB_DATA}" | nc -w 29 -N "${TSDB_HOST}" "${TSDB_PORT}"
fi
