#!/usr/bin/env python

# GetBiosSetting.py created by Mark Contreras mcontreras@llnw.com
# This script is used to gather current bios configuration from cds hosts.
# It uses Redfish API to gather the data on Dell hosts.
# It uses sm-sum to gather the data on SuperMicro hosts.
# Text file with output is stored in /usr/local/llnw/zabbix/json/currbios.txt.

import requests
import json
import sys
import re
import time
import os
import warnings
import argparse
from datetime import datetime
import subprocess
import logging

warnings.filterwarnings("ignore")

ipmi_user = "InReach"
ipmi_pass = "Iktc0d3!"
file_path = '/usr/local/llnw/zabbix/json/currbios.txt'
file_temp = '/tmp/currbios.txt'


def remove_temp_file():
    if os.path.exists(file_temp):
        try:
            logging.debug("Deleting file %s" % file_temp)
            os.remove(file_temp)
        except:
            logging.error("Error deleting file %s" % file_temp)


def cleanup_file():
    try:
        logging.debug("Renaming file %s to %s" % (file_temp, file_path))
        os.rename(file_temp, file_path)
    except:
        logging.error("Error renaming file %s" % file_temp)


def main():
    host_type = subprocess.Popen(['dmidecode', '-t', 'baseboard'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = host_type.communicate()
    system = re.search('Manufacturer: (\\w+)', stdout).group(1)
    model = re.search('Product Name: (\\w+)', stdout).group(1)
    logging.debug("System: %s" % system)
    logging.debug("Model: %s" % model)

    if "Dell" in system:
        logging.debug("System is Dell")
        logging.debug("Getting IPMI address")
        process = subprocess.Popen(['ipmitool','lan', 'print'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout, stderr = process.communicate()
        ipmi_ip = re.search('IP Address\s+: (\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})', stdout).group(1)
        
        if ipmi_ip:
            logging.debug("Retrieving IPMI IP was successful: %s" % ipmi_ip)
            url = 'https://%s/redfish/v1/Managers/iDRAC.Embedded.1/Actions/Oem/EID_674_Manager.ExportSystemConfiguration' % ipmi_ip
            payload = {"ExportFormat":"XML","ShareParameters":{"Target":"ALL"}}
            headers = {'content-type':'application/json'}

            response = requests.post(url, data=json.dumps(payload), headers=headers, verify=False, auth=(ipmi_user,ipmi_pass))

            if response.status_code != 202:
                with open(file_temp, "w+") as file:
                    file.write("sysprofile=PerfOptimized BIOS check error")
                    file.write("status_code: %s" % response.status_code)
                    file.write("error: %s" % response.__dict__['_content'])
                logging.error("Fail, status code %s" % response.status_code)
                logging.error(response.__dict__['_content'])
                cleanup_file()
                sys.exit()
            else:
                logging.debug(response.__dict__)
                response_output=response.__dict__

        job_id=response_output["headers"]["Location"]
        try:
            job_id=re.search("JID_.+", job_id).group()
        except:
            logging.error("Failed to load job_id")
            with open(file_temp, "w+") as file:
                file.write("sysprofile=PerfOptimized BIOS check error")
            cleanup_file()
            sys.exit()
        
        logging.debug("Success JobID: %s" % job_id)

        while True:
            logging.debug("Getting job")
            req = requests.get('https://%s/redfish/v1/TaskService/Tasks/%s' % (ipmi_ip, job_id), verify=False, auth=(ipmi_user,ipmi_pass))
            logging.debug(req.status_code)
            if req.status_code == 202:
                d=req.__dict__
                logging.debug(d)
                time.sleep(5)
            elif req.status_code == 200:
                d=req.__dict__['_content']
                logging.debug(d)
                logging.debug("Writing temp file")
                with open(file_temp, "w+") as file:
                    file.write(d)
                logging.debug("Finished getting job")
                break
            else:
                with open(file_temp, "w+") as file:
                    file.write("sysprofile=PerfOptimized redfish failed")
                cleanup_file()
                sys.exit()
    elif "Supermicro" in system:
        logging.debug("System is Supermicro")
        processB = subprocess.Popen(['sm-sum', '-c', 'GetCurrentBiosCfg'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout, stderr = processB.communicate()
        if stdout:
            logging.debug("StdOut: %s" % stdout)
            with open(file_temp, "w+") as file:
                file.write("Supermicro sysprofile=PerfOptimized BIOS check not supported\n")
                file.write(stdout)
        else:
            try:
                with open(file_temp, "w+") as file:
                    file.write("Supermicro sysprofile=PerfOptimized BIOS check not supported")
            except:
                logging.error("Error writing to file %s" % file_temp)
                pass
    else:
        with open(file_temp, "w+") as file:
            try:
                with open(file_temp, "w+"):
                    file.write("Other host type sysprofile=PerfOptimized BIOS check not supported")
            except:
                logging.error("Error writing to file %s" % file_temp)
                pass

if __name__ == '__main__':
    logging.basicConfig(stream=sys.stderr, level=logging.ERROR)
    if '-v' in sys.argv:
        logging.getLogger().setLevel(logging.DEBUG)
    remove_temp_file()
    main()
    cleanup_file()