#!/usr/bin/env python

import argparse
import logging
import logging.handlers
import syslog
import socket
import re
import glob
import sys
import subprocess
import StringIO
import os.path
import json
import yaml
import time
import os
import random
from collections import defaultdict

import pprint
pp = pprint.PrettyPrinter()

TMP_EXTENSION = '.tmp'

CPU_MIN = 10
CPU_LOW_RATE = 0.85
CPU_MID_RATE = 0.95

NET_MIN_RATE = 0.10
NET_LOW_RATE = 0.85
NET_MID_RATE = 0.95

METRIC_TYPE_NET = 'net'
METRIC_TYPE_VCPU = 'vcpu'

USAGE_SUMMARY_LOW = 'low'
USAGE_SUMMARY_MEDIUM = 'medium'
USAGE_SUMMARY_HIGH = 'high'

# The SyslogHandler class is a custom syslog logging handler that uses the
# underlying syslog library. This ensures the correct syslog configuration.
# Whereas, the logging.handlers module in the Python standard library provides a
# syslog handler, but it requires explicit configuration, such as port number.
# Our custom class should prove more robust.
class SyslogHandler(logging.Handler):
    def __init__(self):
        logging.Handler.__init__(self)
        syslog.openlog('autorate', syslog.LOG_PERROR, syslog.LOG_USER)

    def emit(self, record):
        m = self.format(record)
        priority = syslog.LOG_ERR
        if record.levelno >= logging.CRITICAL:
            priority = syslog.LOG_CRIT
        elif record.levelno >= logging.ERROR:
            priority = syslog.LOG_ERR
        elif record.levelno >= logging.WARNING:
            priority = syslog.LOG_WARNING
        elif record.levelno >= logging.INFO:
            priority = syslog.LOG_INFO
        elif record.levelno >= logging.DEBUG:
            priority = syslog.LOG_DEBUG
        syslog.syslog(priority, m)

log = logging.getLogger('autorate')
syslog_handler = SyslogHandler()
log.addHandler(syslog_handler)

args = None # command-line arguments

def nf(val):
    if val is None:
        return 'None'
    return "%.0f" % (val,)

def actual_rate_limit_with_rounding(x):
    return int(x)

def pop_from_fq_domain_name(fqdn):
    # e.g., "cds1234.lax.llnw.net" -> "lax"
    labels = fqdn.split('.')
    if len(labels) < 2:
        return None
    return labels[1]

def bool_to_tsdb_value(x):
    if x:
        return 1
    return 0

def report_to_opentsdb(model, current_time, conn_info=None, output_file=None):
    local_hostname = socket.gethostname()
    local_pop = pop_from_fq_domain_name(local_hostname)
    tsdb_commands = ''
    for label in model.labels:
        key = 'app.ep.server_qos.label_name'
        common_tags = 'fqdn={}, port={}, pop={}, kind={}, label_name={}'.format(
                local_hostname, label.squid_port, local_pop, label.kind,
                label.name)
        # Throughput values are internally represented as kBps, and we report to
        # OpenTSDB as bps.
        tsdb_commands += 'put app.ep.server_qos.thru_bps {} {} {}\n'.format(
                current_time, 8 * 1024 * label.avgkbps, common_tags)
        tsdb_commands += 'put app.ep.server_qos.reqs_per_sec {} {} {}\n'.format(
                current_time, label.rps, common_tags)
        tsdb_commands += 'put app.ep.server_qos.rate_limit_bps {} {} {}\n'.format(
                current_time,
                8 * 1024 * actual_rate_limit_with_rounding(label.newlimkbps),
                common_tags)
        tsdb_commands += 'put app.ep.server_qos.reject {} {} {}\n'.format(
                current_time, bool_to_tsdb_value(label.reject), common_tags)
    if output_file:
        output_file.write(tsdb_commands)
        output_file.flush()
    if not conn_info:
        return
    # If we fail to send after RETRY_COUNT tries, then drop the TSDB data.
    tsdb_host, tsdb_port, retry_count, timeout, delay = conn_info
    log.info('sending stats to OpenTSDB, host={}, port={}'.format(
        repr(tsdb_host), tsdb_port))
    log.debug('sleeping for {} second(s) (random value) before connecting'.format(delay))
    time.sleep(delay)
    for i in range(retry_count):
        try:
            s = socket.create_connection((tsdb_host, tsdb_port), timeout)
            s.sendall(tsdb_commands)
            s.close()
        except socket.error as e:
            log.warning('failed to send stats to OpenTSDB, attempt {} of {}: {}'.format(
                i + 1, retry_count, e))

            continue
        s.close()
        return # success
    log.warning('failed to send stats to OpenTSDB, giving up')

class CacheMgrFetcher:
    def __init__(self, host, port):
        self.__host = host
        self.__port = port

    def fetch_section(self, section):
        max_bytes = 10e6
        req = ("GET cache_object://menu/%s HTTP/1.0\r\n"
               "User-Agent: AutoRate/1.0\r\n\r\n"
              ) % section
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.settimeout(10)
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        s.connect((self.__host, self.__port))
        s.send(req)
        buff = StringIO.StringIO()
        while max_bytes > 0:
            data = s.recv(32768)
            max_bytes -= 32768
            if not data:
                break
            buff.write(data)
        s.close()
        result = buff.getvalue()
        buff.close()
        lines = result.splitlines()
        # read and ignore header
        n = 0
        for line in lines:
            n += 1
            line = line.strip()
            if line == "":
                break
        # iteratively return body
        lines = lines[n:]
        return lines

# returns dictionary of dictionaries per rate label in the ep cache manager
def fetch_labels(cache_mgr_fetcher, autorate_conf_reader, expected_mpep_child_id):
    useful_labels = ('label', 'numwait', 'avgKB/s', 'limKB/s', 'req/s', 'conc')
    labels = {}
    lines = cache_mgr_fetcher.fetch_section('rate_stats')
    if len(lines) < 1:
        return labels
    headers = lines.pop(0).split()
    for line in lines:
        row = {}
        for key, val in zip(headers, line.split()):
            if key in useful_labels:
                if key == 'label':
                    row[key] = val
                else:
                    row[key] = float(val)
        label = row.pop('label', None)
        row['oldlimkbps'] = 0.0
        row['oldreject'] = False
        if label is not None:
            labels[label] = row

    kbyte_limit_re = re.compile('^\s*(if-child\s+(\d+)\s+)?rate_label_kbyte_per_sec_(limit|reject)\s+(\S+)\s+(\d+)')
    kbyte_code_re = re.compile('^\s*(if-child\s+(\d+)\s+)?rate_label_kbyte_per_sec_reject_code\s+(\S+)\s+(\S+)')
    if autorate_conf_reader:
        for line in autorate_conf_reader:
            m1 = kbyte_limit_re.search(line)
            if m1:
                _, mpep_child_id, reject, label, rate = m1.groups()
                if mpep_child_id:
                    mpep_child_id = int(mpep_child_id)
                if label not in labels or mpep_child_id != expected_mpep_child_id:
                    continue
                labels[label]['oldlimkbps'] = float(rate)
                labels[label]['oldreject'] = reject == "reject"
                continue
            m2 = kbyte_code_re.search(line)
            if m2:
                _, mpep_child_id, label, code = m2.groups()
                if mpep_child_id:
                    mpep_child_id = int(mpep_child_id)
                if label not in labels or mpep_child_id != expected_mpep_child_id:
                    continue
                labels[m2.group(3)]['oldrejectcode'] = m2.group(4)
                labels[label]['oldrejectcode'] = code
                continue
    return labels

def fetch_counters(cache_mgr_fetcher):
    stats = {}
    lines = cache_mgr_fetcher.fetch_section('counters')
    if len(lines) < 0:
        return stats
    eq_re = re.compile(r'^(\S+) = ([0-9.-]+)')
    for line in lines:
        m = eq_re.search(line)
        if not m:
            continue
        stats[m.group(1)] = float(m.group(2))
    return stats

# Returns true if and only if the given filename is for an autorate-specific
# squid config file.
def is_autorate_include(fname):
    return 'autorate' in fname

# This function recursively reads the given squid config file and all included
# files. It yields each line stripped of whitespace, and it excludes empty lines
# and comment lines.
#
# This function does *not* recursively follow any include file that's an
# autorate-specific squid config file. Such an `include` line is yielded as
# is--same as a normal line.
#
# This function does *not* understand `if-child` directives, meaning it yields
# any line containing an `if-child include` directive as is--same as a normal
# line.
#
def read_squid_config_lines(fname):
    lines = []
    with open(fname) as f:
        lines = f.readlines()
    for line in lines:
        line = line.lstrip()
        if not line or line[0] == '#':
            pass
        elif line.startswith('include'):
            orig_line = line
            line = line[7:]
            fname = line.strip()
            if is_autorate_include(fname):
                yield orig_line.rstrip()
            if 'rewrite' in fname:
                continue # ignore line
            else:
                for line in read_squid_config_lines(fname):
                    yield line
        else:
            yield line.rstrip()

def parse_squid_config(lines):
    # connect to squid2 unless there is a squid1 http_port config
    pid_filename = None
    autorate_conf = None
    port = None
    mpep_num_children = None
    mpep_base_port = None
    mpep_pid_filename_prefix = None
    for line in lines:
        assert line
        assert line[0] != '#'
        assert not line[0].isspace()
        parts = line.split(None, 1)
        assert 1 <= len(parts) <= 2
        directive = parts[0]
        remainder = '' if 1 == len(parts) else parts[1]
        if directive == 'mpep_num_children':
            if mpep_num_children == None and remainder.isdigit():
                mpep_num_children = int(remainder)
        elif directive == 'mpep_base_port':
            if mpep_base_port == None and remainder.isdigit():
                mpep_base_port = int(remainder)
        elif directive == 'mpep_pid_filename_prefix':
            if mpep_pid_filename_prefix == None:
                mpep_pid_filename_prefix = remainder
        elif directive == 'http_port':
            if port != None:
                pass
            elif remainder.startswith('127.0.0.1:') and remainder[10:].isdigit():
                port = int(remainder[10:])
            elif remainder.isdigit():
                p = int(remainder)
                if 8000 <= p and p < 9000:
                    port = p
        elif directive == 'pid_filename':
            if pid_filename == None:
                pid_filename = remainder
        elif directive == 'include':
            if autorate_conf == None and is_autorate_include(remainder):
                autorate_conf = remainder
    if not autorate_conf:
        raise RuntimeError('no autorate configuration filename found in squid configuration')
    # yield tuple: (host, port, mpep_child_id, pid filename, autorate config)
    if mpep_num_children and mpep_base_port and mpep_pid_filename_prefix:
        assert 1 <= mpep_num_children
        assert 1 <= mpep_base_port
        pid_fname = ('{}.pid'.format(mpep_pid_filename_prefix))
        yield '127.0.0.1', mpep_base_port, 1, pid_fname, autorate_conf
        for i in range(1, mpep_num_children):
            pid_fname = '{}-{}.pid'.format(mpep_pid_filename_prefix, i + 1)
            yield '127.0.0.1', mpep_base_port + i, i + 1, pid_fname, autorate_conf
        return
    if not pid_filename:
        raise RuntimeError('no pid filename found in squid configuration')
    if not port:
        raise RuntimeError('no port found in squid configuration')
    yield '127.0.0.1', port, None, pid_filename, autorate_conf

def get_squid_conf_info(fname):
    def read_pid(pid_fname):
        with open(pid_fname) as f:
            return f.readline().strip()
    config_lines = read_squid_config_lines(fname)
    for host, port, mpep_child_id, pid_fname, autorate_conf in parse_squid_config(config_lines):
        pid = read_pid(pid_fname)
        if pid:
            yield host, port, mpep_child_id, pid, autorate_conf

get_cpu_time_re = re.compile('^(\d+):(\d+).(\d+)')

def get_cpu(pid):
    res = subprocess.check_output(['ps', '-Ho', 'lwp,time,etimes,comm', '-p', pid])
    for line in res.splitlines():
        words = line.split()
        if len(words) < 4:
            continue
        if words[3] == 'squid/squid':
            elapsed = int(words[2])
            m = get_cpu_time_re.search(words[1])
            return float(m.group(1)) * 60 + float(m.group(2)) + float(m.group(3)) / 100.0, elapsed
    return None, 0

def get_squids():
    pairs = (('squid1', args.squid1glob), ('squid2', args.squid2glob))
    for kind, g in pairs:
        found_any = False
        for squid in glob.glob(g):
            for host, port, mpep_child_id, pid, autorate_conf in get_squid_conf_info(squid):
                found_any = True
                yield kind, host, port, mpep_child_id, pid, autorate_conf
        if not found_any:
            log.info("no squid configurations found for %s using glob %s", kind, g)

def get_iface_byte_count(iface_name):
    netstat = subprocess.check_output(['netstat', '-i', '-b', '-I', iface_name])
    byte_counter = 0
    for line in netstat.splitlines():
        words = line.split()
        if len(words) < 11:
            continue
        if "Link" in words[2]:
            return int(words[10])
    raise RuntimeError('could not get byte count for iface {}'.format(
        iface_name))

def get_net_counters():
    # The logic here is to return the lagg if it exists, otherwise return the
    # gateway NIC. One complication is the nicinfo.py script marks all lagg
    # members as gateways, so we can't return a gateway NIC until we're sure
    # there's no lagg. Another complication is we're working within the original
    # design of ADRL, which was lagg-agnostic and picked the NIC based on which
    # iface is most-used. This is why we return a dictionary of NICs even though
    # there's only ever one NIC to return.
    nic_json = subprocess.check_output(['nicinfo.py'])
    nic_info = json.loads(nic_json)
    net = defaultdict(defaultdict)
    for iface in nic_info:
        gateway = bool(nic_info[iface]['gateway'])
        log.debug('found net iface {}, gateway={}'.format(iface, gateway))
        if not gateway:
            continue
        if iface.startswith('lagg') and iface[4:].isdigit():
            net = defaultdict(defaultdict)
            net[iface]['speed'] = nic_info[iface]['negotiated_speed_bps'] / 8
            net[iface]['bytes'] = get_iface_byte_count(iface)
            return net
        net[iface]['speed'] = nic_info[iface]['negotiated_speed_bps'] / 8
    for iface in net:
        net[iface]['bytes'] = get_iface_byte_count(iface)
    return net

def validate_positive_key(d, k):
    if k not in d:
        return False
    if not isinstance(d[k], (int, float)):
        return False
    return d[k] > 0

def validate_all_positive_keys(ds, ks):
    for d in ds:
        for k in ks:
            if not validate_positive_key(d, k):
                return False
    return True

# if there are more ssds than not, return true
def is_ssd(force_ssd):
    if force_ssd == 0:
        return False
    if force_ssd == 1:
        return True
    geom = subprocess.check_output(['geom', 'disk', 'list'])
    rotation_re = re.compile(r'rotationrate:\s+(\d+)')
    ssd_difference = 0
    for line in geom.splitlines():
        m = rotation_re.search(line)
        if m is None:
            continue
        rate = int(m.group(1))
        if rate == 0:
            ssd_difference += 1
        else:
            ssd_difference -= 1
    if ssd_difference > 0:
        return True
    return False

# the usage, target, metric, increment, and util must all be in the same unit of measure
# usage: the actual usage of the combined metric
# target: the desired usage of the combined metric
# target_mid: when above target, what should we attempt to reduce utilization to
# target_low: generally should be about 85% of the target, but this is the low threshold
# get_priority: obtain the priority of the label
# get_metric: obtain the utilization metric of the label
# set_metric: set the utilization metric for the label
# is_shaped: is the current label being shaped
# min_increment: the largest value to decrement from a label in one step
# min_util: the lowest value to reduce the metric of any label
def priority_qos(metric_type, labels, usage, target, target_mid,
        target_low, get_priority, get_metric, set_metric, is_shaped,
        max_decrement, min_util, process=None):
    priority_map = defaultdict(list)
    for label in labels:
        priority_map[label.priority].append(label)
    # if we have to pull traffic down
    process_str = ''
    if process:
        process_str = ', pid={}, mpep_child_id={}'.format(process['pid'],
                process['mpep_child_id'])
    log.debug("priority start, type={}{}, usage={}, target={}".format(
        metric_type, process_str, usage, target))
    if usage > target:
        for priority in sorted(priority_map):
            log.debug("priority={}".format(priority))
            # add a small epsilon to prevent infinite looping
            while usage > target_mid + 0.01:
                log.debug("usage={}, target_mid={}".format(usage, target_mid))
                # if we loop through all of the labels and don't find eligible reductions
                # then look to the next priority
                found_label = False
                cur_labels = sorted(priority_map[priority], key=lambda x: get_metric(x), reverse=True)
                for label in cur_labels:
                    label_metric = get_metric(label)
                    log.debug("label={}, metric={}".format(
                        label.id, label_metric))
                    if label_metric <= min_util:
                        continue
                    # full reduction target
                    reduction = usage - target_mid
                    # prevent the reduction from lowering metric below min_util
                    reduction = min(reduction, label_metric - min_util)
                    # prevent the reduction from exceeding max_decrement
                    reduction = min(max_decrement, reduction)
                    if reduction < 0.01:
                        continue
                    if not set_metric(label, label_metric - reduction):
                        # update was unsuccessful
                        continue
                    usage -= reduction
                    log.debug("label={}, reduction={}, usage={}".format(
                        label.id, reduction, usage))
                    found_label = True
                    # every time we reduce a label, start over including re-sorting
                    # cur_labels looking for the next
                    break
                if found_label == False:
                    break
        return USAGE_SUMMARY_HIGH
    # we are below the high threshold
    available = target - usage
    for label in labels:
        # if we are between the low and high threshold, do not relieve any
        # active shaping
        if usage >= target_low and is_shaped(label):
            log.debug("label={}, usage={}, target_low={}, is_shaped={}".format(
                label.id, usage, target_low, is_shaped(label)))
            continue
        # allow the label to, by an increase in its own traffic alone,
        # raise the metric by the available amount
        label_metric = get_metric(label)
        if label_metric is None:
            log.debug("label={}, metric=None, available={}, min_util={}".format(
                label.id, available, min_util))
            continue
        log.debug("label={}, metric={}, available={}, min_util={}".format(
            label.id, label_metric, available, min_util))
        set_metric(label, max(label_metric + available, min_util))
    if usage >= target_low:
        return USAGE_SUMMARY_MEDIUM
    return USAGE_SUMMARY_LOW

class Metrics:
    def __init__(self, old_state, new_state, force_ssd):
        assert 'time' in new_state
        self.old_state = old_state
        self.new_state = new_state
        self.ssd = is_ssd(force_ssd)
        self.dur = 0
        if not validate_all_positive_keys((self.old_state, self.new_state), ('time',)):
            log.debug('do not have time from previous run so the time delta is unavailable')
            return
        dur = self.new_state['time'] - self.old_state['time']
        log.debug('time delta from previous run is {} seconds'.format(dur))
        MIN_DELTA = 0
        MAX_DELTA = 600
        if (args and args.no_sanity_checks) or (MIN_DELTA < dur <= MAX_DELTA):
            self.dur = dur
        else:
            log.info("time since previous run is out of range, time_delta={}, min={}, max={}".format(
                dur, MIN_DELTA, MAX_DELTA))

    def get_network_usage(self):
        if 'net' not in self.old_state:
            return 0.0, 0.0, 0.0
        if 'net' not in self.new_state:
            return 0.0, 0.0, 0.0
        if self.dur == 0:
            return 0.0, 0.0, 0.0
        max_pct = 0
        max_util_kbps = 0
        max_speed_kbps = 0
        for iface, dat in self.new_state['net'].iteritems():
            if iface not in self.old_state['net']:
                continue
            old_dat = self.old_state['net'][iface]
            if not validate_all_positive_keys((old_dat, dat), ('speed', 'bytes')):
                continue
            max_bytes = self.dur * dat['speed']
            util = dat['bytes'] - old_dat['bytes']
            pct = 100.0 * util / max_bytes
            if pct > 102:
                continue
            if pct > max_pct:
                max_pct = pct
                max_speed_kbps = dat['speed'] / 1024
                max_util_kbps = max_pct / 100 * max_speed_kbps
        return max_pct, max_util_kbps, max_speed_kbps

    def get_ep_usage(self, config):
        ep_dat = defaultdict(lambda: defaultdict(lambda: defaultdict(float)))
        labels = []
        if self.dur == 0:
            return ep_dat, labels
        if 'ep' not in self.new_state:
            return ep_dat, labels

        for pid in self.new_state['ep']:
            if pid not in self.old_state['ep']:
                continue
            dat = self.new_state['ep'][pid]
            old_dat = self.old_state['ep'][pid]
            kind = dat['kind']
            mpep_child_id = dat['mpep_child_id']
            if not validate_all_positive_keys((old_dat, dat),
                    ('kbytes_out', 'requests', 'cpu')):
                continue
            ep_dat[pid]['kind'] = kind
            ep_dat[pid]['kbps_out'] = (dat['kbytes_out'] - old_dat['kbytes_out']) / self.dur
            ep_dat[pid]['rps'] = (dat['requests'] - old_dat['requests']) / self.dur
            ep_dat[pid]['cpu'] = 100 * (dat['cpu'] - old_dat['cpu']) / self.dur
            for label, label_dat in dat['labels'].iteritems():
                if label not in config['labels']:
                    continue
                lconf = config['labels'][label]
                if 'priority' not in lconf:
                    continue

                def get_config_override(key, t, default):
                    val = default
                    if key in config:
                        val = t(config[key])
                    if key in lconf:
                        val = t(lconf[key])
                    return val

                req_kb = get_config_override('req_kb', int, 500)
                reject_code = get_config_override('reject_code', str, None)
                reject = False
                conc_reject = get_config_override('conc_reject', int, None)
                if conc_reject is not None:
                    if conc_reject <= label_dat['conc']:
                        reject = True
                numwait_reject = get_config_override('numwait_reject', int, None)
                if numwait_reject is not None:
                    if numwait_reject <= label_dat['numwait']:
                        reject = True
                l = Label(label, lconf['priority'], kind, pid,
                        self.new_state['ep'][pid]['port'], mpep_child_id,
                        req_kb, reject, reject_code, label_dat)
                labels.append(l)
        return ep_dat, labels


class Label:
    def __init__(self, name, priority, kind, pid, squid_port, mpep_child_id,
            req_kb, reject, reject_code, dat):
        self.name = name
        self.priority = priority
        self.kind = kind
        self.pid = pid
        self.squid_port = squid_port
        self.mpep_child_id = mpep_child_id
        self.id = ":".join((name, kind, pid))
        self.req_kb = req_kb
        self.avgkbps = float(dat['avgKB/s'])
        # per EP-14611 this value is not reliably updated
        # use value from config instead
        # self.limkbps = float(dat['limKB/s'])
        self.limkbps = dat.get('oldlimkbps', 0)
        self.cpulimitkbps = None
        self.netlimitkbps = None
        self.newlimkbps = None
        self.conc = int(dat['conc'])
        self.numwait = int(dat['numwait'])
        self.rps = float(dat['req/s'])
        self.limit_reason = ''
        self.parent = None
        self.reject = reject
        self.reject_code = reject_code
        # this is used by limit_cpu_pid() to set the average response kilobytes
        # when set_metric() is successfully called
        self.avg_resp_kb = None

        self.oldlimkbps = dat.get('oldlimkbps', 0)
        self.oldreject = dat.get('oldreject', False)
        self.oldrejectcode = dat.get('oldrejectcode', None)

class Model:
    def __init__(self, metrics, config):
        self.metrics = metrics
        self.config = config
        ep_dat, labels = self.metrics.get_ep_usage(config)
        label_map = {l.pid:l for l in labels}
        for pid in ep_dat:
            v = ep_dat[pid]
            log.debug('squid stats, pid={}, cpu={}%, thru={} (KB/s), req_rate={} (rps)'.format(
                pid, v['cpu'], v['kbps_out'], v['rps']))
            for l in [l for l in labels if l.pid == pid]:
                log.debug('label stats, name={}, thru={} (KB/s), req_rate={} (rps), old_limit={} (KB/s), conc={}, numwait={}'.format(
                l.name, l.avgkbps, l.rps, l.oldlimkbps, l.conc, l.numwait))
        self.ep_dat = ep_dat
        self.labels = labels
        self.network_pct_usage, self.network_kbps_usage, self.network_speed = self.metrics.get_network_usage()
        log.debug("network_pct_usage={}%, network_kbps_usage={} (KB/s), network_speed={} (KB/s)".format(
            self.network_pct_usage, self.network_kbps_usage,
            self.network_speed))
        self.pid_map = {}
        self.should_reconfig_cached = None
        for pid, items in ep_dat.iteritems():
            self.pid_map[pid] = items

    def label_info(self, label):
        if label.pid not in self.pid_map:
            return None
        ep = self.pid_map[label.pid]
        return 'pid {} mpep_child_id {} name {} cpu {} network {} oldlimkbps {} netlimitkbps {} cpulimitkbps {} newlimitkbps {} oldreject {} reject {} oldrejectcode {} reject_code {} limit_reason {} should_reconfig {}'.format(
                label.pid,
                label.mpep_child_id,
                label.name,
                nf(ep['cpu']),
                nf(self.network_pct_usage),
                nf(label.oldlimkbps),
                nf(label.netlimitkbps),
                nf(label.cpulimitkbps),
                nf(label.newlimkbps),
                label.oldreject,
                label.reject,
                label.oldrejectcode,
                label.reject_code,
                label.limit_reason,
                self.should_reconfig(),
                )

    def limit_cpu_pid(self, cpu_target, pid, mpep_child_id, pid_dat, labels):
        cur_cpu = pid_dat['cpu']

        label_dict = {}
        for label in labels:
            label_dict[label.name] = label
        for label in labels:
            if 'parent' in self.config['labels'][label.name]:
                parent_name = self.config['labels'][label.name]['parent']
                label.parent = label_dict.get(parent_name, None)

        # account for labels with req_kb overrides to ensure the total virtual
        # kilobytes per second is consistent
        corrected_rps_vkbps = pid_dat['rps'] * self.config['req_kb']
        for label in labels:
            corrected_rps_vkbps += label.rps * (label.req_kb - self.config['req_kb'])
        total_vkbps = corrected_rps_vkbps + pid_dat['kbps_out']

        if not (args and args.no_sanity_checks) and total_vkbps < 10:
            return
        if not (args and args.no_sanity_checks) and cur_cpu < 2:
            return

        vkbps_c = cur_cpu / total_vkbps
        cpu_low = CPU_LOW_RATE * cpu_target
        cpu_mid = CPU_MID_RATE * cpu_target

        # set avg_resp_kb
        def get_avg_resp_kbps(label, level):
            if level == 10:
                return None
            if label.avgkbps < 1000:
                if label.parent is None:
                    return None
                return get_avg_resp_kbps(label.parent, level + 1)
            return label.avgkbps / (label.rps + 1)

        for label in labels:
            label.avg_resp_kb = get_avg_resp_kbps(label, 0)


        def get_priority(label):
            return label.priority

        def get_metric(label):
            if label.avg_resp_kb is None:
                return None
            if label.cpulimitkbps:
                return vkbps_c * (label.cpulimitkbps + label.req_kb * label.cpulimitkbps / (label.avg_resp_kb + 1))
            return vkbps_c * (label.avgkbps + label.req_kb * label.avgkbps / (label.avg_resp_kb + 1))

        def cpu_to_kbps(label, cpu):
            if label.avg_resp_kb is None:
                return None
            # cpu = vkbps_c * kbps + vkpbs_c * rps * req_kb
            # rps = kbps / avg_resp_kb
            # cpu = vkbps_c * kbps + vkbps_c * req_kb * kbps / avg_resp_kb
            # kbps * (1 + req_kb / avg_resp_kb) = cpu / vkbps_c
            # kbps = cpu / vkbps_c / (1 + req_kb / avg_resp_kb)
            result = cpu / vkbps_c / (1 + label.req_kb / (label.avg_resp_kb + 1))
            return result

        def set_metric(label, cpu):
            result = cpu_to_kbps(label, cpu)
            if result is None:
                log.debug("set_metric={}, cpu={}, avg_resp_kb=None".format(
                    label.id, cpu))
                return False
            label.cpulimitkbps = result
            log.debug("set_metric={}, cpu={}, val={}, avg_resp_kb={}, vkpbs_c={}, req_kb={}".format(
                label.id, cpu, label.cpulimitkbps, label.avg_resp_kb, vkbps_c,
                label.req_kb))
            return True

        def is_shaped(label):
            return label.numwait > 5

        usage_summary = priority_qos(METRIC_TYPE_VCPU, labels, cur_cpu,
                cpu_target, cpu_mid, cpu_low, get_priority, get_metric,
                set_metric, is_shaped, 1, CPU_MIN,
                process={'pid': pid, 'mpep_child_id': mpep_child_id})
        self.metrics.new_state['usage_summary'][METRIC_TYPE_VCPU][pid] = \
                usage_summary
        for label in labels:
            if label.cpulimitkbps is None and label.oldlimkbps != 0:
                label.cpulimitkbps = label.oldlimkbps

    def limit_cpu(self, pct):
        pid_map = defaultdict(list)
        mpep_child_id_map = {}
        for label in self.labels:
            pid_map[label.pid].append(label)
            mpep_child_id_map[label.pid] = label.mpep_child_id
        for pid, labels in pid_map.iteritems():
            self.limit_cpu_pid(pct, pid, mpep_child_id_map[pid],
                    self.pid_map[pid], labels)

    def limit_network(self, pct):
        if self.network_pct_usage is None:
            return

        # if nic is less than 800 Mbps, abort
        if self.network_speed < 100000:
            return

        pct = float(pct)

        net_target = pct / 100 * self.network_speed
        net_low = NET_LOW_RATE * net_target
        net_mid = NET_MID_RATE * net_target
        max_decrement = 0.01 * net_target
        min_util = NET_MIN_RATE * net_target

        # the methods for priority_qos do not need to translate kbps values
        # since the target, min, mid, usage, and stored values will all be
        # in kbps (kilobytes per second)
        def get_priority(label):
            return label.priority

        def get_metric(label):
            if label.netlimitkbps is not None:
                return label.netlimitkbps
            return label.avgkbps

        def set_metric(label, val):
            label.netlimitkbps = val
            return True

        def is_shaped(label):
            return label.numwait > 5

        usage_summary = priority_qos(METRIC_TYPE_NET, self.labels,
                self.network_kbps_usage, net_target, net_mid, net_low,
                get_priority, get_metric, set_metric, is_shaped, max_decrement,
                min_util)
        self.metrics.new_state['usage_summary'][METRIC_TYPE_NET] = \
                usage_summary

        for label in self.labels:
            if label.netlimitkbps is None and label.oldlimkbps != 0:
                label.netlimitkbps = label.oldlimkbps

    def set_limits(self):
        if self.metrics.ssd:
            self.limit_cpu(self.config['cpu_target'])
        else:
            self.limit_cpu(self.config['cpu_target_hdd'])
        self.limit_network(self.config['net_target'])

        for label in self.labels:
            if label.cpulimitkbps is None and label.netlimitkbps is None:
                continue

            if label.netlimitkbps is None:
                label.newlimkbps = label.cpulimitkbps
                label.limit_reason = 'cpu'
            elif label.cpulimitkbps is None:
                label.newlimkbps = label.netlimitkbps
                label.limit_reason = 'net'
            elif label.netlimitkbps < label.cpulimitkbps:
                label.newlimkbps = label.netlimitkbps
                label.limit_reason = 'net'
            else:
                label.newlimkbps = label.cpulimitkbps
                label.limit_reason = 'cpu'

        def log_usage_summary(metric_type, old_summary, new_summary, pid=None,
                mpep_child_id=None):
            # Don't handle None values specially. Just log any difference.
            assert old_summary == None or type(old_summary) == str or type(old_summary) == unicode, \
                    'Wrong type, got {}'.format(type(old_summary))
            assert type(new_summary) == str, \
                    'Wrong type, got {}'.format(type(new_summary))
            if old_summary == new_summary:
                return # no change, so log nothing
            pid_text = ''
            if pid:
                pid_text = 'pid {} '.format(pid)
            mpep_child_id_text = ''
            if mpep_child_id:
                mpep_child_id_text = 'mpep_child_id {} '.format(mpep_child_id)
            log.info(('usage threshold changed, type {} {}{}previous {} '
                    'current {}').format(
                        metric_type, pid_text, mpep_child_id_text, old_summary,
                        new_summary))
            return

        new_summaries = self.metrics.new_state['usage_summary']
        old_summaries = self.metrics.old_state['usage_summary']
        for metric_type in new_summaries:
            if metric_type == METRIC_TYPE_VCPU:
                for pid in new_summaries[metric_type]:
                    mpep_child_id = \
                            self.metrics.new_state['ep'][pid]['mpep_child_id']
                    new_summary = new_summaries[metric_type][pid]
                    old_summary = None
                    if pid in old_summaries[metric_type]:
                        old_summary = old_summaries[metric_type][pid]
                    log_usage_summary(metric_type, old_summary,
                            new_summary, pid, mpep_child_id)
            else:
                new_summary = new_summaries[metric_type]
                old_summary = old_summaries[metric_type]
                log_usage_summary(metric_type, old_summary, new_summary)

        def limit_summary(x):
            if x is None or x == 0:
                return '-'
            return '{} (KB/s)'.format(x)
        for label in self.labels:
            if label.newlimkbps != label.oldlimkbps:
                log.info('label rate limit changed, label={}, previous={}, current={}'.format(
                    label.name,
                    limit_summary(label.oldlimkbps),
                    limit_summary(label.newlimkbps)))

    def should_reconfig(self):
        if self.should_reconfig_cached is None:
            self.should_reconfig_cached = self.should_reconfig_internal()
        return self.should_reconfig_cached

    def should_reconfig_internal(self):
        for label in self.labels:
            if label.oldlimkbps == 0 and label.newlimkbps == 0:
                continue
            if label.oldreject != label.reject:
                return True
            if label.reject and label.oldrejectcode != label.reject_code:
                return True
            if label.oldlimkbps == 0 and label.newlimkbps != 0:
                return True
            if label.newlimkbps == 0 and label.oldlimkbps != 0:
                return True
            if label.oldlimkbps > label.newlimkbps and label.oldlimkbps * 0.99 > label.newlimkbps:
                return True
            if label.newlimkbps > label.oldlimkbps and label.newlimkbps * 0.99 > label.oldlimkbps:
                return True
        return False

def main():
    parser = argparse.ArgumentParser(description="Dynamically set rate label limits for EdgePrism.")
    parser.add_argument('--config', dest="config_file", default="/etc/sq/autorateconfig.yml", help="location for the main configuration file")
    parser.add_argument('--state', dest="state_file", default="/etc/sq/autoratestate.json", help="location for the state file")
    parser.add_argument('--squid1glob', default="/etc/squid1*.conf", help="glob to match squid1 config files")
    parser.add_argument('--squid2glob', default="/etc/squid2*.conf", help="glob to match squid2 config files")
    parser.add_argument('--reconfigpath', default="/etc/sq/reconfig", help="path to the reconfig file")
    parser.add_argument('--forcessd', default=-1, type=int, help="set to 0 to force a non ssd machine, 1 to force an ssd machine")
    parser.add_argument('--send-to-tsdb', metavar='ADDR', dest="send_to_tsdb", help='send stats to OpenTSDB at the given host and port (e.g "tsdb-data.lldns.net:4242")')
    parser.add_argument('--write-tsdb-commands', metavar='FILE', dest="write_tsdb_commands", help="write OpenTSDB commands to the given file, or - for stdout")
    parser.add_argument('--tsdb-time-alignment', metavar='SECONDS', default=300, type=int, help='factor that time values reported to OpenTSDB will be rounded down to OR zero for no rounding')
    parser.add_argument('--tsdb-retry-count', metavar='COUNT', default=3, type=int, help='number of times to try connecting to OpenTSDB server')
    parser.add_argument('--tsdb-timeout', metavar='SECONDS', default=10, type=int, help='number of seconds before timing out communication with the OpenTSDB server')
    parser.add_argument('--tsdb-max-delay', metavar='SECONDS', default=60, type=int, help='max number of seconds to delay before connecting to OpenTSDB server')
    parser.add_argument('--debug', action="store_true", help="same as --no-reconfig --level=debug --no-syslog")
    parser.add_argument('--no-reconfig', action="store_true", help="do not trigger reconfig")
    parser.add_argument('--squelch', action="store_true", help="if enabled, do not write configurations to stdout or to config files")
    parser.add_argument('--level', default='warning', choices=['debug', 'info', 'warning', 'error', 'critical'])
    parser.add_argument('--no-syslog', action='store_false', dest='syslog', help="write log output to stderr but not syslog")
    parser.add_argument('--debugcfgout', action="store_true", help="write configs to console instead of detected locations")
    parser.add_argument('--debugcfgin', default="", help="use specified config as the existing autorate config instead of detected locations")
    parser.add_argument('--no-sanity-checks', action="store_true", dest="no_sanity_checks", help="do not run environmental sanity checks--useful for debugging")
    global args
    args = parser.parse_args()

    if args.debug:
        args.no_reconfig = True
        args.level = 'debug'
        args.syslog = False

    if not args.syslog:
        log.removeHandler(syslog_handler)
        stream_handler = logging.StreamHandler()
        formatter = logging.Formatter('%(pathname)s:%(lineno)d:%(message)s', datefmt='%Y-%m-%d %H:%M:%S')
        stream_handler.setFormatter(formatter)
        log.addHandler(stream_handler)
    log.setLevel(getattr(logging, args.level.upper()))

    tsdb_conn_info = None
    if args.send_to_tsdb:
        parts = args.send_to_tsdb.rsplit(':', 1)
        if len(parts) != 2:
            log.error("OpenTSDB server address does not contain host and port (e.g., 'host:1234'), cannot report stats to OpenTSDB")
        else:
            host = parts[0]
            port = parts[1]
            if host[0] == '[' and host[-1] == ']':
                host = host[1:-1]
            args.tsdb_max_delay = max(0, args.tsdb_max_delay)
            delay = random.randint(0, args.tsdb_max_delay)
            tsdb_conn_info = (host, port, args.tsdb_retry_count,
                    args.tsdb_timeout, delay)

    config = {}
    with open(args.config_file) as f:
        config = yaml.load(f)

    for val in ('req_kb', 'net_target', 'cpu_target', 'cpu_target_hdd'):
        if val not in config:
            raise ValueError("%s is required at the global configuration scope" % (val,))

    old_state = {}

    if os.path.isfile(args.state_file):
        with open(args.state_file) as f:
            old_state = json.load(f)
    if 'usage_summary' not in old_state:
        old_state['usage_summary'] = {}
    if METRIC_TYPE_VCPU not in old_state['usage_summary']:
        old_state['usage_summary'][METRIC_TYPE_VCPU] = {}
    if METRIC_TYPE_NET not in old_state['usage_summary']:
        old_state['usage_summary'][METRIC_TYPE_NET] = None

    new_state = {}
    new_state['time'] = time.time()
    new_state['usage_summary'] = {}
    new_state['usage_summary'][METRIC_TYPE_VCPU] = {}
    new_state['net'] = get_net_counters()
    new_state['ep'] = {}

    if not new_state['net']:
        raise RuntimeError('no NICs found')

    conf_files_by_name = {}
    conf_files_by_pid = {}
    conf_files_to_close = []
    for kind, host, port, mpep_child_id, pid, autorate_conf in get_squids():
        log.debug('found squid instance: kind={}, host={}, port={}, pid={}'.format(
            kind, host, port, pid))
        cache_mgr_fetcher = CacheMgrFetcher(host, port)
        autorate_conf_reader = None
        fetch_autorate_conf = autorate_conf
        if args.debugcfgin != "":
            fetch_autorate_conf = args.debugcfgin
        if not os.path.isfile(fetch_autorate_conf):
            log.warning("config file does not exist, kind={}, pid={}, config={}".format(
                kind, pid, fetch_autorate_conf))
        else:
            autorate_conf_reader = open(fetch_autorate_conf)
        counters = fetch_counters(cache_mgr_fetcher)
        new_state['ep'][pid] = {}
        new_state['ep'][pid]['kind'] = kind
        new_state['ep'][pid]['mpep_child_id'] = mpep_child_id
        new_state['ep'][pid]['port'] = port
        new_state['ep'][pid]['kbytes_out'] = counters['client_http.kbytes_out']
        new_state['ep'][pid]['requests'] = counters['client_http.requests']
        new_state['ep'][pid]['labels'] = fetch_labels(
                cache_mgr_fetcher, autorate_conf_reader, mpep_child_id)
        if autorate_conf_reader:
            autorate_conf_reader.close()
        cpu, elapsed = get_cpu(pid)
        # if the process doesn't exist or hasn't been running for at least 10 minutes
        if not args.no_sanity_checks and (cpu is None or elapsed < 600):
            del new_state['ep'][pid]
            log.info("squid process is too young, kind={}, pid={}, elapsed={} (s)".format(
                kind, pid, elapsed))
            continue
        new_state['ep'][pid]['cpu'] = cpu

        if not args.squelch:
            assert pid not in conf_files_by_pid
            if args.debugcfgout:
                conf_files_by_pid[pid] = sys.stdout
            elif autorate_conf in conf_files_by_name:
                conf_files_by_pid[pid] = conf_files_by_name[autorate_conf]
            else:
                f = open(autorate_conf + TMP_EXTENSION, 'w')
                conf_files_to_close.append(f)
                conf_files_by_pid[pid] = f
                conf_files_by_name[autorate_conf] = f

    metrics = Metrics(old_state, new_state, args.forcessd)

    model = Model(metrics, config)
    model.set_limits()

    # OpenTSDB reporting
    t = int(time.time())
    m = args.tsdb_time_alignment
    if m:
        t -= t % m
    if args.write_tsdb_commands == '-':
        report_to_opentsdb(model, t, tsdb_conn_info, sys.stdout)
    elif args.write_tsdb_commands:
        with open(args.write_tsdb_commands, 'w') as f:
            report_to_opentsdb(model, t, tsdb_conn_info, f)
    else:
        report_to_opentsdb(model, t, tsdb_conn_info)

    with open(args.state_file, "w") as f:
        json.dump(new_state, f, indent=4)

    # write logs
    for label in model.labels:
        log_line = model.label_info(label)
        log.info(log_line)

    # write configs
    if not args.squelch:
        for label in model.labels:
            child_prefix = '' if label.mpep_child_id is None else 'if-child {} '.format(label.mpep_child_id)
            limit = 'limit'
            if label.reject:
                limit = 'reject'
            if label.newlimkbps > 0:
                f = conf_files_by_pid[label.pid]
                f.write('{}rate_label_kbyte_per_sec_{} {} {} # {} pid {} limited {}\n'.format(
                    child_prefix,
                    limit,
                    label.name,
                    actual_rate_limit_with_rounding(label.newlimkbps),
                    label.kind,
                    label.pid,
                    label.limit_reason,
                    ))
                if label.reject and label.reject_code is not None:
                    f.write('%srate_label_kbyte_per_sec_reject_code %s %s\n' % (
                        child_prefix, label.name, label.reject_code))

        for f in conf_files_to_close:
            f.close()
        config_renamed = False
        for filename in conf_files_by_name.iterkeys():
            t = filename + TMP_EXTENSION
            os.rename(t, filename)
            config_renamed = True

        if config_renamed and not args.no_reconfig and model.should_reconfig():
            open(args.reconfigpath, 'a').close()

if __name__ == "__main__":
    try:
        main()
    except Exception:
        log.exception('a fatal error occurred')
