#!/bin/bash
########################################################################
#   Script: cachedir_clean.sh
#
# Overview: When we remove a multiproc instance the disk space used by 
#           that instance is not cleaned up. This script checks the 
#           squid conf files to find what directories should exist and 
#           what directories should be removed.
#
#   Author: Chris Gutekanst
#    Group: SRE Delivery
#
# Location: https://github.llnw.net/llnw/salt-sae/blob/master/sae-cds/files/cachedir_clean.sh

# init option vars
script_name="cachedir_clean.sh"
dryrun=0;
auto=0;

# Basic logging function to format logs
function log(){
    code=$1
    msg=$2
    timestamp=$(date '+%Y/%M/%d %H:%M:%S')
    echo "${timestamp} [${code}]: ${msg}"
    logger -p daemon.err -t "$script_name" "[${code}]: ${msg}"

}

# Get EP option value from /etc/squid<n>.conf or /etc/sq/squid<n>_options.conf
get_option() {
    local instance="$1"
    local primary="${instance%-[0-9]*}"
    local option="$2"
    sed -E -ne "s/^(if-(before|after)-release +[0-9]+ +)?$option *//p" \
        /etc/${instance}.conf /etc/sq/${primary}_options.conf
}

# Finds all cache_dir entries in the EP configs, and expands them based on
# the mpep_num_children setting for the corresponding EP.
get_all_cache_dirs() {
    local epconf
    for epconf in /etc/squid*.conf; do
        local instance=`echo $epconf | egrep -o 'squid[-0-9]+'`
        local numkids=`get_option $instance mpep_num_children`

        local cachedir
        for cachedir in `get_option $instance cache_dir | awk '{print $2}'`; do
            echo $cachedir

            # if there's more than 1 child, add dirs for them
            if echo $numkids | egrep -v '^[01]$' | egrep -q '^[0-9]+$'; then
                for kidnum in `seq 2 $numkids`; do
                    echo $cachedir-$kidnum
                done
            fi
        done
    done
}

# Check that we are root
if [ $UID -ne 0 ]; then
    log "ERROR" "Must be root to run the script"
    exit 1
fi



# =================
# Parse arguments
# =================


function usage(){
        cat <<END
cachedir_clean.sh - cleans up unused cache dirs

Documentation: http://confluence

Usage example:
  $0 [-h] [-d] [-a]

Options:
  -h : Displays this information.
  -d : Dryrun output do not remove any files.
  -a : auto remove files without prompting.
END
        exit 1
}

while getopts "hda" OPTION; do
    case $OPTION in
        d)
            dryrun="1"
            ;;
        a)
            auto="1"
            ;;
        h)
            usage
            ;;
        *)
            usage
            ;;
    esac
done



# =================================
# Check that graceful not running
# =================================
log "INFO" "Checking for running graceful"

# Get list of lock files on disk
lock_files=$(ls /logs/squid/*graceful_lock)

# Check each squid graceful lock file
for file in $lock_files; do
    # Exit if we are doing a graceful
    pgrep -U squid -F "$file" 2>/dev/null 1>/dev/null
    if [[ $? == 0 ]]; then
        log "ERROR" "Graceful currently running PID:$(cat "$file")"
        exit 1
    fi
done



# ============================================
# Validate that run config files are on disk
# ============================================
log "INFO" "Checking that running configs are on disk"

# Get all squid conf files used by the running squid processes
running_configs=$(ps -U squid | egrep -o '\/etc\/squid[0-9\-]+\.conf')

# Check that the running files are on disk
for file in $running_configs; do
    # If file missing from disk error and quit
    if [ ! -f "$file" ]; then
        log "ERROR" "Config file ${file} running but not on disk. Graceful needed."
        exit 1
    fi
done



# ==============================
# Find extra cache directories 
# ==============================
log "INFO" "Looking for extra cache directories"

# Get sqcache directories from the squid config
config_cachedirs=$(get_all_cache_dirs)

# Get sqcache directories on the disk
disk_cachedirs=$(find /d* -maxdepth 1 -type d -name "sqcache*")

# Use comm to parse the two lists. Find files that are only in the second list. This is a "right excluding join".
extra_directories=$(comm -1 -3  <(echo "$config_cachedirs" | sort) <(echo "$disk_cachedirs" | sort))

# exit if we have no work to do
if [ ! "$extra_directories" ]; then
    log "INFO" "No mismatched directories, exiting."
    exit 1
fi



# ================================
# Remove extra cache directories
# ================================
log "INFO" "Processing extra directories"


# Remove directory with some extra checks
function safe_remove(){
    directory=$1

    log "INFO" "Processing $directory for removal"
    
    # If squid is accessing the folder something is wrong
    if [ "$(lsof +D "$directory")" ]; then
        log "ERROR" "Process currently accessing directory: ${directory}"
        exit 1
    fi

    log "INFO" "Removing directory $directory"
    rm -rf "$directory"

    if [[ $? = 0 ]]; then
        log "INFO" "$directory removed"
    else
        log "ERROR" "Removing $directory failed"
        exit 1
    fi

}

# Exit if we are in dry run
if [[ $dryrun = 1 ]]; then
    log "INFO" "dryrun mode enabled listing files for removal"
    for directory in $extra_directories; do 
        log "INFO" "$directory"
    done
    exit 0
fi

# Check if we are in interactive mode
if [[ $auto = 0 ]]; then
    log "INFO" "Starting interactive removal of files"

    # Print out file list for removal
    echo "Following files are up for deletion"
    echo "$extra_directories"

    read -r -p "Do you want to remove these files? [y/N] " response
    if [[ "$response" =~ ^([yY][eE][sS]|[yY])$ ]]
    then
        for d in $extra_directories; do 
            safe_remove "$d"
        done
    else
        log "INFO" "No files removed"
    fi
# Auto remove files if flag is set
else
    for d in $extra_directories; do 
        safe_remove "$d"
    done
fi
