#!/bin/bash
#
# CDS Virtual Memory Stats Collector
# Collects statistics listed in STATS variable and outputs to TSDB
#
# Usage:
#   cds_vm_stats_collector.sh [options]

### Vars ###
# Constants
METRIC="sys.memory"
ALIGN_INTERVAL=300
TSDB_HOST="tsdb.lldns.net"
TSDB_PORT=4242
TSDB_TIMEOUT=5

# Command-line options
dry_run=0
immediate=0

# TSDB key to sysctl name mappings
# <TSDB Key>:<sysctl name>
read -d '' STATS << 'EOF'
pdshortfalls:vm.stats.vm.v_pdshortfalls
io_faults:vm.stats.vm.v_io_faults
swap_out:vm.stats.vm.v_swapout
swap_in:vm.stats.vm.v_swapin
swap_pgs_out:vm.stats.vm.v_swappgsout
swap_pgs_in:vm.stats.vm.v_swappgsin
vm_faults:vm.stats.vm.v_vm_faults
laundry_count:vm.stats.vm.v_laundry_count
inactive_count:vm.stats.vm.v_inactive_count
active_count:vm.stats.vm.v_active_count
wire_count:vm.stats.vm.v_wire_count
free_count:vm.stats.vm.v_free_count
page_count:vm.stats.vm.v_page_count
EOF

# Globals
tsdb_output=()
fqdn=$( hostname -f )
host=$( echo "${fqdn}" | /usr/bin/awk -F "." '{ print $1 }' )
pop=$( echo "${fqdn}" | /usr/bin/awk -F "." '{ print $2 }' )
host_type=$( echo "${host}" | /usr/bin/sed -r "s/^([a-zA-Z]{1,5}).*/\1/" )
host_num=$( echo "${host}" | /usr/bin/sed -r "s/^[a-zA-Z]*([0-9]{1,5}).*$/\1/" );
splay_time=$(( "${host_num}" % ALIGN_INTERVAL ))
sample_time=$( /bin/date +%s )
sample_time_aligned=$(( "${sample_time}" - "${sample_time}" % "${ALIGN_INTERVAL}" ))

### Functions ###
# Output error messages to STDERR with timestamp
error(){
  echo "[$( /bin/date +'%Y-%m-%dT%H:%M:%S%z' )]: $@" 1>&2
}

# Print usage summary
usage(){
    echo "Usage: $0 [options]"
    echo "Options:"
    echo "  -d  Print to STDOUT instead of sending to TSDB"
    echo "  -i  Run immediately instead of wating for splay"
    echo "Example:"
    echo "  $0 -d -i"
    exit 1
}

# Retrieve stats from sysctl and store in tsdb_output
get_stats(){
    for record in ${STATS}; do
        local key=$( echo "${record}" | /usr/bin/awk -F ":" '{ print $1 }' )
        local name=$( echo "${record}" | /usr/bin/awk -F ":" '{ print $2 }' )
        local counter=$( /sbin/sysctl -n "${name}" )
        
        # Save results in tsdb_output, but don't write
        # garbage data to TSDB if there is a problem fetching
        # the stat
        if [ $? -ne 0 ]; then
            error "Counter ${name} not found or problem with sysctl. Skipping this metric"
        else
            tsdb_output+="put ${METRIC}.${key} ${sample_time_aligned} ${counter} fqdn=${fqdn} pop=${pop}\n"
        fi
    done
}

# Transmit data to tsdb
send_data(){
    if [ "${dry_run}" -ne 1 ]; then
        echo -ne "${tsdb_output}" | /usr/bin/nc -N -w "${TSDB_TIMEOUT}" "${TSDB_HOST}" "${TSDB_PORT}"
        if [ $? -ne 0 ]; then
            error "Problem sending output to TSDB"
            exit 1
        fi
    else   
        echo -ne "${tsdb_output}"
    fi
}

# Wait for 'random' number of seconds based on 
# host number
splay(){
    if [ "${immediate}" -eq 0 ]; then
        /bin/sleep "${splay_time}"
    fi
}

# Main function
main(){
    if [ "${host_type}" != 'cds' ]; then
        error "Not running on cds host. Exiting."
        exit 1
    fi

    splay
    get_stats

    if [ -n "$tsdb_output" ]; then
        send_data
    else
        error "No data to send. Exiting"
        exit 1
    fi
}

# Process command line options
while getopts "di" opt; do
    case "${opt}" in 
        d)
            dry_run=1
            ;;
        i)
            immediate=1
            ;;
        *)
            usage
            ;;
    esac
done

main
