#!/bin/bash                                    

#
# CDS Log Expander
# Author: Dominick Maggio
# 
# Description:
# This is a multi-use script that was created to address the need for expanding the log
# partitions of CDS hosts.  It is only meant to be used on hosts with contiguous /logs and
# /d0 partitions.  If the host has an unexpected configuration, the script will exit
# gracefully and inform the user of what went wrong.  The user must understand that using 
# this script will wipe the data on the first cache drive (/d0) in order to obtain the
# free space necessary to expand /logs.
# 

KILL_MAXIMUM=8

mbr=0                            
gpt=0                            
ada0_present=0                   
da0_present=0                    
force=0
nostart=0
fill=0
forcekill=0
drive=""
logs=""
d0=""
logsize="300G"
kerndisks=$(sysctl -n kern.disks)

usage() {                                            
  echo "$0 [-fl]"                                    
  echo "  -h: You made it here once, use -h to do it again"
  echo "  -f: Force the script to run"               
  echo "  -k: Force kill all remaining processes writing to logs.  This will only kill up to a maximum of $KILL_MAXIMUM processes."
  echo "  -l: /logs partition size, default is 300G.  Format: 0-999[K,M,G,T]"
  echo "  -n: Do NOT start squid once the script has completed"
  echo "  -N: Do NOT start any services once the script has completed"
  echo "  -i: Fill remainder of disk with /logs partition, do not recreate /d0, and comment out the entry for /d0 in /etc/fstab"
  exit 1
}


while [ -n "$1" ]                                    
do
  case $1 in
    -h) usage;;
    -f) force=1; shift 1;;
    -k) forcekill=1; shift 1;;
    -l) logsize="$2"; shift 2;;
    -n) nostart=1; shift 1;;
    -N) nostart=2; shift 1;;
    -i) fill=1; shift 1;;
    *) echo "Unkown Option '$1'"; usage;;
  esac
done

# Check for root
if (( $EUID != 0 )); then
  echo "This script must be run as root."
  exit
fi

# Perform a light check against the motd to determine if the host may be taking traffic                   
if grep -iq "In Service" /etc/motd
then
  if grep -iq "HOST ISSUE" /etc/motd
  then
    :
  else
    if [ $force = 0 ];
    then
      echo "This host may be taking traffic! -f to force."                                                  
      exit 1
    fi
  fi
fi

# Check for mid-tier and verify the host is a cds host.
if egrep -iq ".*mid.*tier.*" /etc/sq/dedicated_id.txt
then
  echo "This script is not configured for mid-tier hosts."
  exit 1
elif egrep -iq "hostname=\"cds" /etc/rc.conf || [ $force -eq 1 ]
then
  :
else
  echo "This script is tailored to CDS hosts, and not meant to run elsewhere.  -f to force."
  exit 1 
fi

# Verify logsize formatting and normalize to K.
if echo "$logsize" | egrep -q '^[0-9]{1,3}[K,M,G,T]';                        
then                                                                      
  logint=$(echo "$logsize" | sed "s/^\([0-9]\{1,3\}\).*/\1/")             
else                                                                      
  echo "Check -l, this is not a recognized log size."                     
  exit 1                                                                  
fi                                                                        
                                                                                
case $(echo "$logsize" | sed "s/^[0-9]\{1,3\}\([K,M,G,T]\).*/\1/") in     
  K)logsize=$(($logint));;                                                
  M)logsize=$(($logint * 1024));;                                         
  G)logsize=$(($logint * 1024 * 1024));;                                  
  T)logsize=$(($logint * 1024 * 1024 * 1024));;                           
  *) echo "Log size is not recognized.  Exiting."; exit 1;;               
esac                                                                      
		                                                                          

# Fail if logs is already equal to or greater than target
if [ $(df | grep "/logs" | tr -s "\t" " " | cut -d ' ' -f 2) -ge $logsize ];
then
  echo "This log partition is already greater than or equal to ${logsize}K."
  exit 1
fi

# Check for presence of ada0 and da0                 
for disk in $kerndisks; do
  if [ $disk = "ada0" ]; then
    ada0_present="1"
    drive="ada0"
  fi
  if [ $disk = "da0" ]; then
    da0_present="1"
    drive="da0"
  fi
done

# Fail if unexpected configuration                   
if [ $ada0_present = 1 ] && [ $da0_present = 1 ];
then
  echo "Both ada0 and da0 are present."              
  exit 1
elif [ $ada0_present = 0 ] && [ $da0_present = 0 ];
then
  echo "Neither ada0 nor da0 are present."           
  exit 1
fi

# Detect GPT/MBR
if ls /dev/*da0* | egrep ".*da0s[1-9][a-j].*" > /dev/null
then
  mbr="1"
  drive="${drive}s1"
fi
if ls /dev/*da0* | egrep ".*da0p[1-9].*" > /dev/null
then
  gpt="1"
fi

# Fail if something goes wrong with mbr/gpt detection, else assign index variables.                       
if [ $mbr = 1 ] && [ $gpt = 1 ];
then
  echo "System cannot be both MBR and GPT."          
  exit 1
elif [ $mbr = 1 ];
then
  logs=$(grep "/logs" /etc/fstab | grep -v "#" | sed "s/.*da0s1\([a-j]\).*/\1/" | tr [a-j] [1-9])
  d0=$(grep "/d0" /etc/fstab | grep -v "#" | sed "s/.*da0s1\([a-j]\).*/\1/" | tr [a-j] [1-9])
elif [ $gpt = 1 ];
then
  logs=$(grep "/logs" /etc/fstab | grep -v "#" | sed "s/.*da0[p,s]\([0-9]\).*/\1/")
  d0=$(grep "/d0" /etc/fstab | grep -v "#" | sed "s/.*da0[p,s]\([0-9]\).*/\1/")
else
  echo "Cannot determine whether the system is MBR or GPT."                                               
  exit 1
fi

# Drives are commented out in /etc/fstab
if [ "$(grep "/logs" /etc/fstab)" = "" ];
then
  echo "I have no active fstab records for logs, please verify entries and re-run script."
  exit 1
elif [ "$(grep "/d0" /etc/fstab)" = "" ];
then
  echo "I have no active fstab records for d0, please verify entries and re-run script."
  exit 1
fi

# Make sure /logs is immediately before /d0, if they are not contiguous this script will fail.
if [ $((logs + 1)) -ne $d0 ]
then
  echo "/logs and /d0 partitions were not found where they were expected to be."
  gpart show $drive
  exit 1
fi

# Check to see if squid is running 
if ps aux | egrep -iq squid[12]    
then                               
  echo "Squid is running, executing '/etc/sq/rc.squid stop'"
  /etc/sq/rc.squid stop > /dev/null
  sleep 8
fi                                 

service eqrldp stop
service edgequeryd stop
service eqbilling stop
service edgequery_logstreamer stop
service syslog-ng stop
killall logpusher
kill $(pgrep -f logfeed) &> /dev/null
kill $(pgrep -f loghit) &> /dev/null
sleep 5

# First attempt to unmount /logs and /d0
umount /logs &> /dev/null
umount /d0 &> /dev/null

activepids=$(lsof | grep -E "/d0|/logs" |  tr -s '\t' ' ' | cut -d' ' -f2 | uniq)

if mount | egrep "d0|logs" > /dev/null && [ $forcekill -eq 1 ]
then
  # Force kill not more than 4 processes
  if [ $(echo "$activepids" | wc -l) -gt $KILL_MAXIMUM ];
  then
    echo "Too many active pids writing to /d0 and /logs.  Quitting..."
    exit 1
  else
    for i in $activepids;
    do
      pname=$(ps -p $i -o comm)
      pcomm=$(ps -p $i -o command)
      echo "Killing: $pname PID: $i"
      echo "Started with:"
      echo "$pcomm"
      kill $i
    done
    # Attempt to unmount again
    umount /logs &> /dev/null
    umount /d0 &> /dev/null
    if mount | egrep "d0|logs" > /dev/null
      then
      echo "Partitions are still mounted after force kill.  Exiting..."
      exit 1
    fi
    sleep 5
  fi
  echo "Both partitions successfully unmounted, beginning resize..."
elif mount | egrep "d0|logs" > /dev/null
then
  echo "The following processes were still running when attempting to umount /logs and /d0" 
  echo "Stop these processes manually, or attempt to run the script with -k"                
  # Partitions were not able to unmount
  for i in $activepids;
  do
    ps -p $i -o pid,user,command
  done
  exit 1
else
  # Partitions successfully unmounted
  echo "Both partitions successfully unmounted, beginning resize..."
fi

# Define partition lettering for MBR
if [ $mbr = 1 ]
then
  logsltr=$(echo "$logs" | tr [1-9] [a-j])
  d0ltr=$(echo "$d0" | tr [1-9] [a-j])
fi

gpart delete -i $d0 $drive
if [ $fill -eq 0 ]
then
  gpart resize -i $logs -s ${logsize}K $drive
else
  gpart resize -i $logs $drive
fi

# Different commands for MBR/GPT.
if [ $mbr = 1 ];
then
  growfs -y /dev/${drive}${logsltr}
  if [ $fill = 0 ];
  then
    gpart add -t freebsd-ufs -i $d0 ${drive}
  fi
elif [ $gpt = 1 ];
then
  growfs -y /dev/${drive}p$logs
  if [ $fill = 0 ];
  then
    gpart add -a 1M -t freebsd-ufs -l cache0 ${drive}
  fi
fi
if [ $fill = 1 ];
then
  sed -e '/\/d0/ s/^#*/#/' -i /etc/fstab
fi

/etc/sq/cdsdisk.pl
mount -a

if [ $nostart -eq 2 ]
then
  exit 0
elif [ $nostart -lt 2 ]
then
  service eqrldp start
  service edgequeryd start
  service eqbilling start
  service edgequery_logstreamer start
  service syslog-ng start
fi
if [ $nostart -lt 1 ]    
then                     
  /etc/sq/rc.squid start 
fi                       

exit 0
