#!/usr/bin/env python

import yaml
import unittest
import logging
import argparse
from collections import defaultdict
import StringIO

import sys
sys.dont_write_bytecode = True

import autorate

DEFAULT_TIME_DELTA = 300 # expect to run every 5 minutes

class ModelTest:
    def __init__(self, config_yaml, net_util, net_speed, time_delta = None):
        self.time_delta = time_delta
        if self.time_delta is None:
            self.time_delta = DEFAULT_TIME_DELTA
        self.net_speed = float(net_speed) * 1024
        self.time = 1.0
        self.bytes = 1
        self.config = yaml.load(config_yaml)
        self.old = self.__new_state(0, 0)
        self.new = self.__new_state(self.time_delta, net_util)
        self.old_model = None

    def __new_state(self, time_delta, net_util):
        bytes_delta = self.net_speed * net_util * time_delta / 100
        s = {}
        s['ep'] = defaultdict(lambda: defaultdict(lambda: defaultdict(float)))
        s['net'] = defaultdict(lambda: defaultdict(float))
        s['time'] = self.time + time_delta
        s['usage_summary'] = {}
        s['usage_summary'][autorate.METRIC_TYPE_VCPU] = {}
        s['usage_summary'][autorate.METRIC_TYPE_NET] = None
        s['net']['ix0']['bytes'] = self.bytes + bytes_delta
        s['net']['ix0']['speed'] = self.net_speed
        self.bytes += bytes_delta
        self.time += time_delta
        return s

    def next_period(self, old_model, net_util, time_delta = None):
        self.time_delta = time_delta
        if self.time_delta is None:
            self.time_delta = DEFAULT_TIME_DELTA
        self.old = self.new
        self.new = self.__new_state(self.time_delta, net_util)
        self.old_model = old_model

    def set_ep(self, kind, pid, kbps, rps, cpu):
        ep = self.new['ep'][str(pid)]
        ep['kind'] = kind
        ep['requests'] = float(rps) * self.time_delta + 1
        ep['cpu'] = float(cpu) * self.time_delta / 100 + 1
        ep['kbytes_out'] = float(kbps) * self.time_delta + 1
        ep['labels'] = defaultdict(lambda: defaultdict(float))
        ep['mpep_child_id'] = None
        ep = self.old['ep'][str(pid)]
        if 'kind' not in ep:
            ep['kind'] = kind
            ep['requests'] = 1.0
            ep['cpu'] = 1.0
            ep['kbytes_out'] = 1.0
            ep['mpep_child_id'] = None

    def set_label(self, kind, pid, name, kbps, rps, limit, conc, numwait):
        label = self.new['ep'][str(pid)]['labels'][name]
        label['avgKB/s'] = float(kbps)
        label['req/s'] = float(rps)
        label['conc'] = float(conc)
        label['oldlimkbps'] = float(limit)
        label['numwait'] = float(numwait)
        # The following is a hack to carry over the `oldlimkbps` value from the
        # previous run. This matters only when using the `next_period` feature.
        if self.old_model:
            labels_by_name = {label.name: label for label in self.old_model.labels}
            if name in labels_by_name and \
                labels_by_name[name].newlimkbps is not None:
                label['oldlimkbps'] = labels_by_name[name].newlimkbps

    def get_model(self):
        metrics = autorate.Metrics(self.old, self.new, 1)
        return autorate.Model(metrics, self.config)

# Mocks the CacheMgrFetcher class to provide CacheMgr results without needing to
# connect to an actual squid instance.
class MockCacheMgrFetcher:
    def __init__(self):
        self.__sections = {}
    def set_section(self, section, lines):
        self.__sections[section] = lines
    def fetch_section(self, section):
        return self.__sections[section]

class AutoRateTestCase(unittest.TestCase):
    def test_cpu_limit(self):
        config_yaml = """---
req_kb: 500
net_target: 85
cpu_target: 80
cpu_target_hdd: 50
labels:
  highcpu:
    priority: 0
  lowcpu:
    priority: 1
"""
        mt = ModelTest(config_yaml, 50, 2500000)
        mt.set_ep('squid1', 1, 500000, 1000, 85)
        mt.set_label('squid1', 1, 'highcpu', 400000, 800, 0, 0, 0)
        mt.set_label('squid1', 1, 'lowcpu', 50000, 100, 0, 0, 0)
        mt.set_ep('squid1', 2, 500000, 1000, 85)
        mt.set_label('squid1', 2, 'highcpu', 400000, 800, 0, 0, 0)
        mt.set_label('squid1', 2, 'lowcpu', 50000, 100, 0, 0, 0)
        mt.set_ep('squid2', 3, 500000, 1000, 85)
        mt.set_label('squid2', 3, 'highcpu', 200000, 400, 0, 0, 0)
        mt.set_label('squid2', 3, 'lowcpu', 25000, 50, 0, 0, 0)
        model = mt.get_model()

        model.set_limits()
        label_map = {"%s:%s:%s" % (label.kind, label.pid, label.name): label for label in model.labels}
        self.assertAlmostEqual(label_map['squid1:1:highcpu'].newlimkbps, 347038, delta=10)
        self.assertAlmostEqual(label_map['squid1:2:highcpu'].newlimkbps, 347038, delta=10)
        self.assertAlmostEqual(label_map['squid2:3:highcpu'].newlimkbps, 147071, delta=10)
        self.assertAlmostEqual(label_map['squid1:1:lowcpu'].newlimkbps, 925000, delta=10)
        self.assertAlmostEqual(label_map['squid1:2:lowcpu'].newlimkbps, 925000, delta=10)
        self.assertAlmostEqual(label_map['squid2:3:lowcpu'].newlimkbps, 900000, delta=10)

    def test_net_limit(self):
        config_yaml = """---
req_kb: 500
net_target: 85
cpu_target: 80
cpu_target_hdd: 50
labels:
  highnet:
    priority: 0
    conc_reject: 50
    numwait_reject: 50
  lownet:
    priority: 1
"""
        mt = ModelTest(config_yaml, 90, 2500000)
        mt.set_ep('squid1', 1, 1125000, 1000, 35)
        mt.set_label('squid1', 1, 'highnet', 787500, 800, 0, 100, 0)
        mt.set_label('squid1', 1, 'lownet', 337500, 100, 0, 0, 0)
        mt.set_ep('squid1', 2, 1125000, 1000, 35)
        mt.set_label('squid1', 2, 'highnet', 787500, 800, 0, 0, 100)
        mt.set_label('squid1', 2, 'lownet', 337500, 100, 0, 0, 0)
        model = mt.get_model()

        model.set_limits()
        label_map = {"%s:%s:%s" % (label.kind, label.pid, label.name): label for label in model.labels}
        self.assertTrue(label_map['squid1:1:highnet'].reject)
        self.assertTrue(label_map['squid1:2:highnet'].reject)
        self.assertFalse(label_map['squid1:1:lownet'].reject)
        self.assertFalse(label_map['squid1:2:lownet'].reject)
        self.assertAlmostEqual(label_map['squid1:1:highnet'].newlimkbps, 662500, delta=10)
        self.assertAlmostEqual(label_map['squid1:2:highnet'].newlimkbps, 681250, delta=10)
        self.assertAlmostEqual(label_map['squid1:1:lownet'].newlimkbps, 2154926, delta=10)
        self.assertAlmostEqual(label_map['squid1:2:lownet'].newlimkbps, 2154926, delta=10)

    def test_cpu_limit_priority(self):
        config_yaml = """---
req_kb: 500
net_target: 85
cpu_target: 80
cpu_target_hdd: 50
labels:
  lowpri:
    priority: 0
  highpri:
    priority: 1
"""
        mt = ModelTest(config_yaml, 50, 2500000)
        mt.set_ep('squid1', 1, 190000, 0, 75)
        mt.set_label('squid1', 1, 'lowpri', 100000, 0, 0, 0, 0)
        mt.set_label('squid1', 1, 'highpri', 90000, 0, 0, 0, 0)
        model = mt.get_model()

        model.set_limits()
        for label in model.labels:
            log_line = model.label_info(label)
            autorate.log.debug(log_line)

        label_map = {"%s:%s:%s" % (label.kind, label.pid, label.name): label for label in model.labels}
        self.assertAlmostEqual(label_map['squid1:1:lowpri'].newlimkbps, 112603, delta=10)
        self.assertAlmostEqual(label_map['squid1:1:highpri'].newlimkbps, 102596, delta=10)

    def test_cpu_limit_low(self):
        config_yaml = """---
req_kb: 500
net_target: 85
cpu_target: 5
cpu_target_hdd: 50
labels:
  lowpri:
    priority: 0
  highpri:
    priority: 1
"""
        mt = ModelTest(config_yaml, 30, 750000)
        # 30% of 20 Gbps
        mt.set_ep('squid1', 1, 750000, 0, 30)
        mt.set_label('squid1', 1, 'lowpri', 375000, 100, 0, 0, 0)
        mt.set_label('squid1', 1, 'highpri', 375000, 200, 0, 0, 0)
        model = mt.get_model()

        model.set_limits()
        for label in model.labels:
            log_line = model.label_info(label)
            autorate.log.debug(log_line)

        label_map = {"%s:%s:%s" % (label.kind, label.pid, label.name): label for label in model.labels}
        self.assertAlmostEqual(label_map['squid1:1:lowpri'].newlimkbps, 220329, delta=10)
        self.assertAlmostEqual(label_map['squid1:1:highpri'].newlimkbps, 197183, delta=10)

    def test_parse_squid_config_non_mpep_squid1(self):
        got = set(autorate.parse_squid_config([
            'include /etc/autorate.conf',
            'http_port 127.0.0.1:1234',
            'pid_filename /var/run/squid1.pid',
        ]))
        expected = set([
            ('127.0.0.1', 1234, None, '/var/run/squid1.pid', '/etc/autorate.conf'),
        ])
        self.assertEqual(got, expected)

    def test_parse_squid_config_non_mpep_squid2(self):
        got = set(autorate.parse_squid_config([
            'include /etc/autorate.conf',
            'http_port 80', # ignore port value out of range
            'http_port 8080',
            'pid_filename /var/run/squid1.pid',
        ]))
        expected = set([
            ('127.0.0.1', 8080, None, '/var/run/squid1.pid', '/etc/autorate.conf'),
        ])
        self.assertEqual(got, expected)

    def test_parse_squid_config_mpep(self):
        got = set(autorate.parse_squid_config([
            'include /etc/autorate.conf',
            'mpep_num_children 3',
            'mpep_base_port 1234',
            'mpep_pid_filename_prefix /var/run/squid-pid',
        ]))
        expected = set([
            ('127.0.0.1', 1234, 1, '/var/run/squid-pid.pid', '/etc/autorate.conf'),
            ('127.0.0.1', 1235, 2, '/var/run/squid-pid-2.pid', '/etc/autorate.conf'),
            ('127.0.0.1', 1236, 3, '/var/run/squid-pid-3.pid', '/etc/autorate.conf'),
        ])
        self.assertEqual(got, expected)

    def test_ep_14703(self):
        NET_SPEED = 1000000 # arbitrary value for testing
        for net_target in [20, 50, 80, 95]:
            config_yaml = """---
req_kb: 500
net_target: {}
cpu_target: 80
labels:
  highnet:
    priority: 0
""" \
                .format(net_target)
            for num_squids in range(1, 5):
                mt = ModelTest(config_yaml, 100, NET_SPEED)
                for pid in range(num_squids):
                    # CPU use is zero to ensure that only network limiting occurs, not
                    # CPU limiting. Request rates don't matter.
                    kind = 'squid1' if pid % 2 == 0 else 'squid2'
                    mt.set_ep(kind, pid, NET_SPEED / 4, 1000, 0)
                    mt.set_label(kind, pid, 'highnet', NET_SPEED / 5, 100, 0, 0, 0)
                model = mt.get_model()
                model.set_limits()
                low_limit = NET_SPEED * net_target / 100.0 * autorate.NET_MIN_RATE
                self.assertTrue(
                    all([low_limit <= x.newlimkbps for x in model.labels]),
                    'One or more limits set to below minimum (limits={}, net_target={}, num_squids={})'.format(
                        [x.newlimkbps for x in model.labels], net_target,
                        num_squids))
                rate_target = NET_SPEED * net_target / 100.0 * autorate.NET_MID_RATE
                max_reduction = NET_SPEED - rate_target
                actual_reduction = sum(
                    [NET_SPEED / 5 - x.newlimkbps for x in model.labels])
                self.assertLessEqual(
                    actual_reduction, max_reduction,
                    'Total rate-limit reduction exceeds target (limits={}, max_reduction={}, net_target={}, num_squids={})'.format(
                        [x.newlimkbps for x in model.labels], max_reduction,
                        net_target, num_squids))

    # This test case doesn't target any specific functionality. It exists merely
    # as an example for simulating multiple runs, which I initially used to
    # manually verify logging.
    def test_multi_period(self):
        config_yaml = """---
req_kb: 500
net_target: 85
cpu_target: 5
cpu_target_hdd: 50
labels:
  alpha:
    priority: 0
"""
        mt = ModelTest(config_yaml, 86, 100 * 1000)
        mt.set_ep('squid1', 1, 50 * 1000, 0, 1)
        mt.set_label('squid1', 1, 'alpha', 30 * 1000, 0, 0, 0, 0)
        model = mt.get_model()
        model.set_limits()
        for label in model.labels:
            log_line = model.label_info(label)
            autorate.log.debug(log_line)
        mt.next_period(model, 30)
        mt.set_ep('squid1', 1, 40 * 1000, 0, 1)
        mt.set_label('squid1', 1, 'alpha', 20 * 1000, 0, 0, 0, 0)
        model = mt.get_model()
        model.set_limits()
        for label in model.labels:
            log_line = model.label_info(label)
            autorate.log.debug(log_line)

class FetchLabelsTestCase(unittest.TestCase):

    def setUp(self):
        self.cache_mgr_fetcher = MockCacheMgrFetcher()
        self.cache_mgr_fetcher.set_section('rate_stats', [
            'label                avgKB/s numwait    avgwait limKB/s      req/s  conc    avgconc maxconc conc-rej/s maxKB/s   KB-rej/s maxreq/s  req-rej/s',
            'sonycoment             18012       0     0.0000  787626     3.9078    28    29.3961       0     0.0000       0     0.0000        0     0.0000',
            'msftsrvcs                  0       0     0.0000   42293     0.0000     0     0.0000       0     0.0000       0     0.0000        0     0.0000'
        ])
        self.autorate_config_reader = StringIO.StringIO('\n'.join([
            'rate_label_kbyte_per_sec_limit sonycoment 1234 # squid1 pid 58202 limited net',
            'rate_label_kbyte_per_sec_reject msftsrvcs 2345 # squid1 pid 58202 limited net',
            'rate_label_kbyte_per_sec_reject_code msftsrvcs 429',
            'if-child 1 rate_label_kbyte_per_sec_limit sonycoment 4567 # squid1 pid 58203 limited net',
            'if-child 1 rate_label_kbyte_per_sec_reject msftsrvcs 5678 # squid1 pid 58203 limited net',
            'if-child 1 rate_label_kbyte_per_sec_reject_code msftsrvcs 429',
            'if-child 3 rate_label_kbyte_per_sec_limit sonycoment 6789 # squid1 pid 58203 limited net',
            ''
        ]))

    def test_mpep_follows_if_child(self):
        got = autorate.fetch_labels(self.cache_mgr_fetcher,
                self.autorate_config_reader, 1)
        expected = {
            'sonycoment': {
                'req/s': 3.9078,
                'numwait': 0.0,
                'conc': 28.0,
                'oldlimkbps': 4567.0,
                'oldreject': False,
                'limKB/s': 787626.0,
                'avgKB/s': 18012.0,
            },
            'msftsrvcs': {
                'req/s': 0.0,
                'numwait': 0.0,
                'conc': 0.0,
                'oldlimkbps': 5678.0,
                'oldreject': True,
                'oldrejectcode': '429',
                'limKB/s': 42293.0,
                'avgKB/s': 0.0
            }
        }
        self.assertEqual(got, expected)

    def test_mpep_ignores_non_matching_if_child(self):
        got = autorate.fetch_labels(self.cache_mgr_fetcher,
                self.autorate_config_reader, 2)
        expected = {
            'sonycoment': {
                'req/s': 3.9078,
                'numwait': 0.0,
                'conc': 28.0,
                'oldlimkbps': 0.0,
                'oldreject': False,
                'limKB/s': 787626.0,
                'avgKB/s': 18012.0,
            },
            'msftsrvcs': {
                'req/s': 0.0,
                'numwait': 0.0,
                'conc': 0.0,
                'oldlimkbps': 0.0,
                'oldreject': False,
                'limKB/s': 42293.0,
                'avgKB/s': 0.0
            }
        }
        self.assertEqual(got, expected)

    def test_non_mpep_follows_non_if_child(self):
        got = autorate.fetch_labels(self.cache_mgr_fetcher,
                self.autorate_config_reader, None)
        expected = {
            'sonycoment': {
                'req/s': 3.9078,
                'numwait': 0.0,
                'conc': 28.0,
                'oldlimkbps': 1234.0,
                'oldreject': False,
                'limKB/s': 787626.0,
                'avgKB/s': 18012.0,
            },
            'msftsrvcs': {
                'req/s': 0.0,
                'numwait': 0.0,
                'conc': 0.0,
                'oldlimkbps': 2345.0,
                'oldreject': True,
                'oldrejectcode': '429',
                'limKB/s': 42293.0,
                'avgKB/s': 0.0
            }
        }
        self.assertEqual(got, expected)

if __name__ == "__main__":
    # -e for extra debug
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument('-e', action="store_true")
    options, args = parser.parse_known_args()
    if options.e:
        autorate.log = logging.getLogger('testing')
        autorate.log.setLevel(logging.DEBUG)
        autorate.log.addHandler(logging.StreamHandler(sys.stdout))
    unittest.main(argv=sys.argv[:1] + args)
