#!/usr/bin/perl
#
# Attempts to profile a running process by executing pmcstat against it, collects
# the resulting samples, dependency libraries and binaries, packages them up and
# sends them off to a remote collection host for further processing or archiving.
#
# Copyright (c) 2015, Limelight Networks, Inc.
#

use POSIX qw(getcwd uname);
use Getopt::Std;
use File::Basename;
use File::Copy qw(cp);
use File::Temp;
use File::Path qw(mkpath);
use File::Find;
use Archive::Tar;
use Sys::Hostname;
use IO::Socket;

our($opt_h, $opt_p, $opt_s, $opt_d, $opt_a, $opt_u);

$Getopt::Std::STANDARD_HELP_VERSION = 1;

my $MIN_DURATION             = 1;
my $MAX_DURATION             = 60;
my $DEFAULT_DURATION         = 10;
my $DEFAULT_ARCHIVE_LOCATION = "pmc-collector.llnw.net:9876";
my $archivehost              = $DEFAULT_ARCHIVE_LOCATION;
my $version                  = "1.0";

# gets the startup timestamp of the specified pid
sub getprocesstime {
	my ($pid) = @_;
	my $file = sprintf("/proc/%d/status", $pid);
	my $buff = ''x4096;

	open(F, $file) || warn "can't read status file $file, $!\n";
	sysread(F, $buff, 4096);
	close(F);
	my @fields = split(/ /, $buff);
	my ($ts, undef) = split(/,/, $fields[7]);
	return $ts;
}

# gets the process name associated with a given pid
sub getprocessname {
	my ($pid) = @_;
	my $file = sprintf("/proc/%d/status", $pid);
	my $buff = ''x4096;

	open(F, $file) || warn "can't read status file $file, $!\n";
	sysread(F, $buff, 4096);
	close(F);
	my @fields = split(/ /, $buff);
	return $fields[0];
}

# locates a process by regex match / name and returns the listing of pids
sub findproc {
	my ($name) = @_;
	my @results;
	opendir(D, "/proc") || die "can't chdir to /proc, $!\n";
	my @pids = grep { /\d+/ } readdir(D);
	closedir(D);
	foreach my $pid (sort @pids) {
		my $file = sprintf("/proc/%d/cmdline", $pid);
		my $buff = ''x4096;
		open(F, $file) || next;
		sysread(F, $buff, 4096);
		my @fields = split(/\0/, $buff);
		close(F);
		my $cmdline = join(' ', @fields);
		next if ($cmdline !~ m/$name/); # skip anything that doesnt match
		next if ($cmdline =~ m/$0/);    # skip self
		push(@results, $pid);
		my $ts = getprocesstime($pid);
		printf("pid %d matches (timestamp: %d, cmdline: %s)\n", $pid, $ts, $cmdline);
	}
	return @results;
}

# enables the hwpmc module if not already enabled on the system
sub enablepmcstat {
	my $found = 0;
	open(PROC, "/sbin/kldstat |") || die "fatal: can't execute kldstat, $!\n";
	while (<PROC>) {
		#Id Refs Address            Size     Name
		# 6    1 0xffffffff81617000 6538     ftcp.ko
		# 7    1 0xffffffff8161e000 1a673    hwpmc.ko
		next if ( m/^Id/ );
		if ( m/hwpmc.ko/ ) { $found = 1; last; }
	}
	close(PROC);

	if ($found) {
		# don't do anything if the module is already loaded
		return;
	}

	system("/sbin/kldload hwpmc") == 0 || die "fatal: failed to load hwpmc module, exiting!\n";
}

# disables the hwpmc module on the system
sub disablepmcstat {
	system("/sbin/kldunload hwpmc") == 0 || warn "warning: failed to unload hwpmc module\n";
}

# copies all dependencies found within a given pmc sample mapfile.   This is done
# by first analyzing the mapfile for map-in entries which indicate what kernel
# module, libraries, and binaries were mapped in during the profiling attempt. These
# files are all then copied to the archive location to provide a snapshot of what
# the system actually had for binaries and libraries during the capture.
sub copydeps {
	my ($mapfile, $dropdir) = @_;

	printf("archiving dependencies...");
	open(MAP, $mapfile) || warn "can't read mapfile $mapfile, $!\n";
	while (<MAP>) {
		# if call-chain entries start, then no more map-in entries are expected
		last if ( m/^call-chain/ );

		# look for map-in entries that tell us what dependencies exist
		if ( m/^map-in\s+(\d+)\s+\S+\s+\"(\S+)\"/ ) {
			my $mode = $1;
			my $dependency = $2;
			my $normalized = $dependency;
			next if ( $mode == -1 );
			$normalized =~ s#^/##;
			$normalized = sprintf("%s/%s", $dropdir, $normalized);
			mkpath(dirname($normalized), 0, 0755);

			# copy the files to the destination so we can tar them up easily
			#printf("copying %s to %s...\n", $dependency, $normalized);
			cp($dependency, $normalized) 
			    || warn "failed to copy $dependency to $normalized, $!\n";
			printf(".");
		}
	}
	close(MAP);
	printf("done.\n");
}

# creates the tar archive of the profiling data, including the mapfile, binary and
# library dependencies.
sub createarchive {
	my ($dir, $pid) = @_;
	my $filename = sprintf("%s.tar.bz2", $dir);
	my $prefix = sprintf("%s.%d.%d.%s.profile", getprocessname($pid), $pid, time(), hostname);
	printf("creating archive %s...", $filename);
	my $cwd = getcwd();
	chdir($dir);
	my $arc = Archive::Tar->new();
	my @files = ();
	find( sub { push(@files, $File::Find::name) }, ".");
	$arc->add_files(@files);
	$arc->write($filename, COMPRESS_BZIP, $prefix);
	printf("done.\n");
	chdir($cwd);
	return $filename;
}

# generates the mapfile of symbols and dependencies from the collected
# pmcstat sample.  This is to identify the dependencies primarily.
sub generatemap {
	my ($samplefile, $mapfile) = @_;

	# stupid pmcstat -M always outputs to stdout
	open(MAP, ">$mapfile");
	printf("generating mapfile from pmcstat sample...");
	open(PMC, "/usr/sbin/pmcstat -R $samplefile -M - |") || 
	     warn "can't get mapfile information, ignoring\n";
	while (<PMC>) {
		print MAP $_;
	}
	close(PMC) || 
	     warn "execution of pmcstat failed, $!\n";
	printf("done.\n");
	close(MAP);
}

# uploads the given archive/tarball to the desitnation collection host.
# It expects to just straight dump the entire contents to the remote
# host.  At this time, no handshakes or other data should be sent to the
# collection host.
sub uploadarchive {
	my ($archive) = @_;
	my $progress = 0;

	# send the archive to the remote collection host
	printf("sending archive to collection host (%s)...", $archivehost);
	my $socket = IO::Socket::INET->new(PeerAddr => $archivehost, Proto => 'tcp', Timeout => '20');
	if (not defined $socket) {
		warn "failed to connect to $archivehost, $!\n";
		return 0;
	}
	open(ARCHIVE, $archive) || warn "failed to open $archive, $!\n";
	while (1) {
		my $buff   = '' x 4096;
		my $nbytes = sysread(ARCHIVE, $buff, 4096);
		if ($nbytes <= 0) {
			last;
		}
		if ($progress++ % 57 == 0) {
			printf(".");
		}
		syswrite($socket, $buff, $nbytes);
	}
	close(ARCHIVE);
	$socket->close;
	printf("done.\n");
	return 1;
}

# executes pmcstat on the specified process id for the number of specfiied
# seconds.  The process is execd and then waited on till completion.  Upon
# completion, the depdendencies are identified from the mapfile and then
# copied to the archive for reference.  After completion, the archive is
# uploaded to the collection host, and on success will cleanup the files.
sub runpmcstat {
	my ($procpid, $duration) = @_;
	my $dir = File::Temp->newdir("/tmp/pmcstatXXXXXX");
	my $samplefile = sprintf("%s/sample.pmc", $dir);
	my $mapfile = sprintf("%s/mapfile", $dir);
	my $pid = fork();
	if ($pid == 0) {
		# child pid
		exec("/usr/sbin/pmcstat",
		     "-P", "unhalted-cycles",
		     "-O", $samplefile,
		     "-l", $duration,
		     "-t", $procpid);
		exit(0);
	}

	$| = 1;
	printf("waiting for pmcstat command to complete...");
	while (waitpid($pid, WNOHANG) > 0) {
		printf(".");
		sleep(1);
	}
	printf("done.\n");
	generatemap($samplefile, $mapfile);
	copydeps($mapfile, $dir);
	my $archive = createarchive($dir, $procpid);
	if ( !uploadarchive($archive) ) {
		warn "failed to upload $archive, $! (leaving local copy intact)\n";
		return;
	}
	printf("cleaning up...");
	if ( -f $archive ) {
		# cleanup
		unlink($archive);
	}
	printf("done.\n");
}

# checks to make sure we are running on a valid operating system
sub validateos {
	my ($sysname, undef, $release, undef, undef) = uname();
	if ($sysname ne "FreeBSD") {
		printf("Sorry, but this script is only valid on FreeBSD.\n");
		exit(-1);
	}
	if ($release =~ m/^(\d+)\.(\d+)/) {
		my $major = $1; my $minor = $2;
		if ($major >= 10) {
			return;
		}
	}
	printf("Sorry, but this script can only run on FreeBSD 10.x or above.\n");
	exit(-1);
}

# prints the usage and exits
sub usage {
	printf("Usage: %s [-h] [-p pid | -s regex] [-d secs] [-u] [-a host:port]\n", $0);
	printf("Profiles a running process and ships off the samples to an archive host.\n");
	printf("\n");
	printf("  -a host:port    set archive host (default: %s)\n", $archivehost);
	printf("  -d secs         duration in seconds to run the profile (default: %d secs)\n", $DEFAULT_DURATION);
	printf("  -p pid          profile the specified pid\n");
	printf("  -s regex        profile a running process matching the regex specified\n");
	printf("  -u              unload kernel module after completion\n");
	printf("  -h, --help      display this help and exit\n");
	printf("      --version   output version information and exit\n");
	exit(0);
}

sub HELP_MESSAGE {
	usage();
}

sub VERSION_MESSAGE {
	printf("%s version %s\n", $0, $version);
	printf("Copyright (c) 2015, Limelight Networks, Incorporated\n");
	printf("\n");
}

sub main {
	my $duration = $DEFAULT_DURATION;
	my $pid = 0;

	getopts("hp:s:d:a:u") || usage();
	usage() if ($opt_h);

	validateos();

	if ($opt_d) {
		$duration = int($opt_d);
		if ($duration < $MIN_DURATION || $duration > $MAX_DURATION) {
			$duration = $DEFAULT_DURATION;
		}
	}

	if ($opt_a) {
		$archivehost = $opt_a;
		usage() if ( $archivehost !~ m/\S+:\d+/ );
	}

	# only one of string match or pid can be specified, not both
	if ($opt_s) {
		my @pids = findproc($opt_s);
		if ($#pids > 0) {
			printf("too many matches, please select one from above, and rerun with -p <pid>\n");
			exit(-1);
		} elsif ($#pids == -1) {
			printf("no matching process found\n");
			exit(-1);
		}
		$pid = $pids[0];
	} elsif ($opt_p) {
		$pid = int($opt_p);
		if ( ! -f "/proc/$pid/status" ) {
			printf("no matching process found\n");
			exit(-1);
		}
	} else {
		# if -s or -p arent specified, we're missing key information
		usage();
	}

	# run pmcstat on the supplied pid
	if ( $pid > 0 ) {
		enablepmcstat();
		runpmcstat($pid, $duration);
		disablepmcstat() if ($opt_u);
	}
}

&main;

1;
