#!/usr/bin/env perl
use strict;
use warnings;
use JSON;
use Time::HiRes;
use Sys::Hostname;
use Getopt::Std;
use IO::Socket;
use Socket qw(:crlf);
use POSIX 'strftime';

# We need to save the last line number read for next time
my $ics_instance = 0;
my $apache_log_dir = "/logs/apache24/";
my $counter_state_file;

my $counter_state_dir = "/var/tmp/";

# Hashes used in reading and writing persistant storage
my $usage_counters_old = {};
my $usage_counters_new = {};


my $fqdn = hostname();
my $pop = extract_pop(); 
my $debug = 0;
my $prefix = 'app.ics';

my $tsdb_server = "tsdb-data.lldns.net";
my $tsdb_port = "4242";
my $tsdb_sock;


#set default target values at which the metric is considered full or in trouble
my $cpu_target = 80; 
my $disk_target = 40;  #adjusted below based on type of disk hdd 7200 ssd etc
my $network_target = 65; #


######################################################################################
# Return the pop from the hostname
######################################################################################
sub extract_pop {
  my $POP = 'unknown';
  ($POP) = (hostname() =~ /^\w+\.(\w+)\.llnw\.net/);
  chomp($POP);
  return lc($POP);
}


######################################################################################
# Return the type of disk assemblys found on this host
######################################################################################
sub get_disk_type {
    my $cmd = '/sbin/geom disk list';

    my %vote;
    open(my $f, "-|", $cmd) or die "failed to execute $cmd: $!";
    while (<$f>) {
        if (/rotationrate:\s+(\d+)/) {
            my $rate = $1;
            if ($rate eq 0) {
                $rate = 100;#"ssd";
            } elsif ($rate < 8000) {
                $rate = 7200;
            } elsif ($rate < 12500) {
                $rate = 10000;
            } else {
                $rate = 15000;
            }
            $vote{$rate}++;
        }
    }
    close($f);
    my $winner = "none";
    my $max_count = 0;
    for my $rate (sort { $a <=> $b } keys(%vote)) {
        my $count = $vote{$rate};
        if ($count > $max_count) {
            $winner = $rate;
            $max_count = $count;
        }
    }
    $winner = "ssd" if ($winner == 100);
    return $winner;
}


######################################################################################
# Returns the number of access request since the last run
######################################################################################
sub get_apache_accesses
{
    my $old = shift;
    my $new = shift;

    return 0 unless validate_positive_key($old->{'apache'},'TotalAccesses' );
    return 0 unless validate_positive_key($new->{'apache'},'TotalAccesses' );

    return($new->{'apache'}{'TotalAccesses'} - $old->{'apache'}{'TotalAccesses'});
}


######################################################################################
# Return the number of byte sent by apache since the last run
######################################################################################
sub get_apache_kbytes
{
    my $old = shift;
    my $new = shift;

    return 0 unless validate_positive_key($old->{'apache'}, 'TotalkBytes');
    return 0 unless validate_positive_key($new->{'apache'}, 'TotalkBytes');

    return($new->{'apache'}{'TotalkBytes'} - $old->{'apache'}{'TotalkBytes'});

}


######################################################################################
# Returns the number of cpu clocks the processor was busy and a percentage since last run
######################################################################################
sub get_cpu_usage {
    my $old = shift;
    my $new = shift;

    my @cp_time;
    my @cp_old;
    my %cpu_states;
    my $CPUSTATES = 5; #user nice sys interrupt idle

    return 0 unless validate_positive_key($old, 'cpu');
    return 0 unless validate_positive_key($new, 'cpu');

    #local copy of old values
    $cp_old[0] = $old->{'cpu'}{'user'};
    $cp_old[1] = $old->{'cpu'}{'nice'};
    $cp_old[2] = $old->{'cpu'}{'sys'};
    $cp_old[3] = $old->{'cpu'}{'interrupt'};
    $cp_old[4] = $old->{'cpu'}{'idle'};
    #local copy of new
    $cp_time[0] = $new->{'cpu'}{'user'};
    $cp_time[1] = $new->{'cpu'}{'nice'};
    $cp_time[2] = $new->{'cpu'}{'sys'};
    $cp_time[3] = $new->{'cpu'}{'interrupt'};
    $cp_time[4] = $new->{'cpu'}{'idle'};
    
    # convert cp_time counts to percentages 
    percentages($CPUSTATES, \%cpu_states, \@cp_time, \@cp_old );

    my $pct = (10000 - int($cpu_states{'idle'}*100) )/100;
    return 0 if $pct < 0 || $pct > 100;
    return $pct;
}

######################################################################################
# Returns network though put as a percentage of the max since last called
######################################################################################
sub get_network_usage {
    my $old = shift;
    my $new = shift;

    return 0 unless defined $old->{'net'};
    return 0 unless defined $new->{'net'};
    return 0 unless validate_positive_key($old, 'time');

    my $max_pct = 0;
    while (my ($iface, $dat) = each %{$new->{'net'}}) {
        next unless validate_positive_key($dat, 'speed');
        next unless validate_positive_key($dat, 'bytes');
        next unless validate_positive_key($old->{'net'}{$iface}, 'speed');
        next unless validate_positive_key($old->{'net'}{$iface}, 'bytes');
        my $old_dat = $old->{'net'}{$iface};

        my $dur = $new->{'time'} - $old->{'time'};
        return 0 if $dur <= 0 || $dur > 600;
        my $max_bytes = $dur * $dat->{'speed'};
        my $util = $dat->{'bytes'} - $old_dat->{'bytes'};
        my $pct = 100 * $util / $max_bytes;
        next if $pct > 102;
        $max_pct = $pct if $pct > $max_pct;
    }
    return $max_pct;
}

######################################################################################
#
######################################################################################
sub get_disk_usage {
    my $old = shift;
    my $new = shift;

    return 0 unless validate_positive_key($old, 'disk');
    return 0 unless validate_positive_key($new, 'disk');
    return 0 unless validate_positive_key($old, 'time');

    my $dur = $new->{'time'} - $old->{'time'};
    return 0 if $dur <= 0 || $dur > 600;
    my $util = ($new->{'disk'} - $old->{'disk'});
    my $pct = 100 * $util / $dur;
    return 0 if $pct < 0 || $pct > 100;
    return $pct;
}


######################################################################################
#
######################################################################################
sub validate_positive_key {
    my $h = shift;
    my $k = shift;

    return 0 unless defined $h->{$k};
    return 0 if $h->{$k} <= 0;
    return 1;
}

######################################################################################
#
######################################################################################
sub get_net_counters {
    my %net;
    my $ifaces_str = `/sbin/ifconfig -l`;
    chomp $ifaces_str;
    my @ifaces = split /\s+/, $ifaces_str;

    for my $iface (@ifaces) {
        my $cmd = "/sbin/ifconfig $iface";
        open(my $ifconfig, '-|', $cmd) or die "failed to run $cmd: $!";
        my $speed = 0;
        while (<$ifconfig>) {
            if (/media:.*?(\d+)(\S*base)/i) {
                my $speed_val = $1;
                my $speed_label = $2;
                if ($speed_label =~ /^base/i) {
                    $speed = $speed_val * 1e6 / 8;
                } elsif ($speed_label =~ /^gbase/i) {
                    $speed = $speed_val * 1e9 / 8;
                # terabit, we can dream
                } elsif ($speed_label =~ /^tbase/i) {
                    $speed = $speed_val * 1e12 / 8;
                }
            }
        }
        close($ifconfig) or die "error while closing $cmd: $!";
        next if $speed eq 0;
        $cmd = "/usr/bin/netstat -i -b -I $iface";
        open($ifconfig, '-|', $cmd) or die "failed to run $cmd: $!";
        my $bytes = 0;
        while (<$ifconfig>) {
            my @words = split /\s+/, $_;
            if ($words[2] =~ /Link/) {
                $bytes = $words[10];
            }
        }
        close($ifconfig) or die "error while closing $cmd: $!";
        next if $bytes <= 0;
        $net{$iface}{'speed'} = $speed;
        $net{$iface}{'bytes'} = 0 + $bytes;
    }
    return \%net;
}


######################################################################################
# grab raw disk busy from iostat if available
######################################################################################
sub run_iostat {
    my @iostat = `/usr/sbin/iostat -x -I`;
    my $field_sb = -1;
    my $field_device = -1;
    my $count = 0;
    my $total = 0;
    for my $line (@iostat) {
        my @words = split /\s+/, $line;
        if ($field_sb == -1 || $field_device == -1) {
            # determine field numbers for device and seconds busy
            for (my $i=0; $i < @words; $i++) {
                $field_device = $i if $words[$i] eq "device";
                $field_sb = $i if $words[$i] eq "sb/i";
            }
        } elsif (defined $words[$field_device] && defined $words[$field_sb]) {
            # if we have valid field numbers and enough words, extract the data
            my $disk = $words[$field_device];
            my $sb = $words[$field_sb];
            next unless $disk =~ /^a?da\d+$/;
            $count++;
            $total += $sb;
        }
    }
    if ($count > 0) {
        my $avg = sprintf "%.3lf", $total/$count;
        return 0 + $avg;
    }
    return 0;
}


######################################################################################
#  Gets the cpu counters and stash them in our hash
######################################################################################
sub get_cpu
{
    # Vars used to monitor cpu usage
    my @cp_time;
    my %cpu_states;

    #Read the cpu values
    get_cpu_stats(\@cp_time);
    $cpu_states{'user'} = $cp_time[0];
    $cpu_states{'nice'} = $cp_time[1];
    $cpu_states{'sys'} = $cp_time[2];
    $cpu_states{'interrupt'} = $cp_time[3];
    $cpu_states{'idle'} = $cp_time[4];
    return(\%cpu_states);

}

######################################################################################
# get the CPU counters from the kernel
######################################################################################
sub get_cpu_stats
{
    my $cp_time = shift;
    #sysctl kern.cp_time
    #kern.cp_time: 155271910 161335 171166393 29105139 10695726059
    my $sys_cmd = "/sbin/sysctl kern.cp_time";
    open(my $ps_f, "$sys_cmd 2>&1|") or die "cannot execute $sys_cmd: $!";
    while (<$ps_f>) {
        next unless /^kern.cp_time: (\d+) (\d+) (\d+) (\d+) (\d+)/;
        #user nice sys interrupt idle
        @{$cp_time}[0] = $1;  
        @{$cp_time}[1] = $2;  
        @{$cp_time}[2] = $3;  
        @{$cp_time}[3] = $4;  
        @{$cp_time}[4] = $5;  
    }
    close $ps_f;
}

######################################################################################
#   percentages(cnt, out, new, old) - calculate percentage change
#   between array "old" and "new", putting the percentages in "out".
#   "cnt" is size of each array.
#   The array "old" is updated on each call.
#   The routine assumes modulo arithmetic.  This function is especially
#   useful on BSD mchines for calculating cpu state percentages.
######################################################################################
sub percentages 
{
    my $cnt = shift;
    my $out = shift; 
    my $new = shift; 
    my $old = shift; 

    #Local vars
    my $i;
    my $change;
    my $total_change;
    my $half_total;
    my @diffs;

    # initialization 
    $total_change = 0;

    # calculate changes for each state and the overall change 
    for ($i = 0; $i < $cnt; $i++)
    {
        if (($change = ${$new}[$i] - ${$old}[$i]) < 0)
        {
            # this only happens when the counter wraps 
            $change = ${$new}[$i] - ${$old}[$i];
        }       
        $total_change += ($diffs[$i] = $change);        
    }
    # avoid divide by zero potential 
    if ($total_change == 0)
    {
        $total_change = 1;
    }
    # calculate percentages based on overall change, rounding up 
    $half_total = $total_change / 2;

    # Do not divide by 0. Causes Floating point exception 
    if($total_change) {
        ${$out}{user} = (($diffs[0] * 100 + $half_total) / $total_change);
        ${$out}{nice} = (($diffs[1] * 100 + $half_total) / $total_change);
        ${$out}{sys}  = (($diffs[2] * 100 + $half_total) / $total_change);
        ${$out}{interrupt} = (($diffs[3] * 100 + $half_total) / $total_change);
        ${$out}{idle} = (($diffs[4] * 100 + $half_total) / $total_change);
    }
    #return the total in case the caller wants to use it 
    return($total_change);
}

######################################################################################
#
######################################################################################
sub get_log_counters
{
    my %log_resultcnt;     #where the counts go
    my $log_start = 0;     # read from persistant storage  #2068439 - 1000;
    my $log_filesize = 0;  # read from persistant storage
    $log_start = $usage_counters_old->{'apachelog'}{'line'} if($usage_counters_old->{'apachelog'}{'line'});     
    $log_filesize = $usage_counters_old->{'apachelog'}{'size'} if($usage_counters_old->{'apachelog'}{'size'});
    # get the counts for the main host
    ($log_filesize,$log_start) = fetch_counts($ics_instance,$log_filesize,$log_start,\%log_resultcnt); 
    $log_resultcnt{'size'} = $log_filesize;
    $log_resultcnt{'line'} = $log_start;    
    return(\%log_resultcnt);
}


######################################################################################
#    get_result_counts(log,line_number,hashref)
#
#   Reads though the apache log file and starts counting result codes at
#   line_number storing the results in hashref. returns the line number
#   of the last line of the file proccesed. By passing the last line 
#   back into this function we can skip the lines already proccessed and 
#   fetch the status codes since the last time called.
######################################################################################
sub get_result_counts
{
    my $log = shift;   # The full path to the log to be examined
    my $ignor_lines = shift;   # Number of lines to skip
    my $result_counts = shift; # Hash ref to return the results
    my $line_count=0;          # Line counter
    my $linesproced=0;         #limit to 5000 lines at a time for testing
    open FILE1, $log or die;
    while (my $line=<FILE1>) 
    {
        $linesproced++;
        $line_count++;
        if($line_count >= $ignor_lines)
        {
            if($line =~ m/(\d\d\d) \S+ ".*" ".*"$/ )
            {
                if($1 == 400 || $1 == 404 || $1 == 500)
                {
                    $result_counts->{$1}++;
                    if($1 eq '')
                    {
                       print $line . "\n";
                    }
                }elsif($1 != 200)
                {
                    $result_counts->{other}++;
                }
            }
        }
    }
    close FILE1;
    return($line_count);
}


######################################################################################
##  gen_logname(yesterday,jail)
#    create a file name using the current date or yesterdays date if
#    yesterday is not 0 for the ics apache log on the host if jail 
#    is 0 or for the jailed instance is jail is not 0
#    The file names returned will looks something like this
# mfs-access-2018-01-08.log         yesterdays log
# mfs-ics10-access-2018-01-08.log   yesterdays jailed log
# mfs-access-2018-01-09.log         todays log
# mfs-ics10-access-2018-01-09.log   todays jailed log
######################################################################################
sub gen_logname
{
    my $yesterday = shift;  #0 = today anyting else = yesterday  
    my $jail = shift;       #0 = host anything else = jailed instance
    my $logname = "";    # place to build the name of the file
    #Figuer out the date for which logs to pull
    my $then = time;
    $then = time - 24 * 60 * 60 if($yesterday); #subtract a day to get yesterdays logs
    my $dt = strftime ('%Y-%m-%d', localtime $then);#DateTime->now(time_zone  => 'local');# Stores current date and time as datetime object
    my $dom = 0;
    my $mon = 0;
    my $yr = 0;
    
    if( $dt =~ m/\d\d\d\d-\d\d-(\d\d)/) {$dom = $1;}
    if( $dt =~ m/\d\d\d\d-(\d\d)-\d\d/) {$mon = $1;}
    if( $dt =~ m/(\d\d\d\d)-\d\d-\d\d/) {$yr = $1;}

    if($jail)
    {
        $logname = $apache_log_dir."mfs-ics10-access-" . $yr . "-" . $mon . "-" . $dom . ".log";
    }
    else
    {
        $logname = $apache_log_dir."mfs-access-" . $yr . "-" . $mon . "-" . $dom . ".log";
    }
    return($logname);
}



######################################################################################
## fetch_counts(jailed,last_size,result_ref)
#    Handles log rollovers and non-existant files
#     last_size is the size of the file the last time we checked
#     if jailed != 0 then fetch the counts for the main host
#     else fetch the counts for the jailed instance
#     Return the file size and last line read for use the 
#     next time the check the file
######################################################################################
sub fetch_counts
{
    my $jailed = shift;
    my $last_size = shift;
    my $startline = shift;
    my $result_ref = shift;
    my $size = 0;
    #ensures the hash fields exist and are empty
    $result_ref->{400} = 0;
    $result_ref->{404} = 0;
    $result_ref->{500} = 0;
    $result_ref->{other} = 0;
    #get the log file name
    my $logfile = gen_logname(0,$jailed);
    #if the file exist
    if (-e $logfile) {
        # use the file size to see if we rolled over
        # there is a big assumtion that the file size will
        # be smaller if we rolled over
        $size = -s $logfile;
        if($size < $last_size)
        {
            if ($debug) {
                print "The apache log file $logfile has rolled over\n";
            }
            # read the last lines from the yesterdays log
            my $oldlog = gen_logname(1,$jailed);
            #if yesterdays log exist try to process the last lines
            if (-e $oldlog) {
                get_result_counts($oldlog,$startline,$result_ref);
            }
            else
            {
                if ($debug) {
                    print "Yesterdays Apache log file $logfile does not exist\n";
                }
            }
            $startline = 0; # we will be starting a new file so reset
        }
        #read the lines in the log and get the result counts and capture
        #the new startline
        $startline = get_result_counts($logfile,$startline,$result_ref);
    }
    else #the logfile does not exist
    {
        if ($debug) {
            print "The apache log file $logfile does not exist\n";
        }
        $startline = 0;
        $size = 0;
    }
    return($size,$startline);
}


######################################################################################
# read the apache status page and scrape the counts from the page 
######################################################################################
sub get_apache_counters
{
    my %counters;
    $counters{'BytesPerSec'} = 0;
    $counters{'BytesPerReq'} = 0;
    $counters{'ReqPerSec'} = 0;
    $counters{'ConnsTotal'} = 0;
    $counters{'BusyWorkers'} = 0;
    $counters{'IdleWorkers'} = 0;
    $counters{'CPUUser'} = 0;
    $counters{'CPULoad'} = 0;
    $counters{'CPUSystem'} = 0;
    my $loop = $ics_instance+1;
    my $curl_cmd = "/usr/local/bin/curl -s -interface 127.0.0.1 'http://127.0.0." . $loop . "/server-status?auto' ";
    open(my $ps_f, "$curl_cmd 2>&1|") or die "cannot execute $curl_cmd: $!";
    while (<$ps_f>) {       
        if (/^Load1: (.+)/)
        {
            $counters{'Load1'} = $1;
        }
        if (/^Load5: (.+)/)
        {
            $counters{'Load5'} = $1;
        }
        if (/^Load15: (.+)/)
        {
            $counters{'Load15'} = $1;
        }
        if (/^Total Accesses: (.+)/)
        {
            $counters{'TotalAccesses'} = $1;
        }
        if (/^Total kBytes: (.+)/)
        {
            $counters{'TotalkBytes'} = $1;
        }
        if (/^CPUUser: (.+)/)
        {
            $counters{'CPUUser'} = $1;
        }
        if (/^CPUSystem: (.+)/)
        {
            $counters{'CPUSystem'} = $1;
        }
        if (/^CPUChildrenUser: (.+)/)
        {
            $counters{'CPUChildrenUser'} = $1;
        }
        if (/^CPUChildrenSystem: (.+)/)
        {
            $counters{'CPUChildrenSystem'} = $1;
        }
        if (/^CPULoad: (.+)/)
        {
            $counters{'CPULoad'} = $1;
        }
        if (/^Uptime: (.+)/)
        {
            $counters{'Uptime'} = $1;
        }
        if (/^ReqPerSec: (.+)/)
        {
            $counters{'ReqPerSec'} = $1;
        }
        if (/^BytesPerSec: (.+)/)
        {
            $counters{'BytesPerSec'} = $1;
        }
        if (/^BytesPerReq: (.+)/)
        {
            $counters{'BytesPerReq'} = $1;
        }
        if (/^BusyWorkers: (.+)/)
        {
            $counters{'BusyWorkers'} = $1;
        }
        if (/^IdleWorkers: (.+)/)
        {
            $counters{'IdleWorkers'} = $1;
        }
        if (/^ConnsTotal: (.+)/)
        {
            $counters{'ConnsTotal'} = $1;
        }
        if (/^ConnsAsyncWriting: (.+)/)
        {
            $counters{'ConnsAsyncWriting'} = $1;
        }
        if (/^ConnsAsyncKeepAlive: (.+)/)
        {
            $counters{'ConnsAsyncKeepAlive'} = $1;
        }
        if (/^ConnsAsyncClosing: (.+)/)     
        {
            $counters{'ConnsAsyncClosing'} = $1;
        }       
    }
    close $ps_f;
    return \%counters;
}


######################################################################################
######################################################################################
#
#       The Main Event
#
######################################################################################
######################################################################################
my %options = ();
getopts('di:', \%options);
$debug = 1 if defined $options{'d'};
$ics_instance = $options{'i'} if defined $options{'i'};

# Set som vars
$counter_state_file = $counter_state_dir . "fetchics.$ics_instance.json";

######################################################################################
# Retrive persistent values from storage
######################################################################################
# if the file exists, complain about errors on open
# if not, $usage_counters_old will just be an empty hash and the rest of the code will
# handle it
if (-e $counter_state_file) {
    local $/;
    open(my $fh, '<', $counter_state_file) or die "Can't open $counter_state_file: $!";
    my $json_text = <$fh>;
    $usage_counters_old = {};
    $usage_counters_old = eval {
        decode_json($json_text);
    };
    close($fh);
} 


######################################################################################
# Obtain the new counters/metrics
######################################################################################
$usage_counters_new->{'time'} = Time::HiRes::time;
$usage_counters_new->{'cpu'} = get_cpu();
$usage_counters_new->{'disk'} = run_iostat();
$usage_counters_new->{'net'} = get_net_counters();
$usage_counters_new->{'apachelog'} = get_log_counters();
$usage_counters_new->{'apache'} = get_apache_counters();

my $cpu_usage = get_cpu_usage($usage_counters_old, $usage_counters_new);
my $disk_usage = get_disk_usage($usage_counters_old, $usage_counters_new);
my $network_usage = get_network_usage($usage_counters_old, $usage_counters_new);

my $apache_accesses = get_apache_accesses($usage_counters_old, $usage_counters_new);
my $apache_kbytes = get_apache_kbytes($usage_counters_old, $usage_counters_new);



#############################  Calculations and scaling  ####################

#Gather some basic system information
my $disk_type = get_disk_type();

my $cpu = `sysctl hw.model`;
chomp $cpu;

my $dmi_name = `/usr/local/bin/sudo /usr/local/sbin/dmidecode -s system-product-name | head -1`;
chomp $dmi_name;

if ($dmi_name =~ /^X8/) {
    if ($disk_type == 7200) {
        $disk_target = 25;
    }
    if ($dmi_name eq "X8DTL") {
        $disk_target = 20;
    }
} elsif ($dmi_name =~ /^(X9|SYS-2UDP-|SYS-2027TR-)/) {
    if ($disk_type eq "ssd") {
        if ($cpu =~ /E5-2640 0/) {
            $dmi_name = "X9v1";
        } else {
            $dmi_name = "X9v2";
        }
    } else {
        if ($disk_type == 7200) {
            $disk_target = 25;
        }
        $dmi_name = "X9";
    }
} elsif ($dmi_name =~ /^SYS-1018R-WC0R/) {
    $dmi_name = "X10";
    $dmi_name .= "v2" if $cpu =~ /v4/;
} elsif ($dmi_name =~ /R430/) {
    $dmi_name = "R430";
    $dmi_name .= "v2" if $cpu =~ /v4/;
}

my $capacity_class = "$dmi_name $disk_type";
$capacity_class =~ s/\s+/_/g;

#Scale cpu and disk
#$cpu_usage = 100 * $cpu_usage;
#$disk_usage = 100 * $disk_usage;

#calculate health values
my $cpu_health = 100 * $cpu_usage / $cpu_target;
my $disk_health = 100 * $disk_usage / $disk_target;
my $network_health = 100 * $network_usage / $network_target;




my $combined_usage;
if ($network_health > $cpu_health) {
    $combined_usage = $network_health;
} else {
    $combined_usage = $cpu_health;
}
if ($disk_health > $combined_usage) {
    $combined_usage = $disk_health;
} 


###############################  Output  #############################

######################################################################################
# Send the key,value and tag to tsdb
######################################################################################
sub send_metric {
    my $key = shift;
    my $value = shift;
    my $addtag = shift // "";
    my $time = time();
    if ($debug) {
        print STDOUT "* $key $time $value fqdn=$fqdn instance=$ics_instance capacity_class=$capacity_class pop=$pop $addtag\n";
    } else {
        print $tsdb_sock "put $key $time $value fqdn=$fqdn instance=$ics_instance capacity_class=$capacity_class pop=$pop $addtag", CRLF;
    }
}

######################################################################################
# Save the persistent values to storage
######################################################################################
open(my $fh, '>', $counter_state_file) or die "Can't open $counter_state_file for writing: $!";
my $json_out = encode_json $usage_counters_new;
print $fh $json_out or die "cannot write to $counter_state_file: $!";
close($fh) or die "error closing $counter_state_file: $!";

# tsdb feed
$tsdb_sock = IO::Socket::INET->new(
    PeerHost => $tsdb_server,
    PeerPort => $tsdb_port,
    Proto    => 'tcp',
) or die "Failed to open tsdb socket: $@";


######################################################################################
#
#   Send system type info to tsdb
#
######################################################################################
send_metric("$prefix.sys.cpu_usage", $cpu_usage);
send_metric("$prefix.sys.network_usage", $network_usage);
send_metric("$prefix.sys.disk_usage", $disk_usage);
send_metric("$prefix.sys.cpu_health", $cpu_health);
send_metric("$prefix.sys.network_health", $network_health);
send_metric("$prefix.sys.disk_health", $disk_health);
send_metric("$prefix.sys.server_health", $combined_usage);

######################################################################################
#
#   Send the result codes from the apache logs to tsdb
#
######################################################################################
send_metric("$prefix.apachelog.http_400", $usage_counters_new->{'apachelog'}{'400'});
send_metric("$prefix.apachelog.http_404", $usage_counters_new->{'apachelog'}{'404'});
send_metric("$prefix.apachelog.http_500", $usage_counters_new->{'apachelog'}{'500'});
send_metric("$prefix.apachelog.http_other", $usage_counters_new->{'apachelog'}{'other'});

######################################################################################
#
#   Send the values scraped from the apache status page to tsdb
#
######################################################################################
send_metric("$prefix.apache.TotalAccesses", $apache_accesses);
send_metric("$prefix.apache.TotalkBytes", $apache_kbytes);

send_metric("$prefix.apache.MbytesPerSec", int($usage_counters_new->{'apache'}{'BytesPerSec'}/1048576));
send_metric("$prefix.apache.KbytesPerReq", int($usage_counters_new->{'apache'}{'BytesPerReq'}/1024));
send_metric("$prefix.apache.ReqPerSec", $usage_counters_new->{'apache'}{'ReqPerSec'});

send_metric("$prefix.apache.ConnsTotal", $usage_counters_new->{'apache'}{'ConnsTotal'});
send_metric("$prefix.apache.BusyWorkers", $usage_counters_new->{'apache'}{'BusyWorkers'});
send_metric("$prefix.apache.IdleWorkers", $usage_counters_new->{'apache'}{'IdleWorkers'});

############################
send_metric("$prefix.apache.CPUUser", $usage_counters_new->{'apache'}{'CPUUser'});
send_metric("$prefix.apache.CPULoad", $usage_counters_new->{'apache'}{'CPULoad'});
send_metric("$prefix.apache.CPUSystem", $usage_counters_new->{'apache'}{'CPUSystem'});

#explicitly close the tsdb socket gracefully
close($tsdb_sock);
exit(0);

