#!/bin/sh
PATH='/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin'
HOSTNAME=`hostname | cut -d "." -f 1`
RHOSTNAME=`hostname`
RSYNC=`which rsync`
LOGGER=`which logger`
egrep=`which egrep`
df=`which df`
host=`which host`
mail=`which mail`;
date=`which date`;
wget=`which wget`
header='healthcheck-noexist.llnw.net'
file='healthcheck.txt'

DATE=`date`;

# if this is a shared origin we have to manipulate the filename more
echo $RHOSTNAME | $egrep "origin.+\."
RCODE=$?
if [ "$RCODE" = "0" ]; then
  # we're a shared origin
  HOSTNAME=`echo $HOSTNAME | awk '{print substr($0, 2, 10) ".llnw.net"}'`
else
  # we're a dedicated!
  HOSTNAME="`hostname | sed 's/\.bocs\./\.ocs\./'`"
fi

$LOGGER -p daemon.info "newsync.sh starting run at $DATE"

# Are we on backup and someone's just retarded?
ORIGIN=`$host $HOSTNAME`
echo $ORIGIN | $egrep "borigin|bocs"
RCODE=$?
if [ "$RCODE" = "0" ]; then 
  $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME failed because $HOSTNAME is pointed to the backup - exiting"
  echo "Someone tried to start newsync on $RHOSTNAME while it's acting as primary" | $mail -s "RSYNC ERROR" opseng@llnw.com
  exit 253
fi

# Are we on backup and want SYNCs not enabled
if [ -e "/etc/NOSYNC" ]; then
  # disable sync
  $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME has sync disabled due to /etc/NOSYNC file"
  exit 254
fi

if [ -e "/var/run/proftpd.pid" ]; then
  $LOGGER -p daemon.error -t ORIGIN_SYNC "$RHOSTNAME has a proftpd.pd file, assuming proftpd is running - exiting"
  echo "$RHOSTNAME has proftpd.pid, is proftpd running? exiting newsync" | $mail -s "RSYNC ERROR" opseng@llnw.com
  exit 252
fi

# see if we're already running
#NEWSYNCPID=`pgrep -f newsync`
NEWSYNCPID=`pgrep -f "/bin/sh /usr/local/bin/newsync.sh"`
if [ "X$NEWSYNCPID" != "X" ]; then
  if [ "$NEWSYNCPID" != "$$" ]; then
    # we're already running, make a note and exit
    $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME already running newsync.sh with pid $NEWSYNCPID - exiting"
    exit 0
  fi
fi

# make sure /content is mounted
mounted=`$df | $egrep '/content$'`
if [ "x$mounted" = "x" ]; then
  $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME failed sync because /content is not mounted"
  exit 255
fi

# First make sure that primary sees /content mounted, then check to make sure we are able to read and write from the /content filesystem as a health check 
MOUNTTEST=`ssh -q $HOSTNAME mount |grep content |grep ^\/dev |awk '{print $1}'`
if [ -n "$MOUNTTEST" ]
then
  TOUCHTEST=`ssh -q $HOSTNAME 'touch /content/hc/healthcheck/content/touchtest && stat /content/hc/healthcheck/content/touchtest && rm /content/hc/healthcheck/content/touchtest; echo $?'`
  if [ "$TOUCHTEST" -ne 0 ]
  then
    $LOGGER -p daemon.err -t ORIGIN_SYNC "$HOSTNAME sync error: read/write failed on /content. Please check health of RAID and /content mount on $HOSTNAME"
    exit 255
  fi
else
  $LOGGER -p daemon.err -t ORIGIN_SYNC "$HOSTNAME sync error: primary /content not mounted"
  exit 255
fi

# touch our lockfile
touch /var/run/newsync.running

# run our syncs
echo -e "Syncing ftp\n"
$RSYNC -vrtogpulH --delete-during --ignore-errors --exclude-from /home/xfer/exclude.file $HOSTNAME::ftpd/ /etc/proftpd/
RCODE=$?
if [ $RCODE != "0" ]; then
  if [ $RCODE = "1" ]; then
    $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME sync error: syntax error - proftpd"
  elif [ $RCODE = "2" ]; then
    $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME sync error: protocol error - proftpd"
  elif [ $RCODE = "10" ]; then 
    $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME failed to sync from primary: fatal connection error - exiting"
    rm /var/run/newsync.running
    exit 255
  elif [ $RCODE = "11" ]; then 
    $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME failed to sync from primary: fatal I/O error - exiting"
    rm /var/run/newsync.running
    exit 255
  elif [ $RCODE = "12" ]; then
    $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME sync error: datastream error - proftpd"
  elif [ $RCODE = "22" ]; then
    $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME sync error: memory error - proftpd"
  elif [ $RCODE = "23" ]; then
    $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME sync error: partial content - proftpd"
  elif [ $RCODE = "24" ]; then
    $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME sync error: source vanished - proftpd"
  elif [ $RCODE = "30" ]; then
    $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME sync error: data timeout - proftpd"
  elif [ $RCODE = "35" ]; then
    $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME sync error: connection timeout - proftpd"
  elif [ $RCODE = "5" ]; then
    # do nothing, it's semi-intentional - it means there's nothing there to sync
    echo "do nothing"
  else 
      $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME sync error: unknown code $RCODE - proftpd"
  fi
else 
    $LOGGER -p daemon.info -t ORIGIN_SYNC "$RHOSTNAME sync completed for proftpd"
    # touch a file to show when newsync last completed without errors
    touch /etc/newsync.lastcontact
fi

sleep 3
echo -e "Syncing srv\n"
$RSYNC -vrtogpulH --delete-during --ignore-errors --exclude-from /home/xfer/exclude.file $HOSTNAME::srv/ /srv/
RCODE=$?
if [ $RCODE != "0" ]; then
  if [ $RCODE = "1" ]; then
    $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME sync error: syntax error - proftpd"
  elif [ $RCODE = "2" ]; then
    $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME sync error: protocol error - proftpd"
  elif [ $RCODE = "10" ]; then
    $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME failed to sync from primary: fatal connection error - exiting"
    rm /var/run/newsync.running
    exit 255
  elif [ $RCODE = "11" ]; then
    $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME failed to sync from primary: fatal I/O error - exiting"
    rm /var/run/newsync.running
    exit 255
  elif [ $RCODE = "12" ]; then
    $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME sync error: datastream error - proftpd"
  elif [ $RCODE = "22" ]; then
    $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME sync error: memory error - proftpd"
  elif [ $RCODE = "23" ]; then
    $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME sync error: partial content - proftpd"
  elif [ $RCODE = "24" ]; then
    $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME sync error: source vanished - proftpd"
  elif [ $RCODE = "30" ]; then
    $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME sync error: data timeout - proftpd"
  elif [ $RCODE = "35" ]; then
    $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME sync error: connection timeout - proftpd"
  elif [ $RCODE = "5" ]; then
    # do nothing, it's semi-intentional - it means there's nothing there to sync
    echo "do nothing"
  else
      $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME sync error: unknown code $RCODE - proftpd"
  fi
else
    $LOGGER -p daemon.info -t ORIGIN_SYNC "$RHOSTNAME sync completed for proftpd"
    # touch a file to show when newsync last completed without errors
    touch /etc/newsync.lastcontact
fi

sleep 3
echo -e "Syncing /etc/sv\n"
$RSYNC -vrtogpulH --delete-during --ignore-errors --exclude-from /home/xfer/exclude.file $HOSTNAME::sv/ /etc/sv/
RCODE=$?
if [ $RCODE != "0" ]; then
  if [ $RCODE = "1" ]; then
    $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME sync error: syntax error - proftpd"
  elif [ $RCODE = "2" ]; then
    $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME sync error: protocol error - proftpd"
  elif [ $RCODE = "10" ]; then
    $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME failed to sync from primary: fatal connection error - exiting"
    rm /var/run/newsync.running
    exit 255
  elif [ $RCODE = "11" ]; then
    $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME failed to sync from primary: fatal I/O error - exiting"
    rm /var/run/newsync.running
    exit 255
  elif [ $RCODE = "12" ]; then
    $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME sync error: datastream error - proftpd"
  elif [ $RCODE = "22" ]; then
    $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME sync error: memory error - proftpd"
  elif [ $RCODE = "23" ]; then
    $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME sync error: partial content - proftpd"
  elif [ $RCODE = "24" ]; then
    $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME sync error: source vanished - proftpd"
  elif [ $RCODE = "30" ]; then
    $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME sync error: data timeout - proftpd"
  elif [ $RCODE = "35" ]; then
    $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME sync error: connection timeout - proftpd"
  elif [ $RCODE = "5" ]; then
    # do nothing, it's semi-intentional - it means there's nothing there to sync
    echo "do nothing"
  else
      $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME sync error: unknown code $RCODE - proftpd"
  fi
else
    $LOGGER -p daemon.info -t ORIGIN_SYNC "$RHOSTNAME sync completed for proftpd"
    # touch a file to show when newsync last completed without errors
    touch /etc/newsync.lastcontact
fi

sleep 3
if [ -e "/usr/local/nginx/sbin/nginx" ]; then
  echo -e "Syncing nginx\n"
  $RSYNC -vrtogpulH --delete-during --ignore-errors --exclude-from /home/xfer/exclude.file $HOSTNAME::nginx/ /usr/local/nginx/conf/
  RCODE=$?
  if [ $RCODE != "0" ]; then
    if [ $RCODE = "1" ]; then
      $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME sync error: syntax error - nginx"
    elif [ $RCODE = "2" ]; then
      $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME sync error: protocol error - nginx"
    elif [ $RCODE = "10" ]; then
      $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME failed to sync from primary: fatal connection error - exiting"
      rm /var/run/newsync.running
      exit 255
    elif [ $RCODE = "11" ]; then
      $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME failed to sync from primary: fatal I/O error - exiting"
      rm /var/run/newsync.running
      exit 255
    elif [ $RCODE = "12" ]; then
      $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME sync error: datastream error - nginx"
    elif [ $RCODE = "22" ]; then
      $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME sync error: memory error - nginx"
    elif [ $RCODE = "23" ]; then
      $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME sync error: partial content - nginx"
    elif [ $RCODE = "24" ]; then
      $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME sync error: source vanished - nginx"
    elif [ $RCODE = "30" ]; then
      $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME sync error: data timeout - nginx"
    elif [ $RCODE = "35" ]; then
      $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME sync error: connection timeout - nginx"
    elif [ $RCODE = "5" ]; then
      # do nothing, it's semi-intentional - it means there's nothing there to sync
    echo "do nothing"
    else 
      $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME sync error: unknown code $RCODE - nginx"
    fi
  else 
      $LOGGER -p daemon.info -t ORIGIN_SYNC "$RHOSTNAME sync completed for nginx"
    # touch a file to show when newsync last completed without errors
    touch /etc/newsync.lastcontact
  fi
fi

sleep 3
if [ -e "/usr/local/apache2/bin/httpd" ]; then
  echo -e "Syncing apache\n"
  $RSYNC -vrtogpulH --delete-during --ignore-errors --exclude-from /home/xfer/exclude.file $HOSTNAME::apache/ /usr/local/apache2/conf/
  RCODE=$?
  if [ $RCODE != "0" ]; then
    if [ $RCODE = "1" ]; then
      $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME sync error: syntax error - apache"
    elif [ $RCODE = "2" ]; then
      $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME sync error: protocol error - apache"
    elif [ $RCODE = "10" ]; then
      $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME failed to sync from primary: fatal connection error - exiting"
      rm /var/run/newsync.running
      exit 255
    elif [ $RCODE = "11" ]; then
      $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME failed to sync from primary: fatal I/O error - exiting"
      rm /var/run/newsync.running
      exit 255
    elif [ $RCODE = "12" ]; then
      $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME sync error: datastream error - apache"
    elif [ $RCODE = "22" ]; then
      $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME sync error: memory error - apache"
    elif [ $RCODE = "23" ]; then
      $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME sync error: partial content - apache"
    elif [ $RCODE = "24" ]; then
      $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME sync error: source vanished - apache"
    elif [ $RCODE = "30" ]; then
      $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME sync error: data timeout - apache"
    elif [ $RCODE = "35" ]; then
      $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME sync error: connection timeout - apache"
    elif [ $RCODE = "5" ]; then
      # do nothing, it's semi-intentional - it means there's nothing there to sync
    echo "do nothing"
    else 
      $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME sync error: unknown code $RCODE - apache"
    fi
  else 
      $LOGGER -p daemon.info -t ORIGIN_SYNC "$RHOSTNAME sync completed for apache"
    # touch a file to show when newsync last completed without errors
    touch /etc/newsync.lastcontact
  fi
fi 

# cycle the standard syncs

for type in ht fc wm rm qt disabled; do
  echo -e "Syncing $type\n"
  $RSYNC -vrtogpulH --delete-during --ignore-errors --bwlimit=90000 --exclude-from /home/xfer/exclude.file $HOSTNAME::$type/ /content/$type/
  RCODE=$?
  if [ $RCODE != "0" ]; then
    if [ $RCODE = "1" ]; then
      $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME sync error: syntax error - $type"
    elif [ $RCODE = "2" ]; then
      $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME sync error: protocol error - $type"
    elif [ $RCODE = "10" ]; then
      $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME failed to sync from primary: fatal connection error - exiting"
      rm /var/run/newsync.running
      exit 255
    elif [ $RCODE = "11" ]; then
      $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME failed to sync from primary: fatal I/O error - exiting"
      rm /var/run/newsync.running
      exit 255
    elif [ $RCODE = "12" ]; then
      $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME sync error: datastream error - $type"
    elif [ $RCODE = "22" ]; then
      $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME sync error: memory error - $type"
    elif [ $RCODE = "23" ]; then
      $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME sync error: partial content - $type"
    elif [ $RCODE = "24" ]; then
      $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME sync error: source vanished - $type"
    elif [ $RCODE = "30" ]; then
      $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME sync error: data timeout - $type"
    elif [ $RCODE = "35" ]; then
      $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME sync error: connection timeout - $type"
    elif [ $RCODE = "5" ]; then
      # do nothing, it's semi-intentional - it means there's nothing there to sync
    echo "do nothing"
    else 
      $LOGGER -p daemon.err -t ORIGIN_SYNC "$RHOSTNAME sync error: unknown code $RCODE - $type"
    fi
  else 
      $LOGGER -p daemon.info -t ORIGIN_SYNC "$RHOSTNAME sync completed for $type"
    # touch a file to show when newsync last completed without errors
    touch /etc/newsync.lastcontact.$type
  fi
done

# notify we're all done
$LOGGER -p daemon.info "newsync.sh finished run at $DATE"

# lets not forget to remove our lock file
rm /var/run/newsync.running
