#!/usr/bin/perl
#
use lib '/usr/local/llnw/backstage';

use strict;
use JSON;
use POSIX;
use YAML qw(LoadFile);
use LWP::Simple;
use Sys::Hostname; my $host = hostname;
use Sys::Syslog qw(:DEFAULT setlogsock);  # default set, plus setlogsock()
use Data::Dumper;

my $debug; $debug++ if $ARGV[0] eq 'debug';
print "I am: $host\n" if $debug;

my $config = LoadFile('/usr/local/llnw/backstage/etc/watchdogMQ.yml'); # Read config options
my $backstage = $config->{backstageAPI};

# End user configurables. Touch anything else and it's your ass.

# Initialization block
my $OS = $^O;

my $daemon = $0; $daemon =~ s/^(.*)\/(\S+)$/$2/;        # identify ourselves, sans path if we were called that way.
openlog("$daemon\[$$\]", 'ndelay', 'daemon');       # Initialize syslog

# End initialization

# main process
my %fstab = readFStab();
my %blocks = dfBlocks();

unless ( -e '/etc/origin.pair') { # Attempt to generate a default configuration if we don't have one
    sendlog("No configuration found, attempting to generate a default version for $host");
    my $remote = determineSibling($host);

    if ($remote) {
    	if ( open(OUT, '>', '/etc/origin.pair') ) {
	        print OUT "$remote /content 5000\n";
       		 close OUT;
        	sendlog("Default configuration written");
    	}
    	else {
        	sendlog("Unable to write new/default configuration file, /etc/origin.pair: $!");
    	}

    }
    else {
        sendlog("Unable to determine sibling origin from hostname, cannot generate default configuration, aborting");
	exit;
    }
}

originParityCheck() ;

# end main process
sub originParityCheck {
	if ( open(IN, '/etc/origin.pair') ) { 			# Read hostnames and sister partitions from /etc/origin.pair

		while(<IN>) {
			my $line = $_;
			chomp($line);
			next if ($line =~ /^#|^$/); 			# Skip blank/commented lines

			my ($hostname, $partition, $parity, $group) = split(/\s+/, $line); # Foreach host and partition

			# Ask backstage for the disk util of partition X on device Y, compare to local partition X
			if ( my $remoteUtil =  partitionPairCheck($hostname, $partition) ) {	
				my $disparity = abs($blocks{$partition}{bused} - $remoteUtil->{Counter}) / 1000;


				# Send alarm if disparity greater than Z megabytes
				if($disparity >= $parity) {
					sendlog( printf( "%s:%s usage (%s MB) <-> %s:%s usage (%s) disparity of %s MB is greater than %s MB parity threshold", $host, $partition, $blocks{$partition}{bused} / 1000, $hostname, $partition, $remoteUtil->{Counter} / 1000, $disparity, $parity ));

				}
			}
			else {
				# Send alarm if stats are missing from backstage (misconfiguration, sister device not reporting)
				sendlog( "Pair check of $hostname failed to return any data via the API." );
			}
		}
	}
	else {
		sendlog("Failed to read /etc/origin.pair: $!");
	}
}

# Functions
#
sub determineSibling {
    my ($hostname) = @_;

    my $sibling = $hostname;

    if($hostname =~ /bocs|^borigin/) {  # is backup
            $sibling =~ s/b(ocs|origin)/$1/;
    }
    elsif($hostname =~ /ocs|^origin/) { # is origin
            $sibling =~ s/(ocs|origin)/b$1/;
    }
    else {          # something else entirely
        sendlog("Hostname doesn't match expected origin formats: $hostname");
        return();
    }

    my $json = backstageFetch( sprintf('metrics/%s?metricType=%s&instance=%s', $sibling, 'Disk Used', '/content' ) );
    my @arrayRef = @{from_json($json)};
    my $metricRef = pop @arrayRef;
	
    return($sibling) if $metricRef;

    print "Assumed $sibling was paired origin, could not get metrics for /content";
    return(0);
}

sub partitionPairCheck {
	my ($host, $partition, $parity) = @_;

	print "Checking $host, $partition via backstage\n" if $debug;

	# Ask backstage for the disk util of partition X on device Y, compare to local partition X
	if( my $json = backstageFetch( sprintf('metrics/%s?metricType=%s&instance=%s', $host, 'Disk Used', $partition) ) ) {

		my @arrayRef = @{from_json($json)};
		my $metricRef = pop @arrayRef;

		unless($metricRef) {
		 	sendlog("Parity check failed, unable to find metrics for $host via backstage");
			print "Parity check failed, unable to find metrics for $host via backstage";
			exit;
		}

		my $time = time - 3600;
		my $usage = backstageFetch( sprintf('getMetrics/%s?start=%d', $metricRef->{MetricID}, $time ));
		my %hash = %{from_json($usage)};

		if( my $metricRef = pop(@{$hash{$metricRef->{MetricID}}}) ) {
			return($metricRef);
		}
	}
	else {
		sendlog("Fetch of metric instances failed.\n");
	}

	sendlog("partitionPairCheck: Fetch of disk util for $host:$partition failed");
	return(0);
}

sub readFStab {
	# Fstab formats between freebsd and debian are identical

	my $acceptFileSystems = $config->{acceptFileSystems};
	$acceptFileSystems =~ s/\s+/|/g;

	my $ignoreFileSystems = $config->{ignoreFileSystems};
	$ignoreFileSystems =~ s/\s+/|/g;

	if ( open(IN, '/etc/fstab') ) {
		my %hash;
		while(<IN>) {
			next if ($_ =~ /^#|^$/); # Skip blank/commented lines
			my ($device, $mountpoint, $fsType, $options, $dump, $pass) = split(/\s+/, $_);
			next unless $fsType =~ $acceptFileSystems;
			next if $mountpoint =~ $ignoreFileSystems;

			$hash{$mountpoint} = (
				Device => $device,
				FSType => $fsType,
				Options => $options,
				Dump => $dump,
				Pass => $pass
			);
		}
		close(IN);

		return(%hash);
	}
	else {
		sendlog("Unable to read /etc/fstab: $!");
		return(0);
	}
}

sub dfBlocks {
	my %hash;

        $SIG{ALRM} = sub { die "timeout" };
	eval {
                alarm(2);

		if ( open(IN, '/bin/df -kl|') ) {
			while(<IN>) {
				chomp;
				my $value = $_;
       	 			next if $value =~ /procfs|devfs|File|udev/;

				my ($disk, $totalBlocks, $usedBlocks, $availBlocks, $blockCapacity, $used, $avail, $diskUtil, $mountPoint, $total);
				my @values = split(/\s+/, $value);

        			if(scalar @values >= 9) {       #freebsd inode return
              			# freebsd df -i
              			#Filesystem  1K-blocks    Used     Avail Capacity iused   ifree %iused  Mounted on
              			#/dev/da0s1a  10154158 1572118   7769708    17%  162833 1156077   12%   /

              				($disk, $totalBlocks, $usedBlocks, $availBlocks, $blockCapacity, $used, $avail, $diskUtil, $mountPoint) = @values;
					$diskUtil =~ s/(\d+)\%/$1/g;
	      				next if $disk eq 'none';
              				$total = $used + $avail;
        			}
        			else {                          #linux inode/block return, freebsd block return
				# Freebsd df -k
				# Filesystem  1K-blocks    Used     Avail Capacity  Mounted on
				# /dev/da0s1a  10154158 1572120   7769706    17%    /
	
				# linux df -k
				# Filesystem           1K-blocks      Used Available Use% Mounted on
				# /dev/sda1            138353996  11020812 120305152   9% /

					($disk, $total, $used, $avail, $diskUtil, $mountPoint) = @values;
					$diskUtil =~ s/(\d+)\%/$1/g;
					next if $disk eq 'none';
				}

				$hash{$mountPoint} =  { 
					blocks => $total,
					bused => $used,
					per => $diskUtil
				};			
			}
		}
		else {
			sendlog( "Unable to read df output: $!");
		}

                alarm(0);
	};

        alarm(0);

        if ($@) {
                sendlog("Error reading partition info: $@");
        }
	return %hash;
}

sub dfInodes {
	my %hash;

        $SIG{ALRM} = sub { die "timeout" };
	eval {
                alarm(2);

		if ( open(IN, '/bin/df -i|') ) {
			while(<IN>) {
				chomp;
				my $value = $_;
       	 			next if $value =~ /procfs|devfs|File|udev/;

				my ($disk, $totalBlocks, $usedBlocks, $availBlocks, $blockCapacity, $used, $avail, $diskUtil, $mountPoint, $total);
				my @values = split(/\s+/, $value);

        			if(scalar @values >= 9) {       #freebsd inode return
              			# freebsd df -i
              			#Filesystem  1K-blocks    Used     Avail Capacity iused   ifree %iused  Mounted on
              			#/dev/da0s1a  10154158 1572118   7769708    17%  162833 1156077   12%   /

              				($disk, $totalBlocks, $usedBlocks, $availBlocks, $blockCapacity, $used, $avail, $diskUtil, $mountPoint) = @values;
	      				next if $disk eq 'none';
              				$total = $used + $avail;
        			}
        			else {                          #linux inode/block return, freebsd block return
				# Freebsd df -k
				# Filesystem  1K-blocks    Used     Avail Capacity  Mounted on
				# /dev/da0s1a  10154158 1572120   7769706    17%    /
	
				# linux df -k
				# Filesystem           1K-blocks      Used Available Use% Mounted on
				# /dev/sda1            138353996  11020812 120305152   9% /

					($disk, $total, $used, $avail, $diskUtil, $mountPoint) = @values;
					next if $disk eq 'none';
				}

				$diskUtil =~ s/(\d+)\%/$1/g;

				$hash{$mountPoint} =  { 
					inodes => $total,
					iused => $used,
					per => $diskUtil
				};			
			}
		}
		else {
			sendlog( "Unable to read df output: $!");
		}

                alarm(0);
	};

        alarm(0);

        if ($@) {
                sendlog("Error reading partition info: $@");
        }
	return %hash;
}

# Backstage/MQ bus internals. If you touch these, your ass is the least of your worries.

sub backstageFetch {
	my ($route) = @_;
	print "Fetching $route\n" if $debug;
	if ( my $json = get( sprintf('http://%s/%s', $backstage, $route) ) ) {
		return($json)
	}
	else { sendlog("Backstage fetch of $route failed"); return; }
}

sub sendlog {
        my($error)= @_;
        warn $error . "\n" if $debug;
        syslog( 'warning', '%s', $error);
}

