#!/usr/bin/env perl

use strict;
use warnings;

#######################################################################
#Description: Script to apply firewall rules on dns[123]xx servers
#to block port 179, 2601, and 2605 from the outside world.
#
#https://confluence.atlas.llnw.com/display/OPS/dns_ipfw.pl-runbook
#Author: JWillbanks
#Date: 2015-03-11
#Updated: 2018-03-27 SRE-1986 
#######################################################################


my @path = qw{
  /etc
  /usr/local/sbin
  /usr/local/bin
  /sbin
  /usr/sbin
  /bin
  /usr/bin
};

$ENV{'PATH'} = join ":", @path;

use constant {
  OK    => 0,
  FATAL => 1,
};

my $base_name = (split m{/}, $0)[-1];
my $tmp_address;
my $ip_address;
my @ip_address;
my $ip_address_file;
my $bgp_ipv4_neighbor;
my $OS = $^O;
my $linux_interface;


if ( $OS =~ /freebsd/ ) {
  $ip_address_file = "/etc/rc.conf";

  get_bsd_ip();
  get_bgp_neighbor_ip();

  if ( check_bsd_firewall_rules() == 0 ) {
    apply_bsd_firewall_rules();
  }
  else {
    flush_bsd_firewall_rules();
    apply_bsd_firewall_rules();
  }
}
elsif ( $OS =~ /linux/ ) {
  $linux_interface = get_linux_interface();
  get_bgp_neighbor_ip();

  if ( check_linux_firewall_rules() == 0 && check_linux_v6_firewall_rules() == 0 ) {
    apply_linux_firewall_rules();
    apply_linux_v6_firewall_rules();
  }
  else {
    flush_linux_firewall_rules();
    flush_linux_v6_firewall_rules();
    apply_linux_firewall_rules();
    apply_linux_v6_firewall_rules();
  }
}


write_zabbix_status(OK);

###### SUBS BELOW ######

sub fatal {
  my ($msg) = @_;
  print STDERR "FATAL: $msg\n";
  write_zabbix_status(FATAL);
  exit 1;
}

sub write_zabbix_status {
  my $status = shift;
  my $file = "/tmp/$base_name.status";

  open my $out, '>', $file or die "Failed to open status file: $!";
  print $out $status or die "Failed to write status file: $!";
  close $out or die "Failed to close status file: $!";
}

sub get_bsd_ip {
  open(my $RC_CONF, $ip_address_file) || fatal("Failed to open $ip_address_file: $!\n");
    while (my $row = <$RC_CONF>) {
      if ($row =~ /^ifconfig_.*inet (\d{1,3})\.(\d{1,3})\.(\d{1,3})\.(\d{1,3})/) {
        ($tmp_address) = (split ' ',$row)[1];
      }
    }
    unless ( $tmp_address =~ /(\d{1,3})\.(\d{1,3})\.(\d{1,3})\.(\d{1,3})/ ) {
          fatal("ip address $tmp_address from $ip_address_file does not appear to be valid\n");
    }
  ($ip_address) = ($tmp_address);
  close($RC_CONF) || fatal("Failed to close: $ip_address_file $!\n");
}

sub get_linux_interface {
  open(my $NETSTAT,"netstat -r |") || die "Failed: $!\n";
    while (my $row = <$NETSTAT>) {
      if ($row =~ /default/) {
        ($linux_interface) = (split ' ',$row)[7];
        return $linux_interface;
      }
    }
  close($NETSTAT) || fatal("Failed to close: $NETSTAT $!\n");
}

sub get_bgp_neighbor_ip {
    #different versions of quagga behave differently, ugly but universal way to extract running information from the config
    open(my $BGP_CONF, "-|", "( echo \"term length 0\" ; echo \"show run\") | vtysh") || fatal("Failed to get bgp neighbor data $!\n");
        while (my $row = <$BGP_CONF>) {
            if ($row =~ m/neighbor (\d{1,3})\.(\d{1,3})\.(\d{1,3})\.(\d{1,3}) remote-as/) {
                ($bgp_ipv4_neighbor) = (split ' ',$row)[1];
            }
        }
    close($BGP_CONF) || fatal("Failed to close: 'vtysh -c show run' $!\n");
}

sub check_bsd_firewall_rules {
    #Check to make sure the rules are applied appropriately and increment the counter if they are
    my $fw_rule_counter = 0;

    open(my $IPFW_RULES, "ipfw set 1 list |") || fatal("Failed to execute: $!\n");
    
    while (my $row = <$IPFW_RULES>) {
        chomp $row;
        if ( $row eq "00001 allow ip from any to any dst-port 53" ) {
            $fw_rule_counter ++;
        }
        elsif ( $row eq "00300 allow ip from $bgp_ipv4_neighbor to any in recv $ip_address" ) {
            $fw_rule_counter ++;
        }
        elsif ( $row eq "00301 deny ip from any to any dst-port 179 in recv $ip_address" ) {
            $fw_rule_counter ++;
        }
        elsif ( $row eq "00302 deny ip from any to any dst-port 2605 in recv $ip_address" ) {
            $fw_rule_counter ++;
        }
        elsif ( $row eq "00303 deny ip from any to any dst-port 2601 in recv $ip_address" ) {
            $fw_rule_counter ++;
        }
    }
    close($IPFW_RULES) || fatal("Failed to close: $!\n");
    return $fw_rule_counter;
}

sub check_linux_firewall_rules {
    #Check to make sure the rules are applied appropriately and increment the counter if they are
    my $fw_rule_counter = 0;

    open(my $IPTABLES_RULES, "iptables -S |") || fatal("Failed to execute: $!\n");
    
    while (my $row = <$IPTABLES_RULES>) {
        chomp $row;
        if ( $row eq "-A INPUT -i $linux_interface -p udp --dport 53 -j ACCEPT" ) {
          $fw_rule_counter ++;
        }
        elsif ( $row eq "-A INPUT -i $linux_interface -p tcp --dport 53 -j ACCEPT" ) {
          $fw_rule_counter ++;
        }
        elsif ( $row eq "-A INPUT -i $linux_interface -p tcp --dport 179 -s $bgp_ipv4_neighbor -j ACCEPT" ) {
          $fw_rule_counter ++;
        }
        elsif ( $row eq "-A INPUT -i $linux_interface -p tcp -m multiport --dports 179,2605,2601 -j DROP" ) {
          $fw_rule_counter ++;
        }
    }
    close($IPTABLES_RULES) || fatal("Failed to close: $!\n");
    return $fw_rule_counter;
}

sub check_linux_v6_firewall_rules {
    #Check to make sure the rules are applied appropriately and increment the counter if they are
    my $fw_rule_counter = 0;

    open(my $IP6TABLES_RULES, "ip6tables -S |") || fatal("Failed to execute: $!\n");

    while (my $row = <$IP6TABLES_RULES>) {
        chomp $row;
        if ( $row eq "-A INPUT -p tcp -m multiport --dports 179,2605,2601 -j DROP" ) {
          $fw_rule_counter ++;
        }
    }
    close($IP6TABLES_RULES) || fatal("Failed to close: $!\n");
    return $fw_rule_counter;
}

#Apply ipfw rules under 'set 1' which will segment them from the default 'set 0' rules.
#Since rules are applied under 'set 1' a flush of 'set 1' will only flush the rules
#associated with 'set 1'.  It was done this way so we don't accidently flush the default
#permit any any rule at the end of the ipfw rules associated with set 31 and rule number
#65535.
#dns101# ipfw -S list
#65535 set 31 allow ip from any to any

sub apply_bsd_firewall_rules {
    my @ipfw_commands = (
        "ipfw -q add 1 set 1 accept ip from any to any 53",
        "ipfw -q add 300 set 1 accept ip from $bgp_ipv4_neighbor to any in recv $ip_address",
        "ipfw -q add 301 set 1 deny ip from any to any 179 in recv $ip_address",
        "ipfw -q add 302 set 1 deny ip from any to any 2605 in recv $ip_address",
        "ipfw -q add 303 set 1 deny ip from any to any 2601 in recv $ip_address"
    );

    exec_bsd_firewall(@ipfw_commands);
}

sub apply_linux_firewall_rules {
  my @iptables_commands = (
    "iptables -A INPUT -i $linux_interface -p udp --dport 53 -j ACCEPT",
    "iptables -A INPUT -i $linux_interface -p tcp --dport 53 -j ACCEPT",
    "iptables -A INPUT -i $linux_interface -p tcp --dport 179 -s $bgp_ipv4_neighbor -j ACCEPT",
    "iptables -A INPUT -i $linux_interface -p tcp -m multiport --dport 179,2605,2601 -j DROP"
  );

  exec_linux_firewall(@iptables_commands);
}

sub apply_linux_v6_firewall_rules {
  my @ip6tables_commands = (
    "ip6tables -A INPUT -p tcp -m multiport --dport 179,2605,2601 -j DROP"
  );

  exec_linux_v6_firewall(@ip6tables_commands);
}

sub flush_bsd_firewall_rules {
  my @ipfw_flush_commands = (
    "ipfw -q -f set 1 flush"
  );

  exec_bsd_firewall(@ipfw_flush_commands);
}

sub flush_linux_firewall_rules {
  my @iptables_flush_commands = (
    "iptables --flush"
  );

  exec_linux_firewall(@iptables_flush_commands);
}

sub flush_linux_v6_firewall_rules {
  my @ip6tables_flush_commands = (
    "ip6tables --flush"
  );

  exec_linux_v6_firewall(@ip6tables_flush_commands);
}

sub exec_bsd_firewall {
  my @commands = @_;

  for my $cmd (@commands) {
    if (system($cmd) != 0) {
      system("ipfw -q -f set 1 flush");
        fatal("$cmd failed to execute properly, flushing ipfw rules\n");
    }
  }
}

sub exec_linux_firewall {
  my @commands = @_;

  for my $cmd (@commands) {
    if (system($cmd) != 0) {
      system("iptables --flush");
        fatal("$cmd failed to execute properly, flushing iptables rules\n");
    }
  }
}

sub exec_linux_v6_firewall {
  my @commands = @_;

  for my $cmd (@commands) {
    if (system($cmd) != 0) {
      system("ip6tables --flush");
        fatal("$cmd failed to execute properly, flushing ip6tables rules\n");
    }
  }
}
