#!/usr/bin/perl -w
#
# copy files over http - Brad Harvell 2009-03-07
#
# http has the benefits of near zero overhead, if-modified-since
# to check if we already have current files, and can be used through
# caching proxy servers (use good expires headers!) so can be very
# scalable and still get updates frequently / quickly
#
# multiple config files can be specified, each one is optional (it is
# not an error for some or all to not exist)
#
# configured http urls will be fetched if/when they are newer than
# current local files (or if there is no local file yet)
#
# if the urls are tarballs (.tar, .tgz, .tbz2) the tarballs will be
# fetched, files extracted, then these files will be put in place.
# tarballs should not have a directory structure. this is especially
# handy since you can't simply transfer all files that match a
# provided pattern since http does not have a call to get directory
# listings like ftp and rsync.
#
# if they are not tarballs, the files can still be compressed (.gz,
# .tbz2), and will get decompressed automatically
#
# if we do successfully get some new files, we'll run whatever
# reconfigure type scripts that are defined
#
# no dependancies. even though it would have been easy to use some
# external modules for some functions, maybe wget, it probably would
# end up being more hassle than it is worth from having to make sure
# every box this script runs on has all the extra stuff.

use strict;
use warnings;
use Socket;
use IO::Socket;
use Time::Local;
use Time::HiRes qw(sleep);
use Fcntl qw(:flock);
use CGI::Util qw();

# tuning
my @confs = qw( /etc/geo/config_urls.txt );
my $logfile = '/var/log/htcopy';
my $lockfile = '/var/run/htcopy.lock';
my $targetdir = '/tftpboot';
my $reconfig = undef;
# my @confs = qw(/tmp/config_urls.txt);
# my $logfile = "/tmp/htcopy.log";
# my $lockfile = '/tmp/htcopy.lock';
# my $targetdir = '/tmp/sq';
# my $reconfig = "/tmp/sq/reconfig";
umask(022);

my $ctimeout = 10; # seconds for connection timeout
my $randsleep = 60; # max seconds to sleep (to offset requests)

# initialization
my @entries;
my @loglevelcounts;
my $warns = 0;
my $fails = 0;
my $logfh;

my $interactive = -t STDIN;
if (! $interactive) {
  my $sleep = rand($randsleep);
  &plog(3, "sleeping $sleep seconds");
  sleep($sleep);
}

&getlock;
&readconfs;
&checkfiles;
# XXX if we downloaded something new, run whatever "reconfig" scripts
&mayberunscripts;
&savelog;
exit(0);

sub checkfiles {
  foreach my $entry (@entries) {
    my $res = &trydownload($entry);
    # XXXX ok. gotta decompress / untar / shuffle files around, then run scripts
    if ($res) {
	if ($reconfig) {
	    &plog(3, "Touching $reconfig\n");
	    open(RECONFIG, ">>$reconfig");
	    close(RECONFIG);
	}
    }
  }
};

sub trydownload {
  my($entry) = @_;
  my $file = $entry->{'filename'};
  my $res;

  # connect to server
  my($sock) = &tryconnecthosts($entry);
  if (! $sock) {
    return undef;
  }
  my $addr = $entry->{'addr'};
  my $hostname = $entry->{'hostname'};

  # send http request
  my $req = &makehttpreq($entry);
  &plog(3, "sending request\n", $req);
  print $sock $req;

  # read http response
  my $line = <$sock>;
  $line =~ s/[\r\n]+$//;
  if ($line !~ m@^HTTP/1.\d (\d\d\d)($| )@) {
    &plog(0, "trydownload($file) unexpected response from $hostname/$addr: '$line'");
    $entry->{'failed'} = 1;
    return undef;
  }
  my $rcode = $1;
  if ($rcode == 304) {
    &plog(3, "trydownload($file) got 304 from $hostname/$addr: '$line'");
    $entry->{'unchanged'} = 1;
    return undef;
  }
  if ($rcode != 200) {
    # could be 404 not found, etc
    &plog(1, "trydownload($file) got $rcode from $hostname/$addr: '$line'");
    $entry->{'failed'} = 1;
    return undef;
  }
  my %headers;
  while (<$sock>) {
    s/[\r\n]+$//;
    if ($_ eq '') {
      &plog(3, "trydownload($file): end of headers");
      last;
    }
    my ($header, $val) = split(/:\s/, $_, 2);
    &plog(3, "header='$header' val='$val'");
    $headers{$header} = $val;
  }

  # sanity check for headers
  if (!defined $headers{'Last-Modified'}) {
    &plog(1, "trydownload($file) response from $hostname is missing Last-Modified header");
    return undef;
  }
  my $lastmod = &httptime2epoch($headers{'Last-Modified'});
  if (!defined $lastmod) {
    &plog(1, "trydownload($file) can't decode Last-Modified header from $hostname");
    return undef;
  }
  if (!defined $headers{'Content-Length'}) {
    &plog(1, "trydownload($file) response from $hostname is missing Content-Length header");
    return undef;
  }

  # open output file
  my $path = "$entry->{targetdir}/$file";
  open(NEW, ">$path.new") or die "can't write $path: $!";

  # read http body
  my $clen = $headers{'Content-Length'};
  my $havebytes = 0;
  while ($clen > $havebytes) {
    my $left = $clen - $havebytes;
    my $buflen = $left;
    if ($left > 32768) {
      $buflen = 32768;
    }
    my $buf;
    my $bytes = read($sock, $buf, $buflen);
    if (!defined $bytes) {
      &plog(0, "trydownload($file) $hostname read() of $buflen bytes failed: $!");
      $res = unlink("$path.new");
      if (! $res) {
        &plog(0, "can't unlink $path.new: $!");
      }
      return undef;
    }
    if ($bytes == 0) {
      &plog(0, "trydownload($file) $hostname read() of $buflen returned zero: $!");
      $res = unlink("$path.new");
      if (! $res) {
        &plog(0, "can't unlink $path.new: $!");
      }
      return undef;
    }
    $havebytes += $bytes;
    &plog(4, "trydownload($file) $hostname read() received $bytes (now $havebytes)");
    $res = print NEW $buf;
    if (! $res) {
      &plog(0, "trydownload($file) writing $bytes bytes to $path.new at $havebytes bytes failed: $!");
      $res = unlink("$path.new");
      if (! $res) {
        &plog(0, "can't unlink $path: $!");
      }
      return undef;
    }
  }
  &plog(3, "trydownload($file) $hostname $havebytes bytes download complete");
  close $sock;
  close NEW;

  # set the file timestamp
  $res = utime($lastmod, $lastmod, "$path.new");
  if (! $res) {
    &plog(0, "trydownload($file) can't set utime on $path.new: $!");
    $res = unlink("$path.new");
    if (! $res) {
      &plog(0, "can't unlink $path.new: $!");
    }
    return undef;
  }
  $res = rename("$path.new", $path);
  if (! $res) {
      &plog(0, "trydownload($file) can't rename $path into place: $!");
      return undef
  }
  $entry->{'downloaded'} = 1;
  return($havebytes);
}

sub makehttpreq {
  my($entry) = @_;
  # generate and send http request
  my $mtime = &getmtime($entry);
  my $req = sprintf "GET %s HTTP/1.0\r\n", $entry->{'httppath'};
  $req .= sprintf "Host: %s\r\n", $entry->{'httphost'};
  if ($mtime > 0) {
    $req .= sprintf "If-Modified-Since: %s\r\n", CGI::Util::expires($mtime);
  }
  $req .= "User-Agent: htcopy/0.1\r\n";
  $req .= "X-LDebug: 1\r\n";
  $req .= "\r\n";
  return($req);
}

sub getmtime {
  my($entry) = @_;
  my($path, $mtime);
  my @st;
  my $targetdir = $entry->{'targetdir'};
  my $file = $entry->{'filename'};
  # this function is used get a timstamp to generate the
  # if-modified-since header
  $path = "$targetdir/$file";
  @st = stat($path);
  if ($#st == -1) {
    &plog(3, "getmtime: $path: no such file");
    if ($file =~ /^(.*)\.(gz|bz2)$/) {
      $path = "$targetdir/$1";
    }
    else {
      $path = "$targetdir/$file";
    }
    @st = stat($path);
  }
  if ($#st == -1) {
    &plog(3, "getmtime: $path: no such file");
    return -1;
  }
  $mtime = $st[9];
  &plogf(3, "getmtime(%s, %s): path=%s mtime='%s'\n",
	$targetdir, $file, $path, $mtime);
  return($mtime);
}

sub tryconnecthosts {
  my($entry) = @_;
  # until successful, try each parent and as last resort try direct
  if ($entry->{'parents'}) {
    my @parents = split(" ", $entry->{'parents'});
    #@parents = sort { &hash("$file $a") <=> &hash("$file $b") } @parents;
    @parents = sort { rand() <=> rand() } @parents;
    &plogf(3, "tryconnect: parents(%s): %s\n",
	$entry->{'filename'}, join(",", @parents));
    while (my $parent = shift(@parents)) {
      my $sock = &tryconnecthostname($entry, $parent);
      if ($sock) {
        return($sock);
      }
    }
  }
  my $sock = &tryconnecthostname($entry, $entry->{'httphost'});
  if ($sock) {
    return($sock);
  }
  &plogf(2, "trydownload(%s): could not connect to %s: %s/%s",
	$entry->{'filename'}, $entry->{'httphost'}, $!, $@);
}

sub tryconnecthostname {
  my($entry, $hostname) = @_;
  my(@iaddrs);
  my(@addrs);
  my $port = 80;
  if ($hostname =~ s/:(\d+)$//) {
    $port = $1;
  }
  (undef, undef, undef, undef, @iaddrs) = gethostbyname($hostname);
  if ($#iaddrs < 0) {
    &plog(2, "unable to look up hostname '$hostname': '$!'");
    return undef;
  }
  foreach my $iaddr (@iaddrs) {
    push(@addrs, inet_ntoa($iaddr));
  }
  &plogf(3, "tryconnecthostname: resolved %s to %s", $hostname, join(" ", @addrs));

  # try to connect to each address
  foreach my $addr (@addrs) {
    &plogf(3, "connecting to %s at %s:%s\n", $hostname, $addr, $port);
    my $sock = IO::Socket::INET->new(
                PeerAddr => $addr,
                PeerPort => $port,
                Proto => 'tcp',
                Timeout => $ctimeout);
    if ($sock) {
      &plogf(3, "connection established to %s at %s:%s\n", $hostname, $addr, $port);
      $entry->{'hostname'} = $hostname;
      $entry->{'addr'} = $addr;
      $entry->{'port'} = $port;
      return($sock);
    }
    else {
      &plogf(2, "failed to connect to %s at %s:%s: %s/%s\n", $hostname, $addr, $port, $!, $@);
      next; # try the next addr if there is one
    }
  }
  return undef; # we couldn't connect anywhere
}

sub readconfs {
  foreach my $conf (@confs) {
    my $res = open(IN, $conf);
    if (! $res) {
      # each config is optional - not a failure condition
      &plogf(2, "can not read $conf: $!");
      next;
    }
    while (<IN>) {
	my $entry;
	s/[\r\n]+$//;
	my $line = $_;
	s/\#.*$//;
	s/[ \t]+$//;
	next if ($_ eq '');
	if (s/^targetdir\s+//) {
	    $targetdir = $_;
	    next;
	}
	my ($file, $url, $parents) = split(" ", $_, 3);
	if ($url !~ m@http://([^/]+)(/.*)@) {
	    &plog(1, "$conf: bad url $url");
	    next;
	}
	$entry->{'httphost'} = $1;
	$entry->{'httppath'} = $2;
	$entry->{'filename'} = $file;
	$entry->{'url'} = $url;
	$entry->{'parents'} = $parents;
	$entry->{'targetdir'} = $targetdir;
	push(@entries, $entry);
    }
    close IN;
  }
}

sub httptime2epoch {
  my($str) = @_;
  my($sec,$min,$hour,$mday,$mon,$year);
  my($monstr);
  my %mon2num = qw( Jan 0 Feb 1 Mar 2 Apr 3 May 4 Jun 5 Jul 6 Aug 7 Sep 8 Oct 9 Nov 10 Dec 11 );
  my $res = ($mday, $monstr, $year, $hour, $min, $sec) = ($str =~ /^\w\w\w, (\d\d) (\w\w\w) (\d\d\d\d) (\d\d):(\d\d):(\d\d) GMT$/);
  if (! $res) {
    &plog(1, print "httptime2epoch: could not get a pattern match on '$str'");
    return undef;
  }
  $mon = $mon2num{$monstr};
  if (! defined $mon) {
    &plog(1, "httptime2epoch: could not get a month match on '$str' ($monstr)");
    return undef;
  }
  $year -= 1900;
  my $time = timegm($sec,$min,$hour,$mday,$mon,$year);
  # confirmation test:
  #$newstr = CGI::Util::expires($time);
  #print "orig: $str\n";
  #print "new:  $newstr\n";
  return($time);
}

sub getlock {
  open(LOCK, ">>$lockfile") or die "can't write to $lockfile: $!";
  my $res = flock(LOCK, LOCK_EX|LOCK_NB);
  if (! $res) {
    if (-t STDIN) {
      print "$0: looks like i'm already running\n";
    }
    exit 0;
  }
}

sub mayberunscripts {
  # XXXX maybe do some reconfiguring?
}

sub savelog {
  # rename log file depending on whether warnings/errors happened
  my $newname;
  if ($loglevelcounts[0]) {
    $newname = "$logfile.crit";
  }
  elsif ($loglevelcounts[1]) {
    $newname = "$logfile.err";
  }
  elsif ($loglevelcounts[2]) {
    $newname = "$logfile.warn";
  }
  else {
    $newname = "$logfile.info";
  }
  my $res = rename($logfile, $newname);
  if (! $res) {
    # odd. if we can't rename it, do we even have a log? print error
    warn "can't rename $logfile to $newname: $!";
  }
}

sub plog {
  my($level, @rest) = @_;
  # levels:
  # - 0 unusual error - interrupted downloads, write failure
  # - 1 file not found, denied, content-lenght or last-modified missing
  # - 2 intermittant error types - dns error, connect error
  # - 3+ normal operation messages
  $loglevelcounts[$level]++;
  if (! $logfh) {
    $logfh = new IO::Handle;
    my $res = open($logfh, ">>$logfile");
    if (! $res) {
      undef $logfh;
      warn "can not append to '$logfile': $!";
      return;
    }
    $logfh->autoflush(1);
  }
  foreach my $str (split(/\n/, join('', @rest))) {
    if (-t STDIN) {
      print $str . "\n";
    }
    print $logfh $str . "\n";
  }
}

sub plogf {
  my($level, $fmt, @rest) = @_;
  &plog($level, sprintf($fmt, @rest));
}

# todo - being smarter, clearer
# XXX differentiate "strange or unusual" errors from normal errors. maybe don't put files in place in some cases
# XXX save a copy of replaced files?
# XXX do a 'cmp' of "new" and "old" if same size before blindly replacing? (don't reconfig if content is the same)
# XXX only replace if "new" file has a newer timestamp? or replace but don't run commands?
# XXX have an option to force http transfer / file copy / commands?

# consider:
# XXX syslog under some error conditions?
# XXX check free disk space (blocks and inodes)
# XXX take command line argument for config file(s) maybe http?
# XXX any need to set owner/group/perms/attribs?
# XXX mark a group/collection of files, and for each group in
#    order, download all files, then install all files and run
#    scripts, then move on to the next group
