#!/usr/bin/perl -w

use strict;
use Socket;
use IO::Handle;

my $route_uri = "/cgi-bin/receive-route-data";
my $update_interval = 89;
my $runtime = 3600*(6 + int(6*rand())) - 3*$update_interval;
my $old_upload_interval = 45*60;
my $bgp_logdir = "/logs/bgp";
my $timeout = 600; # die if we hang for this long

use Fcntl qw(:flock);
my $lockfile = '/var/run/lockedrun.routetracker';
open(LOCK, ">>$lockfile") or die "$0: can't write to $lockfile: $!";
if (!flock(LOCK, LOCK_EX|LOCK_NB)) {
    print "$0: already running\n";
    exit(0);
}

chdir("$bgp_logdir") || die "$bgp_logdir: $!";

if (! -t STDOUT) {
    my $f = fork();
    die "fork: $!" if $f < 0;
    if ($f > 0) {
	# parent
	print "$0 started\n";
	exit(0);
    }

    rename("$bgp_logdir/routetracker.log", "$bgp_logdir/routetracker.log.age");
    open(STDOUT, ">/logs/bgp/routetracker.log") || die "routetracker.log: $!";
    open(STDERR, ">&STDOUT");
    $|=1;
    open(STDIN, "</dev/null");
}

$SIG{ALRM} = sub { die "alarm\n" };

my %submitted_size; # since file headers change based on time, assume that files of the same size are identical
my $now = time();
my $start_time = $now;
my $last_old_upload = $now + $update_interval - $old_upload_interval;

# generate a random offset for uploading, so not all servers
# upload exactly at the beginning of the minute
my $random_offset = int(rand() * $update_interval);
print "Random upload offset is at $random_offset s from $update_interval s interval\n";
my $end_time = $start_time + $runtime;

while ($now < $end_time) {
    alarm($timeout);
    $now = time();
    print "Checking for new files at ".localtime()."; last old is ".localtime($last_old_upload)."; running until ".localtime($end_time)."\n";
    &check_new();

    my $next_check = int(($now - $random_offset) / $update_interval) * $update_interval + $update_interval + $random_offset;
    die "logic error: $next_check > $now" unless $next_check > $now;
    alarm(0);
    sleep($next_check - $now);
}
print "Exiting due to runtime exceeded\n";
exit(0);

sub check_new {
    opendir(BGPDIR, ".") || die "/logs/bgp: opendir: $!";
    my $best_timestamp;
    my $best_filename;
    my @removedumps;
    if ( -f "/tmp/bgp-routes" && ! -f "/tmp/bgp-routes.work" ) {
	# hack to include /tmp/bgp-routes in the list. we can remove this once all
	# the quaggas are updated to log to /logs/bgp
	rename("/tmp/bgp-routes", "/tmp/bgp-routes.work");
    }
    while (my $fn = readdir(BGPDIR)) {
	if ($fn eq '.') {
	    next unless -f "/tmp/bgp-routes.work";
	    # hack to include /tmp/bgp-routes in the list. we can remove this once all
	    # the quaggas are updated to log to /logs/bgp
	    $fn = '/tmp/bgp-routes.work';
	} else {
	    next unless $fn =~ /^bgp-routes\./;
	}
	my @st = stat $fn;
	if (!@st) {
	    die "stat $fn: $!";
	}
	my $mtime = $st[9];
	print "$fn has mtime ".localtime($mtime)." (".($now - $mtime)." seconds ago)\n";
	if (($mtime + 10) > $now) { # assume bgpd is still writing to it
	    print "Skipping $fn due to too new\n";
	    next;
	}
	push(@removedumps, $fn);
	
	if (!defined($best_timestamp) || ($mtime > $best_timestamp)) {
	    $best_timestamp = $mtime;
	    $best_filename = $fn;
	}
    }
    closedir(BGPDIR);
    if ($best_filename) {
	&maybe_upload($best_filename, $best_timestamp);
    }
    for my $fn (@removedumps) {
	print "Removing $fn\n";
	unlink($fn);
    }
}

sub maybe_upload {
    my($filename) = @_;

    my @st = stat($filename);
    die "$filename: stat: $!" unless @st;
    my $size = $st[7];

    if ($submitted_size{$size}) {
	print "Already submitted $filename size $size\n";
	return;
    }

    print "Haven't yet uploaded $filename size $size\n";

    &upload($filename, "loriscfg.llnw.net", "gzip");
    if (!defined($last_old_upload) || $last_old_upload < ($now - $old_upload_interval)) {
	&upload($filename, "cdn-edge-config.llnw.net");
	$last_old_upload = $now;
	undef %submitted_size;
    }
    $submitted_size{$size} = 1;
}


sub upload {
    my ($filename, $server, $encoding) = @_;

    print "Uploading $filename to $server\n";

    if (!open(BGP, $filename)) {
	warn "$filename: $!";
	exit(0);
    }
    my @st = stat(BGP);
    if (!@st) {
	warn "fstat $filename: $!";
	exit(0);
    }
    my $size = $st[7];
    if (!$size) {
	warn "bgp-routes: blank file $filename";
	return
    }
    
    if ($encoding) {
	print "(gzipping)\n";
	system("gzip -9 < $filename > /tmp/bgp-routes.gz");
	open(BGP,"/tmp/bgp-routes.gz") || die "bgp-routes.gz: $!";
	unlink("/tmp/bgp-routes.gz");
	my @st = stat(BGP);
	$size = $st[7];
    }
    if (!socket(S, AF_INET, SOCK_STREAM, 0)) {
	die "socket: $!";
    }
    my $ip = inet_aton($server);
    if (!$ip) {
	warn "server: $!";
	exit(1);
    }
    my $sin = sockaddr_in(80,$ip);
    print "Connecting to $server\n";
    if (!connect(S, $sin)) {
	warn "connect: $!";
	exit(1);
    }
    print "Posting route data...\n";
    print S "POST $route_uri HTTP/1.0\r\nHost: $server\r\nContent-length: $size\r\n";
    if ($encoding) {
	print S "Content-Encoding: $encoding\r\n";
    }
    print S "\r\n";
    for (;;) {
	my $buf;
	if (read(BGP, $buf, 8192) <= 0) {
	    last;
	}
	print S $buf;
    }
    S->flush();
    shutdown(S,1);
    print "Waiting for HTTP reply...\n";
    while (<S>) {
	print;
    }
    close(S);
    print "Done uploading to $server\n";
}

