#!/usr/bin/env bash
PATH="/sbin:/bin:/usr/bin:/usr/sbin:/usr/local/sbin:/usr/local/bin"
shopt -s nocaseglob
shopt -s nocasematch

#TODO:
#1 - Look into re-order logic - must use disk labels to be trustworthy
#2 - Deal with multiple partitions

usage(){
  echo "usage: $0 -[asdfN] [options]"
  echo "  -a      Check the health of All detected drives"
  echo "  -s      Check the health of Specified drive"
  echo "  -d      Dynamic (check for slow op errors in cache1.log for all detected disks)"
  echo "  -f      Forcefully open a ticket for a drive"
  echo "  -N      Check for New drives in the chassis, and put them in service"
  echo "Example:"
  echo "  $0 -a"
  echo "  $0 -ad"
  echo "  $0 -aN"
  echo "  $0 -s da2"
  echo "  $0 -f -s da2"
  echo "  $0 -s da2 -d"
  exit 1
}

while getopts ":as:fNd" opt; do
  case ${opt} in
    a)
      all="1"
      ;;
    d)
      dynamic="1"
      ;;
    s)
      specific_drive=${OPTARG}
      ;;
    f)
      force_issue="1"
      ;;
    N)
      new_drives="1"
      ;;
    *)
      usage
      ;;
  esac
done
shift $(($OPTIND - 1))

##FILES##
lockfile="/tmp/drive-test.lock"
slow_ops_log="/tmp/slowopcache1.log"
drive_test_mail="/tmp/drive-test.mail"
bad_drive_file="/etc/bad_drives"
drive_map="/etc/drive_map"

error_exit(){
  echo "${0}: ${1:-""}" 1>&2
  exit 1
}

if [ -f ${lockfile} ]; then
  echo "${lockfile} exists, tra, tra, tra, try again later!"
  exit 1
fi

touch ${lockfile} ${slow_ops_log} ${drive_test_mail} ${bad_drive_file} || error_exit "Error on Line Line $LINENO"
trap "rm -f ${lockfile} ${slow_ops_log} ${drive_test_mail} ; exit 13" TERM INT EXIT


if [ "${1}" = health ] || [ "${1}" = healthcron ]; then
  all=1
fi

get_system_info(){
  OS=`uname -s`
  free_space=`df / | awk '/\/d/ {print $4}'`
  hostname=`hostname`
  pop=`echo $hostname | cut -d . -f 2`
  date=`date +"%Y/%m/%d"`

  if [[ ! "${hostname}" =~ ^cds|^fcds|^ics ]]; then
    error_exit "Only safe to run on cds, fcds, and ics machines"
  fi

  #Check version of bash, I guess our linux build is really old.
  if ! bash --version | egrep -q "version 4"; then
    error_exit "Bash is too old, needs version 4 or newer"
  fi


  if [ "${specific_drive}" ]; then
    mother_board="null"
  else
    mother_board=`dmidecode -s baseboard-product-name | egrep -v ^# | grep -iv invalid`
  fi

  if [[ "${OS}" =~ freebsd ]]; then
    OS_NAME="freebsd"
  elif [[ "${OS}" =~ linux ]]; then
    OS_NAME="linux"
  else
    error_exit "Unable to determine Operating System: Line $LINENO"
  fi

  if sysctl dev.mrsas.0.%desc; then
    is_mrsas=true
  else
    is_mrsas=false
  fi
}

get_disks(){
  declare -gA detected_disks
  declare -ga all_disks

  #Get disks detected by the kernel and build array
  if [ "${OS_NAME}" = freebsd ]; then
    for disk in `sysctl -n kern.disks|egrep "da[0-9]|ad[0-9]"` ; do
      detected_disks[${disk}]=1
    done
  elif [ "${OS_NAME}" = linux ]; then
    for disk in `sg_map -sd | awk -F '/' '{print $NF}'` ; do
      detected_disks[${disk}]=1
    done
  fi

  #Statically defined disks for Freebsd per motherboard type, because
  #we don't know for sure how many disks should be populated
  #drive_count = last numbered disk
  if [ "${OS_NAME}" = "freebsd" ]; then
    if [ "${mother_board}" = "null" ]; then
        all_disks=( ${specific_drive} )
        return
    elif [[ "${mother_board}" =~ X7DVL ]]; then
      drive_count=3
    elif [[ "${mother_board}" =~ X8DTL|03XKDV ]]; then
      drive_count=7
    elif [[ "${mother_board}" =~ X8DTT|X9DRT ]]; then
      drive_count=11
    elif [[ "${mother_board}" =~ X10SRW|R430|0CN7X8|X11SPW-TF ]]; then
      drive_count=9
    elif [[ "${mother_board}" =~ X10DRH-iT ]]; then
      drive_count=35
    elif [[ "${mother_board}" =~ X10DSC+ ]]; then
      drive_count=44
    elif [[ "${mother_board}" =~ 0HFG24 && "${is_mrsas}" == true ]]; then
      drive_count=7
    elif [[ "${mother_board}" =~ 0HFG24 && "${is_mrsas}" == false ]]; then
      drive_count=9 
    elif [[ "${mother_board}" =~ 06XDD1 && "${is_mrsas}" == true ]]; then
      drive_count=9
    else
      error_exit "Motherboard type not recognized, please contact platform@llnw.com to check this machine out: Line $LINENO"
    fi

    echo -en '\E[030;36m'"Motherboard ${mother_board} ending drive: ${drive_count}\033[0m\n"

    local count=0
    while [ $count -le $drive_count ]; do
      if [ ${detected_disks[ada${count}]} ]; then
        dsk_pfx=ada
        all_disks+=( "${dsk_pfx}${count}" )
      fi
      if [ ${detected_disks[da${count}]} ]; then
        dsk_pfx=da
        all_disks+=( "${dsk_pfx}${count}" )
      fi
      if [ ! $dsk_pfx ]; then
        miss_disk+=( "${count}" )
      fi
      unset dsk_pfx
      (( count++ ))
    done

    for mdisk in ${miss_disk[@]}; do
      #We have no idea what the disk prefix should be, check fstab for hints
      if `egrep -iq "dev\/ada${mdisk}[a-z]" /etc/fstab`; then
        dsk_pfx=ada
        all_disks+=( "${dsk_pfx}${mdisk}" )
      fi
      if `egrep -iq "dev\/da${mdisk}[a-z]" /etc/fstab`; then
        dsk_pfx=da
        all_disks+=( "${dsk_pfx}${mdisk}" )
      fi
      if [ ! $dsk_pfx ]; then
        #unknown disk prefix at this time, just skip it
        echo -en '\E[030;31m'"skipping disk ${mdisk}: disk not detected, and no entry in /etc/fstab\033[0m\n"
      fi
      unset dsk_pfx
    done
  fi

  #Statically defined disks for linux per motherboard type
  if [ "${OS_NAME}" = "linux" ]; then
    if [ "${mother_board}" = "null" ]; then
      all_disks=( ${specific_drive} )
      return
    elif [[ "${mother_board}" =~ X6D|X7DVL ]]; then
      all_disks=( sda sdb sdc sdd )
    elif [[ "${mother_board}" =~ X8DTL|03XKDV ]]; then
      all_disks=( sda sdb sdc sdd sde sdf sdg sdh )
    elif [[ "${mother_board}" =~ X8DTT|X9DRT ]]; then
      all_disks=( sda sdb sdc sdd sde sdf sdg sdh sdi sdj sdk sdl )
    elif [[ "${mother_board}" =~ X10SRW|0HFG24|R430|0CN7X8 ]]; then
      all_disks=( sda sdb sdc sdd sde sdf sdg sdh sdi sdj )
    elif [[ "${mother_board}" =~ X10DRH-iT ]]; then
      all_disks=( sda sdb sdc sdd sde sdf sdg sdh sdi sdj sdk sdl sdm sdn sdo sdp sdq sdr sds sdt sdu sdv sdw sdx sdy sdz
      sdaa sdab sdac sdad sdae sdaf sdag sdah sdai sdaj sdak sdal )
    elif [[ "${mother_board}" =~ 0HFG24 && "${is_mrsas}" == true ]]; then
      all_disks=( sda sdb sdc sdd sde sdf sdg sdh )
    elif [[ "${mother_board}" =~ 0HFG24 && "${is_mrsas}" == false ]]; then
      all_disks=( sda sdb sdc sdd sde sdf sdg sdh sdi sdj )
    elif [[ "${mother_board}" =~ 06XDD1 && "${is_mrsas}" == true ]]; then
      all_disks=( sda sdb sdc sdd sde sdf sdg sdh sdi sdj )
    elif [[ "${mother_board}" =~ X7DCL ]]; then
      drive_count=`sg_map -x | egrep -c "sd[a-z]"`
      if [ "${drive_count}" -gt 4 ]; then
        all_disks=( sda sdb sdc sdd sde sdf sdg sdh )
      else
        all_disks=( sda sdb sdc sdd )
      fi
    else
      error_exit "Motherboard type not recognized, please contact platform@llnw.com to check this machine out: Line $LINENO"
    fi
  fi
}

get_new_mount_point(){
  #$disk should get passed here, if disk is da10, mount point should be /d10
  if [ "${OS_NAME}" = freebsd ]; then
    if [[ ${1} =~ ^ada[0-9] ]]; then
      local mnt_pnt="d${1:3}"
    elif [[ ${1} =~ ^da[0-9]|^ad[0-9] ]]; then
      local mnt_pnt="d${1:2}"
    fi

    if df | egrep -q "${1}[a-z][0-9]"; then
      disk_info[${disk}_mount]="mount_in_use"
    else
      disk_info[${disk}_mount]=${mnt_pnt}
    fi

    return
  fi

  #need to create mapping of letters to numbers for linux, ugh
  if [ "${OS_NAME}" = linux ]; then
    if [[ ${1} =~ ^sd[a-z] ]]; then
      local drive_letter="${1:2}"
      local counter=1
      local alpha=( a b c d e f g h i j k l m n o p q r s t u v w x y z aa ab ac ad
      ae af ag ah ai aj ak al am an ao ap aq ar as at au av aw ax ay az )

      for letter in ${alpha[@]}; do
        if [ "${drive_letter}" = ${letter} ]; then
          break
        else
          (( counter ++ ))
        fi
      done
    fi

    if df | egrep -q "${1}[0-9]"; then
      disk_info[${disk}_mount]="mount_in_use"
    else
      disk_info[${disk}_mount]="d${counter}"
    fi
  fi
} 

get_bsd_disk_slice(){
  #look at gpart status to see if this disk has partition scheme.  If it does then it had a filesystem
  #at some point.  We want new disks to automatically be put into service, but not old disks.
  gpart_status=`gpart status -g | grep -ic ${1}`

  if [ "$gpart_status" -gt 0 ]; then
    gpart_slice=`gpart status ${1}| awk "/^${1}/"'{print $1}'`
  else
    slice="not_configured"
    return
  fi

  if [[ "${gpart_slice}" =~ [ad|ada|da][0-9]*s1 ]]; then
    slice="s1e"
  elif [[ ${gpart_slice} =~ [ad|ada|da][0-9]*p1 ]]; then
    slice="p1"
  fi

  if [ ! ${slice} ]; then
    error_exit "Something went wrong when trying to detect the disk slice $LINENO"
  fi
}

error_check(){
  #Bash bug with -x, verify smart is here and executable
  smartctl -h > /dev/null || error_exit "smartctl not installed or executable: Line $LINENO"

  #Check for raid devices, currently not supported
  raid_devices=`ls /dev | egrep -c "twa[as]*"`
  if [ "${raid_devices}" != 0 ]; then
    error_exit "$0 does not work on devices with raid cards: Line $LINENO"
  fi

  if [ "${#all_disks[@]}" -le 0 ]; then
    error_exit "Unable to detect the correct number of disks: Line $LINENO"
  fi

  if [ ! "${specific_drive}" ] && [ -z "${mother_board}" ]; then
    error_exit "Unable to detect Mother Board type: Line $LINENO"
  fi

  if [ ! "${specific_drive}" ] && [ "${force_issue}" ]; then
    error_exit "-f can only be specified with a specific drive, not all: Line $LINENO"
  fi

  if [ "${all}" ] && [ "${specific_drive}" ]; then
    error_exit "Can not specify -a and -s at the same time: Line $LINENO"
  fi

  if [ "${free_space}" -lt 10000 ]; then
    error_exit "Less than 10MB of space on root partition, clean up / and try again: Line $LINENO"
  fi

  if [ "${new_drives}" ] && [ ! "${all}" ]; then
    error_exit "Must specify -a when using -N: Line $LINENO"
  fi

  if [ ! "${all}" ] && [ ! "${specific_drive}" ]; then
    error_exit "Must specify -a or -s to run against a drive(s): Line $LINENO"
  fi
}

remove_old_motd(){
  #If we install a new disk, lets check and clear out any old existing 0x0000 entries so that
  #if the new disk goes missing, we take action.
  local old_motd=`egrep -ic "${1} .* 0x0000" /etc/motd`
  local old_bad=`egrep -ic "0x0000 ${1}" ${bad_drive_file}`

  if [ "${old_motd}" -gt 0 ] || [ "${old_bad}" -gt 0 ]; then
    sed -i '' "/${1} .* 0x0000/d" /etc/motd
    sed -i '' "/0x0000 \/dev\/${1}/d" ${bad_drive_file}
  fi
}

get_mounts(){
  declare -gA gpart_kv
  declare -gA gpart_fs

  #filesystem to drive maps from gpart
  while read fs status drive; do
    gpart_kv[$drive]=$fs
    gpart_fs[$fs]=$drive
  done< <(gpart status -s)

  #come back to this, may want to loop the mounts instead of taking first two
  for disk in ${all_disks[@]}; do
    if [ ! "${gpart_kv[$disk]}" ]; then
      disk_info[${disk}_mount]="not_mounted"
      continue
    fi

    local mounts=( `df -m | awk -F '/' "/^\/dev\/${gpart_kv[$disk]}/"'{print $NF}'` )
    if [ ${#mounts[@]} -le 0 ]; then
      disk_info[${disk}_mount]="not_mounted"
    elif [ ${#mounts[@]} -gt 1 ]; then
      disk_info[${disk}_mount]="${mounts[0]},${mounts[1]}"
    else
      disk_info[${disk}_mount]=${mounts[0]}
    fi
    unset mounts
  done
}

disk_label_lookup(){
  while read name blah fsys; do
    if [[ $name =~ $1 ]]; then
      drivefs="$fsys"
    fi
  done< <(geom label status) || error_exit "Failed to lookup disk label"
}

get_boot_drive(){
  while read fs blocks used avail cap mount; do
    if [[ $fs =~ a?d[a-z]?[0-9]+ ]]; then
      local split_fs=( ${fs//\// } )
      local boot_fs=${split_fs[1]}
      gbfs=${gpart_fs[$boot_fs]}
      if [ ${gpart_fs[$gbfs]} ]; then
        disk_info[${gpart_fs[$gbfs]}_role]="boot_drive"
      else
        disk_info[${gpart_fs[$boot_fs]}_role]="boot_drive"
      fi
    elif [[ $fs =~ rootfs ]]; then
      disk_label_lookup rootfs
      if [[ ${gpart_fs[$drivefs]} =~ mirror ]]; then
        while read gmfs sync drive status; do
          if [ "${gpart_fs[$drivefs]}" = $gmfs ]; then
            disk_info[${drive}_role]="boot_drive"
          fi
        done< <(gmirror status -s) || error_exit "Failed to exec 'gmirror status -s'"
      else
        disk_info[${gpart_fs[$drivefs]}_role]="boot_drive"
      fi
    fi
  done< <( df /)
}

#Get the health of the drives and sort into good and bad arrays
#Check to make sure that the disk we want to query is even detected.
health_check_disks(){
  declare -gA good_disks
  declare -gA bad_disks
  declare -gA missing_disks
  declare -gA disk_info

  get_mounts
  get_boot_drive

  echo -en '\E[030;36m'"Checking health of disks ${all_disks[@]}\033[0m\n"

  for disk in ${all_disks[@]} ; do
    if [ "${OS_NAME}" = freebsd ]; then
      if [ "${disk_info[${disk}_role]}" != "boot_drive" ]; then
        disk_info[${disk}_role]="cache_drive"
      fi
    fi

    if [ "${OS_NAME}" = linux ]; then
      if `df -m / | egrep -q "^/dev/${disk}[0-9]"`; then
        disk_info[${disk}_mount]=`df | awk -F '/' "/^\/dev\/${disk}[0-9]/"'{print $NF}'`
        disk_info[${disk}_role]=boot_drive
      else
        disk_info[${disk}_mount]=`df | awk -F '/' "/^\/dev\/${disk}[0-9]/"'{print $NF}'`
        disk_info[${disk}_role]="cache_drive"
      fi
      if [ ! "${disk_info[${disk}_mount]}" ]; then
        disk_info[${disk}_mount]="not_mounted"
      fi
    fi

    #Check to see if detected disks is in all disks so we can put missing disks into bad_disk array
    if [ ! "${detected_disks[${disk}]}" ]; then
      missing_disks[${disk}]=${disk_info[${disk}_mount]}
      disk_info[${disk}_serial]=0x0000
      disk_info[${disk}_health]=missing
      disk_info[${disk}_type]=missing
      disk_info[${disk}_model]=missing
      disk_info[${disk}_firmware]=missing
      echo -en '\E[030;31m'"${disk} is missing\033[0m\n"
      continue
    fi

    while read line; do
      if [[ "${line}" =~ "Vendor" ]]; then
        disk_info[${disk}_brand]=`echo ${line} | awk '{print $NF}'`
      elif [[ "${line}" =~ "Model Family" ]]; then
        disk_info[${disk}_brand]=`echo ${line} | awk '{print $3}'`
      elif [[ "${line}" =~ "Device Model"|"Product:" ]]; then
        disk_info[${disk}_model]=`echo ${line} | awk '{print $NF}'`
        if [ ! "${disk_info[${disk}_brand]}" ]; then
          disk_info[${disk}_brand]=`echo ${line} | awk '{print $(NF-1)}'`
        fi
      elif [[ "${line}" =~ "DELL" ]]; then
        disk_info[${disk}_manufacturer]="DELL"
        disk_info[${disk}_brand]="DELL"
      elif [[ "${line}" =~ Serial ]]; then
        disk_info[${disk}_serial]=`echo ${line} | awk '{print $NF}'`
      elif [[ "${line}" =~ rpm ]]; then
        disk_info[${disk}_type]=HDD
      elif [[ "${line}" =~ "Solid State" ]]; then
        disk_info[${disk}_type]=SSD
      elif [[ "${line}" =~ Firmware ]]; then
        disk_info[${disk}_firmware]=`echo ${line} | awk '{print $NF}'`
      fi
    done < <(smartctl -i /dev/${disk})

    if [ ! "${disk_info[${disk}_model]}" ]; then
      smartctl -i /dev/${disk} > /dev/null
      if [ $? != 0 ]; then
        disk_info[${disk}_model]=unknown
        disk_info[${disk}_serial]=0x0000
        disk_info[${disk}_type]=unknown
        disk_info[${disk}_health]=false
        disk_info[${disk}_firmware]=unknown
        bad_disks[${disk}]=${disk_info[${disk}_mount]}
        continue
      fi
    fi

    if [ ! "${disk_info[${disk}_serial]}" ]; then
      disk_info[${disk}_serial]=0x0000
    fi

    if [ ! "${disk_info[${disk}_type]}" ]; then
      disk_info[${disk}_type]=unknown
    fi

    if [ ! "${disk_info[${disk}_firmware]}" ]; then
      disk_info[${disk}_firmware]=unknown
    fi

    #If -f is specified put in bad_disk array regardless, skip rest of checks
    if [ "${force_issue}" ]; then
      echo -en '\E[030;33m'"Force Option used, skipping health checks\033[0m\n"
      bad_disks[${disk}]=${disk_info[${disk}_mount]}
      disk_info[${disk}_health]=false
      continue
    fi

    health_status=`smartctl -H /dev/${disk} | grep -i health` 
    ready_status=`smartctl -i /dev/${disk} | grep -ic "not ready"`
    smart_unsupported=`smartctl -i /dev/${disk} | grep "SMART support is: Unavailable"`

    if [ -n "${smart_unsupported}" ]; then
      echo "Smartctl isn't supported for ${disk} since it's likely serving as the boot mirror on a Marvell RAID Controller"
      disk_info[${disk}_health]=true
    elif [ "${ready_status}" = 0 ] && [[ "${health_status}" =~ ok|passed ]]; then
      disk_info[${disk}_health]=true
      #check serial against motd to see if someone forced a ticket open on it
      if `egrep -q "${disk} .* ${disk_info[${disk}_serial]}" /etc/motd` ; then
        if [ "${disk_info[${disk}_mount]}" = not_mounted ]; then
          echo -en '\E[030;33m'"${disk} ${disk_info[${disk}_serial]} is in /etc/motd and not mounted, ignoring...\033[0m\n"
          continue
        else
          echo -en '\E[030;33m'"${disk} ${disk_info[${disk}_serial]} is in /etc/motd but still mounted\033[0m\n"
          bad_disks[${disk}]=${disk_info[${disk}_mount]}
          continue
        fi
      fi
      good_disks[${disk}]=${disk_info[${disk}_mount]}
      echo -en '\E[030;32m'"${disk} ${health_status}\033[0m\n"
    else
      disk_info[${disk}_health]=false
      bad_disks[${disk}]=${disk_info[${disk}_mount]}
      echo -en '\E[030;31m'"${disk} ${health_status}\033[0m\n"
    fi
  done
}

micron_wear_levels(){
while read line; do
  if [[ "${line}" =~ ^202 ]]; then
    ssd_wear_level=${line: -3}
    if [ "${max_wear_level[$1]}" -gt 100 ]; then
      echo -e Mis-Matched wear level value for ${1} SMART Percent_Lifetime_Remain\
      should be a max value of 100, but ${1} has a max value of ${max_wear_level[$1]}
    fi

    if [ "${ssd_wear_level}" -ge "${max_wear_level[$1]}" ]; then
      ssd_lifetime[$1]=$(( (( ssd_wear_level * 100 ) / ${max_wear_level[$1]} ) ))
      echo "${1} wear level is above threshold ${ssd_wear_level}/${max_wear_level[$1]}"
      break
    else
      ssd_lifetime[$1]=$(( (( ssd_wear_level * 100 ) / ${max_wear_level[$1]} ) ))
      echo "${ssd_drive} current wear level ${ssd_wear_level}/${max_wear_level[$1]} Percent_used: ${ssd_lifetime[$1]}"
      break
    fi
  fi
done < <(smartctl -A /dev/${1})
}

dell_wear_levels(){
while read line; do
  if [[ "${line}" =~ ^245 ]]; then
    if [ ${disk_maker[$1]} = "samsung" ]; then
      max_wear_level[$1]=5
      ssd_life_remaining=${line: -3}
      if [ "${ssd_life_remaining}" -le "${max_wear_level[$1]}" ]; then
        ssd_lifetime[$1]=$(( 100 - ${ssd_life_remaining} ))
        echo "${1} wear level is above threshold, remaining life:${ssd_life_remaining}%"
        break
      else
        ssd_lifetime[$1]=$(( 100 - ${ssd_life_remaining} ))
        echo "${1} Percent_used: ${ssd_lifetime[$1]} Life_remaining:${ssd_life_remaining}/100"
        break
      fi
    elif [ ${disk_maker[$1]} = "intel" ]; then
      ssd_wear_level=${line: -3}
      if [ "${max_wear_level[$1]}" -gt 100 ]; then
        echo -e Mis-Matched wear level value for ${1} SMART Percent_Lifetime_Remain\
        should be a max value of 100, but ${1} has a max value of ${max_wear_level[$1]}
      fi

      if [ "${ssd_wear_level}" -ge "${max_wear_level[$1]}" ]; then
        ssd_lifetime[$1]=$(( (( ssd_wear_level * 100 ) / ${max_wear_level[$1]} ) ))
        echo "${1} wear level is above threshold ${ssd_wear_level}/${max_wear_level[$1]}"
        break
      else
        ssd_lifetime[$1]=$(( (( ssd_wear_level * 100 ) / ${max_wear_level[$1]} ) ))
        echo "${1} current wear level ${ssd_wear_level}/${max_wear_level[$1]} Percent_used: ${ssd_lifetime[$1]}"
        break
      fi
    fi
  fi
done < <(smartctl -A /dev/${1})
}

samsung_wear_levels(){
while read line; do
  if [[ "${line}" =~ "Wear_Leveling_Count" ]]; then
    ssd_wear_level=${line:${#line}<5?0:-5}
    if [ "${ssd_wear_level}" -ge "${max_wear_level[$1]}" ]; then
      echo "${1} wear level is above threshold ${ssd_wear_level}/${max_wear_level[$1]}"
      ssd_lifetime[$1]=$(( (( ssd_wear_level * 100 ) / ${max_wear_level[$1]} ) ))
      break
    else
      ssd_lifetime[$1]=$(( (( ssd_wear_level * 100 ) / ${max_wear_level[$1]} ) ))
      echo "${1} current wear level ${ssd_wear_level}/${max_wear_level[$1]} Percent_used: ${ssd_lifetime[$1]}"
      break
    fi
  fi
done < <(smartctl -A /dev/${1})
}

get_ssd_lifetime(){
#smartctl -a is costly, run a -i first to see which disks are
#ssd's, then run a more informational one if we need to.  Each
#Vendor has different fields and calculates wear level differently
#so we need to hard code what to do with each type.  Currently there
#is no hard action taken here.  TBW will vary depending on the workload
#But when a drive exhausts its nand, it will go into a read only state,
#which will be reported by the controller as "Failed".  The drive will
#be pulled from service at that time as it sits now.

#exit this fuction if no disks are in good array like if we use the force option
if [ "${#good_disks[@]}" -eq 0 ]; then
  return
fi

echo
echo -en '\E[030;36m'"Checking SSD Disk Wear Levels ${!good_disks[@]}\033[0m\n"

declare -gA max_wear_level
declare -gA disk_maker
declare -gA ssd_lifetime

for disk in ${!good_disks[@]} ; do
  if [ "${disk_info[${disk}_type]}" != SSD ]; then
    echo "${disk} is not an SSD"
    continue
  fi

  case ${disk_info[${disk}_model]} in
    *MZ7WD480*)
      #Samsung SM843Tn
      max_wear_level[$disk]=3958
      disk_maker[$disk]=samsung
    ;;
    *MZ7KM480*)
      #Samsung SM863a
      max_wear_level[$disk]=6416
      disk_maker[$disk]=samsung
    ;;
    *MZ7[GL][ME]480H[MC]*)
      #Samsung PM853Tn
      max_wear_level[$disk]=2400
      disk_maker[$disk]=samsung
    ;;
    *MZ7KM960H[AM][HJ]P*)
      #Samsung SM863 960GB
      max_wear_level[$disk]=6416
      disk_maker[$disk]=samsung
    ;;
    *MZ7LM960HCHP*)
      #Samsung PM863 960GB
      max_wear_level[$disk]=1387
      disk_maker[$disk]=samsung
    ;;
    *MZ7LM1T9HCJM*)
      #Samsung PM863 1.9TB
      max_wear_level[$disk]=1387
      disk_maker[$disk]=samsung
    ;;
    *M5[01]0DC*)
      #Micron M500DC||M510DC
      max_wear_level[$disk]=95
      disk_maker[$disk]=micron
    ;;
    *Micron_5100*)
      #Micron_5100
      max_wear_level[$disk]=95
      disk_maker[$disk]=micron
    ;;
    *MTFDDAK*)
      #Micron_5100 / 5200
      max_wear_level[$disk]=95
      disk_maker[$disk]=micron
    ;;
    *SSDSC2BX*)
      #INTEL s3610 400|800|1200|1600
      max_wear_level[$disk]=95
      disk_maker[$disk]=intel
    ;;
  esac

  if [ ! "${max_wear_level[$disk]}" ] && [ "${disk_info[${disk}_type]}" = SSD ]; then
    echo "${disk} is unknown SSD model.  Update drive-test to support ${disk_info[${disk}_model]}"
  fi
done

#Exit this function if no ssds are found
if [ "${#max_wear_level[@]}" -eq 0 ]; then
  return
fi

for ssd_drive in ${!max_wear_level[@]}; do
  if [ "${disk_info[${ssd_drive}_manufacturer]}" ] && [ "${disk_info[${ssd_drive}_manufacturer]}" = "DELL" ]; then
    dell_wear_levels $ssd_drive
    continue
  fi

  if [ "${disk_maker[$ssd_drive]}" = "micron" ]; then
    micron_wear_levels $ssd_drive
  elif [ "${disk_maker[$ssd_drive]}" = "samsung" ]; then
    samsung_wear_levels $ssd_drive
  fi
done
}

build_json_drive_map(){
#Don't build drive map when only one drive is specified
#lifetime value is the percent of an SSD that has been used

if [ "${specific_drive}" ]; then
  return
fi

array_index=${#all_disks[@]}
counter=1

printf "{
  \"drives\": {\n" > ${drive_map}

for disk in ${all_disks[@]}; do

cat << EOF >> /etc/drive_map
    "${disk}": {
        "mounted"   : "${disk_info[${disk}_mount]}",
        "role"      : "${disk_info[${disk}_role]}",
        "healthy"   : "${disk_info[${disk}_health]}",
        "brand"     : "${disk_info[${disk}_brand]}",
        "model"     : "${disk_info[${disk}_model]}",
        "type"      : "${disk_info[${disk}_type]}",
        "lifetime"  : "${ssd_lifetime[${disk}]}",
        "firmware"  : "${disk_info[${disk}_firmware]}",
EOF

if [ "${array_index}" -eq "${counter}" ]; then
  printf "\t\"serial\"    : \"${disk_info[${disk}_serial]}\" }\n" >> ${drive_map}
  printf "  }\n}\n" >> ${drive_map}
else
  printf "\t\"serial\"    : \"${disk_info[${disk}_serial]}\" },\n" >> ${drive_map}
  (( counter++ ))
fi

done
}


check_slow_ops_reassigns(){
  #Need to actually do something with this data, currently it's just informational
  #so you can forcefully open a ticket if necessary.  Next step is to set a threshold
  #for how many bad sectors before opening a ticket.
  if [ "${dynamic}" ]; then
    echo
    echo -en '\E[030;36m'"Checking drives ${all_disks[@]} for slow ops in Cache1.log\033[0m\n"
    if [ -f /logs/squid/cache1.log ]; then
      tail -n 50000 /logs/squid/cache1.log | grep `date -v-1d "+%Y/%m/%d"` | egrep 'slow .* operation' > ${slow_ops_log}
    else
      echo "No cache1.log found in /logs/squid, unable to check for slow ops"
      return
    fi
  else
    return
  fi

  #If drive has more than 300 slow operations, lets check the defects or reallocated_sectors, whichever it supports.
  for disk in ${all_disks[@]} ; do
    slow_count=`egrep -ic "slow .* operation .* /${disk_info[${disk}_mount]}/" ${slow_ops_log}`
    reassigns=0

    if [ "${slow_count}" -ge 300 ]; then
      slow_op_disks=("${slow_op_disks[@]}" "$disk")
      reassigns=`smartctl -A /dev/${disk} | grep -i Reallocated_Sector | awk '{print $NF}'`
      if [ ! "${reassigns}" ]; then
        reassigns=`sg_reassign -g /dev/${disk} | awk '{print $NF}'`
      fi
      echo "${disk} Slow_Ops:${slow_count} Reassgins:${reassigns}"
    fi
  done

  #Break out if no slow ops found, else continue to loop to check gstat ops
  if [ "${#slow_op_disks[@]}" -eq 0 ]; then
    echo "No slow operations found"
    return
  fi

  #Check ops or percent busy to see if disk is just busy
  gstat -b -p -I 3s > ${slow_ops_log} || error_exit "Could not write slowops to ${slow_ops_log} - line: $LINENO"

  for disk in ${slow_op_disks[@]}; do
    disk_busy=`awk "/${disk}/"'{print $9}' ${slow_ops_log} | cut -d . -f 1`
    if [ "${disk_busy}" -gt 80 ]; then
      echo "${disk} is ${disk_busy} percent busy"
    else
      echo "${disk} is less than 80% busy, disk might be bad"
    fi
  done
}

#Need separate bsd/linux checks in here
bsd_new_drive_check(){
  if [ "${OS_NAME}" != freebsd ] || [ ! "${new_drives}" ]; then
    return
  fi

  echo
  echo -en '\E[030;36m'"Checking drives ${!good_disks[@]} to see if any are new\033[0m\n"

  for disk in ${!good_disks[@]}; do
    if [ "${disk_info[${disk}_role]}" = boot_drive ]; then
      #Don't try to do anything with boot drives
      continue
    fi

    get_bsd_disk_slice ${disk}

    #If not mounted, check for filesystem.
    if [ "${disk_info[${disk}_mount]}" = "not_mounted" ]; then
      #If slice is not configured, skip the bologne and put a FS on it.
      if [ "${slice}" = "not_configured" ]; then
        echo -en '\E[030;36m'"No Slice detected on ${disk} running cdsdisk.pl to place filesystem\033[0m\n"
        /etc/sq/cdsdisk.pl ${disk} > /dev/null || error_exit "/etc/sq/cdsdisk.pl not available or encountered an error line: $LINENO"
        remove_old_motd ${disk}
        get_new_mount_point ${disk}
      else
        echo "${disk} has a filesystem on it, or had a filesystem on it, not touching"
        continue
      fi

      get_bsd_disk_slice ${disk}

      local has_fstab=`egrep -ic "^/dev/${disk}${slice}" /etc/fstab`
      local has_commented_fstab=`egrep -ic "^#/dev/${disk}${slice}" /etc/fstab`

      #If disk has a commented fstab, uncomment, and mount it up, and set the mount point.
      if [ "${has_commented_fstab}" -eq 1 ]; then
        sed "s/^#\/dev\/"${disk}${slice}"/\/dev\/"${disk}${slice}"/g" /etc/fstab > /tmp/fstab.new || error_exit "Unable to write to /tmp/fstab Line $LINENO"
        if `mount -F /tmp/fstab.new /${disk_info[${disk}_mount]}`; then
          echo "${disk} : MOUNTED OK"
          sed -i '' "s/^#\/dev\/"${disk}${slice}"/\/dev\/"${disk}${slice}"/g" /etc/fstab || error_exit "Unable to comment ${disk} in /etc/fstab Line $LINENO"
          disk_info[${disk}_mount]=`df | egrep ^/dev/${disk}${slice} | awk -F '/' '{print $NF}'`
          needs_squid_restart=1
        else
          echo "Could not mount /${disk_info[${disk}_mount]}"
          continue
        fi
      fi

      #If no fstab entry or commented fstab entry, write out a new one.
      if [ "${has_fstab}" -eq 0 ] && [ "${has_commented_fstab}" -eq 0 ]; then
        echo -en '\E[030;36m'"Attempting to mount ${disk}${slice} to /${disk_info[${disk}_mount]}\033[0m\n"
        echo -e "/dev/${disk}${slice}\t\t/${disk_info[${disk}_mount]}\t\tufs\trw,noatime\t3\t3" > /tmp/fstab.new || error_exit "Unable to create tmp entry @ line: $LINENO"
        if `mount -F /tmp/fstab.new /${disk_info[${disk}_mount]}`; then
          echo "${disk} : MOUNTED OK"
          echo -e "/dev/${disk}${slice}\t\t/${disk_info[${disk}_mount]}\t\tufs\trw,noatime\t3\t3" >> /etc/fstab || error_exit "Unable to create fstab entry @ line: $LINENO"
          disk_info[${disk}_mount]=`df | egrep ^/dev/${disk}${slice} | awk -F '/' '{print $NF}'`
          needs_squid_restart=1
        else
          echo -en '\E[030;31m'"Could not mount /${disk_info[${disk}_mount]}\033[0m\n"
          continue
        fi
      fi

      #If there is an fstab entry, then just try to mount
      if [ "${has_fstab}" -eq 1 ]; then
        echo -en '\E[030;36m'"Attempting to mount ${disk}${slice} to /${disk_info[${disk}_mount]}\033[0m\n"
        if `mount -F /etc/fstab /${disk_info[${disk}_mount]}`; then
          echo "${disk} : MOUNTED OK"
          disk_info[${disk}_mount]=`df | egrep ^/dev/${disk}${slice} | awk -F '/' '{print $NF}'`
          needs_squid_restart=1
        fi
      fi
    fi
  done
}

check_bad_mounted(){
  #return if no bad disks have been found, missing disks should have no way
  #of being mounted, so no need to check if they are missing to the kernel
  if [ "${#bad_disks[@]}" -eq 0 ]; then
    return
  fi

  echo
  echo -en '\E[030;36m'"Checking bad drive(s) ${!bad_disks[@]} to see if they are still mounted\033[0m\n"

  for disk in ${!bad_disks[@]} ; do
    #Don't try to do anything with boot drives
    if [ "${disk_info[${disk}_role]}" = boot_drive ]; then
      echo -en '\E[030;33m'"Skipping ${disk_info[${disk}_role]} ${disk}\033[0m\n"
      continue
    fi
    #See if drive is still mounted and attempt to unmount
    if [ "${disk_info[${disk}_mount]}" != not_mounted ]; then
      drive_in_use=`lsof -t /${disk_info[${disk}_mount]}|head -n 1`
      if [ ${drive_in_use} ]; then
        echo -en '\E[030;33m'"${disk} is in use, can not unmount\033[0m\n"
      else
        echo -en '\E[030;36m'"Attempting to unmount bad drive: ${disk}\033[0m\n"
        if `umount /${disk_info[${disk}_mount]}`; then
          echo "UNMOUNT OF ${disk}: SUCCESSFUL"
          bad_disks[${disk}]="false"
        else
          echo -en '\E[030;31m'"UNMOUNT OF ${disk}: FAILED! \033[0m\n"
        fi
      fi
    else
      echo -en '\E[030;32m'"${disk} = ${disk_info[${disk}_mount]}\033[0m\n"
    fi
  done
}

comment_out_bad_drives(){
  if [ "${OS_NAME}" != freebsd ]; then
    return
  fi

  #Previously bad disks were unset as part of last function, so exit function if arrays are now empty
  if [ "${#bad_disks[@]}" -eq 0 ] && [ "${#missing_disks[@]}" -eq 0 ]; then
    return
  fi

  echo
  echo -en '\E[030;36m'"Commenting out disks ${!bad_disks[@]} ${!missing_disks[@]} from /etc/fstab and /etc/squid*conf\033[0m\n"

  for disk in ${!bad_disks[@]} ${!missing_disks[@]}; do

    if [ "${disk_info[${disk}_mount]}" != not_mounted ]; then
      touch "/${disk_info[${disk}_mount]}/.bad_drive" || mail -s "Failure to create /${disk}/.bad_drive on $hostname" platform@llnw.com < /dev/null
    fi

    if `egrep -iq "^cache_dir diskd /${disk_info[${disk}_mount]}/sqcache" /etc/squid*conf` ; then
      echo "Bad drive /${disk_info[${disk}_mount]} still being used by EdgePrism, EdgePrism will graceful at the end of this run."
      needs_squid_restart=1
    fi

    if [ "`cat /etc/sq/squid_code`" != "standard" ]; then
      sed -i '' "s/^cache_dir diskd \/"${disk_info[${disk}_mount]}"\/sqcache/\#&/" /etc/squid*.conf
      needs_squid_restart=1
    fi

    if [ "${OS_NAME}" = freebsd ]; then
      if `egrep -iq "^/dev/${disk}[a-z][0-9]" /etc/fstab`; then
        sed -ri '' "s/^\/dev\/${disk}[a-z][0-9].[[:space:]]+\/${disk_info[${disk}_mount]}/\#&/g" /etc/fstab || error_exit "Unable to comment ${disk} in /etc/fstab Line $LINENO"
        echo "COMMENTED FSTAB FOR ${disk}: OK"
      else
        echo "COMMENTED FSTAB FOR ${disk}: Already Commented"
      fi
    elif [ "${OS_NAME}" = linux ]; then
      if `egrep -iq "^/dev/${disk}[0-9]" /etc/fstab`; then
        sed -i '' "s/^\/dev\/"${disk}[0-9]"/\#&/g" /etc/fstab || error_exit "Unable to comment ${disk} in /etc/fstab Line $LINENO"
        echo "COMMENTED FSTAB FOR ${disk}: OK"
      else
        echo "COMMENTED FSTAB FOR ${disk}: Already Commented"
      fi
    fi
  done
}

restart_squid(){
  if [ "${needs_squid_restart}" ]; then
    touch /etc/sq/graceful
  fi
}

mail_ticket(){
  mail -s "Drive Replacement $hostname [Drive ${1}]" network@llnw.com < ${drive_test_mail}
  echo "${1} ticket opened"
}

motd_logs_tickets(){
  #return if no actions are to be taken
  if [ "${#bad_disks[@]}" -eq 0 ] && [ "${#missing_disks[@]}" -eq 0 ]; then
    return
  fi

  if [ -t 0 ]; then
    if [ ${force_issue} ]; then
      automated='force'
    else
      automated='manual'
    fi
  else
    automated='cron'
  fi

  for disk in ${!bad_disks[@]} ${!missing_disks[@]}; do
    local has_motd=`egrep -ic "${disk} .* ${disk_info[${disk}_serial]}" /etc/motd`
    #local has_bad_entry=`egrep -ic "${disk_info[${disk}_serial]} /dev/${disk}" ${bad_drive_file}`

    if [ "${automated}" = manual ]; then
      if [ ${has_motd} -eq 0 ]; then
        logger -t $0 -i "type=drivetest-raw $hostname drive=${disk} pop=${pop} runtype=${automated}"
        echo "[FD] Drive replacement ticket opened for ${disk} on ${date} ${disk_info[${disk}_serial]}" >> /etc/motd
        echo "${disk_info[${disk}_serial]} /dev/${disk} /${disk_info[${disk}_mount]} removed ${date}" >> ${bad_drive_file}

        echo " drive-test was executed manually by someone, and has found ${disk} to be in"> ${drive_test_mail}
        echo " a failed state, or has too many reassigns to be in service.  If this drive" >> ${drive_test_mail}
        echo " is not bad, please notify platform@llnw.com to investigate." >> ${drive_test_mail}
        echo " ${disk} has been removed/commented from EdgePrism, fstab, and has ">> ${drive_test_mail}
        echo " been unmounted. EdgePrism has been gracefully restarted without ${disk}.">> ${drive_test_mail}
        /usr/local/sbin/smartctl -a /dev/${disk} >> ${drive_test_mail}

        mail_ticket ${disk}
      fi
    elif [ "${automated}" = cron ]; then
      if [ ${has_motd} -eq 0 ]; then
        logger -t $0 -i "type=drivetest-raw $hostname drive=${disk} pop=${pop} runtype=${automated}"
        echo "[FD] Drive replacement ticket opened for ${disk} on ${date} ${disk_info[${disk}_serial]}" >> /etc/motd
        echo "${disk_info[${disk}_serial]} /dev/${disk} /${disk_info[${disk}_mount]} removed ${date}" >> ${bad_drive_file}

        echo " ${disk} was deemed bad by the nightly cron job, if this drive"> ${drive_test_mail}
        echo " is not bad, please notify or platform@llnw.com to investigate." >> ${drive_test_mail}
        echo " ${disk} has been removed/commented from EdgePrism, fstab, and has ">> ${drive_test_mail}
        echo " been unmounted. EdgePrism has been gracefully restarted without ${disk}.">> ${drive_test_mail}
        /usr/local/sbin/smartctl -a /dev/${disk} >> ${drive_test_mail}

        mail_ticket ${disk}
      fi
    elif [ "${automated}" = force ]; then
      if [ ${has_motd} -eq 0 ]; then
        logger -t $0 -i "type=drivetest-raw $hostname drive=${disk} pop=${pop} runtype=${automated}"
        echo "[FO] Drive replacement ticket opened for ${disk} on ${date} ${disk_info[${disk}_serial]}" >> /etc/motd
        echo "${disk_info[${disk}_serial]} /dev/${disk} /${disk_info[${disk}_mount]} removed ${date}" >> ${bad_drive_file}

        echo " ${disk} was manually taken out of rotation by using the"> ${drive_test_mail}
        echo " force option in drive-test.  This drive may not register as" >> ${drive_test_mail}
        echo " bad, but someone found it necessary to forcefully remove it" >> ${drive_test_mail}
        echo " from service.  I have removed the drive from EdgePrism and fstab, ">> ${drive_test_mail}
        echo " and restarted EdgePrism.">> ${drive_test_mail}
        /usr/local/sbin/smartctl -a /dev/${disk} >> ${drive_test_mail}

        mail_ticket ${disk}
      fi
    fi

  done
}

#Lets run through these Mother Fu*@%ions
get_system_info
get_disks
error_check
health_check_disks
get_ssd_lifetime
check_slow_ops_reassigns
bsd_new_drive_check
check_bad_mounted
comment_out_bad_drives
build_json_drive_map
restart_squid
motd_logs_tickets
