#!/usr/local/bin/perl -w
use DBI;
use Config::IniFiles;
use strict;
use File::Basename qw(dirname);
use Cwd qw(abs_path);
use JSON;
use URI::Escape;
my $config_dir = ".";
my $config_dbh;
my $do_csv;
my $pop;
my $dates = "";
my $test_api;
my $debug_output=0;
my %servers;
my %ent;
my $recent = 'recent';
my $old = 'old';
my $id = 'id';
my $ded_id = "dedicated_id";
my $test_id = 'enterprise';
my $workingdir = dirname abs_path $0;
my $total_tests = 0;
my $failed_tests = 0;
my $recent_days = 30;
my $limit = 50;
my $attrib_fetch_cnt = 3;

sub print_usage
{
    print "Usage: edgeinfo_api_unit.pl -p pop [-l limit] [debug] [--help]\n";
    print "  -p pop   - 3 char pop abbreviation\n";
    print "  -l limit - maximum number of servers (2-500) to test in a pop (Default 50)\n";
    print "  debug    - generate additional debug output\n";
    print "  --help   - print this screen\n";
    print "Examples:\n";
    print " perl edgeinfo_api_unit.pl -p mia -l 20 debug\n";
    print " perl edgeinfo_api_unit.pl -p dal debug debug\n";
    print "\n";
    exit 1;
}

sub parse_options()
{
    my $num_args = $#ARGV + 1;
    if ($num_args < 2 || $num_args > 6) {
        print_usage();
    }
    for (my $i = 0; $i < $num_args; ++$i) {
        if ($ARGV[$i] eq '-p' && defined $ARGV[$i + 1]) {
            $pop = $ARGV[$i + 1];
            ++$i;
            print "Pop: $pop\n";
        }elsif ($ARGV[$i] eq '-l' && defined $ARGV[$i + 1]) {
            $limit = $ARGV[$i + 1];
            if ($limit < 2) {
                $limit = 2;
            } elsif ($limit > 500) {
                $limit = 500;
            }
            ++$i;
            print "Limit: $limit\n";
        } elsif ($ARGV[$i] =~ /debug/i) {
            ++$debug_output;
            print "Generate debug output\n";
        } elsif ($ARGV[$i] eq '--help') {
            print_usage();
        }
    }
    if (!defined $pop) {
        print_usage();
    }
}

sub parse_db_config()
{
    # read the specified cfg file for server/user/pass
    my $gencfg =Config::IniFiles->new(-file => "$config_dir/eiapi_sql.cfg");
    if (!defined($gencfg)) {
        print "Configuration problem on server - unreadable config";
        exit 1;
    }
    my $dbhost    = $gencfg->val('EDGEINFO', 'Host');
    my $database  = $gencfg->val('EDGEINFO', 'Database');
    my $user      = $gencfg->val('EDGEINFO', 'User');
    my $password  = $gencfg->val('EDGEINFO', 'Password');
    my $dsn = "DBI:mysql:database=$database;host=$dbhost";
    $config_dbh = DBI->connect($dsn, $user, $password) or db_server_error $DBI::errstr;
}

sub fail
{
    print MAIL "Fatal Error terminating\n";
    close MAIL;
    exit -1;
}

# query for a set of servers in the specified pop
sub query_for_servers_in_pop
{
    my $found_data;
    my $json_data;
    my $server_regex = "^cds.*\.$pop\.";
    my $sql =
        "select distinct server from edgeinfo where server REGEXP ? order by modified desc LIMIT $limit";
    my $fieldinfo_sth =$config_dbh->prepare($sql);
    $fieldinfo_sth->bind_param(1, $server_regex);

    # bind attribute values into the in list
    $fieldinfo_sth->execute();
    while (my $row = $fieldinfo_sth->fetchrow_hashref()) {
        my $srv = $row->{server};
        if ($debug_output) {
            print "server: $row->{server}\n";
        }
        $servers{$srv} = undef;
        $servers{$srv}{$old} = undef;
        $servers{$srv}{$recent} = undef;
    }
    $servers{'bogus'} = undef;
    $servers{'bogus'}{$old} = undef;
    $servers{'bogus'}{$recent} = undef;
    if (keys %servers == 0) {
        print 'DB has no servers for POP: $pop';
        exit 1;
    }
}

# Taking the servers selected from the pop - capture the recent, old and dedicated id
sub fetch_server_details
{
    my $row;
    my $edgeinfo_recent = $config_dbh->prepare(
        "select attrib, value from edgeinfo where server = ? and modified > date_sub(now(), interval ? day) order by attrib"
    );
    my $edgeinfo_old = $config_dbh->prepare(
        "select attrib, value from edgeinfo where server = ? and modified <= date_sub(now(), interval ? day) order by attrib"
    );
    my $dedicated_sth =$config_dbh->prepare("select dedicated_id from dedicated where server = ?");
    for my $server (keys %servers) {

        # fetch recent data about the server
        $edgeinfo_recent->bind_param(1, $server);
        $edgeinfo_recent->bind_param(2, $recent_days);
        $edgeinfo_recent->execute();
        while ($row = $edgeinfo_recent->fetchrow_hashref()) {
            $servers{$server}{$recent}{$row->{attrib}} = $row->{value};
        }

        # fetch old data about the server
        $edgeinfo_old->bind_param(1, $server);
        $edgeinfo_old->bind_param(2, $recent_days);
        $edgeinfo_old->execute();
        while ($row = $edgeinfo_old->fetchrow_hashref()) {
            $servers{$server}{$old}{$row->{attrib}} = $row->{value};
        }

        # fetch dedicated id of the server
        $dedicated_sth->bind_param(1, $server);
        $dedicated_sth->execute();
        while ($row = $dedicated_sth->fetchrow_hashref()) {
            $servers{$server}{$id} = $row->{dedicated_id};
            $servers{$server}{$recent}{$ded_id} = $row->{dedicated_id};
        }
    }
}

sub query_enterprise_servers
{
    my $found_data;
    my $json_data;
    my $sql =
          "select distinct E.server from edgeinfo E, dedicated D where"
        . " E.server=D.server and D.dedicated_id = ?";
    my $fieldinfo_sth =$config_dbh->prepare($sql);
    $fieldinfo_sth->bind_param(1, $test_id);

    # bind attribute values into the in list
    $fieldinfo_sth->execute();
    while (my $row = $fieldinfo_sth->fetchrow_hashref()) {
        my $srv = $row->{server};
        if ($debug_output) {
            print "server: $row->{server}\n";
        }
        $ent{$srv} = undef;
        $ent{$srv}{$old} = undef;
        $ent{$srv}{$recent} = undef;
        $ent{$srv}{$recent}{$ded_id} = $test_id;
        $ent{$srv}{$id} = $test_id;
    }
    if (keys %ent == 0) {
        print 'DB has no servers for POP: $pop';
        exit 1;
    }
}

# Taking the servers selected from the pop - capture the recent, old and dedicated id
sub fetch_enterprise_details
{
    my $row;
    my $edgeinfo_recent = $config_dbh->prepare(
        "select E.server, E.attrib, E.value from edgeinfo E, dedicated D where E.server = D.server and D.dedicated_id = ? and E.attrib = 'numsquid1' and E.modified > date_sub(now(), interval ? day)"
    );
    my $edgeinfo_old = $config_dbh->prepare(
        "select E.server,E.attrib, E.value from edgeinfo E, dedicated D where E.server = D.server and D.dedicated_id = ? and E.attrib = 'numsquid1' and E.modified <= date_sub(now(), interval ? day)"
    );

    # fetch recent data about the server
    $edgeinfo_recent->bind_param(1, $test_id);
    $edgeinfo_recent->bind_param(2, $recent_days);
    $edgeinfo_recent->execute();
    while ($row = $edgeinfo_recent->fetchrow_hashref()) {
        $ent{$row->{server}}{$recent}{$row->{attrib}} = $row->{value};
    }

    # fetch old data about the server
    $edgeinfo_old->bind_param(1, $test_id);
    $edgeinfo_old->bind_param(2, $recent_days);
    $edgeinfo_old->execute();
    while ($row = $edgeinfo_old->fetchrow_hashref()) {
        $ent{$row->{server}}{$old}{$row->{attrib}} = $row->{value};
    }
}

# Execute commands and Check for erors
sub execute_command
{
    my ($command, $output) = @_;
    if ($debug_output) {
        print "CMD: $command\n";
    }
    $$output = `$command`;
    if ($? == -1) {
        print "failed to execute: $command\n";
    } elsif ($? & 127) {
        printf " $!\nchild died with signal %d, %s coredump\n",
            ($? & 127),  ($? & 128) ? 'with' : 'without';
    } elsif (${^CHILD_ERROR_NATIVE} != 0) {
        print "$! \nchild exited with value %d\n", ${^CHILD_ERROR_NATIVE};
    }
    return ${^CHILD_ERROR_NATIVE};
}

# fetch and test the dedicated id
sub test_dedicated_id
{
    print "Testing - Dedicated_id lookup\n";
    my $result = '';
    for my $server (keys %servers) {
        my $cmd = "$workingdir/edgeinfo_api.pl dedicated=$server";
        if ($debug_output) {
            $cmd .= ' sql_header=1';
        }
        execute_command($cmd, \$result);

        #print "Result: $result\n";
        my @array = split /\n/,  $result;
        if ($debug_output){
            for my $row (@array) {
                if ($row =~ /sql/i) {
                    print "$row\n";
                }
            }
        }
        my $data = from_json($array[$#array]);
        foreach my $row (@{$data}) {

            # print "Value: $row->{'value'} \n";
            if ($row->{'value'} ne $servers{$server}{$id}) {
                print
                    "Error Dedicated ID match fail for $server: $servers{$server}{'$id'} == $row->{'value'}\n";
                ++$failed_tests;
            }
            ++$total_tests;
        }
    }
}

# fetch and test only recently changed attributes if any
sub test_server_recent
{
    print "Testing - Recent server list lookup\n";
    my $result = '';
    for my $server (keys %servers) {
        my $cmd = "$workingdir/edgeinfo_api.pl server=$server\\&modified_days=$recent_days";
        if ($debug_output) {
            $cmd .= '\\&sql_header=1';
        }
        execute_command($cmd, \$result);
        if ($debug_output > 1) {
            print "Result: $result\n";
        }
        my @array = split /\n/,  $result;
        if ($debug_output){
            for my $row (@array) {
                if ($row =~ /sql/i) {
                    print "$row\n";
                }
            }
        }
        my $data = from_json($array[$#array]);
        my $json_size = scalar @{$data};
        foreach my $row (@{$data}) {

            # print "Value: $row->{'value'} \n";
            if ($row->{'value'} ne $servers{$server}{$recent}{$row->{'attrib'}}) {
                print
                    "Error Recent attrib ID match fail for $server|{$row->{'attrib'}:  $servers{$server}{$recent}{$row->{'attrib'}} == $row->{'value'}\n";
                ++$failed_tests;
            }
            ++$total_tests;
        }
        my $recent_size = keys %{$servers{$server}{$recent}};
        if ($recent_size == $json_size) {
            ++$total_tests;
        } else {
            print "Error attrib count missmatch - All ", $recent_size, " <> ", $json_size, "\n";
            ++$failed_tests;
        }
    }
}

# Fetch and test the all attributes are present by fetching 'old' content
sub test_server_no_date
{
    print "Testing - Old server list lookup\n";
    my $result = '';
    for my $server (keys %servers) {
        my $cmd = "$workingdir/edgeinfo_api.pl server=$server\\&old=1";
        if ($debug_output) {
            $cmd .= '\\&sql_header=1';
        }
        execute_command($cmd, \$result);
        if ($debug_output > 1) {
            print "Result: $result\n";
        }
        my @array = split /\n/,  $result;
        if ($debug_output){
            for my $row (@array) {
                if ($row =~ /sql/i) {
                    print "$row\n";
                }
            }
        }
        my $data = from_json($array[$#array]);
        my $json_size = scalar @{$data};
        foreach my $row (@{$data}) {
            my $age = $old;

            # print "Value: $row->{'value'} \n";
            if (!defined $servers{$server}{$age}{$row->{'attrib'}}) {
                $age = $recent;
            }
            if ($row->{'value'} ne $servers{$server}{$age}{$row->{'attrib'}}) {
                print
                    "Error Recent attrib ID match fail for $server Attrib: '$row->{'attrib'}':",
                    "  $servers{$server}{$age}{$row->{'attrib'}} == $row->{'value'}\n";
                ++$failed_tests;
            }
            ++$total_tests;
        }
        my $recent_size = keys %{$servers{$server}{$recent}};
        my $old_size = keys %{$servers{$server}{$old}};
        if ($recent_size + $old_size == $json_size) {
            ++$total_tests;
        } else {
            print "Error attrib count missmatch - All ", $recent_size,
                " + ", $old_size, " <> ", $json_size, "\n";
            ++$failed_tests;
        }
    }
}

# test the fetch of specific attributes for each server - generate a random list of attributes
# for each server to test variations
sub test_fetch_server_attrib
{
    my ($age,$inc_ded) = @_;
    if (defined $inc_ded) {
        print "Testing - attrib list with dedicated_id in where lookup - $age\n";
    } else {
        print "Testing - attrib list without dedicated_id in where lookup - $age\n";
    }
    my $result = '';
    for my $server (keys %servers) {
        my $size = keys %{$servers{$server}{$age}};

        #build the attribs key
        my @attrib_set;
        next if $size == 0;
        my $attrib_cnt = int(rand($size - 1)) + 1;
        my $offset = int(($size + $attrib_cnt - 1)/$attrib_cnt);
        my @akeys = keys %{$servers{$server}{$age}};
        my $next_pos = $size - 1;
        my $ded_match;
        my $cmd;
        my $conj;

        while ($next_pos >= 0) {
            push(@attrib_set, $akeys[$next_pos]);
            $next_pos -= $offset;
        }
        my $attribs = join ',', @attrib_set;
        $attribs = uri_escape($attribs);
        if ($attribs eq "") {
            print "Empty attribs -error\n";
            next;
        }
        if (defined $inc_ded) {
            if (defined $servers{$server}{$id}){
                $ded_match = "dedicated_id=$servers{$server}{$id}";
            }
        }
        if ($age eq $old) {
            $cmd = "$workingdir/edgeinfo_api.pl regex=^$server old=1";
            $conj = " ";
        } else {
            $cmd = "$workingdir/edgeinfo_api.pl regex=^$server"."\\&modified_days=$recent_days";
            $conj = '\\&';
        }
        if ($attribs ne "") {
            $cmd .= $conj . "attribs=$attribs";
        }
        if ($debug_output) {
            $cmd .= $conj . "sql_header=1";
        }
        if (defined $ded_match) {
            $cmd .= $conj . "$ded_match";
        }
        execute_command("$cmd", \$result);
        if ($debug_output > 1) {
            print "Result: $result\n";
        }
        my @array = split /\n/,  $result;
        if ($debug_output){
            for my $row (@array) {
                if ($row =~ /sql/i) {
                    print "$row\n";
                }
            }
        }
        my $data = from_json($array[$#array]);

        # remove Configured as - row as it messes with validation
        # because it is not in edgeinfo table
        my $json_size = scalar @{$data};
        foreach my $row (@{$data}) {
            if (!exists $servers{$server}{$age}{$row->{'attrib'}}) {
                print
                    "Error $age attrib not found in attrib list for $server Attrib: '$row->{'attrib'}'\n";
                ++$failed_tests;
            } elsif ($row->{'value'} ne $servers{$server}{$age}{$row->{'attrib'}}) {
                print
                    "Error $age attrib value mismatch for $server Attrib: '$row->{'attrib'}':",
                    "  $servers{$server}{$age}{$row->{'attrib'}} <> $row->{'value'}\n";
                ++$failed_tests;
            }
            ++$total_tests;
        }
        if ($json_size != @attrib_set) {
            print "Error attrib count error - Expected: ", scalar @attrib_set,
                " Fetched:", $json_size, "\n";
            ++$failed_tests;
        } else {
            ++$total_tests;
        }
    }
}

# test the fetch of specific attributes for each server - generate a random list of attributes
# for each server to test variations
sub test_fetch_server_no_attrib
{
    my ($age, $inc_ded) = @_;
    if (defined $inc_ded) {
        print "Testing - all attrib list with dedicated_id in where lookup - $age\n";
    } else {
        print "Testing - all attrib list without dedicated_id in where lookup - $age\n";
    }
    my $result = '';
    for my $server (keys %servers) {
        my $cmd;
        my $conj;
        if ($age eq $old) {
            $cmd = "$workingdir/edgeinfo_api.pl regex=^$server old=1";
            $conj = " ";
        } else {
            $cmd = "$workingdir/edgeinfo_api.pl regex=^$server\\&modified_days=$recent_days";
            $conj = '\\&';
        }
        if (defined $inc_ded && defined $servers{$server}{$id}){
            $cmd .= $conj . "dedicated_id=$servers{$server}{$id}";
        }
        if ($debug_output) {
            $cmd .= $conj . "sql_header=1";
        }
        execute_command("$cmd", \$result);
        if ($debug_output > 1) {
            print "Result: $result\n";
        }
        my @array = split /\n/,  $result;
        if ($debug_output){
            for my $row (@array) {
                if ($row =~ /sql/i) {
                    print "$row\n";
                }
            }
        }
        my $data = from_json($array[$#array]);
        my $json_size = scalar @{$data};
        foreach my $row (@{$data}) {
            my $lage = $age;

            # print "Value: $row->{'value'} \n";
            if (!defined $servers{$server}{$lage}{$row->{'attrib'}}) {
                $lage = $recent;
            }
            if (!defined $servers{$server}{$lage}{$row->{'attrib'}}) {
                print "Error $lage attrib not found in list for $server",
                    " Attrib: '$row->{'attrib'}'\n";
                ++$failed_tests;
            } elsif ($row->{'value'} ne $servers{$server}{$lage}{$row->{'attrib'}}) {
                print "Error $lage attrib mismatch for $server Attrib:",
                    " '$row->{'attrib'}': $servers{$server}{$lage}{$row->{'attrib'}} ",
                    "<> $row->{'value'}\n";
                ++$failed_tests;
            }
            ++$total_tests;
        }
        my $recent_size = keys %{$servers{$server}{$recent}};
        my $old_size = keys %{$servers{$server}{$old}};
        my $test_size = $recent_size;
        if ($age eq $old) {
            $test_size += $old_size;
        }
        if ($test_size == $json_size) {
            ++$total_tests;
        } else {
            print "Error attrib count missmatch - All $test_size <> $json_size\n";
            ++$failed_tests;
        }
    }
}

sub test_fetch_dedicated_id_search
{
    my ($age, $inc_ded) = @_;
    my $result = '';
    my $cmd;
    my $conj;
    my $attribs = 'numsquid1';
    my $recent_size = 0;
    my $old_size = 0;
    print "Testing - Query by dedicated_id for numsquid1\n";

    if ($age eq $old) {
        $cmd = "$workingdir/edgeinfo_api.pl dedicated_id=enterprise old=1";
        $conj = " ";
    } else {
        $cmd = "$workingdir/edgeinfo_api.pl dedicated_id=enterprise\\&modified_days=$recent_days";
        $conj = '\\&';
    }
    if (defined $attribs){
        if ($inc_ded){
            $attribs .= ",dedicated_id";
        }
        $attribs = uri_escape($attribs);
        $cmd .= $conj . "attribs=$attribs";
    }
    if ($debug_output) {
        $cmd .= $conj . "sql_header=1";
    }
    execute_command($cmd, \$result);
    if ($debug_output > 1) {
        print "Result: $result\n";
    }
    my @array = split /\n/,  $result;
    if ($debug_output){
        for my $row (@array) {
            if ($row =~ /sql/i) {
                print "$row\n";
            }
        }
    }
    my $data = from_json($array[$#array]);
    my $json_size = scalar @{$data};
    foreach my $row (@{$data}) {
        my $srv = $row->{'server'};
        if (!exists $ent{$srv}) {
            $srv .= ".llnw.net";
        }
        if (!exists $ent{$srv}) {
            print "Error server $row->{'server'} doesn't match\n";
            ++$failed_tests;
            next;
        }
        my $lage = $age;
        if (!defined $ent{$srv}{$lage}{$row->{'attrib'}}) {
            $lage = $recent;
        }
        if (!defined $ent{$srv}{$lage}{$row->{'attrib'}}) {
            print "Error $lage attrib not found in list for $srv"," Attrib: '$row->{'attrib'}'\n";
            ++$failed_tests;
        } elsif ($row->{'value'} ne $ent{$srv}{$lage}{$row->{'attrib'}}) {
            print "Error $lage attrib mismatch for $srv Attrib:",
                " '$row->{'attrib'}': $ent{$srv}{$lage}{$row->{'attrib'}} ",
                "<> $row->{'value'}\n";
            ++$failed_tests;
        }
        ++$total_tests;
    }
}

sub test_failures
{
}

sub main
{
    parse_options;
    parse_db_config;
    chdir($workingdir);
    query_for_servers_in_pop;
    fetch_server_details;
    query_enterprise_servers;
    fetch_enterprise_details;

    # begin testing
    test_dedicated_id;
    test_server_recent;
    test_server_no_date;
    test_fetch_server_attrib $old;
    test_fetch_server_attrib $recent;
    test_fetch_server_attrib $old, 1;
    test_fetch_server_attrib $recent, 1;
    test_fetch_server_no_attrib $old;
    test_fetch_server_no_attrib $recent;
    test_fetch_server_no_attrib $old, 1;
    test_fetch_server_no_attrib $recent, 1;
    test_fetch_dedicated_id_search $old;
    test_fetch_dedicated_id_search $recent;
    test_fetch_dedicated_id_search $old, 1;
    test_fetch_dedicated_id_search $recent, 1;
    print "Unit Test results for POP: $pop\n";
    print "  Servers tested:",  scalar keys %servers, "\n";
    print "  Total tests: $total_tests\n";

    if ($failed_tests > 0) {
        print "Failed tests: $failed_tests\n";
    } else {
        print "All Tests passed \n";
    }
}
main;
