#!/bin/bash

# Ensure that user is root before allowing script to run.
if [ $UID -ne 0 ]; then
  echo "Must be root to run the script, Exiting"
  exit 1
fi


# Global Vars and defaults
PROGRAM_NAME=$(basename "$0" | cut -d'.' -f1)
verbose=false
dry_run=false
max_receive_time=120
ip_version="any"
IP_VERSION_VALIDATION="^(4|6|any)$"
states="LAST_ACK|FIN_WAIT_1|FIN_WAIT_2|CLOSING"
STATES_VALIDATION="^(CLOSED|LISTEN|SYN_SENT|SYN_RCVD|ESTABLISHED|CLOSE_WAIT|FIN_WAIT_1|CLOSING|LAST_ACK|FIN_WAIT_2|TIME_WAIT)$"
wait_interval=-1
dropped_socket_limit=-1
locked_run=true
force=false
log_target=""
report_metrics=false
LOCK_FILE="/var/lock/${PROGRAM_NAME}.lock"


# TSDB vars
TSDB_HOST="tsdb-data.lldns.net"
TSDB_PORT="4242"
TSDB_TIMEOUT=10
TSDB_KEY="app.${PROGRAM_NAME}.dropped_sockets"


# Write out zabbix monitoring file
# Zabbix can collect this with "vfs.file.contents["/tmp/${PROGRAM_NAME}.status"]"
# The trigger can use 'vfs.file.contents["/tmp/${PROGRAM_NAME}.status"].last().regexp("^FAILED")==1' to trigger
#
# Input: 
#  $1 - The the state we want to log. Currently this should be FAILED or GOOD
#  $2 - The detailed message explaining why we are in this state. Stored in the zabbix item so short is better. 
update_zabbix_file () {
	local state=$1
	local msg=$2

	echo "${state}: ${msg}" > /tmp/"${PROGRAM_NAME}".status
}


# Prints out the useage of the script and exit.
#
# Input:
#	$1 - (optional) Specific error message on why you called usage.
usage () {
	# Check if we got a message explaining why usage was called.
	local msg="$1"
	if [ -n "${msg}" ]; then
		printf "\nERROR: %s\n\n" "${msg}" >&2
	fi

	cat <<-EOF >&2
    Usage: ${PROGRAM_NAME} <options>

    The ${PROGRAM_NAME} script is designed to find TCP sockets in a given state and with the
    last packet receive time above a given threshold. Once found we run "tcpdrop" on that socket. 
    This tooling is in response to #ia0004231 to address a FreeBSD Security Advisory. 

    Resource exhaustion due to sessions stuck in LAST_ACK state
    https://www.freebsd.org/security/advisories/FreeBSD-SA-15:13.tcp.asc

    DEFAULT BEHAVIOR
    The default action that will be taken if this script is run with no parameters will be to 
    look for TCP(4|6) sockets in the LAST_ACK,FIN_WAIT_1,FIN_WAIT_2,CLOSING state without a 
    received packet in the last ${max_receive_time} seconds. A summary will be printed and the user  
    will be prompted if they would like to drop the connections. If the user agrees, the sockets
    will be dropped. 

    Generic Options:
        -h, --help 
            This help output.
                
        -v,--verbose
            Explain the actions taken at each step. This can be very noisy depending on the number of sockets. 

        -d,--dry-run
            Generates the sockets that will be dropped. Note this will still log and use normal locking.
    

    Filters:
        -m,--max-receive-time <time in seconds>
            Max time, in seconds, since the last packet received for a given socket.
            Default: 120

        -i,--ip-version <4|6|any>
            Select the IP version to filter.
            Default: any

        -s,--state <comma seperated list>
            Select the tcp states to target. The list can be seen
            from " netstat -sp tcp" command output.
            Default: LAST_ACK,FIN_WAIT_1,FIN_WAIT_2,CLOSING
    
    Runtime:
        -w,--wait-interval <time in seconds>
            Add sleep time in seconds between each drop command to slow down TCP drops.
            Default: None

        -L,--dropped-socket-limit <Max number sockets to drop>
            Limit the number of drops that will happen in a single run. 
            Default: None

        -n,--no-lock
            Skip checking for process lock.

        -f,--force
            Don't prompt for confirmation before running. Useful for automated runs.

    Reporting:
        -l,--log-target <File location|syslog>
            Log output path and file name.
            Default: None

        -r,--report-metrics
            Send metrics into openTSDB on the status of the results of the run. 

            Metrics:
            	- Number of sockets dropped
            Tags:
            	- FQDN
            	- pop
            	- IP Version
            	- State 
            	- Port {80|443|other}
	EOF
	exit 1
}


# Parse the arguments value return the value. 
#
# Input:
# 	$1 - String that is intended to be the value of an argument.
get_arg_value () {
	local arg=$1

	# Check if the value of the are starts with another flag or
	# named argument.
	if [ -n "$arg" ] && [ "${arg:0:1}" != "-" ]; then
		echo "$arg"
	else
		usage "Error: Argument for $arg is missing"
	fi
}


#Parse arg flags and named arguments for the script.
while (( "$#" )); do 
	case "$1" in
		-h|--help)
			usage;
			;;
		-v|--verbose)
			verbose=true
			shift 1
			;;
		-d|--dry-run)
			dry_run=true
			shift 1
			;;
		-m|--max-receive-time)
			max_receive_time=$(get_arg_value "$2")

			# Validate its a positive integer
			[[ -n "${max_receive_time##*[!0-9]*}" ]] || \
				usage "$1 - value ${max_receive_time} is not an integer."
			[[ $max_receive_time -gt 0 ]] || \
				usage "$1:  value ${max_receive_time} is not positive."
				
			shift 2
			;;
		-i|--ip-version)
			ip_version=$(get_arg_value "$2")

			# Validate IP version
			[[ "$ip_version" =~ $IP_VERSION_VALIDATION ]] || \
				usage "$1: Invalid IP Version ${ip_version}"

			shift 2
			;;
		-s|--states)
			states=$(get_arg_value "$2" | tr ',' '|')

			# Validate TCP States
			for state in $(echo "$states" | tr '|' ' '); do
				[[ "$state" =~ $STATES_VALIDATION ]] || \
					usage "$1: Unsupported tcp state ${state}"
			done

			shift 2
			;;
		-w|--wait-interval)
			wait_interval=$(get_arg_value "$2")

			# Validate its a positive integer
			[[ -n "${wait_interval##*[!0-9]*}" ]] || \
				usage "$1 - value ${wait_interval} is not an integer."
			[[ $wait_interval -gt 0 ]] || \
				usage "$1:  value ${wait_interval} is not positive."

			shift 2
			;;
		-L|--dropped-socket-limit)
			dropped_socket_limit=$(get_arg_value "$2")

			# Validate its a positive integer
			[[ -n "${dropped_socket_limit##*[!0-9]*}" ]] || \
				usage "$1 - value ${dropped_socket_limit} is not an integer."
			[[ $dropped_socket_limit -gt 0 ]] || \
				usage "$1:  value ${dropped_socket_limit} is not positive."

			shift 2
			;;
		-n|--no-lock)
			locked_run=false
			shift 1
			;;
		-f|--force)
			force=true
			shift 1
			;;
		-l|--log-target)
			log_target=$(get_arg_value "$2")
			shift 2
			;;
		-r|--report-metrics)
			report_metrics=true
			shift 1
			;;
		-*|--*) 
			usage "Unsupported flag $1"
			;;
		*)
			usage "Unknown positional argument $1"
			;;
		
	esac
done


# Log message to various targets with timestamp and correct formatting. 
# 
# Input:
#	$1 - Log level as indicated by the below table 
#		Log Levels
#		 FATAL	Very severe error events that might cause the application to terminate.
#		 ERROR	Error events of considerable importance that will prevent normal program
#				execution, but might still allow the application to continue running.
#		 WARN	Potentially harmful situations of interest to end users or system managers 
#				that indicate potential problems.
#		 INFO	Informational messages that might make sense to end users and system 
#				administrators, and highlight the progress of the application.
#		 DEBUG	Relatively detailed tracing used by application developers. The exact meaning 
#				of the three debug levels varies among subsystems.
# 	$2 - Detailed log message
log (){
	local level=$1
	local msg=$2

	# Define log format. This was taken from DCE's plan for new microservice design that is
	# is compatable with ELK and splunk. 
	local log_message="time=\"$(date +%Y-%m-%dT%H:%M:%S%z)\" level=\"${level}\" application=\"${PROGRAM_NAME}\" msg=\"${msg}\""

	# Do nothing if we are not in verbose mode and we get a debug log message.
	if [[ "$level" == "DEBUG" ]] && ! "$verbose"; then
		return 0
	fi

	# Log to local file or to syslog - NOTE: this file is setup and confirmed valid in main().
	if [[ -n "$log_target" ]]; then
		if [[ "$log_target" == "syslog" ]]; then
			logger -t "${PROGRAM_NAME}" "$log_message"
		else
			echo "${log_message}" >> "$log_target"
		fi
	fi

	# Always log to STDERR
	echo "$log_message"	>&2
}


# Track a local lock file containing the pid of the current proceess. If a pid is already
# running that matches the contents of the pid file the program will exit. 
process_lock () {
	# Check on the stat of the lock file.
	log "DEBUG" "Checking ${LOCK_FILE} status"
	if [[ -f $LOCK_FILE ]]; then
		log "DEBUG" "${LOCK_FILE} file exists, checking pid state."
		if pgrep -qF "$LOCK_FILE" 2>/dev/null; then
			# Normal state is to not run if we have an active lock 
			if ${locked_run}; then
				log "FATAL" "Unable to start due to script already running."
				exit
			# User has indicated that they would like to run the script even if we have a lock
			else
				log "WARN" "Process is already running but proceeding anyway because force flag. Not updating pid file."
				return 0
			fi
		fi
	fi

	# If nothing failed above we are clear to write our pid to the lock file.
	log "DEBUG" "No active process running. Saving this processes pid:${$} to ${LOCK_FILE}"
	echo $$ > "$LOCK_FILE"
	if [ $? -ne 0 ]; then
		log "FATAL" "Could not create/update pid file ${LOCK_FILE}"
		exit 1
	fi
}


# Collect all sockets that match our filters and joint them to get the state and rcvtime time. 
#
# Returns:
# 	stdout - List of all sockets in the format 
#				Local_Address Foreign_Address Proto rcvtime state
get_sockets () {
	local socket_time_cmd
	local socket_data

	# Get age of all tcp sockets 
	socket_time_cmd=$(netstat -anxWp tcp | grep ^tcp)

	# Remove listen sockets
	socket_time_cmd=$(echo "$socket_time_cmd" | grep --fixed-strings -v '*.*')

	# Filter IP version to v4 or v6 sockets
	if [[ "$ip_version" =~ ^(4|6)$ ]]; then
		socket_time_cmd=$(echo "$socket_time_cmd" | grep ^tcp"${ip_version}")
	fi

	# Filter the rcvtime
	socket_time_cmd=$(echo "$socket_time_cmd" | awk -v max_receive_time="${max_receive_time}" '{if ($NF > max_receive_time) print $4"^"$5, $1, $NF}')

	# Sort final socket time results and remove dupes if any
	socket_time_cmd=$(echo "$socket_time_cmd" | sort -u)
	
	# Get socket states filtered by state
	socket_state_cmd=$(netstat -anWp tcp | awk '{print $4"^"$5, $NF}' | grep -E "${states}" | sort -u)

	# Find sockets that match both filters
	socket_data=$(join <(echo "$socket_time_cmd") <(echo "$socket_state_cmd") | tr '^' ' ')

	log "INFO" "Found $(echo "$socket_data" | wc -l | awk '{print $1}') matching sockets"

	# Return Local Address and Foreign Address of the sockets for processing
	echo "$socket_data"
}


# Drop a list of sockets
#
# Input:
#	$1 - String containing a list of sockets to drop. First two columns should be local_addr and foreign_addr.
#	$2 - Variable namerefs. This is updated with the list of sockets that are actully droppped. 
drop_sockets () {
	local socket_list=$1
	local -n dropped_sockets_array=$2

	if [[ $dropped_socket_limit -gt 0 ]]; then
		log "INFO" "Limiting the number of dropped sockets to ${dropped_socket_limit}"
	fi

	# Init array to track what sockets where dropped for reporting
	dropped_sockets_array=()


	while IFS= read -r line; do
		if [[ ${#dropped_sockets_array[@]} -lt $dropped_socket_limit || $dropped_socket_limit == -1 ]]; then
			local local_addr=$(echo "$line" | cut -d' ' -f1)
			local foreign_addr=$(echo "$line" | cut -d' ' -f2)
			log "DEBUG" "running: tcpdrop ${local_addr} ${foreign_addr} for socket ${line}"

			# Run tcpdrop, capture the output, and if success add to the dropped array.
			if ! output=$(tcpdrop "$local_addr" "$foreign_addr" 2>&1); then
				if echo "$output" | grep -q 'No such process'; then
					log "DEBUG" "Socket already gone (${local_addr} ${foreign_addr}): ${output}"
				else
					log "DEBUG" "Failed to drop socket (${local_addr} ${foreign_addr}): ${output}"
				fi
			else
				dropped_sockets_array+=("$line")
				log "DEBUG" "Success Output: $output"

				# If we have a wait configured, sleep between drops.
				[[ "$wait_interval" -gt 0 ]] && sleep "$wait_interval"
			fi
		else
			break
		fi
	done <<< "$socket_list"

	log "INFO" "Dropped ${#dropped_sockets_array[@]} sockets"
}


# Send metrics into openTSDB on what was dropped.
#
# Input:
#	$1 - String containing the sockets that where dropped in the following format. 
#			Local_Address Foreign_Address Proto rcvtime state
send_metrics () {
	local -n sockets=$1
	local FQDN=$(uname -n)
	local POP=$(echo "$FQDN" | cut -d'.' -f2)

	# Convert array to string
	local socket_list=$( IFS=$'\n'; echo "${sockets[*]}" )

	# Get the time and the time aligned to the rearest 2nd minute
	# This is needed for grafana and tsdb to graph and calulate the
	# data for all cds servers at the same minute. 
	local now=$(date +%s)
	local now_aligned=$(( now - now % 120 ))

	# Convert socket data to TSDB lines
	TSDB_lines=$(
		echo "$socket_list" \
		| awk '{split($1,local_addr,".");local_port=local_addr[length(local_addr)]; if (local_port !~ /^(80|443)$/) local_port="other";print local_port,$3,$5}' \
		| sort | uniq -c | sort -nrk 1 \
		| awk -v fqdn="${FQDN}" -v pop="${POP}" -v timestamp="$now_aligned" -v tsdbkey="$TSDB_KEY" '{print "put", tsdbkey, timestamp, $1, "fqdn="fqdn, "pop="pop, "port="$2, "ip="$3, "state="$4}'
	)

	while IFS= read -r line; do
		log "DEBUG" "openTSDB metric created: ${line}"
	done <<< "$TSDB_lines"

	if $report_metrics; then
		if ! output=$(echo -ne "$TSDB_lines" | nc -w $TSDB_TIMEOUT $TSDB_HOST $TSDB_PORT); then
			# If the metrics are send and TSDB blocks them it will 
			# still report back valid. So this check is missing alot.
			log "FATAL" "Failed to send data to tsdb output: ${output}"

			# Not going to monitor this due to openTSDB issues being very likly.
			#update_zabbix_file "FAILED" "Could not send in metrics to openTSDB"
			exit
		else
			log "INFO" "Metrics sent to openTSDB: ${output}"
		fi
	fi

}


main () {
	# Setup logging
	if [[ -n "$log_target" && "$log_target" != "syslog" ]]; then
		# Ensure path exists or we can make the direcotry else exit
		local dir_error=$(mkdir -p $(dirname "$log_target") 2>&1)
		local result=$?
		if [[ $result -ne 0 ]] ; then
			log "FATAL" "Unable to make dir: ${dir_error}"
			update_zabbix_file "FAILED" "Unable to make dir: ${dir_error}"
			exit 1
		fi
	fi
	log "INFO" "Started ${PROGRAM_NAME}"

	# Check lock file and update as needed
	process_lock

	# Get the list of sockets that are actively matching the filters.
	socket_list=$(get_sockets)

	if $dry_run; then
		echo "The following matched current filters for drop targets."
		echo "Local_Address Foreign_Address Proto rcvtime state"
		echo "$socket_list"
		exit
	fi

	# Check if we should exit before dropping sockets
	if ! $force; then
		echo
		read -p "Would you like to proceed with dropping the sockets? (Y/N) " -n 1 -r
		echo
		if [[ $REPLY =~ ^[Nn]$ ]]; then
			log "INFO" "Nothing dropped, exiting"
			exit 1
		elif [[ ! $REPLY =~ ^[Yy]$ ]]; then
			echo "FATAL" "Unkown user input \"${REPLY}\", exiting..."
			exit 1
		fi
	fi
	
	# Proceed with droping selected sockets
	local dropped_sockets
	drop_sockets "$socket_list" dropped_sockets

	# Send in metrics related to the sockets
	send_metrics dropped_sockets

	# Write that script completed run. 
	update_zabbix_file "GOOD" "Completed full run"
}

main