import re
import os
import psutil
import argparse
import time
import sys
import json
import random
import requests
from requests.auth import HTTPBasicAuth

# Min number of FDs to report on a per proc basis
MAX_OPEN_FDS = 1000

# InfluxDB settings
INFLUX_SERVER = 'opscap.llnw.com'
INFLUX_PORT = '50001'
INFLUX_DATABASE = 'cdssystem'
INFLUX_MEASUREMENT = 'FDs-collection'
INFLUX_USER = 'ingest'
INFLUX_PASSWD = 'senddata'

# General variables
influx_dict = {}
timestamp = int(round(time.time()))
fqdn = os.uname()[1]
EP_PID_FILE = '/usr/local/llnw/zabbix/json/sqprocs.json'
EP_PID_KEY = u'{#SQPROC_PIDFILE}'
ep_pid_map = []
ep_dict = {}


class ProcessFilter():

    # Always report on these programs
    proc_names = ['nginx', 'haproxy',
                  'edgequery-logstream', 'zabbix_agentd']

    # Report these only if the script they are running
    # is matched
    interpreter_names = ['python', 'perl', 'bash']

    # Script names to match
    script_names = ['GridConfigClient.pl', 'edgequeryd.py']

    def __init__(self, max_open_fds, process):
        self.name = process.name()
        self.match_name = self.name
        self.num_fds = process.num_fds()
        self.cmdline = process.cmdline()
        self.max_open_fds = max_open_fds
        self.ep_pid = process.pid

    def match_all(self):
        # Loop over all matching functions

        for match_func in [self.match_procs, self.match_script,
                           self.match_ep_pid, self.match_num_fds]:
            result = match_func()
            if result:
                return result
        self.match_name = "other"
        return self.format_output()

    def match_procs(self):
        # Try to match on proc name

        regexp = re.compile('|'.join(self.proc_names))
        if regexp.search(self.name):
            return self.format_output()
        return False

    def match_script(self):
        # For certain interpreters, try to match a particular script
        # If the cmdline length is 1 or less, it can't match
        if len(self.cmdline) <= 1:
            return False

        proc_regexp = re.compile('|'.join(self.interpreter_names))
        script_regexp = re.compile('|'.join(self.script_names))

        # If both the interpreter and script name match, report the script name
        if proc_regexp.search(self.name) and script_regexp.search(''.join(self.cmdline)):
            self.match_name = self.cmdline[1].split('/')[-1]
            return self.format_output()
        return False

    def match_ep_pid(self):
        # Used to build EP children proc names/FDs
        for k, v in ep_dict.items():
            if self.ep_pid == v:
                self.match_name = k
                return self.format_output()
        return False

    def match_num_fds(self):
        # Match all procs with greater than self.max_open_fds file descriptors
        if self.num_fds > self.max_open_fds:
            return self.format_output()
        return False

    def format_output(self):
        # Return output tuple for a process name match
        return {self.match_name: self.num_fds}


def ep_pid_collect():
    # Build EP child pid file list for specific child reporting
    try:
        with open(EP_PID_FILE) as f:
            data = json.load(f)

        for k in data['data']:
            ep_pid_map.append(k[EP_PID_KEY])

        for k in ep_pid_map:
            with open(k, 'r') as f:
                key = k.split('/')[-1]
                key = key.split('.')[0]
                ep_dict[key] = int(f.read().strip())

    except Exception as e:
        print(e)
        sys.exit(1)


def proc_check(proc):
    try:
        for k, v in proc.items():
            if k in influx_dict:
                influx_dict[k] += v
            else:
                influx_dict[k] = v
    except Exception as e:
        print(e)
        sys.exit(1)


def construct_influx_line(proc_name, fd_count):
    try:
        # for procName, fdCount in proc_dict.items():
        tags = {'fqdn': fqdn, 'process_name': proc_name}
        data = {'file_count': fd_count}
        return process_influx_line(timestamp, 'fdcount', tags, data)
    except Exception as e:
        print(e)
        sys.exit(1)


def process_influx_line(time, measurement, tags, data):
    try:
        tagValues = ','.join({"{}={}".format(k, v) for k, v in tags.items()})
        dataValues = ','.join({"{}={}".format(k, v) for k, v in data.items()})
        influx_line = '{},{} {} {}'.format(
                        '{}.{}'.format(INFLUX_MEASUREMENT, measurement),
                        tagValues,
                        dataValues,
                        time
                    )
        return influx_line
    except Exception as e:
        print(e)
        sys.exit(2)


def send_to_influxDB(influxDataLines):
    # Send the Data to influxDB
    print("Sending {} lines to influxdb".format(len(influxDataLines)))
    r = requests.post('https://%s:%s/write?db=%s&precision=s' % (
            INFLUX_SERVER,
            INFLUX_PORT,
            INFLUX_DATABASE
        ), 
        data='\n'.join(influxDataLines), 
        timeout=30, auth=HTTPBasicAuth(
            INFLUX_USER,
            INFLUX_PASSWD
        ))
    print('Response: {} {} {}'.format(r.text, r.url, r.status_code))


def main():

    # Adds debug optional flag
    parser = argparse.ArgumentParser()
    parser.add_argument('-d', required=False, help="Do not send to InfluxDB", action='store_true')
    parser.add_argument('-i', required=False, help="Do not send to InfluxDB", action='store_true')
    args = parser.parse_args()
    dry_run = args.d
    immediate_run = args.i

    influx_lines = []
    ep_pid_collect()

    # Loop through each process
    for proc in psutil.process_iter():
        proc_obj = ProcessFilter(MAX_OPEN_FDS, proc)
        match = proc_obj.match_all()
        proc_check(match)

    for k, v in influx_dict.items():
        influx_lines.append(construct_influx_line(k, v))

    if dry_run:
        print(influx_lines)
    else:
        # Add a random 1s, up to 3m splay
        if not immediate_run:
            time.sleep(random.randint(1, 180))
        send_to_influxDB(influx_lines)


if __name__ == '__main__':
    sys.exit(main())
