#!/usr/bin/env python


"""
Author: Nate Williams
Created: 1/24/2022
Updated: 2/11/2022

Script to check the firmware of a device and compare
it to an approved verison for monitoring.  Currently
only drive checks are implemented, but the goal is to
support other devices in the future.
"""


import subprocess
import json
import argparse
import sys
import os
import syslog
import re
from operator import itemgetter


parser = argparse.ArgumentParser(description="""
########################################################
##################### Status Codes #####################
########################################################
#0 - device is firmware good
#1 - device firmware is incorrect and monitoring is supressed
#2 - device firmware is incorrect and monitoring is NOT supressed
#3 - device not found in JSON file
#4 - script failed to run""", formatter_class=argparse.RawTextHelpFormatter)
parser.add_argument(
                    "-f",
                    "--file",
                    action="store",
                    help='Provide a JSON file',
                    dest="file",
                    default="/usr/local/llnw/firmware/devicemap.json",
                    type=str,
                    required=False)
parser.add_argument(
                    "-v",
                    "--verbose",
                    action="store_true",
                    help='Show target FW version',
                    dest="verbose",
                    required=False)
args = parser.parse_args()


vb_comment = "%s: Device /dev/%s \
has firmware version %s, and has supported firmware version(s) %s"


def check_root():
    # Check running as root or exit
    if os.geteuid() != 0:
        if args.verbose:
            print("Not root...exiting")
            sys.exit(1)
        else:
            sys.exit(1)


def _run_cmd(cmd, shell=True, quit_on_failure=True):
    # use subprocess to run commands on server with error handling
    try:
        result = subprocess.check_output(
            cmd,
            shell=shell,
            stderr=subprocess.STDOUT)

        return result.rstrip()

    except subprocess.CalledProcessError as e:
        if quit_on_failure:
            # If we dont want to trust the data if the command fails and think
            # its better to exit than report in any data we will quit.
            print("Failed to run command: {}".format(e.cmd))
            print("CODE: {}".format(e.returncode))
            print("OUTPUT: {}".format(e.output))
            quit(1)
        else:
            # raise the error so that the upstream can process the failure.
            raise


def _load_json(filename):
    # Load a json file
    with open(filename) as f:
        data = json.load(f)
    return data


def get_disks():
    # Generate a list of disks from the server filter out
    # cd0 and build a dict
    disk_list = _run_cmd("sysctl -n kern.disks").split()
    disk_list_filtered = {}
    for disk in disk_list:
        if disk != 'cd0':
            disk_list_filtered[disk] = {}
    return disk_list_filtered


def get_disk_model(disks):
    # Iterate over dict of disks, using smartctl
    # to get disk model and append k,v to disk key
    disks_with_models = {}
    for disk in disks.items():
        disk_name = disk[0]
        disk_value = disk[1]
        model = _run_cmd("smartctl -i /dev/%s | \
        egrep 'Device Model:|Product:' | awk '{print $NF}'" % disk_name)
        if model:
            disk_value["model"] = model
            disks_with_models[disk_name] = disk_value
    return disks_with_models


def get_disk_fw(disks):
    # Iterate over dict of disks, using smartctl
    # to get disk firmware and append k,v to disk key
    disks_with_model_and_firmware = {}
    for disk in disks.items():
        disk_name = disk[0]
        disk_value = disk[1]
        fw = _run_cmd("smartctl -i /dev/%s | \
        egrep 'Firmware Version:|Revision:' | awk '{print $NF}'" % disk_name)
        disk_value["firmware"] = fw
        disks_with_model_and_firmware[disk_name] = disk_value
    return disks_with_model_and_firmware


def max_status(current, new):
    # Check status code and only return the highest value
    if new > current:
        return new
    else:
        return current


def sorted_nicely(l, key):
    """ Sort the given iterable in the way that humans expect."""
    convert = lambda text: int(text) if text.isdigit() else text
    alphanum_key = lambda item: [convert(c) for c in re.split('([0-9]+)', key(item))]
    return sorted(l, key=alphanum_key)


def check_disks(data):
    # function, iterate over disks compare disk firmware
    # with firmware in json file, by matching disk model
    status = 0
    disks = get_disks()
    disks = get_disk_model(disks)
    disks = get_disk_fw(disks)
    sorted_disks = sorted_nicely(disks.items(), itemgetter(0))
    try:
        for disk in sorted_disks:
            disk_name = disk[0]
            disk_model = disk[1]["model"]
            disk_fw = disk[1]["firmware"]
            model_info = data["DISK"][disk_model]
            if disk_fw not in model_info["firmware"]["supported"]:
                status = 1
                if model_info["monitored"] is True:
                    status = 2
                vb_status = "FAIL"
                syslog.syslog(
                              syslog.LOG_ERR,
                              vb_comment % (
                                           vb_status,
                                           disk_name,
                                           disk_fw,
                                           model_info["firmware"]
                                           ["supported"]))
            else:
                vb_status = "PASS"
            if args.verbose:
                print(vb_comment % (
                                   vb_status,
                                   disk_name,
                                   disk_fw,
                                   model_info["firmware"]["supported"]))
        return status
    except KeyError as err:
        status = 3
        if args.verbose:
            print('Unable to locate %s in JSON file.' % err)
        return status


def main():
    status = 0
    try:
        data = _load_json(args.file)
        check_root()
        status_checks = [
            check_disks
        ]
        for check in status_checks:
            status = max_status(status, check(data))
    except:
        status = 4
        print(status)
        sys.exit(1)
    print(status)


if __name__ == "__main__":
    main()
