#!/usr/bin/env python2

import sys
import socket
import argparse

IRR_SOCKET = '/var/run/haproxy/master.sock'

def handle(conf, requested_key, verbose):
    """
    Accepting input to display parsed data from the IRR socket

    :param conf: 1 or 2
    :param requested_key: desired key we want the metric for
    :param verbose: displays the original raw socket details
    :return: None
    """
    socket_info = sock_parse(conf, verbose)

    if verbose:
        print >> sys.stderr, '** Start of raw data from socket'
        print >> sys.stderr, socket_info
        print >> sys.stderr, '** End of raw data from socket'

    parsed_data = dict()
    parsed_data['shared_mem_used'] = get_shared_counters(socket_info, 'used')
    parsed_data['shared_mem_max_used'] = get_shared_counters(socket_info, 'max_used')
    parsed_data['shared_mem_available'] = get_shared_counters(socket_info, 'max_available')
    parsed_data['shared_ids_used'] = get_shared_counters(socket_info, 'used', False)
    parsed_data['shared_ids_max_used'] = get_shared_counters(socket_info, 'max_used', False)
    parsed_data['shared_ids_available'] = get_shared_counters(socket_info, 'max_available', False)
    parsed_data['orphan_entries'] = get_orphan_entries(socket_info)

    if verbose:
        print >> sys.stderr, '** Start of parsed data'
        for k, v in parsed_data.items():
            print >> sys.stderr, '[%s]\t%s' % (k, v)
        print >> sys.stderr, '** End of parsed data'

    if requested_key in parsed_data:
        print(parsed_data[requested_key])
    else:
        print(requested_key + ' is not a valid key')
        sys.exit(1)

def sock_parse(conf, verbose):
    """
    Sends show-cdn-conig to irr unix socket, requesting libcdnc data for
    specified squid#-conf

    :param conf: 1 or 2
    :param verbose: specifies if we want to output the raw data as read from the socket 
    :return: raw data from the socket
    """
    command = 'show-cdn-config squid%s-conf\n' % conf

    sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    try:
        sock.connect(IRR_SOCKET)
        if verbose:
            print >> sys.stderr, '** Sending "%s" to socket at %s' % (command.strip(), IRR_SOCKET)
        sock.sendall(command)
        data = sock.recv(4096)
    except socket.error, msg:
        print >> sys.stderr, '** Error during connection with socket: %s' % msg
        sys.exit(1)
    finally:
        if verbose:
            print >>sys.stderr, '** Closing socket'
        sock.close()
    return data

def get_orphan_entries(socket_info):
    """
    Returns the number of orphaned entries, given a string containing the data
    returned from the IRR socket

    :param socket_info: the string data returned from the IRR socket
    :return: an integer containing the number of orphaned entries
    """
    return len([key_val for key_val in __key_val_parse(socket_info) if key_val[0] == 'orphan'])


def get_shared_counters(socket_info, shared_id_type, use_shared_mem=True):
    """
    Returns the values of shared mem and shared_ids, given a string containing
    the data returned from the IRR socket

    :param socket_info: the string data returned from the IRR socket
    :param shared_id_type: specify the keys used for shared_ids
    :param use_shared_mem: specify the keys used for shared mem
    :return: an integer containing the values of the sahred_id_type
    """
    assert(shared_id_type in ['used', 'max_used', 'max_available'])

    shared_ids_list = [key_val for key_val in __key_val_parse(socket_info) if key_val[0] == ('shared mem' if use_shared_mem else 'shared_ids')]
    assert(len(shared_ids_list) == 1) # only 1 shared_id/shared mem element expected
    shared_ids_value = shared_ids_list[0][1] # get first (and only) element from list, and get key from kv pair
    for shared_id in shared_ids_value.split(','):
        kv = shared_id.split('=', 1)
        if kv[0] == shared_id_type:
            return kv[1]
    return -1

def __key_val_parse(data):
    """
    Given a multiline string containing at least one ':' per line to delimit
    key value pairs, return a list containing two element lists of key vals.

    :param data: the multiline list to parse into key value pairs
    :return: a list containing two element lists
    """
    return [line.split(':', 1) for line in data.splitlines()]

def __arg_conf_validate(value):
    if value not in ('1', '2'):
        raise argparse.ArgumentTypeError('expected 1 or 2')
    return value
    
if __name__ == '__main__':
    argparser = argparse.ArgumentParser()
    argparser.description = 'Reads IRR unix socket, parsing libcdnc data'
    argparser.add_argument('-c', '--conf', type=__arg_conf_validate, required=True, help='which IRR conf to use, allowed "1" or "2" (for squid1-conf or squid2-conf)')
    argparser.add_argument('-k', '--key', type=str, required=True, help='which metric is required')
    argparser.add_argument('-v', '--verbose', action='store_true', help='flag to enable verbose output printing')

    args = argparser.parse_args()

    handle(args.conf, args.key, args.verbose)
