#!/usr/bin/env python
import subprocess
import re
import argparse
import json

def _run_cmd(cmd, shell=True, quit_on_failure=True):
    '''
    Runs a system command and can return the failure or exit the program.

    Parameters
    ----------
    cmd : string
        shell command string to be ran

    shell : bool, optional
        If shell is True, the specified command will be executed through
        the shell. (Default is True)

    quit_on_failure : bool, optional
        If we get a bad return code and we want to kill the current running
        python process.


    Returns
    -------
    string
        Output of the command that ran


    Raises
    ------
    CalledProcessError
        Exception raised when a process run by check_call() or check_output() returns a non-zero exit status.

    '''

    try:
        result = subprocess.check_output(
            cmd,
            shell=shell,
            stderr=subprocess.STDOUT
            )

        return result.rstrip()

    except subprocess.CalledProcessError as e:
        if quit_on_failure:
            # If we dont want to trust the data if the command fails and think
            # its better to exit than report in any data we will quit.
            print("Failed to run command: {}".format(e.cmd))
            print("CODE: {}".format(e.returncode))
            print("OUTPUT: {}".format(e.output))
            quit(1)
        else:
            # raise the error so that the upstream can process the failure.
            raise


def get_ifconfig_output(nic_name):
    list_interface = 'ifconfig -v {}'.format(nic_name)
    ifconfig_output = _run_cmd(list_interface)
    return ifconfig_output


def get_mac_address(raw_ifconfig_output):
    mac = {'mac_hwaddr': None, 'mac_ether': None}

    for line in raw_ifconfig_output.split('\n'):
        if 'ether ' in line:
            mac['mac_ether'] = line.split()[1]
        if 'hwaddr ' in line:
            mac['mac_hwaddr'] = line.split()[1]

    return mac


def get_interface_status(raw_ifconfig_output):
    status = None

    for line in raw_ifconfig_output.split('\n'):
        if 'status: ' in line:
            status = ' '.join(line.split()[1:])

    return status


def get_media_data(raw_ifconfig_output):
    data = {
        'potential_speed_bps': 0,
        'duplex': None,
        'negotiated_speed_bps': 0
    }

    media_regex = r'media:.*?(\d+)(\S*base).+<(\S+)>'

    for line in raw_ifconfig_output.split('\n'):
        if 'media: ' in line:
            line = line.lower()
            data['media_type'] = 'null'

            m = re.search(media_regex, line)
            if m:
                data['duplex'] = m.group(3)
                data['media_type'] = '{}{}'.format(m.group(1), m.group(2))

                speed = int(m.group(1))
                base = m.group(2)

                if base.startswith('base'):
                    speed = speed * 1e6
                elif base.startswith('gbase'):
                    speed = speed * 1e9
                elif base.startswith('tbase'):
                    speed = speed * 1e12

                data['potential_speed_bps'] = int(speed)
                data['negotiated_speed_bps'] = int(speed)

    return data


def get_lagg_info(raw_ifconfig_output):
    lagg_data = {
        'lagg_proto': None,
        'lagg_hash': None,
        'lagg_members': [],
        'lagg_flags': None,
        'lagg_member_flags': {}
    }

    flags_regex = r'flags=\S+<(\S*)>'
    lines = raw_ifconfig_output.split('\n')
    for index, line in enumerate(lines):
        if 'laggproto' in line:
            _, laggproto, _, lagghash = line.split()
            lagg_data['lagg_proto'] = laggproto
            lagg_data['lagg_hash'] = lagghash
        elif 'laggport' in line:
            interface = line.split()[1]
            lagg_data['lagg_members'].append(interface)
            if 'flags=' in line:
                m = re.search(flags_regex, line)
                lagg_data['lagg_member_flags'][interface] = m.group(1).lower().split(',')
        elif 'lagg options:' in line:
            if 'flags=' in lines[index + 1] and 'flags=0' not in lines[index + 1]:
                m = re.search(flags_regex, lines[index + 1])
                lagg_data['lagg_flags'] = m.group(1).lower().split(',')

    return lagg_data


def get_interface_flags(raw_ifconfig_output):
    data = None

    flag_regex = r'flags=\S+<(\S+)>'

    for line in raw_ifconfig_output.split('\n'):
        if ': flags=' in line:
            line = line.lower()
            m = re.search(flag_regex, line)

            data = m.group(1).split(',')

    return data


def get_interface_options(raw_ifconfig_output):
    data = {
        'options': None,
        'nd6_options': None
    }

    option_regex = r'options=\S+<(\S+)>'

    for line in raw_ifconfig_output.split('\n'):
        if 'options=' in line:
            line = line.lower()
            m = re.search(option_regex, line)
            if 'nd6 options=' in line:
                data['nd6_options'] = m.group(1).split(',')
            else:
                data['options'] = m.group(1).split(',')

    return data


def get_default_route_iface():
    default_interface = None

    netstat_default_route = 'netstat -4nr'

    netstat_output = _run_cmd(netstat_default_route)

    for line in netstat_output.split('\n'):
        if line.startswith('default'):
            default_interface = line.split()[-1]

    return default_interface


def get_cc_max_speed(interface_number):
    key = 'dev.cc.{}.max_speed'.format(interface_number)
    sysctl_max_speed = 'sysctl {}'.format(
        key
    )

    speed = _run_cmd(sysctl_max_speed).split()[-1]
    speed = float(speed) * 1e9

    return int(speed)


def get_interface_list():
    list_interfaces = 'ifconfig -l'
    interface_list = _run_cmd(list_interfaces).split()
    return interface_list


def get_pciconf(device):
    '''
    Calls 'pciconf -val {interface_name}' to get the data for the nic

    Parameters
    ----------
    device : string
        system device to pass to pciconf

    Returns
    -------
    dict
        dict of keys that pciconf returns
    '''
    pci_data = {'vendor' : None, 'device_string': None}

    pciconf = 'pciconf -val {} | tr -d "\'" | egrep "vendor|device"'.format(device)
    try:
        interface_pciconf = _run_cmd(pciconf, quit_on_failure=False).split('\n')
        for item in interface_pciconf:
            split_item = item.split('=')

            if len(split_item) != 2:
                # Documenting the reason for the exception however nothing will report on this.
                raise ValueError('Unexpeted string found in config line: {}'.format(
                    item
            ))
            key, value = split_item
            value = value.strip()

            if 'vendor' in key:
                pci_data['vendor'] = value

            elif 'device' in key:
                pci_data['device_string'] = value

        return pci_data
    except:
        # Let the calling fuction know that we where not able to find
        # the pciconf data
        return pci_data


def get_sysctl(key):
    '''
    Calls 'sysctl {key}' and returns the value of the sysctl key

    Parameters
    ----------
    key : string
        System sysctl key

    Returns
    -------
    string
        value of the sysctl key. On failure None is returned
    '''
    try:
        return _run_cmd("sysctl -n {}".format(key), quit_on_failure=False)
    except:
        return None


def get_device_parent(interface_name):
    '''
    Find parent device via sysctl.

    Parameters
    ----------
    interface_name : string
        Name of the nic device e.g. ix0

    Returns
    -------
    string
        name of parent value

    '''
    parent = None
    try:
        interface_parts = list(re.findall(r"^(\D+)(\d+)", interface_name)[0])
        sysctl_name = 'dev.{}.{}.%parent'.format(interface_parts[0], interface_parts[1])

        parent = get_sysctl(sysctl_name)
    except:
        pass

    return parent


def get_nic_vendor_model(interface_name):
    '''
    Find interface vendor and model strings.

    Parameters
    ----------
    interface_name : string
        Name of the nic device e.g. ix0

    dict
        Contains 'vendor' and 'device_string'
    '''
    nicinfo = get_pciconf(interface_name)

    # If we have no data, try and get the parent device name
    if nicinfo['vendor'] == None:
        parent_device = get_device_parent(interface_name)

        # Try once to get the info using the parent device name this
        # is needed for Chelsio cards. The interface name cc0 is a child
        # to a t5nex0 device that pciconf can read.
        if parent_device:
            nicinfo = get_pciconf(parent_device)

    return nicinfo


def filter_matches_interface(filters, interface):
    '''
    Takes in a dict of filters and an interface to try and match
    against to find if it matches or not.

    Parameters
    ----------
    filters : dict
        A dict of keys to match with regex values.
        {'key1': r'filter1', 'key2': r'filter2', 'keyn', r'filtern'}

    interface : dict
        A dict contaning all of the interface keys that will be filtered
        on. For exmaple if you filtering on {'vendor': '.*T6.*'}
    '''

    for k, v in filters.iteritems():
        # If the interface is missing the key or the value
        # is None we know it does not match.
        if k not in interface or interface[k] == None:
            return False

        # Use the value of the filter to try a regex match.
        if re.search(v, interface[k]) == None:
            # If we fail to match any of the filters we
            # know that we do not match and return False.
            return False

    # If none of the filters failed then we know that this
    # is a match for the interface.
    return True

def get_potential_override(interface_object):
    '''
    Uses the interface mapping to override potential_speed_bps. First
    pattern match wins.

    Parameters
    ----------
    interface_object : dict
        This is a dictionary must containing 'vendor',
        'device_string' and 'potential_speed_bps' used
        for the lookups and default values.

    Returns
    -------
    int
        The potential speed in bps units.
    '''

    # Setting the default value of potential_speed_bps to the original
    # value incase we find no match.
    potential_speed_bps = interface_object['potential_speed_bps']

    # Ordered list of overrides to seach through for the interface
    potential_speed_override = [
        {'filters': {'vendor': r'.*Chelsio.*', 'device_string': r'.*T6.*'}, 'value': 25000000000},
    ]

    for override in potential_speed_override:
        match = filter_matches_interface(override['filters'], interface_object)


        if match:
            potential_speed_bps = override['value']
            break

    return potential_speed_bps


def get_interface_data():
    # Get all interfaces
    interface_list = get_interface_list()
    interfaces = {}

    for interface in interface_list:
        interfaces[interface] = {
            'gateway': False
        }

        # Get raw output of the interface for parsing
        raw_ifconfig_output = get_ifconfig_output(interface)
        interfaces[interface]['raw'] = raw_ifconfig_output

        # Get mac addresses
        interfaces[interface].update(get_mac_address(raw_ifconfig_output))

        # Get nic status
        interfaces[interface]['status'] = get_interface_status(raw_ifconfig_output)

        # Parse media output
        interfaces[interface].update(get_media_data(raw_ifconfig_output))

        # Detect lagg
        if 'lagg' in raw_ifconfig_output:
            interfaces[interface].update(get_lagg_info(raw_ifconfig_output))

        # Get interface flags
        interfaces[interface]['flags'] = get_interface_flags(raw_ifconfig_output)

        # Get interface options
        interfaces[interface].update(get_interface_options(raw_ifconfig_output))

        # Get interface Vendor and Model
        interfaces[interface].update(get_nic_vendor_model(interface))

        # Chelsio NICs currently have a bug where they dont report the
        # media info correctly if they have a different speed optic.
        if re.search('cc[0-9]+', interface):
            interfaces[interface]['negotiated_speed_bps'] = get_cc_max_speed(
                interface[2:]
            )
        # Reports negotiated_speed_bps as 0 if status not active and flag not up
        if 'up' not in interfaces[interface]['flags'] or interfaces[interface]['status'] != 'active':
            interfaces[interface]['negotiated_speed_bps'] = 0

        # Some Chelsio interfaces do not report the max speed in sysctl correctly
        # so we are able to report on potential_speed_bps by confirming with the
        # capacity team what devices can support 25G nics even if the host does not
        # report correctly.
        interfaces[interface]['potential_speed_bps'] = get_potential_override(interfaces[interface])



    # Process lagg interface
    for interface in interfaces:
        if 'lagg' in interface and 'lagg_members' in interfaces[interface]:
            lagg_negotiated_speed_bps = 0
            lagg_potential_speed_bps = 0

            # Add the members of the laggs speed to the lagg
            for lagg_member in interfaces[interface]['lagg_members']:

                # If lagg_member_flags meets true then add lagg_negotiated_speed_bps and interfaces[lagg_member]['negotiated_speed_bps']
                # and assign new value to lagg_negotiated_speed_bps

                if 'active' in interfaces[interface]['lagg_member_flags'][lagg_member]:
                   lagg_negotiated_speed_bps += interfaces[lagg_member]['negotiated_speed_bps']
                else:
                    interfaces[lagg_member]['negotiated_speed_bps'] = 0
                lagg_potential_speed_bps += interfaces[lagg_member]['potential_speed_bps']

            interfaces[interface]['negotiated_speed_bps'] = lagg_negotiated_speed_bps
            interfaces[interface]['potential_speed_bps'] = lagg_potential_speed_bps

    # Get default route interface. If its a lagg tag the members as the gateway as well
    default_interface = get_default_route_iface()

    if 'lagg' in default_interface and 'lagg_members' in interfaces[default_interface]:
        for lagg_member in interfaces[default_interface]['lagg_members']:
                interfaces[lagg_member]['gateway'] = True

    interfaces[default_interface]['gateway'] = True

    return interfaces


def read_args():
    parser = argparse.ArgumentParser(
        description="Command line tool for purge api actions"
    )

    parser.add_argument(
        '-f',
        '--format',
        choices=['json', 'zabbix'],
        default='json',
        help='Set the output format. json(default), zabbix pipe/comma seperated values).'
    )
    parser.add_argument(
        '-k',
        '--keys',
        help='Comma seperated list of items to return. By default all keys are returned.  Exmaple: status,duplex'
    )
    parser.add_argument(
        '-i',
        '--interface',
        help='Comma seperated list of interfaces to return data for. By default all interfaces are returned.'
    )
    return parser.parse_args()


def filter_keys(data, filter, separator=','):
    filter_keys = filter.split(separator)
    filtered_data = {}

    for interface in data:
        filtered_data[interface] = {}

        for key in filter_keys:
            if key in data[interface]:
                filtered_data[interface][key] = data[interface][key]

    return filtered_data


def filter_interface(data, interface_list, separator=','):
    filtered_interfaces = interface_list.split(separator)
    filtered_data = {}

    for interface in data.keys():
        if interface in filtered_interfaces:
            filtered_data[interface] = data[interface]

    return filtered_data


def print_formated(data, output_format):
    if output_format == 'json':
        print(json.dumps(data, indent=4))

    elif output_format == 'zabbix':
        for interface in data:
            for key in data[interface]:
                if key in ['raw']:
                    continue

                value = data[interface][key]
                if not value or value == 'null':
                    value = ''

                if isinstance(value, list):
                    value = ' '.join(value)

                print('{},{}|{}'.format(
                    interface,
                    key,
                    value
                ))


def main():
    # Read all input arguments
    args = read_args()

    # Get all the interface data
    d = get_interface_data()

    # Process interface filter
    if args.interface:
        d = filter_interface(d, args.interface)

    # Process key filter
    if args.keys:
        d = filter_keys(d, args.keys)

    # Print collected data
    print_formated(d, args.format)


if __name__ == "__main__":
    main()
