#!/usr/bin/env python
import subprocess
import re
import argparse
import json


def _run_cmd(cmd_array):
    try:
        result = subprocess.check_output(cmd_array).rstrip()

        return result
    except Exception as e:
        print('FAILED: Could not execute cmd \'{}\', {}'.format(
            ' '.join(cmd_array),
            e
        ))

        # If we have failed to execute any commands the data
        # cannot be trusted so we exit.
        quit(1)



def get_ifconfig_output(nic_name):
    list_interface = ['ifconfig', '-v', nic_name]
    ifconfig_output = _run_cmd(list_interface)
    return ifconfig_output


def get_mac_address(raw_ifconfig_output):
    mac = {'mac_hwaddr': None, 'mac_ether': None}

    for line in raw_ifconfig_output.split('\n'):
        if 'ether ' in line:
            mac['mac_ether'] = line.split()[1]
        if 'hwaddr ' in line:
            mac['mac_hwaddr'] = line.split()[1]

    return mac


def get_interface_status(raw_ifconfig_output):
    status = None

    for line in raw_ifconfig_output.split('\n'):
        if 'status: ' in line:
            status = ' '.join(line.split()[1:])

    return status


def get_media_data(raw_ifconfig_output):
    data = {
        'potential_speed_bps': 0,
        'duplex': None,
        'negotiated_speed_bps': 0
    }

    media_regex = r'media:.*?(\d+)(\S*base).+<(\S+)>'

    for line in raw_ifconfig_output.split('\n'):
        if 'media: ' in line:
            line = line.lower()
            data['media_type'] = 'null'

            m = re.search(media_regex, line)
            if m:
                data['duplex'] = m.group(3)
                data['media_type'] = '{}{}'.format(m.group(1), m.group(2))

                speed = int(m.group(1))
                base = m.group(2)

                if base.startswith('base'):
                    speed = speed * 1e6
                elif base.startswith('gbase'):
                    speed = speed * 1e9
                elif base.startswith('tbase'):
                    speed = speed * 1e12

                data['potential_speed_bps'] = int(speed)
                data['negotiated_speed_bps'] = int(speed)

    return data


def get_lagg_info(raw_ifconfig_output):
    lagg_data = {
        'lagg_proto': None,
        'lagg_hash': None,
        'lagg_members': [],
        'lagg_flags': None
    }

    flags_regex = r'flags=\S+<(\S+)>'
    lines = raw_ifconfig_output.split('\n')
    for index, line in enumerate(lines):
        if 'laggproto' in line:
            _, laggproto, _, lagghash = line.split()
            lagg_data['lagg_proto'] = laggproto
            lagg_data['lagg_hash'] = lagghash
        elif 'laggport' in line:
            interface = line.split()[1]
            lagg_data['lagg_members'].append(interface)
        elif 'lagg options:' in line:
            if 'flags=' in lines[index + 1]:
                m = re.search(flags_regex, lines[index + 1])
                lagg_data['lagg_flags'] = m.group(1).lower().split(',')

    return lagg_data


def get_interface_flags(raw_ifconfig_output):
    data = None

    flag_regex = r'flags=\S+<(\S+)>'

    for line in raw_ifconfig_output.split('\n'):
        if ': flags=' in line:
            line = line.lower()
            m = re.search(flag_regex, line)

            data = m.group(1).split(',')

    return data


def get_interface_options(raw_ifconfig_output):
    data = {
        'options': None,
        'nd6_options': None
    }

    option_regex = r'options=\S+<(\S+)>'

    for line in raw_ifconfig_output.split('\n'):
        if 'options=' in line:
            line = line.lower()
            m = re.search(option_regex, line)
            if 'nd6 options=' in line:
                data['nd6_options'] = m.group(1).split(',')
            else:
                data['options'] = m.group(1).split(',')

    return data


def get_default_route_iface():
    default_interface = None

    netstat_default_route = ['netstat', '-4r']

    netstat_output = _run_cmd(netstat_default_route)

    for line in netstat_output.split('\n'):
        if line.startswith('default'):
            default_interface = line.split()[-1]

    return default_interface


def get_cc_max_speed(interface_number):
    key = 'dev.cc.{}.max_speed'.format(interface_number)
    sysctl_max_speed = ['sysctl', key]
    speed = _run_cmd(sysctl_max_speed).split()[-1]
    speed = float(speed) * 1e9

    return int(speed)


def get_interface_list():
    list_interfaces = ['ifconfig', '-l']
    interface_list = _run_cmd(list_interfaces).split()
    return interface_list


def get_interface_data():
    # Get all interfaces
    interface_list = get_interface_list()
    interfaces = {}

    for interface in interface_list:
        interfaces[interface] = {
            'gateway': False
        }

        # Get raw output of the interface for parsing
        raw_ifconfig_output = get_ifconfig_output(interface)
        interfaces[interface]['raw'] = raw_ifconfig_output

        # Get mac addresses
        interfaces[interface].update(get_mac_address(raw_ifconfig_output))

        # Get nic status
        interfaces[interface]['status'] = get_interface_status(raw_ifconfig_output)

        # Parse media output
        interfaces[interface].update(get_media_data(raw_ifconfig_output))

        # Detect lagg
        if 'lagg' in raw_ifconfig_output:
            interfaces[interface].update(get_lagg_info(raw_ifconfig_output))

        # Chelsio NICs currently have a bug where they dont report the
        # media info correctly if they have a different speed optic.
        if re.search('cc[0-9]+', interface):
            interfaces[interface]['negotiated_speed_bps'] = get_cc_max_speed(
                interface[2:]
            )

        # Get interface flags
        interfaces[interface]['flags'] = get_interface_flags(raw_ifconfig_output)

        # Get interface options
        interfaces[interface].update(get_interface_options(raw_ifconfig_output))

    # Process lagg interface
    for interface in interfaces:
        if 'lagg' in interface and 'lagg_members' in interfaces[interface]:
            lagg_negotiated_speed_bps = 0
            lagg_potential_speed_bps = 0

            # Add the members of the laggs speed to the lagg
            for lagg_member in interfaces[interface]['lagg_members']:
                lagg_negotiated_speed_bps += interfaces[lagg_member]['negotiated_speed_bps']
                lagg_potential_speed_bps += interfaces[lagg_member]['potential_speed_bps']

            interfaces[interface]['negotiated_speed_bps'] = lagg_negotiated_speed_bps
            interfaces[interface]['potential_speed_bps'] = lagg_potential_speed_bps

    # Get default route interface. If its a lagg tag the members as the gateway as well
    default_interface = get_default_route_iface()

    if 'lagg' in default_interface and 'lagg_members' in interfaces[default_interface]:
        for lagg_member in interfaces[default_interface]['lagg_members']:
                interfaces[lagg_member]['gateway'] = True

    interfaces[default_interface]['gateway'] = True

    return interfaces


def read_args():
    parser = argparse.ArgumentParser(
        description="Command line tool for purge api actions"
    )

    parser.add_argument(
        '-f',
        '--format',
        choices=['json', 'zabbix'],
        default='json',
        help='Set the output format. json(default), zabbix pipe/comma seperated values).'
    )
    parser.add_argument(
        '-k',
        '--keys',
        help='Comma seperated list of items to return. By default all keys are returned.  Exmaple: status,duplex'
    )
    parser.add_argument(
        '-i',
        '--interface',
        help='Comma seperated list of interfaces to return data for. By default all interfaces are returned.'
    )
    return parser.parse_args()


def filter_keys(data, filter, separator=','):
    filter_keys = filter.split(separator)
    filtered_data = {}

    for interface in data:
        filtered_data[interface] = {}

        for key in filter_keys:
            if key in data[interface]:
                filtered_data[interface][key] = data[interface][key]

    return filtered_data


def filter_interface(data, interface_list, separator=','):
    filtered_interfaces = interface_list.split(separator)
    filtered_data = {}

    for interface in data.keys():
        if interface in filtered_interfaces:
            filtered_data[interface] = data[interface]

    return filtered_data


def print_formated(data, output_format):
    if output_format == 'json':
        print(json.dumps(data, indent=4))

    elif output_format == 'zabbix':
        for interface in data:
            for key in data[interface]:
                if key in ['raw']:
                    continue

                value = data[interface][key]
                if not value or value == 'null':
                    value = ''

                if isinstance(value, list):
                    value = ' '.join(value)

                print('{},{}|{}'.format(
                    interface,
                    key,
                    value
                ))


def main():
    # Read all input arguments
    args = read_args()

    # Get all the interface data
    d = get_interface_data()

    # Process interface filter
    if args.interface:
        d = filter_interface(d, args.interface)

    # Process key filter
    if args.keys:
        d = filter_keys(d, args.keys)

    # Print collected data
    print_formated(d, args.format)


if __name__ == "__main__":
    main()
