#!/bin/bash

usage(){
    echo "usage: $0 [options]"
    echo "Options:"
    echo "  -d     dryrun output instead of sending to TSDB  "
    echo "Example:"
    echo "  $0 -d"
    exit 1
}

while getopts ":d" opt; do
  case ${opt} in
      d)
        dryrun=1
        ;;
      *)
        usage
        ;;
    esac
done

# Script Vars
#####################################################################

# TSDB vars
TSDB_HOST="tsdb-data.lldns.net"
TSDB_PORT="4242"
TSDB_DATA=""
TSDB_TIMEOUT=10

# ms-ops file
MSOPS_HOSTS=$(grep -hEo 'ms-ops-[0-9]+\.[a-z0-9]+\.llnw\.net' /usr/local/llnw/ms-ops-hosts.json)

# Ericsson sites
ERICSSON_HOSTS=$(grep -hEo 'cds[0-9]+\.[a-z0-9]+' /usr/local/llnw/ericsson-hosts.json | cut -d'.' -f2 | sort -u | awk '{print "healthd."$1".llnw.net"}')

# Ping command 
test_ping="ping -W 5 -c 10"

# Define source pop
spop=$(uname -n | cut -d'.' -f2)

#####################################################################

getTime() {
    # Get the time and the time aligned to the rearest minute
    # This is needed for grafana and tsdb to graph and calulate the 
    # data for all cds servers at the same minute.
    now=$(date +%s)
    now_aligned=$(( now - now % 60 ))

    echo $now_aligned
}

getPingStats() {
    
    # ms-ops host to ping
    local host=$1
    local dpop=$(echo $host | cut -d'.' -f2)
    local ping_result

    ping_result=$($test_ping $host)

    tmpIFS=$IFS
    IFS=$'\n'
    for line in $(echo "$ping_result");
    do 
        # Parse packet loss stat
        # 2 packets transmitted, 2 packets received, 0.0% packet loss, 2 packets out of wait time
        reg_packet_loss='([0-9]+)% packet loss'
        if [[ $line =~ $reg_packet_loss ]]; then
            sendMetric 'app.ping.packet_loss' ${BASH_REMATCH[1]} $dpop
        fi
        
        # Parse the stat fields for min,max,avg and mdev
        # round-trip min/avg/max/stddev = 261.447/261.472/261.497/0.025 ms
        reg_stats='^round-trip .*stddev = ([0-9\.]+)\/([0-9\.]+)\/([0-9\.]+)\/([0-9\.]+) .*$'
        if [[ $line =~ $reg_stats ]]; then
            sendMetric 'app.ping.rtt.min' ${BASH_REMATCH[1]} $dpop
            sendMetric 'app.ping.rtt.avg' ${BASH_REMATCH[2]} $dpop
            sendMetric 'app.ping.rtt.max' ${BASH_REMATCH[3]} $dpop
            sendMetric 'app.ping.rtt.mdev' ${BASH_REMATCH[4]} $dpop
        fi
    done
    IFS=$tmpIFS
}

sendMetric() {
    local key=$1
    local value=$2
    local dpop=$3

    if [ -n "${dryrun}" ]; then
         echo -ne "put $key $(getTime) $value source=$spop dest=$dpop\n"
    else
         echo -ne "put $key $(getTime) $value source=$spop dest=$dpop\n" | nc -w $TSDB_TIMEOUT $TSDB_HOST $TSDB_PORT
    fi
   
}

for host in $ERICSSON_HOSTS $MSOPS_HOSTS; do
    # Launch tests 
    getPingStats $host &

    # Seperate tests by 1/2 second
    sleep .5
done

# Wait for backkground tests to finish
wait