#!/usr/bin/env bash
PATH="/sbin:/bin:/usr/bin:/usr/sbin:/usr/local/sbin:/usr/local/bin"

int_mtu="1600"
minimum_speed="1000"

# Temp for SRE-5386
PARENT_COMMAND=$(ps $PPID | tail -n 1 | sed -e "s/[[:space:]]\{1,\}/ /g"  | cut -d' ' -f5-)
/usr/bin/logger "bond-ethernet-no-neg.sh called by: " $PARENT_COMMAND

set_route_mtu(){
  ext_nets_v4=$(netstat -anrW -f inet | awk "\$5 == $int_mtu {print \$1}")
  ext_nets_v6=$(netstat -anrW -f inet6 | awk "\$5 == $int_mtu {print \$1}")

  for net4 in $ext_nets_v4; do
    net_gateway=$(route -n get -inet $net4 | grep gateway: | awk '{print $NF}')
    if [[ $net_gateway ]]; then
      route change -inet $net4 $net_gateway -mtu 1500 > /dev/null
    else
      net_interface=$(route -n get -inet $net4 | grep interface: | awk '{print $NF}')
      route change -inet $net4 -interface $net_interface -mtu 1500 > /dev/null
    fi
  done

  # If we didn't collect v6 nets initially, give v6 a few seconds to come up
  if [ ! "${ext_nets_v6}" ]; then
    sleep 5
    ext_nets_v6=$(netstat -anrW -f inet6 | awk "\$5 == $int_mtu {print \$1}")
  fi

  for net6 in $ext_nets_v6; do
    net6_gateway=$(route -n get -inet6 $net6 | grep gateway: | awk '{print $NF}')
    if [[ $net6_gateway ]]; then
      route change -inet6 $net6 $net6_gateway -mtu 1500 > /dev/null
    else
      net6_interface=$(route -n get -inet6 $net6 | grep interface: | awk '{print $NF}')
      route change -inet6 $net6 -interface $net6_interface -mtu 1500 > /dev/null
    fi
  done
}

set_interface_mtu(){
  if [ "${has_lagg}" -eq 0 ]; then
    ifconfig ${1} up mtu ${int_mtu}
  fi
}

remove_v6_rtadv(){
  ndp -i ${1} -- -accept_rtadv > /dev/null 2>&1
  ndp -i ${1} -- -auto_linklocal > /dev/null 2>&1
}

#load variables from rc.conf
. /etc/rc.conf

#set the sysctl so rtsol isn't bugged
#sysctl -w net.inet6.ip6.accept_rtadv=1

if ifconfig -l | grep -q lagg; then
  has_lagg=1
else
  has_lagg=0
fi

get_interface_info(){
  declare -gA interface_info
  declare -gA ip_addrs

  for int in `ifconfig -l` ; do
    if [[ ${int} =~ gif[0-9]|lo[0-9] ]]; then
      #skip gif and loopback
      continue
    fi

    if egrep -iq "^ifconfig_${int}=" /etc/rc.conf; then
      primary_interface="${int}"
      set_interface_mtu ${primary_interface}
    fi
  done

  case ${primary_interface} in
    igb[0])
      net="$ifconfig_igb0"
      secondary_interface="igb1"
      remove_v6_rtadv ${secondary_interface}
      set_interface_mtu ${secondary_interface}
    ;;
    igb[1])
      net="$ifconfig_igb1"
      secondary_interface="igb0"
      remove_v6_rtadv ${secondary_interface}
      set_interface_mtu ${secondary_interface}
    ;;
    em[0])
      net="$ifconfig_em0"
      secondary_interface="em1"
      remove_v6_rtadv ${secondary_interface}
      set_interface_mtu ${secondary_interface}
    ;;
    em[1])
      net="$ifconfig_em1"
      secondary_interface="em0"
      remove_v6_rtadv ${secondary_interface}
      set_interface_mtu ${secondary_interface}
    ;;
    ix[0])
      net="$ifconfig_ix0"
      secondary_interface="ix1"
      remove_v6_rtadv ${secondary_interface}
      set_interface_mtu ${secondary_interface}
    ;;
    ix[1])
      net="$ifconfig_ix1"
      secondary_interface="ix0"
      remove_v6_rtadv ${secondary_interface}
      set_interface_mtu ${secondary_interface}
    ;;
    cc[0])
      net="$ifconfig_cc0"
      secondary_interface="cc1"
      remove_v6_rtadv ${secondary_interface}
      set_interface_mtu ${secondary_interface}
    ;;
    cc[1])
      net="$ifconfig_cc1"
      secondary_interface="cc0"
      remove_v6_rtadv ${secondary_interface}
      set_interface_mtu ${secondary_interface}
    ;;
    cxl[0])
      net="$ifconfig_cxl0"
      secondary_interface="cxl1"
      remove_v6_rtadv ${secondary_interface}
      set_interface_mtu ${secondary_interface}
    ;;
    cxl[1])
      net="$ifconfig_cxl1"
      secondary_interface="cxl0"
      remove_v6_rtadv ${secondary_interface}
      set_interface_mtu ${secondary_interface}
    ;;
    mce[0])
      net="$ifconfig_mce0"
      secondary_interface="mce1"
      remove_v6_rtadv ${secondary_interface}
      set_interface_mtu ${secondary_interface}
    ;;
    mce[1])
      net="$ifconfig_mce1"
      secondary_interface="mce0"
      remove_v6_rtadv ${secondary_interface}
      set_interface_mtu ${secondary_interface}
    ;;
    mlxen[0])
      net="$ifconfig_mlxen0"
      secondary_interface="mlxen1"
      remove_v6_rtadv ${secondary_interface}
      set_interface_mtu ${secondary_interface}
    ;;
    mlxen[1])
      net="$ifconfig_mlxen1"
      secondary_interface="mlxen0"
      remove_v6_rtadv ${secondary_interface}
      set_interface_mtu ${secondary_interface}
    ;;
  esac

  sleep 5

  #add ipv6 to primary interface and enable (fixes a problem introduced
  #with ipv6_activate_all_interfaces option, which doesn't set
  #accept_rtadv or clear ifdisabled)
  ndp -i ${primary_interface} accept_rtadv
  ndp -i ${primary_interface} -- -disabled 

  for int in `ifconfig -l` ; do
    if [[ ${int} =~ gif[0-9]|lo[0-9] ]]; then
      #skip gif and loopback
      continue
    fi

    #remove v6 from everything except the primary interface
    if [ ${int} != ${primary_interface} ] && [ ${int} != ${secondary_interface} ]; then
      remove_v6_rtadv ${int}
    fi

    OLDIFS=$IFS
    IFS=$'\n'

    current_lagg_use_flowid=0
    for line in `ifconfig -v $int`; do
      if [[ "${line}" =~ mtu.[0-9]+ ]]; then
        interface_info[${int}_mtu]=`echo $line | awk '{print $NF}'`
      elif [[ "${line}" =~ inet.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3} ]]; then
        address=`echo $line | awk '{print $2}'`
        subnet=`echo $line | awk '{print $4}'`
        ip_addrs[${address}]=${subnet}
      elif [[ "${line}" =~ "ether:" ]]; then
        echo $line | awk '{print $NF}' > /etc/${int}.mac
      elif [[ "${line}" =~ "25GBase" ]]; then
        interface_info[${int}_speed]=25000
      elif [[ "${line}" =~ "25Gbase" ]]; then
        interface_info[${int}_speed]=25000
      elif [[ "${line}" =~ "10Gbase" ]]; then
        interface_info[${int}_speed]=10000
      elif [[ "${line}" =~ "1000base" ]]; then
        interface_info[${int}_speed]=1000
      elif [[ "${line}" =~ "100base" ]]; then
        interface_info[${int}_speed]=100
      elif [[ "${line}" =~ "status: active" ]]; then
        interface_info[${int}_status]=active
      elif [[ "${line}" =~ "status:" ]]; then
        interface_info[${int}_status]=not_active
      elif [[ "${line}" =~ "laggproto" ]]; then
        current_lagg_proto=`echo $line | awk '{print $2}'`
        current_lagg_hash=`echo $line | awk '{print $4}'`
      elif [[ "${line}" =~ "USE_FLOWID" ]]; then
        current_lagg_use_flowid=1
      fi
    done
    IFS=$OLDIFS
  done
}

error_check(){
  #set in /etc/rc.conf
  if [ "${lacp_enable}" ]; then
    lagg_proto="lacp"
  else
    lagg_proto="loadbalance"
  fi

  # set sane lagg_hash default if not set in /etc/rc.conf
  if [ "${lagg_hash}" == "" ]; then
    lagg_hash="l3,l4"
  fi

  # set sane lagg_use_flowid default if not set in /etc/rc.conf
  if [ "${lagg_use_flowid}" == "" ]; then
    lagg_use_flowid=0
  fi

  # convert flag to ifconfig command
  if [ "${lagg_use_flowid}" == "0" ]; then
    use_flowid_cmd="-use_flowid"
  else
    use_flowid_cmd="use_flowid"
  fi

  #See if primary and secondary ints have matching speeds
  if [ "${interface_info[${primary_interface}_speed]}" != "${interface_info[${secondary_interface}_speed]}" ]; then
    echo "${primary_interface} and ${secondary_interface} do not match in negotiated speed"
    echo "skipping bonding, setting MTU, and exiting"
    set_route_mtu
    exit 1
  fi

  #See if primary int is 1gig or faster, above ^^ check makes sure they match
  if [ "${interface_info[${primary_interface}_speed]}" -lt "${minimum_speed}" ]; then
    echo "Interfaces are negotiated at 100Mbit or less, skipping bond, setting MTU, and exiting"
    set_route_mtu
    exit 1
  fi

  #If no lagg detected, run the bondcheck to see if they can be lagged
  if [ "${lagg_proto}" != "lacp" ]; then
    if [ "${has_lagg}" -eq 0 ]; then
      if ! /usr/local/bin/ethbondcheck ${primary_interface} ${secondary_interface}; then
        echo "seems the ports aren't properly configured, possibly they aren't in the same VLAN?"
        set_route_mtu
        exit 1
      fi
    fi
  fi

  #If there is a lagg built, but the current balance protocol doesn't match what's needed/wanted,
  #we have to tear it down and start from scratch.
  if [ "${has_lagg}" -eq 1 ]; then
    if [ "${lagg_proto}" != "${current_lagg_proto}" ]; then
      ifconfig lagg0 destroy
      /etc/netstart
      sleep 4
      exec $0
    fi
    #If lagg exists and interface mtus are not 1600, we need to tear down and rebuild.
    if [ "${interface_info[${primary_interface}_mtu]}" -ne "${int_mtu}" ] || [ "${interface_info[${secondary_interface}_mtu]}" -ne "${int_mtu}" ]; then 
      echo "Interface MTUs are not set to ${int_mtu}, must destroy lagg and rebuild"
      ifconfig lagg0 destroy
      /etc/netstart
      sleep 4
      exec $0
    fi
    # If lagg exists and lagghash is incorrect, just change it
    if [ "${current_lagg_hash}" != "${lagg_hash}" ]; then
      echo "lagghash is '${current_lagg_hash}', should be '${lagg_hash}', fixing"
      ifconfig lagg0 lagghash ${lagg_hash}
    fi
    # If lagg exists and use_flowid is wrong, just change it
    if [ "${current_lagg_use_flowid}" != "${lagg_use_flowid}" ]; then
      echo "lagg0 use_flowid is '${current_lagg_use_flowid}', should be '${lagg_use_flowid}', fixing"
      ifconfig lagg0 "${use_flowid_cmd}"
    fi
  fi
}

set_ipv6_things(){
  net6_0=`ifconfig ${primary_interface} | awk '/inet6/ {print $2}'`
  net6_1=`ifconfig ${secondary_interface} | awk '/inet6/ {print $2}'`

  if [ "${net6_0}" ] || [ "${net6_1}" ]; then
    for net6addr_0 in $net6_0; do
      ifconfig ${primary_interface} inet6 $net6addr_0 delete
      ndp -d $net6addr_0 > /dev/null 2>&1
    done

    for net6addr_1 in $net6_1; do
      ifconfig ${secondary_interface} inet6 $net6addr_1 delete
      ndp -d $net6addr_1 > /dev/null 2>&1
    done

    ndp -c
    route delete -inet6 default > /dev/null 2>&1
  fi
}

build_the_lagg(){
  net=`echo $net |sed -e 's/inet //' -e 's/ polling//'`

  # bypass freebsd lagg build crash bug
  /sbin/ifconfig "${primary_interface}" down
  /sbin/ifconfig "${secondary_interface}" down

  if [ ! "${in_memory_lagg}" ]; then
    kldload if_lagg > /dev/null 2>&1
  fi

  if [ "${has_lagg}" -eq 0 ]; then
    for ip in ${!ip_addrs[@]}; do
      ifconfig "${primary_interface}" "${ip}" -alias >/dev/null 2>&1
    done

    ifconfig lagg0 create
    sleep 2
    ifconfig lagg0 up laggproto "${lagg_proto}" lagghash "${lagg_hash}" "${use_flowid_cmd}" laggport "${primary_interface}" laggport "${secondary_interface}" $net
    ndp -i lagg0 accept_rtadv
    sleep 1
    rtsol lagg0

    # bypass freebsd lagg build crash bug
    /sbin/ifconfig "${primary_interface}" up
    /sbin/ifconfig "${secondary_interface}" up
    /bin/sleep 1

    for ip in ${!ip_addrs[@]}; do
      ifconfig lagg0 ${ip} netmask ${ip_addrs[${ip}]} alias
    done

    route add default ${defaultrouter}
  fi
}

check_connectivity(){
  # This connectivity check causes issues for our loadbalanced lagg
  # It seems they can take a while to initialize in rare cases
  if [ "${lagg_proto}" != "lacp" ]; then
    return
  fi

  #if we can't ping our gateway after setting up the lagg, rip it out
  ping -c 20 -W 500 -t 4 -Qqno $defaultrouter > /dev/null

  if [ "$?" != "0" ]; then
    # Slow negotiations can be caused by varying switch configs
    # If we failed the first attempt wait and try again before tearing down
    sleep 10
    ping -c 20 -W 500 -t 4 -Qqno $defaultrouter > /dev/null
    if [ "$?" != "0" ]; then
      for int in `ifconfig -l`; do
        if [ "${int}" = "lagg0" ]; then
          /etc/rc.d/netif stop
          sleep 1
          ifconfig lagg0 destroy
          sleep 2
          /etc/netstart
          sleep 2
          remove_v6_rtadv ${secondary_interface}
          ndp -i ${primary_interface} accept_rtadv
          ndp -i ${primary_interface} auto_linklocal
          rtsol ${primary_interface}
          set_route_mtu
          return
        fi
      done
      /etc/rc.d/netif stop
      sleep 1
      /etc/netstart
      sleep 2
      remove_v6_rtadv ${secondary_interface}
      ndp -i ${primary_interface} accept_rtadv
      ndp -i ${primary_interface} auto_linklocal
      rtsol ${primary_interface}
      set_route_mtu
    fi
  fi
}

get_interface_info
error_check
set_ipv6_things
build_the_lagg
set_route_mtu
check_connectivity
