#!/usr/bin/env perl

use Fcntl ':mode';

use strict;
use warnings;

# calculate size target based off of percent of total space according to df
# for example, 0.7 would mean, if the target partition is greater than 70% full,
# delete files until the percentage would be no greater than 70%
our $max_pct = 0.7;

# max_cap only comes into play if df fails for any reason.  In that case,
# we limit the data in the list of files selected by $glob to be no greater
# than this value.  The units are in bytes.
our $max_cap = 512 * 1024 * 1024;

our $partition = "/logs";

# these are the files subject to deletion
our $glob = "/logs/*/old/*";

# use -Pk options to get consistent output across both FreeBSD and Linux
our $df = "df -Pk $partition";



# get the desired quantity of bytes to delete
my $target_bytes = &get_target();


if ($target_bytes == -1) {
	# df failed, resort to capping to $max_cap
	&cap_files($max_cap);
} else {
	&delete_files($target_bytes);
}


sub delete_files {
	my $target = shift;
	my $files = &get_files;

	my @files = sort {$files->{$a}->[9] <=> $files->{$b}->[9]} keys %{$files};

	for my $fname (@files) {
		if ($target > 0) {
			my $size = $files->{$fname}->[7];
			$target -= $size;
			&unlink($fname);
		} else {
			last;
		}
	}
}

sub cap_files {
	my $target = shift;
	my $files = &get_files;

	my @files = sort {$files->{$b}->[9] <=> $files->{$a}->[9]} keys %{$files};

	for my $fname (@files) {
		if ($target > 0) {
			my $size = $files->{$fname}->[7];
			$target -= $size;
		} else {
			&unlink($fname);
		}
	}
}

# returns reference to hash of file name and the result of a stat
sub get_files {
	my %files;
	for my $fname (glob $glob) {
		my @stat = stat $fname;
		next if (scalar(@stat) < 10);
		next unless $stat[2] & S_IFREG;
		$files{$fname} = \@stat;
	}
	return \%files;
}

sub get_target {
	open(DF, "-|", $df) || return $max_cap;
	if ($?) {
		close DF;
		return -1;
	}

	my $line;

	while ($line = <DF>) {
		if ($line =~ /^\s*\S+\s+(\d+)\s+(\d+)\s.*/) {
			close DF;
			my $target = $1 * $max_pct * 1024;
			my $usage = $2 * 1024;
			my $delete = $usage - $target;
			close DF;
			if ($delete > 0) {
				return $delete;
			} else {
				return 0;
			}
		}
	}
	close DF;
	return -1;
}

sub unlink {
	my $fname = shift;

	unlink $fname;
}
