#!/usr/bin/perl

# Collect squid stats from localhost, generate metrics, and feed to monitoring MQ bus

use Data::Dumper;
use Sys::Syslog qw(:DEFAULT setlogsock);  # default set, plus setlogsock()
use Net::RabbitMQ;
use JSON; my $json = new JSON; $json->allow_nonref;
use strict;
use POSIX;

use Net::HTTP;

my %hostdata;
my $diskid;
my @disklist;
my @diskdata;
my @info = uname();

my $self = $info[1];

# silence stdout/stderr if not run from command line
if (! -t STDIN) { close STDOUT;close STDERR }

sleep(rand(60));

@disklist = `/bin/cat /etc/fstab | /usr/bin/egrep "noatime"`;

my %queueBinds = (
        'events'        => { Channel => 1,  Exchange => '/exchange/events', Queue => '/events', Key => 'events' },
        'alerts'        => { Channel => 2,  Exchange => '/exchange/alerts', Queue => '/alerts', Key => 'alerts' },
        'metrics'       => { Channel => 3,  Exchange => '/exchange/metrics', Queue => '/metrics', Key => 'metrics'},
        'heartbeat'     => { Channel => 4,  Exchange => '/exchange/events', Queue => '/heartbeat', Key => 'heartbeat'},
        'inventory'     => { Channel => 5,  Exchange => '/exchange/inventory', Queue => '/inventory', Key => 'inventory'},
        'traffic'       => { Channel => 6,  Exchange => '/exchange/traffic', Queue => '/traffic', Key => 'traffic'}
);

my $mq = connectMQ('billn', 'testing', 'anymq.llnw.com');

my %funcs = (
        storedir => \&storedir,
        utilization => \&utilization
);

if ( my @data = get($self, 80, 'info') ) {
                my %info = info(@data);

                printf "%s: %s\n", $self, $info{version};
                exit unless $info{version} =~ /^3\.8/;

                foreach my $port (80, 8000, 8001) {
                        print "Calling $self:$port\n";

                        my $req;
                        foreach my $q (keys %funcs) {
                                my @results = get($self, $port, $q);
                                &{$funcs{$q}}($self, $port, @results);
                        }
                }
}

#disk_stats();

# SQ stats fetch pieces
sub get {
        my ($host, $port, $func) = @_;

	print "Get: $host, $port, $func\n";

        if ( my $s = Net::HTTP->new(Host => $host, LocalHost => '127.0.0.1') ) {
        	$s->write_request(
			'GET' 		=> "cache_object://mycache99.example.com/$func", 
        		'Authorization'	=> "Basic c3F1aWQ6c3F6emFjbw==", # /usr/local/bin/b64enc squid:sqzzaco
			'httpVersion'	=> 'HTTP/1.0',
			'User-Agent' 	=> "httpSQ/1.0"
		);

        	my($code, $mess, %h) = $s->read_response_headers;

		my $response;
        	while (1) {
           		my $buf;
           		my $n = $s->read_entity_body($buf, 1024);
           		warn "read failed: $!" unless defined $n;
           		last unless $n;

			$response .= $buf;
        	}
		return(split(/\n/, $response));
	}
	else {
		sendlog($@);
        	return(0);
	}
}

sub info {
        my @input = @_;

        my %objectHash;

        my $lock;
        my $section;
        my $data;
        foreach my $line (@input) {
                $line =~ s/[\r\n]//g;
                chomp $line;

                if ( my @foo = $line =~ /^(Connection information|Cache information|Median Service Times|Resource usage|Memory accounted|File descriptor usage)/ ) {
                        $section = $1;

#                       print "Locking on section $section\n";
                        $lock++;
                }
                elsif($lock && $line =~ /^\s*$/) {     # release lock, end of stanza
                        $lock = 0;
                        $section = undef;
                }
                elsif($lock && $section) {           # process line
                        $line =~ s/^\s+(.*)$/$1/;
                        my ($key, $data) = $line =~ /^\s*(.*):\s+(.*)$/;
                        $objectHash{$section}{$key} = $data;
                }
        }

        foreach my $line (@input) {
                $line =~ s/[\r\n]//g;
                chomp $line;

                if ( my @foo = $line =~ /^(Internal Data Structures)/ ) {
                        $section = $1;

#                        print "Locking on section $section\n";
                        $lock++;
                }
                elsif($lock && $line =~ /^\s*$/) {     # release lock, end of stanza
                        $lock = 0;
                        $section = undef;
                }
                elsif($lock && $section) {           # process line
                        $line =~ s/^\s+(.*)$/$1/;
                        my ($data, $key) = $line =~ /^\s*(\d+)\s+(.*)$/;
                        $objectHash{$section}{$key} = $data;
                }
        }

        foreach my $line (@input) {
                if( $line =~ /Squid Object Cache: Version (\S+)/ ) {
                        $objectHash{version} = $1;
                }

                if( $line =~ /Version info: (\S+)/ ) {
                        $objectHash{'Version info'} = $1;
                }

                if( $line =~ /Start Time:\s+(.*)/ ) {
                        $objectHash{'Start Time'} = $1;
                }
        }

        return(%objectHash);
}

sub storedir {
        my ($host, $port, @input) = @_;

        my %metricConversion = (
                'Age' => "EdgePrism:$port Cache Age",
                'Percent Files Used' => "EdgePrism:$port Store File Usage",
                'Percent Size Used' => "EdgePrism:$port Store Partition Capacity",
                'Filemap bits in use' => "EdgePrism:$port Store Filemap Usage",
                'Filesystem Inodes in used' => "EdgePrism:$port Store Inode Util",
                'Pending operations' => "EdgePrism:$port Store Pending Operations"
        );

        my %transform = (
                'Percent Size Used' => '(\S+)\%',
                'Percent Files Used' => '(\S+)\%',
                'Filemap bits in use' => '\((\d+)\%\)',
        );

        my %objectHash;

        my $partition;
        my $lock;

        foreach my $line (@input) {
                $line =~ s/[\r\n]//g;
                chomp $line;

                if ( my @foo = $line =~ /^Store Directory \#\d+ \(\w+\): (\S+)/ ) {      # match the start of an disk store stanza: Store Directory #0 (diskd): /d0/sqcache1
                        $partition = $1;
                        $lock++;
                        print "Found $partition - $line\n";
                }
                elsif($lock && $line =~ /^\s*$/) {     # release lock, end of stanza
                        $lock = 0;
                        $partition = undef;
                }
                elsif($lock && $partition) {           # process line
                        $line =~ s/^\s+(.*)$/$1/;
                        my ($key, $data) = split(/:\s*/, $line);
                        $objectHash{$partition}{$key} = $data;

                        print "Storing $partition, $key: $data\n";
                }


                print $line . "\n";
        }

        foreach my $partition ( keys %objectHash ) {
                foreach my $key ( keys %{$objectHash{$partition}} ) {
                        next unless $metricConversion{$key};
                        my $value;

                        if($transform{$key}) {
                                ($value) = $objectHash{$partition}{$key} =~ $transform{$key};
                        #       print "Transform regex: $transform{$key} found '$value'\n";
                        }
                        else {
                                $value = $objectHash{$partition}{$key};
                        }

                        print "Metric: $host, $key -> $metricConversion{$key}, $partition, $value\n";

                        metric( $mq, (
                                Timestamp => time,
                                Host => $host,
                                MetricName => $metricConversion{$key},
                                InstanceName => $partition,
                                Value => $value,
                                Available => 1,
                        ));
                }
        }
}

sub utilization {
        my ($host, $port, @input) = @_;

        my %metricConversion = (
                'client_http.requests' => "EdgePrism:$port Client HTTP Requests",
                'client_http.hits' => "EdgePrism:$port Client HTTP Cache Hits",
                'client_http.errors' => "EdgePrism:$port Client HTTP Errors",
                'diskd.backoff_drain_some' => "EdgePrism:$port Disk Drain Some",
                'diskd.backoff_drain_all' => "EdgePrism:$port Disk Drain All",
                'diskd.read.ops' => "EdgePrism:$port Disk Read Ops",
                'diskd.read.fail' => "EdgePrism:$port Disk Read Failures",
                'diskd.read.bytes' => "EdgePrism:$port Disk Bytes Read",
                'diskd.write.ops' => "EdgePrism:$port Disk Write Ops",
                'diskd.write.fail' => "EdgePrism:$port Disk Write Failures",
                'diskd.write.bytes' => "EdgePrism:$port Disk Bytes Written"
        );

        my %transform = (
                'client_http.requests' => '(\S+)\/sec',
                'client_http.hits' => '(\S+)\/sec',
                'client_http.errors' => '(\S+)\/sec',
                'diskd.backoff_drain_some' => '(\S+)\/sec',
                'diskd.backoff_drain_all' => '(\S+)\/sec',
                'diskd.read.ops' => '(\S+)\/sec',
                'diskd.read.fail' => '(\S+)\/sec',
                'diskd.read.bytes' => '(\S+)\/sec',
                'diskd.write.ops' => '(\S+)\/sec',
                'diskd.write.fail' => '(\S+)\/sec',
                'diskd.write.bytes' => '(\S+)\/sec',
        );

        my %objectHash;

        my $section;
        my $lock;

        foreach my $line (@input) {
                $line =~ s/[\r\n]//g;

                chomp $line;
                if ( my @foo = $line =~ /^(Last.*):$|(Totals) since cache startup/ ) {
                        $section = $1;

                        #print "Locking on section $section\n";
                        $lock++;
                }
                elsif($lock && $line =~ /^\s*$/) {     # release lock, end of stanza
                        $lock = 0;
                        $section = undef;
                }
                elsif($lock && $section) {           # process line
                        $line =~ s/^\s+(.*)$/$1/;
                        my ($key, $data) = split(/ = /, $line);
                        $objectHash{$section}{$key} = $data;
                }
        }

        $section = 'Last 5 minutes';
        foreach my $key ( keys %{$objectHash{$section}} ) {
                next unless $metricConversion{$key};
                my $value;

                if($transform{$key}) {
                        ($value) = $objectHash{$section}{$key} =~ $transform{$key};
#                       print "Transform regex: $transform{$key} found '$value'\n";
                }
                else {
                        $value = $objectHash{$section}{$key};
                }

                print "Metric: $host, $key -> $metricConversion{$key}, $section, $value\n";
                metric( $mq, (
                                Timestamp => time,
                                Host => $host,
                                MetricName => $metricConversion{$key},
                                InstanceName => '',
                                Value => $value,
                                Available => 1,
                ));
        }
}


sub disk_stats
{

	foreach my $disk (@disklist)
	{

		if ( $disk =~ /\/dev\/(...)/ )
		{

			$diskid = $1;
			$hostdata{$self}{$diskid}{DiskID} = $diskid;

                	# Debug for the moment
                	print "Disk ID: $diskid\n";

       		        # Fetch our smartctl output
	                @diskdata = `/usr/local/sbin/smartctl --attributes --info --log=error /dev/$diskid`;

               		# default some data that not all drives will return
           	     	$hostdata{$self}{$diskid}{Hours} = 0;

               		# loop through the diskdata
               		foreach my $line (@diskdata)
               		{

				# grab device info
                      		 send_out_data($diskid, "Drive Temperature", $1) if ( $line =~ /Current Drive Temperature:\s+(\d+)/);
                       		if ($line =~ /number of hours powered up = (\S+)/)
                       		{
                               		if ($1) {
                                      	 	send_out_data($diskid, "Drive Runtime", $1);
                               		} else
                               		{
                                      		 send_out_data($diskid, "Drive Runtime", 0);
                               		}
                       		}
				send_out_data($diskid, "Drive Defects", $1) if ($line =~ /Elements in grown defect list: (\d+)/);
	                        # get read data
       		                 if ( $line =~ /read:.*?(\d+).*?(\d+).*?(\d+).*?(\d+).*?(\d+).*?(\d+).*?(\d+)/ )
       		                 {
       		                         send_out_data($diskid, "Corrected Read Errors", $1);
       		                         send_out_data($diskid, "Delayed Read Errors", $2);
       		                         send_out_data($diskid, "ReReads", $3);
       		                         send_out_data($diskid, "Read Total Errors", $4);
       		                         send_out_data($diskid, "Read Correction Invocations", $5);
       		                         send_out_data($diskid, "Data Read", $6);
       		                         send_out_data($diskid, "Read Uncorrected Errors", $7);
       		                 }
       		                 # get write data
       		                 if ( $line =~ /write:.*?(\d+).*?(\d+).*?(\d+).*?(\d+).*?(\d+).*?(\d+).*?(\d+)/ )
       		                 {
       		                         send_out_data($diskid, "Corrected Write Errors", $1);
       		                         send_out_data($diskid, "Delayed Write Errors", $2);
       		                         send_out_data($diskid, "ReWrites", $3);
       		                         send_out_data($diskid, "Write Total Errors", $4);
       		                         send_out_data($diskid, "Write Correction Invocations", $5);
       		                         send_out_data($diskid, "Data Write", $6);
       		                         send_out_data($diskid, "Write Uncorrected Errors", $7);
       		                 }
       		                 # non media errors
       		                 send_out_data($diskid, "Non Medium Error Count", $1) if ( $line =~ /Non-medium error count:\s+(\d+)/ );

       		          }
       		  }
 	}
}


#---------
#MQ and framework support routines
#---------

sub sendlog {
        my($error)= @_;
        warn $error . "\n";
        syslog 'warning', '%s', $error;

        if ($error =~ /went away/i) { # lost db handle  
                die $error;
        }
}

sub connectMQ {
        my ($user, $password, $host) = @_;

        my $mq = Net::RabbitMQ->new();

        print "$0: Attempting connect to $host as $user, $password\n";
        eval { $mq->connect($host, { user => $user, password => $password }); };
	unless ($@) {
                print "$0: Connect successful to $host\n";

        	foreach my $set ('events', 'alerts', 'metrics', 'heartbeat', 'inventory', 'traffic') {


                	my $exchange = $queueBinds{$set}{Exchange};
                	my $queue    = $queueBinds{$set}{Queue};
                	my $key      = $queueBinds{$set}{Key};
                	my $channel  = $queueBinds{$set}{Channel};

			print Dumper($queueBinds{$set});

			print "Opening $channel\n";

                	eval { $mq->channel_open($channel); }; print $@ if defined $@;
                	eval { $mq->queue_bind($channel, $queue, $exchange, $key); }; print $@ if defined $@;
                }
                return($mq);
        }
        else {
                sendlog("MQ connect to $host as $user/$password failed: $@");
        }

        sendlog("MQ connection failed: $@");
        return(0);
}

sub alert {
        my ($mq, %alert) = @_;

        unless(defined $alert{Node} && defined $alert{Summary} && defined $alert{SubMethod} && defined $alert{AlarmGroup} && defined $alert{Unique} && defined $alert{Severity}) {
                warn "Alert is missing elements: " . Dumper(%alert);
                return(0);
        }

        my $datagram = encode_json({ results => scalar keys %alert, %alert });
        eval { $mq->publish(2, 'alerts', $datagram, { exchange => '/exchange/alerts', immediate => 1, mandatory => 1, deliveryMode => 1 }) };  sendlog($@) if $@;
}

sub heartbeat {
        my ($mq, %heartbeat) = @_;

        $heartbeat{Pulse} = 0 unless $heartbeat{Pulse};

        unless(defined $heartbeat{Host} && defined $heartbeat{ProcessName} && defined $heartbeat{Pulse}) {
                warn "Heartbeat is missing elements: " . Dumper(%heartbeat);
                return(0);
        }

        my $datagram = encode_json({ results => scalar keys %heartbeat, %heartbeat });
        eval { $mq->publish(4, 'heartbeat', $datagram, { exchange => '/exchange/events', immediate => 1, mandatory => 1, deliveryMode => 2 }) }; sendlog("$0: $@") if $@;
}

sub metric {
        my ($mq, %metric) = @_;

        unless(defined $metric{Host} && defined $metric{MetricName} && defined $metric{InstanceName} && defined $metric{Value} && defined $metric{Available}) {
                warn "Metric is missing elements: " . Dumper(%metric);
                return(0);
        }

        my $datagram = encode_json({ results => scalar keys %metric, %metric });
        eval { $mq->publish(3, 'metrics', $datagram, { exchange => '/exchange/metrics', immediate => 1, mandatory => 1, deliveryMode => 2 }) }; sendlog($@) if $@;
}


sub send_out_data
{

        # unshift our values
        my ($instance, $metric, $value) = @_;


        # get a timestamp
        my $time = time();


        metric ($mq, ( Timestamp => $time, Available => 1, Host => $self, MetricName => $metric, InstanceName => $instance, Value => $value ));

}

