#!/bin/bash

# IPv6 MTU default gateway adjustment
# SRE-5781 for details

##########################################
# Exit codes
# 0 - MTU already set to $mtuoverride or was successfully configured for $mtuoverride
# 1 - Unable to determine default IPv6 gw route MTU value
# 2 - Unable to determine default IPv6 gw route - /usr/bin/netstat -rnW -6 | egrep '^default'
# 3 - Current MTU variable assignment failed
# 4 - MTU cannot be lower than 1280 for IPv6 per RFC
# 5 - No IPv6 assignment found on interface
# 6 - Script must be run as root
# 7 - IPv6 gw did not pass regex check
# 8 - Script not enabled. MTU not found in /etc/sq/config_options.yaml or not force with '-f'
# 9 - Default router interface could not be determined
# 10 - IPv6 MTU cannot be higher than 64K per RFC. This script does not account for jumbo currently.
# 11 - Failed to update IPv6 gw route MTU
# 12 - Unable to read MTU from config_options.yaml
##########################################

##########################################
# Log & File locations
# /etc/sq/config_options.yaml - Used to read in desired MTU set by through salt
# /var/log/ipv6_status - Used to report if IPv6 is found on the default interface
# /var/log/ipv6_mtu_script_status - exit code of script
##########################################

#Set bash options
set -o pipefail

# Help menu
helpmenu() {
    echo "Description:"
    echo -e "\t Used to adjust the MTU for the default IPv6 route gateway. No flags are necessary, but allow for manual control and functions."
    echo
    echo "Flags:"
    echo -e "\t-6: Only check for IPv6 address on default gateway interface"
    echo -e "\t-d: Enable debug output"
    echo -e "\t-h: Help"
    echo -e "\t-f: Force script to run"
    echo -e "\t-m: <IPv6_MTU_value>: Value must be >= 1280"
    echo
    echo "Exit Codes:"
    echo -e "\t0 - MTU sucessfully updated or already at desired value. Prints only IPv6 address(if found) when -6 option specified"
    echo -e "\t1 - Unable to determine default IPv6 gw route MTU value"
    echo -e "\t2 - Unable to determine default IPv6 gw route - /usr/bin/netstat -rnW -6 | egrep '^default'"
    echo -e "\t3 - Current MTU variable assignment failed"
    echo -e "\t4 - MTU cannot be lower than 1280 for IPv6 per RFC"
    echo -e "\t5 - No IPv6 assignment found on interface"
    echo -e "\t6 - Script requires root"
    echo -e "\t7 - IPv6 did not pass regex check"
    echo -e "\t8 - Script not enabled. MTU not found in /etc/sq/config_options.yaml or not forced with '-f'"
    echo -e "\t9 - Unable to determine default route interface"
    echo -e "\t10 - MTU cannot be higher than 64K without jumbo per RFC"
    echo -e "\t11 - Failed to update IPv6 gw route MTU"
    echo -e "\t12 - Unable to read MTU from config_options.yaml"
    echo -e "\t100 - Help menu"
    echo
    echo "Log & File locations:"
    echo -e "\t /etc/sq/config_options.yaml"
    echo -e "\t\t ^-- Used to read in desired MTU set through salt (config_options_input.yaml)"
    echo -e "\t /var/log/ipv6_status"
    echo -e "\t\t ^-- Used to report if IPv6 is found on the default interface"
    echo -e "\t /var/log/ipv6_mtu_script_status"
    echo -e "\t\t ^-- exit code of script"
    exit 100
}

# Get cli input
while getopts "dhfm:6" opt; do
    case ${opt} in
        d )
            debug=1
            ;;
        f ) 
            enableScript=1
            ;;
        h)
            helpmenu
            ;;
        m)
            mtuoverride=${OPTARG}
            enforce=1
            ;;
        6)
            onlyv6chk=1
            ;;
        *)
            helpmenu
            ;;
    esac
done

##########################################
# Primary files and checks
##########################################

# Salt conf file - checks for ipv6_mtu value
conf_in="/etc/sq/config_options.yaml"
v6_file_out="/var/log/ipv6_status"
v6_mtu_status="/var/log/ipv6_mtu_script_status"

# Confirm netstat bin path  
if [[ -e /usr/bin/netstat ]];then
    v6_route_gateway=$(/usr/bin/netstat -rnW -6 | grep -E '^default')
    v4_int=$(/usr/bin/netstat -nrW -4 | grep -E '^default' | awk '{print $NF}')
fi

# If reading in from config_options, set to enable. Otherwise, '-f' is needed
if [[ -e $conf_in ]];then
    # Allow for manual override if needed
    if [[ $enforce -ne 1 ]];then
        mtuoverride=$(grep -E 'ipv6_mtu' $conf_in | awk '{print $2}')
        if [[ -z $mtuoverride ]];then
            conf_fail=1
        else
            # Trusting salt config_options.yaml key-val to enable MTU adjustment
            enableScript=1
        fi
    fi
fi

##########################################
# General variables
##########################################

enableScript="${enableScript:-0}"
gateway=$(awk '{print $2}' <<< "${v6_route_gateway}")
current_mtu=$(awk '{print $5}' <<< "${v6_route_gateway}")
modifymtu=0
debug="${debug:-0}"
noexit=0
mtuoverride="${mtuoverride:-1280}"
script_name=$(basename "$0")
uid=$(id -u)
onlyv6chk="${onlyv6chk:-0}"
conf_fail="${conf_fail:-0}"

##########################################
# Functions
##########################################

# Determine current IPv6 gw route MTU and set mod bit
check_ipv6_mtu() {
    if [[ $mtuoverride -lt 1280 ]];then
        exit_func "4" "MTU cannot be lower than 1280!"
    elif [[ $mtuoverride -gt 64000 ]];then
        exit_func "10" "MTU cannot be higher than 64K without jumbo!"
    else
        if [[ -z $current_mtu ]];then
            exit_func "1" "Missing current MTU assignment!"
        elif [[ $current_mtu -ne $mtuoverride ]];then
            modifymtu=1
        elif [[ $current_mtu -eq $mtuoverride ]];then
            exit_func "0" "MTU already set to ${mtuoverride}"
        else
            exit_func "3" "Unable to verify MTU"
        fi
    fi
}

# Modify IPv6 gw route MTU
override_ipv6_mtu() {
    IPv6_valid "${gateway}" || exit_func "$0"
    route change -inet6 -mtu "${mtuoverride}" default "${gateway}" &> /dev/null || exit_func "11" "Failed to update IPv6 gw route MTU"
}

# Exit codes and call to logging
exit_func() {
    code=$1
    msg=$2

    if [[ $debug -eq 1 ]];then
        echo "Code:$code - $msg"
    fi

    log_func "$code" "$msg"
    if [[ $noexit -eq 0 ]];then
        exit "${code}"
    fi
}

# Logs to /var/log/messages if FBSD
log_func() {
    logger -t "${script_name} -" "Code:${1} - ${2}"
    echo "$1" > ${v6_mtu_status}
}

# Validates IPv6 adresses
IPv6_valid() {
    v6_string=$1
    v6_regex="(([0-9a-fA-F]{1,4}:){7,7}[0-9a-fA-F]{1,4}|([0-9a-fA-F]{1,4}:){1,7}:|([0-9a-fA-F]{1,4}:){1,6}:[0-9a-fA-F]{1,4}|([0-9a-fA-F]{1,4}:){1,5}(:[0-9a-fA-F]{1,4}){1,2}|([0-9a-fA-F]{1,4}:){1,4}(:[0-9a-fA-F]{1,4}){1,3}|([0-9a-fA-F]{1,4}:){1,3}(:[0-9a-fA-F]{1,4}){1,4}|([0-9a-fA-F]{1,4}:){1,2}(:[0-9a-fA-F]{1,4}){1,5}|[0-9a-fA-F]{1,4}:((:[0-9a-fA-F]{1,4}){1,6})|:((:[0-9a-fA-F]{1,4}){1,7}|:)|fe80:(:[0-9a-fA-F]{0,4}){0,4}%[0-9a-zA-Z]{1,}|::(ffff(:0{1,4}){0,1}:){0,1}((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])|([0-9a-fA-F]{1,4}:){1,4}:((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9]))"
    
    if [[ ! $v6_string =~ $v6_regex ]];then
        exit_func "7" "IPv6 (${v6_string}) not valid"
    fi
}
 
# Route attribute modification requires root
check_root() {
    if [[ $uid -ne 0 ]];then
        exit_func "6" "Root required!"
    fi
}

# LLNW uses NDP for automatic IPv6 allocation. Checks for IPv6 against the primary interface
check_for_ipv6() {
    #Exclude link local address
    v6_exist=$(ifconfig "${v4_int}" | grep autoconf | grep -v 'inet6 fe80' | head -1 | awk '{print $2}')

    #Found that a route can still exist even when IPv6 has been dropped from the interface
    if [[ -z $v6_exist ]];then
        noexit=1
        exit_func "5" "No IPv6 assignment found on ${v4_int}"
        noexit=0
        echo "Missing" > $v6_file_out
    else
        echo "Found" > $v6_file_out
    fi
    if [[ $onlyv6chk -eq 1 ]];then
        exit_func "0" "IPv6 only check found address ${v6_exist} on ${v4_int}"
    fi
}


# Main func 
main() {
    if [[ $conf_fail -eq 1 ]];then
        exit_func "12" "Unable to read MTU from ${conf_in}"
    fi
    if [[ $enableScript -eq 1 ]];then
        check_root
        check_for_ipv6
        if [[ -z $v6_route_gateway ]];then
            exit_func "2" "Failed to determine IPv6 gw route"
        fi
        if [[ -z $v4_int ]];then
            exit_func "9" "Unable to determine default route interface"
        fi
        check_ipv6_mtu
        if [[ $modifymtu -eq 1 ]];then
            override_ipv6_mtu
        fi
        exit_func "0" "MTU set to ${mtuoverride}"
    else
        exit_func "8" "Script is disabled. See help menu"
    fi
}

# Run main - requires script to have enabled attribute to run
main