#!/usr/bin/perl

use Carp 'verbose';
use Socket;
use IO::Socket;
use IO::Seekable;
use Time::HiRes qw(time);
use Sys::Hostname;
use POSIX qw(strftime setsid);
use Digest::MD5 qw(md5_base64);
use Fcntl qw(:flock);
use JSON;

@logdirs = qw(/logs/squid);
$lockfile = '/var/run/logfeed.lock';
$logfile = '/logs/logfeed';
$monportnum = 2001;
$serverstring = 'lf/0.9.3';
$agentstring = 'logfeed/0.9.3';

$sel->{'rbits'} = '';
$sel->{'wbits'} = '';
$newconnections = 0;
$newgetaddrs = 0;
$replies = 0;
$last_scanlogs = 0;
$last_checkconf = 0;
$last_config_size = 0;
$last_config_mtime = 0;
$read_cds_ips = 0;
$activesock = 0;
$goodconf = 0;
$loglevel = 1; # readconf will set this later
%getaddrs_res = ();
%acctid_queues = ();

$configfile = '/etc/sq/logfeed.conf';
if (! -f $configfile) {
	$configfile = '/etc/logfeed.conf';
	if (! -f $configfile) {
		die "no config file!";
	}
}

@fields = qw(
	time duration addr res bytes method srcurl ident hier contenttype
	aceid acctid acecli aceuser
	referer agent cookie range puburl contentlen contentrange ddev lastref
	forwarded_for progress_times accept_language bytes_miss
	transfer sess_id bytes_read tmuf_info is_peer stream_type req_header
	req_glue realorigurl log_req_header2 log_req_header3 log_req_header4
	log_req_header5 log_keyval tls_ver ssl_enc ssl_cert_v ssl_cert_f
	ssl_expiry ssl_sess_reuse cacheability tcphost_port client_dest_addr
	term_code ipgeo orig_host req_id hier2 ref_count ssl_proto ssl_cipher
	ssl_reused http_proto client_port mpep_sibling
	);
foreach $num (0 .. $#fields) {
	$name = $fields[$num];
	${"fnum_$name"} = $num;
}

$SIG{'CHLD'} = 'IGNORE';
$SIG{'PIPE'} = 'IGNORE';
$now = time;
$last_checktimeouts = $now;

open(LOCK, ">>$lockfile") or die "can't write to $lockfile: $!";
$res = flock(LOCK, LOCK_EX|LOCK_NB);
if (! $res) {
	if (-t STDIN) {
		print "$0: looks like i'm already running\n";
	}
	exit 0;
}

$monport = IO::Socket::INET->new(
		Listen => 5,
		LocalPort => $monportnum,
		Proto => 'tcp',
		Reuse => 1);
if (! $monport) {
	die "can't listen on port $monportnum: $!";
}
&fh_watch_readable($monport, 'monport_accept');

if (-t STDIN) {
	print "$0: backgrounding\n";
}
setsid;
chdir('/');
close STDIN;
close STDOUT;
close STDERR;
fork && exit;

while (1) {
	$now = time;
	&checkconf;
	if (! $goodconf) {
		sleep 30;
		next;
	}
	if (abs($read_cds_ips - $now) > 600) {
		&read_cds_ips;
		$read_cds_ips = $now;
	}
	&checktimeouts;
	&scanlogfiles;
	while (($path, $fh) = each %logopen) {
		&readlogentries($fh, $path);
	}
	foreach (@queues) { &queue_run($_); }
	&dump_fdstate;
	&run_select(1);
	&dump_fdstate;
}

sub processlogentry {
	my($line, $path) = @_;
	if ($line !~ /^[0-9]/) {
		warn "$path: bad line start: $line";
		return;
	}
	@words = split(" ", $line);
	$numwords = $#words + 1;
	if ($numwords < 34) {
		warn "$path: not enough words:", $line;
		return;
	}
	#&dumpfields(@words);
	# edgeprism is currently bugged, sometimes it records a 206
	# for a non range request when partial caching is enabled
	if ($words[$fnum_res] =~ /206/ && $words[$fnum_range] eq "-") {
		$words[$fnum_res] =~ s/206/200/;
		$line =~ s/(^\S+\s+\S+\s+\S+\s+\S+)\/206/$1\/200/;
	}
	$acctid = $words[11];
	foreach $queue (&acctidqueues($acctid)) {
		&queue_addentry($queue, $line, $path);
	}
}

sub queue_addentry {
	my($queue, $line, $path) = @_;
	my($rstr, $rcode);
	my($entry);
	my(@words);
	@words = split(" ", $line);
	($rstr, $rcode) = split("/", $words[$fnum_res]);

	# filter out by client IP address
	if ($cf{$queue}{'filter_addr'}) {
		if (! $cf{$queue}{'filter_addr'}{$words[$fnum_addr]}) {
			return;
		}
	}
	# filter out by user agent
	if ($cf{$queue}{'filter_agent'}) {
		if ($words[$fnum_agent] !~ /$cf{$queue}{'filter_agent'}/) {
			return;
		}
	}
	# filter out by src url
	if ($cf{$queue}{'filter_srcurl'}) {
		if ($words[$fnum_srcurl] !~ /$cf{$queue}{'filter_srcurl'}/) {
			return;
		}
	}
	# exclude by src url
	if ($cf{$queue}{'exclude_srcurl'}) {
		if ($words[$fnum_srcurl] =~ /$cf{$queue}{'exclude_srcurl'}/) {
			return;
		}
	}

	# filter out by pub url
	if ($cf{$queue}{'filter_puburl'}) {
		if ($words[$fnum_puburl] !~ /$cf{$queue}{'filter_puburl'}/) {
			return;
		}
	}
	# exclude by pub url
	if ($cf{$queue}{'exclude_puburl'}) {
		if ($words[$fnum_puburl] =~ /$cf{$queue}{'exclude_puburl'}/) {
			return;
		}
	}

	# filter out by status codes
	if ($cf{$queue}{'filter_status'}) {
		if (! ($cf{$queue}{'filter_status'}{$rcode} or $cf{$queue}{'filter_status'}{$rstr})) {
			return;
		}
	}

	# filter out by method
	if ($cf{$queue}{'filter_method'}) {
		if (! $cf{$queue}{'filter_method'}{$words[$fnum_method]}) {
			return;
		}
	}

	# filter out by duration greater than threshold
	if ($cf{$queue}{'filter_duration_gt_thresh'}) {
		if ($words[$fnum_duration] <= $cf{$queue}{'filter_duration_gt_thresh'}) {
			return;
		}
	}

	# filter by computed header length
	if ($cf{$queue}{'filter_bytes_cl_delta'}) {
		if ($words[$fnum_contentlen] < 0 || $words[$fnum_bytes] - $words[$fnum_contentlen] < $cf{$queue}{'filter_bytes_cl_delta'}) {
			return;
		}
	}

	if (($rcode == 200) or ($rcode == 206)) {
		$portion = &portionsent($rcode,
			$words[$fnum_bytes],
			$words[$fnum_contentlen],
			$words[$fnum_contentrange]);
	}
	else {
		$portion = 'na';
	}

	if (! $cf{$queue}{'include_intermediate'}) {
		if ($cds_ips{$words[$fnum_addr]}) {
			next;
		}
	}

	if (! $cf{$queue}{'internal'}) {
		if (! $words[$fnum_bytes]) {
			next;
		}
	}

	if (! $cf{$queue}{'include_middle'}) {
		if ($portion eq 'middle') {
			&plog(2, "$path: skipping entry portion = '$portion' line = '$line'\n");
			return;
		}
		if ($portion eq 'none') {
			&plog(2, "$path: skipping - portion = '$portion' line = '$line'\n");
			return;
		}
		if ($portion eq 'unknown') {
			&plog(2, "$path: skipping - portion = '$portion' line = '$line'\n");
			return;
		}
	}
	&plog(2, "$path: portion = '$portion' adding to queue '$queue': '$line'\n");
	$cf{$queue}{'entrycount'}++;
	$entry->{'time'} = $words[$fnum_time];
	$entry->{'line'} = $line;
	$entry->{'portion'} = $portion;
	$entry->{'queue'} = $queue;
	push(@{$cf{$queue}{'entries'}}, $entry);
}

sub buildreq {
	my($queue, $entry) = @_;
	my($req, $path, $res, $rcode, $devent, $wincode, $lastbyte, $authn, $authsign, $dstr, $rurl);
	my(@words);
	my(@terms);
	my(%queryterm);

	$devent = '-';
	$lastbyte = '-';
	@words = split(" ", $entry->{'line'});

	$res = $words[$fnum_res];
	(undef, $rcode) = split('/', $res);
	if ($rcode == 200) {
		if ($words[$fnum_bytes] > $words[$fnum_contentlen]) {
			$devent = 'success';
			$wincode = 0;
			if ($cf{$queue}{'ms'}) {
				$lastbyte = 0;
			}
			else {
				$lastbyte = 1;
			}
		}
		else {
			$devent = 'failure';
			$wincode = 121;
			$lastbyte = 0;
		}
	}
	elsif ($rcode == 206) {
		$wincode = 0;
		if ($entry->{'portion'} eq 'start') {
			$devent = 'begin';
			$lastbyte = 0;
		}
		if ($entry->{'portion'} eq 'end') {
			$devent = 'success';
			$lastbyte = 1;
		}
		if ($entry->{'portion'} eq 'full') {
			$devent = 'success';
			$lastbyte = 1;
		}
	}
	$authn = join(', ', 1, $myaddr, $words[$fnum_addr], int($now),
		&getunique, $ms_keyselector);

	if ($cf{$queue}{'ms'}) {
		# mscom special treatment
		#$path = $words[$fnum_srcurl];
		#$path =~ s#^http://[^/]+/#/#;
		$getpath = $cf{$queue}{'submitpath'};
		$getpath .= '?Source=' . $ms_source;
		$getpath .= '&UTC=' . strftime("%y%m%d%H%M%SZ", gmtime($entry->{'time'}));
		#$getpath .= '&URI=' . $path;
		$getpath .= '&URI=' . $words[$fnum_puburl];
		$getpath .= '&dEvent=' . $devent;
		$getpath .= '&dsize=' . $words[$fnum_bytes];
		$getpath .= '&win32=' . $wincode;
		$getpath .= '&http=' . $rcode;
		$getpath .= '&lastbyte=' . $lastbyte;
		$getpath .= '&geoid=UNK';
		$getpath .= '&MSID=' . &getcookie($words[$fnum_cookie], 'MSID');
		$getpath .= '&cip=' . $words[$fnum_addr];
		$getpath .= '&Dtime=' . int($words[$fnum_duration] / 1000);
		$getpath .= '&ruri=' . $words[$fnum_referer];
		$getpath .= '&UA=' . $words[$fnum_agent];
		$getpath .= '&DType=1';
		$dstr = $ms_key . $authn . $getpath;
		&plog(2, "dstr = '$dstr'\n");
		$authsign = md5_base64($dstr) . '==';
		$url = sprintf "http://%s%s", $cf{$queue}{'submithost'}, $getpath;
		$req .= "GET $url HTTP/1.0\r\n";
		$req .= "Connection: Keep-Alive\r\n";
		$req .= "User-Agent: $agentstring\r\n";
		$req .= sprintf "X-%s-Authn: %s\r\n", $ms_id, $authn;
		$req .= sprintf "X-%s-AuthSign: %s\r\n", $ms_id, $authsign;
		$req .= "\r\n";
		return($req);
	} elsif ($cf{$queue}{'appleformat'}) {
		# apple special treatment
		my $req = '';
		my $event;
		my $ttfb = '-';
		my $rex = '-';
		my $rexb = '-';
		# Apple asked for billing region to be included...
		my $localpop = `/bin/hostname | awk -F. {'print $2'}`;
		my $billingregion = '-';
		if ($localpop =~ m/(bkk|dmk|hkg|kix|kul|mnl|nrt|itm|eaaa|eaac|eaad|eaae|eaaf|eaaj|eaak|eaap|eaar|eaas|eaax|eaay|eaaz|eaba|eabb|eabc|eabd|eabo|eabk|eabn|eabr|eabs|eabw|eabx|eacc|eacd|eace|eacg|eadb|eadl)/) {
			$billingregion = 'apac';
		} elsif ($localpop =~ m/(ams|arn|cdg|dus|fra|lcy|lin|lon|man|mrs|ory|pmo|qqs|qqu|qqw|stn|toj|vie|eaab|eaaw|eabg|eabh|eabl|eabm|eabq|eaby|eabz|eacf|eact|eacu|eacv|eacw|eacx|eacy|eadk|eadm|eadn|eado)/) {
			$billingregion = 'eu';
		} elsif ($localpop =~ m/(atl|aza|bur|crs|dal|dca|fll|iad|jfk|lax|lga|msp|ord|sea|sjc|yhz|yul|yxx|yyz|eaal|eaao|eaca|eacb|eadc|eade)/) {
			$billingregion = 'na';
		} elsif ($localpop =~ m/(bog|eze|gru|sce|eaag|eaah|eaat|eaau|eabp|each|eaci|eacj|eack|eacl|eacm|eacn|eacp|eacq|eacr)/) {
			$billingregion = 'sa';
		} else {
			$billingregion = 'row';
		}
		# AS0002178 - Add new fields to Apple receipts: ttfb, rex
		if ($words[$fnum_progress_times] !~ "-") {
			my @progress = split(/,/, $words[$fnum_progress_times]);
			foreach my $subfield (@progress) {
				if ($subfield =~ m/^s:(\d+)/) {
					$ttfb = $1;
				}
				if ($subfield =~ m/^rex:(\d+)/) {
					$rex = $1;
				}
				if ($subfield =~ m/^rexb:(\d+)/) {
					$rexb = $1;
				}
			}
		}
		# AS0006020 - Add new proto field to apple receipts
		my $proto = '-';
		if ($words[$fnum_puburl] =~ /https:/) {
			$proto = "https";
		} else {
			$proto = "http";
		}
		# ------------------------------------------------------------------------
		# Apple needs a field called billableBytes. Per conversations with Apple (Emily & Ruud),
		# billableBytes field should more closely reflect the actual monthly billed amount.
		# See Jira ASA-270 https://jira.atlas.llnw.com/browse/ASA-270 for more background.
		# Control billing reports uses EQ query 301 - https://eq-billing.llnw.com/portal/queries/overview#%7B%22queryId%22%3A301%7D
		# EQ query 301 references ascendant EQ query 29 - https://eq-billing.llnw.com/portal/queries/overview#%7B%22queryId%22%3A29%7D
		# To understand the qlang2 code in EQ query 29 check out this reference:  https://confluence.atlas.llnw.com/display/SWENG/QLang2+API
		# ------------------------------------------------------------------------

		# Initiatialize some variables before we get started
		my $hier_status = $words[$fnum_hier];
		my $billablebytes = '-';
		my $tcpoverhead1 = 0;
		my $tcpoverhead2 = 0;
		my $raw_bytes = 0;
		my $raw_bytes_from_origin = 0;
		my $bb_b_out = 0;
		my $bb_b_in = 0;

		# Extra logging for debugging during development & testing
		#&plogf(2, "Receipts: hier_status = %s\n", $hier_status);
		#&plogf(2, "Receipts: client_is_a_peer = %d\n", $words[$fnum_is_peer]);
		#&plogf(2, "Receipts: rex = %d\n", $rex);
		#&plogf(2, "Receipts: rexb = %d\n", $rexb);

		# If this connection is from an end user, then we sum up egress bytes + overhead
		if (not $words[$fnum_is_peer]) {
				$raw_bytes = $words[$fnum_bytes];
				# The next two lines are converted from EQ-Billing query #29 to add tcpoverhead to the raw bytes, 
				# plus retransmitted bytes (and their tcp overhead)
				$tcpoverhead1 = (($raw_bytes + 1453) / 1454) * 46;
				$bb_b_out = ($raw_bytes + $tcpoverhead1 + $rexb + (46 * $rex));
		}

		# If this is not a CARP transaction then add TCP overhead for the origin ingress bytes
		if (not $hier_status =~ /CARP\/cds*/) {
				$raw_bytes_from_origin = $words[$fnum_bytes_miss];
				# The next two lines are converted from EQ-Billing query #29 to add tcp overhead to the raw bytes
				$tcpoverhead2 = (($raw_bytes_from_origin + 1453) / 1454) * 46;
				$bb_b_in = ($raw_bytes_from_origin + $tcpoverhead2);
		}
		# Extra logging for debugging during development & testing
		#&plogf(2, "Receipts: raw_bytes = %d\n", $raw_bytes);
		#&plogf(2, "Receipts: tcpoverhead1 = %d\n", $tcpoverhead1);
		#&plogf(2, "Receipts: bb_b_out = %d\n", $bb_b_out);
		#&plogf(2, "Receipts: raw_bytes_from_origin = %d\n", $raw_bytes_from_origin);
		#&plogf(2, "Receipts: tcpoverhead2 = %d\n", $tcpoverhead2);
		#&plogf(2, "Receipts: bb_b_in = %d\n", $bb_b_in);
		#&plogf(2, "Receipts: request_id = $words[$fnum_req_id]");

		$billablebytes = $bb_b_out + $bb_b_in;
		#&plogf(2, "Receipts: billablebytes = %d\n", $billablebytes);
    # ------------------------------------------------------------------------

		my $path = $words[$fnum_puburl];
		my $hostname = "-";
		if ($path =~ m#https?://([^/]+)#) {
			$hostname = $1;
		}
		$path =~ s#^(http|https)://[^/]+/#/#;
#		$path =~ s/accessKey=[^&]+//;

		if ($words[$fnum_bytes] > $words[$fnum_contentlen]) {
			$event = "&event=success";
		} else {
			$event = "&event=abort";
		}

		my $firstByteServed = 0;
		my $lastByteServed = 0;
		if ($entry->{'portion'} eq 'start') {
			$firstByteServed = 1;
		} elsif ($entry->{'portion'} eq 'end') {
			$lastByteServed = 1;
		} elsif ($entry->{'portion'} eq 'full') {
			$firstByteServed = 1;
			$lastByteServed = 1;
		}

		my $objectContentLength = '&objectContentLength=-';
		if ($words[$fnum_contentrange] ne '-') {
			my(undef, $len) = split(/\//, $words[$fnum_contentrange]);
			if ($len) {
				$objectContentLength = "&objectContentLength=" . $len;
			}

		} elsif ($words[$fnum_contentlen] != -1) {
			$objectContentLength = "&objectContentLength=" . $words[$fnum_contentlen];
		}

		my $submithost = $cf{$queue}{'submithost'};
		if ($cf{$queue}{'submitport'}) {
			if ($cf{$queue}{'https'}) {
				if ($cf{$queue}{'submitport'} != 443) {
					$submithost .= ':' . $cf{$queue}{'submitport'};
				}
			} else {
				if ($cf{$queue}{'submitport'} != 80) {
					$submithost .= ':' . $cf{$queue}{'submitport'};
				}
			}
		}

		$event =~ s/&event=//;
		$objectContentLength =~ s/&objectContentLength=//;
		$send{time} = int($entry->{'time'});
		$send{startTime} = $send{time};
		$send{cdn} = 'llnw';
		$send{hostHeader} = $hostname;
		$send{requestedUrl} = $path;
		$send{clientIP} = $words[$fnum_addr];
		$send{cdnIP} = $myaddr;
		$send{userAgent} = $words[$fnum_agent];
		$send{downloadTime} = $words[$fnum_duration];
		$send{bytesServed} = $words[$fnum_bytes];
		$send{bytesServedFromCache} = ($words[$fnum_bytes] - $words[$fnum_bytes_miss]);
		$send{bytesDownloadedForRequest} = $words[$fnum_bytes_miss];
		if (!$words[$fnum_bytes_miss] == 1) {
			$send{servedFromCache} = 1;
		} else {
			$send{servedFromCache} = 0;
		}
		$send{event} = $event;
		$send{firstByteServed} = $firstByteServed;
		$send{lastByteServed} = $lastByteServed;
		$send{objectContentLength} = $objectContentLength;
		$send{statusCode} = $rcode;
		$send{referer} = $words[$fnum_referer];
		$send{timeToFirstByte} = $ttfb;
		$send{retransmits} = $rex;
		if ($hostname =~ /hls.itunes.apple.com/) {
			$send{playbackSessionID} = $words[$fnum_req_header];
		} else {
			$send{playbackSessionID} = '-';
		}
		$send{proto} = $proto;
		$send{billableBytes} = $billablebytes;
		$send{billingCode} = $billingregion;
		# iosapps specific configuration
		if ($hostname =~ /iosapps.itunes.apple.com/) {
			$send{appleDownloadType} = $words[$fnum_req_header];
		} else {
			$send{appleDownloadType} = '-';
		}
		# Format Apple-specific json POST data
		$tmp_json = encode_json \%send;
		$prepend = '{"events":[';
		$json = "$prepend" . "$tmp_json]}\r\n";
		# Calculate Content-Length header
		my $clength = length($json);
		my $postlocation = '-';
		my $postheader = '-';
		if ($hostname =~ /cdn-sre.cdn-apple.com/) {
			$postlocation = "POST https://logfeed-gw/cdn-xp-ingest.edge.apple/report/2/xp_cdn_receipt";
			$postheader = "Host: cdn-xp-ingest.edge.apple\r\n";
		} else {
			$postlocation = "POST https://logfeed-gw/xp-cdn.apple.com/report/2/xp_cdn_receipt";
			$postheader = "Host: xp-cdn.apple.com\r\n";
		}
		# Build raw socket request
		$req = $postlocation;
		$req .= " HTTP/1.0\r\n";
		$req .= $postheader;
		$req .= "Connection: Keep-Alive\r\n";
		$req .= "User-Agent: $agentstring\r\n";
		$req .= "Content-Length: $clength\r\n";
		$req .= "Content-Type: application/json\r\n\r\n";
		$req .= $json;
		return($req);
	}
	# normal handling
	if ($cf{$queue}{'send'}{'puburl'}) {
		$queryterm{'puburl'}		= $words[$fnum_puburl];
	}
	if ($cf{$queue}{'send'}{'srcurl'}) {
		$queryterm{'srcurl'}		= $words[$fnum_srcurl];
	}
	if ($cf{$queue}{'send'}{'statuscode'}) {
		$queryterm{'statuscode'}	= $rcode;
	}
	if ($cf{$queue}{'send'}{'bytes'}) {
		$queryterm{'bytes'}		= $words[$fnum_bytes];
	}
	if ($cf{$queue}{'send'}{'clientaddr'}) {
		$queryterm{'clientaddr'}	= $words[$fnum_addr];
	}
	if ($cf{$queue}{'send'}{'duration'}) {
		$queryterm{'duration'}		= int($words[$fnum_duration] / 1000);
	}
	if ($cf{$queue}{'send'}{'referer'}) {
		$queryterm{'referer'}		= $words[$fnum_referer];
	}
	if ($cf{$queue}{'send'}{'useragent'}) {
		$queryterm{'useragent'}		= $words[$fnum_agent];
	}
	if ($cf{$queue}{'send'}{'cookie'}) {
		$queryterm{'cookie'}		= $words[$fnum_cookie];
	}
	if ($cf{$queue}{'send'}{'time'}) {
		$queryterm{'time'}		= int($words[$fnum_time]);
	}
	if ($cf{$queue}{'send'}{'timestring'}) {
		$queryterm{'timestring'}	= strftime("%Y%m%d%H%M%SZ", gmtime($words[$fnum_time]));
	}
	if ($cf{$queue}{'send'}{'contentlength'}) {
		$queryterm{'contentlength'}	= $words[$fnum_contentlen];
	}
	if ($cf{$queue}{'send'}{'contentrange'}) {
		$queryterm{'contentrange'}	= $words[$fnum_contentrange];
	}
	if ($cf{$queue}{'send'}{'devent'}) {
		$queryterm{'devent'}		= $devent;
	}
	if ($cf{$queue}{'send'}{'lastbyte'}) {
		$queryterm{'lastbyte'}		= $lastbyte;
	}
	if ($cf{$queue}{'send'}{'complete'}) {
		if ($words[$fnum_contentlen] eq '-') {
			# exclude
		}
		elsif ($words[$fnum_bytes] > $words[$fnum_contentlen]) {
			$queryterm{'complete'}	= 'yes';
		}
		else {
			$queryterm{'complete'}	= 'no';
		}
	}
	if ($cf{$queue}{'internal'}) {
		if ($cf{$queue}{'send'}{'duration_msec'}) {
			$queryterm{'duration_msec'} = $words[$fnum_duration];
		}
		foreach $num (0 .. $#fields) {
			$name = $fields[$num];
			$queryterm{$name} = $words[$num];
		}
	}
	foreach $key (sort keys %queryterm) {
		$val = $queryterm{$key};
		next if ($val eq '-');
		# most odd characters are already encoded but for query
		# terms we also need to escape ampersand, equals, semicolon
		#$val =~ s/([&=;\?])/uc sprintf("%%%02x",ord($1))/eg;
		$val =~ s/([^a-zA-Z0-9\%_.-])/uc sprintf("%%%02x",ord($1))/eg;
		push(@terms, "$key=$val");
	}
	$req .= sprintf "GET %s?%s HTTP/1.0\r\n", $cf{$queue}{'submitpath'}, join('&', @terms);
	$req .= sprintf "Host: %s\r\n", $cf{$queue}{'submithost'};
	$req .= "Connection: Keep-Alive\r\n";
	$req .= "User-Agent: $agentstring\r\n";
	if ($cf{$queue}{'custom_header'}) {
		$req .= $cf{$queue}{'custom_header'} . "\r\n";
	}
	$req .= "\r\n";
	return($req);
}

sub getcookie {
	my($str, $name) = @_;
	my(@list);
	my($key, $val);
	@list = split(/;%20/, $str);
	foreach (@list) {
		($key, $val) = split(/=/, $_, 2);
		if ($key eq $name) {
			return($val);
		}
	}
	return('');
}

sub portionsent {
	my($rcode, $sent, $clen, $crange) = @_;
	my($x, $y, $z);
	my($start, $end);

	$start = 0;
	$end = 0;
	if ($sent == 0) {
		&plog(2, "portionsent: sent = '$sent' line = '$line'\n");
		return('none');
	}
	if ($clen == 0) {
		&plog(2, "portionsent: clen = '$clen' line = '$line'\n");
		return('unknown');
	}
	# it is more reliable to look at the crange variable than the
	# return code, mainly due to a current bug in edgeprism.
	# however, I suggest we keep it this way even after the ep bug
	# is fixed, no reason to rely on the return code. -jegg
	if ($rcode != 200 && $rcode != 206) {
		&plog(2, "portionsent: unknown for rcode '$rcode'\n");
		return("unknown");
	}
	if ($crange eq "-") {
		if ($sent) {
			$start = 1;
		}
		if ($sent > $clen) {
			$end = 1;
		}
	}
	else {
		if ($crange !~ m@^bytes%20(\d+)-(\d+)/(\d+)$@) {
			&plog(1, "portionsent: crange not matched: '$crange' line = '$line'\n");
			return('unknown');
		}
		$x = $1;
		$y = $2;
		$z = $3;
		if ($x == 0) {
			&plog(2, "portionsent: we sent the start with a 206\n");
			$start = 1;
		}
		if (((1+$y) == $z) and ($sent > $clen)) {
			&plog(2, "portionsent: we sent the end with a 206\n");
			$end = 1;
		}
	}
	# summarize..
	if ($start and $end) {
		return("full");
	}
	if ($start) {
		return("start");
	}
	if ($end) {
		return("end");
	}
	return("middle");
}


sub dumpfields {
	my(@words) = @_;
	my($fn);
	foreach $fn (0 .. $#fields) {
		&plogf(2, "%s %s\n", $fields[$fn], $words[$fn]);
	}
}
sub acctidqueues {
	my(@acctid) = @_;
	my(@queues);
	next if (! defined $acctid_queues{$acctid});
	@queues = @{$acctid_queues{$acctid}};
	#push(@queues, 'full'); # XXX for a global feed
	return(@queues);
}
sub watchacctid {
	my($acctid) = @_;
	if ($acctid_queues{$acctid}) {
		return(1);
	}
	return(0);
}

sub readlogentries {
	my($fh, $path) = @_;
	while (<$fh>) {
		$_ =~ s/[\r\n]+$//;
		&plogf(5, "%s:%s: %s\n", $path, $., $_);
		&processlogentry($_, $path);
	}
	if (! $logexists{$path}) {
		&plog(1, "$path does not exist anymore, closing\n");
		close($fh);
		delete($logopen{$path});
	}
}

sub scanlogfiles {
	my($tdiff, $dir, $file, $path, $res, $age);
	$tdiff = abs($now - $last_scanlogs);
	return if ($tdiff < 3);
	$last_scanlogs = $now;
	%logexists = ();
	foreach $dir (@logdirs) {
		$res = opendir(D, $dir);
		if (! $res) {
			warn "can not opendir $dir: $!";
			next;
		}
		while ($file = readdir(D)) {
			$path = "$dir/$file";
			next if ($file !~ /access(?:-\d+)?\.(\d+)\..*log$/);
			$created = $1;
			if ($file =~ /\.a([0-9]+)\./) {
				if (! &watchacctid($1)) {
					# not one of the ones we need to monitor, for now?
					next;
				}
			}
			@st = stat($path);
			$age = $now - $st[10];
			if ($age > 7200) {
				# file hasn't been modified anytime recently, ignore it
				next;
			}
			$logexists{$path} = 1;
			if ($logopen{$path}) {
				next;
			}
			$fh = new IO::Handle;
			$res = open($fh, $path);
			if (! $res) {
				warn "scanlogfiles could not open $path: $!";
				next;
			}
			$logopen{$path} = $fh;
			$age = $now - $created;
			if ($age > 15) {
				# if this file has been around a while then seek to the end
				$res = seek($fh, 0, SEEK_END) or warn "can't seek to end on $file: $!";
				if (! $res) {
					warn "seek to end of $path failed: $!";
				}
				&plog(1, "opened $path and seeked to end (age=$age)\n");
			}
			else {
				&plog(1, "opened $path without seeking to end (age=$age)\n");
			}
		}
		closedir D;
	}
}

sub checktimeouts {
	my($tdiff, $host, $age);
	$tdiff = abs($now - $last_checktimeouts);
	return if ($tdiff < 1);
	$last_checktimeouts = $now;

	# old dns lookup results
	foreach $host (keys %getaddrs_tstamp) {
		$age = abs($now - $getaddrs_tstamp{$host});
		if ($age > 30) {
			&plog(2, "deleting cached getaddrs data for $host\n");
			delete($getaddrs_tstamp{$host});
			delete($getaddrs_res{$host});
		}
	}

	foreach $fileno (0 .. $#fdstate) {
		$sock = $fdstate[$fileno]{'sock'};
		next if (! defined $sock);
		if ($fdstate[$fileno]{'state'} eq 'connecting') {
			$age = abs($now - $fdstate[$fileno]{'connecting'});
			if ($age > 30) {
				$addr = $fdstate[$fileno]{'addr'};
				$port = $fdstate[$fileno]{'port'};
				&plog(1, "connection to $addr:$port timed out\n");
				$connect_pending{"$addr:$port"} = 0;
				&closesock($sock);
			}
		}
		elsif ($fdstate[$fileno]{'state'} eq 'connected') {
			if ($fdstate[$fileno]{'idlesince'}) {
				$age = abs($now - $fdstate[$fileno]{'idlesince'});
				if ($age > 30) {
					&plog(2, "=== inactive connection fd=$fileno\n");
					&closesock($sock);
				}
			}
			elsif ($fdstate[$fileno]{'lastwrite'}) {
				if ($fdstate[$fileno]{'lastread'}) {
					$age = abs($fdstate[$fileno]{'lastwrite'} -
						$fdstate[$fileno]{'lastread'});
					if ($age > 30) {
						&plog(1, "=== too long with no response #1 fd=$fileno\n");
						&closesock($sock);
					}
				}
				else {
					$age = abs($now - $fdstate[$fileno]{'lastwrite'});
					if ($age > 30) {
						&plog(1, "=== too long with no response #2 fd=$fileno\n");
						&closesock($sock);
					}
				}
			}
		}
	}
}

sub queue_trim {
	my($queue) = @_;
	# prune old unsendable entries
	while ($cf{$queue}{'entrycount'} > 0) {
		$time = $cf{$queue}{'entries'}[0]->{'time'};
		$tdiff = $now - $time;
		&plog(2, "entry time is '$time', log entry age = $tdiff\n");
		if ($tdiff < 30) {
			return;
		}
		shift(@{$cf{$queue}{'entries'}}); # take it off, too old..
		$cf{$queue}{'entrycount'}--;
	}
}

sub queue_run {
	my($queue) = @_;
	my($addr, $sock, $fileno, $sent, $loop, $filled);
	my(@socks);
	my(@newsocks);
	my($host) = $cf{$queue}{'submithost'};
	my($port) = $cf{$queue}{'submitport'};

	&queue_trim($queue);
	if ($cf{$queue}{'entrycount'} == 0) {
		# empty, nothing to run..
		return;
	}
	&plogf(2, "queue '%s' entrycount = '%s' sent = '%s'\n", $queue,
		$cf{$queue}{'entrycount'}, $cf{$queue}{'submits'});
	@socks = &queue_getsocks($queue);
	if ($#socks == -1) {
		&plogf(2, "queue %s: no socks found\n", $queue);
		&queue_connect($queue);
		return;
	}
	$filled = 0;
	while ($cf{$queue}{'entrycount'} > 0) {
		if ($filled > $#socks) {
			&plogf(2, "%s: all %s sockets full\n", $queue, 1 + $#socks);
			&queue_connect($queue);
			return;
		}
		$sock = $socks[$cf{$queue}{'sockpos'} % (1 + $#socks)];
		$cf{$queue}{'sockpos'}++;
		$fileno = $sock->fileno;
		$addr = $fdstate[$fileno]{'addr'};
		$port = $fdstate[$fileno]{'port'};
		if (! defined $fdstate[$fileno]{'maxreqonwire'}) {
			# initializing to 1
			# raise it later if keep-alive is detected
			$fdstate[$fileno]{'maxreqonwire'} = 1;
		}
		if (! defined $fdstate[$fileno]{'reqonwire'}) {
			# initializing reqonwire to 0
			$fdstate[$fileno]{'reqonwire'} = 0;
		}
		if ($fdstate[$fileno]{'maxkeepalive'} and
				($fdstate[$fileno]{'reqsent'} >= $fdstate[$fileno]{'maxkeepalive'})) {
			&plogf(2, "FD %s (%s:%s): max keep alives reached\n", $fileno, $addr, $port);
			$filled++;
			next;
		}
		if (defined $fdstate[$fileno]{'writebuf'} && length($fdstate[$fileno]{'writebuf'}) > 0) {
			&plogf(2, "FD %s (%s:%s): existing write buffer\n", $fileno, $addr, $port);
			$filled++;
			next;
		}
		if ($fdstate[$fileno]{'reqonwire'} < $fdstate[$fileno]{'maxreqonwire'}) {
			$fdstate[$fileno]{'writebuf'} = &buildreq($queue, $cf{$queue}{'entries'}[0]);
			&plogf(2, "%s: sending entry over FD %s (%s:%s)\n", $queue, $fileno,
				$fdstate[$fileno]{'addr'}, $fdstate[$fileno]{'port'});
			&fh_watch_readable($sock, 'read_http_reply');
			$res = &writebuf($sock);
			if ($res < 0) {
				# sock was closed
				@socks = &queue_getsocks($queue);
				$filled = 0;
				next;
			}
			$fdstate[$fileno]{'reqonwire'}++;
			$fdstate[$fileno]{'reqsent'}++;
			$fdstate[$fileno]{'idlesince'} = 0;
			$cf{$queue}{'submits'}++;
			$cf{$queue}{'entrycount'}--;
			shift(@{$cf{$queue}{'entries'}}); # take entry off
			if (length($fdstate[$fileno]{'writebuf'})) {
				$filled++;
				next;
			}
			else {
				$filled = 0;
				next;
			}
		}
		else {
			&plog(2, "$queue: FD $fileno is full\n");
			$filled++;
			next;
		}
	}
	if ($cf{$queue}{'entrycount'} == 0) {
		&plogf(2, "queue_run: %s is empty\n", $queue);
		return;
	}
}

sub queue_connect {
	my($queue) = @_;
	my($host, $port, $addr);
	my($min, $waddr, $tdiff);
	my $maxconns;
	# may be called when we've already made the max connections
	$host = $cf{$queue}{'submithost'};
	$port = $cf{$queue}{'submitport'};
	$pipeline = $cf{$queue}{'pipeline'};

	if ($cf{$queue}{'https'}) {
		@addrs = ('127.0.0.1');
		$port = 80;
		$maxconns = 20;
	} else {
		$maxconns = $cf{$queue}{'maxconns'};

		if (! $getaddrs_res{$host}) {
			if (! $getaddrs_pending{$host}) {
				&plog(2, "host '$host' not yet known, need to look up\n");
				&getaddrs_start($host);
				$getaddrs_pending{$host} = time;
			}
			else {
				&plog(2, "host '$host' not yet known, lookup is pending\n");
			}
			return;
		}
		@addrs = @{$getaddrs_res{$host}};
	}
	# find the first addr with the fewest connections
	$min = undef;
	foreach $addr (@addrs) {
		&plog(2, "queue_connect: looking for fewest connections.. addr $addr\n");
		# XXX allow for multiple pending connect()s?
		if ($connect_pending{"$addr:$port"}) {
			&plog(2, "connection already pending to $host ($addr)\n");
			return;
		}
		@socks = values %{$connected{"$addr:$port"}};
		if ((1 + $#socks) >= $maxconns) {
			&plogf(2, "queue_connect: %s %s already max (%s socks)\n", $host, $addr, (1 + $#socks));
			next;
		}
		if ((!defined $min) or ($#socks < $min)) {
			$min = $#socks;
			$waddr = $addr;
		}
	}
	if (!defined $min) {
		&plog(2, "$queue $host:$port looks like all possible connections are already made\n");
		return;
	}
	$addr = $waddr;
	&plog(2, "$queue $host - chose to connect to $addr:$port\n");
	&connect_start($addr, $port, $pipeline);
}
sub queue_getsocks {
	my($queue) = @_;
	my($addr);
	my(@socks);
	my($host) = $cf{$queue}{'submithost'};
	my($port) = $cf{$queue}{'submitport'};

	if ($cf{$queue}{'https'}) {
		@socks = values %{$connected{"127.0.0.1:80"}};
	} else {
		if ($getaddrs_res{$host}) {
			my(@addrs) = @{$getaddrs_res{$host}};
			foreach $addr (@addrs) {
				push(@socks, values %{$connected{"$addr:$port"}});
			}
		}
	}
	@socks = sort { fileno($a) <=> fileno($b) } @socks;
	return(@socks);
}

sub run_select {
	my($timeout) = @_;
	my($nfound, $timeleft, $rbits, $wbits, $ebits);
	my($fileno, $subname, $fdcount);
	$rbits = $sel->{'rbits'};
	$wbits = $sel->{'wbits'};
	$ebits = $rbits | $wbits;
	#&plogf(5, "rbits in: %20s\n", unpack("b*", $rbits));
	#&plogf(5, "wbits in: %20s\n", unpack("b*", $wbits));
	#&plogf(5, "ebits in: %20s\n", unpack("b*", $ebits));
	($nfound, $timeleft) = select($rbits, $wbits, $ebits, $timeout);
	$now = time;
	if ($nfound < 0) {
		&plog(2, "nfound was '$nfound': $!/$@\n");
		return(0);
	}
	if ($nfound > 1) {
		&plog(2, "run_select: nfound = '$nfound', timeleft = '$timeleft'\n");
	}
	#&plogf(5, "rbits out: %20s\n", unpack("b*", $rbits));
	#&plogf(5, "wbits out: %20s\n", unpack("b*", $wbits));
	#&plogf(5, "ebits out: %20s\n", unpack("b*", $ebits));
	if ($nfound == 0) {
		return(0);
	}
	$fdcount = length($ebits) * 8;
	foreach $fileno (0 .. $fdcount) {
		if (vec($rbits, $fileno, 1)) {
			$subname = $fdstate[$fileno]{'canread'};
			$sock = $fdstate[$fileno]{'sock'};
			&plog(2, "fileno $fileno is readable, has subname '$subname'\n");
			&{$subname}($sock);
		}
		if (vec($wbits, $fileno, 1)) {
			$subname = $fdstate[$fileno]{'canwrite'};
			$sock = $fdstate[$fileno]{'sock'};
			&plog(2, "fileno $fileno is writable, has subname '$subname'\n");
			&{$subname}($sock);
		}
		if (vec($ebits, $fileno, 1)) {
			$subname = $fdstate[$fileno]{'haserror'};
			$sock = $fdstate[$fileno]{'sock'};
			&plog(2, "fileno $fileno has error, has subname '$subname'\n");
			&{$subname}($sock);
		}
	}
	return($nfound);
}

sub connect_start {
	my($addr, $port, $pipeline) = @_;
	my($sock, $iaddr, $paddr, $res);
	my(@addr);

	$sock = IO::Socket::INET->new(Proto => 'tcp');
	if (! $sock) {
		die "can't create socket: $!/$@";
	}
	$sock->blocking(0) or die "can't set non block for $sock: $!";
	$iaddr = inet_aton($addr);
	$paddr = sockaddr_in($port, $iaddr);
	$res = connect($sock, $paddr);
	if ($res) {
		&plog(1, "non blocking connect to $addr:$port unexpectedly immediately succeeded\n");
	}
	&fh_watch_writable($sock, 'connect_finish');
	my($fileno) = $sock->fileno;
	$fdstate[$fileno]{'addr'} = $addr;
	$fdstate[$fileno]{'port'} = $port;
	$fdstate[$fileno]{'paddr'} = $paddr;
	$fdstate[$fileno]{'sock'} = $sock;
	$fdstate[$fileno]{'pipeline'} = $pipeline;
	$fdstate[$fileno]{'state'} = 'connecting';
	$fdstate[$fileno]{'connecting'} = $now;
	$connect_pending{"$addr:$port"} = time;
	&plog(2, "connection to '$addr:$port' started\n");
}
sub connect_finish {
	my($sock) = @_;
	my($fileno) = $sock->fileno;
	my($addr) = $fdstate[$fileno]{'addr'};
	my($port) = $fdstate[$fileno]{'port'};
	my($paddr) = $fdstate[$fileno]{'paddr'};
	my(@socks);

	$activesock = 1;
	&fh_nowatch_writable($sock);
	my($res) = connect($sock, $paddr);
	$connect_pending{"$addr:$port"} = 0;
	if ((! $res) && $! ne 'Socket is already connected') {
		&plog(1, "connect to $addr:$port returned '$res': $!");
		&closesock($sock);
		return;
	}
	&plog(2, "connected to $addr:$port on fd $fileno\n");
	$connected{"$addr:$port"}{$fileno} = $sock;
	$fdstate[$fileno]{'state'} = 'connected';
	$fdstate[$fileno]{'connected'} = $now;
	$fdstate[$fileno]{'idlesince'} = $now;
	$newconnections++;
}

sub monport_accept {
	my($remote, $paddr);
	my($iaddr, $addr, $port, $reply);
	($remote, $paddr) = $monport->accept;
	if (! $remote) {
		&plog(1, "monport accept failed: $!\n");
		return;
	}
	($port, $iaddr) = unpack_sockaddr_in($paddr);
	if (length($iaddr) != 4) {
		&plog(1, "monport iaddr '$iaddr' wrong size\n");
		close $remote;
		return;
	}
	$addr = inet_ntoa($iaddr);
	&plog(2, "monport accepted connection from $addr:$port\n");
	# spit out a quick message
	if ($goodconf) {
		$reply = "HTTP/1.0 200 OK\r\n";
		$reply .= sprintf "X-C-Size: %s\r\n", $last_config_size;
		$reply .= sprintf "X-C-Time: %s\r\n", $last_config_mtime;
	}
	else {
		$reply = "HTTP/1.0 500 internal server error\r\n";
	}
	$reply .= "Server: $serverstring\r\n";
	$reply .= "Content-Length: 0\r\n";
	$reply .= "Connection: close\r\n";
	$reply .= "\r\n";
	print $remote $reply;
	close $remote;
}

sub writebuf {
	my($sock) = @_;
	my($fileno, $addr, $port, $buf, $res);

	$activesock = 1;
	$fileno = $sock->fileno;
	$buf = $fdstate[$fileno]{'writebuf'};
	$port = $fdstate[$fileno]{'port'};
	&plog(2, "==== writebuf:\n", $buf, "==== \n");
	$res = syswrite($sock, $buf);
	if (! defined $res) {
		&plogf(1, "FD %s: writebuf could not write %s bytes: '%s'\n", $fileno, length($buf), $!);
		if ($! eq 'Resource temporarily unavailable') {
			&plogf(1, "FD %s: reducing maxreqonwire from %s to %s\n",
				$fileno, $fdstate[$fileno]{'maxreqonwire'},
				$fdstate[$fileno]{'reqonwire'});
			$fdstate[$fileno]{'maxreqonwire'} = $fdstate[$fileno]{'reqonwire'};
			&fh_watch_writable($sock, 'writebuf');
			return(0);
		}
		else {
			&closesock($sock);
			return(-1);
		}
	}
	$fdstate[$fileno]{'lastwrite'} = $now;
	if ($res < length($buf)) {
		#open(OUT, ">>x.$$.$fileno");
		#print OUT substr($buf, 0, $res);
		#close OUT;
		&plogf(1, "FD %s: writebuf only wrote %s of %s bytes\n", $fileno, $res, length($buf));
		&plogf(1, "FD %s: reducing maxreqonwire from %s to %s\n",
			$fileno, $fdstate[$fileno]{'maxreqonwire'},
			$fdstate[$fileno]{'reqonwire'});
		$fdstate[$fileno]{'maxreqonwire'} = $fdstate[$fileno]{'reqonwire'};
		$fdstate[$fileno]{'writebuf'} = substr($buf, $res);
		&fh_watch_writable($sock, 'writebuf');
		return(0);
	}
	#open(OUT, ">>x.$$.$fileno");
	#print OUT $buf;
	#close OUT;
	&plogf(2, "FD %s: writebuf wrote %s of %s bytes\n", $fileno, $res, length($buf));
	$fdstate[$fileno]{'writebuf'} = '';
	&fh_nowatch_writable($sock);
	return(0);
}
sub read_http_reply {
	my($sock) = @_;
	my($newbuf, $buf, $bytes, $reply);
	my($fileno) = $sock->fileno;
	my($addr) = $fdstate[$fileno]{'addr'};
	my($port) = $fdstate[$fileno]{'port'};

	$activesock = 1;
	$bytes = sysread($sock, $newbuf, 131072);
	if (! defined $bytes) {
		&plogf(1, "read_http_reply FD %s (%s:%s) read failed: '%s'\n", $fileno, $addr, $port, $!);
		&closesock($sock);
		return;
	}
	if ($bytes == 0) {
		&plogf(1, "read_http_reply FD %s (%s:%s) returned zero: '%s'\n", $fileno, $addr, $port, $!);
		&closesock($sock);
		return;
	}
	$fdstate[$fileno]{'lastread'} = $now;
	if ($fdstate[$fileno]{'buf'}) {
		$buf = $fdstate[$fileno]{'buf'} . $newbuf;
	}
	else {
		$buf = $newbuf;
	}
	# next, remove any remaining body if we have been waiting for it.
	if (defined $fdstate[$fileno]{'readbytes'} and ($fdstate[$fileno]{'readbytes'} > 0)) {
		&plogf(2, "read_http_reply FD %s (%s:%s) continuing body read (need %s bytes, got %s bytes)\n",
			$fileno, $addr, $port, $fdstate[$fileno]{'readbytes'}, length($buf));
		if (length($buf) < $fdstate[$fileno]{'readbytes'}) {
			$fdstate[$fileno]{'readbytes'} -= length($buf);
			&plogf(1, "read_http_reply FD %s (%s:%s) body read incomplete (%s bytes still remain)\n",
				$fileno, $addr, $port, $fdstate[$fileno]{'readbytes'});
			&plog(1, "incomplete continuation:\n", $buf);
			return;
		}
		$oldbodyremnant = substr($buf, 0, $fdstate[$fileno]{'readbytes'}, '');
		&plog(1, "old body remnant:\n", $oldbodyremnant);
		$fdstate[$fileno]{'readbytes'} = 0;
		if ($fdstate[$fileno]{'reqonwire'} == 0) {
			$fdstate[$fileno]{'lastwrite'} = 0;
			$fdstate[$fileno]{'lastread'} = 0;
			$fdstate[$fileno]{'idlesince'} = $now;
		}
	}
	while ($buf =~ s/^(HTTP.*?)\r\n\r\n//s) {
		($status, @headers) = split("\r\n", $1);
		&plog(1, "== response status line: $status\n");
		$replies++;
		$fdstate[$fileno]{'reqonwire'}--;
		%hkv = ();
		foreach $header (@headers) {
			($key, $val) = split(/: ?/, $header, 2);
			$key =~ tr/A-Z/a-z/;
			$hkv{$key} = $val;
			&plog(1, "== '$key' => '$val'\n");
		}
		if ($hkv{'connection'}) {
			if ($hkv{'connection'} !~ /keep-alive/i) {
				&plogf(2, "read_http_reply FD %s (%s:%s) closing (connection: '%s')\n",
					$fileno, $addr, $port, $hkv{'connection'});
				&closesock($sock);
				return;
			}
			if ($hkv{'keep-alive'} and ($hkv{'keep-alive'} =~ /max=(\d+)/i)) {
				&plogf(2, "read_http_reply FD %s (%s:%s) max keep alives = %s\n",
					$fileno, $addr, $port, $1);
				if (! $fdstate[$fileno]{'maxkeepalive'}) {
					$fdstate[$fileno]{'maxkeepalive'} = $1;
				}
			}
			if ($fdstate[$fileno]{'maxreqonwire'} == 1) {
				if ($fdstate[$fileno]{'pipeline'} > 0) {
					$fdstate[$fileno]{'maxreqonwire'} = $fdstate[$fileno]{'pipeline'};
					&plogf(2, "read_http_reply FD %s (%s:%s) piplining %s\n",
						$fileno, $addr, $port, $fdstate[$fileno]{'maxreqonwire'});
				}
			}
		}
		else {
			&plogf(2, "read_http_reply FD %s (%s:%s) closing (no connection: header)\n",
				$fileno, $addr, $port, $hkv{'connection'});
			&closesock($sock);
			return;
		}
		if (! defined $hkv{'content-length'}) {
			&plogf(2, "read_http_reply FD %s (%s:%s) closing (no content-length: header)\n",
				$fileno, $addr, $port, $hkv{'connection'});
			&closesock($sock);
			return;
		}
		if (length($buf) >= $hkv{'content-length'}) {
			$body = substr($buf, 0, $hkv{'content-length'}, '');
			&plogf(2, "read_http_reply FD %s (%s:%s) full body was read, can read next req now\n",
				$fileno, $addr, $port);
			&plog(2, $body);
			next;
		}
		$remains = $hkv{'content-length'} - length($buf);
		$buf = '';
		$fdstate[$fileno]{'readbytes'} = $remains;
		&plogf(2, "read_http_reply FD %s (%s:%s) waitiing for %s more bytes of body\n",
			$fileno, $addr, $port, $remains);
	}
	$fdstate[$fileno]{'buf'} = $buf;
	if (length($buf) > 0) {
		&plogf(2, "read_http_reply FD %s (%s:%s) has pending buffer:\n", $fileno, $addr, $port);
		&plog(2, "====\n", $buf, "====\n");
		$fdstate[$fileno]{'buf'} = $buf;
		return;
	}
	if ($fdstate[$fileno]{'reqonwire'} == 0) {
		$fdstate[$fileno]{'lastwrite'} = 0;
		$fdstate[$fileno]{'lastread'} = 0;
		$fdstate[$fileno]{'idlesince'} = $now;
	}
}
sub closesock {
	my($sock) = @_;
	my($fileno) = $sock->fileno;
	my($addr) = $fdstate[$fileno]{'addr'};
	my($port) = $fdstate[$fileno]{'port'};

	$activesock = 1;
	&plogf(2, "closing FD %s (%s:%s)\n", $fileno, $addr, $port);
	vec($sel->{'rbits'}, $fileno, 1) = 0;
	vec($sel->{'wbits'}, $fileno, 1) = 0;
	delete($connected{"$addr:$port"}{$fileno});
	$fdstate[$fileno] = undef;
	close $sock;
	#die "connection closed"; # just handy to exit here
}

sub getaddrs_start {
	my($host) = @_;
	my($res);
	my($child, $parent);
	my(@iaddrs);
	my(@addrs);

	if ($host =~ /^\d+\.\d+\.\d+$/) {
		die "do not use getaddrs_start for an ip address";
	}

	$activesock = 1;
	$child = new IO::Handle;
	$parent = new IO::Handle;
	$res = socketpair($child, $parent, AF_UNIX, SOCK_DGRAM, PF_UNSPEC) or die "socketpair: $!";
	if (! $res) {
		die("getaddrs_start socketpair() returned '$res': $!");
	}
	$res = fork();
	if (! defined $res) {
		die("getaddrs_start fork() failed: $!");
	}
	if ($res == 0) {
		# child process
		$0 .= " - looking up $host";
		(undef, undef, undef, undef, @iaddrs) = gethostbyname($host);
		@addrs = ();
		foreach (@iaddrs) {
			push(@addrs, inet_ntoa($_));
		}
		# force cds465.lax:
		#print $parent "208.111.144.77";
		print $parent join(" ", @addrs);
		exit 0;
	}
	# parent
	&fh_watch_readable($child, 'getaddrs_finish');
	$fdstate[$child->fileno]{'getaddr'} = $host;
}

sub getaddrs_finish {
	my($sock) = @_;
	my($buf);
	my($res) = recv($sock, $buf, 32768, 0);
	my($fileno) = $sock->fileno;
	my($host) = $fdstate[$fileno]{'getaddr'};
	delete($getaddrs_pending{$host});
	if (! defined $res) {
		die "recv failed while looking up '$host': $!";
	}
	if ($buf ne '') {
		&plogf(2, "getaddrs_finish: looking up %s returned '%s'\n", $host, $buf);
		$getaddrs_res{$host} = [split(' ', $buf)];
		$getaddrs_tstamp{$host} = time;
		$newgetaddrs++;
	}
	&fh_nowatch_readable($sock);
	$fdstate[$fileno] = undef;
	close $sock;
}

sub fh_watch_readable {
	my($sock, $subname) = @_;
	my($fileno) = $sock->fileno;
	die "fileno() for $sock failed: $!" unless defined $fileno;
	if (defined $fdstate[$fileno]{'canread'} and
		$fdstate[$fileno]{'canread'} ne $subname) {
		&plog(2, "watching readable for $fileno: '$subname'\n");
	}
	vec($sel->{'rbits'}, $fileno, 1) = 1;
	$fdstate[$fileno]{'sock'} = $sock;
	$fdstate[$fileno]{'canread'} = $subname;
}
sub fh_watch_writable {
	my($sock, $subname) = @_;
	my($fileno) = $sock->fileno;
	die "fileno() for $sock failed: $!" unless defined $fileno;
	&plog(2, "watching writable for $fileno: '$subname'\n");
	vec($sel->{'wbits'}, $fileno, 1) = 1;
	$fdstate[$fileno]{'sock'} = $sock;
	$fdstate[$fileno]{'canwrite'} = $subname;
}
sub fh_nowatch_readable {
	my($sock) = @_;
	my($fileno) = $sock->fileno;
	die "fileno() for $sock failed: $!" unless defined $fileno;
	&plog(2, "not watching readable for $fileno\n");
	vec($sel->{'rbits'}, $fileno, 1) = 0;
	$fdstate[$fileno]{'canread'} = '';
}
sub fh_nowatch_writable {
	my($sock) = @_;
	my($fileno) = $sock->fileno;
	if (! defined $fileno) {
		warn "fh_nowatch_writable fileno() for $sock failed: $!";
		return;
	}
	&plog(2, "not watching writable for $fileno\n");
	vec($sel->{'wbits'}, $fileno, 1) = 0;
	$fdstate[$fileno]{'canwrite'} = '';
}
sub dump_fdstate {
	my($fileno, $key);
	return if ($activesock == 0);
	$activesock = 0;
	$str = '';
	foreach $fileno (0 .. $#fdstate) {
		foreach $key (sort keys %{$fdstate[$fileno]}) {
			next if ($key eq 'paddr');
			$str .= sprintf "fdstate[%s]{%s} = '%s'\n", $fileno, $key, $fdstate[$fileno]{$key};
		}
	}
	return if ($str eq '');
	&plog(2, $str);
}

sub access_log_unescape {
	my($str) = @_;
	$str =~ s/%(20|5b|5d|7f|[0189abcdef][0-9a-f])/chr(hex($1))/eg;
	return($str);
}

sub qtencode {
	my($str) = @_;
	$str =~ s/([^a-zA-Z0-9\/\%_.,:-])/uc sprintf("%%%02x",ord($1))/eg;
	return($str);
}

sub getunique {
	if (! defined $unique) {
		$unique = rand(1_000_000);
	}
	$unique += (1+rand(10));
	return(int($unique % 1_000_000));
}

sub myaddr {
	my($sock, $addr);
	$sock = IO::Socket::INET->new(
		PeerAddr => '9.9.9.9',
		PeerPort => 9999,
		Proto => 'udp');
	if (! $sock) {
		return("0.0.0.0");
	}
	$addr = $sock->sockhost;
	close $sock;
	return($addr);
}

sub plog {
	my($level, @rest) = @_;
	my($res);
	my(@st);
	if ($level > $loglevel) {
		return;
	}
	if (! $logfh) {
		$logfh = new IO::Handle;
		$res = open($logfh, ">>$logfile");
		if (! $res) {
			undef $logfh;
			warn "can not append to '$logfile': $!";
			return;
		}
		$logfh->autoflush(1);
		open STDOUT, '>& $logfh';
		open STDERR, '>& $logfh';
		my $old = select(STDOUT);
		$| = 1;
		select(STDERR);
		$| = 1;
		select $old;
	}
	if (substr($rest[$#rest], -1) ne "\n") {
		$rest[$#rest] .= "\n";
	}
	print $logfh @rest;
	@st = stat($logfh);
	if ($st[7] > 20_000_000) {
		close $logfh;
		undef $logfh;
		rename($logfile, "$logfile.old");
	}
}
sub plogf {
	my($level, $fmt, @rest) = @_;
	if ($level > $loglevel) {
		return;
	}
	&plog($level, sprintf($fmt, @rest));
}

sub readconf {
	my $new_loglevel;
	my %new_acctid_queues;
	my %newcf;

	$new_loglevel = 4;
	$res = open(CONFIG, $configfile);
	if (! $res) {
		&plog(1, "readconf failed: '$queue' missing required '$key'\n");
		return;
	}
	# save info about the config so we can tell when it changes
	@st = stat(CONFIG);
	$last_config_size = $st[7];
	$last_config_mtime = $st[10];
	while (<CONFIG>) {
		s/\s+$//;
		if (/^queue\s/) {
			(undef, $queue, $key, $val) = split(" ", $_, 4);
			if (($key eq 'filter_status') or ($key eq 'filter_method') or ($key eq 'send') or ($key eq 'filter_addr')) {
				foreach $item (split(" ", $val)) {
					$newcf{$queue}{$key}{$item} = 1;
				}
			}
			else {
				$newcf{$queue}{$key} = $val;
			}
		}
		elsif (/^loglevel\s/) {
			(undef, $new_loglevel) = split(" ", $_, 2);
		}
	}
	close CONFIG;

	@queues = sort keys %newcf;
	foreach $queue (@queues) {
		if ($newcf{$queue}{'submiturl'}) {
			if ($newcf{$queue}{'submiturl'} =~ m#http://([^/]+)(/.*)$#) {
				$newcf{$queue}{'submithost'} = $1;
				$newcf{$queue}{'submitpath'} = $2;
			}
			elsif ($newcf{$queue}{'submiturl'} =~ m#https://([^/]+)(/.*)$#) {
				$newcf{$queue}{'submithost'} = $1;
				$newcf{$queue}{'submitpath'} = $2;
				$newcf{$queue}{'https'} = 1;
			}
		}
		foreach $key ('submithost', 'acctid') {
			if (! $newcf{$queue}{$key}) {
				&plog(1, "readconf failed: '$queue' missing required '$key'\n");
				return;
			}
		}
		if ($newcf{$queue}{'submithost'} =~ s/:(\d+)$//) {
			$newcf{$queue}{'submitport'} = $1;
		}
		if (! $newcf{$queue}{'submitport'}) {
			if ($newcf{$queue}{'https'}) {
				$newcf{$queue}{'submitport'} = 443;
			} else {
				$newcf{$queue}{'submitport'} = 80;
			}
		}
		if (! $newcf{$queue}{'submits'}) {
			$newcf{$queue}{'submits'} = 0;
		}
		if (! $newcf{$queue}{'sockpos'}) {
			$newcf{$queue}{'sockpos'} = 0;
		}
		if (! $newcf{$queue}{'entrycount'}) {
			$newcf{$queue}{'entrycount'} = 0;
		}
		if (! $newcf{$queue}{'maxconns'}) {
			$newcf{$queue}{'maxconns'} = 1;
		}
		if (! $newcf{$queue}{'pipeline'}) {
			$newcf{$queue}{'pipeline'} = 0;
		}
		if (! $newcf{$queue}{'internal'}) {
			# turn on default fields
			foreach $name ('statuscode', 'bytes', 'clientaddr') {
				if (! defined $newcf{$queue}{'send'}{$name}) {
					$newcf{$queue}{'send'}{$name} = 1;
				}
			}
			if (! $newcf{$queue}{'send'}{'puburl'}) {
				# if we are not sending the puburl, it is
				# implicit that we will send the srcurl.
				# both can be sent if needed.
				$newcf{$queue}{'send'}{'srcurl'} = 1;
			}
		}
		foreach $acctid (split(" ", $newcf{$queue}{'acctid'})) {
			&plog(1, "readconf $acctid => $queue\n");
			push(@{$new_acctid_queues{$acctid}}, $queue);
		}

		# debug
		foreach $key (sort keys %{$newcf{$queue}}) {
			if (($key eq 'send') or ($key eq 'filter_status')) {
				foreach $subkey (sort keys %{$newcf{$queue}{$key}}) {
					$val = $newcf{$queue}{$key}{$subkey};
					&plog(1, "readconf: '$queue': '$key' '$subkey' => '$val'\n");
				}
			}
			else {
				$val = $newcf{$queue}{$key};
				&plog(1, "readconf: '$queue': '$key' => '$val'\n");
			}
		}
	}
	# [re-]initialize
	$hostname = hostname;
	$myaddr = &myaddr;
	if ($hostname =~ /(9\.llnw\.net|msecn.net)$/) {
		$ms_source = 'EC';
		$ms_key = 'f338038e87ac440da57b6c77dfd8b4bfa81205e3624d489b';
		$ms_keyselector = '1';
		$ms_id = 'ECN';
	}
	else {
		$ms_source = 'LL';
		$ms_key = 'ff6ba4b0ca0e429088a2e2e53ff9eb06acf7253a5e1f4d5a';
		$ms_keyselector = '1';
		$ms_id = 'LimeLight';
	}
	# no errors, commit the new config
	$goodconf = 1;
	%acctid_queues = %new_acctid_queues;
	%cf = %newcf;
	$loglevel = $new_loglevel;
}
sub checkconf {
	# check for a new config, and re-read if new
	my $tdiff = abs($now - $last_checkconf);
	return if ($tdiff < 3);
	my @st = stat($configfile);
	if (($st[7] != $last_config_size) or ($st[10] != $last_config_mtime)) {
		&plogf(1, "%s: was %s:%s, now %s:%s\n", $configfile,
			 $last_config_size, $last_config_mtime, $st[7], $st[10]);
		&readconf;
	}
	return;
}

sub read_cds_ips {
	my $file = '/etc/sq/cds_ips.dat';
	open(IN, $file);
	while (<IN>) {
		chop;
		$cds_ips{$_} = $_;
	}
	close IN;
	&plogf(1, "finished reading $file\n", $file);
}

# always use the sock / handle (not the fileno) in sub arguments
# $getaddrs_pending{$host} = time;			# for pending dns lookups
# $getaddrs_res{$host} = ['1.1.1.1', '2.2.2.2'];	# has the ip addresses in space delimited text form
# $getaddrs_tstamp{$host} = time;			# has the timestamp of the lookup
# $connect_pending{"$addr:$port"} = time;		# for pending connections
# $connected{"$addr:$port"}{$fileno} = $sock;		# has the established connections
# $connected_tstamp{"$addr:$port"}{$fileno} = time;	# time connection was established

# read conf
# check current logs
# open new applicable logs
# if conf was just read, and new applicable logs were found, seek to end of file
# close logs that are gone after they've been read to the end
# read log entries
# determine which queue the log entry applies to (by acctid plus filters)
# queue up log entries that have new-enough timestamps
# per queue with pending entries:
# if we haven't looked up the host look it up
# if we haven't connected, connect
# if we are connected and blocked, connect again if not maxed
# if entries are new enough, write them
# disconnect after N connected seconds even if not busy


# https://jira.atlas.llnw.com/browse/DLRCPT-1
#queue apple_swcdn acctid 2431
#queue apple_swcdn appleformat 1
#queue apple_swcdn include_middle 1
#queue apple_swcdn submiturl https://mzuserxp.itunes.apple.com/WebObjects/MZUserXP.woa/wa/receiptStats?rcpt=swcdn
#queue apple_swcdn filter_puburl http://swcdn.apple.com/
#queue apple_swcdn exclude_puburl \.(pkm|dist|smd)($|\?)

#cdn_rewrite https://logfeed-gw/ https:// acct_id 0 keepalive_timeout 60

# http://swcdn.apple.com/content/downloads/26/31/041-3545/8dRscG2Y7LhqxMwdBccdTvN8mzVc4KSCxv/iTunes.msi
