#!/usr/bin/env python
from collections import namedtuple
import sys
import argparse


# Global Vars
##########################
CustReq = namedtuple('CustReq', 'acct_id FBms')
TimeCust = {}
TimeGlobal = {'req': 0.0, 'total': 0.0, 'avg': None}


def process_args():
    parser = argparse.ArgumentParser(description='Customer breakdown TTFB SLI report')
    parser.add_argument('-t', '--top', help='Set the top N customers to return in the report. Default 20', default='20')
    parser.add_argument('-w', '--weight', help='Set the weighted threshold to show abnormal accounts. Default: .15', default='.15')
    return parser.parse_args()

# Parse log line to include it in the TTFB calculation or skip it
# if valid we return a namedtuple else we return None
def filterLog(logLine):
    lineArray = logLine.split()

    # Extract needed fields
    validLine={
        'STATUS': lineArray[3],
        'METHOD': lineArray[5],
        'HIER_STATUS': lineArray[54],
        'CACHE_HIT_STATUS': lineArray[10],
        'ACCT_ID': int(lineArray[11]),
        'PROGRESS_TIMES': lineArray[24],
        'FBS': None
    }

    # Return failed for unwanted status
    if 'NONE' not in validLine['HIER_STATUS']:
        return None
    if 'TCP_HIT/2' not in validLine['STATUS']:
        return None
    if 'GET' not in validLine['METHOD'] and 'HEAD' not in validLine['METHOD']:
        return None
    if 's:' not in validLine['PROGRESS_TIMES']:
        return None
    if validLine['ACCT_ID'] == 0:
        return None
    if validLine['CACHE_HIT_STATUS'] != '1' and validLine['CACHE_HIT_STATUS'] != '2':
        return None

    # Parse s: param
    for param in validLine['PROGRESS_TIMES'].split(','):
        if 's:' in param:
            validLine['FBS'] = float(param.split(':')[1])

    return CustReq(validLine['ACCT_ID'], validLine['FBS'])

# Add req and times to global and customer totals
def addTimes(request):
    # Process Global Times
    TimeGlobal['req'] +=  1
    TimeGlobal['total'] += request.FBms

    # Process Customer Times
    if request.acct_id not in TimeCust:
        TimeCust[request.acct_id] = {'req': 0.0, 'total': 0.0, 'avg': None, 'weight': None}

    TimeCust[request.acct_id]['req'] += 1
    TimeCust[request.acct_id]['total'] += request.FBms

# Read in logs from stdin and store times of valid logs
def processLogs():
    for line in sys.stdin:

        # Parse log line
        logLine=filterLog(line)

        # If we get a valid log add it to the total times
        if logLine:
            addTimes(logLine)

# Tell who is pulling the averge the hardst up or down.
def weightedAvg(customer):
    # (Direction that the customer is pulling the avg) * (how hard they are pulling)
    # (customer avg fb time - total avg fb time) * (customer total req / total request)
    if TimeGlobal['req'] == 0:
        return -1
    else:
        return (TimeCust[customer]['avg'] - TimeGlobal['avg']) * (TimeCust[customer]['req'] / TimeGlobal['req'])

# Do the basic math to find the avg and who is effecting the avg
def calcTotals():
    # Global Times
    if TimeGlobal['req'] == 0:
        print "No requests found"
        quit()
    TimeGlobal['avg'] = float(TimeGlobal['total'])/float(TimeGlobal['req'])

    # Customer Times
    for customer in TimeCust:
        TimeCust[customer]['avg'] = float(TimeCust[customer]['total'])/float(TimeCust[customer]['req'])
        TimeCust[customer]['weight'] = weightedAvg(customer)

# Print out the report
def generateReport(top, weight, reportFormat='pretty'):
    if reportFormat == 'pretty':
        print '\n  Total Requests: %-12.0f' % TimeGlobal['req']
        print '  AvgFB: %-12.3f' % TimeGlobal['avg']
        print '\n'
        print '  Top %s of requesting accounts' % (top)
        print '    %-12s%-12s%-12s%-12s' % ('acct_id', 'Requests', 'AvgTTFB', 'Weight')
        for customer in sorted(TimeCust ,key=lambda x:TimeCust[x]['weight'], reverse=True)[:top]:
            print '    %-12s%-12.0f%-12.3f%-12.3f' % (customer, TimeCust[customer]['req'], TimeCust[customer]['avg'], TimeCust[customer]['weight'])
        print '\n'
        print '  Abnormal Accounts > %s' % weight
        print '    %-12s%-12s%-12s%-12s%-12s' % ('acct_id','Requests','AvgTTFB','Weight','Variance ms')
        for customer in TimeCust:
            if TimeCust[customer]['weight'] > weight:
                print '    %-12s%-12.0f%-12.2f%-12.3f%-12.2f' % (customer, TimeCust[customer]['req'], TimeCust[customer]['avg'], TimeCust[customer]['weight'], TimeCust[customer]['avg'] - TimeGlobal['avg'])


if __name__ == '__main__':
    # Read in args
    args = process_args()

    # Read in logs and filter
    processLogs()

    # Calculate values
    calcTotals()

    # Print out report
    generateReport(int(args.top), float(args.weight), reportFormat='pretty')
