#!/bin/sh

usage () {
	cat <<END
usage: $0 [options]

options:
 -d             if set, do a dry run, not changing anything
 -m <maxage>	specify the allowed delay after a graceful update for existing connections. default 14400
 -s		if set, start edgeprisms that should be started unless they are all dead

END
	exit 1
}

dryrun=no
maxage=14400
start=no

while getopts "dsm:" OPTION; do
	case $OPTION in
		d)
			dryrun=yes
			;;
		m)
			maxage="$OPTARG"
			;;
		s)
			start=yes
			;;
		*)
			usage
			;;
	esac
done

curtime=$(date +%s)

if [ "$(uname)" != FreeBSD ]; then
	echo Requires FreeBSD
	exit 1
fi

killsquid () {
	# only kill the parent, ppid, if the parent is squid
	# this is an important check, becuase if the parent squid has already
	# been killed, the parent pid of the main squid pid will be 1
	# and killing pid 1 will send the system into single user mode

	pid="$1"
	ppid="$2"

	vppid="$(ps -U root -o pid,comm "$ppid" | awk '$2 == "squid" {print $1}')"
	if [ "$ppid" = "$vppid" ]; then
		echo kill parent $ppid
		if [ "$dryrun" != "yes" ]; then
			kill $ppid
			sleep 2
		fi
	fi

	# kill the pid twice with TERM
	echo kill child $pid
	if [ "$dryrun" != "yes" ]; then
		kill $pid
		sleep 2
		echo kill child again $pid
		kill $pid
	fi
}

# Returns the value of the specified EP option if it exists.
get_option() {
	local instance="$1"
	local option="$2"

	local primary=`echo $instance | sed -E 's/-[0-9]+//'`

	cat /etc/squid${instance}.conf /etc/sq/squid${primary}_options.conf \
	  | sed -Ene 's/^(if-(before|after)-release[[:blank:]]+[0-9]+[[:blank:]]+)?'${option}'[[:blank:]]+(.*)$/\3/p'
}

# Returns the approriate pid filename for the instance/child.
get_pidfile() {
	local squidnum="$1"
	local kidnum="$2"

	if [ "$kidnum" = "" ]; then
		local pidfile=`get_option $squidnum pid_filename`
		pidfile=${pidfile:-/logs/squid/squid$squidnum.pid}
	else
		local pid_prefix=`get_option $squidnum mpep_pid_filename_prefix`
		pid_prefix=${pid_prefix:-/logs/squid/squid$squidnum}
		if [ "$kidnum" = "1" ]; then
			local pidfile="$pid_prefix.pid"
		else
			local pidfile="$pid_prefix-$kidnum.pid"
		fi
	fi

	echo -n "$pidfile"
}

# Determines if the pidfile matches the newest running process.
# Updates the pidfile if it's missing or wrong.
# Removes the pidfile if the process is not running.
# Sets global $startsquid to "yes" if an expected process is not running.
# TODO reconsider $startsquid in context of child processes?
check_pidfile() {
	local squidnum="$1"
	local pidfile="$2"
	local prep_re="$3"

	fpid=$(cat $pidfile 2>/dev/null)

	newpid=$(pgrep -U squid -n -f "$pgrep_re")

	# if there is no squid running, then skip
	if [ -z "$newpid" ]; then
		echo WARNING squid"$squidnum" is down
		if [ -f "$pidfile" ]; then
			echo WARNING removing "$pidfile"
			[ "$dryrun" = "yes" ] || rm -f "$pidfile"
		fi
		if [ $start = yes ]; then
			echo WARNING starting "squid$squidnum"
			echo WARNING queueing rc.squid
			startsquid=yes
		fi
		return 0
	fi

	if [ -z "$fpid" ]; then
		# if the pidfile is empty, write the newest squid's pid into it
		echo WARNING writing "$newpid" to "$pidfile"
		[ "$dryrun" = "yes" ] || echo "$newpid" > "$pidfile"
	else
		# if the pidfile has something in it, make sure it is one of the running squids
		pgrep -U squid -f "$pgrep_re" | grep -q "$fpid"
		if [ $? != 0 ]; then
			# the pidfile contains a pid, but not of one of the running squidX's
			echo WARNING changing "$pidfile" from "$fpid" to "$newpid"
			[ "$dryrun" = "yes" ] || echo "$newpid" > "$pidfile"
		fi
	fi
}

# Check all child pidfiles for Native Multi-Process EP.
check_mpep_pidfiles() {
	local squidnum=$1
	local numkids=$2

	for kidnum in `seq 1 $numkids`; do
		local pidfile=`get_pidfile $squidnum $kidnum`
		local pgrep_re="squid.*child.*squid$squidnum.conf.*-c $kidnum( |$)"

		check_pidfile "$squidnum" "$pidfile" "$pgrep_re"
	done
}

# check the lone pidfile for Single-Process EP
check_spep_pidfile() {
	local squidnum=$1

	local pidfile=`get_pidfile $squidnum`
	local pgrep_re="squid.*child.*squid$squidnum.conf"

        check_pidfile "$squidnum" "$pidfile" "$pgrep_re"
}

# if all squids are down, that is outside of the scope of this script, just exit
# also, kill any squids that don't have an associated squidX.conf

ps -U squid -o pid,ppid,command | sed -Ene 's,^[[:blank:]]*([0-9]+)[[:blank:]]+([0-9]+)[[:blank:]]+[^[:blank:]]*squid[[:blank:]].*/etc/squid(([0-9]+)(-[0-9]+)?)\.conf.*,\1 \2 \3,p' | (
	anysquids=no; while read args; do
		set -- $args
		pid="$1"
		ppid="$2"
		squidnum="$3"
		squidconf="/etc/squid$squidnum.conf"

		if [ ! -f "$squidconf" ]; then
			echo WARNING "$pid" is running with no corresponding "$squidconf"
			killsquid $pid $ppid
			continue
		fi

		anysquids=yes
	done

	# this code is running in an inner shell, exiting here sets $? which is read
	# just a few lines below here
	if [ $anysquids = yes ]; then
		exit 0
	fi
	exit 1
)

# if there are no squids running, exit cleanly
if [ $? = 1 ]; then
	exit 0
fi

# make sure pidfiles are sane when squid*.conf exists

startsquid=no

for squidconf in /etc/squid*.conf; do
	instance=$(echo "$squidconf" | egrep -o "[0-9]+(-[0-9]+)?")

	numkids=`get_option $instance mpep_num_children`
	if echo $numkids | grep -q '[1-9]'; then
		check_mpep_pidfiles $instance $numkids
	else
		check_spep_pidfile $instance
	fi
done

if [ $start = yes -a $startsquid = yes -a "$dryrun" != "yes" ]; then
	/etc/sq/rc.squid
	# sleep a little to make sure pidfiles exist
	sleep 5
fi


# kill off other edgeprisms

ps -U squid -o pid,ppid,command | sed -Ene 's,^[[:blank:]]*([0-9]+)[[:blank:]]+([0-9]+)[[:blank:]]+[^[:blank:]]*squid[[:blank:]].*/etc/squid(([0-9]+)(-[0-9]+)?)\.conf([[:blank:]]+-c[[:blank:]]+([0-9]+))?.*$,\1 \2 \3 \7,p' | while read args; do
	set -- $args
	pid="$1"
	ppid="$2"
	squidnum="$3"
	kidnum="$4"

	pidfile=`get_pidfile $squidnum $kidnum`
	squidconf="/etc/squid$squidnum.conf"
	fpid=$(cat $pidfile 2>/dev/null)

	# if squidconf doesn't exist, skip, because we've already killed it
	# in the first loop above
	if [ ! -f "$squidconf" ]; then
		continue
	fi

	if [ -z "$fpid" ]; then
		echo WARNING assuming newly started "$pid" "$pidfile" is unreadable
		continue
	fi

	# don't kill if pid matches pidfile
	if [ "$pid" = "$fpid" ]; then
		echo pid "$pid" matches pid in "$pidfile"
		continue
	fi

	# abort if pidfile references something other than squid
	vfpid="$(ps -U squid -o pid,comm "$fpid" | awk '$2 == "squid" {print $1}')"
	if [ "$fpid" != "$vfpid" ]; then
		echo WARNING pid in "$pidfile" does not point to an active squid process
		continue
	fi

	# don't kill if there are any open handles to any swap.state files
	lsof -n -p "$pid" 2>/dev/null | grep -q swap.state
	if [ $? = 0 ]; then
		echo pid "$pid" has swap.state files open
		continue
	fi

	# calculate the age of the new squid1
	fmtime=$(awk '{split($8,tval,","); print tval[1]}' /proc/"$fpid"/status)

	if [ $? != 0 -o "$fmtime" -lt 1000000000 ]; then
		echo WARNING unable to determine age of $fpid
		continue
	fi

	age=$(echo $curtime - $fmtime | bc)
	
	if [ $? != 0 ]; then
		echo WARNING age calculation failed for fpid $fpid
		continue
	fi

	# only check for existing connections if the new process is young
	if [ "$age" -lt "$maxage" ]; then
		# don't kill if there are any established tcp connections
		lsof -n -p "$pid" 2>/dev/null | grep -q ESTABLISHED
		if [ $? = 0 ]; then
			echo pid "$pid" has established tcp connections
			continue
		fi
	fi

	# now we have decided to kill this pid
	killsquid $pid $ppid
done
