#!/usr/bin/env perl
use strict;
use warnings;

# various required modules
use Digest::MD5;
use Data::Dumper;
use Sys::Syslog;
use Sys::Hostname;
use File::stat;
use File::Path qw(make_path);
use File::Basename;
use Fcntl qw(:flock);
use Getopt::Std;
use MIME::QuotedPrint qw(encode_qp);
use POSIX;
use JSON;
my $VERSION = "4.1";

# globals
# setup options
my %args = ();
getopts( "d:s:a:c:", \%args );
my $default_config = do { local $/; <DATA> };
my $config_file = $args{'c'}
  if defined $args{'c'} || '/usr/local/etc/squid_dump.json';
my $config    = read_json($config_file)        || decode_json($default_config);
my $gdb_bin   = $config->{Global}->{gdb_bin}   || '/usr/local/bin/gdb';
my $ident_bin = $config->{Global}->{ident_bin} || '/usr/bin/ident';
my $squid_bin = $config->{Global}->{squid_bin} || '/usr/local/squid/bin/squid';
my $tail_bin  = $config->{Global}->{tail_bin}  || '/usr/bin/tail';
my $tar_bin   = $config->{Global}->{tar_bin}   || '/usr/bin/tar';
my $squid_bin_path =
  $config->{Global}->{squid_bin_path} || '/usr/local/squid/bin/';
my $cachexlog_dir = $config->{Global}->{cachexlog_dir} || '/logs/squid/';
my $dedicated_id_file =
  $config->{Global}->{dedicated_id_file} || '/etc/sq/dedicated_id.txt';
my $optimized_cfg_file =
  $config->{Global}->{optimized_cfg_file} || '/etc/sq/optimized';
my $oldsquid_bin =
  $config->{Global}->{oldsquid_bin} || '/usr/local/squid/bin/squid.old';
my $archive_dir   = $config->{Global}->{archive_dir}   || '/logs/crash/';
my $lockfile_path = $config->{Global}->{lockfile_path} || '/var/run/';
my $backtrace_cmd =
  $config->{Global}->{backtracecmd} || '/usr/local/bin/submitbacktrace.sh';
my $sqconfigdir      = $config->{Global}->{sqconfigdir}      || '/etc/sq/';
my $sqconfigpath     = $config->{Global}->{sqconfigpath}     || '/etc/';
my $enablecfgbackup  = $config->{Global}->{enablecfgbackup}  || 1;
my $maxconfigbackups = $config->{Global}->{maxconfigbackups} || 10;
my $mode             = 0644;

# globals we'll modify if we need
my $debug           = -1;
my $emailreport     = $config->{Global}->{emailreport} || 1;
my $enablebacktrace = $config->{Global}->{enablebacktrace} || 1;
my $sqnum           = $config->{Global}->{sqnum} || "*";
my $age             = $config->{Global}->{age} || 300;
my $sleeptime       = $config->{Global}->{sleeptime} || 10;

# this is the maximum amount of cores to keep per ep instance
# this is in addition to the normal core dump file, so it will
# only keep N extra cores..
my $maxcores = $config->{Global}->{maxcores} || 1;
my $backuppaths = $config->{backuppaths};

# setup some shortcut mapping for convenience
my %cachemap = (
    0   => 'squid2',
    1   => 'squid1',
    2   => 'squid1',
    3   => 'squid3',
    '*' => '*'
);

%cachemap = %{ $config->{Global}->{cachemap} }
  if defined $config->{Global}->{cachemap};

# create a lockfile or bail
getlock( $lockfile_path . "." . basename($0) );

# parse our options for age , squid number, debug
$debug = $args{'d'} if defined $args{'d'};
$sqnum = $args{'s'} if defined $args{'s'};
$age   = $args{'a'} if defined $args{'a'};

doDbg( 1, $debug, $sqnum, $age );

# make sure we can save our exported crash data
File::Path::make_path( $archive_dir, { error => \my $patherr } );
if (@$patherr) {
    warn("Could not create $archive_dir, will not save crash info");
}

# and a reverse map
my %squidmap = reverse %cachemap;

# find any coredumps
my @coredumps = </d[0-9]/sqcache${squidmap{$sqnum}}/squid.core>;

# process any cores we found that match the necessary parameters
foreach my $dump (@coredumps) {
    doDbg( 1, "Found $dump" );

    # gather file stat data on our coredump
    my $core = stat($dump);

    # check age on core, if it's older than our configured time, skip it
    next if ( $core->mtime < ( time() - $age ) );

    # which squid crashed
    my ($sq) = ( $dump =~ m#/d0/sqcache([0-9])/squid.core$# );

    # rename the core to a temp name
    $dump = renameCore( $dump, $dump . ".processing" );

    # wait for the core to finish writing out
    waitCoreComplete( $dump, $sleeptime );
    my $hash = doFingerprint( $dump, $sq, $core );
    $hash->{squid_dump_ver} = $VERSION;
    doLog($hash);
    $hash = backup_unknown_cores( $hash, $maxcores );
    $hash = backup_configs( $hash, $maxconfigbackups );

    # build a report and store it
    my $report = generateReport($hash);
    my $json   = generateReportv2($hash);
    next if check_known_signatures($hash);
    next unless $config->{Global}->{emailaddress};
    email_report( $hash, $report, $json, $config->{Global}->{emailaddress} );
}
exit(0);

###############
# sub functions
###############

sub email_report {
    return unless $emailreport;
    my ( $hash, $report, $json, $emailaddress ) = @_;
    my $boundary = $hash->{stack}->{md5v2};
    my $hostname = $hash->{sys}{hostname};
    next unless $hash;
    next unless $report;
    next unless $json;
    next unless $emailaddress;
    my $report_qp = encode_qp($report);
    my $json_qp = encode_qp($json);
    open( MAIL, '|-', '/usr/sbin/sendmail -t' );
    print MAIL <<"HEADER";
From: root\@$hostname
To: $emailaddress
Subject: Squid Dump Version: $hash->{ep}{version} Signature $boundary
MIME-Version: 1.0
Content-Type: multipart/mixed; boundary="----=_$boundary"

------=_$boundary
Content-Type: text/plain;
Content-Transfer-Encoding: quoted-printable

HEADER

    print MAIL "$report_qp";

    print MAIL <<"JSONBOUNDARY";

------=_$boundary
Content-Type: text/json; name="$boundary-report_v2.json"
Content-Transfer-Encoding: quoted-printable
Content-Disposition: attachment; filename="$boundary-report_v2.json"

JSONBOUNDARY

    print MAIL $json_qp;

    print MAIL <<"END";

------=_$boundary--
END

    close(MAIL) or warn "sendmail failed $? : $!";

}

# if the signature matches a known signature
# then just unlink the file, no reason to continue.
sub check_known_signatures {
    my $hash = shift;

    if (
        (
            defined $config->{Version}->{1}->{KnownSigatures}
            ->{ $hash->{stack}{md5} }
        )
        or (
            defined $config->{Version}->{2}->{KnownSigatures}
            ->{ $hash->{stack}{md5v2} } )
      )
    {
        unlink $hash->{core}{path};
        return 1;
    }
    return 0;

}

sub read_json {
    my $config_file = shift || return 0;
    my $return;
    return unless -e $config_file;
    local $/;
    open( my $fh, '<', $config_file ) || return;
    my $json_text = <$fh>;
    $return = decode_json($json_text);
    close $fh || return;
    return $return;
}

# rename wrapper function
# to keep things clean for later
sub renameCore {
    my ( $core, $newname ) = @_;
    rename $core, $newname || die "unable to move core to new name";
    return $newname;
}

# Simple Sub Routine to see if a core has finished writing
# within a given timeframe (say 10 seconds.
sub waitCoreComplete {
    my ( $dump, $sleeptime ) = @_;
    my $core_stat = stat($dump);
    my $old_size  = 0;
    while ( $core_stat->size != $old_size ) {
        $old_size = $core_stat->size;
        sleep $sleeptime;
        $core_stat = stat($dump);
    }
    return 1;
}

# this function will help to keep unknown / none blacklisted core dumps
# we will set a limit to the total number of core dumps
# in order to help with debugging
sub backup_unknown_cores {
    my ( $hash, $maxcores ) = @_;
    my $sq  = $hash->{core}{path};
    my $sig = $hash->{stack}{md5v2};
    if ( check_known_signatures($hash) ) {
        unlink($sq);
        $hash->{core}{path} = "KNOWN SIGNATURE CORE REMOVED";
        return $hash;
    }
    ( my $sq_n = $sq ) =~ s/\.processing//;
    my $newname = "$sq_n.$sig";

    # if the signature already exists then keep the new one

    if ( -e $newname ) {
        rename $sq, "$newname";

    }
    else {
        my @totalcores =
          sort { ( stat $a )->mtime <=> ( stat $b )->mtime } <$sq_n*>;
        while ( ( $#totalcores - 1 ) >= $maxcores ) {
            my $remove = shift(@totalcores);
            next if $remove =~ m/processing/;
            doDbg( 2, $remove );
            unlink $remove;
        }
        rename $sq, "$newname";

    }
    $hash->{core}{path} = "$newname";
    change_mode( $mode, $hash->{core}{path} );
    return $hash;
}

sub change_mode {
    my ( $mode, $file ) = @_;
    if ( -e $file ) {
        chmod $mode, $file;
    }
}

sub tailCacheXlog {
    my ( $sqnum, $lines ) = @_;
    my $squidinstance = $cachemap{$sqnum};
    my ($cachelognum) = ( $squidinstance =~ m/squid(\d+)/ );
    my $cachexlog     = $cachexlog_dir . "/cache" . $cachelognum . ".log";
    my $output        = executeBin( $tail_bin, "-n", $lines, $cachexlog );
    my $return        = join( "\n", @{$output} );
    return ( $return, $cachexlog );
}

sub backup_configs {
    my $hash = shift;
    return $hash unless ($enablecfgbackup);
    my @backupfiles;

    $hash->{ep}{configs} =
        $archive_dir
      . "ConfigBackup" . "."
      . $hash->{ep}{version} . "."
      . $hash->{stack}{md5v2} . "."
      . "tar.bz2";

    if ( -e $hash->{ep}{configs} ) {
        return $hash;
    }

    foreach my $path ( @{$backuppaths} ) {
        push( @backupfiles, </$path> );
    }
    executeBin( $tar_bin, '-cjf', $hash->{ep}{configs}, @backupfiles );
    my @configbackups =
      sort { ( stat $a )->mtime <=> ( stat $b )->mtime }
      <$archive_dir/ConfigBackup*.tar.bz2>;
    while ( ( $#configbackups - 1 ) >= $maxconfigbackups ) {
        my $remove = shift(@configbackups);
        doDbg( 2, $remove );
        unlink $remove;
    }
    change_mode( $mode, $hash->{ep}{configs} );
    return $hash;
}

sub getFileContents {
    my $file = shift;
    return "Unknown" unless -e $file;
    open my $file_fh, '<', $file || warn "cannot open file $file: $!";
    my $contents = do { local $/; <$file_fh> }
      || "Unknown";
    close $file_fh;
    return $contents;
}

sub parseIdentInfo {
    my $ident_data = shift;
    my %sq_ident_hash;
    foreach ( split( "\n", $ident_data ) ) {
        next unless $_ =~ m/^\s+\$.+\: .+\s+\$/;
        my ( $codesource, $params ) = ( $_ =~ m/^\s+\$(.+)\: (.+)\s+\$/ );
        if ( $params =~ m/\=/ ) {
            foreach my $token ( split( /\s\s/, $params ) ) {
                if ( $token =~ m/\=/ ) {
                    my ( $key, $raw_val ) = split( /=/, $token );
                    ( my $val = $raw_val ) =~ s/\s+\$$//;
                    $sq_ident_hash{$key} = $val;
                }
            }
        }
    }
    return \%sq_ident_hash;
}

sub findSquidBin {
    my ( $sq_bin_path, $dump, $cmds ) = @_;
    my @squid_bins = <$sq_bin_path/squid*>;
    my $best       = 'Unknown';
    my ( $release, $stack, $build );

    # New method to find correct binary (This is for EP  >= 4.3.1.0 )
    foreach my $sqbin (@squid_bins) {
        my $releasecheck =
          traceCore( $dump, $sqbin, [ 'print RELEASE', 'quit' ] );
        my $releasevar = checkRelease($releasecheck);
        next if $releasevar eq 'Unknown';
        $best    = $sqbin;
        $release = $releasevar;
    }

    # Include an enhanced  historic method of finding the
    # binary for backwards compat for versions
    # for pre 4.3.1.0 version of EP that doesnt have
    # a RELEASE variable defined
    if ( $best eq 'Unknown' ) {
        my %candidates;

        foreach my $sqbin (@squid_bins) {
            my $hasmain = 0;
            $candidates{$sqbin}->{stack} = traceCore( $dump, $sqbin, $cmds );
            $candidates{$sqbin}->{functcount} = 0;

            # Foundout that the buildInfo string
            # will be found even with incorrect binary
            # instead i will count the number of functions
            # named "??"
            my ( $md5, $functions ) =
              fingerprintTraceV2( $candidates{$sqbin}->{stack} );

            foreach ( @{$functions} ) {
                next if $_ =~ m/^\?\?$/;

                # make sure that there is a main() function
                # defined in the list of functions recognized
                $hasmain = 1 if $_ = m/^main$/;
                $candidates{$sqbin}->{functcount}++;
            }

            # unless main is defined, then set the total funtion count
            # to zero
            $candidates{$sqbin}->{functcount} = 0 unless $hasmain;

        }

        # find the candidate of EP that has the MOST recognized Function names
        my @bestchoices =
          sort {
            $candidates{$a}->{functcount} <=> $candidates{$b}->{functcount}
          }
          keys %candidates;

        # get the top result
        $best = pop @bestchoices;

        # if the count of recognized function names , or the buildinfo
        # is equal to zero define best as "Unknown"
        $best = "Unknown" if $candidates{$best}->{functcount} eq 0;
        $release = "Pre 4.3.1.0 version, no RELEASE var defined";
    }

    # exit out if we cant identify the binary for the core
    exit(1) if $best eq 'Unknown';
    $stack = traceCore( $dump, $best, $cmds );
    $build = checkInfo($stack);

    # exit out if we cant identify the build, or relase of the core
    exit(1) if $build eq 'Unknown';
    exit(1) if $release eq 'Unknown';

    # get process uptime
    my $uptimeCmds = [ 'print current_time.tv_sec-squid_start.tv_sec', 'quit' ];
    my $uptimeOutput = traceCore( $dump, $best, $uptimeCmds );
    my $uptime = checkUptime($uptimeOutput);

    return ( $best, $stack, $build, $release, $uptime );
}

# this is essentially a 2nd main, its the primary loop for doing all the meat of the work
sub doFingerprint {
    my ( $dump, $sq, $core ) = @_;
    my $hash;
    my $sqbin = $squid_bin;

    $hash->{sqnum}         = $sq;
    $hash->{squidinstance} = $cachemap{$sq};
    $hash->{core}{stat}    = $core;
    $hash->{core}{mtime}   = $core->mtime;
    $hash->{core}{ctime}   = $core->ctime;
    $hash->{core}{size}    = $core->size;
    $hash->{core}{path}    = $dump;

    # get our stack trace from gdb
    (
        $squid_bin,            $hash->{stack}{data},
        $hash->{stack}{build}, $hash->{stack}{release},
        $hash->{ep}{uptime}
    ) = findSquidBin( $squid_bin_path, $dump, $config->{GDBCommands} );

    # get some lines of cacheX log
    ( $hash->{ep}{cachelog}, $hash->{ep}{cachelogfile} ) =
      tailCacheXlog( $sq, $config->{Global}->{tail_lines} );

    # gather some vitals on our squid binary
    $hash->{ep}{path}        = $squid_bin;
    $hash->{ep}{md5}         = doMD5($squid_bin);
    $hash->{ep}{stat}        = stat($squid_bin);
    $hash->{ep}{ident}       = ident($squid_bin);
    $hash->{ep}{identstruct} = parseIdentInfo( $hash->{ep}{ident} );
    $hash->{ep}{optimized} =
      $hash->{ep}{identstruct}->{build_mode} || "unknown";
    my $ep_version = squid('-v');
    $ep_version =~ s/\W/\./g;
    $ep_version =~ s/\W/\./g;
    $hash->{ep}{version} = $ep_version;
    my $dedicatedid = getFileContents($dedicated_id_file);

    $hash->{ep}{dedicatedid} = getFileContents($dedicated_id_file);
    chomp( $hash->{ep}{dedicatedid} );
    $hash->{ep}{optimizedcfg} = getFileContents($optimized_cfg_file);

    $hash->{sys}{uname}    = join( " ", POSIX::uname() );
    $hash->{sys}{hostname} = hostname();
    $hash->{sys}{time}     = localtime( time() );

    # generate a fingerprint based on the stack trace
    ( $hash->{stack}{md5}, $hash->{stack}{functions} ) =
      fingerprintTrace( $hash->{stack}{data} );

    # generate a v2 fingerprint for the stacktrace
    ( $hash->{stack}{md5v2}, $hash->{stack}{functionsv2} ) =
      fingerprintTraceV2( $hash->{stack}{data} );

    # create our report name
    $hash->{stack}{report} =
        $archive_dir
      . $cachemap{$sq} . "."
      . $hash->{ep}{version} . "."
      . $hash->{stack}{md5v2} . "."
      . $core->ctime
      . ".report";

    $hash->{stack}{reportv2} =
        $archive_dir
      . $cachemap{$sq} . "."
      . $hash->{ep}{version} . "."
      . $hash->{stack}{md5v2} . "."
      . $core->ctime . ".v2"
      . ".report";

    doBacktrace( $dump, $squid_bin, $hash );
    return $hash;

}

sub doBacktrace {
    my ( $dump, $squid, $hash ) = @_;
    return unless $enablebacktrace;
    return unless -x $backtrace_cmd;
    return unless -e $dump;
    my %kv_pair = (
        version     => '@version_string:string',  # ptrace extracts these
        build       => '@buildInfo:string',
        uptime      => $hash->{ep}{uptime},
        signature   => $hash->{stack}{md5},
        signatureV2 => $hash->{stack}{md5v2},
        optimized   => $hash->{ep}{optimized}
    );

    my $tags = join( ",", map { "$_:$kv_pair{$_}" } keys %kv_pair );

    my $data =
      executeBin( $backtrace_cmd, $dump, $squid, 'squid', 'squid', $tags );
    return $data;
}

# this function calls our gdb script generator then runs gdb to get the output from the commands
sub traceCore {
    my ( $dump, $binary, $gdbcommands ) = @_;
    my ( $stacktrace, $gdb_script );
    return unless -e $dump;
    $gdb_script = createScript($gdbcommands);
    $stacktrace =
      gdb( '-nx', '-x', $gdb_script, '--batch', '-q', '-c', $dump, $binary );
    doDbgDump( 2, $stacktrace );
    unlink($gdb_script);
    return ($stacktrace);
}

# a convience function for calling executables and getting their output
sub executeBin {
    my @args = @_;
    doDbgDump( 2, @args );
    my $command = shift @args;
    my @output  = ();

    die "$command is not executable" if ( !-x $command );

    open( my $fh, '-|', $command, @args ) or die "$command @args failed: $!";
    chomp( @output = <$fh> );
    close($fh) or warn "failure to execute $command: $!";
    return \@output;
}

# convenience method for executing gdb
sub gdb {
    my @args = @_;
    my $output = executeBin( $gdb_bin, @args );
    return $output;
}

sub squid {
    my @args   = @_;
    my $output = executeBin( $squid_bin, @args );
    my $return = join( "\n", @{$output} );
    return $return;
}

sub ident {
    my @args   = @_;
    my $output = executeBin( $ident_bin, @args );
    my $return = join( "\n", @{$output} );
    return $return;
}

# generate a temporary gdb script for running commands as you can't easily
# stream commands to gdb directly
sub createScript {
    my $gdbcmds    = shift;
    my $gdb_script = "/tmp/gdbscript." . time();

    open( my $fh, '>', $gdb_script ) or die "Could not open $gdb_script: $!";
    foreach ( @{$gdbcmds} ) {
        print $fh $_ . "\n";
    }
    close($fh);
    return $gdb_script;
}

# look for a valid buildInfo variable, return "Unknown" if we don't find one,
# otherwise return the buildInfo
sub checkInfo {
    my $stacktrace = shift;
    my $buildinfo;

    foreach ( @{$stacktrace} ) {
        if ( $_ =~ /^\$\d = \"Version info: (.*?)\"$/ ) {
            $buildinfo = $1;
            last;
        }
    }
    unless ($buildinfo) {
        $buildinfo = "Unknown";
    }
    return $buildinfo;
}

sub checkRelease {
    my $stacktrace = shift;
    my $release;

    foreach ( @{$stacktrace} ) {

        #$1 = 0x802818280 "release=4.3.1.0-010  build_mode=unopt"
        if ( $_ =~ /^\$\d = 0x[a-f0-9A-F]+ \"(release=.+  build_mode=\w+)\"/ ) {

            $release = $1;
            last;
        }
    }
    unless ($release) {
        $release = "Unknown";
    }
    return $release;
}

sub checkUptime {
    my $stacktrace = shift;
    my $uptime;

    foreach ( @{$stacktrace} ) {

        if ( $_ =~ /^\$\d = (\d+)/ ) {
            $uptime = $1;
            last;
        }
    }
    unless ($uptime) {
        $uptime = "-1";
    }
    return $uptime;
}

# take the functions from the stacktrace, ignoring the buildInfo variable, and create a fingerprint against those
# functions. Crashes caused by the same routines will have the same fingerprint as we're only using function names
# for fingerprinting.
sub fingerprintTrace {
    my $stacktrace = shift;
    my @functions  = ();
    my $ctx        = Digest::MD5->new;

    foreach ( @{$stacktrace} ) {
        if ( $_ =~ m/^\#(\d+)\s+\S+\sin\s(\S+).+\.c\:[0-9]+/ ) {
            doDbg( 3, "Function: $2" );
            next
              if ( defined $config->{Version}->{1}->{FunctionBlacklist}->{$2} );

            push( @functions, $2 );
        }
    }
    doDbgDump( 3, @functions );
    $ctx->add(@functions);
    return ( $ctx->hexdigest, \@functions );
}

# Slightly different fingerprint for a v2 signature
sub fingerprintTraceV2 {
    my $stacktrace = shift;
    my @functions  = ();
    my $ctx        = Digest::MD5->new;
    foreach ( @{$stacktrace} ) {
        if ( $_ =~ m/^\#(\d+)\s+\S+\sin\s(\S+).+/ ) {
            doDbg( 3, "Function: $2" );
            next
              if ( defined $config->{Version}->{2}->{FunctionBlacklist}->{$2} );
            push( @functions, $2 );
        }
    }
    doDbgDump( 3, @functions );
    $ctx->add(@functions);
    my $md5sum = $ctx->hexdigest;
    return ( $md5sum, \@functions );
}

sub generateReportv2 {
    my $hash = shift;
    my $file = $hash->{stack}{reportv2};
    my $json = JSON->new->utf8;
    $json->convert_blessed(1);
    $json->pretty(1);
    $json->allow_blessed(1);

    my $report = $json->allow_blessed->convert_blessed->encode($hash);
    open( my $fh, '>', $file ) or warn "Unable to write to $file: $!";
    print $fh "$report";
    close($fh);
    doDbg( 2, $report );
    change_mode( $mode, $file );
    return $report;
}

# generate and write a report containing all the vital information
sub generateReport {
    my $hash     = shift;
    my $time     = localtime( time() );
    my $fp       = $hash->{stack}{md5};
    my $fpv2     = $hash->{stack}{md5v2};
    my $st       = join( "\n", @{ $hash->{stack}{data} } );
    my $hostname = $hash->{sys}{hostname};
    my $file     = $hash->{stack}{report};
    my $uname    = $hash->{sys}{uname};

    my $report;

    $report .= "REPORT:\n";
    $report .= "=============================================\n";
    $report .= "DATE: $time\n";
    $report .= "Host: $hostname\n";
    $report .= "UNAME: $uname\n";
    $report .= "Squid Dump Version: " . $hash->{squid_dump_ver} . "\n";
    $report .= "\n\n";
    $report .= "COREDUMP:\n";
    $report .= "=============================================\n";
    $report .= "PATH: " . $hash->{core}{path} . "\n";
    $report .= "SIZE: " . $hash->{core}{stat}->size . "\n";
    $report .=
      "CTIME: " . scalar( localtime( $hash->{core}{stat}->ctime ) ) . "\n";
    $report .=
      "MTIME: " . scalar( localtime( $hash->{core}{stat}->mtime ) ) . "\n";
    $report .= "\n\n";
    $report .= "EDGEPRISM:\n";
    $report .= "=============================================\n";
    $report .= "MD5: " . $hash->{ep}{md5} . "\n";
    $report .= "PATH: " . $hash->{ep}{path} . "\n";
    $report .= "VERSION: " . $hash->{ep}{version} . "\n";
    $report .= "BUILD: " . $hash->{stack}{build} . "\n";
    $report .= "IDENT: \n" . $hash->{ep}{ident} . "\n";
    $report .= "DEDICATED ID: " . $hash->{ep}{dedicatedid} . "\n";
    $report .= "CONFIG BACKUP: " . $hash->{ep}{configs} . "\n"
      if $hash->{ep}{configs};
    $report .= "RELEASE: " . $hash->{stack}{release} . "\n";
    $report .= "SQUID INSTANCE: " . $hash->{squidinstance} . "\n";
    $report .= "UPTIME: " . $hash->{ep}{uptime} . "\n";
    $report .= "\n\n";
    $report .= "STACKTRACE:\n";
    $report .= "=============================================\n";
    $report .= "FINGERPRINT: $fp\n";
    $report .= "FINGERPRINT_V2: $fpv2\n";
    $report .= "\n\n";
    $report .= "Raw:\n";
    $report .= "$st\n\n";
    $report .= "CACHEX.LOG: $hash->{ep}{cachelogfile}\n";
    $report .= "=============================================\n";
    $report .= $hash->{ep}{cachelog} . "\n\n";
    open( my $fh, '>', $file ) or warn "Unable to write to $file: $!";
    print $fh $report;
    close($fh);
    doDbg( 2, $report );
    change_mode( $mode, $file );
    return $report;
}

# convenience routine for generating an md5 hash of a given file
sub doMD5 {
    my $file = shift;
    my $ctx  = Digest::MD5->new;
    doDbg( 2, "Creating md5 for $file" );
    $ctx->reset;
    open( my $fh, '<', $file ) or die "Could not open $file: $!";
    binmode($fh);
    $ctx->addfile($fh);
    close($fh);
    return ( $ctx->hexdigest );
}

# debug routine for printing simple statements
sub doDbg {
    my $dbglevel  = shift;
    my $statement = shift;
    if ( $debug >= $dbglevel ) {
        print STDERR "DBG $dbglevel: " . $statement . "\n";
    }
}

# debug routine for dumping data structures
sub doDbgDump {
    my $dbglevel = shift;
    if ( $debug >= $dbglevel ) {
        print "\n";
        print "DBG $dbglevel: " . Dumper(@_) . "\n";
        print "\n";
    }
}

#create a lockfile
sub getlock {
    my ($lockfile) = @_;
    open( LOCK, ">$lockfile" ) or die "can't write to $lockfile: $!";
    my $res = flock( LOCK, LOCK_EX | LOCK_NB )
      || die "$0: can't lock $lockfile ($!), am i already running?";
}

# notify syslog of our information for splunking purposes
sub doLog {

    my $hash = shift;
    syslog( 'warning',
            "squid="
          . $cachemap{ $hash->{sqnum} }
          . " signature="
          . $hash->{stack}{md5}
          . " signatureV2="
          . $hash->{stack}{md5v2}
          . " version="
          . $hash->{ep}{version}
          . " build="
          . $hash->{stack}{build}
          . " dedicatedid="
          . $hash->{ep}{dedicatedid}
          . " optimized="
          . $hash->{ep}{optimized} );
}

__DATA__
{
  "Global": {
    "cachemap": {
      "0": "squid2",
      "1": "squid1",
      "2": "squid1",
      "3": "squid3",
      "*": "*"
    },
    "archive_dir": "/logs/crash/",
    "gdb_bin": "/usr/local/bin/gdb",
    "ident_bin": "/usr/bin/ident",
    "tail_bin": "/usr/bin/tail",
    "tar_bin": "/usr/bin/tar",
    "cachexlog_dir": "/logs/squid/",
    "squid_bin": "/usr/local/squid/bin/squid",
    "oldsquid_bin": "/usr/local/squid/bin/squid.old",
    "squid_bin_path": "/usr/local/squid/bin/",
    "dedicated_id_file": "/etc/sq/dedicated_id.txt",
    "optimized_cfg_file": "/etc/sq/optimized",
    "sqnum": "*",
    "age": "300",
    "tail_lines": "100",
    "lockfile_path": "/var/run/",
    "sleeptime": "10",
    "maxcores": "1",
    "emailaddress": "core.ep.dev.debug@llnw.com",
    "emailreport": "1",
    "enablebacktrace": "1",
    "backtracecmd": "/usr/local/bin/submitbacktrace.sh",
    "enablecfgbackup": "1",
    "maxconfigbackups": "10"
  },
  "GDBCommands": [
    "set print element 4000",
    "set width 0",
    "set print pretty",
    "print buildInfo",
    "bt full",
    "quit"
  ],
  "backuppaths": [
    "/etc/squid*",
    "/etc/sq/*.conf",
    "/etc/sq/squid_code",
    "/etc/sq/*.txt",
    "/etc/sq/optimized",
    "/etc/sq/*.dat",
    "/etc/sq/*.cfg",
    "/etc/sq/min_squid*",
    "/etc/sq/*.json",
    "/etc/sq/viplist",
    "/etc/sq/*.cf",
    "/etc/sq/*.confg",
    "/etc/sq/rules/*",
    "/etc/sq/*.config",
    "/etc/sq/pcds_*",
    "/etc/sq/pg_ca*"
  ],
  "Version": {
    "1": {
      "FunctionBlacklist": {
        "kill": 1,
        "abort": 1,
        "comm_call_handlers": 1,
        "do_comm_select": 1,
        "main": 1,
        "comm_poll": 1
      },
      "KnownSigatures": {
        "4352330f646aefcbf46b0025c258e41e": 1
      }
    },
    "2": {
      "KnownSigatures": {
        "4352330f646aefcbf46b0025c258e41e": 1
      },
      "FunctionBlacklist": {
      }
    }
  }
}
