#!/bin/bash

# vim: ts=4 sts=4 sw=4 et ai :

# This wrapper transitions a cds host from running GridConfigClient.pl as root
# to running as xfer. If the lock file already exists and is owned by xfer,
# then this has been completed and we can just run the script as xfer. If not,
# then either it does not exist (new install or reboot clears out /var/lock)
# or it is owned by root, so before running the script we make sure all output
# directories and files exist and are owned by xfer. We get the lock while
# doing that so that we do not go changing things while a previous invocation
# is still running. If we cannot get the lock or have other problems, we log
# it to syslog + stderr and give up.

SCRIPT="/usr/local/gridconfig/bin/GridConfigClient.pl"
CACHE_DIR="/usr/local/gridconfig/cache/"
OUTPUT_DIRS=("$CACHE_DIR" "/etc/sq/" "/usr/local/llnw/mcdn/" "/usr/local/etc/cert-mgmt/pems/")
LOCK_DIR="/var/lock"
LOCK_FILE="$LOCK_DIR/GridConfigClient.lock"
KERNEL=$( uname -s )

# WRAPPER_FLOCKED should only be set when we call ourselves below.
# The "flock" tool on Linux provides a cleaner way to put a lock around a
# sections of a shell script, but the corresponding "lockf" tool on FreeBSD
# does not, so we jump through this lock-and-reexecute hoop.
if [ "$WRAPPER_FLOCKED" == "yes" ]; then
    set -e
    for dir in "${OUTPUT_DIRS[@]}"; do
        mkdir -p "$dir"
        chown -R xfer:xfer "$dir"
    done
    touch "$LOCK_FILE"
    chown xfer:xfer "$LOCK_FILE"
    exit 0
fi

do_locked_stuff() {
    # Ensure we have a place to put the lock file.
    # A fresh Vagrant salt-epos-411 does not have /var/lock.
    mkdir -p "$LOCK_DIR" || return "$?"
    chmod 1777 "$LOCK_DIR" || return "$?"

    # Lock and re-execute with env var to indicate as much.
    if [ "$KERNEL" == "FreeBSD" ]; then
        WRAPPER_FLOCKED="yes" lockf -k -t 0 "$LOCK_FILE" "$@"
    else
        WRAPPER_FLOCKED="yes" flock --verbose -n "$LOCK_FILE" "$@"
    fi
}

# The 'stat' utility has different formatting options on the two platforms.
if [ "$KERNEL" == "FreeBSD" ]; then
    LOCK_FILE_OWNER=$( stat -f "%Su" "$LOCK_FILE" 2>/dev/null )
else
    LOCK_FILE_OWNER=$( stat -c "%U" "$LOCK_FILE" 2>/dev/null )
fi

# If the lockfile exists and is owned by xfer, then we've already made the
# ownership transition and we're good to go. Otherwise, try to get the lock
# and make the transition, bailing out and logging about it if there is any
# kind of failure.
if [ -f "$LOCK_FILE" ] && [ "$LOCK_FILE_OWNER" == "xfer" ]; then
    # already good to go
    true
elif ! output=$( do_locked_stuff "$0" "$@" 2>&1 ); then
    echo "$0: $output" | logger -s
    exit 1
fi

# Remove root check from script if necessary.
if grep -q 'Not running as root' "${SCRIPT}"; then
    sed -i.ORIG '/Not running as root/ d' "${SCRIPT}"
fi

# Run the script
exec sudo -u xfer "${SCRIPT}" "$@"

