#!/usr/bin/perl -w

use strict;
use JSON; 
use List::Util qw( shuffle );
use Sys::Hostname;
use IO::Socket;
use Socket qw(:crlf);
use Sys::Syslog qw(:DEFAULT setlogsock);

my $debug = 0;
my $hostname = hostname;
my $hostfile = "/usr/local/llnw/ms-ops-hosts.json";

my ($spop,$dpop,$desthost) = "";

print "Host: $hostname\n";
if ($hostname =~ /ms-ops-1\.(\S+)\.llnw\.net/) {
    $spop = $1;
} else {
    exit(1);
} 

# tsdb feed
my $tsdb_server = "tsdb-data.lldns.net";
my $tsdb_port = "4242";
my $tsdb_sock = IO::Socket::INET->new(
    PeerHost => $tsdb_server,
    PeerPort => $tsdb_port,
    Proto    => 'tcp',
) or die "Failed to open tsdb socket: $@";

my $json_data = do {
    open(my $json_fh, "<:encoding(UTF-8)", $hostfile) or die "Could not open $hostfile: $!\n";
    local $/;
    <$json_fh>
};

my $js = new JSON;
my $jsdata = $js->decode($json_data);

# shuffle our hosts so we're not banging on them concurrently
my @boxes = shuffle(@{$jsdata->{boxes}});

foreach (@boxes) {
    # add a random sleep of up to 10 seconds between hosts to tweak things even more
    sleep(rand(5));
    $desthost = $_;
    if ($desthost =~ /ms-ops-1\.(\S+)\.llnw\.net/) {
        $dpop = $1;
    } else {
        next;
    }
    next if ($dpop eq $spop);
    my $json = new JSON; 
    my $data = `iperf3 -J -i 0 -n 2M -b 100M -V -S 0x04 -c $desthost`; 
    if ($? == 0) {
        my $text = $json->allow_nonref->utf8->relaxed->decode($data); 
        my @streams = @{$text->{'end'}->{'streams'}};

        send_metric("app.iperf3.seconds", sprintf("%.02f", $streams[0]->{'sender'}->{'seconds'}));
        send_metric("app.iperf3.Mbps", sprintf("%.02f", $streams[0]->{'sender'}->{'bits_per_second'} / 1024 / 1024));
        send_metric("app.iperf3.retransmits", $streams[0]->{'sender'}->{'retransmits'});
        send_metric("app.iperf3.bytes", sprintf("%.02f", $streams[0]->{'sender'}->{'bytes'} / 1024 / 1024));

        if ($streams[0]->{'sender'}->{'retransmits'} > 20) {
            syslog("error", "iperf3: [ERROR] %s -> %s: %d retransmits", $spop, $dpop, $streams[0]->{'sender'}->{'retransmits'});
        } elsif ($streams[0]->{'sender'}->{'retransmits'} > 10) {
            syslog("warning", "iperf3: [WARNING] %s -> %s: %d retransmits", $spop, $dpop, $streams[0]->{'sender'}->{'retransmits'});
        } elsif ($streams[0]->{'sender'}->{'retransmits'} > 0) {
            syslog("info", "iperf3: [INFO] %s -> %s: %d retransmits", $spop, $dpop, $streams[0]->{'sender'}->{'retransmits'});
        }
    }

    my @ping = `/bin/ping -U -W 5 -c 10 $desthost`;
    foreach my $line (@ping) {
        if ($line =~ /(\d+)% packet loss/) {
            send_metric("app.ping.packet_loss", $1);
        }
        if ($line =~ /^rtt .*?mdev = (\S+)\/(\S+)\/(\S+)\/(\S+) .*$/) {
            send_metric("app.ping.rtt.min", $1);
            send_metric("app.ping.rtt.avg", $2);
            send_metric("app.ping.rtt.max", $3);
            send_metric("app.ping.rtt.mdev", $4);
        }
    }
}

close($tsdb_sock);

sub send_metric {
    my ($key, $value) = @_;

    my $time = time();

    if ($debug) {
        print STDOUT "put $key $time $value source=$spop dest=$dpop\n";
    } else {
        print $tsdb_sock "put $key $time $value source=$spop dest=$dpop", CRLF;
    }
}
