#!/usr/bin/perl

use IO::Socket;
use Sys::Syscall qw(:epoll);
use POSIX qw(nice);
use Fcntl qw(:flock);
use Time::HiRes;
use Data::Dumper;
use JSON;

my $port = 7203;
my $logfile = '/var/log/lhbd3';
my $lockfile = '/var/run/lockedrun.lhbd3';

my $POLLIN = 1;

my $res;
my $now;

my %addrhostname;
my %sock;
my %remaddr;
my %remport;
my %hostname;
my %addrsock;
my %active;
my %ctime;
my %stats;
my %buf;
my %buflen;
my %acctmap;

my $tsdb;
	
my $debug = 0;

open(LOCK, ">$lockfile") or die "can't write to $lockfile: $!";
$res = flock(LOCK, LOCK_EX|LOCK_NB);
if (! $res) {
	if (-t STDIN) {
		die "$0: can't lock $lockfile ($!), am i already running?";
	}
	exit 0;
}

my $lfh = IO::Socket::INET->new(Listen => 1024,
			LocalPort => $port,
			Reuse => 1,
			Proto => 'tcp') or die "can't listen on $port: $!";
print "$0: listening on port $port, backgrounding\n";
fork && exit;

# fork to two processes
fork;
&log("$0: ready for connections\n");

nice(-3) or warn "nice: $!";

$SIG{PIPE} = 'IGNORE';
my $epoll = epoll_create(16384);
if (! defined $epoll) {
	die "can't epoll_create(): $!";
}
if ($epoll < 0) {
	die "epoll_create() returned $epoll: $!";
}

my $lfd = fileno($lfh);

$res = IO::Handle::blocking($lfh, 0);
if (! defined $res) {
	&log("Could not set non-blocking for '$lfh'");
	die "can not set non-blocking for '$lfh'\n";
}

$res = epoll_ctl($epoll, EPOLL_CTL_ADD, $lfd, $POLLIN);
if ($res < 0) {
	&log("Could not add listen port to epoll: $!");
	die "can't add listen port to epoll: $!";
}

my $lastreadzone = 0;

while (1) {
	$now = Time::HiRes::time;
	if (abs($now - $lastreadzone) > 600) {
		&readzone;
		&readmap;
		$lastreadzone = $now;
	}
	&checkevents;
}

sub checkevents {
	my($timeout) = @_;
	my($numevents, $event, $fd, $state);
	my($slot, $lastfile);
	my($res, $remaddr, $sock, $numbytes);
	my(@events);

	if (! defined $timeout) {
		$timeout = 5_000;
	}
	#print "running epoll_wait with timeout = '$timeout'\n";
	$numevents = epoll_wait($epoll, 10, $timeout, \@events);
	#print "numevents = '$numevents'\n";
	foreach $event (@events) {
		($fd, $state) = @$event;
		#printf "event... fd='%s' state='%s'\n", $fd, $state;
		if ($fd == $lfd) {
			&accept_client;
			next;
		}
		#print "this must be a client socket\n";
		$hostname = $hostname{$fd};
		$remaddr = $remaddr{$fd};
		$sock = $sock{$fd};
		if (! $sock) {
			&remove_client($fd, "sock is gone?");
			next;
		}
		if ($state & EPOLLIN) {
			$numbytes = sysread($sock, $buf{$fd}, 32768, $buflen{$fd});
			if (! defined $numbytes) {
				if ($! =~ /Resource temporarily unavailable/) {
					&log("sysread failed on $fd from $hostname: $!\n");
				} else {
					&remove_client($fd, "sysread failed on $fd from $hostname: $!");
				}
				next;
			}
			if ($numbytes == 0) {
				&process_buf($fd);
				&remove_client($fd, "client disconnected (read returned zero)");
				next;
			}
			$buflen{$fd} += $numbytes;
			if ($buflen{$fd} > 500_000) {
				&remove_client($fd, "buffer has grown too large ($buflen{$fd})");
				next;
			}
		}
		else {
			&remove_client($fd, "poll state other than read ($state)");
		}
	}
}

sub process_buf {
	#&log("process_buf_mq running");
	my($fd) = @_;
	my($interval, $dtype);
	my($dir, $path, $tmp);
	my %txt = ();
	my $hostname = $hostname{$fd};
	my $tstamp = Time::HiRes::time;
	my ($totalreq, $totalmissreq, $totaldur, $totalmissdur, $totalbytes, $totalmissbytes) = 0;
	my $msg;
	my $pop;
	if ($hostname =~ /\S+\.(\S+)\.ll/) {
		$pop = $1;
	}
	$tsdb = IO::Socket::INET->new(PeerAddr => "tsdb-proxy.lldns.net:4242");
	foreach (split(/[\r\n]+/, $buf{$fd})) {
		if (! /^rate (\d+) (rbmrb2) (.*)$/) {
			# line does not match
			#&log("Received $_ .. no match\n");
			next;
		}
		#&log("Processing bytes\n");
		$interval = $1;
		$dtype = $2;
		my $metrics = $3;
		$stats{$fd}++;
		$metrics =~ /(\d+) (\S+) (\S+) (\S+) (\S+)/;
		$acct = $1;
		$req = $2;
		$bytes = $3;
		$missreq = $4;
		$missbytes = $5;

		# don't bother writing if they're all 0s or something is missing
		#next if (! $req || ! $bytes || ! $missreq || ! $missbytes);
		#next if ($req == 0 && $bytes == 0 && $missreq == 0 && $missbytes == 0);

		# get our shortname from the map
		my $shortname = $acctmap{"$acct"};

		if ($debug > 1) {
			&log("Acct: $acct -> $req $bytes $missreq $missbytes\n");
		}

		$time = time();
		$msg = "put app.lhb.cust.requests $time $req fqdn=$hostname pop=$pop acct=$acct shortname=$shortname\n";
		$msg .= "put app.lhb.cust.miss_requests $time $missreq fqdn=$hostname pop=$pop acct=$acct shortname=$shortname\n";
		$msg .= "put app.lhb.cust.bytes $time $bytes fqdn=$hostname pop=$pop acct=$acct shortname=$shortname\n";
		$msg .= "put app.lhb.cust.miss_bytes $time $missbytes fqdn=$hostname pop=$pop acct=$acct shortname=$shortname\n";
		$tsdb->send($msg);
	}
	$tsdb->close();

};
	
sub accept_client {
	my($res, $newfh, $newfd);
	my($peer, $remiaddr, $remaddr, $remport, $hostname);
	$newfh = new IO::Handle;
	$peer = accept($newfh, $lfh);
	if (! $peer) {
		&log("accept() failed: $!\n");
		return;
	}
	$newfd = fileno($newfh);
	($remport, $remiaddr) = sockaddr_in($peer);
	$remaddr = inet_ntoa($remiaddr);

	$hostname = $addrhostname{$remaddr};
	if (! $hostname) {
		&remove_client($newfd, "unrecognised address");
		close $newfh;
		return;
	}
	$hostname{$newfd} = $hostname;

	&log("$remaddr:$remport ($hostname) connection received, fd $newfd\n") if ($debug == 1);
	$res = IO::Handle::blocking($newfh, 0);
	if (! defined $res) {
		&log("could not set non-blocking for: $!");
		die "can not set non-blocking for '$newfh'\n";
	}
	#print "non-blocking set for new client socket\n";
	$res = epoll_ctl($epoll, EPOLL_CTL_ADD, $newfd, $POLLIN);
	if ($res < 0) {
		&log("could not add client socket to epoll: $!");
		die "can't add client socket to epoll: $!";
	}
	#print "added $remaddr:$remport to epoll list\n";
	if ($addrsock{$remaddr}) {
		#&remove_client(fileno($addrsock{$remaddr}), "closing previous conenction");
	}
	$remaddr{$newfd} = $remaddr;
	$remport{$newfd} = $remport;
	$ctime{$newfd} = $now;
	$sock{$newfd} = $newfh;
	$active{$newfd} = $now;
	$stats{$newfd} = 0;
	$buf{$newfd} = '';
	$buflen{$newfd} = 0;
	$addrsock{$remaddr} = $newfh;
}

sub remove_client {
	my($fd, $reason) = @_;
	my($res, $remaddr, $remport, $stats, $duration);

	if (! $remport{$fd}) {
		&log("trying to remove client that is already removed $fd\n");
		return;
	}

	$remaddr = $remaddr{$fd};
	$remport = $remport{$fd};
	$hostname = $hostname{$fd};
	$duration = sprintf "%.2f", $now - $ctime{$fd};
	$stats = $stats{$fd};
	&log("$remaddr:$remport ($hostname) closed($stats stats, $duration sec): $reason\n");
	$res = epoll_ctl($epoll, EPOLL_CTL_DEL, $fd, 0);
	if ($res < 0) {
		warn "can't remove client $fd sock from epoll: $!";
	}
	else {
		#print "removed $remaddr{$fd}:$remport{$fd} from epoll list\n";
	}
	close $sock{$fd};
	delete $sock{$fd};
	delete $addrsock{$remaddr{$fd}};
	delete $remaddr{$fd};
	delete $remport{$fd};
	delete $active{$fd};
	delete $ctime{$fd};
	delete $stats{$fd};
	delete $buf{$fd};
}

sub log {
	my($line) = @_;
	$now = time;
	open(LOG, ">>$logfile") or warn "can't write $logfile: $!";
	print LOG $now, ' [', $$, '] ', $line;
	close LOG;
	my @st = stat($logfile);
	if ($st[7] > 50_000_000) {
		rename($logfile, "$logfile.old");
	}
}

sub readzone {
	my($file) = '/local/dns/pull/llnw.net';
	my($hostname, $type, $val);
	open(IN, $file) or die "can't open $file: $!";
	while (<IN>) {
		# cds1.ams.llnw.net.      3600    IN      A       87.248.197.11
		chop;
		next if ($_ eq '');
		($hostname, undef, undef, $type, $val) = split;
		next if ($hostname !~ /^(cds|pec)\d+\.([a-z0-9]+)\./);
		next if ($type ne 'A');
		$hostname =~ s/\.$//;
		$addrhostname{$val} = $hostname;
	}
	close IN;
}

sub readmap {
	my($file) = '/var/tmp/account.map';
	open(IN, $file) or die "can't open $file: $!";
	while(<IN>) {
		chop;
		if ($_ =~ /(\d+):(\S+)/) {
			$acctmap{"$1"} = $2;
		}
	}
	close IN;
}
