#!/usr/bin/env bash
set -x
source jenkinslib.sh

CREDENTIALS_JSON=$1
PILLAR=$2
declare -A CREDENTIALS_ARRAY
PILLAR_HASH_FILE="/var/lib/jenkins/${PILLAR}-pillar_hash.txt"
REQUEST_ERRORS=''

function parse_json_to_array(){
    ## populates $CREDENTIALS_ARRAY from json input
    local json_input=$1
    local out_format='"[\(.key)]=\"\(.value|tostring)\""' # [key1]="value1"   [key2]="value2"
    local jq_query="to_entries|map($out_format)|@tsv"
    eval "CREDENTIALS_ARRAY=( $(jq -r ${jq_query} <<<${json_input}) )"
}

function compare_pillar_hash(){
    local input_json=$1
    NEW_HASH=$(md5sum <<<${input_json} |awk '{print $1}' )
    if [[ -f "${PILLAR_HASH_FILE}" ]]
        then local old_hash=$(cat ${PILLAR_HASH_FILE})
        if [[ "${NEW_HASH}" != "${old_hash}" ]]
           then  echo "Hashes are not equal:
                [1] ${old_hash}
                [2] ${NEW_HASH}"
            else
                echo "changed=no comment='No jenkins credentials to create, pillar has not changed'"
                exit 0
        fi
    else
        echo "First highstate or hash file missing"
    fi
    echo "Proceeding with new pillar data."
}

function delete_credentials_domain() {
    local pillar=$1

    echo "Deleting ${pillar} credentials domain if it exists."
    url="http://127.0.0.1:8080/credentials/store/system/domain/${pillar}/config.xml"
    curl -s -X 'DELETE' "${url}"
     # returns '200' if exists or '404' if not.  We do not check first. So 404 is valid on first highstate.
}

function create_credentials_domain() {
    local pillar=$1
    echo "Creating new ${pillar} credentials domain."

    local url='http://127.0.0.1:8080/credentials/store/system/createDomain'
    response=$(curl -s -X 'POST' -w %{http_code} "${url}" \
        --data-urlencode "json={
         \"name\": \"${pillar}\",
         \"description\": \"Contains secrets loaded from pillar: secrets:${pillar}\"
        }" -o /dev/null) || REQUEST_ERRORS=True

        if [[ ${response} == "302" ]] # returns generic '302 Found' response to indicate valid request
            then echo "${pillar} Credentials Domain Created."
            else
                REQUEST_ERRORS=True
                echo "Request Unsuccessful! createDomain HTTP Response Code = ${response}"
        fi
}

function get_credential_type(){
    local credential=$1
    local credential_type=$(echo ${credential}|awk -F '_' '{print $(NF-1)}')  # split by _ and return second to last field
    echo ${credential_type}
}

function get_credential_username() {
    local name=$1
    local type=$2
    local username=''

    if [[ "${type}"=="passwd" || "${type}"=="sshkey" ]]
        then  username=$(echo ${name}|awk -F "_${type}" '{print $1}')  # split by/get everything before  _credential_type
    fi
    echo ${username}
}

function create_credential() {
    domain=$1
    name=$2
    body=$3

    url="http://127.0.0.1:8080/credentials/store/system/domain/${domain}/createCredentials"
    response=$(curl -s -X 'POST' -w %{http_code} "${url}" \
        --data-urlencode "${body}" -o /dev/null) || REQUEST_ERRORS=True

    if [[ ${response} == "302" ]] # returns generic '302 Found' response to indicate valid request
        then echo "${domain}/${name} credential created."
        else
            REQUEST_ERRORS=True
            echo "Request Unsuccessful! createCredentials HTTP Response Code = ${response}"
    fi
}

function gen_create_credential_body() {
    local name=$1
    local type=$2
    local username=$3
    local password=$4

    case ${type} in
    passwd)
        body="json={
				\"\": \"0\",
				\"credentials\": {
					\"scope\": \"GLOBAL\",
					\"id\": \"${name}\",
					\"username\": \"${username}\",
					\"password\": \"${password}\",
					\"description\": \"secrets:${PILLAR}:${name}\",
					\"\$class\": \"com.cloudbees.plugins.credentials.impl.UsernamePasswordCredentialsImpl\"
					}
				}"
        ;;
    sshkey)
        body="json={
				\"\": \"7\",
				\"credentials\": {
					\"scope\": \"GLOBAL\",
					\"id\": \"${name}\",
					\"description\": \"secrets:${PILLAR}:${name}\",
					\"username\": \"${username}\",
					\"privateKeySource\": {\"value\": \"0\",
					\"privateKey\": \"${password}\",
					\"stapler-class\": \"com.cloudbees.jenkins.plugins.sshcredentials.impl.BasicSSHUserPrivateKey\$DirectEntryPrivateKeySource\",
					\"\$class\": \"com.cloudbees.jenkins.plugins.sshcredentials.impl.BasicSSHUserPrivateKey\$DirectEntryPrivateKeySource\"},
					\"passphrase\": \"\",
					\"\$redact\": \"passphrase\",
					\"stapler-class\": \"com.cloudbees.jenkins.plugins.sshcredentials.impl.BasicSSHUserPrivateKey\",
					\"\$class\": \"com.cloudbees.jenkins.plugins.sshcredentials.impl.BasicSSHUserPrivateKey\"
					}
				}"
        ;;
    secret)
        body="json={
				\"\": \"9\",
				\"credentials\": {
					\"scope\": \"GLOBAL\",
					\"id\": \"${name}\",
					\"secret\": \"${password}\",
					\"\$redact\": \"${password}\",
					\"description\": \"secrets:${PILLAR}:${name}\",
					\"stapler-class\": \"org.jenkinsci.plugins.plaincredentials.impl.StringCredentialsImpl\",
					\"\$class\": \"org.jenkinsci.plugins.plaincredentials.impl.StringCredentialsImpl\"
					}
				}"
        ;;
    esac

    echo ${body}
}

function loop_create_credentials_from_global_array() {
    for credential in "${!CREDENTIALS_ARRAY[@]}"
        do echo "creating credential for ${credential}."

        ## parse data
        type=$(get_credential_type "${credential}")
        username=$(get_credential_username "${credential}" "${type}")
        password=${CREDENTIALS_ARRAY[${credential}]}
        body=$(gen_create_credential_body "${credential}" "${type}" "${username}" "${password}")

        ## create credential
        create_credential "${PILLAR}" "${credential}" "${body}"
    done
}

function check_for_errors() {
    if [[ ${REQUEST_ERRORS} ]]
        then echo "State Unsuccessful! See errors in logs."
        exit 1
    fi
}

function update_pillar_hash() {
    # uses globals from previously run compare_pillar_hash()
    echo -n ${NEW_HASH} > ${PILLAR_HASH_FILE}
}

## BEGIN SCRIPT
compare_pillar_hash "${CREDENTIALS_JSON}" # will exit 0 if no changes since last highstate
parse_json_to_array "${CREDENTIALS_JSON}" # populates global CREDENTIALS_ARRAY

disable_security
# disable_security now turns off nginx to prevent unauthorized access. It also restarts jenkins and does wait_for_jenkins

delete_credentials_domain "${PILLAR}" # delete (if exists) 1st since createCredentials does not update secrets
create_credentials_domain "${PILLAR}" # (re-)create domain to load fresh credentials into
loop_create_credentials_from_global_array # (re-)create each jenkins credential in the (re-)created domain

enable_security
# enable_security now has restart and wait_for_jenkins builtin and it starts nginx (stopped to prevent unauthorized access)

check_for_errors # will exit 1 if REQUEST_ERRORS
update_pillar_hash # success! update hash for comparison next highstate

echo
echo "changed=yes comment='Created ${PILLAR} credentials'"