#!/usr/bin/env bash

export EXIT_CODE=0

write_log()
{
    local log_dir="/var/log/opsramp"
    local log_file="${log_dir}/configure-opsramp-agent.log"
    local now=$(date +'%Y%m%dT%H%M%S')
    if [ ! -d "${log_dir}" ]; then
        mkdir -p "${log_dir}"
    fi
    touch "${log_file}"
    chmod 666 "${log_file}"
    echo "$now $$ $@" >> "${log_file}"
}

test_gateway()
{
    local gateway="$1"
    local port="$2"
    nc -zvw 2 "$gateway" "$port" &>/dev/null
    result="$?"
    if [ "$result" == "0" ];
    then
        echo "success"
    else
        echo "fail"
    fi
}

report_result()
{
    local quiet="$1"
    local changed="$2"
    local comment="$3"

    write_log "changed=${changed} comment='${comment}'"
    if [ "$quiet" != "quiet" ]; then
        echo "changed=${changed} comment='${comment}'"
    fi
}

configure_agent()
{
    local current_gateway_file=""
    local gateway=""
    local agent_key=""
    local agent_secret=""
    local opsramp_api_endpoint=""
    local port=""
    local results=""
    local quiet="false"

    while (( "$#" )); do
        case "${1}" in
            -d|--direct)
                # blank gateway means "connect directly to opsramp_api_endpoint"
                gateway=""
                shift
                ;;
            -q|--quiet)
                shift
                quiet="${1}"
                shift
                ;;
            -g|--gateway)
                shift
                gateway="${1}"
                shift
                ;;
            -f|--file)
                shift
                current_gateway_file="${1}"
                shift
                ;;
            -K|--key)
                shift
                agent_key="${1}"
                shift
                ;;
            -S|--secret)
                shift
                agent_secret="${1}"
                shift
                ;;
            -e|--endpoint)
                shift
                opsramp_api_endpoint="${1}"
                shift
                ;;
            -p|--port)
                shift
                port="${1}"
                shift
                ;;
        esac
    done

    results=$(service opsramp-agent stop)
    write_log "${results}"
    results=$(service opsramp-shield stop)
    write_log "${results}"

    if [ "${gateway}" == "" ]; then
        write_log /opt/opsramp/agent/bin/configure -K "${agent_key}" -S "${agent_secret}" -s "${opsramp_api_endpoint}"
        results=$(/opt/opsramp/agent/bin/configure -K "${agent_key}" -S "${agent_secret}" -s "${opsramp_api_endpoint}")
        write_log "${results}"
    else
        write_log /opt/opsramp/agent/bin/configure -K "${agent_key}" -S "${agent_secret}" -s "${opsramp_api_endpoint}" -m proxy -x "${gateway}" -p ${port}
        results=$(/opt/opsramp/agent/bin/configure -K "${agent_key}" -S "${agent_secret}" -s "${opsramp_api_endpoint}" -m proxy -x "${gateway}" -p ${port})
        write_log "${results}"
    fi

    if [[ "${results}" =~ "Fail" ]]; then
        write_log "Failure detected -- deleted ${current_gateway_file} to force retry on next run"
        rm -f "${current_gateway_file}"
        report_result "${quiet}" "false" "Agent failure detected"
        EXIT_CODE=1
    else
        echo -n "${gateway}" > "${current_gateway_file}"
        chmod 666 "${current_gateway_file}"
        write_log "set current gateway=${gateway}"

        if [ "${current_gateway}" == "" ]; then
            report_result "${quiet}" "true" "Connected to ${opsramp_api_endpoint}"
        else
            report_result "${quiet}" "true" "Connected to ${gateway}"
        fi
    fi
}

main()
{
    local quiet="$1"
    local preferred_gateway="{{preferred_gateway}}"
    local port="{{port}}"
    local agent_key="{{agent_key}}"
    local agent_secret="{{agent_secret}}"
    local opsramp_api_endpoint="{{opsramp_api_endpoint}}"
    local go_direct="{{go_direct|lower}}"
    local gateway=""
    local gateway_count=0
    local gateway_list=()
    local current_gateway=""
    local current_gateway_file="/tmp/current_opsramp_gateway"
    local current_gateway_file_exists="false"
    local i=0

    if [ -f "${current_gateway_file}" ]; then
        current_gateway=$(cat "${current_gateway_file}")
        current_gateway_file_exists="true"
    fi

    if [ "${go_direct}" != "true" ]; then

        # use "end" to end it because the list of gateways may be empty
        for gateway in {{gateways|join(" ")}} "end"
        do
            if [ "${gateway}" == "end" ]; then
                break
            fi
            result=$(test_gateway "${gateway}" "${port}")

            if [ "${result}" == "success" ]; then
                if [ "${preferred_gateway}" == "None" ] && [ "${gateway}" == "${current_gateway}" ]; then
                    # no preferred gateway and the current one is still working
                    report_result "${quiet}" "false" "Still using ${current_gateway}"
                    exit $EXIT_CODE
                elif [ "${preferred_gateway}" == "${current_gateway}" ] && [ "${gateway}" == "${current_gateway}" ]; then
                    # the current gateway is already the preferred gateway and it's still working
                    report_result "${quiet}" "false" "Still using ${current_gateway}"
                    exit $EXIT_CODE
                elif [ "${gateway}" == "${preferred_gateway}" ]; then
                    configure_agent --quiet "${quiet}" --file "${current_gateway_file}" --gateway "${gateway}" --key "${agent_key}" --secret "${agent_secret}" --endpoint "${opsramp_api_endpoint}" --port "${port}"
                    exit $EXIT_CODE
                else
                    gateway_list+=("${gateway}")
                fi
            fi
        done
    fi

    gateway_count=${{'{#'}}gateway_list[@]}
    if [ $gateway_count -eq 0 ]; then
        go_direct="true"
    fi

    if [ "${go_direct}" == "true" ]; then
        # connect directly to API 

        if [ "${current_gateway}" == "" ] && [ "${current_gateway_file_exists}" == "true" ]; then
            # if there is no current gateway and this script already ran once (that's when the
            # current gateway file was generated), the agent is already configured

            report_result "${quiet}" "false" "Already connected to ${opsramp_api_endpoint}"
        else
            configure_agent --quiet "${quiet}" --file "${current_gateway_file}" --direct --key "${agent_key}" --secret "${agent_secret}" --endpoint "${opsramp_api_endpoint}"
        fi
    else
        # use a random, working gateway as a proxy
        i=$(( $RANDOM % $gateway_count ))
        gateway=${gateway_list[$i]}
        configure_agent --quiet "${quiet}" --file "${current_gateway_file}" --gateway "${gateway}" --key "${agent_key}" --secret "${agent_secret}" --endpoint "${opsramp_api_endpoint}" --port "${port}"
    fi
    exit $EXIT_CODE
}
main "$@"
