#!/usr/bin/python
# -*- coding: utf-8 -*-
'''
syslog-proxy.py

syslog data_fields:
['<134>Sep', '24', '21:25:29', 'cds317.dfw.llnw.net', 'logpusher[74883]:', 
'finished', 'sending', 'ssl_sw_access-3.1632535200.3of64.https.70680.log.bz2', 
'(https,', 'size=4283343', 'checksum=d287099d627a23462ecaf9cde4f2d615affd126a', 
'modified', '1632542399)']
'''

__author__ = 'christian kuzmanic'
__copyright__ = 'copyright © 2021, Limelight Networks'
__version__ = '1.1'
__maintainer__ = 'christian kuzmanic'
__email__ = 'ckuzmanic@llnw.com'
__status__ = 'Production'

import multiprocessing as mp
import signal
import time

from collections import OrderedDict
from socket import *
from scapy.all import IP, UDP, send

# { managed via Salt - do not edit
# config vars
sport = 42114           # local source port
lport = 5514            # local listen port
dhost = '10.252.2.16'   # gw ip
dport = 514             # gw destination port
tz_offset = '-07:00'    # MST
dns_cache_ttl = 3600    # seconds
dns_cache_size = 3000   # num. cache entries
# } end: managed via Salt

sock = socket(AF_INET, SOCK_DGRAM)
sock.bind(('localhost', lport))
dns_cache = OrderedDict()


def pool_initializer():
    ''' ignore CTRL+C in the worker process'''
    signal.signal(signal.SIGINT, signal.SIG_IGN)


def dns_cache_lookup(hostname):
    if hostname not in dns_cache.keys():
        src_ip = src_ip_lookup(hostname)
        if not src_ip:
            return
        dns_cache_update(hostname, src_ip)
    else:
        ttl_expired = timedelta(
            int(round(time.time())),
            dns_cache[hostname]['timestamp']
        )
        if ttl_expired:
            src_ip = src_ip_lookup(hostname)
            if src_ip:
                dns_cache_update(hostname, src_ip)
            else:
                del dns_cache[hostname]
        else:
            src_ip = dns_cache[hostname]['src_ip']
    return src_ip


def dns_cache_update(hostname, src_ip):
    dns_cache_prune()
    timestamp = int(round(time.time()))
    dns_cache[hostname] = {
        'src_ip': src_ip,
        'timestamp': timestamp
    }


def dns_cache_prune():
    if len(dns_cache.keys()) > dns_cache_size:
        del dns_cache[next(iter(dns_cache))]


def src_ip_lookup(hostname):
    src_ip = False
    try:
        src_ip = gethostbyname("%s" % (hostname))
    except Exception as e:
        pass
    return src_ip


def timedelta(current_timestamp, last_timestamp):
    if (current_timestamp - last_timestamp) > dns_cache_ttl:
        return True
    else:
        return False


def send_messages(data):
    data_fields = data.rstrip().split()
    data_fields[2] = data_fields[2] + tz_offset
    data = ' '.join([str(x) for x in data_fields])
    try:
        src_ip = dns_cache_lookup(data_fields[3])
        if src_ip:
            send(IP(src=src_ip, dst=dhost) /
                 UDP(sport=sport, dport=dport)/data)
    except Exception as e:
        # print(data)
        # raise
        pass


def process_messages():
    pool = mp.Pool(24, initializer=pool_initializer)
    try:
        while True:
            data, clientaddr = sock.recvfrom(4096)
            pool.map_async(send_messages, [data])
    except KeyboardInterrupt:
        pool.terminate()
        pool.join()


if __name__ == "__main__":
    process_messages()
