#!jinja|pyobjects

import yaml
import logging
import re
from collections import defaultdict

{% from "sae-prometheus/map.jinja" import prometheus_config %}

log = logging.getLogger(__name__)

FQDN = '{{ grains['fqdn'] }}'
PROMETHEUS_CONFIG = {{prometheus_config}}
SITES = []
BUILDINGS = {}
FILTERS = defaultdict(lambda: defaultdict(list))
FILTER_TYPES = ['tag', 'server', 'region', 'pop', 'longitude']
OFFNET_TYPES = ['smartpop', 'off-net']


def represent_dict(self, data):
    """Helper method to order keys in dumped yaml."""
    def key_function((key,value)):
        """Prioritize key names"""
        priority = {"targets":0, "labels":1}.get(key, 99)
        return (priority, key)
    items = data.items()
    items.sort(key=key_function)
    return self.represent_mapping(u'tag:yaml.org,2002:map', items)


def underscore(input):
    replace_chars = ['.', ' ', '+', '-']
    for character in replace_chars:
        if character in input:
            input = input.replace(character, '_')
    return input.lower()


def set_cs_sites():
    global SITES
    global BUILDINGS

    cs_result = __salt__['publish.runner'](
        'cs.cslist', [ 'building_group', '{"id": ".*"}'])
    log.debug('CSRESULT SITES: {}'.format(cs_result))
    for site in cs_result:
        SITES.append(re.sub("\d", '', site['name']).lower())

    cs_result = __salt__['publish.runner'](
        'cs.cslist', [ 'building', '{"id": ".*"}'])
    log.debug('CSRESULT BUILDINGS: {}'.format(cs_result))
    for building in cs_result:
        BUILDINGS[re.sub("\d", '', building['group_name']).lower()] = building['longitude']
    log.debug("BUILDINGS: {}".format(BUILDINGS))


def get_cs_data():
    enabled_tags = []
    try:
        for job in PROMETHEUS_CONFIG['jobs']:
            enabled_tags.extend(PROMETHEUS_CONFIG['jobs'][job]['tags'].keys())
    except (AttributeError, KeyError), e:
        pass

    cs_result = __salt__['publish.runner'](
        'cs.device_tag_search',
        [
            enabled_tags,
            'include=["tags","parents"]',
            'match_any=True'
        ])

    log.debug("CSRESULT: {}".format(cs_result))
    return cs_result


def filter_host(host_data, host, pop, tag_list):
    """Whitelist hosts based on filters configured in map.yaml"""

    region = get_region(host_data)
    log.debug("DEBUG region: {}".format(region))
    building_type = get_building_type(host_data)
    log.debug("DEBUG building_type: {}".format(building_type))

    if host in FILTERS['server']:
        log.debug("FILTER: Host {} found in server filter".format(host))
        return True
    else:
        log.debug("FILTER: {} not in server filter {}".format(host, FILTERS['server']))

    if FILTERS['tag']:
        log.debug("FILTER: Tag filter configured")
        if [i for i in tag_list if i in FILTERS['tag']]:
            log.debug("FILTER: Tag {} present".format(i))
            if FILTERS['pop']:
                log.debug("FILTER: Pop filter configured (with tag filter)")
                if pop in FILTERS['pop']:
                    log.debug("FILTER: Pop {} present".format(pop))
                    return True
                else:
                    log.debug("FILTER: Pop {} not present".format(pop))
                    return False

            if FILTERS['region']:
                log.debug("FILTER: Region filter configured (with tag filter)")
                if region in FILTERS['region']:
                    log.debug("FILTER: Region {} present".format(region))
                    return True
                else:
                    log.debug("FILTER: Region {} not present".format(region))
                    return False

            log.debug("FILTER: Filtered tag found in {}".format(tag_list))
            return True

    if FILTERS['pop']:
        log.debug("FILTER: Pop filter configured")
        if pop in FILTERS['pop']:
            log.debug("FILTER: Pop {} present".format(pop))
            return True

    if FILTERS['region']:
        log.debug("FILTER: Region filter configured")
        if building_type in OFFNET_TYPES:
            log.debug("FILTER: Offnet host.  Skipping")
            return False
        if region in FILTERS['region']:
            log.debug("FILTER: Region {} present".format(region))
            if FILTERS['longitude']:
                log.debug('FILTER: Longitude present')
                if pop in BUILDINGS:
                    log.debug("FILTER: Longitude check {} <= {} <= {}".format(
                        FILTERS['longitude'][0],
                        BUILDINGS[pop],
                        FILTERS['longitude'][1]))
                    if int(FILTERS['longitude'][0]) <= int(float(BUILDINGS[pop])) <= int(FILTERS['longitude'][1]):
                        return True
                    else:
                        return False
                else:
                    log.debug("FILTER: {} not found in BUILDINGS, including as fallback".format(pop))
                    return True

            return True

    if FILTERS['offnet']:
        log.debug("FILTER: Offnet filter configured")
        if building_type in OFFNET_TYPES:
            log.debug("FILTER: Offnet type {} matched".format(building_type))
            return True

    log.debug("FILTER: {} not found any any whitelist filter".format(host))
    return False


def get_building_type(host_data):
    """Retrieve the building type for the host.
    Used to identify offnet hosts."""
    if host_data['parent_id']:
        return host_data['parent']['building_type_name'].lower()
    else:
        return host_data['building_type_name'].lower()


def get_region(host_data):
    """Retrieve the host's region.  Useful for guest os, since they typically
    don't have the correct building_region_name for hosts outside NA"""
    if host_data['parent_id']:
        return host_data['parent']['building_region_name']
    else:
        return host_data['building_region_name']


def get_pop(host, cs_data):
    m = re.match(".+\.(.+)\.ll.+\.(!?com|net)", host)
    if m:
        pop = m.group(1)
    else:
        pop = 'none'
    if pop not in SITES:
        if cs_data['building_name'].lower() == "virtual":
            if cs_data['parent_id']:
                cs_data = salt.publish.runner('cs.get_cs_id', cs_data['parent_id'])

        if cs_data['building_name_short']:
            pop = cs_data['building_name_short']
        else:
            pop = 'none'
        pop = re.sub("\d", '', pop)
    return pop.lower()


def get_targets():
    data = dict()
    targets = defaultdict(list)
    for job in PROMETHEUS_CONFIG['jobs']:
        for tag in PROMETHEUS_CONFIG['jobs'][job]['tags']:
            data.update(
                {
                    tag : {
                        "ports": PROMETHEUS_CONFIG['jobs'][job]['tags'][tag],
                        "job": job,
                        "type": PROMETHEUS_CONFIG['jobs'][job]['type']
                    }
                })
    log.debug("TARGET_DATA: {}".format(data))
    cs_result = get_cs_data()

    if not cs_result:
        #exiting should prevent overwriting a 'good' file if CS is down
        exit

    for host in cs_result:
        log.debug("{}:{}".format(host, cs_result[host]))
        tag_list = map(lambda tags: tags['name'], cs_result[host]['tags'])
        pop = get_pop(host, cs_result[host])
        if pop is 'none':

            if not pop:
                pop = 'none'

        whitelisted = filter_host(cs_result[host], host, pop, tag_list)

        if not whitelisted:
            continue

        """Allow only dev|qa|stage|lab resources in devnet prometheus.
        Compare the list of strings against the FQDN of the prometheus server
        and the list of tags supplied from the CS query results.  Allow on
        prometheus dev hosts only if the hosts contain the strings in their
        hostname or in their CS tags.

        If prometheus host is non-production and a there is no non-prod CS tag
        and there is no non-prod hostname, then continue."""

        env_list = ['dev', 'qa', 'stage', 'staging', 'lab', 'laaa']
        if 'dev' in FQDN and \
            not any(env in env_list for env in tag_list) and \
            not any(env in host for env in env_list):
            log.debug('Skipping prod host')
            continue

        """Allow only production tagged or hostname matches into prod
        prometheus, using the same env_list to determine non-prod resources.

        If prometheus host is production and a non-prod tag in tag_list or
        non-prod string in hostname except stg, then continue."""

        if 'stage' in FQDN and \
            any(env in 'staging' for env in tag_list):
            log.debug('Allowing stg host')
            pass
        elif 'dev' not in FQDN and \
            (any(env in env_list for env in tag_list) or \
            any(env in host for env in env_list)):
            log.debug('Skipping non-prod host')
            continue

        found_tags = [i for i in tag_list if i in data]
        for tag in found_tags:
            if tag in data:
                if 'null' not in data[tag]['ports']:
                    formatted_target = ["%s:%s" % (host, port) for port in data[tag]['ports']]
                else:
                    formatted_target = ["%s" % host] 
                region = get_region(cs_result[host]) or 'none'

                targets[data[tag]['type']].append(
                    {
                        "labels": {
                            "job": data[tag]['job'],
                            "region": underscore(region)
                        },
                        "targets": formatted_target
                    })
    return targets

# BEGIN
if FQDN in PROMETHEUS_CONFIG['prometheus_servers']:
    """Create list of filters from the yaml config"""
    for index in range(len(PROMETHEUS_CONFIG['prometheus_servers'][FQDN]['cs_filters'])):
        for type in PROMETHEUS_CONFIG['prometheus_servers'][FQDN]['cs_filters'][index]:
            FILTERS[type] = PROMETHEUS_CONFIG['prometheus_servers'][FQDN]['cs_filters'][index][type]
    log.debug("FILTERS: {}".format(FILTERS))

    set_cs_sites()
    log.debug("SITES: {}".format(SITES))

    targets = get_targets()
    log.debug("TARGETS: {}".format(targets))

    yaml.add_representer(dict, represent_dict)
    yaml.Dumper.ignore_aliases = lambda *args : True

    if len(targets['common']) > 0:
        yaml_targets_common = yaml.dump(targets['common'])
        File.accumulated(
            "sae|promometheus|targets-common-accumulator",
            filename="/etc/prometheus/targets_common.yml",
            text=yaml_targets_common,
            require_in=File("sae|prometheus|config"))

    if len(targets['common-https']) > 0:
        yaml_targets_common_https = yaml.dump(targets['common-https'])
        File.accumulated(
            "sae|promometheus|targets-common-https-accumulator",
            filename="/etc/prometheus/targets_common_https.yml",
            text=yaml_targets_common_https,
            require_in=File("sae|prometheus|config"))

    if len(targets['netdata']) > 0:
        yaml_targets_netdata = yaml.dump(targets['netdata'])
        File.accumulated(
            "sae|promometheus|targets-netdata-accumulator",
            filename="/etc/prometheus/targets_netdata.yml",
            text=yaml_targets_netdata,
            require_in=File("sae|prometheus|config"))

    if len(targets['rtsv2']) > 0:
        yaml_targets_rtsv2 = yaml.dump(targets['rtsv2'])
        File.accumulated(
            "sae|promometheus|targets-rtsv2-accumulator",
            filename="/etc/prometheus/targets_rtsv2.yml",
            text=yaml_targets_rtsv2,
            require_in=File("sae|prometheus|config"))

    if len(targets['ltf_exporter']) > 0:
        yaml_targets_ltf_exporter = yaml.dump(targets['ltf_exporter'])
        File.accumulated(
            "sae|promometheus|targets-ltf-accumulator",
            filename="/etc/prometheus/targets_ltf.yml",
            text=yaml_targets_ltf_exporter,
            require_in=File("sae|prometheus|config"))

    if len(targets['snmp_ifmib']) > 0:
        yaml_targets_snmp_ifmib = yaml.dump(targets['snmp_ifmib'])
        File.accumulated(
            "sae|promometheus|targets-snmp-ifmib-accumulator",
            filename="/etc/prometheus/targets_snmp_ifmib.yml",
            text=yaml_targets_snmp_ifmib,
            require_in=File("sae|prometheus|config"))

    if len(targets['snmp_arista']) > 0:
        yaml_targets_snmp_arista = yaml.dump(targets['snmp_arista'])
        File.accumulated(
            "sae|promometheus|targets-snmp-arista-accumulator",
            filename="/etc/prometheus/targets_snmp_arista.yml",
            text=yaml_targets_snmp_arista,
            require_in=File("sae|prometheus|config"))

    if len(targets['snmp_brocade']) > 0:
        yaml_targets_snmp_brocade = yaml.dump(targets['snmp_brocade'])
        File.accumulated(
            "sae|promometheus|targets-snmp-brocade-accumulator",
            filename="/etc/prometheus/targets_snmp_brocade.yml",
            text=yaml_targets_snmp_brocade,
            require_in=File("sae|prometheus|config"))
