#!/usr/bin/env python

# This script runs on Linux server epdev-sirperfulous as a daemon,
# receives the data sent by sirperfulous from CDS systems, and
# generates flame graphs
#
# The files and data are under /sirPerfulous on the server:
#   /sirPerfulous
#       |
#       |-- bin
#       |    |
#       |    |-- sirperfulous-process.py, flamegraph.pl
#       |-- data
#       |    |
#       |    |-- <client>
#       |    |     |
#       |    |     |-- <user>
#       |    |           |
#       |    |           |-- <timestamp>
#       |    |
#       |    |-- archived
#       |    |     |-- <client>
#       |    |            |
#       |    |            |-- <user>
#       |    |                  |
#       |    |                  |-- <timestamp>
#       |-- log
#       |    |
#       |    |-- sirperfulous.log
#
# This script is under /sirPerfulous/bin
# The data collected and processed are under:
#     /sirPerfulous/data/<client>/<timestamp>
# The log folder holds the logs of this script
#
# The archiver looks thorough /sirPerfulous/data/<client> folders
# creation/modification time, and compress the folder into gzip file
# if it is more than 7 days old.
#

import sys
import os
import subprocess
import argparse
import textwrap
import socket
import multiprocessing
import time
import stat
import logging
from datetime import datetime


# Information "constants"
bindir = ''
sudo = 'sudo'
usr = 0
host = ''
port = 9876
indir = ''
outdir = '/sirPerfulous'
logfile = '/sirPerfulous/log/sirperfulous.log'
filepath = ''
clienttype = 'Linux'
debug = False


def log(s):
    logger = logging.getLogger('sirperfulous')
    logger.info(s)
    if debug:
        print s


def execute_shell_cmd(cmd):
    log("Running cmd: %s" % cmd)
    process = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True)
    return process.communicate()[0].strip()


def check_create_path(path):
    if not os.path.exists(path):
        try:
            os.mkdir(path)
        except OSError:
            print("Creating directory %s failed." % path)
            sys.exit()
        else:
            log("Directory %s is created." % path)


def checkenv():
    global usr, sudo, outdir, bindir, logfile, logger

    host_type = subprocess.check_output(['uname']).rstrip()
    if host_type != "Linux":
        print "This script only runs on Linux"
        sys.exit()

    bindir = "%s/bin" % outdir
    bindir = execute_shell_cmd("cd %s && pwd" % bindir)

    usr = subprocess.check_output(['id', '-u']).rstrip()
    if usr == 0:
        sudo = ''

    check_create_path(outdir)

    logger = logging.getLogger('sirperfulous')
    hdlr = logging.FileHandler(logfile)
    formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
    hdlr.setFormatter(formatter)
    logger.addHandler(hdlr)
    logger.setLevel(logging.INFO)


def check_perf():
    if (execute_shell_cmd("dpkg --list linux-tools-common | \
            grep linux-tools-common | grep ii") != ""):
        s = "linux-tools-`uname -r`"
        if (execute_shell_cmd("dpkg --list %s | grep %s | \
                grep ii" % (s, s)) != ""):
            log("perf is installed on the system")
        else:
            print "%s is not installed" % s
            print "Install %s by running \"sudo apt-get install \
                linux-tools-`uname -r`\"" % s
            sys.exit()
    else:
        print "linux-tools-common is not installed"
        print "Install linux-tools-common by running \"sudo apt install \
                linux-tools-common\""
        sys.exit()


def readArgs():
    global indir

    parser = argparse.ArgumentParser(
                description="Receives and process profiling and statistics "
                "data from CDS systems",
                formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('-p',
                        help=textwrap.dedent('''\
                                generate flame graph with existing profiling
                                data in this folder'''),
                        required=False, action='store', dest='indir')

    args = parser.parse_args()
    indir = args.indir if args.indir else indir


def uncompress(filepath, clienttype):
    zipfile = ("%s/perf_stats_data_%s.tar.gz" % (filepath, clienttype))
    log("Received zip file %s ...\n" % zipfile)

    if os.path.exists(zipfile):
        execute_shell_cmd("tar -zxf %s -C %s" % (zipfile, filepath))
        if os.path.exists("%s/perf.tar.gz" % filepath):
            execute_shell_cmd("tar -zxf %s/perf.tar.gz -C %s \
                    " % (filepath, filepath))
        if os.path.exists("%s/stats.tar.gz" % filepath):
            execute_shell_cmd("tar -zxf %s/stats.tar.gz -C %s \
                    " % (filepath, filepath))


def flame(checking, filepath, clienttype, pid):
    if (not checking) or os.path.exists("%s/perf.tar.gz" % filepath):
        if clienttype == 'Linux':
            # CPU
            if pid == '':
                execute_shell_cmd(("cat %s/perf_cpu.folded | %s/flamegraph.pl \
                    --title=\"CPU Flame Graph\" > %s/perf-`whoami`-cpu-all.svg\
                    ") % (filepath, bindir, filepath))

                execute_shell_cmd(("grep -v cpu_idle %s/perf_cpu.folded | %s/flamegraph.pl \
                    --title=\"Non-idle CPU Flame Graph\" > \
                    %s/perf-`whoami`-cpu-nonidle.svg") % (filepath,
                                  bindir, filepath))
                execute_shell_cmd(("grep -i squid %s/perf_cpu.folded | %s/flamegraph.pl \
                    --title=\"Squid CPU Flame Graph\" > \
                    %s/perf-`whoami`-cpu-squid.svg") % (filepath,
                                  bindir, filepath))
            else:
                execute_shell_cmd(("cat %s/perf_cpu.folded | %s/flamegraph.pl \
                    --title=\"CPU Flame Graph\" > %s/perf-`whoami`-cpu-%s.svg\
                    ") % (filepath, bindir, filepath, pid))
        else:
            # CPU
            if pid == '':
                execute_shell_cmd(("cat %s/perf_cpu_*.folded | %s/flamegraph.pl \
                    --title=\"CPU Flame Graph\" > %s/perf-`whoami`-cpu-all.svg\
                    ") % (filepath, bindir, filepath))
                execute_shell_cmd(("egrep -v '^idle;-' %s/perf_cpu_*.folded | \
                    %s/flamegraph.pl --title=\"Non-idle CPU Flame Graph\" > \
                    %s/perf-`whoami`-cpu-nonidle.svg") % (filepath,
                                  bindir, filepath))
                execute_shell_cmd(("grep -i squid %s/perf_cpu_*.folded | %s/flamegraph.pl \
                    --title=\"Squid CPU Flame Graph\" > \
                    %s/perf-`whoami`-cpu-squid.svg") % (filepath,
                                  bindir, filepath))
            else:
                execute_shell_cmd(("cat %s/perf_cpu_pmcstat.folded | %s/flamegraph.pl \
                    --title=\"CPU Flame Graph\" > %s/perf-`whoami`-cpu-%s.svg\
                    ") % (filepath, bindir, filepath, pid))


def process_data(filepath, clienttype, pid):
    log("Start flame graph generating in folder %s ..." % filepath)
    uncompress(filepath, clienttype)
    flame(True, filepath, clienttype, pid)
    if os.path.exists("%s/perf_stats_data_%s.tar.gz" % (filepath, clienttype)):
        execute_shell_cmd("rm -rf %s/perf_stats_data_%s.tar.gz %s/perf.tar.gz \
                %s/stats.tar.gz" % (filepath, clienttype, filepath, filepath))
    log("Done flame graph generating in folder %s ..." % filepath)
    return


def recv_process_data():
    global filepath, clienttype, debug

    if indir != '':
        debug = True
        filepath = indir
        uncompress(filepath, clienttype)
        flame(False, filepath, clienttype, '')
        return

    sock = socket.socket()
    sock.bind((host, port))
    sock.listen(5)
    print 'Server is listening ...'

    while True:
        conn, addr = sock.accept()
        data = conn.recv(4096)
        if data:
            client, timestamp, filename, filesize, pid, user = data.split(" ")
            filesize = int(filesize)
            filename = os.path.basename(filename)
            clienttype = 'Linux'
            if 'FreeBSD' in filename:
                clienttype = 'FreeBSD'

            filepath = "%s/data/%s" % (outdir, client)
            check_create_path(filepath)
            filepath = "%s/%s" % (filepath, user)
            check_create_path(filepath)
            filepath = "%s/%s" % (filepath, timestamp)
            check_create_path(filepath)
            conn.send("ack")
            log("Receiving data from client %s" % client)

            with open(("%s/%s" % (filepath, filename)), 'wb') as f:
                while True:
                    bytes_read = conn.recv(4096)
                    if not bytes_read:
                        break
                    f.write(bytes_read)
                f.close()
                conn.close()

                multiprocessing.Process(target=process_data,
                                        args=(filepath, clienttype, pid)).start()


def archive_data():
    archive_period = 7 * 24 * 60 * 60  # archive weekly
    print "Archiver is running ..."

    # data are being archived if the folder timestamp has not changed in one week.
    # data will be archived to "data/archived/<client>/<user>/<timestamp>"
    # from "data/<client>/<user>/</timestamp>"
    while True:
        for d in os.listdir("%s/data" % outdir):
            clientd = "%s/data/%s" % (outdir, d)
            if clientd == ("%s/data/archived" % outdir):
                continue
            else:
                check_create_path("%s/data/archived/%s" % (outdir, d))
                for d2 in os.listdir(clientd):
                    check_create_path("%s/data/archived/%s/%s" % (outdir, d, d2))
                    userd = "%s/%s" % (clientd, d2)
                    for d3 in os.listdir(userd):
                        timed = "%s/%s" % (userd, d3)

                        c = os.path.getctime(timed)
                        ct = time.strftime('%Y-%m-%d_%H_%M_%S', time.localtime(c))

                        time_elapsed = time.time() - c
                        if (time_elapsed > archive_period):
                            log("Archiving directory %s" % timed)
                            execute_shell_cmd("cd %s; tar -zcf data/archived/%s/%s/%s.tar.gz \
                                              --absolute-names %s --remove-files" %
                                              (outdir, d, d2, ct, timed))

        # wake up daily and check/archive
        time.sleep(24 * 60 * 60)


def main():
    readArgs()
    checkenv()
    check_perf()
    print "\nResults at: %s\n" % outdir

    if indir == '':
        multiprocessing.Process(target=archive_data).start()
    recv_process_data()


if __name__ == '__main__':
    main()
