#!/usr/bin/env python

# This script runs on live FreeBSD/Linux CDS systems.
# It collects profiling and stat data and sends to epdev-sirperfulous
# host for processing.
# By default, the collected data are located at /tmp/sirPerfulous

import sys
import os
import subprocess
import argparse
import textwrap
import socket
from datetime import datetime

# Configurable values
sampletime = 120
server = 'epdev-sirperfulous.devnet.llnw.net'
port = 9876
datatype = 'profile'
senddataonly = False
logging = False
pid = ''
usedtrace = False
flame_locally = False
hz = 99

# Non-configurable Values
host_type = 'Linux'
sudo = 'sudo'
usr = 0
outdir = '/tmp/sirPerfulous'
argpid = ''


def log(s):
    if logging:
        print s


def execute_shell_cmd(cmd):
    log("Running cmd: %s" % cmd)
    process = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True)
    return process.communicate()[0].strip()


def exec_stat_cmd(cmd, filen):
    filename = "%s/stats/%s" % (outdir, filen)
    execute_shell_cmd("echo \"Output from cmd '%s':\" > %s" % (cmd, filename))
    execute_shell_cmd("%s >> %s" % (cmd, filename))


# Return:
#   1: path is existed
#   2: path is not existed and successfully created
def check_create_path(path):
    if not os.path.exists(path):
        try:
            os.mkdir(path)
        except OSError:
            print("Creating directory %s failed." % path)
            sys.exit()
        else:
            log("Directory %s is created." % path)
            return 2
    else:
        log("Directory %s is existed." % path)
        return 1


def checkenv():
    global usr, sudo, outdir, host_type

    host_type = subprocess.check_output(['uname']).rstrip()
    if not (host_type == "Linux" or host_type == "FreeBSD"):
        print "This script only runs on Linux or FreeBSD"
        sys.exit()

    if host_type == 'FreeBSD' and usedtrace and pid != '':
        print "\nWarning: pmcstat will be used for profiling on \
              specific pid on FreeBSD"

    usr = subprocess.check_output(['id', '-u']).rstrip()
    if usr == 0:
        sudo = ''

    outdir = outdir + '_' + execute_shell_cmd('whoami')
    if check_create_path(outdir) == 1:
        sys.exit("\nERROR: Directory %s is existed. Delete the \
                    directory and re-run the cmd.\n" % outdir)


def readArgs():
    global sampletime, server, senddataonly, datatype
    global logging, argpid, pid, usedtrace, hz, flame_locally

    parser = argparse.ArgumentParser(
        description="Collects profiling and statistics data and "
                    "send to server",
        formatter_class=argparse.RawTextHelpFormatter
    )
    parser.add_argument('-a',
                        help=textwrap.dedent('''\
                        sets process server name or ip to receive collected
                        data (default: %s)''' % server),
                        required=False, action='store', dest='server')
    parser.add_argument('-d',
                        help='profile the specific pid or process',
                        required=False, action='store', dest='pid')
    parser.add_argument('-n', default=False,
                        help=textwrap.dedent('''\
                        not collect new data. Just send the previously
                        collected data'''),
                        required=False, action='store_true',
                        dest='senddataonly')
    parser.add_argument('-p',
                        help=textwrap.dedent('''\
                        sets process server port to receive collected
                        data (default: %s)''' % port),
                        required=False, action='store', dest='port')
    parser.add_argument('-r',
                        help=textwrap.dedent('''\
                        profiling all processes using DTrace on FreeBSD.
                        Default is using pmcstat.'''),
                        required=False, action='store_true', dest='usedtrace')
    parser.add_argument('-s',
                        help='sets the time in seconds to take samples '
                        'over (default 120).',
                        required=False, action='store', dest='sampletime')
    parser.add_argument('-t',
                        help=textwrap.dedent('''\
                        sets the data type to be collected.
                        profile:   collect profiling data only. Default
                        stats:     collect performance statistics data only
                        all:       collect both profiling and statistics data
                        '''),
                        required=False, action='store', dest='datatype')
    parser.add_argument('-v',
                        help='print out more messages', default=False,
                        required=False, action='store_true', dest='verbose')
    parser.add_argument('-z',
                        help='profiling Hertz (99, 997, 999, etc). Default: 99',
                        required=False, action='store', dest='hz')
    # hidden option: generate flame graph locally
    parser.add_argument('-f',
                        help=argparse.SUPPRESS,
                        required=False, action='store_true', dest='flame')

    args = parser.parse_args()
    sampletime = args.sampletime if args.sampletime else sampletime
    server = args.server if args.server else server
    senddataonly = args.senddataonly if args.senddataonly else senddataonly
    datatype = args.datatype if args.datatype else datatype
    logging = args.verbose if args.verbose else logging
    usedtrace = args.usedtrace if args.usedtrace else usedtrace
    flame_locally = args.flame if args.flame else flame_locally
    hz = args.hz if args.hz else hz

    if args.pid:
        try:
            int(args.pid)
        except ValueError:
            # convert process into comma separated pid or pid list
            pid = execute_shell_cmd("ps -aux | grep -w %s | grep -v grep | "
                                    "grep -v sirperfulous.py |awk {'print $2'}"
                                    " | paste -s -d, -" % args.pid)
            if (pid == ""):
                print "Process %s is not running" % args.pid
                sys.exit()
        else:
            pid = args.pid
        log("pid is: %s" % pid)
        argpid = args.pid


def check_perf():
    if (execute_shell_cmd("dpkg --list linux-tools-common | "
                          "grep linux-tools-common | grep ii") != ""):
        s = "linux-tools-`uname -r`"
        if (execute_shell_cmd("dpkg --list %s | grep %s | "
                              "grep ii" % (s, s)) != ""):
            log("perf is installed on the system")
        else:
            print "%s is not installed" % s
            print "Install %s by running \"sudo apt-get install \
                    linux-tools-`uname -r`\"" % s
            sys.exit()
    else:
        print "linux-tools-common is not installed"
        print "Install linux-tools-common by running \"sudo apt install \
                linux-tools-common\""
        sys.exit()


def stackcollapse():
    if host_type == 'Linux':
        if (argpid != ''):
            cmd = "%s perf script -f -i %s/perf_cpu_%s.data | stackcollapse-perf.pl > \
                %s/perf_cpu.folded; if [ %s -ne 0 ]; then %s chown %s \
                %s/perf_cpu.folded; fi" % (sudo, outdir, argpid,
                                           outdir, usr, sudo, usr, outdir)
        else:
            cmd = "%s perf script -f -i %s/perf_cpu.data | stackcollapse-perf.pl > \
                %s/perf_cpu.folded; if [ %s -ne 0 ]; then %s chown %s \
                %s/perf_cpu.folded; fi" % (sudo, outdir, outdir, usr,
                                           sudo, usr, outdir)

        execute_shell_cmd(cmd)
    else:
        if (usedtrace and argpid == ''):
            cmd = "%s stackcollapse.pl %s/perf_cpu_dtrace.data > %s/perf_cpu_dtraces.folded \
                " % (sudo, outdir, outdir)
        else:
            cmd = "%s stackcollapse-pmc.pl %s/perf_cpu_pmcstat.data> %s/perf_cpu_pmcstat.folded \
                " % (sudo, outdir, outdir)
            execute_shell_cmd(cmd)
            cmd = "%s /sbin/kldunload hwpmc" % sudo
        execute_shell_cmd(cmd)


def collect_profiling():
    print("Profiling. Will take %s seconds ..." % sampletime)

    if host_type == 'Linux':
        # CPU
        if (argpid != ''):
            cmd = "cd %s; sudo perf record -F %s  -p %s -g -o perf_cpu_%s.data -- sleep %s; \
                if [ %s -ne 0 ]; then sudo chown %s perf_cpu_%s.data; fi \
                " % (outdir, hz, pid, argpid, sampletime, usr, usr, argpid)
        else:
            cmd = "cd %s; sudo perf record -F %s -a -g -o perf_cpu.data -- sleep %s; \
                if [ %s -ne 0 ]; then sudo chown %s perf_cpu.data; fi \
                " % (outdir, hz, sampletime, usr, usr)

        execute_shell_cmd(cmd)
    else:
        # CPU
        if (usedtrace and argpid == ''):
            cmd = 'cd %s; ' % outdir
            cmd += 'sudo dtrace -x stackframes=100 -x ustackframes=100 -n "profile-%s { \
                    @[stack(), ustack(), execname] = count(); } tick-%ss \
                    ' % (hz, sampletime)
            cmd += '{ printa(\\"%k-%k%s\\\\n%@d\\\\n\\", @); exit(0); }" \
                -o perf_cpu_dtrace.data; '
            cmd += 'if [ %s -ne 0 ]; then sudo chown %s perf_cpu_dtrace.data; \
                fi' % (usr, usr)
            execute_shell_cmd(cmd)
        else:
            cmd = "%s /sbin/kldload hwpmc" % sudo
            execute_shell_cmd(cmd)
            tmp = '-S unhalted-cycles' if argpid == '' \
                  else ("-P unhalted-cycles -t %s" % argpid)
            cmd = "cd %s; sudo pmcstat %s -O perf_cpu_pmcstat.raw -l %s; if \
                   [ %s -ne 0 ]; then sudo chown %s perf_cpu_pmcstat.raw; fi" \
                   % (outdir, tmp, sampletime, usr, usr)
            execute_shell_cmd(cmd)
            cmd = "cd %s; pmcstat -R perf_cpu_pmcstat.raw -G \
                    perf_cpu_pmcstat.data" % outdir
            execute_shell_cmd(cmd)
            execute_shell_cmd("cd %s; rm perf_cpu_pmcstat.raw" % outdir)

    stackcollapse()

    if not flame_locally:
        # zip all the files
        execute_shell_cmd("cd %s; tar -zcf perf.tar.gz perf*.data \
                perf*.folded" % outdir)
        execute_shell_cmd("cd %s; rm -f perf*.data perf*.folded" % outdir)

    print "\nDone profiling ...\n"


def collect_stats():
    print "\nCollecting stats. Will take 1 minute ...\n"

    check_create_path("%s/stats" % outdir)

    if host_type == 'Linux':
        exec_stat_cmd("uptime", "uptime.txt")
        exec_stat_cmd("free -m", "free.txt")
        exec_stat_cmd("dmesg", "dmesg.txt")

        exec_stat_cmd("ifconfig -a", "ifconfig.txt")
        exec_stat_cmd("netstat -natuw", "netstat.txt")
        exec_stat_cmd("top -b -n 5", "top.txt")
        exec_stat_cmd("ps -aux", "ps-aux.txt")
        exec_stat_cmd("ps -aux | grep squid", "ps-squid.txt")

        nics = execute_shell_cmd("ifconfig -a | cut -c 1-4 | sort | uniq -u | \
                grep eth | xargs")
        for n in nics.split():
            cmd = "echo \"Output from cmd 'ethtool -k':\" > %s \
                    " % ("%s/stats/ethtool-%s.txt" % (outdir, n))
            execute_shell_cmd(cmd)
            execute_shell_cmd("ethtool -k %s >> %s/stats/ethtool-%s.txt"
                              % (n, outdir, n))

        exec_stat_cmd("timeout 10 vmstat 1", "vmstat.txt")
        exec_stat_cmd("timeout 10 mpstat -P ALL 1", "mpstat.txt")
        exec_stat_cmd("timeout 10 pidstat 1", "pidstat.txt")
        exec_stat_cmd("timeout 10 iostat -xz 1", "iostat.txt")
        exec_stat_cmd("sar -u -r -S -b -w -q 2 10", "sar.txt")
        exec_stat_cmd("timeout 10 sar -n DEV 1", "sar-DEV.txt")
        exec_stat_cmd("timeout 10 sar -n TCP,ETCP,DEV 1", "sar-TCP.txt")
    else:
        exec_stat_cmd("uptime", "uptime.txt")
        exec_stat_cmd("dmesg -a", "dmesg.txt")

        exec_stat_cmd("ifconfig -a", "ifconfig.txt")
        exec_stat_cmd("netstat -na", "netstat.txt")
        exec_stat_cmd("top -b -n 5", "top.txt")
        exec_stat_cmd("ps -aux", "ps-aux.txt")
        exec_stat_cmd("ps -aux | grep squid", "ps-squid.txt")

        nics = execute_shell_cmd("ifconfig -a | cut -c 1-4 | sort | uniq -u \
                | grep eth | xargs")
        for n in nics.split():
            cmd = "echo \"Output from cmd 'ethtool -k':\" > \
                    %s" % ("%s/stats/ethtool-%s.txt" % (outdir, n))
            execute_shell_cmd(cmd)
            execute_shell_cmd("ethtool -k %s >> %s/stats/ethtool-%s.txt" %
                              (n, outdir, n))

        exec_stat_cmd("timeout 10 vmstat 1", "vmstat.txt")
        exec_stat_cmd("vmstat -m", "vmstat-m.txt")
        exec_stat_cmd("swapinfo", "swapinfo.txt")
        exec_stat_cmd("timeout 10 iostat -xz 1", "iostat.txt")

    if not flame_locally:
        # zip all the files
        execute_shell_cmd("cd %s; tar -zcf stats.tar.gz stats/*.txt" % outdir)
        execute_shell_cmd("rm -rf %s/stats" % outdir)

    print "\nDone collecting stats ...\n"


def generate_flame():
    if host_type == 'Linux':
        # CPU
        if argpid == '':
            execute_shell_cmd(("cat %s/perf_cpu.folded | flamegraph.pl \
                --title=\"CPU Flame Graph\" > %s/perf-`whoami`-cpu-all.svg\
                ") % (outdir, outdir))

            execute_shell_cmd(("grep -v cpu_idle %s/perf_cpu.folded | flamegraph.pl \
                --title=\"Non-idle CPU Flame Graph\" > \
                %s/perf-`whoami`-cpu-nonidle.svg") % (outdir, outdir))
            execute_shell_cmd(("grep -i squid %s/perf_cpu.folded | flamegraph.pl \
                --title=\"Squid CPU Flame Graph\" > \
                %s/perf-`whoami`-cpu-squid.svg") % (outdir, outdir))
        else:
            execute_shell_cmd(("cat %s/perf_cpu.folded | flamegraph.pl \
                --title=\"CPU Flame Graph\" > %s/perf-`whoami`-cpu-%s.svg\
                ") % (outdir, outdir, argpid))
    else:
        # CPU
        if argpid == '':
            execute_shell_cmd(("cat %s/perf_cpu_*.folded | flamegraph.pl \
                --title=\"CPU Flame Graph\" > %s/perf-`whoami`-cpu-all.svg\
                ") % (outdir, outdir))
            execute_shell_cmd(("egrep -v '^idle;-' %s/perf_cpu_*.folded | \
                flamegraph.pl --title=\"Non-idle CPU Flame Graph\" > \
                %s/perf-`whoami`-cpu-nonidle.svg") % (outdir, outdir))
            execute_shell_cmd(("grep -i squid %s/perf_cpu_*.folded | flamegraph.pl \
                --title=\"Squid CPU Flame Graph\" > \
                %s/perf-`whoami`-cpu-squid.svg") % (outdir, outdir))
        else:
            execute_shell_cmd(("cat %s/perf_cpu_pmcstat.folded | flamegraph.pl \
                --title=\"CPU Flame Graph\" > %s/perf-`whoami`-cpu-%s.svg\
                ") % (outdir, outdir, argpid))


def send_data():
    if not os.path.exists("%s/perf_stats_data_%s.tar.gz"
                          % (outdir, host_type)):
        sys.exit("Data not existed: %s/perf_stats_data_%s.tar.gz"
                 % (outdir, host_type))

    print "\nSending collected data to server %s:%s" % (server, port)

    host = execute_shell_cmd("hostname").rstrip()
    ts = datetime.now().strftime("%Y_%m_%d_%H_%M_%S")
    sock = socket.socket()
    sock.connect((server, port))

    print "Connected to server: %s:%s" % (server, port)
    filename = "%s/perf_stats_data_%s.tar.gz" % (outdir, host_type)
    filesize = os.path.getsize(filename)
    me = execute_shell_cmd("whoami").rstrip()
    sock.send("%s %s %s %d %s %s" % (host, ts, filename, filesize, argpid, me))
    data = sock.recv(16)

    with open(filename, "rb") as f:
        while True:
            bytes_read = f.read(4096)
            if not bytes_read:
                break
            sock.sendall(bytes_read)
    f.close()
    sock.close()
    print "Sending data done"


def main():
    readArgs()
    checkenv()

    print "\nResults at %s\n" % outdir

    if host_type == 'Linux':
        check_perf()
    if senddataonly:
        send_data()
        sys.exit()

    if datatype == "profile":
        collect_profiling()
    elif datatype == "stats":
        collect_stats()
    else:
        collect_profiling()
        collect_stats()

    if flame_locally:
        print "Generating flame graph ..."
        generate_flame()
    else:
        if (os.path.exists("%s/perf.tar.gz" % outdir) or
                os.path.exists("%s/stats.tar.gz" % outdir)):
            execute_shell_cmd("cd %s; tar -zcf perf_stats_data_%s.tar.gz \
                    *.tar.gz" % (outdir, host_type))
            execute_shell_cmd("cd %s; rm -f perf.tar.gz stats.tar.gz" % outdir)

        send_data()

    print "\nDone...\n"

    if flame_locally:
        print "Flame graphs are generated at: %s\n" % outdir
    else:
        print "Flame graphs are sent to server %s:/sirPerfulous/data/%s\n" \
                % (server, execute_shell_cmd("hostname").rstrip())


if __name__ == '__main__':
    main()
