#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <error.h>
#include <limits.h>
#include <assert.h>
#include <string.h>
#include <errno.h>
#include <ifaddrs.h>
#include <signal.h>
#include <fcntl.h>
#include <syslog.h>
#include <stddef.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/mman.h>
#include <netinet/ether.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip6.h>
#include <netinet/ip_icmp.h>
#include <netinet/icmp6.h>
#include <arpa/inet.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <net/if_arp.h>
#include <net/ethernet.h>
#include <linux/if_ether.h>
#include <linux/if_packet.h>
#include <linux/filter.h>
#include <linux/rtnetlink.h>

/* protocol specific constants */
/* Ethernet header length */
#define ETH_HDRLEN 14
/* max Ethernet data length */
#define ETH_DATLEN 1500
/* max Ethernet frame length */
#define ETH_FRMLEN (ETH_HDRLEN + ETH_DATLEN)
/* IPv4 header length */
#define IP4_HDRLEN 20
/* IPv6 header length */
#define IP6_HDRLEN 40
/* ARP header length */
#define ARP_HDRLEN 28
/* NDP header length */
#define NDP_HDRLEN 24
/* ICMP header length */
#define ICMP_HDRLEN 8
/* max ICMPv4 data length */
#define ICMP4_DATLEN (ETH_DATLEN - IP4_HDRLEN - ICMP_HDRLEN)
/* max ICMPv6 data length */
#define ICMP6_DATLEN (ETH_DATLEN - IP6_HDRLEN - ICMP_HDRLEN)

/* command line switches related constants */
#define MIN_PKT_SIZE 1
#define MAX_PKT_SIZE 65535
#define MIN_PKT_COUNT 1
#define MAX_PKT_COUNT 65535
#define MIN_PKT_TIME 1
#define MAX_PKT_TIME 3600
#define MAX_IF_NAME 256

/* bit operations on a char[] mask of unlimited length */
#define CBIT_BIT(bit)           (1 << ((bit) % 8))
#define CBIT_BIN(mask, bit)     (mask)[(bit) >> 3]
#define CBIT_SET(mask, bit)     ((void)(CBIT_BIN(mask, bit) |= CBIT_BIT(bit)))
#define CBIT_CLR(mask, bit)     ((void)(CBIT_BIN(mask, bit) &= ~CBIT_BIT(bit)))
#define CBIT_TEST(mask, bit)    (CBIT_BIN(mask, bit) & CBIT_BIT(bit))

/* sleep between daemon connect retries */
#define DAEMON_WAIT_SLEEP 10000 /* us */

/* 128 blocks of 32768 bytes = 4MB ring per interface */
#define RING_FRAME_SIZE 2048
#define RING_BLOCK_SIZE 32768
#define RING_BLOCK_COUNT 128
#define RING_BLOCK_TIMEOUT 60 /* ms */

/* internal error exit code */
#define EXIT_INTERNAL_ERROR 2

/* helper macro for parsing routing attributes (rtnetlink) */
#define NDA_RTA(r) ((struct rtattr *)(((char *)(r)) + NLMSG_ALIGN(sizeof(struct ndmsg))))
/* size of buffer for netlink response */
#define NETLINK_RCV_BUF 65536

/* debug helper */
#define debug(fmt, ...) cond_print(debug, fmt, ##__VA_ARGS__)
/* quiet print helper */
#define quiet_print(fmt, ...) cond_print(!quiet, fmt, ##__VA_ARGS__)
/* conditional print */
#define cond_print(cond, fmt, ...) \
    do { \
        if (cond) { \
            if (in_daemon) {\
                syslog(LOG_DEBUG, fmt, ##__VA_ARGS__);\
            } else {\
                printf(fmt, ##__VA_ARGS__); \
            }\
        } \
    } while (0);
/* exit with error report in stderr/syslog */
#define error_exit(status, errnum, fmt, ...) \
    do { \
        if (in_daemon) {\
            syslog(LOG_CRIT, fmt": %s", ##__VA_ARGS__, strerror(errnum));\
            exit(status);\
        } else {\
            error(status, errnum, fmt, ##__VA_ARGS__);\
        }\
    } while (0);

/* define a struct for ARP header */
struct arp {
    uint16_t htype;
    uint16_t ptype;
    uint8_t hlen;
    uint8_t plen;
    uint16_t opcode;
    struct ether_addr sender_ether;
    struct in_addr sender_inet;
    struct ether_addr target_ether;
    struct in_addr target_inet;
} __attribute__ ((__packed__));

/* define a struct for inet address */
struct inaddr {
    int family;
    union {
        struct in_addr in;
        struct in6_addr in6;
    } addr;
};

/* define a struct for ring buffer block descriptor */
struct ring_bdesc {
    uint32_t version;
    uint32_t offset_to_priv;
    struct tpacket_hdr_v1 h1;
};

/* define a struct for ring buffer */
struct ring {
    int fd;
    int iface_index;
    struct iovec *rd;
    uint8_t *map;
    unsigned block;
};

/* socket filter for daemon mode (filters out vlan tagged packets)
 * generated by the following:
 * tcpdump '(arp && arp[6:2] == 2) ||
 * (ip && ip[0] & 0xf == 5 && icmp && icmp[icmptype] == icmp-echoreply) ||
 * (ip6 && ip6[6] == 58 && (ip6[40] == 136 || ip6[40] == 129)) && not vlan' -dd */
static struct sock_fprog daemonfilter = {
    .filter = (struct sock_filter []) {
        {0x28, 0, 0, 0x0000000c},
        {0x15, 0, 2, 0x00000806},
        {0x28, 0, 0, 0x00000014},
        {0x15, 17, 20, 0x00000002},
        {0x15, 0, 10, 0x00000800},
        {0x30, 0, 0, 0x0000000e},
        {0x54, 0, 0, 0x0000000f},
        {0x15, 0, 16, 0x00000005},
        {0x30, 0, 0, 0x00000017},
        {0x15, 0, 14, 0x00000001},
        {0x28, 0, 0, 0x00000014},
        {0x45, 12, 0, 0x00001fff},
        {0xb1, 0, 0, 0x0000000e},
        {0x50, 0, 0, 0x0000000e},
        {0x15, 6, 9, 0x00000000},
        {0x15, 0, 8, 0x000086dd},
        {0x30, 0, 0, 0x00000014},
        {0x15, 0, 6, 0x0000003a},
        {0x30, 0, 0, 0x00000036},
        {0x15, 1, 0, 0x00000088},
        {0x15, 0, 3, 0x00000081},
        {0x30, 0, 0, 0xfffff030},
        {0x15, 1, 0, 0x00000001},
        {0x6, 0, 0, ETH_FRMLEN},
        {0x6, 0, 0, 0x00000000}
    },
    .len = 25
};

/* socket filter for ARP replies (filters out vlan tagged packets),
 * generated by the following:
 * tcpdump 'arp && arp[6:2] == 2 && not vlan' -dd */
static struct sock_fprog arpfilter = {
    .filter = (struct sock_filter []) {
        {0x28, 0, 0, 0x0000000c},
        {0x15, 0, 5, 0x00000806},
        {0x28, 0, 0, 0x00000014},
        {0x15, 0, 3, 0x00000002},
        {0x30, 0, 0, 0xfffff030},
        {0x15, 1, 0, 0x00000001},
        {0x6, 0, 0, ETH_FRMLEN},
        {0x6, 0, 0, 0x00000000}
    },
    .len = 8
};

/* socket filter for ICMP replies with ident (filters out IP packets with
 * options and vlan tagged packets) generated by the following:
 * tcpdump 'ip && ip[0] & 0xf == 5 && icmp && icmp[icmptype] == icmp-echoreply
 * && icmp[4:2] == 0xAAAA && not vlan' -dd */
static struct sock_fprog icmpfilter = {
    .filter = (struct sock_filter []) {
        {0x28, 0, 0, 0x0000000c},
        {0x15, 0, 15, 0x00000800},
        {0x30, 0, 0, 0x0000000e},
        {0x54, 0, 0, 0x0000000f},
        {0x15, 0, 12, 0x00000005},
        {0x30, 0, 0, 0x00000017},
        {0x15, 0, 10, 0x00000001},
        {0x28, 0, 0, 0x00000014},
        {0x45, 8, 0, 0x00001fff},
        {0xb1, 0, 0, 0x0000000e},
        {0x50, 0, 0, 0x0000000e},
        {0x15, 0, 5, 0x00000000},
        {0x48, 0, 0, 0x00000012},
        {0x15, 0, 3, 0x0000aaaa},
        {0x30, 0, 0, 0xfffff030},
        {0x15, 1, 0, 0x00000001},
        {0x6, 0, 0, ETH_FRMLEN},
        {0x6, 0, 0, 0x00000000}
    },
    .len = 18
};

/* socket filter for NDP NA replies (filters out vlan tagged packets),
 * generated by the following:
 * tcpdump 'ip6 && ip6[6] == 58 && ip6[40] == 136 && not vlan' -dd */
static struct sock_fprog ndpnafilter = {
    .filter = (struct sock_filter []) {
        {0x28, 0, 0, 0x0000000c},
        {0x15, 0, 7, 0x000086dd},
        {0x30, 0, 0, 0x00000014},
        {0x15, 0, 5, 0x0000003a},
        {0x30, 0, 0, 0x00000036},
        {0x15, 0, 3, 0x00000088},
        {0x30, 0, 0, 0xfffff030},
        {0x15, 1, 0, 0x00000001},
        {0x6, 0, 0, ETH_FRMLEN},
        {0x6, 0, 0, 0x00000000}
    },
    .len = 10
};

/* socket filter for ICMPv6 replies with ident (filters out IP packets with
 * options and vlan tagged packets) generated by the following:
 * tcpdump 'ip6 && ip6[6] == 58 && ip6[40] == 129 && ip6[44:2] == 0xAAAA
 * && not vlan' -dd */
static struct sock_fprog icmp6filter = {
    .filter = (struct sock_filter []) {
        {0x28, 0, 0, 0x0000000c},
        {0x15, 0, 9, 0x000086dd},
        {0x30, 0, 0, 0x00000014},
        {0x15, 0, 7, 0x0000003a},
        {0x30, 0, 0, 0x00000036},
        {0x15, 0, 5, 0x00000081},
        {0x28, 0, 0, 0x0000003a},
        {0x15, 0, 3, 0x0000aaaa},
        {0x30, 0, 0, 0xfffff030},
        {0x15, 1, 0, 0x00000001},
        {0x6, 0, 0, ETH_FRMLEN},
        {0x6, 0, 0, 0x00000000},
    },
    .len = 12
};

/* our identity for ICMP packets */
static uint16_t ident = 0;
/* seq of last sent ICMP packet */
static uint16_t seq = 0;
/* bitset of received ICMP replies (seq) */
static uint8_t acks[MAX_PKT_COUNT / 8 + 1] = {0};
/* unix socket name (abstract address namespace) */
static const char sockname[] = "macping.socket";
/* pid file path */
static const char pidfile[] = "/var/run/macping.pid";
/* true if running as daemon */
static int in_daemon = 0;
/* maps client fd to state (0: unknown, 1: connected, 2+: iface index) */
static int clients[FD_SETSIZE] = {0};
/* maps iface index to raw fd */
static int ifacefd[FD_SETSIZE] = {0};
/* maps iface raw fd to ring */
static struct ring *ifacerg[FD_SETSIZE] = {0};
/* maps iface index to ref count */
static int ifacerc[FD_SETSIZE] = {0};
/* number of current client connections */
static size_t client_fds = 0;

/* command line parameters (with defaults) */
static unsigned packet_size = 56;
static unsigned packet_count = MAX_PKT_COUNT;
static unsigned send_interval = 1;
static unsigned wait = 0;
static unsigned cached_ether = 1;
static unsigned quiet = 0;
static unsigned debug = 0;
static unsigned daemon_mode = 0;
static const char *iface = NULL;
static const char *vip = NULL;
static const char *server = NULL;

static void
usage(const char *name)
{
    fprintf(stderr, "usage: %s -I [<ip>][%%][<iface>] [OPTIONS] <vip> <server ip>\n"
        "\nOPTIONS\n"
        " -s packet size\n"
        " -c count\n"
        " -i interval\n"
        " -w wait\n"
        " -l do not use arp/ndp cache\n"
        " -b background/daemon mode\n"
        " -q quiet\n",
        name);
    exit(EXIT_INTERNAL_ERROR);
}

static unsigned
safe_atou(char arg, const char *value, unsigned min, unsigned max)
{
    unsigned long res = strtoul(value, NULL, 10);
    if (res < min || res > max) {
        error_exit(EXIT_INTERNAL_ERROR, 0, "invalid value to -%c switch: '%s', must be an "
            "integer in [%u, %u] range", arg, value, min, max);
    }
    return res;
}

static void
parse_arguments(int argc, char *argv[])
{
    int opt = -1;

    if (argc < 4) {
        usage(argv[0]);
    }

    while ((opt = getopt(argc, argv, "I:s:c:i:w:ldqb")) != -1) {
        switch (opt) {
            case 'I':
                iface = optarg;
                break;
            case 's':
                packet_size = safe_atou(opt, optarg, MIN_PKT_SIZE, MAX_PKT_SIZE);
                break;
            case 'c':
                packet_count = safe_atou(opt, optarg, MIN_PKT_COUNT, MAX_PKT_COUNT);
                break;
            case 'i':
                send_interval = safe_atou(opt, optarg, MIN_PKT_TIME, MAX_PKT_TIME);
                break;
            case 'w':
                wait = safe_atou(opt, optarg, MIN_PKT_TIME, MAX_PKT_TIME);
                break;
            case 'l':
                cached_ether = 0;
                break;
            case 'd':
                debug = 1;
                break;
            case 'q':
                quiet = 1;
                break;
            case 'b':
                daemon_mode = 1;
                break;
            default:
                usage(argv[0]);
        }
    }

    if (iface == NULL || optind >= argc || (argc - optind) != 2) {
        usage(argv[0]);
    }

    vip = argv[optind++];
    server = argv[optind];

    debug("vip: %s server: %s iface: %s\n", vip, server, iface);
}

static uint16_t
count_acks()
{
    uint16_t i = 1, res = 0;
    for (; i <= seq; ++i) {
        res += CBIT_TEST(acks, i) ? 1 : 0;
    }
    return res;
}

static int
summary()
{
    quiet_print("\n--- %s ping statistics ---\n", vip);
    uint16_t ack_count = count_acks();
    uint8_t loss = seq ? 100 - ack_count * 100 / seq : 0;
    quiet_print("%u packets transmitted, %u received, %u%% packet loss\n", seq,
        ack_count, loss);
    return !seq || loss;
}

static void
sighandler(int signal)
{
    exit(summary());
}

/* computing the internet checksum (RFC 1071) */
static uint32_t
checksum_update(uint32_t sum, const uint16_t *addr, int len)
{
    int count = len;

    /* sum up 2-byte values until none or only one byte left */
    while (count > 1) {
        sum += *(addr++);
        count -= 2;
    }

    /* add left-over byte, if any */
    if (count > 0) {
        sum += *(uint8_t *)addr;
    }

    return sum;
}

static uint32_t
checksum_init(const uint16_t *addr, int len)
{
    return checksum_update(0, addr, len);
}

static uint16_t
checksum_final(uint32_t sum)
{
    uint16_t answer = 0;

    /* fold 32-bit sum into 16 bits; we lose information by doing this, */
    /* increasing the chances of a collision */
    /* sum = (lower 16 bits) + (upper 16 bits shifted right 16 bits) */
    while (sum >> 16) {
        sum = (sum & 0xffff) + (sum >> 16);
    }

    /* checksum is one's compliment of sum */
    answer = ~sum;

    return answer;
}

static char *
strlcpy(char *dst, const char *src, size_t size)
{
    strncpy(dst, src, size - 1);
    dst[size - 1] = 0;
    return dst;
}

static void
safe_send(int sockfd, const void *buf, size_t len)
{
    ssize_t bytes = -1;
    while ((bytes = send(sockfd, buf, len, MSG_DONTWAIT)) < 0 && errno == EINTR);

    if (bytes < 0) {
        error_exit(EXIT_INTERNAL_ERROR, errno, "send() failed");
    }

    /* TODO retry here? */
    if (bytes != len) {
        error_exit(EXIT_INTERNAL_ERROR, 0, "partial send()");
    }
}

static void
safe_send_and_forget(int sockfd, const void *buf, size_t len)
{
    ssize_t bytes = -1;
    while ((bytes = send(sockfd, buf, len, MSG_DONTWAIT)) < 0 && errno == EINTR);
}

static ssize_t
safe_recv(int sockfd, void *buf, size_t len)
{
    ssize_t bytes = -1;
    while ((bytes = recv(sockfd, buf, len, 0)) < 0 && errno == EINTR);

    if (!in_daemon && !bytes) {
        error_exit(EXIT_INTERNAL_ERROR, 0, "EOF from raw socket");
    }

    if (!in_daemon && bytes < 0) {
        error_exit(EXIT_INTERNAL_ERROR, errno, "recv() failed");
    }

    return bytes;
}

static int
safe_select(int sockfd, struct timeval *tmo)
{
    fd_set readfds;
    FD_ZERO(&readfds);
    FD_SET(sockfd, &readfds);

    int res = 0;
    while ((res = select(sockfd + 1, &readfds, NULL, NULL, tmo)) < 0 && errno == EINTR);

    if (res < 0) {
        error_exit(EXIT_INTERNAL_ERROR, errno, "select() failed");
    }

    return res;
}

static void
set_filter_and_drain(int sockfd, const struct sock_fprog *filter)
{
    char b;
    ssize_t bytes;
    if (filter) {
        /* attach (replace) socket filter */
        if (setsockopt(sockfd, SOL_SOCKET, SO_ATTACH_FILTER, filter, sizeof(*filter)) < 0) {
            error_exit(EXIT_INTERNAL_ERROR, errno, "setsockopt(SO_ATTACH_FILTER) failed");
        }
    }
    /* drain socket receive queue */
    while ((bytes = recv(sockfd, &b, sizeof(b), MSG_TRUNC | MSG_DONTWAIT)) > 0 ||
        (bytes < 0 && errno == EINTR));
}

static struct ring *
create_socket_ring(int iface_index)
{
    struct ring *ring = calloc(1, sizeof(struct ring));
    assert(ring);

    ring->iface_index = iface_index;

    ring->fd = socket(PF_PACKET, SOCK_RAW, htons(ETH_P_ALL));
    if (ring->fd < 0) {
        error_exit(EXIT_INTERNAL_ERROR, errno, "PF_PACKET socket() failed");
    }
    set_filter_and_drain(ring->fd, &daemonfilter);

    int v = TPACKET_V3;
    int res = setsockopt(ring->fd, SOL_PACKET, PACKET_VERSION, &v, sizeof(v));
    if (res < 0) {
        error_exit(EXIT_INTERNAL_ERROR, errno, "PF_PACKET setsockopt(PACKET_VERSION) failed");
    }

    struct tpacket_req3 req = {0};
    req.tp_block_size = RING_BLOCK_SIZE;
    req.tp_frame_size = RING_FRAME_SIZE;
    req.tp_block_nr = RING_BLOCK_COUNT;
    req.tp_frame_nr = (RING_BLOCK_SIZE * RING_BLOCK_COUNT) / RING_FRAME_SIZE;
    req.tp_retire_blk_tov = RING_BLOCK_TIMEOUT;
    req.tp_feature_req_word = TP_FT_REQ_FILL_RXHASH;

    res = setsockopt(ring->fd, SOL_PACKET, PACKET_RX_RING, &req, sizeof(req));
    if (res < 0) {
        error_exit(EXIT_INTERNAL_ERROR, errno, "PF_PACKET setsockopt(PACKET_RX_RING) failed");
    }

    ring->map = mmap(NULL, RING_BLOCK_SIZE * RING_BLOCK_COUNT,
             PROT_READ | PROT_WRITE, MAP_SHARED | MAP_LOCKED, ring->fd, 0);
    if (ring->map == MAP_FAILED) {
        error_exit(EXIT_INTERNAL_ERROR, errno, "PF_PACKET ring mmap failed");
    }

    ring->rd = calloc(RING_BLOCK_COUNT, sizeof(*ring->rd));
    assert(ring->rd);
    int i = 0;
    for (; i < RING_BLOCK_COUNT; ++i) {
        ring->rd[i].iov_base = ring->map + (i * RING_BLOCK_SIZE);
        ring->rd[i].iov_len = RING_BLOCK_SIZE;
    }

    /* fill out sockaddr_ll */
    struct sockaddr_ll device = {0};
    device.sll_ifindex = iface_index;
    device.sll_family = PF_PACKET;
    device.sll_protocol = htons(ETH_P_ALL);
    device.sll_pkttype = PACKET_HOST;

    res = bind(ring->fd, (struct sockaddr *)&device, sizeof(device));
    if (res < 0) {
        error_exit(EXIT_INTERNAL_ERROR, errno, "PF_PACKET bind() failed");
    }

    return ring;
}

static void
destroy_socket_ring(struct ring *ring)
{
    munmap(ring->map, RING_BLOCK_SIZE * RING_BLOCK_COUNT);
    free(ring->rd);
    close(ring->fd);
    free(ring);
}

static void
handle_connection(int listenfd, fd_set *fds)
{
    int clientfd = -1;
    if ((clientfd = accept(listenfd, NULL, NULL)) < 0) {
        if (errno != EMFILE && errno != ENFILE) {
            error_exit(EXIT_INTERNAL_ERROR, errno, "AF_UNIX accept() failed");
        }
        return;
    }
    FD_SET(clientfd, fds);
    client_fds++;
    clients[clientfd] = 1;
    debug("accepted client fd: %d\n", clientfd);
}

static void
handle_client_event(int clientfd, fd_set *fds)
{
    /* packet buffer */
    char frame[ETH_FRMLEN] = {0};
    /* bytes received */
    ssize_t bytes = safe_recv(clientfd, frame, ETH_FRMLEN);
    debug("read from client fd: %d bytes: %zd\n", clientfd, bytes);
    if (bytes == sizeof(int) && clients[clientfd] == 1) {
        /* client binds itself to iface index */
        char *tmp = frame;
        int iface_index = *((int *)tmp);
        clients[clientfd] = iface_index + 2;
        ifacerc[iface_index]++;
        debug("client fd: %d at iface: %d refc: %d\n", clientfd, iface_index, ifacerc[iface_index]);
    } else if (bytes > 0 && clients[clientfd] > 1) {
        /* client sends network frame */
        int iface_index = clients[clientfd] - 2;
        if (!ifacefd[iface_index]) {
            struct ring *ring = create_socket_ring(iface_index);
            FD_SET(ring->fd, fds);
            ifacefd[iface_index] = ring->fd;
            ifacerg[ring->fd] = ring;
            debug("opened rawfd: %d for iface: %d\n", ring->fd, iface_index);
        }
        /* inject frame into iface raw fd */
        safe_send(ifacefd[iface_index], frame, bytes);
    } else {
        /* error or EOF from client */
        close(clientfd);
        if (clients[clientfd] > 1) {
            int iface_index = clients[clientfd] - 2;
            if (--ifacerc[iface_index] == 0 && ifacefd[iface_index]) {
                FD_CLR(ifacefd[iface_index], fds);
                debug("closed rawfd: %d\n", ifacefd[iface_index]);
                destroy_socket_ring(ifacerg[ifacefd[iface_index]]);
                ifacerg[ifacefd[iface_index]] = NULL;
                ifacefd[iface_index] = 0;
            }
        }
        clients[clientfd] = 0;
        debug("closed client fd: %d\n", clientfd);
        FD_CLR(clientfd, fds);
        client_fds--;
    }
}

static void
handle_iface_event(int rawfd, fd_set *fds)
{
    debug("iface event rawfd: %d\n", rawfd);
    struct ring *ring = ifacerg[rawfd];
    struct ring_bdesc *pbd = ring->rd[ring->block].iov_base;
    /* loop over blocks with new frames */
    while (pbd->h1.block_status & TP_STATUS_USER) {
        struct tpacket3_hdr *ppd = (struct tpacket3_hdr *)
            ((uint8_t *)pbd + pbd->h1.offset_to_first_pkt);
        int i = 0;
        /* loop over frames in the current block */
        for (; i < pbd->h1.num_pkts; ++i) {
            int j = 0;
            /* forward to clients bound to this iface */
            for (; j < FD_SETSIZE; ++j) {
                if (clients[j] == ring->iface_index + 2) {
                    safe_send_and_forget(j, (uint8_t *)ppd + ppd->tp_mac, ppd->tp_snaplen);
                }
            }
            ppd = (struct tpacket3_hdr *)((uint8_t *)ppd + ppd->tp_next_offset);
        }
        debug("processed %u packets in iface block %u\n", pbd->h1.num_pkts, ring->block);
        /* tell kernel that it can take this block */
        pbd->h1.block_status = TP_STATUS_KERNEL;
        /* proceed to the next block */
        ring->block = (ring->block + 1) % RING_BLOCK_COUNT;
        pbd = ring->rd[ring->block].iov_base;
    }
}

static void
handle_socket_event(int fd, fd_set *fds)
{
    if (clients[fd]) {
        handle_client_event(fd, fds);
    } else {
        assert(ifacerg[fd]);
        handle_iface_event(fd, fds);
    }
}

static void
daemon_main(int listenfd, size_t max_clients)
{
    /* write pid file */
    FILE *pidf = fopen(pidfile, "w");
    if (!pidf) {
        error_exit(EXIT_INTERNAL_ERROR, errno, "failed to create %s", pidfile);
    }
    fprintf(pidf, "%d\n", getpid());
    fclose(pidf);

    quiet_print("macping daemon started");

    fd_set readfds, tempfds;
    FD_ZERO(&readfds);
    FD_SET(listenfd, &readfds);

    /* main loop */
    while (1) {
        debug("calling select()\n");
        int res = 0;
        tempfds = readfds;
        while ((res = select(FD_SETSIZE, &tempfds, NULL, NULL, NULL)) < 0 && errno == EINTR);
        debug("select() returned: %d\n", res);

        if (res < 0) {
            error_exit(EXIT_INTERNAL_ERROR, errno, "select() failed");
        }

        int i = 0;
        for (; i < FD_SETSIZE; ++i) {
            if (!FD_ISSET(i, &tempfds) || !FD_ISSET(i, &readfds)) {
                continue;
            }
            debug("read event fd: %d\n", i);
            if (i == listenfd) {
                handle_connection(i, &readfds);
            } else {
                handle_socket_event(i, &readfds);
            }
            /* stop accepting new connections if we have max */
            if (client_fds == max_clients && FD_ISSET(listenfd, &readfds)) {
                debug("not accepting new clients\n");
                FD_CLR(listenfd, &readfds);
            }
            /* start accepting new connections if we don't have max */
            if (client_fds < max_clients && !FD_ISSET(listenfd, &readfds)) {
                debug("accepting new clients\n");
                FD_SET(listenfd, &readfds);
            }
        }
    }
    /* not reached */
}

static void
start_daemon(int argc, char *argv[])
{
    /* open listen socket descriptor */
    int listenfd = -1;
    if ((listenfd = socket(AF_UNIX, SOCK_SEQPACKET, 0)) < 0) {
        error_exit(EXIT_INTERNAL_ERROR, errno, "AF_UNIX socket() failed");
    }

    /* fill out sockaddr_un */
    struct sockaddr_un sockaddr = {0};
    sockaddr.sun_family = AF_UNIX;
    /* linux abstract address namespace */
    snprintf(sockaddr.sun_path, sizeof(sockaddr.sun_path), "%c%s", 0, sockname);

    /* bind socket to the abstract address */
    if (bind(listenfd, (struct sockaddr *)&sockaddr, sizeof(sa_family_t) + sizeof(sockname)) < 0) {
        if (errno != EADDRINUSE) {
            error_exit(EXIT_INTERNAL_ERROR, errno, "AF_UNIX bind() failed");
        }
        /* another instance is running/starting */
        return;
    }

    if ((listen(listenfd, FD_SETSIZE)) < 0) {
        error_exit(EXIT_INTERNAL_ERROR, errno, "AF_UNIX listen() failed");
    }

    /* fork off the parent process */
    pid_t pid = fork();

    /* an error occurred */
    if (pid < 0) {
        error_exit(EXIT_INTERNAL_ERROR, errno, "fork() failed");
    }

    /* success: let the parent continue */
    if (pid > 0) {
        close(listenfd);
        return;
    }

    /* open the log file */
    openlog("macping", LOG_PID, LOG_DAEMON);
    in_daemon = 1;

    /* on success: The child process becomes session leader */
    if (setsid() < 0) {
        error_exit(EXIT_INTERNAL_ERROR, errno, "setsid() failed");
    }

    /* catch, ignore and handle signals */
    signal(SIGCHLD, SIG_IGN);
    signal(SIGHUP, SIG_IGN);

    /* fork off for the second time*/
    pid = fork();

    /* an error occurred */
    if (pid < 0) {
        error_exit(EXIT_INTERNAL_ERROR, errno, "fork() failed");
    }

    /* success: let the parent terminate */
    if (pid > 0) {
        exit(EXIT_SUCCESS);
    }

    /* set new file permissions */
    umask(0);

    /* change the working directory to the root directory */
    assert(chdir("/") == 0);

    size_t max_fds = sysconf(_SC_OPEN_MAX);
    if (max_fds > FD_SETSIZE) {
        max_fds = FD_SETSIZE;
    }

    /* close all open file descriptors */
    int i;
    for (i = max_fds; i >= 0; --i) {
        if (i != listenfd) {
            close(i);
        }
    }

    /* reopen std{in|out|err} to point to /dev/null */
    assert(open("/dev/null", O_RDONLY) == 0);
    assert(open("/dev/null", O_WRONLY) == 1);
    assert(open("/dev/null", O_WRONLY) == 2);

    /* clear all command line arguments */
    for (i = 1; i < argc; ++i) {
        memset(argv[i], 0, strlen(argv[i]));
    }

    daemon_main(listenfd, max_fds * 0.8);
    /* not reached */
}

static int
connect_to_daemon(int iface_index)
{
    int sockfd = -1;

    /* open socket descriptor */
    if ((sockfd = socket(AF_UNIX, SOCK_SEQPACKET, 0)) < 0) {
        error_exit(EXIT_INTERNAL_ERROR, errno, "AF_UNIX socket() failed");
    }

    /* fill out sockaddr_un */
    struct sockaddr_un sockaddr = {0};
    sockaddr.sun_family = AF_UNIX;
    /* linux abstract address namespace */
    snprintf(sockaddr.sun_path, sizeof(sockaddr.sun_path), "%c%s", 0, sockname);

    /* connect to the abstract address */
    if (connect(sockfd, (struct sockaddr *)&sockaddr, sizeof(sa_family_t) + sizeof(sockname)) < 0) {
        if (errno == ECONNREFUSED) {
            close(sockfd);
            return -1;
        }
        error_exit(EXIT_INTERNAL_ERROR, errno, "AF_UNIX connect() failed");
    }

    quiet_print("CONNECTED TO DAEMON\n");

    /* tell daemon our iface index */
    safe_send(sockfd, &iface_index, sizeof(iface_index));

    return sockfd;
}

static int
create_raw_socket(int iface_index)
{
    int sockfd = -1;

    /* open raw socket descriptor */
    if ((sockfd = socket(PF_PACKET, SOCK_RAW, htons(ETH_P_ALL))) < 0) {
        error_exit(EXIT_INTERNAL_ERROR, errno, "PF_PACKET socket() failed");
    }

    /* fill out sockaddr_ll */
    struct sockaddr_ll device = {0};
    device.sll_ifindex = iface_index;
    device.sll_family = PF_PACKET;
    device.sll_protocol = htons(ETH_P_ALL);
    device.sll_pkttype = PACKET_HOST;

    /* bind socket to the interface */
    if (bind(sockfd, (struct sockaddr *)&device, sizeof(device)) < 0) {
        error_exit(EXIT_INTERNAL_ERROR, errno, "PF_PACKET bind() failed");
    }

    return sockfd;
}

static int
inaddr_pton(const char *saddr, struct inaddr *iaddr)
{
    if (inet_pton(AF_INET, saddr, &iaddr->addr.in)) {
        iaddr->family = AF_INET;
        return 1;
    }
    if (inet_pton(AF_INET6, saddr, &iaddr->addr.in6)) {
        iaddr->family = AF_INET6;
        return 1;
    }
    return 0;
}

static const char *
inaddr_ntop(const struct inaddr *iaddr)
{
    static char saddr[INET6_ADDRSTRLEN];
    return inet_ntop(iaddr->family, &iaddr->addr, saddr, sizeof(saddr));
}

static int
inaddr_cmp(const struct inaddr *addr1, const void *addr2)
{
    if (addr1->family == AF_INET) {
        return memcmp(&addr1->addr.in, addr2, sizeof(addr1->addr.in));
    } else if (addr1->family == AF_INET6) {
        return memcmp(&addr1->addr.in6, addr2, sizeof(addr1->addr.in6));
    }
    return 0;
}

static int
if_lookup(struct inaddr *iaddr, struct ether_addr *eaddr, int family)
{
    /* make a copy of -I switch value */
    char iname[MAX_IF_NAME];
    strlcpy(iname, iface, sizeof(iname));

    struct ifaddrs *ifaddr = NULL, *ifa = NULL;
    /* check to see if we have zone index */
    char *zone = index(iname, '%');
    if (zone) {
        *zone++ = 0;
    }
    debug("iname: %s zone: %s\n", iname, zone);
    /* type of lookup: 0 - by iname, 1 - by iaddr */
    int type = inaddr_pton(iname, iaddr);
    /* iface name or inet based on type of lookup */
    int found_other = 0;
    int found_ether = 0;
    int if_index = -1;

    debug("lookup type: %d\n", type);

    /* if we had zone and inaddr_pton failed to parse address */
    if (type == 0 && zone != NULL) {
        error_exit(EXIT_INTERNAL_ERROR, 0, "inaddr_pton() on '%s' failed", iname);
    }

    /* if we lookup by iaddr and we had zone - we know iface name */
    if (type == 1 && zone != NULL) {
        strlcpy(iname, zone, sizeof(iname));
    }

    if (getifaddrs(&ifaddr) == -1) {
        error_exit(EXIT_INTERNAL_ERROR, errno, "getifaddrs");
    }

    /* find iface name or inet based on type of lookup */
    for (ifa = ifaddr; ifa != NULL && !found_other; ifa = ifa->ifa_next) {
        if (ifa->ifa_addr == NULL) {
            continue;
        }
        /* find iface inet by it's name */
        if (type == 0 && ifa->ifa_addr->sa_family == family &&
            strcmp(ifa->ifa_name, iname) == 0) {
            if (family == AF_INET) {
                found_other = 1;
                iaddr->family = AF_INET;
                memcpy(&iaddr->addr.in, &((struct sockaddr_in *)ifa->ifa_addr)->sin_addr,
                    sizeof(iaddr->addr.in));
            } else if (family == AF_INET6 &&
                /* prefer link-local addresses for IPv6 */
                IN6_IS_ADDR_LINKLOCAL(&((struct sockaddr_in6 *)ifa->ifa_addr)->sin6_addr)) {
                found_other = 1;
                iaddr->family = AF_INET6;
                memcpy(&iaddr->addr.in6, &((struct sockaddr_in6 *)ifa->ifa_addr)->sin6_addr,
                    sizeof(iaddr->addr.in6));
            }
        }
        /* find iface name by it's inet */
        if (type == 1 && ifa->ifa_addr->sa_family == iaddr->family) {
            if (iaddr->family == AF_INET &&
                memcmp(&((struct sockaddr_in *)ifa->ifa_addr)->sin_addr,
                    &iaddr->addr.in, sizeof(iaddr->addr.in)) == 0) {
                /* if we had zone, verify that we have inet on the iface */
                found_other = zone == NULL ? 1 : strcmp(iname, ifa->ifa_name) == 0;
            } else if (iaddr->family == AF_INET6 &&
                memcmp(&((struct sockaddr_in6 *)ifa->ifa_addr)->sin6_addr,
                    &iaddr->addr.in6, sizeof(iaddr->addr.in6)) == 0) {
                /* if we had zone, verify that we have inet on the iface */
                found_other = zone == NULL ? 1 : strcmp(iname, ifa->ifa_name) == 0;
            }
            if (found_other) {
                strlcpy(iname, ifa->ifa_name, sizeof(iname));
            }
        }
    }

    /* length of the base interface in case of alias (i.e. eth0:0) */
    size_t base = strcspn(iname, ":");

    /* find iface ether and index by it's name */
    for (ifa = ifaddr; ifa != NULL && found_other && !found_ether; ifa = ifa->ifa_next) {
        if (ifa->ifa_addr == NULL) {
            continue;
        }
        /* check for ether address, use base interface for aliases */
        if (ifa->ifa_addr->sa_family == AF_PACKET && strlen(ifa->ifa_name) == base &&
            strncmp(ifa->ifa_name, iname, base) == 0) {
            found_ether = 1;
            memcpy(eaddr, &((struct sockaddr_ll *)ifa->ifa_addr)->sll_addr,
                sizeof(*eaddr));
            if_index = ((struct sockaddr_ll *)ifa->ifa_addr)->sll_ifindex;
        }
    }

    freeifaddrs(ifaddr);
    debug("found_other: %d found_ether: %d\n", found_other, found_ether);
    return (found_other && found_ether) ? if_index : -1;
}

static unsigned
make_arp_packet(src_ether, src_inet, dst_inet, frame)
    const struct ether_addr *src_ether;
    const struct inaddr *src_inet;
    const struct inaddr *dst_inet;
    uint8_t *frame;
{
    assert(src_inet->family == AF_INET && dst_inet->family == AF_INET);
    /* broadcast ether address */
    struct ether_addr dst_ether = {{0xff, 0xff, 0xff, 0xff, 0xff, 0xff}};

    struct ether_header etherhdr = {0};
    struct arp arphdr = {0};

    /* ARP header */
    /* hardware type (16 bits) */
    arphdr.htype = htons(ARPHRD_ETHER);
    /* protocol type (16 bits) */
    arphdr.ptype = htons(ETH_P_IP);

    /* hardware address length (8 bits) */
    arphdr.hlen = sizeof(struct ether_addr);
    /* protocol address length (8 bits) */
    arphdr.plen = sizeof(struct in_addr);

    /* opcode */
    arphdr.opcode = htons(ARPOP_REQUEST);
    /* sender hardware address (48 bits) */
    memcpy(&arphdr.sender_ether, src_ether, sizeof(arphdr.sender_ether));
    /* sender protocol address (32 bits) */
    memcpy(&arphdr.sender_inet, &src_inet->addr.in, sizeof(arphdr.sender_inet));

    /* target hardware address (48 bits) */
    memset(&arphdr.target_ether, 0, sizeof(arphdr.target_ether));
    /* target protocol address (32 bits) */
    memcpy(&arphdr.target_inet, &dst_inet->addr.in, sizeof(arphdr.target_inet));

    /* fill out Ethernet frame header */
    memcpy(&etherhdr.ether_dhost, &dst_ether, sizeof(etherhdr.ether_dhost));
    memcpy(&etherhdr.ether_shost, src_ether, sizeof(etherhdr.ether_shost));
    etherhdr.ether_type = htons(ETH_P_ARP);

    /* copy Ethernet header */
    memcpy(frame, &etherhdr, ETH_HDRLEN);
    /* copy ARP header */
    memcpy(frame + ETH_HDRLEN, &arphdr, ARP_HDRLEN);

    return ETH_HDRLEN + ARP_HDRLEN;
}

static unsigned
make_icmp_packet(src_ether, src_inet, dst_ether, dst_inet, frame)
    const struct ether_addr *src_ether;
    const struct inaddr *src_inet;
    const struct ether_addr *dst_ether;
    const struct inaddr *dst_inet;
    uint8_t *frame;
{
    assert(src_inet->family == AF_INET && dst_inet->family == AF_INET);

    struct ether_header etherhdr = {0};
    struct ip iphdr = {0};
    struct icmp icmphdr = {0};

    unsigned ip_flags[4] = {0};
    uint8_t data[ICMP4_DATLEN] = {0};

    int i = 0;
    /* ICMP data */
    for (; i < packet_size; ++i) {
        data[i] = "LLNW"[i % 4];
    }

    /* IPv4 header length (4 bits): Number of 32-bit words in header = 5 */
    iphdr.ip_hl = IP4_HDRLEN / sizeof(uint32_t);
    /* Internet Protocol version (4 bits): IPv4 */
    iphdr.ip_v = 4;
    /* type of service (8 bits) */
    iphdr.ip_tos = 0;
    /* total length of datagram (16 bits): IP header + ICMP header + ICMP data */
    iphdr.ip_len = htons(IP4_HDRLEN + ICMP_HDRLEN + packet_size);
    /* ID sequence number (16 bits): unused, since single datagram */
    iphdr.ip_id = htons(0);

    /* flags, and fragmentation offset (3, 13 bits): 0 since single datagram */
    /* zero (1 bit) */
    ip_flags[0] = 0;
    /* do not fragment flag (1 bit) */
    ip_flags[1] = 1;
    /* more fragments following flag (1 bit) */
    ip_flags[2] = 0;
    /* fragmentation offset (13 bits) */
    ip_flags[3] = 0;

    iphdr.ip_off = htons(
        (ip_flags[0] << 15) +
        (ip_flags[1] << 14) +
        (ip_flags[2] << 13) +
        (ip_flags[3])
    );

    /* time-to-Live (8 bits) */
    iphdr.ip_ttl = IPDEFTTL;
    /* transport layer protocol (8 bits) */
    iphdr.ip_p = IPPROTO_ICMP;

    /* source IPv4 address (32 bits) */
    memcpy(&iphdr.ip_src, &src_inet->addr.in, sizeof(iphdr.ip_src));
    /* destination IPv4 address (32 bits) */
    memcpy(&iphdr.ip_dst, &dst_inet->addr.in, sizeof(iphdr.ip_dst));

    /* IPv4 header checksum (16 bits): set to 0 when calculating checksum */
    iphdr.ip_sum = 0;
    iphdr.ip_sum = checksum_final(checksum_init((uint16_t *)&iphdr, IP4_HDRLEN));

    /* ICMP header */
    /* message type (8 bits) */
    icmphdr.icmp_type = ICMP_ECHO;
    /* message code (8 bits) */
    icmphdr.icmp_code = 0;
    /* identifier (16 bits) */
    icmphdr.icmp_id = htons(ident);
    /* sequence number (16 bits): starts at 0 */
    icmphdr.icmp_seq = htons(seq);
    /* ICMP header checksum (16 bits) */
    icmphdr.icmp_cksum = 0;
    icmphdr.icmp_cksum = checksum_final(checksum_update(checksum_init(
        (uint16_t *)&icmphdr, ICMP_HDRLEN), (uint16_t *)data, packet_size));

    /* fill out Ethernet frame header */
    memcpy(&etherhdr.ether_dhost, dst_ether, sizeof(etherhdr.ether_dhost));
    memcpy(&etherhdr.ether_shost, src_ether, sizeof(etherhdr.ether_shost));
    etherhdr.ether_type = htons(ETH_P_IP);

    /* copy Ethernet header */
    memcpy(frame, &etherhdr, ETH_HDRLEN);
    /* copy IPv4 header */
    memcpy(frame + ETH_HDRLEN, &iphdr, IP4_HDRLEN);
    /* copy ICMP header */
    memcpy(frame + ETH_HDRLEN + IP4_HDRLEN, &icmphdr, ICMP_HDRLEN);
    /* copy ICMP data */
    memcpy(frame + ETH_HDRLEN + IP4_HDRLEN + ICMP_HDRLEN, data, packet_size);

    return ETH_HDRLEN + IP4_HDRLEN + ICMP_HDRLEN + packet_size;
}

static unsigned
make_ndp_packet(src_ether, src_inet, dst_inet, frame)
    const struct ether_addr *src_ether;
    const struct inaddr *src_inet;
    const struct inaddr *dst_inet;
    uint8_t *frame;
{
    assert(src_inet->family == AF_INET6 && dst_inet->family == AF_INET6);
    /* fill out multicast inet address */
    struct in6_addr dst_mcast = {{{0xFF, 0x02, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0x01, 0xFF}}};
    memcpy(&dst_mcast.__in6_u.__u6_addr8[13], &dst_inet->addr.in6.__in6_u.__u6_addr8[13], 3);
    /* fill out multicast ether address */
    struct ether_addr dst_ether = {{0x33, 0x33}};
    memcpy(&dst_ether.ether_addr_octet[2], &dst_mcast.__in6_u.__u6_addr32[3],
        sizeof(uint32_t));

    struct ether_header etherhdr = {0};
    struct ip6_hdr ip6hdr = {0};
    struct nd_neighbor_solicit ndphdr = {0};
    struct nd_opt_hdr opthdr = {0};

    /* IPv6 header */
    /* version (4 bits) = 6, traffic class (8 bits), flow label (20 bits) */
    ip6hdr.ip6_flow = htonl(0x60000000);
    ip6hdr.ip6_plen = htons(NDP_HDRLEN + sizeof(opthdr) + sizeof(*src_ether));
    ip6hdr.ip6_nxt = IPPROTO_ICMPV6;
    ip6hdr.ip6_hlim = 255;
    /* source IPv6 address (128 bits) */
    memcpy(&ip6hdr.ip6_src, &src_inet->addr, sizeof(ip6hdr.ip6_src));
    /* destination IPv6 address (128 bits) */
    memcpy(&ip6hdr.ip6_dst, &dst_mcast, sizeof(ip6hdr.ip6_dst));

    /* neighbor solicitation ICMPv6 header */
    ndphdr.nd_ns_type = ND_NEIGHBOR_SOLICIT;
    ndphdr.nd_ns_code = 0;
    ndphdr.nd_ns_cksum = 0;
    ndphdr.nd_ns_reserved = 0;
    /* neighbor solicitation target protocol address (128 bits) */
    memcpy(&ndphdr.nd_ns_target, &dst_inet->addr, sizeof(ndphdr.nd_ns_target));

    /* neighbor solicitation ICMPv6 option header */
    opthdr.nd_opt_type = ND_OPT_SOURCE_LINKADDR;
    opthdr.nd_opt_len = (sizeof(opthdr) + sizeof(*src_ether)) >> 3;

    /* calculate ICMPv6 checksum */
    ndphdr.nd_ns_cksum = checksum_final(checksum_update(checksum_update(checksum_update(
        checksum_update(checksum_update(checksum_update(checksum_init(
            (uint16_t *)&ip6hdr.ip6_src, sizeof(ip6hdr.ip6_src)),
            (uint16_t *)&ip6hdr.ip6_dst, sizeof(ip6hdr.ip6_dst)),
            (uint16_t *)&ip6hdr.ip6_plen, sizeof(ip6hdr.ip6_plen)),
            (uint16_t *)&((uint8_t []){0, 0, 0, ip6hdr.ip6_nxt}), 4),
            (uint16_t *)&ndphdr, NDP_HDRLEN),
            (uint16_t *)&opthdr, sizeof(opthdr)),
            (uint16_t *)src_ether, sizeof(*src_ether)));

    /* fill out Ethernet frame header */
    memcpy(&etherhdr.ether_dhost, &dst_ether, sizeof(etherhdr.ether_dhost));
    memcpy(&etherhdr.ether_shost, src_ether, sizeof(etherhdr.ether_shost));
    etherhdr.ether_type = htons(ETH_P_IPV6);

    /* copy Ethernet header */
    memcpy(frame, &etherhdr, ETH_HDRLEN);
    /* copy IPv6 header */
    memcpy(frame + ETH_HDRLEN, &ip6hdr, IP6_HDRLEN);
    /* copy neighbor solicitation ICMPv6 header */
    memcpy(frame + ETH_HDRLEN + IP6_HDRLEN, &ndphdr, NDP_HDRLEN);
    /* copy neighbor solicitation ICMPv6 option header */
    memcpy(frame + ETH_HDRLEN + IP6_HDRLEN + NDP_HDRLEN, &opthdr, sizeof(opthdr));
    /* copy neighbor solicitation ICMPv6 option souce ether address */
    memcpy(frame + ETH_HDRLEN + IP6_HDRLEN + NDP_HDRLEN + sizeof(opthdr), src_ether,
        sizeof(*src_ether));

    return ETH_HDRLEN + IP6_HDRLEN + NDP_HDRLEN + sizeof(opthdr) + sizeof(*src_ether);
}

static unsigned
make_icmp6_packet(src_ether, src_inet, dst_ether, dst_inet, frame)
    const struct ether_addr *src_ether;
    const struct inaddr *src_inet;
    const struct ether_addr *dst_ether;
    const struct inaddr *dst_inet;
    uint8_t *frame;
{
    assert(src_inet->family == AF_INET6 && dst_inet->family == AF_INET6);

    struct ether_header etherhdr = {0};
    struct ip6_hdr ip6hdr = {0};
    struct icmp6_hdr icmp6hdr = {0};

    uint8_t data[ICMP6_DATLEN] = {0};

    int i = 0;
    /* ICMP data */
    for (; i < packet_size; ++i) {
        data[i] = "LLNW"[i % 4];
    }

    /* IPv6 header */
    /* version (4 bits) = 6, traffic class (8 bits), flow label (20 bits) */
    ip6hdr.ip6_flow = htonl(0x60000000);
    ip6hdr.ip6_plen = htons(ICMP_HDRLEN + packet_size);
    ip6hdr.ip6_nxt = IPPROTO_ICMPV6;
    ip6hdr.ip6_hlim = IPDEFTTL;
    /* source IPv6 address (128 bits) */
    memcpy(&ip6hdr.ip6_src, &src_inet->addr, sizeof(ip6hdr.ip6_src));
    /* destination IPv6 address (128 bits) */
    memcpy(&ip6hdr.ip6_dst, &dst_inet->addr, sizeof(ip6hdr.ip6_dst));

    /* ICMPv6 header */
    icmp6hdr.icmp6_type = ICMP6_ECHO_REQUEST;
    icmp6hdr.icmp6_code = 0;
    icmp6hdr.icmp6_cksum = 0;
    icmp6hdr.icmp6_id = htons(ident);
    icmp6hdr.icmp6_seq = htons(seq);

    /* calculate ICMPv6 checksum */
    icmp6hdr.icmp6_cksum = checksum_final(checksum_update(checksum_update(
        checksum_update(checksum_update(checksum_update(checksum_init(
            (uint16_t *)&ip6hdr.ip6_src, sizeof(ip6hdr.ip6_src)),
            (uint16_t *)&ip6hdr.ip6_dst, sizeof(ip6hdr.ip6_dst)),
            (uint16_t *)&ip6hdr.ip6_plen, sizeof(ip6hdr.ip6_plen)),
            (uint16_t *)&((uint8_t []){0, 0, 0, ip6hdr.ip6_nxt}), 4),
            (uint16_t *)&icmp6hdr, ICMP_HDRLEN),
            (uint16_t *)data, packet_size));

    /* fill out Ethernet frame header */
    memcpy(&etherhdr.ether_dhost, dst_ether, sizeof(etherhdr.ether_dhost));
    memcpy(&etherhdr.ether_shost, src_ether, sizeof(etherhdr.ether_shost));
    etherhdr.ether_type = htons(ETH_P_IPV6);

    /* copy Ethernet header */
    memcpy(frame, &etherhdr, ETH_HDRLEN);
    /* copy IPv6 header */
    memcpy(frame + ETH_HDRLEN, &ip6hdr, IP6_HDRLEN);
    /* copy ICMPv6 header */
    memcpy(frame + ETH_HDRLEN + IP6_HDRLEN, &icmp6hdr, ICMP_HDRLEN);
    /* copy ICMPv6 data */
    memcpy(frame + ETH_HDRLEN + IP6_HDRLEN + ICMP_HDRLEN, data, packet_size);

    return ETH_HDRLEN + IP6_HDRLEN + ICMP_HDRLEN + packet_size;
}

static void
cached_ether_lookup(iaddr, result)
    const struct inaddr *iaddr;
    struct ether_addr *result;
{
    int sockfd = socket(AF_NETLINK, SOCK_RAW, NETLINK_ROUTE);

    if (sockfd < 0) {
        error_exit(EXIT_INTERNAL_ERROR, errno, "AF_NETLINK socket() failed");
    }

    struct sockaddr_nl sa = {0};
    sa.nl_family = AF_NETLINK;

    if (bind(sockfd, (struct sockaddr *)&sa, sizeof(sa)) < 0) {
        error_exit(EXIT_INTERNAL_ERROR, errno, "AF_NETLINK bind() failed");
    }

    struct nlmsghdr nlh = {0};
    struct rtgenmsg g = {0};

    struct iovec iov[2] = {
        { .iov_base = &nlh, .iov_len = sizeof(nlh) },
        { .iov_base = &g, .iov_len = sizeof(g) }
    };

    struct msghdr msg = {
        .msg_name = &sa,
        .msg_namelen = sizeof(sa),
        .msg_iov = iov,
        .msg_iovlen = 2,
        .msg_control = NULL,
        .msg_controllen = 0,
        .msg_flags = 0
    };

    nlh.nlmsg_len = NLMSG_LENGTH(sizeof(g));
    nlh.nlmsg_type = RTM_GETNEIGH;
    nlh.nlmsg_flags = NLM_F_ROOT | NLM_F_MATCH | NLM_F_REQUEST;
    nlh.nlmsg_pid = 0;
    nlh.nlmsg_seq = 1;

    /* get this family addresses only */
    g.rtgen_family = iaddr->family;

    debug("trying to get cache table for family: %u\n", iaddr->family);

    ssize_t bytes = -1;
    /* send the request */
    while ((bytes = sendmsg(sockfd, &msg, 0)) < 0 && errno == EINTR);
    debug("sent %zd bytes to netlik\n", bytes);

    /* prepare to receive the response */
    char buf[NETLINK_RCV_BUF];
    iov[0].iov_base = buf;
    iov[0].iov_len = sizeof(buf);
    msg.msg_iovlen = 1;

    int done = 0;

    while (!done) {
        while ((bytes = recvmsg(sockfd, &msg, 0)) < 0 && errno == EINTR);
        debug("received %zd bytes from netlik\n", bytes);

        if (!bytes) {
            debug("EOF from netlink\n");
            break;
        }
        if (msg.msg_flags & MSG_TRUNC) {
            debug("truncated response from netlink\n");
        }

        struct nlmsghdr *nh = (struct nlmsghdr *)buf;

        for (; !done && NLMSG_OK(nh, bytes); nh = NLMSG_NEXT(nh, bytes)) {
            /* the end of multipart message */
            if (nh->nlmsg_type == NLMSG_DONE) {
                done = 1;
                break;
            }

            if (nh->nlmsg_type == NLMSG_ERROR) {
                debug("error response from netlink\n");
                continue;
            }

            struct ndmsg *r = NLMSG_DATA(nh);
            struct rtattr *rta = NDA_RTA(r);
            int rta_len = nh->nlmsg_len - NLMSG_LENGTH(sizeof(*r));

            static char saddr[INET6_ADDRSTRLEN];

            for (; RTA_OK(rta, rta_len); rta = RTA_NEXT(rta, rta_len)) {
                switch (rta->rta_type) {
                    case NDA_DST:
                        done = done ? done : !inaddr_cmp(iaddr, RTA_DATA(rta));
                        inet_ntop(r->ndm_family, RTA_DATA(rta), saddr, sizeof(saddr));
                        debug("rta DST address %s, %d\n", saddr, done);
                        break;
                    case NDA_LLADDR:
                        if (done == 1) {
                            /* now we know the server ether address */
                            memcpy(result, RTA_DATA(rta), sizeof(*result));
                            done = 2;
                        }
                        debug("rta LL address: %s, %d\n", ether_ntoa(RTA_DATA(rta)),
                            done);
                        break;
                    default:
                        break;
                }
            }
        }
    }
    close(sockfd);

    if (done != 2) {
        error_exit(EXIT_INTERNAL_ERROR, 0, "there is no cached MAC address for %s",
            inaddr_ntop(iaddr));
    }
}

static void
ether_lookup(sockfd, src_ether, src_inet, iaddr, result)
    int sockfd;
    const struct ether_addr *src_ether;
    const struct inaddr *src_inet;
    const struct inaddr *iaddr;
    struct ether_addr *result;
{
    uint8_t frame[ETH_FRMLEN] = {0};

    if (iaddr->family == AF_INET) {
        struct arp *arphdr = (struct arp *)&frame[ETH_HDRLEN];

        quiet_print("ARP REQUEST %s\n", inaddr_ntop(iaddr));

        /* prepare for ARP reply */
        set_filter_and_drain(sockfd, daemon_mode ? NULL : &arpfilter);

        /* make and send ethernet frame to socket */
        unsigned size = make_arp_packet(src_ether, src_inet, iaddr, frame);
        safe_send(sockfd, frame, size);

        /* wait for ARP reply from *our* server */
        do {
            ssize_t bytes = safe_recv(sockfd, frame, ETH_FRMLEN);
            if (bytes < ETH_HDRLEN + ARP_HDRLEN) {
                /* insufficient frame size */
                debug("received frame of insufficient size %ld\n", bytes);
                continue;
            }
        } while (inaddr_cmp(iaddr, &arphdr->sender_inet) != 0);

        /* now we know the server ether address */
        memcpy(result, &arphdr->sender_ether, sizeof(*result));
    } else if (iaddr->family == AF_INET6) {
        struct ether_header *etherhdr = (struct ether_header *)frame;
        struct ip6_hdr *ip6hdr = (struct ip6_hdr *)&frame[ETH_HDRLEN];
        struct nd_neighbor_advert *ndphdr =
            (struct nd_neighbor_advert *)&frame[ETH_HDRLEN + IP6_HDRLEN];

        quiet_print("NDP REQUEST %s\n", inaddr_ntop(iaddr));

        /* prepare for advertisement reply */
        set_filter_and_drain(sockfd, daemon_mode ? NULL : &ndpnafilter);

        /* make and send ethernet frame to socket */
        unsigned size = make_ndp_packet(src_ether, src_inet, iaddr, frame);

        safe_send(sockfd, frame, size);

        /* wait for advertisement from *our* server */
        do {
            ssize_t bytes = safe_recv(sockfd, frame, ETH_FRMLEN);
            if (bytes < ETH_HDRLEN + IP6_HDRLEN + NDP_HDRLEN) {
                /* insufficient frame size */
                debug("received frame of insufficient size %ld\n", bytes);
                continue;
            }
            /* verify ICMPv6 checksum */
            uint16_t cksum = ndphdr->nd_na_cksum;
            ndphdr->nd_na_cksum = 0;
            if (cksum != checksum_final(checksum_update(checksum_update(
                checksum_update(checksum_update(checksum_init(
                (uint16_t *)&ip6hdr->ip6_src, sizeof(ip6hdr->ip6_src)),
                (uint16_t *)&ip6hdr->ip6_dst, sizeof(ip6hdr->ip6_dst)),
                (uint16_t *)&ip6hdr->ip6_plen, sizeof(ip6hdr->ip6_plen)),
                (uint16_t *)&((uint8_t []){0, 0, 0, ip6hdr->ip6_nxt}), 4),
                (uint16_t *)ndphdr, bytes - ETH_HDRLEN - IP6_HDRLEN))) {
                /* wrong ICMPv6 header checksum */
                debug("received frame with wrong ICMPv6 header checksum\n");
                continue;
            }
        } while (inaddr_cmp(iaddr, &ndphdr->nd_na_target) != 0);

        /* now we know the server ether address */
        memcpy(result, &etherhdr->ether_shost, sizeof(*result));
    }
}

static void
ping_init(int sockfd, const struct inaddr *dst_inet, const struct ether_addr *dst_ether)
{
    if (dst_inet->family == AF_INET) {
        quiet_print("PING %s (%s) %u(%u) bytes of data.\n",
            inaddr_ntop(dst_inet), ether_ntoa(dst_ether), packet_size,
            packet_size + IP4_HDRLEN + ICMP_HDRLEN);

        /* patch in our ident */
        icmpfilter.filter[13].k = ident;
        /* prepare for ICMP reply */
        set_filter_and_drain(sockfd, daemon_mode ? NULL : &icmpfilter);
    } else if (dst_inet->family == AF_INET6) {
        quiet_print("PING6 %s (%s) %u(%u) bytes of data.\n",
            inaddr_ntop(dst_inet), ether_ntoa(dst_ether), packet_size,
            packet_size + IP6_HDRLEN + ICMP_HDRLEN);

        /* patch in our ident */
        icmp6filter.filter[7].k = ident;
        /* prepare for ICMPv6 reply */
        set_filter_and_drain(sockfd, daemon_mode ? NULL : &icmp6filter);
    }
}

static void
ping_send(sockfd, src_ether, src_inet, dst_ether, dst_inet)
    int sockfd;
    const struct ether_addr *src_ether;
    const struct inaddr *src_inet;
    const struct ether_addr *dst_ether;
    const struct inaddr *dst_inet;
{
    assert(src_inet->family == dst_inet->family);
    uint8_t frame[ETH_FRMLEN] = {0};
    unsigned size = 0;
    if (src_inet->family == AF_INET) {
        size = make_icmp_packet(src_ether, src_inet, dst_ether, dst_inet, frame);
    } else if (src_inet->family == AF_INET6) {
        size = make_icmp6_packet(src_ether, src_inet, dst_ether, dst_inet, frame);
    }
    /* send ethernet frame to socket */
    safe_send(sockfd, frame, size);
}

static void
ping_recv(int sockfd, const struct inaddr *from)
{
    uint8_t frame[ETH_FRMLEN] = {0};
    if (from->family == AF_INET) {
        struct ip *iphdr = (struct ip *)&frame[ETH_HDRLEN];
        struct icmp *icmphdr = (struct icmp *)&frame[ETH_HDRLEN + IP4_HDRLEN];

        ssize_t bytes = safe_recv(sockfd, frame, ETH_FRMLEN);
        if (bytes < ETH_HDRLEN + IP4_HDRLEN + ICMP_HDRLEN + packet_size) {
            /* insufficient frame size */
            debug("received frame of insufficient size %ld\n", bytes);
            return;
        }
        uint16_t cksum = iphdr->ip_sum;
        iphdr->ip_sum = 0;
        /* verify IPv4 checksum */
        if (cksum != checksum_final(checksum_init((uint16_t *)iphdr, IP4_HDRLEN))) {
            /* wrong IPv4 header checksum */
            debug("received frame with wrong IPv4 header checksum\n");
            return;
        }
        if (ntohs(iphdr->ip_len) != IP4_HDRLEN + ICMP_HDRLEN + packet_size) {
            /* wrong IPv4 header packet size */
            debug("received frame with wrong IPv4 packet size\n");
            return;
        }
        cksum = icmphdr->icmp_cksum;
        icmphdr->icmp_cksum = 0;
        /* verify ICMPv4 checksum */
        if (cksum != checksum_final(checksum_init(
            (uint16_t *)icmphdr, ICMP_HDRLEN + packet_size))) {
            /* wrong ICMP header checksum */
            debug("received frame with wrong ICMP header checksum\n");
            return;
        }
        /* check if ICMP reply is from *our* ident && vip */
        if (icmphdr->icmp_id == htons(ident) && inaddr_cmp(from, &iphdr->ip_src) == 0) {
            quiet_print("%u bytes from %s: icmp_seq=%u ttl=%u\n",
                ntohs(iphdr->ip_len) - IP4_HDRLEN, inaddr_ntop(from),
                ntohs(icmphdr->icmp_seq), iphdr->ip_ttl);
            /* record received ack (echo reply) seq */
            CBIT_SET(acks, ntohs(icmphdr->icmp_seq));
        }
    } else if (from->family == AF_INET6) {
        struct ip6_hdr *ip6hdr = (struct ip6_hdr *)&frame[ETH_HDRLEN];
        struct icmp6_hdr *icmp6hdr = (struct icmp6_hdr *)&frame[ETH_HDRLEN + IP6_HDRLEN];

        ssize_t bytes = safe_recv(sockfd, frame, ETH_FRMLEN);
        if (bytes < ETH_HDRLEN + IP6_HDRLEN + ICMP_HDRLEN + packet_size) {
            /* insufficient frame size */
            debug("received frame of insufficient size %ld\n", bytes);
            return;
        }
        if (ntohs(ip6hdr->ip6_plen) != ICMP_HDRLEN + packet_size) {
            /* wrong IPv6 header packet size */
            debug("received frame with wrong IPv6 packet size\n");
            return;
        }
        /* verify ICMPv6 checksum */
        uint16_t cksum = icmp6hdr->icmp6_cksum;
        icmp6hdr->icmp6_cksum = 0;
        if (cksum != checksum_final(checksum_update(checksum_update(
            checksum_update(checksum_update(checksum_init(
            (uint16_t *)&ip6hdr->ip6_src, sizeof(ip6hdr->ip6_src)),
            (uint16_t *)&ip6hdr->ip6_dst, sizeof(ip6hdr->ip6_dst)),
            (uint16_t *)&ip6hdr->ip6_plen, sizeof(ip6hdr->ip6_plen)),
            (uint16_t *)&((uint8_t []){0, 0, 0, ip6hdr->ip6_nxt}), 4),
            (uint16_t *)icmp6hdr, ICMP_HDRLEN + packet_size))) {
            /* wrong ICMPv6 header checksum */
            debug("received frame with wrong ICMPv6 header checksum\n");
            return;
        }
        /* check if ICMP reply is from *our* ident && vip */
        if (icmp6hdr->icmp6_id == htons(ident) && inaddr_cmp(from, &ip6hdr->ip6_src) == 0) {
            quiet_print("%u bytes from %s: icmp_seq=%u ttl=%u\n",
                ntohs(ip6hdr->ip6_plen), inaddr_ntop(from),
                ntohs(icmp6hdr->icmp6_seq), ip6hdr->ip6_hops);
            /* record received ack (echo reply) seq */
            CBIT_SET(acks, ntohs(icmp6hdr->icmp6_seq));
        }
    }
}

int
main(int argc, char *argv[])
{
    /* socket descriptor */
    int sockfd = -1;

    /* outgoing iface inet, ether addresses and index */
    struct inaddr iface_inet = {0};
    struct ether_addr iface_ether = {0};
    int iface_index = -1;

    /* vip inet address */
    struct inaddr vip_inet = {0};

    /* server inet and ether addresses */
    struct inaddr server_inet = {0};
    struct ether_addr server_ether = {0};

    /* parse command line arguments */
    parse_arguments(argc, argv);

    /* schedule deadline alarm signal */
    if (wait != 0) {
        signal(SIGALRM, sighandler);
        alarm(wait);
    }
    /* set user interrupt hanlder */
    signal(SIGINT, sighandler);

    /* use pid as ident for icmp packets */
    ident = getpid();
    debug("my ident: %u\n", ident);

    /* parse server inet address */
    if (inaddr_pton(server, &server_inet) != 1) {
        error_exit(EXIT_INTERNAL_ERROR, 0, "inaddr_pton() on '%s' failed", server);
    }
    /* parse vip inet address */
    if (inaddr_pton(vip, &vip_inet) != 1) {
        error_exit(EXIT_INTERNAL_ERROR, 0, "inaddr_pton() on '%s' failed", vip);
    }
    /* find interface inet and ether addresses from interface name/address */
    if ((iface_index = if_lookup(&iface_inet, &iface_ether, vip_inet.family)) < 0) {
        error_exit(EXIT_INTERNAL_ERROR, 0, "if_lookup() on '%s' failed", iface);
    }

    /* we cannot ping IPv4 address from IPv6 address and vice versa */
    if (server_inet.family != vip_inet.family) {
        error_exit(EXIT_INTERNAL_ERROR, 0, "server and vip addresses must be of the same family");
    }
    if (iface_inet.family != vip_inet.family) {
        error_exit(EXIT_INTERNAL_ERROR, 0, "source and vip addresses must be of the same family");
    }

    /* make sure we don't need to fragment */
    if (vip_inet.family == AF_INET && packet_size > ICMP4_DATLEN) {
        error_exit(EXIT_INTERNAL_ERROR, 0, "maximum ICMP packet size is %u", ICMP4_DATLEN);
    }
    if (vip_inet.family == AF_INET6 && packet_size > ICMP6_DATLEN) {
        error_exit(EXIT_INTERNAL_ERROR, 0, "maximum ICMP6 packet size is %u", ICMP6_DATLEN);
    }

    debug("interface %s is at index %i\n", iface, iface_index);
    debug("interface %s ether address is %s\n", iface, ether_ntoa(&iface_ether));
    debug("interface %s inet address is %s\n", iface, inaddr_ntop(&iface_inet));

    if (daemon_mode) {
        /* connect to daemon and start it if necessary */
        while ((sockfd = connect_to_daemon(iface_index)) < 0) {
            start_daemon(argc, argv);
            /* sleep before retrying */
            usleep(DAEMON_WAIT_SLEEP);
        }
    } else {
        /* create raw socket bound to iface */
        sockfd = create_raw_socket(iface_index);
    }

    /* lookup ether address of server inet address */
    if (cached_ether) {
        cached_ether_lookup(&server_inet, &server_ether);
    } else {
        ether_lookup(sockfd, &iface_ether, &iface_inet, &server_inet, &server_ether);
    }

    /* prepare to ping vip inet at server ether */
    ping_init(sockfd, &vip_inet, &server_ether);

    struct timeval tmo = {0};
    int done = 0;

    /* main loop */
    while (!done) {
        if (seq != packet_count) {
            seq++;
            ping_send(sockfd, &iface_ether, &iface_inet, &server_ether, &vip_inet);
        } else if (wait == 0) {
            break;
        }

        /* wait for ICMP replies for send_interval */
        tmo.tv_sec = send_interval;
        tmo.tv_usec = 0;

        /* receive loop */
        while (!(done = seq == packet_count && seq == count_acks())) {
            /* on Linux, select updates tmo with time not slept */
            if (safe_select(sockfd, &tmo)) {
                ping_recv(sockfd, &vip_inet);
            } else {
                /* we waited for send_interval seconds */
                break;
            }
        }
    }

    /* farewell my raw friend */
    close(sockfd);

    return summary();
}
