#!/usr/bin/env bash
PATH="/sbin:/bin:/usr/bin:/usr/sbin:/usr/local/sbin:/usr/local/bin"
shopt -s nocaseglob

#script to look for .new configs in /etc/keepalived/config_containers and reconfigure the keepalived process.
#Will also run some simple tests and roll back configs if health checks fail, to help mitigating bad configs.

#setting default values if not specified
lock_minutes=60
restore_seconds=15

usage(){
  echo "usage: $0 -r [seconds] -l [minutes]"
  echo "Required:"
  echo "  -r      if error detected, roll back configs after N seconds (default: 15)"
  echo "  -l      if error detected, lock reconfigs for N minutes (default: 60)"
  echo "  -R      Immediately restore from backup config server files"
  echo "Options:"
  echo "  -s      log warning and error messages to syslog"
  echo "Example:"
  echo "  $0 -r 30"
  echo "  $0 -r 30 -l 120"
  echo "  $0 -R"
  exit 1
}

while getopts ":r:l:Rs" opt; do
  case ${opt} in
    r)
      restore_seconds=${OPTARG}
      ;;
    l)
      lock_minutes=${OPTARG}
      ;;
    R)
      rollback=1
      ;;
    s)
      log_to_syslog=1
      ;;
    *)
      usage
      ;;
  esac
done
shift $(($OPTIND - 1))

# the script name from ${0} wasn't getting passed to syslog properly, so specifying here
script_name="slb_reconfig.sh"

error_exit(){
  if [ ${log_to_syslog} ]; then
    logger -p daemon.err -s -i -t ${script_name} "[ERROR] ${1:-""}" 1>&2
  else
    echo "${script_name}: [ERROR] ${1:-""}" 1>&2
  fi
  exit ${2:-"1"}
}

warn(){
  if [ ${log_to_syslog} ]; then
    logger -p daemon.warn -s -i -t ${script_name} "[WARN] ${1:-""}" 1>&2
  else
    echo "${script_name}: [WARN] ${1:-""}" 1>&2
  fi
}

error_check(){
  lockfile="/tmp/slb_reconfig.lock"
  trouble="trouble.lock"

  if [ -f ${lockfile} ]; then
    error_exit "reconfig is currently running" "1"
  fi

  #entries in overrides is a temp operation, clear it each time we reconfig if it's older than 1 hour
  overrides_file="/etc/keepalived/config_containers/overrides.yml"
  if [ -s ${overrides_file} ]; then
    old_override=`find ${overrides_file} -mmin +60`
    if [ ${old_override} ]; then
      rm -f ${old_override} || error_exit "Could not remove ${overrides_file}, try manually removing" "1"
      touch ${overrides_file} || error_exit "Could not create new ${overrides_file}" "1"
    fi
  fi

  #trouble_lock gets set if file is older than the specified lock_minutes
  trouble_lock=`find /tmp -mmin +"${lock_minutes}" -name "${trouble}"`

  if [ -f /tmp/${trouble} ]; then
    if [ "${trouble_lock}" ]; then
      rm -f ${trouble_lock} || error_exit "Could not remove /tmp/${trouble}, try manually removing" "1"
    else
      error_exit "/tmp/$trouble was found and is less than $lock_minutes, will not reconfigure" "1"
    fi
  fi

  touch ${lockfile} || error_exit "Could not create lock file ${lockfile}, line: $LINENO" "1"
  trap "rm -f ${lockfile}; exit" TERM INT EXIT
}

conf_dir="/etc/keepalived/config_containers/"

get_and_backup_configs(){
  #Look for .new config files and back them up if needed
  declare -gA config_files

  for config in `find ${conf_dir} -name "*.new" | cut -d . -f 1` ; do
    if [ ! -f ${config}.conf ]; then
      new_install=1
      config_files[${config}.conf]=${config}.new
    else
      config_files[${config}.conf]=${config}.new
      cp -fp ${config}.conf ${config}.backup || error_exit "Could not backup ${config}.conf" "1"
    fi
  done
}

virt_conf="${conf_dir}virtual_server.conf"
vrrp_conf="${conf_dir}vrrp_instance.conf"

put_new_files_inplace(){
  #Diff the files and only mv into place if they are different.  This avoids unnecessary reconfigs if the files are
  #the same, but servers are listed in a different order.
  for new_conf in ${!config_files[@]}; do
    if [ ! $new_install ]; then
      diff -q <(sort ${config_files[$new_conf]}) <(sort $new_conf) > /dev/null
      if [ $? -ne 0 ]; then
        mv -u ${config_files[$new_conf]} $new_conf || error_exit "Could not move ${config_files[$new_conf]} to $new_conf" "1"
        needs_reload=1
        if [ ${new_conf} = ${virt_conf} ] || [ ${new_conf} = ${vrrp_conf} ]; then
          needs_bgp=1
        fi
      else
        rm ${config_files[$new_conf]}
      fi
    else
      mv -u ${config_files[$new_conf]} $new_conf || error_exit "Could not move ${config_files[$new_conf]} to $new_conf" "1"
      needs_reload=1
      needs_bgp=1
    fi
  done
}

revert_configs(){
  #KeepaliveD will run this script with the rollback arg if quorum was lost and a reconfig was done
  #within the last 10 minutes.

  local min_time=10
  needs_reload=1

  for backup_conf in `find ${conf_dir} -mmin -${min_time} -name "*.backup" | cut -d . -f 1`; do
    cp -f ${backup_conf}.backup ${backup_conf}.conf || error_exit "Could not move ${backup_conf}.backup to ${backup_conf}.conf" "1"
  done

  #We had to revert, lock down reconfigs for the next $lock_minutes default is 60 minutes
  if [ ${lock_minutes} ]; then
    touch /tmp/${trouble} || error_exit "Could not create trouble file /tmp/${trouble}, line: $LINENO" "1"
  fi
}

reload_configs(){
  #reload configs only if keepalived is running and we actually changed something 
  if [ ${needs_bgp} ]; then
    #/etc/quagga/mk_bgp_conf.pl || error_exit "problem running /etc/quagga/mk_bgp_conf.pl" "1"
    /etc/quagga/gen_bgp_config || error_exit "problem running /etc/quagga/gen_bgp_config" "1"
  fi

  if [ ${needs_reload} ]; then
    is_running=`pgrep -o -u root -f /usr/sbin/keepalived`
    if [ ${is_running} ]; then
      /etc/init.d/keepalived force-reload || error_exit "keepalived reload failed line: $LINENO" "1"
    else
      error_exit "KeepaliveD is not running line: $LINENO" "1"
    fi
  fi
}


if [ ${rollback} ]; then
  error_check
  revert_configs
  reload_configs
else
  error_check
  get_and_backup_configs
  put_new_files_inplace
  reload_configs
fi
