#!/bin/bash

if [ $UID -ne 0 ]; then
  echo "Must be root to run the script"
  exit 0
fi

usage(){
    echo "usage: $0 [options]"
    echo "Options:"
    echo "  -d     dryrun output instead of sending to TSDB  "
    echo "Example:"
    echo "  $0 -d"
    exit 1
}

while getopts ":d" opt; do
  case ${opt} in
      d)
        dryrun=1
        ;;
      *)
        usage
        ;;
    esac
done

LAST_PROCSTAT_FILE="/run/procstat.lastrun"

TSDB_SERVER="tsdb-data.lldns.net"
TSDB_PORT="4242"
TSDB_DATA=""


# Meta data for SLB
# =============================

# Distribution - "Ubuntu_16.04.1_LTS"
DIST_DESC=$(grep DISTRIB_DESCRIPTION /etc/lsb-release | cut -d'=' -f2 | tr -d '"' | tr ' ' '_')

# Kernel Version - "4.4.0-98-generic"
KERN_VER=$(uname -r)

# CPU Model - Intel(R)_Core(TM)_i5-2400_CPU_@_3.10GHz
CPU_MODEL=$(grep -m1 "model name" /proc/cpuinfo | sed 's/: /:/g' | cut -d':' -f2 | sed 's/(R)//g;s/(TM)//g;s/@ //g;s/CPU //g' | tr ' ' '_' )

# Chassis Type - SYS-1018R-WC0R
CHASSIS_TYPE=$(dmidecode | grep -A 5 "System Information" | grep 'Product Name:' | awk '{print $3}')

# Hostname - slb03.lab.llnw.net
FQDN=$(uname -n)

# Pop - lab
POP=$(echo "$FQDN" | cut -d'.' -f2 | egrep -o '[a-z]+')


get_cpu_util() {
    # Collect Procstat for each CPU core and the collection time
    new_procstat=$(grep -E  '^cpu[0-9]+' /proc/stat)
    now=$(date +%s)

    # Check for old procstat file
    if [ -f "$LAST_PROCSTAT_FILE" ]; then
        # Read current file line by line
        while read -r now_core now_user now_nice now_system now_idle now_iowait now_irq now_softirq now_steal now_guest now_guest_nice; do
            
            # Grab old procstat line
            read -r last_core last_user last_nice last_system last_idle last_iowait last_irq last_softirq last_steal last_guest last_guest_nice <<< $(grep -w $now_core $LAST_PROCSTAT_FILE)

            # Define cpu time differances
            declare -A procstat_diff
            procstat_diff[user]=$((now_user-last_user))
            procstat_diff[nice]=$((now_nice-last_nice))
            procstat_diff[system]=$((now_system-last_system))
            procstat_diff[idle]=$((now_idle-last_idle))
            procstat_diff[iowait]=$((now_iowait-last_iowait))
            procstat_diff[irq]=$((now_irq-last_irq))
            procstat_diff[softirq]=$((now_softirq-last_softirq))
            procstat_diff[steal]=$((now_steal-last_steal))


            # Sum all cpu counters to get total CPU time
            total_cpu=0
            for k in "${!procstat_diff[@]}"; do 
                # If any numbers are nugative (file rolled over) we need
                # to wait for the next run of the script
                if [[ "${procstat_diff[$k]}" -lt "0" ]]; then
                    # Write current file and exit
                    echo "$new_procstat" > $LAST_PROCSTAT_FILE
                    exit
                fi

                total_cpu=$((total_cpu+procstat_diff[$k]))
            done


            # Get percents for each key
            for k in "${!procstat_diff[@]}"; do
                #used=$(echo "scale=2;(${procstat_diff[$k]}/$total_cpu)*100" | bc)
                used=$(echo "${procstat_diff[$k]}" "$total_cpu" |   awk '{printf "%.2f\n", ($1/$2)*100}')

                # Print TSDB items
                TSDB_DATA+="put sys.cpu.${k} ${now} ${used} core=${now_core} dist_name=${DIST_DESC} kern_ver=${KERN_VER} cpu_model=${CPU_MODEL} chassis_type=${CHASSIS_TYPE} fqdn=${FQDN} pop=${POP}\n"
            done    


        done <<< "$new_procstat"

        # write this run as the last run
        echo "$new_procstat" > $LAST_PROCSTAT_FILE
    else
        # File missing, we write the file and exit
        echo "$new_procstat" > $LAST_PROCSTAT_FILE
    fi

}

get_cpu_util

if [ "${dryrun}" ]; then
    echo -ne "$TSDB_DATA"
else
    echo -ne "$TSDB_DATA" | netcat tsdb-data.lldns.net 4242
fi

