#!/bin/bash

#Script to health check the backend machines.  First check to see if we can do an
#https check, if that is successful, we do the dsr ping check.  If both are successful
#we return a value to keepalived that determines the servers weight.

machine=${1}
vip=${2}
option=${3}

#exit 1 on place holder entries in the server configs
if [ ${machine} = "192.168.255.255" ] || [ ${machine} = "fc00:aaaa::1" ]; then
  exit 1
fi

debug=`global_conf_get.pl macping_debug`
dlogger(){
  if [ "$debug" = "on" ] ; then
    logger $@
  fi
}

verify_v4(){
  if [[ ${1} =~ ^[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3} ]]; then
    return 0
  else
    return 1
  fi
}

expand_ipv6(){
  #ipv6 entries are stored fully uncompressed in /proc/net/ip_vs so we need
  #to expand to the full 32 char length before looking up
  local hextets=( ${1//::/ } )
  local hextet_count=${#hextets[@]}
  local split_hextets=( ${1//:/ } )
  local split_count=${#split_hextets[@]}
  local deficit=$(( 8 - split_count ))
  local temp_pad=${#hextets[0]}
  local split_pad=${temp_pad//:/ }
  local start_pad=${#split_pad}
  local counter=0

  for hex in ${split_hextets[@]}; do
    length=$(( 4 - ${#hex} ))
    if [ ${length} -eq 0 ]; then
      nc_addr+=("${hex}")
      (( counter++ ))
    elif [ ${length} -eq 1 ]; then
      nc_addr+=("0${hex}")
      (( counter++ ))
    elif [ ${length} -eq 2 ]; then
      nc_addr+=("00${hex}")
      (( counter++ ))
    elif [ ${length} -eq 3 ]; then
      nc_addr+=("000${hex}")
      (( counter++ ))
    fi

    if [ $counter -gt $start_pad ]; then
      while [ $deficit -ne 0 ]; do
        nc_addr+=( "0000" )
        (( deficit-- ))
      done
    fi
  done
  temp_addr=${nc_addr[@]}
  expanded_addr=${temp_addr// /:}
}

#Weights are stored in this file by gen_slb_configs.pl
get_weights(){
  local file="/etc/keepalived/server_weights/${machine}"
  if [ -f ${file} ]; then
    current_weight=$(awk '/^current/{print $2}' $file)
    default_weight=$(awk '/^default/{print $2}' $file)
  fi
  if [ -z ${current_weight} ]; then
    current_weight=100
  fi
  if [ -z ${default_weight} ]; then
    default_weight=100
  fi
}

save_weights(){
  local file="/etc/keepalived/server_weights/${machine}"
  cat << EOF > ${file}
---
current: ${new_weight}
default: ${default_weight}
EOF
}

http_check(){
  #use curl to capture connect time, 1st byte response time, http response code.
  #If times are abnormally long, we decrement the server weight to give it less traffic.

  local options=( -x [${1}]:8000 -I --head --http1.1 --max-time 6 -s -o /dev/null )
  local curl_output=( `curl ${options[*]} -H 'Connection: close' -H 'Proxy-connection: close' -w '%{http_code} %{time_connect} %{time_pretransfer}' http://test.vo.llnwd.net/o1/geo-health.txt` )
  local http_code=${curl_output[0]}
  local time_connect=${curl_output[1]}
  local first_byte=${curl_output[2]}

  #Read current and default weights
  get_weights
  local min_weight=$(( default_weight / 2  ))

  #Machine is busy if time_connect gt 1 second
  #EP is busy if first_byte gt 1 second
  if [ ${time_connect:0:1} -ge 2 ] || [ ${first_byte:0:1} -ge 2 ]; then
    if [ ${current_weight} -le ${min_weight} ]; then
      #at the minimum or lower, set to minimum weight
      new_weight=${min_weight}
    else
      #adjust down 3% of _current_ weight,
      local delta_min=1
      local delta_calc=$(( 3 * current_weight / 100  ))
      local delta=$(( delta_min > delta_calc ? delta_min : delta_calc ))
      new_weight=$(( current_weight - delta ))
      dlogger -t $0 "${machine} adjust weight down by ${delta} to {new_weight}"
    fi
  else
    if [ ${current_weight} -lt ${min_weight} ]; then
      #we are below minimum, set to minimum
      new_weight=${min_weight}
    elif [ ${current_weight} -ge ${default_weight} ]; then
      #at default weight or higher, set to default
      new_weight=${default_weight}
    else
      #adjust up 3% of _default_ weight (to adjust up faster than we adjusted down)
      local delta_min=1
      local delta_calc=$(( 3 * default_weight / 100  ))
      local delta=$(( delta_min > delta_calc ? delta_min : delta_calc ))
      new_weight=$(( current_weight + delta ))
      if [ ${new_weight} -gt ${default_weight} ]; then
        new_weight=${default_weight}
      fi
      dlogger -t $0 "${machine} adjust weight up by ${delta} to {new_weight}"
    fi
  fi

  #Save the new weight as current before we add error codes
  save_weights

  #Add 2 to accommodate for error codes
  new_weight=$(( new_weight + 2 ))

  #
  # Check the return code. Our servers will return 666 or 667 when draining.
  # This routine returning a 2 actually tells keepalived the weight is zero, so stop sending new
  # connections to this server, preserve what connections you have.
  #
  if [ ! ${http_code} ]; then
    weight_return=1
    dlogger -t $0 "Failed to return http code"
  elif [ ${http_code} = 200 ]; then
    weight_return=${new_weight}
  elif [ ${http_code} = 666 ] || [ ${http_code} = 667 ]; then
    weight_return=2
  else
    weight_return=1
    dlogger -t $0 "${machine} returned unknown http code ${http_code}"
  fi
}

split_requests(){
  #Use modulo to split ipv4 and ipv6 requests evenly

  ctime=`date +"%-M"`

  if [[ ${machine} =~ ^[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3} ]]; then
    route_list=( `ip route show to match ${machine} scope link` )
    IPN=( ${machine//./ } )
    #if last octet even and minute of hour even, do macping check
    if ! (( $ctime % 2 )) && ! (( ${IPN[3]} % 2 )); then
      macping=1
    fi
    #odd last octet and odd minute of hour time, do macping check
    if (( $ctime % 2 )) && (( ${IPN[3]} % 2 )); then
      macping=1
    fi
  else
    #Machine is ipv6, grab last octet and convert to decimal
    route_list=( `ip -6 route show to match ${machine} scope global` )
    IPN6=( ${machine//:/ } )
    LIST_LEN=${#IPN6[@]}
    LAST_OCT=$(( $LIST_LEN - 1 ))
    IPN6_DEC=`printf '%d' 0x${IPN6[$LAST_OCT]}`
    #If minute is even and last octet divisible by 4, do macping check
    if ! (( $ctime % 2 )) && ! (( $IPN6_DEC % 4 )); then
      macping=1
    fi
    #If minute is odd and last octet not divisible by 4, do macping check
    if (( $ctime % 2 )) && (( ${IPN6_DEC} % 4 )); then
      macping=1
    fi
  fi
}

dsr_ping_check(){
  split_requests

  #route should be attached to eth or bond interface else we have no route
  if [[ ${route_list[2]} =~ eth|bond ]]; then
    ping_interface=${route_list[2]}
  else
    dlogger -t $0 "could not find ping interface"
    return 1
  fi

  #If we are a backup vip, unset macping var
  if [ ${option} ] && [ ${option} = "backup" ]; then
    unset macping
  fi

  server_state="/etc/keepalived/server_states/${vip}/"
  if [ ${macping} ]; then
    if [ "$debug" = "on" ] ; then
      /etc/keepalived/scripts/macping -I ${ping_interface} -d -b -q -c 1 -w 5 ${vip} ${machine} > /dev/null || /etc/keepalived/scripts/macping -I ${ping_interface} -d -b -l -q -c 1 -w 5 ${vip} ${machine} > /dev/null
    else
      /etc/keepalived/scripts/macping -I ${ping_interface} -b -q -c 1 -w 5 ${vip} ${machine} || /etc/keepalived/scripts/macping -I ${ping_interface} -b -l -q -c 1 -w 5 ${vip} ${machine}
    fi
    mac_exit=$?
    #If macping succeeds, check for bad state file and remove it
    if [ $mac_exit = 0 ]; then
      if [ -f ${server_state}${machine} ]; then
        rm ${server_state}${machine}
        return 0
      fi
    #macping returned 2 which is an internal error and we can't trust the check, return 0 unless a statefile exists
    elif [ $mac_exit = 2 ]; then
      if [ -f ${server_state}${machine} ]; then
        dlogger -t $0 "Macping exited with exit code 2, ${server_state}${machine} exists"
        return 1
      else
        dlogger -t $0 "Macping exited with exit code 2, ${server_state}${machine} does not exist"
        return 0
      fi
    else
      #we failed, create state file and return 1
      if [ ! -d ${server_state} ]; then
        mkdir $server_state
      fi
      dlogger -t $0 "Macping exited with exit code $mac_exit"
      touch ${server_state}${machine} && return 1
    fi
  else
    #we didn't have to run macping this time, but still check for bad file state and go off previous fail/pass
    if [ -f ${server_state}${machine} ]; then
      dlogger -t $0 "Macping did not run but ${server_state}${machine} exists, so returning 1"
      return 1
    else
      return 0
    fi
  fi
}

check_options(){
  #If options are set to drain or remove set return code to 2
  if [ ${option} ]; then
    if [ ${option} = "drain" ] || [ ${option} = "remove" ]; then
      exit 2
    fi
  fi
}

main(){
  #exit status 0: svc check success, weight unchanged
  #exit status 1: svc check failed, server removed
  #exit status 2-255: svc check success && weight changed to 2 less than exit status

  http_check ${1}

  #If everything ok, return dynamic weight to keepalived
  if [ ${weight_return} -gt 2 ]; then
    if dsr_ping_check; then
      check_options
      return ${weight_return}
    else
      return 1
    fi
  fi

  #Traffic draining on this server, no new requests should go to this box, set weight 0 by returning 2
  if [ ${weight_return} -eq 2 ]; then
    if dsr_ping_check; then
      return 2
    else
      return 1
    fi
  fi

  #Failed requests, remove the server from the pool by returning 1
  if [ ${weight_return} -eq 1 ]; then
    return 1
  fi
}

main ${1} ${2}
