#!/usr/bin/env bash
PATH="/sbin:/bin:/usr/bin:/usr/sbin:/usr/local/sbin:/usr/local/bin"
shopt -s nocaseglob
shopt -s nocasematch

# This script is use in a virtual server when quorum changes in a vip group.
#
# $1 = "UP" or "DOWN"
# $2 = the native interface for BGP in the form ifName.vlan
# $3 = the vip range in the form x.x.x.x-end
# $4 - $n = list of vip addrs to advertize
#

virtual_server_file=/etc/keepalived/config_containers/virtual_server.conf

error_exit(){
  echo "${0}: ${1:-""}" 1>&2
  exit 1
}

warn(){
  echo "${0}: ${1:-""}" 1>&2
}

ipv4_match(){
  #No $2 means check $1 for regular IPv4 address: x.x.x.x
  if [ ! ${2} ]; then
    if [[ ${1} =~ ^[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3} ]]; then
      return 0
    else
      return 1
    fi
  fi

  #Match for ip_addr-[###]
  if [ ${2} = range ]; then
    if [[ ${1} =~ ^[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\-[0-9]{1,3} ]]; then
      return 0
    else
      return 1
    fi
  fi

  #Match for ip_addr/[##]
  if [ ${2} = subnet ]; then
    if [[ "${1}" =~ ^[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\/[0-9]{1,2} ]]; then
      return 0
    else
      return 1
    fi
  fi

  return 1
}

verify_v6(){
  v6_octet_count=0
  local counter=1
  declare -gA v6_octets

  if [ $v6_ip ]; then
    unset v6_ip
  fi

  #Verify each octet is a valid ipv6 hex
  IFS=':' read -a tmp_octets <<< ${1}

  #Need to count for compressed
  for octet in ${tmp_octets[@]}; do
    (( v6_octet_count++ ))
  done

  if [ ${v6_octet_count} -lt 8 ]; then
    compressed=1
  fi

  for hex in ${tmp_octets[@]}; do
    #Verify the hex to see if valid
    if [[ ${hex} =~ ^[0-9A-Fa-f]{1,4}$ ]]; then
      if [ ${counter} -lt ${v6_octet_count} ]; then
        v6_ip+=( "${hex}:" )
        (( counter++ ))
      elif [ ${counter} -le ${v6_octet_count} ] && [ ${compressed} ]; then
        v6_ip+=( ":" )
        (( octet_count++ ))
      fi
    else
      unset tmp_octets
      return 1
    fi
  done

  unset tmp_octets
  return 0
}

ipv6_match(){
  #IF $2 NOT DEFINED "expect xxxx:xxxx:xxxx:xxxx::HEX"
  if [ ! ${2} ]; then
    if verify_v6 ${1}; then
      return 0
    else
      return 1
    fi
  fi

  #IF $2 = RANGE "expect xxxx:xxxx:xxxx:xxxx::HEX-HEX"
  if [ ${2} = range ]; then
    IFS='-' read -a tmp_range <<< ${1}
    if verify_v6 ${tmp_range[0]}; then
      return 0
      unset tmp_range
    else
      return 1
      unset tmp_range
    fi
  fi

  #IF $2 = SUBNET "expect xxxx:xxxx:xxxx:xxxx::HEX/cidr"
  if [ ${2} = subnet ]; then
    IFS='/' read -a tmp_subnet <<< ${1}
    if verify_v6 ${tmp_subnet[0]}; then
      return 0
      unset tmp_subnet
    else
      return 1
      unset tmp_subnet
    fi
  fi
}

# ${1}=vip addr range
# Output: vip_list array for v4 addresses
vip_list_build(){
  declare -gA vip_list
  split_addr=( ${1//-/ } )
  ip=${split_addr[0]}
  OIFS=$IFS
  IFS='.'
  ip=($ip)
  IFS=$OIFS
  
  # loop through the range building the vip list
  for i in `seq ${ip[3]} ${split_addr[1]}`; do
    vip_list[${ip[0]}.${ip[1]}.${ip[2]}.${i}]=1
  done
}

# ${1}=vip addr range
# Output: vip_list array for v6 addresses
vip_v6_list_build(){
  declare -gA vip_v6_list
  split_addr=( ${1//-/ } )
  ip=$(perl -MNet::IP -e'my $var=new Net::IP($ARGV[0]);  print $var->ip()' ${split_addr[0]})
  ip_full=$ip
  OIFS=$IFS
  IFS=':'
  ip=($ip)
  IFS=$OIFS

  for i in `seq 0x${ip[7]} 0x${split_addr[1]}`; do
    hex_i=$(printf '%04x\n' $i)
    vip_v6_list[${ip[0]}:${ip[1]}:${ip[2]}:${ip[3]}:${ip[4]}:${ip[5]}:${ip[6]}:${hex_i}]=1
  done
}

# ${1}=UP/DOWN ${2}=vip addr range
vip_dummy() {
  if [ ${1} = "UP" ]; then
    op='add'
  else
    op='del'
  fi
  
  if ipv4_match ${2} range; then
    vip_list_build ${2}
    for entry in ${!vip_list[@]}; do
      ip addr ${op} ${entry}/32 dev dummy0 noprefixroute || warn "failed command: ip addr ${op} ${entry}/32 dev dummy0"
    done
  elif ipv6_match ${2} range; then
    vip_v6_list_build ${2}
    for v6_entry in ${!vip_v6_list[@]}; do
      ip -6 addr ${op} ${v6_entry}/128 dev dummy0 noprefixroute || warn "failed command: ip -6 addr ${op} ${v6_entry}/128 dev dummy0"
    done
  fi
}

# Add vip_advertise addresses to the BGP interface
# ${1}=ARGC ${2}=ARGV
vip_bgp_interface() {
  # No vips to add to bgp interface
  if [ ${1} -le 3 ]; then
    return 0
  fi
  
  adv_cnt=$(expr "${1}" - "3")

  # Set the inteface addr operation
  if [ ${2} = "UP" ]; then
    op='add'
  else
    op='del'
  fi
  
  if ipv4_match ${5}; then
    for (( c=0, pos=5; c<adv_cnt; c++, pos++ )); do
      ip addr ${op} ${!pos} dev ${3} || warn "failed command: ip addr ${op} ${!pos} dev ${3}"
    done
  else 
    for (( c=0, pos=5; c<adv_cnt; c++, pos++ )); do
      ip -6 addr ${op} ${!pos} dev ${3} || warn "failed command: ip -6 addr ${op} ${!pos} dev ${3}"
    done
  fi

}


################### MAIN #########################


# Handle the vip addresses on the dummy0 interface.
vip_dummy $1 $3

# Handle the vip advertise addresses on the BGP interface
vip_bgp_interface $# $@
