#!/usr/bin/env perl
use strict;
use warnings;
use Sys::Hostname;
use Data::Dumper;
use Regexp::IPv6 qw($IPv6_re);
use POSIX;
use YAML qw(LoadFile Dump DumpFile);
use Net::CIDR;
use Net::IP;
use IO::Socket;


# vip_traffic.pl
#
# This script calculates the PPS rate per VIP, VIP range and port and sends
# it to OpenTSDB.


# Dry run prints the output to STDOUT instead of sending it to OpenTSDB.
my $dry_run = 0;

# Any VIPs with PPS rate lower than this will be grouped
# together in a VIP named "other".
my $pps_limit = 10000;

main();

sub usage {
  print STDERR "Usage: vip-traffic.pl [<options>]\n";
  print STDERR "   -d: dry run\n";
  print STDERR "\n";
  exit(1);
}

sub get_my_pop {
  my $arg = shift || die "Missing pop arg $!";
  my $local_pop = (split '\.',$arg)[1];
  return $local_pop;
}

sub main {
  my $hostname = hostname;

  if ($ARGV[0] && $ARGV[0] eq '-d') {
    $dry_run = 1;
  } elsif ($ARGV[0] && $ARGV[0] eq '-h') {
    usage();
  }

  unless ( $hostname =~ /^slb\d\d/ ) {
    die "This script is only meant for software load balancers\n";
  }

  my $pop = get_my_pop($hostname);
  my $yml = LoadFile("/etc/keepalived/slb_conf.yml");
  my $yml_config = undef;
  my $yml_vips = undef;

  if ($yml->{config_version} && (($yml->{config_version} eq "2") || ($yml->{config_version} eq "2_ecmp")) ) {
    $yml_config = $yml->{slb_configs}->{$hostname} or die "No config for $hostname found in /etc/keepalived/slb_conf.yml $!";
    $yml_vips = $yml->{vips} or die "No vip config found in /etc/keepalived/slb_conf.yml $!";
  } else {
    $yml_config = $yml->{$hostname}->{slb_config} or die "No config for $hostname found in /etc/keepalived/slb_conf.yml $!";
    $yml_vips = $yml->{$hostname}->{vips} or die "No vip config for $hostname found in /etc/keepalived/slb_conf.yml $!";
  }
  
  my $command = "ipvsadm -Ln --rate";
  my %traffic_data;

  open(my $fh, '-|', $command) or die $!;
  my $ip;
  while (my $line = <$fh>) {
    next if $. < 3; # Skip first three lines
    chomp($line);
    if ($line =~ /^TCP/) {
      my @words = split ' ', $line;
      my $ip_port = $words[1];
      my $pps = $words[3];
      if ($pps > 0) {
        $traffic_data{$ip_port} = $pps;
      }
    }
  }
  close($fh) or die "close $!";

  my %results;
  my %ranges;
  my @range_names = keys %$yml_vips;
  for my $range_name ( @range_names ) {

    my $vip = $yml_vips->{$range_name};
    my $vip_cidr = $vip->{'vip_addr'};
    my $range_start = (split '\/', $vip_cidr)[0];
    my $range_end;
    my $ip_range_start = new Net::IP ($range_start) or die (Net::IP::Error());
    my $ip_range_end;

    if ($ip_range_start->version() == 4) {

      my @range_parts = split '\.', $range_start;
      pop @range_parts;
      push @range_parts, $vip->{'vip_range'};
      $range_end = join('.', @range_parts);
      $ip_range_end = new Net::IP ($range_end) or die (Net::IP::Error());
      my $s = $ip_range_start->ip();
      my $e = $ip_range_end->ip();
      my $range = "${s} - ${e}";
      my $ip_range = new Net::IP ($range) or die (Net::IP::Error());
      my @keys = keys %traffic_data;
      for my $ip_port ( @keys ) {
        # Only IPv4
        if ( scalar (split ':', $ip_port) == 2 ) {
          my $port = (split ':', $ip_port)[1];
          my $pps = $traffic_data{$ip_port};
          my $ip = new Net::IP ((split ':', $ip_port)[0]) or die (Net::IP::Error());
          if ($ip_range->overlaps($ip)) {
            if ($pps > $pps_limit) {
              $results{$range_name}->{$ip->ip()}->{$port} = $pps;
            } else {
              $results{$range_name}->{'other'}->{$port} += $pps;
            }
          }
        }
      }

    } else {

      my $ip_start = new Net::IP ($range_start) or die (Net::IP::Error());
      my @range_parts = split ':', $ip_start->ip();
      pop @range_parts;
      push @range_parts, $vip->{'vip_range'};
      my $ip_end = new Net::IP (join(':', @range_parts)) or die (Net::IP::Error());
      my $s = $ip_start->ip();
      my $e = $ip_end->ip();
      my $str_range = "${s} - ${e}";
      my $ip_range = new Net::IP ($str_range) or die (Net::IP::Error());
      my @keys = keys %traffic_data;
      for my $ip_port ( @keys ) {
        # Only IPv6
        if ( scalar (split ':', $ip_port) > 2 ) {
          my @ip_parts = split ':', $ip_port;
          my $port = pop @ip_parts;
          my $pps = $traffic_data{$ip_port};
          my $ip = join(':', @ip_parts);
          $ip = substr $ip, 1, -1;
          $ip = new Net::IP ($ip) or die (Net::IP::Error());
          if ($ip_range->overlaps($ip)) {
            if ($pps > $pps_limit) {
              $results{$range_name}->{$ip->ip()}->{$port} = $pps;
            } else {
              $results{$range_name}->{'other'}->{$port} += $pps;
            }
          }
        }
      }
    }
  }

  my $tsdb = "";
  my @range_keys = keys %results;
  my $time = time();
  for my $range_key ( @range_keys ) {
    my $range = $results{$range_key};
    my @ip_keys = keys %$range;
    for my $ip_key ( @ip_keys ) {
      my $ip = $range->{$ip_key};
      my @port_keys = keys %$ip;
      for my $port_key ( @port_keys ) {
        my $pps = $ip->{$port_key};
        my $line = sprintf("put net.slb.viptraffic ${time} ${pps} pop=${pop} fqdn=${hostname} range=${range_key} vip=${ip_key} port=${port_key}\n");
        $tsdb = join('', $tsdb, $line);
      }
    }
  }

  if ($dry_run == 1) {
    print "Dry run, printing output:\n";
    if ($tsdb ne "") {
      print $tsdb;
    } else {
      print "<No data>\n";
    }
  } elsif ($tsdb ne "") {
    my $addr = 'tsdb-data.lldns.net';
    my $port = '4242';
    my $socket = IO::Socket::INET->new(PeerAddr => $addr,
                                PeerPort => $port,
                                Proto    => "tcp",
                                Type     => SOCK_STREAM)
    or die "Couldn't connect to $addr:$port : $@\n";
    print $socket $tsdb;
    close($socket);
  }
}
