#!/usr/bin/perl
use strict;
use warnings;
use Sys::Hostname;
use Data::Dumper;
use POSIX;
use YAML qw(LoadFile);
use Socket;

main();

sub main {
  my $hostname = hostname;
  my $gen_all = 0;
  my $gen_vlan = 0;
  my $gen_ints = 0;
  my $bind_ips = 0;

  unless ( $hostname =~ /^slb\d\d/ ) {
    die "This script is only meant for software load balancers\n";
  }

  if ( $ARGV[0] ) {
    if ( $ARGV[0] eq '-A' ) {
      $gen_all=1;
    } elsif ( $ARGV[0] eq '-V' ) {
      $gen_vlan=1;
    } elsif ( $ARGV[0] eq '-I' ) {
      $gen_ints=1;
    } elsif ( $ARGV[0] eq '-B' ) {
      $bind_ips=1;
    } elsif ( $ARGV[0] =~ /^-/) {
      usage();
    }
  }
  else {
    usage();
  }

  # Make sure interfaces.d directory is there. If not, create it.
  system ("mkdir -p /etc/network/interfaces.d");

  my $pop = get_my_pop($hostname);
  my $yml_conf = load_yml_file($pop,$hostname);

  if ( $gen_all == 1 ) {
    gen_vlan_files($yml_conf);
    gen_interfaces_file($yml_conf);
    print "Please verify /etc/network/interfaces and files in /etc/network/interfaces.d/*.cfg before rebooting\n";
  } elsif ( $gen_vlan == 1 ) {
    gen_vlan_files($yml_conf);
    print "Please verify files in /etc/network/interfaces.d/*.cfg before rebooting\n";
  } elsif ( $gen_ints == 1 ) {
    gen_interfaces_file($yml_conf);
    print "Please verify /etc/network/interfaces before rebooting\n";
  } elsif ( $bind_ips == 1 ) {
    #ipv6 doesn't reliably come up under bonded vlan interfaces
    bind_v6_addresses($yml_conf);
  }

}

sub usage {
  print STDERR "Usage: gen_network_configs.pl [<options>]\n";
  print STDERR "   -A: generate all network and vlan files\n";
  print STDERR "   -I: generate only /etc/network/interfaces\n";
  print STDERR "   -V: generate only /etc/network/interfaces.d/vlan.cfg files\n";
  print STDERR "   -B: check and bind vlan v6 addresses due to bugs in ubuntu\n";
  print STDERR "\n";
  print STDERR "Example:\n";
  print STDERR "   ./gen_network_configs.pl -V (builds only vlan files)\n";
  exit(1);
}

sub get_my_pop {
  my $arg = shift || die "Missing pop arg $!";

  my $local_pop = (split '\.',$arg)[1];
  return $local_pop;
}

sub load_yml_file {
  my $pop = shift || die "Did not receive a valid pop $!";
  my $hostname = shift || die "Did not receive a valid hostname $!";
  my $yml = LoadFile("/etc/keepalived/slb_conf.yml");
  my $yml_config = undef;

  if ($yml->{config_version} && (($yml->{config_version} eq "2") || ($yml->{config_version} eq "2_ecmp")) ) {
    $yml_config = $yml->{slb_configs}->{$hostname} or die "No config for $hostname found in /etc/keepalived/slb_conf.yml $!";
  } else {
    $yml_config = $yml->{$hostname}->{slb_config} or die "No config for $hostname found in /etc/keepalived/slb_conf.yml $!";
  }
  return $yml_config;
}

sub bind_v6_addresses {
  my $slb_conf = shift || die "Did not receive yml ref $!";
  my $yml_vlans = $slb_conf->{cdn_vlans};
  my $slb_interface = $slb_conf->{interface};
  my $slb_vlan = $slb_conf->{native_vlan};
  my $v6_address = (split '/',$slb_conf->{v6_address})[0];
  my $v6_netmask = (split '/',$slb_conf->{v6_address})[1];

  # if v6 is not defined in the config then skip it
  if ( !(defined($v6_address) && defined($v6_netmask)) ) {
    print STDERR "Skipping IPv6 address binding - IPv6 info is not defined in config";
    return;
  }

  system("ip -6 addr add ${v6_address}/${v6_netmask} dev ${slb_interface}.${slb_vlan}");

  my @keys = keys %$yml_vlans;

  for my $vlan ( @keys ) {
    my $vlan_v6 = $yml_vlans->{$vlan}->{v6_address};

    if ( defined($vlan_v6) && $vlan_v6 ne "" ) {
      my $vlan_v6_address = (split '/',$yml_vlans->{$vlan}->{v6_address})[0];
      my $vlan_v6_netmask = (split '/',$yml_vlans->{$vlan}->{v6_address})[1];

      system("ip -6 addr add ${vlan_v6_address}/${vlan_v6_netmask} dev ${slb_interface}.${vlan}");
    }
  }
}

sub gen_vlan_files {
  my $slb_conf = shift || die "Did not receive yml ref $!";
  my $yml_vlans = $slb_conf->{cdn_vlans};
  my $slb_interface = $slb_conf->{interface};
  my $slb_vlan = $slb_conf->{native_vlan};

  # if CDN vlan is not defined, default to using the Native VLAN as a single VLAN for both
  # inbound and CDN/DSR communication.
  if (!defined $yml_vlans) {
    print "WARNING: no CDN VLAN configured -- defaulting to Native VLAN ${slb_vlan}\n";
    return
  }

  my @keys = keys %$yml_vlans;

  for my $vlan ( @keys ) {
    my $vlan_ip = $yml_vlans->{$vlan}->{address};
    my $vlan_v6 = $yml_vlans->{$vlan}->{v6_address};
    my $vlan_tmp_file = "/etc/network/interfaces.d/${slb_interface}.${vlan}.cfg.new";
    my $vlan_file = "/etc/network/interfaces.d/${slb_interface}.${vlan}.cfg";

    open( my $nfh,'>',"$vlan_tmp_file" ) or die "Failed to open $vlan_tmp_file $!";

    #If array we need to loop through all the entries and create the primary interface and then
    #secondary interfaces, which need to be virtuals of the primary to apply multiple addresses
    #to a single vlan
    my $virt_int = 0;
    if ( $slb_vlan == $vlan ) {
      $virt_int++;
    }
    #ipv4 setup
    if (ref($vlan_ip) eq 'ARRAY') {
      foreach my $address ( @{ $vlan_ip } ) {
        my $v4_vlan_ip = (split '/',${address})[0] or die "Failed to parse address from $address";
        my $v4_vlan_cidr = (split '/',${address})[1] or die "Failed to parse netmask from $address";
        if ( $virt_int == 0 ) {
          print $nfh "auto ${slb_interface}.${vlan}\n" or die "Failed to write to $vlan_tmp_file $!";
          print $nfh "iface ${slb_interface}.${vlan} inet static\n" or die "Failed to write to $vlan_tmp_file $!";
          print $nfh "  address $v4_vlan_ip\n" or die "Failed to write to $vlan_tmp_file $!";
          print $nfh "  netmask $v4_vlan_cidr\n" or die "Failed to write to $vlan_tmp_file $!";
          print $nfh "  vlan-raw-device $slb_interface\n" or die "Failed to write to $vlan_tmp_file $!";
          $virt_int++;
          check_route("${slb_interface}.${vlan}", ${v4_vlan_ip}, ${v4_vlan_cidr});
        } else {
          print $nfh "auto ${slb_interface}.${vlan}:$virt_int\n" or die "Failed to write to $vlan_tmp_file $!";
          print $nfh "iface ${slb_interface}.${vlan}:$virt_int inet static\n" or die "Failed to write to $vlan_tmp_file $!";
          print $nfh "  address $v4_vlan_ip\n" or die "Failed to write to $vlan_tmp_file $!";
          print $nfh "  netmask $v4_vlan_cidr\n" or die "Failed to write to $vlan_tmp_file $!";
          $virt_int++;
          check_route("${slb_interface}.${vlan}", ${v4_vlan_ip}, ${v4_vlan_cidr});
        }
      }
    } else {
      #This is not an array, no need to iterate
      my $v4_vlan_ip = (split '/',${vlan_ip})[0] or die "Failed to parse address from ${vlan_ip}";
      my $v4_vlan_cidr = (split '/',${vlan_ip})[1] or die "Failed to parse netmask from ${vlan_ip}";
      if ( $virt_int == 0 ) {
        print $nfh "auto ${slb_interface}.${vlan}\n" or die "Failed to write to $vlan_tmp_file $!";
        print $nfh "iface ${slb_interface}.${vlan} inet static\n" or die "Failed to write to $vlan_tmp_file $!";
      } else {
        print $nfh "auto ${slb_interface}.${vlan}:$virt_int\n" or die "Failed to write to $vlan_tmp_file $!";
        print $nfh "iface ${slb_interface}.${vlan}:$virt_int inet static\n" or die "Failed to write to $vlan_tmp_file $!";
      }
      print $nfh "  address $v4_vlan_ip\n" or die "Failed to write to $vlan_tmp_file $!";
      print $nfh "  netmask $v4_vlan_cidr\n" or die "Failed to write to $vlan_tmp_file $!";
      print $nfh "  vlan-raw-device $slb_interface\n" or die "Failed to write to $vlan_tmp_file $!";
      check_route("${slb_interface}.${vlan}", ${v4_vlan_ip}, ${v4_vlan_cidr});
    }

    #ipv6 setup - only process v6 if defined in config
    if (defined($vlan_v6)) {
      $virt_int = 0;
      if ($slb_vlan == $vlan) {
        $virt_int++;
      }

      if (ref($vlan_v6) eq 'ARRAY') {
        foreach my $v6_address (@{$vlan_v6}) {
          my $vlan_v6_address = (split '/', ${v6_address})[0] or die "Failed to parse v6 address from $v6_address";
          my $vlan_v6_netmask = (split '/', ${v6_address})[1] or die "Failed to parse v6 netmask from $v6_address";
          if ($virt_int == 0) {
            print $nfh "iface ${slb_interface}.${vlan} inet6 static\n" or die "Failed to write to $vlan_tmp_file $!";
          }
          else {
            print $nfh "iface ${slb_interface}.${vlan}:$virt_int inet6 static\n" or die "Failed to write to $vlan_tmp_file $!";
          }
          print $nfh "  address $vlan_v6_address\n" or die "Failed to write to $vlan_tmp_file $!";
          print $nfh "  netmask $vlan_v6_netmask\n" or die "Failed to write to $vlan_tmp_file $!";
          $virt_int++;
          check_route("${slb_interface}.${vlan}", ${vlan_v6_address}, ${vlan_v6_netmask});
        }
      }
      else {
        if (defined($vlan_v6) && $vlan_v6 ne "") {
          my $vlan_v6_address = (split '/', $yml_vlans->{$vlan}->{v6_address})[0] or die "Failed to parse v6 address from $vlan_v6";
          my $vlan_v6_netmask = (split '/', $yml_vlans->{$vlan}->{v6_address})[1] or die "Failed to parse v6 netmask from $vlan_v6";
          if ($virt_int == 0) {
            print $nfh "iface ${slb_interface}.${vlan} inet6 static\n" or die "Failed to write to $vlan_tmp_file $!";
          }
          else {
            print $nfh "iface ${slb_interface}.${vlan}:$virt_int inet6 static\n" or die "Failed to write to $vlan_tmp_file $!";
          }
          print $nfh "  address $vlan_v6_address\n" or die "Failed to write to $vlan_tmp_file $!";
          print $nfh "  netmask $vlan_v6_netmask\n" or die "Failed to write to $vlan_tmp_file $!";
          check_route("${slb_interface}.${vlan}", ${vlan_v6_address}, ${vlan_v6_netmask});
        }
      }
    }
    close ($nfh) or die "Failed to close $vlan_tmp_file $!";
    rename_file_perms($vlan_tmp_file,$vlan_file); 
  }
}

sub gen_interfaces_file {
  my $slb_conf = shift || die "Did not receive yml ref $!";
  my $slb_interface = $slb_conf->{interface};
  my $slb_vlan = $slb_conf->{native_vlan};
  my $address = (split '/',$slb_conf->{address})[0] or die "Failed to parse address from $slb_conf->{address}";
  my $netmask = (split '/',$slb_conf->{address})[1] or die "Failed to parse netmask from $slb_conf->{address}";
  my $gateway = $slb_conf->{gateway};
  my $v6_addr_info = $slb_conf->{v6_address};
  my $v6_address = undef;
  my $v6_netmask = undef;
  my $lacp_ints = $slb_conf->{lacp_ints};
  my $interfaces_tmp_file = "/etc/network/interfaces.new";
  my $interfaces_file = "/etc/network/interfaces";

  # if v6 addr info is defined then create address and netmask arrays
  if ( defined($v6_addr_info)) {
    $v6_address = (split '/', $slb_conf->{v6_address})[0] or die "Failed to parse address from $slb_conf->{v6_address}";
    $v6_netmask = (split '/', $slb_conf->{v6_address})[1] or die "Failed to parse subnet from $slb_conf->{v6_address}";
  }

  open( my $ifh,'>',"$interfaces_tmp_file" ) or die "Failed to open $interfaces_tmp_file $!";

  print $ifh "auto lo\n" or die "Failed to write to $interfaces_tmp_file $!";
  print $ifh "iface lo inet loopback\n" or die "Failed to write to $interfaces_tmp_file $!";
  print $ifh "\n" or die "Failed to write to $interfaces_tmp_file $!";

  for my $int ( @{$lacp_ints} ) {
    print $ifh "auto $int\n" or die "Failed to write to $interfaces_tmp_file $!";
    print $ifh "iface $int inet manual\n" or die "Failed to write to $interfaces_tmp_file $!";
    print $ifh "  pre-up /sbin/ethtool -s \$IFACE speed 10000 duplex full autoneg off\n" or die "Failed to write to $interfaces_tmp_file $!";
    print $ifh "  pre-up /sbin/ethtool -K \$IFACE gso off gro off lro off tso off\n" or die "Failed to write to $interfaces_tmp_file $!";
    print $ifh "  pre-up /sbin/ifconfig \$IFACE txqueuelen 3000\n" or die "Failed to write to $interfaces_tmp_file $!";
    print $ifh "  post-up /sbin/ethtool -C \$IFACE rx-usecs 50\n" or die "Failed to write to $interfaces_tmp_file $!";
    print $ifh "  pre-up /sbin/ethtool -A \$IFACE autoneg off rx off tx off\n" or die "Failed to write to $interfaces_tmp_file $!";
    print $ifh "  bond-master ${slb_interface}\n" or die "Failed to write to $interfaces_tmp_file $!";
    print $ifh "\n" or die "Failed to write to $interfaces_tmp_file $!";
  }

  print $ifh "auto ${slb_interface}\n" or die "Failed to write to $interfaces_tmp_file $!";
  print $ifh "iface ${slb_interface} inet manual\n" or die "Failed to write to $interfaces_tmp_file $!";
  print $ifh "  bond-mode 4\n" or die "Failed to write to $interfaces_tmp_file $!";
  print $ifh "  bond-miimon 100\n" or die "Failed to write to $interfaces_tmp_file $!";
  print $ifh "  bond-lacp-rate 1\n" or die "Failed to write to $interfaces_tmp_file $!";
  print $ifh "  bond-xmit_hash_policy layer3+4\n" or die "Failed to write to $interfaces_tmp_file $!";
  print $ifh "  bond-slaves @{$lacp_ints}\n" or die "Failed to write to $interfaces_tmp_file $!";
  print $ifh "\n" or die "Failed to write to $interfaces_tmp_file $!";

  print $ifh "auto ${slb_interface}.${slb_vlan}\n" or die "Failed to write to $interfaces_tmp_file $!";
  print $ifh "iface ${slb_interface}.${slb_vlan} inet static\n" or die "Failed to write to $interfaces_tmp_file $!";
  print $ifh "  address $address\n" or die "Failed to write to $interfaces_tmp_file $!";
  print $ifh "  netmask $netmask\n" or die "Failed to write to $interfaces_tmp_file $!";
  print $ifh "  vlan-raw-device ${slb_interface}\n" or die "Failed to write to $interfaces_tmp_file $!";
  print $ifh "\n" or die "Failed to write to $interfaces_tmp_file $!";

  if ( defined($v6_address) ) {
    print $ifh "iface ${slb_interface}.${slb_vlan} inet6 static\n" or die "Failed to write to $interfaces_tmp_file $!";
    print $ifh "  address $v6_address\n" or die "Failed to write to $interfaces_tmp_file $!";
    print $ifh "  netmask $v6_netmask\n" or die "Failed to write to $interfaces_tmp_file $!";
    print $ifh "\n" or die "Failed to write to $interfaces_tmp_file $!";
  }
  print $ifh "source /etc/network/interfaces.d/*.cfg\n" or die "Failed to write to $interfaces_tmp_file $!";

  close ($ifh) or die "Failed to close $interfaces_tmp_file $!";

  rename_file_perms($interfaces_tmp_file,$interfaces_file); 
}

sub check_route {
  my $v_int = shift || die "Missing interface";
  my $interface = (split '\.', $v_int )[0];
  my $vlan = (split '\.', $v_int )[1];
  my $addr_to_check = shift || die "Missing address";
  my $cidr_to_check = shift || die "Missing subnet";
  my @bound_address;
  my $int = system("ifconfig $v_int > /dev/null 2>&1");
  my $nobind = 0;

  if ( $int == 0 ) {
    open( my $vlan_int, "ip address list dev $v_int 2> /dev/null |") or warn "error listing addresses on $int $!";
      while ( my $line = <$vlan_int> ) {
        if ( $line =~ /inet/ ) {
          push @bound_address, (split ' ', $line )[1];
        }
      }
  } else {
    #Interface does not exist, need to create before we can bind
    if ( system("ifconfig $interface > /dev/null") == 0 ) {
      create_interface($interface,$vlan);
    } else {
      $nobind = 1;
    }
  }

  #IF array is not empty loop through addresses to see if it's already bound
  my $addr_match = 0 ;
  if ( @bound_address ) {
    foreach my $addr ( @bound_address ) {
      my $parsed_ip = (split '/', $addr )[0];
      if ( $parsed_ip eq $addr_to_check ) {
        $addr_match = 1 ;
      }
    }
    if ( $addr_match == 0 ) {
      if ( $nobind != 1 ) {
        #Address not found, we need to add it to the interface
        if ( system("ip addr add ${addr_to_check}/${cidr_to_check} dev $v_int") != 0 ) {
          die "Failed to bind ${addr_to_check}/${cidr_to_check} to $v_int $!";
        }
      }
    }
  } else {
    #No addresses found on new interface, bind up the address
    if ( $nobind != 1 ) {
      if ( system("ip addr add ${addr_to_check}/${cidr_to_check} dev $v_int") != 0 ) {
        die "Failed to bind ${addr_to_check}/${cidr_to_check} to $v_int $!";
      }
    }
  }
}

sub create_interface {
  my $interface = shift || die "Missing interface arg $!";
  my $vlan = shift || die "Missing vlan arg $!";

  if ( system("vconfig add $interface $vlan > /dev/null") != 0 ) {
    die "failed to add vlan config for $interface $vlan $!";
  }
}

sub rename_file_perms {
  my $tmp_file = shift || die "Missing temp file arg $!";
  my $file = shift || die "Missing vip file arg $!";

  chmod 0644, $tmp_file or die "Failed to change permissions on file $tmp_file $!";
  my $uid = getpwnam "root";
  my $gid = getgrnam "root";
  chown $uid, $gid, $tmp_file or die "Failed to chown $tmp_file to root:root $!";
  rename $tmp_file,$file or die "Failed to rename $tmp_file to $file $!";
}

