#!/usr/bin/env bash
PATH="/sbin:/bin:/usr/bin:/usr/sbin:/usr/local/sbin:/usr/local/bin"
shopt -s nocaseglob
shopt -s nocasematch

#Script to read the vrrp and virtual server files for keepalived and bind the necessary
#ip addresses to the dummy interface.  This keeps keepalived from sending out massive
#amounts of gratuitous arp entries which overflow the netlink buffers.  This script is
#called by keepalived with 3 args
#$1 = “GROUP” or “INSTANCE”
#$2 = name of group or instance
#$3 = target state of transition (“MASTER”, “BACKUP”, “FAULT”)

TYPE=${1}
NAME=${2}
STATE=${3}

vrrp_file=/etc/keepalived/config_containers/vrrp_instance.conf
virtual_server_file=/etc/keepalived/config_containers/virtual_server.conf

error_exit(){
  echo "${0}: ${1:-""}" 1>&2
  exit 1
}

warn(){
  echo "${0}: ${1:-""}" 1>&2
}

ipv4_match(){
  if [ ! ${2} ]; then
    if [[ ${1} =~ ^[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3} ]]; then
      return 0
    else
      return 1
    fi
  fi

  #Match for ip_addr-[###]
  if [ ${2} = range ]; then
    if [[ ${1} =~ ^[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\-[0-9]{1,3} ]]; then
      return 0
    else
      return 1
    fi
  fi

  #Match for ip_addr/[##]
  if [ ${2} = subnet ]; then
    if [[ "${1}" =~ ^[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\/[0-9]{1,2} ]]; then
      return 0
    else
      return 1
    fi
  fi

  return 1
}

verify_v6(){
  v6_octet_count=0
  local counter=1
  declare -gA v6_octets

  if [ $v6_ip ]; then
    unset v6_ip
  fi

  #Verify each octet is a valid ipv6 hex
  IFS=':' read -a tmp_octets <<< ${1}

  #Need to count for compressed
  for octet in ${tmp_octets[@]}; do
    (( v6_octet_count++ ))
  done

  if [ ${v6_octet_count} -lt 8 ]; then
    compressed=1
  fi

  for hex in ${tmp_octets[@]}; do
    #Verify the hex to see if valid
    if [[ ${hex} =~ ^[0-9A-Fa-f]{1,4}$ ]]; then
      if [ ${counter} -lt ${v6_octet_count} ]; then
        v6_ip+=( "${hex}:" )
        (( counter++ ))
      elif [ ${counter} -le ${v6_octet_count} ] && [ ${compressed} ]; then
        v6_ip+=( ":" )
        (( octet_count++ ))
      fi
    else
      unset tmp_octets
      return 1
    fi
  done

  unset tmp_octets
  return 0
}

ipv6_match(){
  #IF $2 NOT DEFINED "expect xxxx:xxxx:xxxx:xxxx::HEX"
  if [ ! ${2} ]; then
    if verify_v6 ${1}; then
      return 0
    else
      return 1
    fi
  fi

  #IF $2 = RANGE "expect xxxx:xxxx:xxxx:xxxx::HEX-HEX"
  if [ ${2} = range ]; then
    IFS='-' read -a tmp_range <<< ${1}
    if verify_v6 ${tmp_range[0]}; then
      return 0
      unset tmp_range
    else
      return 1
      unset tmp_range
    fi
  fi

  #IF $2 = SUBNET "expect xxxx:xxxx:xxxx:xxxx::HEX/cidr"
  if [ ${2} = subnet ]; then
    IFS='/' read -a tmp_subnet <<< ${1}
    if verify_v6 ${tmp_subnet[0]}; then
      return 0
      unset tmp_subnet
    else
      return 1
      unset tmp_subnet
    fi
  fi
}

get_ranges(){
  declare -gA vip_ranges
  declare -gA vip_ranges_v6

  #Get the ip address ranges from the virtual_server "VIP" config files
  while read ip port; do
    #Match v4 or v6 match ip range xxx-xxx
    if ipv4_match ${ip} range; then
      split_addr=( ${ip//-/ } )
      vip_ranges[${split_addr[0]}]=${split_addr[1]}
      continue
    fi

    if ipv6_match ${ip} range; then
      split_addr=( ${ip//-/ } )
      vip_ranges_v6[${split_addr[0]}]=${split_addr[1]}
    fi
  done < ${virtual_server_file} || error_exit "Failed to open file ${virtual_server_file}"
}

get_vrrp_ips(){
  #Get VRRP specific ips, these will be removed from the vip_binds and vip_binds_v6 key/val
  declare -gA vrrp_ips
  local vip_config_block=0

  while read key value; do
    if [ "${key}" = "virtual_ipaddress" ]; then
      vip_config_block=1
      continue
    fi

    if [ ${vip_config_block} -eq 1 ]; then
      #V4 or V6 Match
      if ipv4_match ${key} subnet || ipv6_match ${key} subnet; then
        IFS='/' read -a temp_array <<< "${key}"
        #Key is the ip address value is the subnet length
        vrrp_ips[${temp_array[0]}]=${temp_array[1]}
      else
        vip_config_block=0
      fi
    fi
  done < ${vrrp_file} || error_exit "Failed to open ${vrrp_file}"
}

get_ipvs_vips() {
  #Get all of the vips that ipvs created and is trying to listen on
  declare -gA vip_binds
  declare -gA vip_binds_v6

  while read protocol ip rest_of_line; do
    if [ ${protocol} = "TCP" ]; then
      if [[ ${ip} =~ ^[0-9]{1,3}\. ]]; then
        split_v4=( ${ip//:/ } )
        vip_binds[${split_v4[0]}]=1
      elif [[ ${ip} =~ ^\[ ]]; then
        split_v6=${ip//\]/ }
        split_list=( ${split_v6//\[/ } )
        vip_binds_v6[${split_list[0]}]=1
      fi
    fi
  done< <(ipvsadm -Ln | egrep -i "TCP|UDP" )

  #Remove the VRRP ip if it's in hash, this needs a non dummy bind.
  for vrrp_ip in ${!vrrp_ips[@]}; do
    if [ ${vip_binds[$vrrp_ip]} ]; then
      unset vip_binds[$vrrp_ip]
    fi
  done

  for vrrp_ip6 in ${!vrrp_ips[@]}; do
    if [ ${vip_binds_v6[$vrrp_ip6]} ]; then
      unset vip_binds_v6[$vrrp_ip6]
    fi
  done
}

get_current_binds(){
  #read current binds from dummy interface to get a list of addresses already bound so we can ignore
  while read inet ip_addr scope global interface; do
    if ipv4_match ${ip_addr} subnet; then
      bound_v4=( ${ip_addr//\// } )
      #remove current v4 bound addresses found on dummy interface from the vip_binds array
      if [ ${vip_binds[${bound_v4[0]}]} ]; then
        unset vip_binds[${bound_v4[0]}]
      fi
    elif ipv6_match ${ip_addr} subnet; then
      bound_v6=( ${ip_addr//\// } )
      #remove current v6 bound addresses found on dummy interface from the vip_binds array
      if [ ${vip_binds_v6[${bound_v6[0]}]} ]; then
        unset vip_binds_v6[${bound_v6[0]}]
      fi
    fi
  done< <(ip addr list dev dummy0) || error_exit "failed command: ip addr list dev dummy0"
}

apply_remaining_binds(){
  #If any binds remain in vip_binds array, apply them to dummy interface
  if [ "${#vip_binds[@]}" -gt 0 ]; then
    for entry in ${!vip_binds[@]}; do
      ip addr add ${entry}/32 dev dummy0 noprefixroute || warn "failed command: ip addr add ${entry}/32 dev dummy0"
    done
  fi

  #Apply remaining v6 binds in array
  if [ "${#vip_binds_v6[@]}" -gt 0 ]; then
    for v6_entry in ${!vip_binds_v6[@]}; do
      ip -6 addr add ${v6_entry}/128 dev dummy0 noprefixroute || warn "failed command: ip -6 addr add ${v6_entry}/128 dev dummy0"
    done
  fi
}

#Need to implement some other function for the fault state, maybe an alert that gets sent or something
case $STATE in
        "MASTER") get_ranges
                  get_vrrp_ips
                  get_ipvs_vips
                  get_current_binds
                  apply_remaining_binds
                  exit 0
                  ;;
        "BACKUP") get_ranges
                  get_vrrp_ips
                  get_ipvs_vips
                  get_current_binds
                  apply_remaining_binds
                  exit 0
                  ;;
        "FAULT")  echo "${NAME} ${TYPE} is in ${STATE}" >> /var/log/syslog
                  #FROM SYSLOG: SSL_RANGE_1 INSTANCE is in FAULT
                  exit 0
                  ;;
        *)        echo "unknown state"
                  exit 1
                  ;;
esac
