#!/usr/bin/env python

# General : Used to collect SLB interface statistics such as egress/ingress for specific bonded vlan interfaces
# Repo : https://github.llnw.net/llnw/salt-sae/blob/master/sae-slb/v#/scripts/slb_interface_collector.py
# Author : cbranch@llnw.com

import yaml
import os
import sys
import socket
import re
import time
import subprocess
import platform
import json
import argparse
import requests


slbProcStats = '/proc/net/dev'
interface = 'bond0'
slbYamlFile = '/etc/keepalived/slb_conf.yml'
procStatsTemp = '/tmp/slb_interface_collector_stats'
maxTimeOffset = 300
frequency = 60

INFLUX_SERVER = 'opscap-ingest.llnw.net'
INFLUX_PORT = '50002'
INFLUX_DATABASE = 'slbperf'
INFLUX_USER = 'ingest'
INFLUX_PASSWD = 'senddata'
INFLUX_METRICDB = 'slbperf'
DB_PRECISION = 's'
DB_RET = 'oneyear'
INFLUX_METRIC = "SLB-PPS"


def readArgs():
    parser = argparse.ArgumentParser(
        description="Returns the bonded interface network metrics"
    )
    parser.add_argument(
        '-d',
        '--debug',
        help='Print debug output rather than sending data to influx',
        required=False,
        action='store_true',
    )
    args = parser.parse_args()
    return args


def _dpkgVersion(cmd_array):
    try:
        result = subprocess.check_output(cmd_array).split('\n')
        for line in result:
            if re.match("Version:", line):
                result = line.split()[1].strip()
                return result
    except Exception as e:
        print('FAILED: Could not execute cmd \'{}\', {}'.format(
            ' '.join(cmd_array),
            e
        ))
        sys.exit(1)


def fileCheck(file_name):
    if not os.path.exists(file_name):
        print("File not found: \'{}\'".format(file_name))
        return False
    else:
        return True


def obtainFQDN():
    fqdn = socket.gethostname()
    hostCheck = re.match("^[Ss][lL][bB][0-9]+.[a-z]+[0-9]?.llnw.net", fqdn)
    if not hostCheck:
        print("Script must be ran on an SLB! FQDN not recognized.")
        sys.exit(1)
    else:
        return fqdn


def obtainPOP(fqdn):
    try:
        fqdnPOP = fqdn.split('.')[1].strip()
        fqdnPOP = re.sub(r'\d+', '', fqdnPOP)
    except:
        print("Failed to identify POP")
        sys.exit(1)
    return fqdnPOP


def obtainKernel():
    kernelVersion = 'unknown'
    try:
        kernelVersion = str(os.uname()[2])
    except:
        print("Failed to retrieve kernel version!")
    return kernelVersion


def obtainDistro():
    distro = 'unknown'
    distro_ver = 'unknown'
    try:
        distro = platform.linux_distribution()[0]
        distro_ver = platform.linux_distribution()[1]
    except:
        print("Failed to retrieve distribution information")
    return distro, distro_ver


def obtainIpvsVer():
    ipvsVer = 'unknown'
    ipvsCmd = ['dpkg', '-s', 'ipvsadm']
    ipvsVer = _dpkgVersion(ipvsCmd)
    return ipvsVer


def obtainKeepAliveVer():
    keepVer = 'unknown'
    keepCmd = ['dpkg', '-s', 'keepalived']
    keepVer = _dpkgVersion(keepCmd)
    return keepVer


def obtainCPU():
    cpuModel = 'unknown'
    cpuInfoFile = '/proc/cpuinfo'
    if fileCheck(cpuInfoFile):
        with open(cpuInfoFile, 'r') as stream:
            for s in stream:
                if re.match("model name", s):
                    cpuModel = str(s.split(':')[1].strip()).replace(" ", "_")
                    break # exit on first match
    return cpuModel


def yamlLoader():
    with open(slbYamlFile, 'r') as stream:
        yamlObj = yaml.load(stream, Loader=yaml.BaseLoader)
    return yamlObj, len(yamlObj)


def procStatLoader(procStatFile, nativeVlan, cdnVlan):
    procStats = open(procStatFile, 'r')
    nativeVlanPattern = ".*" + re.escape(nativeVlan) + ":.*"
    CdnVlanPattern = ".*" + re.escape(cdnVlan) + ":.*"
    for line in procStats.readlines():
        if re.match(nativeVlanPattern, line):
            nativeVlanStats = tuple(line.lstrip().split())
            continue
        elif re.match(CdnVlanPattern, line):
            cdnVlanStats = tuple(line.lstrip().split())
    procStats.close()
    return nativeVlanStats, cdnVlanStats


def obtainVlan(yamlObjSize, yamlObj):
    fqdn = obtainFQDN()
    for size in range(0, yamlObjSize):
        if yamlObj.keys()[size] == "config_version":
            if yamlObj["config_version"] == "2_ecmp":
                try:
                    nativeVlanBond = interface + "." + str(yamlObj["slb_configs"][fqdn]["native_vlan"])
                    cdnVlanBond = interface + "." + str(yamlObj["slb_configs"][fqdn]["cdn_vlans"].keys()[0])
                    return fqdn, nativeVlanBond, cdnVlanBond
                except KeyError:
                    print("Invalid yaml key detected. Check formatting...")
                    sys.exit(1)
            else:
                print("Version 2 yaml detected, but ecmp not set...exiting")
                sys.exit(1)
    try:
        nativeVlanBond = interface + "." + str(yamlObj[fqdn]["slb_config"]["native_vlan"])
        cdnVlanBond = interface + "." + str(yamlObj[fqdn]["slb_config"]["cdn_vlans"].keys()[0])
        return fqdn, nativeVlanBond, cdnVlanBond
    except KeyError:
        print("Unable to parse yaml. Check formatting...")
        sys.exit(1)


def statMap(vlanMap):
    intStat = {}
    procMap = [
        "Interface",
        "Rx_bytes",
        "Rx_packets",
        "Rx_errors",
        "Rx_drop",
        "Rx_fifo",
        "Rx_frame",
        "Rx_compressed",
        "Rx_multicast",
        "Tx_bytes",
        "Tx_packets",
        "Tx_errors",
        "Tx_drop",
        "Tx_fifo",
        "Tx_colls",
        "Tx_carrier",
        "Tx_compressed"
    ]
    for key in procMap:
        intStat[key] = vlanMap[int(procMap.index(key))]
    return intStat


def loadJsonFile():
    with open('/tmp/slb_interface_collector_stats') as json_file:
        return json.load(json_file)


def saveJsonFile(data):
    with open(procStatsTemp, 'w') as stream:
        json.dump(data, stream)
    return True


def _constructor():
    # Load yaml file
    loadedyml, lenyml = yamlLoader()

    # Generate fqdn, and specific bonded VLANs
    fqdn, nativeVlanBond, cdnVlanBond = obtainVlan(lenyml, loadedyml)

    # Set POP
    pop = obtainPOP(fqdn)

    # Gather system info
    distro, distro_ver = obtainDistro()
    kernelVersion = obtainKernel()
    ipvsVersion = obtainIpvsVer()
    keepaliveVersion = obtainKeepAliveVer()
    cpuVersion = obtainCPU()

    # Generate vlan stat objects & timestamp
    epochTime = time.time()
    nativeVlanStats, cdnVlanStats = procStatLoader(slbProcStats, nativeVlanBond, cdnVlanBond)

    # Use vlan stat objects to build dictionary MAPs
    nativeMap = statMap(nativeVlanStats)
    cdnMap = statMap(cdnVlanStats)

    # Build full object
    structMap = {'timestamp': epochTime, "nativeVlan": nativeMap, "cdnVlan": cdnMap}
    tagMap = {
        'fqdn':fqdn, 'pop':pop,
        'distribution':distro, 'distribution_version':distro_ver,
        'kernel_version':kernelVersion, 'ipvs_version':ipvsVersion,
        'keepalive_version':keepaliveVersion, 'cpu':cpuVersion
        }
    return structMap, nativeVlanBond, cdnVlanBond, tagMap


def _calcDeltas(objMap, initMap):
    deltaObj = {}
    dTime = objMap["timestamp"] - initMap["timestamp"]
    if not dTime:
        return False, 0
    deltaMap = [
        "Rx_bytes",
        "Rx_packets",
        "Rx_errors",
        "Rx_drop",
        "Rx_fifo",
        "Rx_frame",
        "Rx_compressed",
        "Rx_multicast",
        "Tx_bytes",
        "Tx_packets",
        "Tx_errors",
        "Tx_drop",
        "Tx_fifo",
        "Tx_colls",
        "Tx_carrier",
        "Tx_compressed"
    ]
    for key in deltaMap:
        for vlan in ["nativeVlan", "cdnVlan"]:
            # Get the avg value per second
            deltaObj[vlan+"_avg_per_sec"+key] = (int(objMap[vlan][key])-int(initMap[vlan][key]))/dTime

            # Extrapolate the sum from the average (this is the same logic influxdb does)
            deltaObj[vlan+"_sum"+key] = deltaObj[vlan+"_avg_per_sec"+key]*frequency
            if deltaObj[vlan+"_avg_per_sec"+key] < 0:
                return False, dTime

    return deltaObj, dTime


def buildInfluxData(deltaObj, tagMap, timestamp):
    tagValues = ','.join({"{}={}".format(k,v) for k, v in tagMap.items()})
    dataValues = ','.join({"{}={}".format(k,v) for k, v in deltaObj.items()})

    influxData = '{},{} {} {}'.format(
        INFLUX_METRIC,
        tagValues,
        dataValues,
        timestamp
    )

    return influxData


def sendToInfluxDB(influxDataLines):
    # Send the Data to influxDB
    r = requests.post('https://%s:%s/write?db=%s&precision=%s&rp=%s' % (INFLUX_SERVER, INFLUX_PORT, INFLUX_DATABASE, DB_PRECISION, DB_RET), data=(influxDataLines), timeout=30, auth=requests.auth.HTTPBasicAuth(INFLUX_USER, INFLUX_PASSWD))
    print r.text
    print r.url
    print r.status_code
    print len(influxDataLines)


# Main function
def main():
    args = readArgs()

    if not (fileCheck(slbYamlFile) or fileCheck(slbProcStats)):
        sys.exit(1)
    objMap, _, _, tagMap = _constructor()
    deltaObj = {}
    # Read current file
    if fileCheck(procStatsTemp):
        initMap = loadJsonFile()
        if ((objMap["timestamp"] - initMap["timestamp"] <= maxTimeOffset) and (objMap["timestamp"] - initMap["timestamp"] != 0)):
            deltaObj, deltaTime = _calcDeltas(objMap, initMap)
    else:
        saveJsonFile(objMap)
        sys.exit(1)

    alignedTimestamp = int(objMap["timestamp"]/frequency)*frequency
    if len(deltaObj.keys()) <= 0:
        saveJsonFile(objMap)
        sys.exit(1)
    else:
        influxData = buildInfluxData(deltaObj, tagMap, alignedTimestamp)

    if args.debug:
        print("==== Debug ====")
        print(influxData)
    else:
        sendToInfluxDB(influxData)
    saveJsonFile(objMap)


if __name__ == '__main__':
    main()
