from __future__ import print_function
import os
import platform
import re
import sys
from argparse import ArgumentParser

import socket

import yaml
import logging
import logging.handlers

# configure logging
log = logging.getLogger(__name__)
log.setLevel(logging.WARNING)
handler = logging.handlers.SysLogHandler(address='/dev/log')
formatter = logging.Formatter('%(module)s.%(funcName)s: %(message)s')
handler.setFormatter(formatter)

log.addHandler(handler)


# file paths
slb_global_config_file = r"/etc/keepalived/global_conf.yml"
macping = '/etc/keepalived/scripts/macping'

# get the hostname
hostname = platform.node()

# dns cache
local_dns_cache = dict()

# add reference to socket.getaddrinfo
prv_getaddrinfo = socket.getaddrinfo


def override_dns(domain, ip):
    """
    Helper: Updates local_dns_cache with domain and override IP
    Args:
        domain: (str) FQDN of domain to override with ip
        ip: (str) IPv4 or v6 IP Address
    """
    local_dns_cache[domain] = ip


def new_getaddrinfo(*args):
    """
    Overrides socket.getaddrinfo to support pinning an IP to a domain
    Args:
        *args: (tuple) IP Address 4-tuple (ex: ('test.vo.llnwd.net', 443, 0, 1))

    Returns:
        (list) of tuples (ex: [(2, 1, 6, '', ('111.119.11.41', 443))] )
    """
    log.debug('args: {}'.format(args))
    if args[0] in local_dns_cache:
        log.debug("Forcing FQDN: {} to IP: {}".format(args[0], local_dns_cache[args[0]]))
        _res = prv_getaddrinfo(local_dns_cache[args[0]], *args[1:])
    else:
        _res = prv_getaddrinfo(*args)
    log.debug('getaddrinfo: results: {}'.format(_res))
    return _res


# override socket.getaddrinfo with our version
socket.getaddrinfo = new_getaddrinfo


def get_pop():
    """
    Parse the POP from the hostname
    Returns:
        (str) POP name parsed from hostname
    """
    if hostname.startswith('slb'):
        try:
            return hostname.split('.')[1]
        except:
            raise ValueError('invalid SLB hostname: {} -- a valid SLB hostname must be fully qualified')
    else:
        raise ValueError('invalid SLB hostname: {} -- a valid SLB hostname start with "slb"'.format(hostname) )


def get_global_config_option(value, raise_error_on_not_found=False):
    """
    Get the specified option value from the global config file.
    Note that this function's logic is a port of global_conf_get.pl
    Args:
        value: (str) global config option to retrieve
        raise_error_on_not_found: (bool) raise an error if option value not found (default False)

    Returns:
        (str) value of global config option if found
    """
    _data = None
    with open(slb_global_config_file) as f:
        _data = yaml.load(f)

    if not _data or value not in _data:
        raise ValueError('no entry for option: {} found in {}'.format(value, slb_global_config_file))

    _data = _data.get(value)

    # try fqdn first
    if hostname in _data:
        return _data[hostname]

    # next try pop
    _pop = get_pop()
    if _pop in _data:
        return _data[_pop]

    # try default
    if 'default' in _data:
        return _data['default']

    if raise_error_on_not_found:
        raise ValueError('ERROR: A default value is missing for option {}'.format(value))

    return _data


def get_ping_interface(ip_address):
    """
    Helper: Get the interface associated with specified IP Address from route table.  Note that command:
    "ip route show to match" is used which will find an entry in the routing table for the specified IP.
    Args:
        ip_address: (str) IP Address in v4 dot or v6 quad notation

    Returns:
        (str) interface name for specified IP Address
    """
    s = os.popen("ip route show to match " + ip_address + " scope link")
    o = s.read()
    if s.close() is None:
        # the results could be an empty string if no match is found
        # a misconfigured CDN vlan subnet is possible
        _results = o.split()
        if _results and len(_results) > 2:
            if re.search(r'^(bond|eth)', _results[2]):
                # only process if the interface is bond or eth
                # probably want to change this to 'bond' only
                return _results[2]

            raise ValueError('unsupported interface: {} found for {}'.format(_results[2], ip_address))

    raise ValueError('no interface found for {}'.format(ip_address))


def touch(path):
    """
    Emulate touch command
    Args:
        path: (str) file path to create/touch
    """
    with open(path, 'a'):
        os.utime(path, None)


def dsr_ping(host, vip, retries=3):
    """
    Performs a DSR ping using macping.
    Args:
        host: (str) host IP Address
        vip: (str) VIP
        retries: (int) number of retries (default 3)

    Returns:
        (int) 0 on successful else macping's error level
    """
    if not (host and vip):
        raise ValueError('parameters host and vip are required')

    # get the interface associated with host IP Address
    _ping_interface = get_ping_interface(host)

    _macping_command = ("{macping} -I {ping_interface} {debug} -b -q -c 1 -w 3 {vip} {host} || {macping} -I "
                        "{ping_interface} {debug} -b -l -q -c 1 -w 3 {vip} {host}").format(
            macping=macping, ping_interface=_ping_interface, vip=vip, host=host,
            debug='-d' if log.isEnabledFor(logging.DEBUG) else '')

    log.debug('macping command: {}'.format(_macping_command))

    # Check DSR
    _exit_level = 0
    for x in range(retries):
        s = os.popen(_macping_command)
        _exit_status = s.close()
        if _exit_status is None:
            # success - return True
            return 0

        # get the exit status
        _exit_level = os.WEXITSTATUS(_exit_status)
        log.debug('dsr_ping exit level: {}'.format(_exit_level))

    # failed
    return _exit_level


class ArgParser(ArgumentParser):
    def error(self, message):
        """
        override ArgumentParser.error() to support specific app logging
        Args:
            message: (str) error message
        """
        print('usage error: {}'.format(message), file=sys.stderr)
        print(self.format_usage(), file=sys.stderr)
        # log a warning so that the usage error is logged to syslog
        log.warning('usage error: {}'.format(message))

        # must exist here -- do not return
        exit(1)
