#!/usr/bin/env perl
use strict;
use warnings;
use Sys::Hostname;
use Data::Dumper;
use Regexp::IPv6 qw($IPv6_re);
use POSIX;
use YAML qw(LoadFile Dump DumpFile);
use Net::CIDR;
use List::Util qw(max);
use Storable qw(dclone);

# Adjust these to set the Max and Min limits for
# normalized load balance weights.
my $g_max_weight = 100;
my $g_min_weight = 0;

# This is the prefix to where this script's folder is located and where config
# and data files will be loaded from. Change this if you want to test run the
# script from your home folder (requires the content of /etc/keepalived to be
# copied there).
my $g_prefix = "/etc/keepalived";

# This is the prefix where generated files will be saved. In case of dry run it
# is /tmp/keepalived, otherwise the same as g_prefix.
my $g_saveprefix = $g_prefix;

# Don't make any changes, print to STDOUT and save files to /tmp/keepalived.
my $g_dryrun = 0;

# supported health checks
my %supported_health_checks = (
    '443'  => {'TYPE' => 'TCP_CHECK',  'PORT' => '443',  'SCRIPT' => undef},
    'http' => {'TYPE' => 'MISC_CHECK', 'PORT' => '80',   'SCRIPT' => 'http_check.sh'},
    '80'   => {'TYPE' => 'MISC_CHECK', 'PORT' => '80',   'SCRIPT' => 'http_check.sh'}
);


main();

sub main {
  my $hostname = hostname;
  my $slb_state = "default";
  my $ecmp_enabled = 0;

  if ( $ARGV[0] ) {
    if ( $ARGV[0] eq '-h' || $ARGV[0] eq '--help' ) {
      usage();
    } elsif ( $ARGV[0] eq '-b' ) {
      $slb_state = "backup";
    } elsif ( $ARGV[0] eq '-m' ) {
      $slb_state = "master";
    } elsif ( $ARGV[0] eq '-d' ) {
      $g_saveprefix = "/tmp/keepalived";
      $g_dryrun = 1;
      print STDOUT "Dry run, saving files to ${g_saveprefix}\n";
      my @dirs = ("${g_saveprefix}", "${g_saveprefix}/config_containers",
        "${g_saveprefix}/server_weights", "${g_saveprefix}/server_states");
      for my $dir ( @dirs ) {
        if ( ! -d $dir ) {
          mkdir($dir, 0755) || die "mkdir $dir: $!";
        }
      }
    } elsif ( $ARGV[0] =~ /^-/) {
      usage();
    }
  }

  unless ( $hostname =~ /^slb\d\d/ ) {
    die "This script is only meant for software load balancers\n";
  }

  my $pop = get_my_pop($hostname);
  my $yml_file = load_yml_file($pop);
  if ($yml_file->{config_version} && (($yml_file->{config_version} eq "2") || ($yml_file->{config_version} eq "2_ecmp")) ) {

    if ($yml_file->{config_version} eq "2_ecmp") {
      $ecmp_enabled = 1;
    }

    # Got new config file format. Take this as input and distribute VIPs across
    # SLBs according to pop_config section. The output is compatible with the
    # previous slb_conf.yml format so the functions below remain the same.
    sanity_check_vips($yml_file);
    $yml_file = gen_yml_conf_from_v2($yml_file, $hostname, $ecmp_enabled);
    if ($g_dryrun == 1) {
      print "\nGot config version 2.\n";
      print "Saving generated intermittent config to /tmp/keepalived/slb_conf.yml.\n";
      DumpFile "/tmp/keepalived/slb_conf.yml", $yml_file;
    }
  } else {
    sanity_check_vips($yml_file->{$hostname});
  }

  my $yml_conf = $yml_file->{$hostname} or die "No config for $hostname found in ${g_prefix}/slb_conf.yml $!";
  my $down_servers = load_down_servers($pop);

  my $v6_enabled = $yml_conf->{slb_config}->{v6_address};
  my $v4_bgp_enabled = $yml_conf->{slb_config}->{bgp_config}->{v4_neighbor};
  my $v6_bgp_enabled = $yml_conf->{slb_config}->{bgp_config}->{v6_neighbor};

  # if v6 BGP is configured but v4 BGP is not, then exit
  if ( defined($v6_bgp_enabled) and !defined($v4_bgp_enabled) ) {
    die "ERROR: v6 BGP requires v4 BGP, v4 BGP is not configured";
  }

  # if v6 BBP is configured but no native VLAN v6 IP Address configured, then exit
  if ( defined($v6_bgp_enabled) and !defined($v6_enabled) ) {
    die "ERROR: v6 BGP requires v6, v6 IP Address is not configured";
  }

  gen_real_configs($yml_conf, $down_servers, $v6_enabled);
  gen_virtual_configs($yml_conf, $v6_enabled, $ecmp_enabled);
  gen_vrrp_configs($pop,$yml_conf, $slb_state, $v6_enabled, $ecmp_enabled);
}

sub usage {
  print STDERR "Usage: gen_slb_configs.pl [<options>]\n";
  print STDERR "   -b: set vip priority to 50 (Make all vips Backup)\n";
  print STDERR "   -m: set vip priority to 254 (Make all vips Master)\n";
  print STDERR "   -d: dry run with debug output (Saving files to /tmp/keepalived)\n";
  print STDERR "\n";
  print STDERR "Example:\n";
  print STDERR "   ./gen_slb_configs.pl (builds real server, vrrp, and virtual_server confs)\n";
  print STDERR "   ./gen_slb_configs.pl -b (changes priority of existing vips to backup)\n";
  print STDERR "   ./gen_slb_configs.pl -m (changes priority of existing vips to master)\n";
  exit(1);
}

sub get_my_pop {
  my $arg = shift || die "Missing pop arg $!";

  my $local_pop = (split '\.',$arg)[1];
  return $local_pop;
}

sub load_yml_file {
  my $pop = shift || die "Did not receive a valid pop $!";
  my $yml = LoadFile("${g_prefix}/slb_conf.yml");
  return $yml;
}

sub load_manual_changes {
  my $file = "${g_prefix}/config_containers/overrides.yml";
  my $yml_overrides;

  if ( -f "$file" ) {
    $yml_overrides = LoadFile("${g_prefix}/config_containers/overrides.yml");
  } else {
    return;
  }
  return $yml_overrides;
}

sub load_down_servers {
  my $pop = shift || die "Did not receive a valid pop $!";

  # trim to just the alphabetic site name
  $pop =~ s/\d+$//;

  my $file = "${g_prefix}/downservers.dat";
  my %down_servers;

  if ( -f $file ) {
    open(my $fh, '<', $file) or die "open $file: $!";
    while (my $line = <$fh>) {
      chomp($line);
      if ($line =~ /^\S+\.$pop\.llnw\.net$/) {
        $down_servers{$line} = 1;
      }
    }
    close($fh) or die "close $file: $!";
  }

  return \%down_servers;
}

sub get_ip_version {
  my $ip = shift;
  my $ip_version;

  if ( $ip =~ /^(([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\.){3}([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])$/ ) {
    $ip_version = "v4";
  } elsif ( $ip =~ /^$IPv6_re$/ ) {
    $ip_version = "v6";
  } else {
    die "$ip is not valid ipv4 or ipv6";
  }

  return $ip_version;
}

# Save the server weight to a file that http_check.sh can read.
# http_check.sh might change the current weight when throttling
# the traffic so don't change this value if lower than default weight.
sub save_server_weight {
  my $ip = shift || die "Did not receive ip address $!";
  my $weight = shift;
  my $state = shift;
  my $filename = "${g_saveprefix}/server_weights/${ip}";
  my $data;

  if (-e $filename) {
    $data = LoadFile($filename);
    # if $weight defined then apply it
    if (defined $weight) {
      $data->{default} = $weight;
      if ($data->{current} > $weight) {
        $data->{current} = $weight;
      }
    }

    # if state does not exists apply default
    if ( !exists($data->{state}) ) {
      $data->{'state'} = 'active';
    }

    # if $state defined then apply it
    if (defined $state) {
      $data->{'state'} = $state;
    }
  } else {
    # file doesn't exists so set defaults
    $data = {
        default => $weight || 100,
        current => $weight || 100,
        state   => $state || 'active'
    };
  }
  DumpFile( $filename, $data );
}

#
# Build a hash table with key='server name' and value='carp weight'
# from load factor file.
#
sub gen_raw_weights {
  my $file = shift || die "Did not receive filename $!";
  my %all_weights = ();

  open (my $fh, "<", $file) or return %all_weights;

  while (my $line =<$fh>) {
    chomp ($line);
    my($name, $weight) = split(" ", $line);
    if ( !defined($name) || $name eq "" || !defined($weight) || $weight eq "" ) {
      next;
    }
    $all_weights{$name} = $weight;
  }

  return %all_weights;
}

#
# Collect real servers used by all vip ranges and map load factor to weight.
# Save each weight in separate file for use by http_check.sh.
# Return a hashtable with all weights, key='ip' value='normalized weight'.
#
sub gen_normalized_weights {
  my $yml_conf = shift || die "Did not receive yml ref $!";
  my $yml_vips = $yml_conf->{vips};
  my @keys = keys %$yml_vips;

  my %all_raw_weights = ();
  my $cfg_val = `global_conf_get.pl loadfactors`;
  if ($cfg_val eq "on") {
    %all_raw_weights = gen_raw_weights("${g_prefix}/load_factors2.dat");
  }

  my %normalized_weights = ();
  my @weights = ();

  for my $vip_name ( @keys ) {
    my $vip_ip = (split '/',$yml_vips->{$vip_name}->{vip_addr})[0];
    my $vip_ip_cidr = $yml_vips->{$vip_name}->{vip_addr};
    my $vip_port = $yml_vips->{$vip_name}->{vip_port};
    my $srv_port = $yml_vips->{$vip_name}->{srv_port};
    my $health_ck = $yml_vips->{$vip_name}->{vip_healthck};
    my $srv_group = $yml_vips->{$vip_name}->{vip_servers};
    my $geo_state = $yml_vips->{$vip_name}->{geo_state};
    my $config_gen = $yml_conf->{slb_config}->{config_gen};
    my $vip_priority = $yml_vips->{$vip_name}->{vip_priority};

    if ( $config_gen ne "auto" ) {
      print "$vip_name is set to ${config_gen}, skipping\n";
      next;
    }

    if ( ! defined($srv_port) || $srv_port eq "" ) {
      $srv_port = "0"
    }

    # 'v4' or 'v6'
    my $ip_version = get_ip_version($vip_ip);

    # Array of { 'ip' => [ip address], 'name' => [server name] } hashes
    my @footprint = get_servers_from_dat_file($srv_group, $ip_version);

    for my $server (@footprint) {
      my $raw_weight = $all_raw_weights{$server->{'name'}};
      my $ip = $server->{'ip'};
      if ( !defined($raw_weight) ) {
        # "-1" is a placeholder for normalized max when value missing.
        $normalized_weights{$ip} = -1;
        next;
      }
      my $weight = int($raw_weight + 0.5);
      push(@weights, $weight);
      $normalized_weights{$ip} = $weight;
    }
  } # end for each vip range

  my $max = max(@weights);

  # If $max is NULL or 0 we probably did not have a load_factors.dat
  # file to read. Set to default max so that we get default old
  # behavior in this case.
  if ( !defined($max) || $max == 0 ) {
    $max = $g_max_weight;
  }

  # Replace raw weight with normalized weight.
  # The highest raw value in the vip ranges will receive a weight of $g_max_weight.
  # The lowest raw value will receive a weight not lower than $g_min_weight.
  # Anything in between will be scaled linearly.
  foreach my $ip (keys %normalized_weights)
  {
    my $raw_weight = $normalized_weights{$ip};
    if ( $raw_weight < 0 ) {
      $raw_weight = $max;
    }
    my $normalized = int($raw_weight * ($g_max_weight - $g_min_weight) / $max + $g_min_weight);
    $normalized_weights{$ip} = $normalized;
    save_server_weight($ip, $normalized);
  }

  return %normalized_weights;
}

sub gen_real_configs {
  my $yml_conf = shift || die "Did not receive yml ref $!";
  my $down_servers = shift // die "Did not receive down servers";
  my $v6_enabled = shift // 0;

  my $yml_vips = $yml_conf->{vips};
  my %normalized_weights = gen_normalized_weights($yml_conf);

  my @keys = keys %$yml_vips;
  for my $vip_name ( @keys ) {
    my $vip_ip = (split '/',$yml_vips->{$vip_name}->{vip_addr})[0];
    # 'v4' or 'v6'
    my $ip_version = get_ip_version($vip_ip);

    # if v6 is disabled (not configured for native vlan) then skip v6 VIPS
    if ( $ip_version eq "v6" and !$v6_enabled) {
      print STDERR "WARNING: v6 is not enabled, ignoring VIP: ${vip_name} in real server config\n";
      next;
    }

    my $vip_ip_cidr = $yml_vips->{$vip_name}->{vip_addr};
    my $vip_port = $yml_vips->{$vip_name}->{vip_port};
    my $srv_port = $yml_vips->{$vip_name}->{srv_port};
    my $health_ck = $yml_vips->{$vip_name}->{vip_healthck};
    my $srv_group = $yml_vips->{$vip_name}->{vip_servers};
    my $geo_state = $yml_vips->{$vip_name}->{geo_state};
    my $config_gen = $yml_conf->{slb_config}->{config_gen};
    my $vip_priority = $yml_vips->{$vip_name}->{vip_priority};

    my $interface = $yml_conf->{slb_config}->{interface};
    my $vlan = $yml_conf->{slb_config}->{native_vlan};

    my $active_vips = get_states($interface,$vlan);

    if ( $config_gen ne "auto" ) {
      print "$vip_name is set to ${config_gen}, skipping\n";
      next;
    }

    if ( ! defined($srv_port) || $srv_port eq "" ) {
      $srv_port = "0"
    }

    # Array of { 'ip' => [ip address], 'name' => [server name] } hashes
    my @footprint = get_servers_from_dat_file($srv_group,$ip_version);

    # Add normalized weight to footprint hash.
    for my $server (@footprint) {
      my $ip = $server->{'ip'};
      my $weight = $normalized_weights{$ip};
      my $real_hostname = $server->{'name'};
      my $server_state;

      if ( !defined($weight) ) {
        $weight = $g_max_weight;
      }
      $server->{'weight'} = $weight;

      # if server is in down_servers, mark state as 'drain' otherwise 'active'
      $server_state = $down_servers->{$real_hostname} ? 'drain': 'active';

      # VRRP only
      if ( defined $vip_priority && $vip_priority < 200) {
        # this is for VRRP support
        # if VIP is active but vip priority is a `backup` (< 200), need to set state to `active`
        # so DSR health checking gets added back in
        $server_state = $active_vips->{$vip_ip_cidr} ? 'active' : 'backup';
      }

      # save the state to the server weights file
      save_server_weight($ip, undef, $server_state);
    }

    build_server_configs(\@footprint,$health_ck,$vip_name,$vip_ip,$vip_ip_cidr,$srv_port,$ip_version,$vip_priority,$yml_conf,$down_servers);
  }
}

sub get_servers_from_dat_file {
  my $server_group = shift || die "Missing server group arg $!";
  my $ip_version = shift || die "Missing protocol arg $!";
  my @servers;

  open( my $sfh, '<', "${g_prefix}/servers.dat") or die "Failed to open servers.dat for reading: $!";
  if ( $ip_version eq "v4" ) {
    while (my $line = <$sfh>) {
      my ($group, $addr, $hostname) = (split(' ', $line))[1..3];
      next unless ($group eq $server_group);
      next unless ($addr =~ /^(([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\.){3}([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])$/);
      push @servers, { 'ip' => $addr, 'name' => $hostname };
    }
  }

  if ( $ip_version eq "v6" ) {
    while (my $line = <$sfh>) {
      my ($group, $addr, $hostname) = (split(' ', $line))[1,5,3];
      next unless ($group eq $server_group);
      next unless ($addr =~ /^$IPv6_re$/);
      push @servers, { 'ip' => $addr, 'name' => $hostname };
    }
  }

  close ($sfh) or die "Failed to close servers.dat $!";
  return @servers;
}

sub get_states {
  #Currently vips in a backup state do not get DSR health checked to save CPU resources and avoid double checking.  For
  #vips initially in a backup state, we want to verify at the time of writing out the configs that they are still backup,
  #meaning we are not in a failure scenario.  If we are, we need to re-enable DSR checks for the backup vips
  my $interface = shift || die "Did not recieve interface $!";
  my $vlan = shift || die "Did not recieve vlan $!";
  my %bound_address = ();

  open( my $vlan_int, "ip address list dev ${interface}.${vlan} 2> /dev/null |") or warn "error listing addresses on ${interface}.${vlan} $!";
    while ( my $line = <$vlan_int> ) {
      if ( $line =~ /inet/ ) {
        $bound_address{(split ' ', $line )[1]} = '1';
      }
    }
  close $vlan_int;
  return \%bound_address
}

sub build_server_configs {
  my $servers = shift || die "Missing footprint array $!";
  my $health_checks = shift || die "Missing health check array $!";
  my $vip_name = shift || die "Missing vip name $!";
  my $vip_ip = shift || die "Missing vip ip $!";
  my $vip_ip_cidr = shift || die "Missing vip ip $!";
  my $srv_port = shift // die "Missing vip port $!";
  my $ip_version = shift || die "Missing ip version $!";
  my $vip_priority = shift || die "Missing vip priority $!";
  my $yml_conf = shift || die "Missing yml reference $!";
  my $down_servers = shift // die "Missing down servers";
  my $manual_overrides = load_manual_changes();

  my $vip_tmp_file = "${g_saveprefix}/config_containers/${vip_name}_real_servers.new";

  open( my $rfh,'>',"$vip_tmp_file" ) or die "Failed to open $vip_tmp_file $!";
    #Default real servers for v4 and v6 else keepalived freaks out when no servers are referenced initially
    if ( $ip_version eq "v4" ) {
      print $rfh "real_server 192.168.255.255 ${srv_port} {\n" or die "Failed to write to $vip_tmp_file $!";
      print $rfh "  weight 0\n" or die "Failed to write to $vip_tmp_file $!";
      print $rfh "  MISC_CHECK {\n" or die "Failed to write to $vip_tmp_file $!";
      print $rfh "    misc_path \"${g_prefix}/scripts/http_check.sh 192.168.255.255 ${vip_ip}\"\n" or die "Failed to write to $vip_tmp_file $!";
      print $rfh "    misc_timeout 5\n" or die "Failed to write to $vip_tmp_file $!";
      print $rfh "    warmup 1200\n" or die "Failed to write to $vip_tmp_file $!";
      print $rfh "  }\n" or die "Failed to write to $vip_tmp_file $!";
      print $rfh "}\n" or die "Failed to write to $vip_tmp_file $!";
    }
    if ( $ip_version eq "v6" ) {
      print $rfh "real_server fc00:aaaa::1 ${srv_port} {\n" or die "Failed to write to $vip_tmp_file $!";
      print $rfh "  weight 0\n" or die "Failed to write to $vip_tmp_file $!";
      print $rfh "  MISC_CHECK {\n" or die "Failed to write to $vip_tmp_file $!";
      print $rfh "    misc_path \"${g_prefix}/scripts/http_check.sh fc00:aaaa::1 ${vip_ip}\"\n" or die "Failed to write to $vip_tmp_file $!";
      print $rfh "    misc_timeout 5\n" or die "Failed to write to $vip_tmp_file $!";
      print $rfh "    warmup 1200\n" or die "Failed to write to $vip_tmp_file $!";
      print $rfh "  }\n" or die "Failed to write to $vip_tmp_file $!";
      print $rfh "}\n" or die "Failed to write to $vip_tmp_file $!";
    }

    #Loop through all of the servers in servers.dat and configure them up in the real_servers file
    # foreach my $real_hostname (sort(keys(%{$servers}))) {
    foreach my $server ( @{ $servers } ) {
      my $real_ip = $server->{'ip'};
      my $weight = $server->{'weight'};
      my $real_hostname = $server->{'name'};
      if ( defined($manual_overrides->{$real_ip}) and $manual_overrides->{$real_ip} eq "remove" ) {
        next;
      }
      print $rfh "real_server ${real_ip} ${srv_port} {\n" or die "Failed to write to $vip_tmp_file $!";
      print $rfh "  weight ${weight}\n" or die "Failed to write to $vip_tmp_file $!";

      # process each specified health check -- note that health checks are simply specified by name
      my $hc_type = undef;
      my $hc_script = undef;
      my $hc_port = undef;
      foreach my $hc ( @{ $health_checks } ) {
        # check if health script is in $supported_health_checks
        if (exists($supported_health_checks{$hc})) {
          # if exists then use the values
          $hc_type = $supported_health_checks{$hc}{'TYPE'};
          $hc_script = $supported_health_checks{$hc}{'SCRIPT'};
          $hc_port = $supported_health_checks{$hc}{'PORT'};
        } else {
          # $hc is not in $supported_health_checks
          if ($hc =~ /^\d+$/) {
            # $hc is a number; use TCP_CHECK
            $hc_type = 'TCP_CHECK';
            $hc_port = $hc;
          } else {
            # $hc is not a number; use MISC_CHECK and script
            $hc_type = 'MISC_CHECK';
            $hc_script = "${hc}_check";
          }
        }

        if ($hc_type eq 'TCP_CHECK') {
          print $rfh "  TCP_CHECK {\n" or die "Failed to write to $vip_tmp_file $!";
          print $rfh "    connect_port ${hc_port}\n" or die "Failed to write to $vip_tmp_file $!";
          print $rfh "    connect_timeout 8\n" or die "Failed to write to $vip_tmp_file $!";
          print $rfh "    retry 0\n" or die "Failed to write to $vip_tmp_file $!";
          print $rfh "    delay_before_retry 0\n" or die "Failed to write to $vip_tmp_file $!";
          print $rfh "  }\n" or die "Failed to write to $vip_tmp_file $!";
        } elsif ($hc_type eq 'MISC_CHECK') {
          # note this assumes the health check script exists and is both executable and functional

          print $rfh "  MISC_CHECK {\n" or die "Failed to write to $vip_tmp_file $!";
          print $rfh "    misc_path \"${g_prefix}/scripts/${hc_script} ${real_ip} ${vip_ip}\"\n" or die "Failed to write to $vip_tmp_file $!";
          print $rfh "    misc_timeout 12\n" or die "Failed to write to $vip_tmp_file $!";
          print $rfh "    misc_dynamic\n" or die "Failed to write to $vip_tmp_file $!";
          print $rfh "  }\n" or die "Failed to write to $vip_tmp_file $!";
        }

      }
      print $rfh "}\n" or die "Failed to write to $vip_tmp_file $!";
    }
  close ($rfh) or die "Failed to close ${vip_tmp_file} $!";
}

sub gen_virtual_configs {
  my $yml_conf = shift || die "Did not receive yml ref $!";
  my $v6_enabled = shift // 0;
  my $ecmp_enabled = shift // 0;
  my $yml_vips = $yml_conf->{vips};

  my $virtual_tmp_file = "${g_saveprefix}/config_containers/virtual_server.new";
  my $virtual_file = "${g_prefix}/config_containers/virtual_server.conf";
  my $interface = $yml_conf->{slb_config}->{interface};
  my $bgp_vlan = $yml_conf->{slb_config}->{native_vlan};

  open( my $vfh,'>',"$virtual_tmp_file" ) or die "Failed to open $virtual_tmp_file $!";

  my @keys = keys %$yml_vips;
  for my $vip_name ( @keys ) {
    my $vip_ip = (split '/',$yml_vips->{$vip_name}->{vip_addr})[0];
    my $ip_version = get_ip_version($vip_ip);

    # if v6 is disabled (not configured for native vlan) then skip v6 VIPS
    if ( $ip_version eq "v6" and !$v6_enabled) {
      print STDERR "WARNING: v6 is not enabled, ignoring VIP: ${vip_name} in virtual config\n";
      next;
    }

    my $vip_ip_cidr = $yml_vips->{$vip_name}->{vip_addr};
    my $vip_port = $yml_vips->{$vip_name}->{vip_port};
    my $vip_range = $yml_vips->{$vip_name}->{vip_range};
    my $vip_sticky = $yml_vips->{$vip_name}->{vip_sticky};
    my $config_gen = $yml_conf->{slb_config}->{config_gen};
    my @vip_broadcast = Net::CIDR::cidr2range("$vip_ip_cidr");
    my $vip_broadcast = (split '-',$vip_broadcast[0])[1];
    my $vip_hash_method = $yml_vips->{$vip_name}->{vip_hash};

    # No hash method will be assigned if v1 config is present, default to wrr
    if ( !$vip_hash_method ) {
      $vip_hash_method = 'wrr';
    }

    if ( $config_gen ne "auto" ) {
      print "$vip_name is set to ${config_gen}, skipping\n";
      next;
    }
    print $vfh "virtual_server_group ${vip_name} {\n" or die "Failed to write to $virtual_tmp_file $!";
    if ( defined($vip_range) ) {
      if (ref($vip_port) eq 'ARRAY') {
        foreach my $port ( @{ $vip_port } ) {
          print $vfh "  ${vip_ip}-${vip_range} ${port}\n" or die "Failed to write to $virtual_tmp_file $!";
        }
      } else {
        print $vfh "  ${vip_ip}-${vip_range} ${vip_port}\n" or die "Failed to write to $virtual_tmp_file $!";
      }
    } else {
      if (ref($vip_port) eq 'ARRAY') {
        foreach my $port ( @{ $vip_port } ) {
          print $vfh "  ${vip_ip} ${port}\n" or die "Failed to write to $virtual_tmp_file $!";
        }
      } else {
        print $vfh "  ${vip_ip} ${vip_port}\n" or die "Failed to write to $virtual_tmp_file $!";
      }
    }
    print $vfh "}\n";
    print $vfh "virtual_server group ${vip_name} {\n" or die "Failed to write to $virtual_tmp_file $!";
    print $vfh "  lvs_sched ${vip_hash_method}\n" or die "Failed to write to $virtual_tmp_file $!";
    if ( ${vip_hash_method} eq "mh" ) {
        print $vfh "  flag-2    # Tells mh scheduler to use source port in hash\n"
                or die "Failed to write to $virtual_tmp_file $!";
    }
    print $vfh "  lvs_method DR\n" or die "Failed to write to $virtual_tmp_file $!";
    if ( $ip_version eq "v4" ) {
      print $vfh "  delay_loop 20\n" or die "Failed to write to $virtual_tmp_file $!";
      if ( defined($vip_sticky) && $vip_sticky eq "true" ) {
        print $vfh "  persistence_timeout 600\n" or die "Failed to write to $virtual_tmp_file $!";
        print $vfh "  persistence_granularity 255.255.255.255\n" or die "Failed to write to $virtual_tmp_file $!";
      }
    } elsif ( $ip_version eq "v6" ) {
      print $vfh "  delay_loop 38\n" or die "Failed to write to $virtual_tmp_file $!";
      if ( defined($vip_sticky) && $vip_sticky eq "true" ) {
        print $vfh "  ip_family inet6\n" or die "Failed to write to $virtual_tmp_file $!";
        print $vfh "  persistence_granularity 128\n" or die "Failed to write to $virtual_tmp_file $!";
      }
    }
    print $vfh "  protocol TCP\n" or die "Failed to write to $virtual_tmp_file $!";
    print $vfh "  alpha\n" or die "Failed to write to $virtual_tmp_file $!";
    print $vfh "  quorum 50\n" or die "Failed to write to $virtual_tmp_file $!";

    my $advertised_ips = $yml_vips->{$vip_name}->{vip_advertise};
    my $geo_state = $yml_vips->{$vip_name}->{geo_state};
    my $q_op = "";
    my $qt = "";

    # ECMP mode uses the quorum scripts to add and delete the interface addresses.
    if ( $ecmp_enabled ) {
      # Adding omega so when a virtual service is reconfigured the quorum down is called to clean up.
      print $vfh "  omega\n" or die "Failed to write to $virtual_tmp_file $!";
      foreach my $q_type ("up","down") {
        if ($q_type eq "up") {
          $q_op = "UP";
          $qt = "quorum_up";
        } else {
          $q_op = "DOWN";
          $qt = "quorum_down";
        }
        print $vfh "  ${qt} \"${g_prefix}/scripts/vip_quorum.sh ${q_op} ${interface}.${bgp_vlan} ${vip_ip}-${vip_range}" or die "Failed to write to $virtual_tmp_file $!";
        #If advertised_ips is not empty, we use these for the quorum calls, which then translate into
        #bgp advertisements.  If it is empty, we default to using the vip_addr for quorum calls.  If geo_state
        #is not equal to 'up' then we leave advertised ips empty so we can still build the configs and
        #vips, but not actually advertise them.
        if ($geo_state eq "up") {
          if ( $advertised_ips ) {
            foreach my $adv_ip ( @$advertised_ips ) {
              print $vfh " ${adv_ip}" or die "Failed to write to $virtual_tmp_file $!";
            }
          } else {
            print $vfh " ${vip_ip_cidr}" or die "Failed to write to $virtual_tmp_file $!";
          }
          print $vfh "\"\n" or die "Failed to write to $virtual_tmp_file $!";
        } else {
          print $vfh "\"\n" or die "Failed to write to $virtual_tmp_file $!";
        }
      }
    }

    print $vfh "  include ${g_prefix}/config_containers/${vip_name}_real_servers.conf\n" or die "Failed to write to $virtual_tmp_file $!";
    print $vfh "}\n" or die "Failed to write to $virtual_tmp_file $!";
  }
  close ($vfh) or die "Failed to close ${virtual_tmp_file} $!";
}

sub gen_vrrp_configs {
  my $pop = shift || die "Did not receive a valid pop $!";
  my $yml_conf = shift || die "Did not receive yml ref $!";
  my $slb_state = shift || die "backup state not specified $!";
  my $v6_enabled = shift // 0;
  my $ecmp_enabled = shift // 0;

  my $yml_vips = $yml_conf->{vips};
  my $local_ip = (split '/',$yml_conf->{slb_config}->{address})[0] or die "Failed to parse address from $yml_conf->{slb_config}->{address})";
  my $local_ip6 = $yml_conf->{slb_config}->{v6_address};
  my $vrrp_script = "${g_prefix}/scripts/check_vips.sh";
  my $vrrp_notify = "${g_prefix}/scripts/parse_ips.sh";
  my $vrrp_tmp_file = "${g_saveprefix}/config_containers/vrrp_instance.new";
  my $vrrp_file = "${g_prefix}/config_containers/vrrp_instance.conf";
  my $v4_neighbors = $yml_conf->{slb_config}->{bgp_config}->{v4_neighbor};
  my $v6_neighbors = $yml_conf->{slb_config}->{bgp_config}->{v6_neighbor};
  my $bgp_script = "${g_prefix}/scripts/bgp_check.sh";

  open( my $vrfh,'>',"$vrrp_tmp_file" ) or die "Failed to open $vrrp_tmp_file $!";

  if ( $ecmp_enabled ) {
    print $vrfh "# This SLB is using ECMP with no VRRP, empty config is intentional\n" or die "Failed to write to $vrrp_tmp_file $!";
    close ($vrfh) or die "Failed to close ${vrrp_tmp_file} $!";
    return;
  }

  if (defined($v4_neighbors)) {
    print $vrfh "vrrp_script check_ipv4_bgp {\n" or die "Failed to write to $vrrp_tmp_file $!";
    if (ref($v4_neighbors) eq 'ARRAY') {
      print $vrfh "  script \"${bgp_script} @$v4_neighbors\"\n" or die "Failed to write to $vrrp_tmp_file $!";
    } else {
      print $vrfh "  script \"${bgp_script} $v4_neighbors\"\n" or die "Failed to write to $vrrp_tmp_file $!";
    }
    print $vrfh "  interval 45\n" or die "Failed to write to $vrrp_tmp_file $!";
    print $vrfh "  timeout 15\n" or die "Failed to write to $vrrp_tmp_file $!";
    print $vrfh "  fall 3\n" or die "Failed to write to $vrrp_tmp_file $!";
    print $vrfh "  rise 3\n" or die "Failed to write to $vrrp_tmp_file $!";
    print $vrfh "}\n" or die "Failed to write to $vrrp_tmp_file $!";
  } else {
    print STDERR "WARNING: No v4 BGP Neighbors defined, defaulting to no BGP in VRRP config\n";
  }

  if (defined($v6_neighbors)) {
    # verify that v4 BGP is configured
    defined($v4_neighbors) or die "v6 BGP requires v4 BGP, v4 BGP not configured";

    print $vrfh "vrrp_script check_ipv6_bgp {\n" or die "Failed to write to $vrrp_tmp_file $!";
    if (ref($v6_neighbors) eq 'ARRAY') {
      print $vrfh "  script \"${bgp_script} @$v6_neighbors\"\n" or die "Failed to write to $vrrp_tmp_file $!";
    } else {
      print $vrfh "  script \"${bgp_script} $v6_neighbors\"\n" or die "Failed to write to $vrrp_tmp_file $!";
    }
    print $vrfh "  interval 45\n" or die "Failed to write to $vrrp_tmp_file $!";
    print $vrfh "  timeout 15\n" or die "Failed to write to $vrrp_tmp_file $!";
    print $vrfh "  fall 3\n" or die "Failed to write to $vrrp_tmp_file $!";
    print $vrfh "  rise 3\n" or die "Failed to write to $vrrp_tmp_file $!";
    print $vrfh "}\n" or die "Failed to write to $vrrp_tmp_file $!";
  } else {
    print STDERR "WARNING: No v6 BGP Neighbors defined, defaulting to no v6 BGP in VRRP config\n";
  }

  my @keys = keys %$yml_vips;
  for my $vip_name ( @keys ) {
    my $vip_ip = (split '/',$yml_vips->{$vip_name}->{vip_addr})[0];
    my $ip_version = get_ip_version($vip_ip);

    # if v6 is disabled (not configured for native vlan) then skip v6 VIPS
    if ( $ip_version eq "v6" and !$v6_enabled) {
      print STDERR "WARNING: v6 is not enabled, ignoring VIP: ${vip_name} in VRRP config\n";
      next;
    }

    my $vip_ip_cidr = $yml_vips->{$vip_name}->{vip_addr};
    my $vip_port = $yml_vips->{$vip_name}->{vip_port};
    my $vrid = $yml_vips->{$vip_name}->{vip_vrid};
    my $vip_priority = $yml_vips->{$vip_name}->{vip_priority};
    my @vip_broadcast = Net::CIDR::cidr2range("$vip_ip_cidr");
    my $vip_broadcast = (split '-',$vip_broadcast[0])[1];
    my $config_gen = $yml_conf->{slb_config}->{config_gen};
    my $interface = $yml_conf->{slb_config}->{interface};
    my $bgp_vlan = $yml_conf->{slb_config}->{native_vlan};


    if ( $config_gen ne "auto" ) {
      print "$vip_name is set to ${config_gen}, skipping\n";
      next;
    }

    my $vip_dir = "${g_saveprefix}/server_states/${vip_ip}";
    if ( ! -d $vip_dir ) {
      mkdir($vip_dir, 0755) || die "mkdir $vip_dir: $!";
    }

    #Sort slb_peers into v4 and v6 arrays, so we don't need to keep separate yaml fields
    my $slb_peers = $yml_conf->{slb_config}->{slb_peers};
    my @v4_slb_peers;
    my @v6_slb_peers;
    foreach my $peer ( @$slb_peers ) {
      my $ip_version = get_ip_version($peer);
      if ( $ip_version eq "v4" ) {
        push @v4_slb_peers, "$peer";
      } elsif ( $ip_version eq "v6" ) {
        push @v6_slb_peers, "$peer";
      }
    }

    if (ref($vip_port) eq 'ARRAY') {
      foreach my $port ( @{ $vip_port } ) {
        print $vrfh "vrrp_script check_${vip_name}_${port} {\n" or die "Failed to write to $vrrp_tmp_file $!";
        print $vrfh "  script \"${vrrp_script} -v ${vip_ip} -p ${port} -s\"\n" or die "Failed to write to $vrrp_tmp_file $!";
        print $vrfh "  interval 10\n" or die "Failed to write to $vrrp_tmp_file $!";
        print $vrfh "  timeout 5\n" or die "Failed to write to $vrrp_tmp_file $!";
        print $vrfh "  fall 3\n" or die "Failed to write to $vrrp_tmp_file $!";
        print $vrfh "  rise 3\n" or die "Failed to write to $vrrp_tmp_file $!";
        print $vrfh "}\n" or die "Failed to write to $vrrp_tmp_file $!";
      }
    } else {
      print $vrfh "vrrp_script check_${vip_name}_${vip_port} {\n" or die "Failed to write to $vrrp_tmp_file $!";
      print $vrfh "  script \"${vrrp_script} -v ${vip_ip} -p ${vip_port} -s\"\n" or die "Failed to write to $vrrp_tmp_file $!";
      print $vrfh "  interval 10\n" or die "Failed to write to $vrrp_tmp_file $!";
      print $vrfh "  timeout 5\n" or die "Failed to write to $vrrp_tmp_file $!";
      print $vrfh "  fall 3\n" or die "Failed to write to $vrrp_tmp_file $!";
      print $vrfh "  rise 3\n" or die "Failed to write to $vrrp_tmp_file $!";
      print $vrfh "}\n" or die "Failed to write to $vrrp_tmp_file $!";
    }

    print $vrfh "vrrp_instance ${vip_name} {\n" or die "Failed to write to $vrrp_tmp_file $!";
    print $vrfh "  state BACKUP\n" or die "Failed to write to $vrrp_tmp_file $!";
    print $vrfh "  interface ${interface}.${bgp_vlan}\n" or die "Failed to write to $vrrp_tmp_file $!";
    print $vrfh "  virtual_router_id ${vrid}\n" or die "Failed to write to $vrrp_tmp_file $!";
    if ( $slb_state eq "backup" ) {
      print $vrfh "  priority 50\n" or die "Failed to write to $vrrp_tmp_file $!";
    } elsif ( $slb_state eq "master" ) {
      print $vrfh "  priority 254\n" or die "Failed to write to $vrrp_tmp_file $!";
    } else {
      print $vrfh "  priority ${vip_priority}\n" or die "Failed to write to $vrrp_tmp_file $!";
    }
    print $vrfh "  advert_int 2\n" or die "Failed to write to $vrrp_tmp_file $!";
    print $vrfh "  preempt_delay 5\n" or die "Failed to write to $vrrp_tmp_file $!";

    if ( $ip_version eq "v6" ) {
      print $vrfh "  native_ipv6\n" or die "Failed to write to $vrrp_tmp_file $!";
      print $vrfh "  unicast_src_ip ${local_ip6}\n" or die "Failed to write to $vrrp_tmp_file $!";
      print $vrfh "  unicast_peer {\n" or die "Failed to write to $vrrp_tmp_file $!";
      foreach my $peer ( @v6_slb_peers ) {
        print $vrfh "    ${peer}\n" or die "Failed to write to $vrrp_tmp_file $!";
      }
      print $vrfh "  }\n" or die "Failed to write to $vrrp_tmp_file $!";
    } else {
      print $vrfh "  unicast_src_ip ${local_ip}\n" or die "Failed to write to $vrrp_tmp_file $!";
      print $vrfh "  unicast_peer {\n" or die "Failed to write to $vrrp_tmp_file $!";
      foreach my $peer ( @v4_slb_peers ) {
        print $vrfh "    ${peer}\n" or die "Failed to write to $vrrp_tmp_file $!";
      }
      print $vrfh "  }\n" or die "Failed to write to $vrrp_tmp_file $!";
    }


    print $vrfh "  track_interface {\n" or die "Failed to write to $vrrp_tmp_file $!";
    print $vrfh "    ${interface}\n" or die "Failed to write to $vrrp_tmp_file $!";
    print $vrfh "  }\n" or die "Failed to write to $vrrp_tmp_file $!";

    print $vrfh "  authentication {\n" or die "Failed to write to $vrrp_tmp_file $!";
    print $vrfh "    auth_type PASS\n" or die "Failed to write to $vrrp_tmp_file $!";
    print $vrfh "    auth_pass vR1d_${vrid}\n" or die "Failed to write to $vrrp_tmp_file $!";
    print $vrfh "  }\n" or die "Failed to write to $vrrp_tmp_file $!";

    #If advertised_ips is not empty, we use these for the vrrp blocks, which then translate into
    #bgp advertisements.  If it is empty, we default to using the vip_addr for vrrp.  If geo_state
    #is not equal to 'up' then we leave advertised ips empty so we can still build the configs and
    #vips, but not actually advertise them.
    my $advertised_ips = $yml_vips->{$vip_name}->{vip_advertise};
    my $virtual_routes = $yml_vips->{$vip_name}->{virtual_routes};
    my $geo_state = $yml_vips->{$vip_name}->{geo_state};

    print $vrfh "  virtual_ipaddress {\n" or die "Failed to write to $vrrp_tmp_file $!";

    if ( $geo_state ne "up" ) {
      print $vrfh "  }\n" or die "Failed to write to $vrrp_tmp_file $!";
    } else {
      if ( $advertised_ips ) {
        foreach my $adv_ip ( @$advertised_ips ) {
          print $vrfh "    ${adv_ip}\n" or die "Failed to write to $vrrp_tmp_file $!";
        }
      } else {
        if ( $ip_version eq "v6" ) {
          print $vrfh "    ${vip_ip_cidr} dev ${interface}.${bgp_vlan}\n" or die "Failed to write to $vrrp_tmp_file $!"
        }
        elsif ( $ip_version eq "v4" ) {
          print $vrfh "    ${vip_ip_cidr} brd ${vip_broadcast} dev ${interface}.${bgp_vlan}\n" or die "Failed to write to $vrrp_tmp_file $!";
        }
      }
      print $vrfh "  }\n" or die "Failed to write to $vrrp_tmp_file $!";
      if ( $virtual_routes ) {
        print $vrfh "  virtual_routes {\n" or die "Failed to write to $vrrp_tmp_file $!";
        foreach my $virt_rt ( @$virtual_routes ) {
          print $vrfh "    ${virt_rt} dev ${interface}.${bgp_vlan}\n" or die "Failed to write to $vrrp_tmp_file $!";
        }
        print $vrfh "  }\n" or die "Failed to write to $vrrp_tmp_file $!";
      }
    }

    print $vrfh "  track_script {\n" or die "Failed to write to $vrrp_tmp_file $!";
    if (ref($vip_port) eq 'ARRAY') {
      foreach my $port ( @{ $vip_port } ) {
        print $vrfh "    check_${vip_name}_${port}\n" or die "Failed to write to $vrrp_tmp_file $!";
      }
    } else {
      print $vrfh "    check_${vip_name}_${vip_port}\n" or die "Failed to write to $vrrp_tmp_file $!";
    }
    if ( $ip_version eq "v6" and defined($v6_neighbors) ) {
      print $vrfh "    check_ipv6_bgp\n" or die "Failed to write to $vrrp_tmp_file $!";
    } elsif ( $ip_version eq "v4" and defined($v4_neighbors) ) {
      print $vrfh "    check_ipv4_bgp\n" or die "Failed to write to $vrrp_tmp_file $!";
    }
    print $vrfh "  }\n" or die "Failed to write to $vrrp_tmp_file $!";
    print $vrfh "  notify ${vrrp_notify}\n";
    print $vrfh "}\n" or die "Failed to write to $vrrp_tmp_file $!";
  }
  close ($vrfh) or die "Failed to close ${vrrp_tmp_file} $!";
}

sub rename_file_perms {
  my $tmp_file = shift || die "Missing temp file arg $!";
  my $file = shift || die "Missing vip file arg $!";

  chmod 0644, $tmp_file or die "Failed to change permissions on file $tmp_file $!";
  my $uid = getpwnam "root";
  my $gid = getgrnam "root";
  chown $uid, $gid, $tmp_file or die "Failed to chown $tmp_file to root:root $!";
  rename $tmp_file,$file or die "Failed to rename $tmp_file to $file $!";
}

#
# Check that advertised vips are inside the range.
#
sub sanity_check_vips() {
  my $yml = shift || die "Did not receive valid yml data $!";
  my %vips = %{$yml->{vips}};
  my @keys = keys %vips;
  for my $vip_name ( @keys ) {
    my $vip = $vips{$vip_name};
    my $vip_addr = $vip->{'vip_addr'};
    if (!$vip->{'vip_advertise'} || scalar($vip->{'vip_advertise'}) == 0) {
      next;
    }
    my @range = Net::CIDR::cidr2range($vip_addr);
    my $addresses = $vip->{'vip_advertise'};
    for my $address ( @$addresses ) {
      Net::CIDR::cidrlookup($address, @range) or die "VIP advertise configuration error: ${address} not in ${vip_addr}";
    }
  }
}

sub gen_yml_conf_from_v2 {
  my $yml = shift || die "Did not receive valid yml data $!";
  my $this_slb = shift || die "Did not receive valid slb name $!";
  my $ecmp_enabled = shift // 0;
  my @pop_cfg_in = @{$yml->{pop_config}{slbs}};
  my %slb_cfgs_in = %{$yml->{slb_configs}};
  my $pop_hash = $yml->{pop_config}{options}{hash};
  my %vips_in = %{$yml->{vips}};
  my $slb_count = scalar(@pop_cfg_in);
  my $vip_count = scalar(keys %vips_in);
  my $backups_max = 2;
  my $backup_count = $slb_count <= $backups_max ? $slb_count-1 : $backups_max;
  my %slbs_out;

  my @backups;
  for my $backup (1..$backup_count) {
    push @backups, $backup;
  }

  # Create priority matrix.
  # 0 is master (highest priority), 1-3 is lower priority (we probably
  # don't want more than 2 or 3 backups due to health check overhead).
  # Set master according to: slb_index = vip_index % slb_count.
  # Rotate the backup prios among the other SLBs for some deterministic
  # randomness. We want to distribute backups evenly so any SLB don't
  # get overloaded when a SLB is taken offline.
  #
  # In the case of 4 SLBs and 3 backups:
  #
  #         s s s s
  #         l l l l
  #         b b b b
  #         0 1 3 4
  #        --------
  # vip 0 | 0 1 2 3
  # vip 1 | 2 0 3 1
  # vip 2 | 3 1 0 2
  # vip 3 | 1 2 3 0
  # and so on...

  # VIP range priorities
  my $prio_slb_override = 254;
  my $prio_vip_override = 250;
  my $prio_master = 200;
  my $prio_backup1 = 150;
  my $prio_backup2 = 100;
  my $prio_backup3 = 50;
  my $prio_offline = 1;
  my @prio_idx_to_val = ($prio_master, $prio_backup1, $prio_backup2, $prio_backup3);

  # @prio_matrix contain the final matrix.
  my @prio_matrix = ();

  # @backup_prios is the array of backup prios we rotate.
  # For 4 SLB setup, this is (1, 2, 3)
  my @backup_prios;
  for my $backup (1..$slb_count-1) {
    push @backup_prios, $backup;
  }

  if ($g_dryrun == 1) {
    print "VIP range count: ${vip_count}\n";
    print "SLB count: ${slb_count}\n";
    print "SLB backup count: ${backup_count}\n\n";
    print "Priority matrix (see script for explanation):\n";
  }

  for my $v (0..$vip_count-1) {
    # Copy the backup prio array and insert master at correct place.
    my @prios = @backup_prios;
    splice @prios, $v % $slb_count, 0, 0;
    push @prio_matrix, \@prios;
    # Rotate backups
    push @backup_prios, shift @backup_prios;
    if ($g_dryrun == 1) {
      for my $p ( @prios ) {
        if ($p > $backup_count) {
          print "  ";
        } else {
          print "${p} ";
        }
      }
      print "\n";
    }
  }

  my $slb_idx = 0;
  for my $slb_in ( @pop_cfg_in ) {

    my $state = $slb_in->{state};
    my $name = $slb_in->{name};
    my @override_vips = ();

    #
    # The hold state is used to tell the SLB to hold the running config. As long
    # as the SLB is set to this state, no changes to the SLB config will affect the
    # SLB running config. Naturally this only applies to the v2 variant of config.
    #
    if ($state eq 'hold' && $name eq $this_slb) {
      exit(0);
    }

    # Someone will probably miss putting vip names in an array...
    if (ref($slb_in->{override_vips}) ne 'ARRAY') {
      push @override_vips, $slb_in->{override_vips};
    } else {
      push @override_vips, @{ $slb_in->{override_vips} };
    }

    $slbs_out{$name}{slb_config} = $slb_cfgs_in{$name};

    my $vip_idx = 0;
    if ($g_dryrun == 1) {
      print "\nSLB: ${name}\n";
    }

    my @keys = keys %vips_in;
    for my $vip_name ( @keys ) {
      my $vip = $vips_in{$vip_name};
      my $override_vip = grep( { defined && $_ eq $vip_name} @override_vips);
      my $vip_prio = 0;

      # In ECMP mode, we never cull the vips and the prio matrix is not used. This code
      # is for the VRRP VIP balancing.
      if ( !$ecmp_enabled ) {
        $vip_prio = $prio_idx_to_val[$prio_matrix[$vip_idx][$slb_idx]];
        # Cull backups we don't want.
        if (!$override_vip && $prio_matrix[$vip_idx][$slb_idx] > $backup_count) {
          $vip_idx++;
          next;
        }
      }

      # Set priority for this SLB-VIP combination
      if ($state eq 'offline') {
        # We are offline (i.e. prio = 1)
        $vip_prio = $prio_offline;
      } elsif ($state eq 'override') {
        # We are master of _all_ VIPs (overriding other SLBs)
        $vip_prio = $prio_slb_override;
      } else {
        if ($override_vip) {
          $vip_prio = $prio_vip_override;
        }
      }

      my $vrid = $vip->{vip_vrid};
      if ($g_dryrun == 1) {
        print "idx: ${vip_idx} name: ${vip_name} prio: ${vip_prio}\n";
      }
      # Set VIP configurations
      $slbs_out{$name}{vips}{$vip_name} = dclone $vip;
      $slbs_out{$name}{vips}{$vip_name}{vip_priority} = $vip_prio;

      # A POP can select one of the supported hashes (maglev-mh or weighted_round_robin-wrr),
      # the default is wrr.
      if ( $pop_hash ) {
        if ( $pop_hash eq 'mh' || $pop_hash eq 'wrr' ) {
          $slbs_out{$name}{vips}{$vip_name}{vip_hash} = $pop_hash;
        } else {
          die "Unsupported hash selected ($pop_hash)";
        }
      } else {
        $slbs_out{$name}{vips}{$vip_name}{vip_hash} = 'wrr';
      }

      # In ECMP mode, taking the SLB offline is just a matter of setting the geo_state
      # down for all VIPs. The SLB will stop advertising the VIPS to BGP.
      if ( $ecmp_enabled ) {
        if ( $vip_prio == $prio_offline ) {
          $slbs_out{$name}{vips}{$vip_name}{geo_state} = 'down';
        } else {
          # If not going offline, priority is always master in ECMP mode
          $slbs_out{$name}{vips}{$vip_name}{vip_priority} = $prio_master;
        }
      }

      $vip_idx++;
    }
    $slb_idx++;
  }

  return \%slbs_out;
}