#!/bin/bash

PATH="/sbin:/bin:/usr/bin:/usr/sbin:/usr/local/sbin:/usr/local/bin"
shopt -s nocaseglob
shopt -s nocasematch

# source required variables
if [ ! -f /etc/keepalived/scripts/.running_vars.inc ]; then
  # provide a friendlier error message vs source <file> || exit 1
  echo "Required source file /etc/keepalived/scripts/.running_vars.inc is missing"
  exit 1
fi
source /etc/keepalived/scripts/.running_vars.inc

if [ $UID -ne 0 ]; then
  echo "Must be root to run the script"
  exit 0
fi

usage(){
  echo "usage: $0 [options]"
  echo "Options:"
  echo "  -s      log warning and error messages to syslog "
  echo "Example:"
  echo "  $0 -s"
  exit 1
}

while getopts ":s" opt; do
  case ${opt} in
    s)
      log_to_syslog=1
      ;;
    *)
      usage
      ;;
  esac
done

hostname=`hostname`
script_name="slb_scan.sh"


echo_status(){
  check=${FUNCNAME[1]}
  status=$1
  msg=${2:-""}

  no_color='\033[0m'

  if [ "$status" = "WARN" ] || [ "$status" = "ERROR" ]; then
    # Set color to RED for WARN and ERROR
    color='\033[030;31m'
  elif [ "$status" = "OK" ] || [ "$status" = "INFO" ]; then
    # Set color to CYAN for OK
    color='\033[030;36m'
  fi

  # Send to syslog if log_to_syslog set
  if [ "${log_to_syslog}" ]; then
    logger -p daemon.err -t "${script_name}:${check}()" "[${status}] ${msg}"
  fi

  # Normal stdout logging
  echo -en "${color}  [$status]   ${no_color}${msg}\\n"

}

echo_txt(){
  #set for normal text yellow:
  echo -en "\E[030;33m${1:-""}\033[0m\n"
}

verify_kernel(){
  echo_txt "VERIFYING KERNEL VERSION:"

  local host_info=( ${hostname//./ } )
  local kernel_ver=`uname -r`
  local kversion=( ${kernel_ver//./ } )

  # Edge Gravity has a different OS for the moment.
  if [ $is_eg = "True" ]; then
    if [ ${kversion[0]} = 4 ] && [ ${kversion[1]} = 15 ]; then
      echo_status "OK" "${kernel_ver}"
    else
      echo_status "WARN"  "${kernel_ver} should be 4.15.0-##-generic"
    fi
  else
    # production should be at kernel 5.3.0-28-generic at the moment except for canary pops
    if [ ${kversion[0]} = 5 ] && [ ${kversion[1]} = 3 ]; then
      echo_status "OK" "${kernel_ver}"
    else
      echo_status "WARN"  "${kernel_ver} should be 5.3.#-##-generic"
    fi
  fi
}

verify_programs(){
  programs_needed=( "ipvsadm" "keepalived" "vlan" "quagga" "ifenslave" "unbound" "linux-crashdump")

  echo_txt "VERIFYING NEEDED PROGRAMS:"
  for dependency in ${programs_needed[@]} ; do
    dpkg -s ${dependency} > /dev/null 2>&1
    if [ $? != 0 ]; then
      echo_status "ERROR" "${dependency}"
      prog_not_installed+=("${dependency}")
    else
      echo_status "OK" "${dependency}"
    fi
    if [ $dependency = "keepalived" ]; then
      keepalived_working='1:1.3.9-1ubuntu0.18.04.2'
      keepalived_installed=`dpkg-query --showformat='${Version}' --show ${dependency}`
      if [ "${keepalived_installed}" != "${keepalived_working}" ]; then
        echo_status "ERROR" "${dependency} is the wrong version, should be $keepalived_working"
      else
        echo_status "OK" "${dependency} version $keepalived_working"
      fi
    fi
  done
}

verify_daemons(){
  local process_count=0
  #Verify programs are running
  echo_txt "VERIFYING KEEPALIVED:"
  for proc in vrrp.pid keepalived.pid checkers.pid; do
    if [ -f /var/run/${proc} ]; then
      process_count=`pgrep -cF /var/run/${proc}`
      if [ $process_count -eq 1 ]; then
        echo_status "OK" "${proc}"
      else
        echo_status "ERROR" "${proc}"
        keepalived_status=0
      fi
    else
      echo_status "ERROR" "${proc}"
      keepalived_status=0
    fi
  done

  if [ $has_v4_bgp = "True" ]; then
    echo_txt "VERIFYING QUAGGA:"
    for proc in zebra.pid bgpd.pid; do
        if [ -f /var/run/quagga/${proc} ]; then
          process_count=`pgrep -cF /var/run/quagga/${proc}`
          if [ $process_count -eq 1 ]; then
            echo_status "OK" "${proc}"
            if [ $proc = "bgpd.pid" ]; then
              bgp_running=1
            fi
          else
            echo_status "ERROR" "${proc}"
          fi
        else
          echo_status "ERROR" "${proc}"
        fi
    done
  fi

  echo_txt "VERIFYING UNBOUND:"
  process_count=`pgrep -cF /var/run/unbound.pid`
  if [ $process_count -eq 1 ]; then
    echo_status "OK" "unbound"
  else
    echo_status "ERROR" "unbound"
  fi
}

verify_modules(){
  #Check for loaded and installed kernel modules
  required_modules=( "ip_vs" "dummy" "8021q" "bonding" )
  banned_modules=( "nf_conntrack_ipv4" "nf_conntrack_ipv6" "nfnetlink" "kvm_intel" )
  
  echo_txt "CHECKING LOADED KERNEL MODULES:"
  while read module rest; do
    for kmod in ${required_modules[@]}; do
      if [ "${module}" = "${kmod}" ]; then
        echo_status "OK" "${kmod}"
        required_modules=(${required_modules[@]/$kmod})
      fi
    done

    for b_kmod in ${banned_modules[@]}; do
      if [ "${module}" = "${b_kmod}" ]; then
        echo_status "ERROR" "${b_kmod} loaded, <modprobe -r ${b_kmod}> to unload"
      fi
    done
  done < <(lsmod)

  if [ "${#required_modules[@]}" -ne 0 ]; then
    for kmod in ${required_modules[@]}; do
      if grep -q ${kmod} /etc/modules; then
        echo_status "ERROR" "${kmod} not loaded"
      else
        echo_status "ERROR" "${kmod} not installed"
      fi
    done
  fi
}

verify_sysctls(){
  #VERIFY SYSCTLS ARE SET PROPERLY
  sysctl_one=( "net.ipv4.ip_forward" "net.ipv6.conf.all.forwarding" "net.ipv4.conf.all.arp_filter" )
  sysctl_zero=( "net.ipv4.conf.all.rp_filter" "net.ipv4.conf.default.rp_filter" "net.ipv4.vs.expire_nodest_conn" )

  echo_txt "CHECKING SYSCTL VALUES:"
  for sysctl in ${sysctl_one[@]}; do
    if [ `sysctl -n ${sysctl}` -eq "1" ]; then
      echo_status "OK" "${sysctl} = 1"
    else
      echo_status "ERROR" "${sysctl} != 1"
    fi
  done

  for sysctl in ${sysctl_zero[@]}; do
    if [ `sysctl -n ${sysctl}` -eq "0" ]; then
      echo_status "OK" "${sysctl} = 0"
    else
      echo_status "ERROR" "${sysctl} != 0"
    fi
  done
}

verify_bond(){
  local IFS=$'\n'
  local int_count=0

  declare -gA bond_info

  echo_txt "CHECKING BOND0 STATUS:"
  if [ ! -e /proc/net/bonding/bond0 ]; then
    echo_status "ERROR" "Bond0 does not exist"
  else
    bond_detected=1
  fi

  for line in `cat /proc/net/bonding/bond0`; do
    if [[ "${line}" =~ "Transmit Hash Policy:" ]]; then
      xmit_hash=`echo $line | cut -d ' ' -f 4`
      continue
    fi
    if [[ "${line}" =~ "MII Status:" ]]; then
      link=`echo $line | cut -d ' ' -f 3`
      continue
    fi
    if [[ "${line}" =~ "Number of ports:" ]]; then
      n_ports=`echo $line | cut -d ' ' -f 4`
      continue
    fi
    if [[ "${line}" =~ "Slave Interface:" ]]; then
      interface=`echo $line | cut -d ' ' -f 3`
      net_ints+=( $interface )
      continue
    fi
    if [[ "${line}" =~ "Speed:" ]]; then
      speed=`echo $line | cut -d ' ' -f 2`
      continue
    fi
    if [[ "${line}" =~ "Link Failure Count:" ]]; then
      failure=`echo $line | cut -d ' ' -f 4`
      continue
    fi

    if [ ${link} ] && [ ${n_ports} ]; then
      # make sure  bonded port count matches number of configured ports
      if [ ${n_ports} -eq ${#lacp_ints[@]} ]; then
        bond_info[bond_status]=healthy
        unset link n_ports
        continue
      else
        bond_info[bond_status]=unhealthy
        unset link n_ports
        continue
      fi
    fi

    if [ ${interface} ] && [ ${link} ] && [ ${speed} ] && [ ${failure} ]; then
      bond_info[${interface}_link]=$link
      bond_info[${interface}_speed]=$speed
      bond_info[${interface}_failures]=$failure
      unset interface link speed failure
    fi
  done

  if [ ${bond_info[bond_status]} = "unhealthy" ]; then
    echo_status "ERROR" "Bond0 in unhealthy state"
  elif [ ${bond_info[bond_status]} = "healthy" ]; then
    echo_status "OK" "Bond0 ${bond_info[bond_status]}"
  fi

  # verify each configured interface
  for int in ${lacp_ints[@]}; do
    # verify the interface is configured
    if [[ ! " ${net_ints[@]} " =~ " ${int} " ]]; then
        echo_status "ERROR" "${int} not found/configured"
        continue
    fi
    if [ "${bond_info[${int}_link]}" != "up" ]; then
      echo_status "ERROR" "${int} LINK=${bond_info[${int}_link]} SPEED=${bond_info[${int}_speed]}"
      continue
    fi

    # get the interface bond state
    bond_state=$(cat /sys/class/net/${int}/bonding_slave/state)
    if [ ${bond_info[${int}_speed]} = "Unknown" ]; then
      echo_status "ERROR" "${int} LINK=${bond_info[${int}_link]} SPEED=${bond_info[${int}_speed]}"
      continue
    elif [ ${bond_info[${int}_speed]} -lt "10000" ]; then
      echo_status "ERROR" "${int} is less than 10gig"
      continue
    elif [ "${bond_state}" != "active" ]; then
      echo_status "ERROR" "${int} bond state is not active"
      continue
    fi
    #Speed mismatch check
    if [ ${bond_info[${int}_speed]} -eq 10000 ]; then
      ten_gig=1
    fi
    if [ ${bond_info[${int}_speed]} -eq 25000 ]; then
      twentyfive_gig=1
    fi
    echo_status "OK" "${int} LINK=${bond_info[${int}_link]} SPEED=${bond_info[${int}_speed]} FLAPS=${bond_info[${int}_failures]}"
  done

  if [ ${ten_gig} ] && [ ${twentyfive_gig} ]; then
    echo_status "ERROR" "Speed mismatch between ports"
  fi

  if [ ${xmit_hash} = "layer3+4" ]; then
    echo_status "OK" "bond0 xmit_hash_policy = $xmit_hash"
  else
    echo_status "ERROR" "bond0 xmit_hash_policy = $xmit_hash (Should be layer3+4)"
  fi
}

verify_nic(){
  #Loop through ethtool info to check performance settings.  net_ints array populated from previous function
  echo_txt "CHECKING NIC SETTINGS:"
  for int in ${net_ints[@]}; do
    flow_control=`ethtool -a $int | grep -ic off`
    irq_coal=`ethtool -c $int | grep -ic 'rx-usecs: 50'`
    txqueue=`ifconfig $int | grep -ic 'txqueuelen 3000'`
    if [ ${flow_control} -ne 3 ]; then
      echo_status "ERROR" "Flow Control is on for $int 'ethtool -a $int'"
    else
      echo_status "OK" "Flow control disabled on $int"
    fi
    if [ ${irq_coal} -lt 1 ]; then
      echo_status "ERROR" "rx-usecs should be set to 50 on $int 'ethtool -c $int'"
    else
      echo_status "OK" "rx-usecs = 50 on $int"
    fi
    if [ ${txqueue} -lt 1 ]; then
      echo_status "ERROR" "txqueuelen should be set to 3000 on $int 'ifconfig $int'"
    else
      echo_status "OK" "txqueue length set to 3000 on $int"
    fi
  done
}

get_bgp_vlan(){
  conf_file="/etc/keepalived/slb_conf.yml"

  if [ ! -f ${conf_file} ]; then
    return 1
  fi

  while read key value; do
    if [ "${key}" = "${hostname}:" ]; then
      my_block=1
      continue
    fi

    if [ ${my_block} ]; then
      if [ "${key}" = "native_vlan:" ]; then
        native_vlan=$value
        unset my_block
      fi
    fi
  done < $conf_file
}

verify_vlans(){
  #We should be using vlan 30 or 31 for SLB connectivity for BGP according to standards
  #but sites may vary due to legacy bits or partner requirements, try to pull in from slb_conf.yml
  #CDN vlans may vary, for the same reason, but anything not bgp/native is assumed to be CDN

  unset bgp_vlan
  unset cdn_vlan

  echo_txt "VERIFYING VLANS:"
  if [ ${bond_detected} ]; then
    vlans=(`ifconfig -a | egrep bond0\.[0-9] | cut -d " " -f 1 | cut -d "." -f 2 | cut -d ":" -f 1`)

    if [ $has_v4_bgp = "True" ]; then
        get_bgp_vlan
    fi
  else
    echo_status "ERROR" "No bond0 interface, can't verify vlans"
    return 0
  fi

  for vlan in ${vlans[@]}; do
    if [[ ${vlan} =~ [0-9]+:[0-9] ]]; then
      #Sub vlan int, no need to check
      continue
    fi

    if [ ${vlan} == ${native_vlan} ]; then
      bgp_vlan+=( $vlan )
      continue
    fi

    if [ ${vlan} != ${native_vlan} ]; then
      cdn_vlan+=( $vlan )
      continue
    fi
  done

  if [ ${#bgp_vlan[@]} -eq 1 ]; then
    echo_status "OK" "BGP Vlan: ${bgp_vlan[@]}"
  elif [ ${#bgp_vlan[@]} -lt 1 ]; then
    echo_status "ERROR" "No BGP vlans found"
  else
    echo_status "ERROR" "Multiple BGP vlans found: ${bgp_vlan}"
  fi

  if [ ${#cdn_vlan[@]} -gt 0 ]; then
    for i in ${cdn_vlan[@]}; do
      echo_status "OK" "CDN Vlan: ${i}"
    done
  else
    # if CDN VLAN is not defined just log INFO message
    if [ $is_eg = "True" ]; then
      echo_status "INFO" "CDN Vlan is not required for EG"
    else
      echo_status "ERROR" "No CDN Vlan Detected"
    fi
  fi
}

verify_dns(){
  #Verify we can resolve hosts with local resolver
  echo_txt "CHECKING LOCAL DNS RESOLVER:"
  dig +short +time=3 @127.0.0.1 $hostname > /dev/null

  if [ $? = "0" ]; then
    echo_status "OK" "Unbound Resolver Working"
  else
    echo_status "ERROR" "Unbound Resolver Working"
  fi
}

verify_bgp(){
  #Verify quagga has established BGP sessions
  local IFS=$'\n'

  echo_txt "VERIFYING BGP NEIGHBOR STATES:"
  if [ ! ${bgp_running} ]; then
    echo_status "ERROR" "BGPD not running"
    return
  fi

  for line in `vtysh -c "show ip bgp neighbors"` ; do
    if [[ "${line}" =~ "BGP neighbor is" ]]; then
      neighbor=`echo $line | awk '{print $4}' | cut -d ',' -f 1`
      continue
    fi

    if [[ "${line}" =~ "BGP state = " ]]; then
      bgp_state=`echo $line | awk '{print $4}' | cut -d ',' -f 1`
      continue
    fi

    if [ ${neighbor} ] && [ ${bgp_state} ]; then
      if [ ${bgp_state} = "Established" ]; then
        echo_status "OK" "${neighbor} = ${bgp_state}"
      else
        echo_status "ERROR" "${neighbor} = ${bgp_state}"
      fi
      unset neighbor bgp_state
    fi
  done
}

verify_virtuals(){
  local good=0
  local bad=0

  echo_txt "VERIFYING VIPS ARE UP WITH SERVERS BOUND:"
  if [ ${keepalived_status} ]; then
    echo_status "ERROR" "KeepaliveD not running, or issue with one of its PIDs, not checking virtuals"
    return 0
  fi

  while read proto ip scheduler junk; do
    if [ ${proto} = "TCP" ] && [ $vip_ip ]; then
      if [ $bad -eq 0 ]; then
        let bad=${bad}+2
        continue
      else
        let bad=${bad}+1
        continue
      fi
    elif [ ${proto} = "TCP" ]; then
      vip_ip=$ip
      continue
    fi

    if [ ${vip_ip} ]; then
      if [ ${scheduler} = "Route" ]; then
        unset vip_ip
        let good=${good}+1
      fi
    fi

  done< <(ipvsadm -Ln)

  if [ ${good} -gt 0 ]; then
    echo_status "OK" "VIPS BOUND WITH SERVERS: $good"
  fi
  if [ ${bad} -gt 0 ]; then
    echo_status "ERROR" "VIPS BOUND WITHOUT SERVERS: $bad"
  fi
}

verify_virtual_and_dns(){
  #Loop through the vips and verify dns for the vip and servers
  local good=0
  local bad=0

  echo_txt "VERIFYING VIPS AND SERVERS ARE BOUND AND HAVE DNS:"
  if [ ${keepalived_status} ]; then
    echo_status "ERROR" "KeepaliveD not running, or issue with one of its PIDs, not checking virtuals"
    return 0
  fi

  while read proto ip junk; do
    if [[ "${ip}" =~ "http" ]]; then
      if [ ${vip} ]; then
        let bad=${bad}+1
        #echo -ne "\rSuccessful: $good Bad: $bad"
        continue
      else
        vip=$ip
        continue
      fi
    fi
    if [[ "${ip}" =~ cds ]]; then
      cds=$ip
      if [ ${vip} ] && [ ${cds} ]; then
        #we have servers under our vip
        unset vip
        unset cds
        let good=${good}+1
        #echo -ne "\rSuccessful: $good Bad: $bad"
      fi
    fi
  done< <(ipvsadm -l)

  if [ ${good} -gt 0 ]; then
    echo_status "OK" "VIPS BOUND WITH SERVERS: $good"
  fi
  if [ ${bad} -gt 0 ]; then
    echo_status "ERROR" "VIPS BOUND WITHOUT SERVERS: $bad"
  fi
}

echo_txt "SITE TYPE:"
echo_status "INFO" "$building_type"

verify_kernel
verify_daemons
verify_modules
verify_programs
verify_sysctls
verify_bond
verify_nic
verify_vlans
verify_dns

if [ $has_v4_bgp = "True" ]; then
    verify_bgp
else
    if [ $is_eg = "True" ]; then
        echo_status "INFO" "BGP is not configured/required for EG"
    else
        echo_status "ERROR" "BGP is not configured"
    fi
fi

verify_virtuals
