#!/usr/bin/env bash
PATH="/sbin:/bin:/usr/bin:/usr/sbin:/usr/local/sbin:/usr/local/bin"
#shopt -s nocaseglob
#shopt -s nocasematch

#Script to check to see if real servers are bound under a vip, if there are then
#they are healthy and the vip is healthy.  If healthy return 0, if unhealth return 1

usage(){
  echo "usage: $0 -v [ip_address] -p [port] [options]"
  echo "Required:"
  echo "  -v      vip ip address v4|v6"
  echo "  -p      vip port number"
  echo "Options:"
  echo "  -l      list virtual server and bound real servers to stdout"
  echo "  -s      log warning and error messages to syslog "
  echo "Example:"
  echo "  $0 -v 192.168.0.1 -p 0 -l"
  echo "  $0 -v fc00:dddd:dddd:dddd::1 -p 0 -l"
  exit 1
}

while getopts ":v:p:l:s" opt; do
  case ${opt} in
    v)
      vip_ip=${OPTARG}
      ;;
    p)
      vip_port=${OPTARG}
      ;;
    l)
      list_servers=1
      ;;
    s)
      log_to_syslog=1
      ;;
    *)
      usage
      ;;
  esac
done
shift $(($OPTIND - 1))

# the script name from ${0} wasn't getting passed to syslog properly, so specifying here
script_name="check_vips.sh"

error_exit(){
  if [ ${log_to_syslog} ]; then
    logger -p daemon.err -s -i -t ${script_name} "[ERROR] ${1:-""}" 1>&2
  else
    echo "${script_name}: [ERROR] ${1:-""}" 1>&2
  fi
  exit ${2:-"1"}
}

error_return(){
  if [ ${log_to_syslog} ]; then
    logger -p daemon.err -s -i -t ${script_name} "[ERROR] ${1:-""}" 1>&2
  else
    echo "${script_name}: [ERROR] ${1:-""}" 1>&2
  fi 
  return ${2:-"1"}
}

warn(){
  if [ ${log_to_syslog} ]; then
    logger -p daemon.warn -s -i -t ${script_name} "[WARN] ${1:-""}" 1>&2
  else
    echo "${script_name}: [WARN] ${1:-""}" 1>&2
  fi
}

verify_v4(){
  if [[ ${1} =~ ^[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3} ]]; then
    return 0
  else
    return 1
  fi
}

port_match(){
  if [[ ${1} =~ [0-9]{1,5} ]]; then
    return 0
  else
    return 1
  fi
}

verify_v6(){
  v6_octet_count=0
  local counter=1
  declare -gA v6_octets

  if [ $v6_ip ]; then
    unset v6_ip
  fi

  #Verify each octet is a valid ipv6 hex
  IFS=':' read -a tmp_octets <<< ${1}

  #Need to count for compressed
  for octet in ${tmp_octets[@]}; do
    (( v6_octet_count++ ))
  done

  if [ ${v6_octet_count} -lt 8 ]; then
    compressed=1
  fi

  for hex in ${tmp_octets[@]}; do
    #Verify the hex to see if valid
    if [[ ${hex} =~ ^[0-9A-Fa-f]{1,4}$ ]]; then
      if [ ${counter} -lt ${v6_octet_count} ]; then
        v6_ip+=( "${hex}:" )
        (( counter++ ))
      elif [ ${counter} -le ${v6_octet_count} ] && [ ${compressed} ]; then
        v6_ip+=( ":" )
        (( octet_count++ ))
      fi
    else
      unset tmp_octets
      return 1
    fi
  done

  unset tmp_octets
  return 0
}

error_check(){
  if [ ! ${vip_ip} ] || [ ! ${vip_port} ]; then
    error_exit "must specify v4|v6 address and port" "1"
  fi

  if ! port_match ${vip_port}; then
    error_exit "port specified is not valid" "1"
  fi

  if verify_v4 ${vip_ip}; then 
    vip_type="v4"
  elif verify_v6 ${vip_ip}; then
    vip_type="v6"
  else
    error_exit "${vip_ip} is not a valid ip address" "1"
  fi
}

check_ipvs_v4(){
declare -gA bound_v4_reals

#scrape the ipvsadm output for the virtual server and the reals.  If at least one real
#is assigned the vip is "healthy" and we return 0, if no servers we return 1
while read junk ip method weight activecon inactivecon rest_of_line; do
  #skip the vip ip itself
  if [ "${ip}" = "${vip_ip}:${vip_port}" ]; then
    continue
  fi

  if verify_v4 ${ip}; then
    if [ ${weight} -eq 0 ]; then
      real_state="Active_Draining"
    else
      real_state="Active_Serving"
    fi
    bound_stats=( "M:${method}" "W:${weight}" "A:${activecon}" "I:${inactivecon}" "S:${real_state}" )
    bound_v4_reals[${ip}]=${bound_stats[@]}
  fi

  if [ ! ${list_servers} ] && [ "${#bound_v4_reals[@]}" -gt 0 ]; then
    break
  fi
done< <(ipvsadm -Ln -t ${vip_ip}:${vip_port} || warn "Vip ${vip_ip}:${vip_port} keepaliveD is not running, or no service runs on this ip:port")

if [ "${#bound_v4_reals[@]}" -gt 0 ]; then
  if [ ${list_servers} ]; then
    if ip route | grep -q ${vip_ip}; then
      local vip_state=MASTER
    else
      local vip_state=BACKUP
    fi
    echo "M=Method W=Weight A=ActiveConn I=InActConn S=State"
    echo "Virtual Server: ${vip_ip} (${vip_state})"
    for real in ${!bound_v4_reals[@]}; do
      echo "   real-server: $real ${bound_v4_reals[$real]}"
    done
  fi
  return 0
else
  if [ ${list_servers} ]; then
    if ip route | grep -q ${vip_ip}; then
      local vip_state=MASTER
    else
      local vip_state=BACKUP
    fi
    echo "Virtual Server: ${vip_ip} (${vip_state})"
    echo "   NO REAL SERVERS BOUND!"
    echo "   NO REAL SERVERS BOUND!"
    #return 1
    error_return "No real servers bound for Virtual Server: ${vip_ip} (${vip_state})" "1"
  fi
  #return 1
  error_return "No real servers bound for Virtual Server: ${vip_ip} (${vip_state})" "1"
fi
}

check_ipvs_v6(){
declare -gA bound_v6_reals

while read junk ip method weight activecon inactivecon rest_of_line; do
  if [ ! "${vip_block}" ]; then
    #skip the vip ip itself
    if [ "${ip}" = "[${vip_ip}]:${vip_port}" ]; then
      vip_block=1
      continue
    fi
    continue
  fi

  IFS='[]' read -a parsed_ip <<< ${ip}

  if verify_v6 ${parsed_ip[1]}; then
    if [ ${weight} -eq 0 ]; then
      real_state="Active_Draining"
    else
      real_state="Active_Serving"
    fi
    bound_stats=( "M:${method}" "W:${weight}" "A:${activecon}" "I:${inactivecon}" "S:${real_state}" )
    bound_v6_reals[${ip}]=${bound_stats[@]}
  fi

  if [ ! ${list_servers} ] && [ "${#bound_v6_reals[@]}" -gt 0 ]; then
    break
  fi
done< <(ipvsadm -Ln -t [${vip_ip}]:${vip_port} || warn "Vip [${vip_ip}]:${vip_port} is broken or keepaliveD is not running")

if [ "${#bound_v6_reals[@]}" -gt 0 ]; then
  if [ "${list_servers}" ]; then
    vip_block=`ip -6 route show to match ${vip_ip} | awk '{print $1}'`
    if [ "${vip_block}" != " " ] && [ "${vip_block}" != "default" ]; then
      vip_state=MASTER
    else
      vip_state=BACKUP
    fi
    echo "M=Method W=Weight A=ActiveConn I=InActConn S=State"
    echo "Virtual Server: ${vip_ip} (${vip_state})"
    for real in ${!bound_v6_reals[@]}; do
      echo "   real-server: $real ${bound_v6_reals[$real]}"
    done
  fi
  return 0
else
  if [ "${list_servers}" ]; then
    vip_block=`ip -6 route show to match ${vip_ip} | awk '{print $1}'`
    if [ "${vip_block}" != " " ] && [ "${vip_block}" != "default" ]; then
      vip_state=MASTER
    else
      vip_state=BACKUP
    fi
    echo "Virtual Server: ${vip_ip} (${vip_state})"
    echo "   NO REAL SERVERS BOUND!"
    echo "   NO REAL SERVERS BOUND!"
  fi
  #return 1
  error_return "NO REAL SERVERS BOUND for Virtual Server: ${vip_ip} (${vip_state})" "1"
fi
#return 1
error_return "check_ipvs_v6 encountered an error for Virtual Server: ${vip_ip} (${vip_state})" "1"
}

main(){
error_check

if [ ${vip_type} = "v4" ]; then
  check_ipvs_v4
elif [ ${vip_type} = "v6" ]; then
  check_ipvs_v6
fi
}

main
