#!/usr/bin/env bash
#-----------------------------------------------------------------------------
# Updates config files from media vault.  
#
# 1) download md5 from media vault
# 2) on download error:
#   a) echo error to stderr
#   b) if required flag then exit with error code 2 otherwise 0
# 3) compare downloaded md5 checksum to current config file's checksum
# 4) if the checksums differ:
#   a) download config from media vault
#   b) verify download checksum to md5 checksum exit on failure
#   c) replace current config file with downloaded file 
#
# Usage:
#   URL - FQDN of media vault's generated config file URL 
#   filename - config file fully qualified path and filename
#   hash URL - FQDN of media vault's generaged config file md5 URL
#   required - indicates whether to fail with an error code if config file
#              isn't updated for any reason (defaults to 1)
#
# Returns: Will exit with the following error level:
#   - 0 no failure or if <required> is 0 -- if the file isn't required then
#     no error is returned 
#   - 1 usage error
#   - 2 other error
#
# Note that errors are logged to stderr while general logging uses stdout. 
#  This script assumes it will return the change state and it is called with
#  cmd.run stateful parameter set to True.  Therefore the last line of
#  stdout will contain "changed=yes comment='something changed'" if the
#  state is True.  Otherwise no stdout is returned which indicates no changes.
#
# Example sls:
# sae|slb|slb-config-{{ edge_conf }}:
#  cmd.run:
#    - name: <path>/update_config_files.sh <URL> <file name> <HASH URL> [0|1]
#    - stateful: True
#
# Additionally note the support for using local sources vs. remote sources.
#  This is only supported in the lab environment. In this case the usage is
#  as follows:
#
# Local/Test mode Usage:
#   URL - Full path of the local source vs the FQDN of media vault's generated
#         config file URL (eg /etc/keepalived/local_servers.dat)
#   filename - config file fully qualified path and filename
#   hash URL - an empty string
#   required - indicates whether to fail with an error code if config file
#              isn't updated for any reason (defaults to 1)
#
#   Example local/test:
#     /etc/keepalived/scripts/update_config_files.sh \
#       "/etc/keepalived/local_servers.dat" \
#       /etc/keepalived/servers.dat "" "1"
#-----------------------------------------------------------------------------

if [ "$#" -lt 3 ]; then
   (>&2 echo "Usage: $0 <URL> <file_name>, <hash URL> [required|not_required]")

   exit 1
fi

# get a unique ID we can use for tmp file path
UUID=$(cat /proc/sys/kernel/random/uuid 2>&1)
if [ $? -ne 0 ]; then
    # just in case uuid changes for some reason
    # use a combination of the process ID for a UUID
    UUID=$$--$$
fi

#-----------------------------------------------------------------------------
# clean_up_exit: removes any temp files created then exits with error code
#-----------------------------------------------------------------------------
clean_up_exit() {
   local exit_code=${1:-0}
   # remove tmp files
   rm /tmp/${UUID}* >/dev/null 2>&1

   exit ${exit_code}
}

#-----------------------------------------------------------------------------
# download_file: downloads specified file, checks for errors, exits on error
#-----------------------------------------------------------------------------
download_file() {
   RESULTS=$(curl -o $1 --fail --write-out "%{http_code}\n" --silent "$2" 2>&1)
   # non zero return is a failure -- error 22 indicated an HTTP error
   if [ $? -ne 0 ]; then
       (>&2 echo "$URL ERROR: $RESULTS - Required: ${REQUIRED}") && clean_up_exit $REQUIRED
   fi

}

# command line args
URL=$1
LOCAL_FILE=$2
HASH_URL=$3

# default to REQUIRED (0 NOT, 1 Required)
REQUIRED=${4:-1}

FILE_NAME=$(basename ${LOCAL_FILE})

# if $URL ends with 'local_' + $FILE_NAME use the local config file vs remote
if [[ $URL == *local_$FILE_NAME ]]; then
    LOCAL_CONFIG_FILE=$URL

    # create MD5 of config and local config files
    LOCAL_CONFIG_HASH=$(md5sum ${LOCAL_CONFIG_FILE} | cut -d ' ' -f1)
    CONFIG_HASH=$(md5sum ${LOCAL_FILE} | cut -d ' ' -f1)

    # if MD5 values do not match, update the config file with the local
    if [ "${LOCAL_CONFIG_HASH}" != "${CONFIG_HASH}" ]; then

        # replace config file with local config file
        if ! cp "${LOCAL_CONFIG_FILE}" "${LOCAL_FILE}"; then
            (>&2 echo "unable to update config file: $? ${RESULT}") && clean_up_exit 2
        fi

        echo # an empty line
        echo "changed=yes comment='replaced $LOCAL_FILE with $LOCAL_CONFIG_FILE'"
    fi
else

    # 1) download MD5 from media vault
    #echo $HASH_URL
    HASH_FILE=/tmp/${UUID}_${FILE_NAME}.md5
    download_file "${HASH_FILE}" "${HASH_URL}"
    HASH=$(cut -d ' ' -f1 ${HASH_FILE})

    # 2) verify MD5 against existing local file, only update local file if there's a difference
    #FILE_MD5=$(md5sum ${FILE_NAME}
    LOCAL_HASH=$(md5sum ${LOCAL_FILE} | cut -d ' ' -f1)
    if [ "${LOCAL_HASH}" != "${HASH}" ]; then
    #if ! cmp "${HASH_FILE}" "${LOCAL_HASH_FILE}" > /dev/null 2>&1; then
        (>&2 echo "download ${HASH} does not match local ${LOCAL_HASH}")

        # get file from media vault
        TMP_FILE=/tmp/${UUID}_${FILE_NAME}
        download_file "${TMP_FILE}" "${URL}"

        # make sure download matches hash
        TMP_HASH=$(md5sum ${TMP_FILE} | cut -d ' ' -f1)
        if [ "${TMP_HASH}" != "${HASH}" ]; then
        #if ! cmp "${HASH_FILE}" "${TMP_HASH_FILE}" > /dev/null 2>&1; then
            (>&2 echo "download of ${FILE_NAME} has invalid hash ${TMP_HASH} - ${HASH}") && clean_up_exit 2
        fi

        # replace current local file with downloaded file
        if ! cp "${TMP_FILE}" "${LOCAL_FILE}"; then
            (>&2 echo "unable to update config file: $? ${RESULT}") && clean_up_exit 2
        fi

        echo # an empty line
        echo "changed=yes comment='replaced local with download'"
    fi
fi
clean_up_exit

