#!/bin/bash

#Script to health check the backend machines.  First check to see if we can do an
#https check, if that is successful, we do the dsr ping check.  If both are successful
#we return a value to keepalived that determines the servers weight.

machine=${1}
vip=${2}

server_state="/etc/keepalived/server_states/${vip}/"

#exit 1 on place holder entries in the server configs
if [ ${machine} = "192.168.255.255" ] || [ ${machine} = "fc00:aaaa::1" ]; then
  exit 1
fi

# global weight return value
# default to 0 which tells KAD to leave the weight unchanged (a no-op)
weight_return=0

# global used to indicate when dsr ping is progress
dsr_ping_in_progress=0

####
# SIGTERM handler -- Exits the process with error level 1 which indicates the health check failed.
#
# Per keepalived.conf(5):
# All scripts should be written so that they will terminate on receipt of a SIGTERM signal. Scripts will be sent
# SIGTERM if their parent terminates, or it is a script the keepalived is awaiting its exit status and it has run
# for too long.
#
# When SIGTERM is caught, call create_server_state to avoid flapping the service. see https://jira.atlas.llnw.com/browse/SLB-266
###
handle_sigterm() {
  logger -t $0 "sigterm: ${machine},${vip},${dsr_ping_in_progress},${weight_return}"

  if [[ ${dsr_ping_in_progress} -eq 1 ]]; then
    create_server_state
  fi

  if [ -f ${server_state}${machine} ]; then
    dlogger -t $0 "${server_state}${machine} exists"
    exit 1
  fi
  exit ${weight_return}
}

trap handle_sigterm SIGTERM

debug=`global_conf_get.pl macping_debug`
dlogger(){
  if [ "$debug" = "on" ] ; then
    logger $@
  fi
}

#Weights are stored in this file by gen_slb_configs.pl
get_weights(){
  local file="/etc/keepalived/server_weights/${machine}"
  if [ -f ${file} ]; then
    current_weight=$(awk '/^current/{print $2}' $file)
    default_weight=$(awk '/^default/{print $2}' $file)
    current_state=$(awk '/^state/{print $2}' $file)
  fi

  # set defaults if values are empty
  if [ -z ${current_weight} ]; then
    current_weight=100
  fi
  if [ -z ${default_weight} ]; then
    default_weight=100
  fi
  if [ -z ${current_state} ]; then
    current_state='active'
  fi
}

save_weights(){
  local file="/etc/keepalived/server_weights/${machine}"
  cat << EOF > ${file}
---
current: ${new_weight}
default: ${default_weight}
state: ${current_state}
EOF
}

http_check(){
  #use curl to capture connect time, 1st byte response time, http response code.
  #If times are abnormally long, we decrement the server weight to give it less traffic.

  local options=( -x [${1}]:8000 -I --head --http1.1 --max-time 6 -s -o /dev/null --user-agent slb_health_check )
  local curl_output=( `curl ${options[*]} -H 'Connection: close' -H 'Proxy-connection: close' -w '%{http_code} %{time_connect} %{time_pretransfer}' http://test.vo.llnwd.net/o1/geo-health.txt` )
  local http_code=${curl_output[0]}
  local time_connect=${curl_output[1]}
  local first_byte=${curl_output[2]}

  #Read current and default weights
  get_weights

  #Machine is busy if time_connect gt 1 second
  #EP is busy if first_byte gt 1 second
  if [ ${time_connect:0:1} -ge 2 ] || [ ${first_byte:0:1} -ge 2 ]; then
      # simply log if time meterics are exceeded
      dlogger -t $0 "${machine} time metrics exceeded: ${time_connect}, ${first_byte}"
  fi

  # set the new_weight from the default_weight
  # Add 2 to accommodate for error codes
  new_weight=$(( default_weight + 2 ))

  #
  # Check the return code. Our servers will return 666 or 667 when draining.
  # This routine returning a 2 actually tells keepalived the weight is zero, so stop sending new
  # connections to this server, preserve what connections you have.
  #
  if [ ! ${http_code} ]; then
    weight_return=1
    dlogger -t $0 "Failed to return http code"
  elif [ ${http_code} = 200 ]; then
    weight_return=${new_weight}
  elif [ ${http_code} = 666 ] || [ ${http_code} = 667 ]; then
    weight_return=2
  else
    weight_return=1
    dlogger -t $0 "${machine} returned unknown http code ${http_code}"
  fi
}

split_requests(){
  #Use modulo to split ipv4 and ipv6 requests evenly

  ctime=`date +"%-M"`

  if [[ ${machine} =~ ^[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3} ]]; then
    route_list=( `ip route show to match ${machine} scope link` )
    IPN=( ${machine//./ } )
    #if last octet even and minute of hour even, do macping check
    if ! (( $ctime % 2 )) && ! (( ${IPN[3]} % 2 )); then
      macping=1
    fi
    #odd last octet and odd minute of hour time, do macping check
    if (( $ctime % 2 )) && (( ${IPN[3]} % 2 )); then
      macping=1
    fi
  else
    #Machine is ipv6, grab last octet and convert to decimal
    route_list=( `ip -6 route show to match ${machine} scope global` )
    IPN6=( ${machine//:/ } )
    LIST_LEN=${#IPN6[@]}
    LAST_OCT=$(( $LIST_LEN - 1 ))
    IPN6_DEC=`printf '%d' 0x${IPN6[$LAST_OCT]}`
    #If minute is even and last octet divisible by 4, do macping check
    if ! (( $ctime % 2 )) && ! (( $IPN6_DEC % 4 )); then
      macping=1
    fi
    #If minute is odd and last octet not divisible by 4, do macping check
    if (( $ctime % 2 )) && (( ${IPN6_DEC} % 4 )); then
      macping=1
    fi
  fi
}

#
# Perform the macping to test DSR. This is perfromed multiple times since the transport is not reliable.
# We dont want some packet loss to take a server down.
#
do_dsr_ping() {
  # set global on to indicate the dsr ping process is in progress
  dsr_ping_in_progress=1
  for ((i=1; i <=3; i++))
  do
    if [ "$debug" = "on" ] ; then
      /etc/keepalived/scripts/macping -I ${ping_interface} -d -b -q -c 1 -w 3 ${vip} ${machine} > /dev/null || /etc/keepalived/scripts/macping -I ${ping_interface} -d -b -l -q -c 1 -w 3 ${vip} ${machine} > /dev/null
    else
      /etc/keepalived/scripts/macping -I ${ping_interface} -b -q -c 1 -w 3 ${vip} ${machine} || /etc/keepalived/scripts/macping -I ${ping_interface} -b -l -q -c 1 -w 3 ${vip} ${machine}
    fi
    dsr_res=$?
    if [ $dsr_res == 0 ]
    then
      break
    fi
  done

  # set global off to indicate dsr ping process is complete
  dsr_ping_in_progress=0

  return $dsr_res
}

dsr_ping_check(){
  split_requests

  #route should be attached to eth or bond interface else we have no route
  if [[ ${route_list[2]} =~ eth|bond ]]; then
    ping_interface=${route_list[2]}
  else
    dlogger -t $0 "could not find ping interface"
    return 1
  fi

  #If we are a backup vip, unset macping var
  if [ ${current_state} ] && [ ${current_state} = "backup" ]; then
    dlogger -t $0 "backup state: ${current_state}"
    unset macping
  fi

  if [ ${macping} ]; then
    do_dsr_ping
    mac_exit=$?
    #If macping succeeds, check for bad state file and remove it
    if [ $mac_exit = 0 ]; then
      if [ -f ${server_state}${machine} ]; then
        rm ${server_state}${machine}
        return 0
      fi
    #macping returned 2 which is an internal error and we can't trust the check, return 0 unless a statefile exists
    elif [ $mac_exit = 2 ]; then
      if [ -f ${server_state}${machine} ]; then
        dlogger -t $0 "Macping exited with exit code 2, ${server_state}${machine} exists"
        return 1
      else
        dlogger -t $0 "Macping exited with exit code 2, ${server_state}${machine} does not exist"
        return 0
      fi
    else
      # dsr ping failed, create state file and return 1
      create_server_state && dlogger -t $0 "Macping exited with exit code $mac_exit" && return 1
    fi
  else
    #we didn't have to run macping this time, but still check for bad file state and go off previous fail/pass
    if [ -f ${server_state}${machine} ]; then
      dlogger -t $0 "Macping did not run but ${server_state}${machine} exists, so returning 1"
      return 1
    else
      return 0
    fi
  fi
}

###
# helper: creates server state file
###
create_server_state() {
  if [ ! -d ${server_state} ]; then
    mkdir $server_state
  fi
  touch ${server_state}${machine}
}

check_options(){
  # exit 2 (drain) if server state = 'drain'
  # this indicates server is in down_servers.dat
  if [ "${current_state}" ] && [ "${current_state}" = "drain" ]; then
    exit 2
  fi

  # exit 2 if machine in overrides.yml and value is 'drain' or 'remove'
  override=$(awk "/$machine/{print \$2}" /etc/keepalived/config_containers/overrides.yml)
  case $override in
    drain)
      exit 2
      ;;
    remove)
      exit 2
      ;;
  esac
}

main(){
  #exit status 0: svc check success, weight unchanged
  #exit status 1: svc check failed, server removed
  #exit status 2-255: svc check success && weight changed to 2 less than exit status

  http_check ${1}

  #If everything ok, return dynamic weight to keepalived
  if [ ${weight_return} -gt 2 ]; then
    if dsr_ping_check; then
      check_options
      return ${weight_return}
    else
      return 1
    fi
  fi

  #Traffic draining on this server, no new requests should go to this box, set weight 0 by returning 2
  if [ ${weight_return} -eq 2 ]; then
    if dsr_ping_check; then
      return 2
    else
      return 1
    fi
  fi

  #Failed requests, remove the server from the pool by returning 1
  if [ ${weight_return} -eq 1 ]; then
    return 1
  fi
}

main ${1} ${2}
