from __future__ import print_function
import os
import platform
import re
import sys
from argparse import ArgumentParser

import socket

import datetime
import netaddr
import yaml
import logging
import logging.handlers

# configure logging
log = logging.getLogger(__name__)
log.setLevel(logging.WARNING)
handler = logging.handlers.SysLogHandler(address='/dev/log')
formatter = logging.Formatter('%(module)s.%(funcName)s: %(message)s')
handler.setFormatter(formatter)

log.addHandler(handler)

# file paths
base_path = r"/etc/keepalived"
slb_global_config_file = r"{}/global_conf.yml".format(base_path)
macping = '{}/scripts/macping'.format(base_path)

# get the hostname
hostname = platform.node()

# dns cache
local_dns_cache = dict()

# add reference to socket.getaddrinfo
prv_getaddrinfo = socket.getaddrinfo


def override_dns(domain, ip):
    """
    Helper: Updates local_dns_cache with domain and override IP
    Args:
        domain: (str) FQDN of domain to override with ip
        ip: (str) IPv4 or v6 IP Address
    """
    local_dns_cache[domain] = ip


def new_getaddrinfo(*args):
    """
    Overrides socket.getaddrinfo to support pinning an IP to a domain
    Args:
        *args: (tuple) IP Address 4-tuple (ex: ('test.vo.llnwd.net', 443, 0, 1))

    Returns:
        (list) of tuples (ex: [(2, 1, 6, '', ('111.119.11.41', 443))] )
    """
    log.debug('args: {}'.format(args))
    if args[0] in local_dns_cache:
        log.debug("Forcing FQDN: {} to IP: {}".format(args[0], local_dns_cache[args[0]]))
        _res = prv_getaddrinfo(local_dns_cache[args[0]], *args[1:])
    else:
        _res = prv_getaddrinfo(*args)
    log.debug('getaddrinfo: results: {}'.format(_res))
    return _res


# override socket.getaddrinfo with our version
socket.getaddrinfo = new_getaddrinfo


def get_pop():
    """
    Parse the POP from the hostname
    Returns:
        (str) POP name parsed from hostname
    """
    if hostname.startswith('slb'):
        try:
            return hostname.split('.')[1]
        except Exception as ex:
            log.debug('error {}'.format(ex))
            raise ValueError('invalid SLB hostname: {} -- a valid SLB hostname must be fully qualified')
    else:
        raise ValueError('invalid SLB hostname: {} -- a valid SLB hostname start with "slb"'.format(hostname))


def drain_override(server_ip):
    """checks the overrides file for server's drain/remove state"""

    _fname = r'{}/config_containers/overrides.yml'
    if not os.path.isfile(_fname):
        # overrides.yml doesn't exist then quietly return False
        return False

    try:
        with open(_fname) as f:
            _data = yaml.load(f)
            if _data.get(server_ip, '') in ['drain', 'remove']:
                # found a match; return True to indicate drain
                log.debug('{} found drain|remove state in overrides'.format(server_ip))
                return True
    except Exception as ex:
        log.debug('error retrieving server weights: {}'.format(ex))

    # return default False which indicates the server has no override drain state
    return False


def get_weights(server_ip):
    """
    retrieve weights from server_weights file
    Args:
        server_ip: (str) server IP Address

    Returns:
        (tuple) (current weight or 100, default weight or 100, current state or 'active')
    """
    _data = {}
    try:
        with open(r'{}/server_weights/{}'.format(base_path, server_ip)) as f:
            _data = yaml.load(f)
    except Exception as ex:
        log.debug('error retrieving server weights: {}'.format(ex))

    return _data.get('current', 100), _data.get('default', 100), _data.get('state', 'active')


def save_weights(server_ip, current=100, default=100, state='active'):
    """save weights to server_weights file"""
    try:
        with open(r'{}/server_weights/{}'.format(base_path, server_ip), "w") as f:
            f.write('---\ncurrent: {}\ndefault: {}\nstate: {}\n'.format(current, default, state))
    except Exception as ex:
        log.debug('error saving server weights: {}'.format(ex))


def get_global_config_option(value, raise_error_on_not_found=False):
    """
    Get the specified option value from the global config file.
    Note that this function's logic is a port of global_conf_get.pl
    Args:
        value: (str) global config option to retrieve
        raise_error_on_not_found: (bool) raise an error if option value not found (default False)

    Returns:
        (str) value of global config option if found
    """
    _data = None
    with open(slb_global_config_file) as f:
        _data = yaml.load(f)

    if not _data or value not in _data:
        raise ValueError('no entry for option: {} found in {}'.format(value, slb_global_config_file))

    _data = _data.get(value)

    # try fqdn first
    if hostname in _data:
        return _data[hostname]

    # next try pop
    _pop = get_pop()
    if _pop in _data:
        return _data[_pop]

    # try default
    if 'default' in _data:
        return _data['default']

    if raise_error_on_not_found:
        raise ValueError('ERROR: A default value is missing for option {}'.format(value))

    return _data


def get_ping_interface(ip_address):
    """
    Helper: Get the interface associated with specified IP Address from route table.  Note that command:
    "ip route show to match" is used which will find an entry in the routing table for the specified IP.
    Args:
        ip_address: (str) IP Address in v4 dot or v6 quad notation

    Returns:
        (str) interface name for specified IP Address
    """
    _ip = to_ip_address(ip_address, True)
    if _ip:
        if _ip.version == 6:
            _cmd = 'ip -6 route show to match {} scope global'.format(ip_address)
        else:
            _cmd = 'ip route show to match {} scope link'.format(ip_address)

        s = os.popen(_cmd)
        o = s.read()
        if s.close() is None:
            # the results could be an empty string if no match is found
            # a misconfigured CDN vlan subnet is possible
            _results = o.split()
            if _results and len(_results) > 2:
                if re.search(r'^(bond|eth)', _results[2]):
                    # only process if the interface is bond or eth
                    # probably want to change this to 'bond' only
                    return _results[2]

                raise ValueError('unsupported interface: {} found for {}'.format(_results[2], ip_address))

    raise ValueError('no interface found for {}'.format(ip_address))


def touch(path):
    """
    Emulate touch command
    Args:
        path: (str) file path to create/touch
    """
    with open(path, 'a'):
        os.utime(path, None)


def dsr_ping(host, vip, retries=3):
    """
    Performs a DSR ping using macping.
    Args:
        host: (str) host IP Address
        vip: (str) VIP
        retries: (int) number of retries (default 3)

    Returns:
        (int)
             0 on successful ping
            -1 if macping wasn't/didn't need to be performed
            otherwise macping's error level
    """
    if not (host and vip):
        raise ValueError('parameters host and vip are required')

    _minute = datetime.datetime.now().minute
    _host_ip = to_ip_address(host, raise_on_error=True)

    # get the interface associated with host IP Address
    _ping_interface = get_ping_interface(host)

    _perform_dsr = False
    if not _minute % 2 and not _host_ip.words[-1] % 2:
        # if last octet even and minute of hour even, do macping check
        log.debug('minute: {} even last: {} even'.format(_minute, _host_ip.words[-1]))
        _perform_dsr = True
    elif _minute % 2 and _host_ip.words[-1] % 2:
        # odd last octet and odd minute of hour time, do macping check
        log.debug('minute: {} odd last: {} odd'.format(_minute, _host_ip.words[-1]))
        _perform_dsr = True

    if not _perform_dsr:
        _exit_level = -1
    else:

        _exit_level = 0
        _macping_command = ("{macping} -I {ping_interface} {debug} -b -q -c 1 -w 3 {vip} {host} || {macping} -I "
                            "{ping_interface} {debug} -b -l -q -c 1 -w 3 {vip} {host}").format(
                macping=macping, ping_interface=_ping_interface, vip=vip, host=host,
                debug='-d' if log.isEnabledFor(logging.DEBUG) else '')

        log.debug('macping command: {}'.format(_macping_command))

        # Check DSR
        for x in range(retries):
            s = os.popen(_macping_command)
            _exit_status = s.close()
            if _exit_status is None:
                # success - return True
                return 0

            # get the exit status
            _exit_level = os.WEXITSTATUS(_exit_status)
            log.debug('dsr_ping exit level: {}'.format(_exit_level))

    return _exit_level


def to_ip_address(ip, raise_on_error=False):
    """
    helper: convert ip to netaddr.IPAddress
    Args:
        ip: (str|netaddr.IPAddress) IP Address to convert
        raise_on_error: (bool) raise error is True

    Returns:
        (netaddr.IPAddress) on success otherwise None
    """
    try:
        if isinstance(ip, netaddr.IPAddress):
            return ip
        elif isinstance(ip, str):
            return netaddr.IPAddress(ip)
        else:
            raise ValueError('invalid type {}'.format(type(ip)))
    except Exception as ex: # noqa
        log.debug('Error {} is an invalid IP Address'.format(ip))
        if raise_on_error:
            raise
        return None


def convert(to, who, default=u''):
    """
    Converts data to a specific datatype.
    In case of error, will return a default value.

    :param to:      datatype to be casted to.
    :param who:     value to cast.
    :param default: value to return in case of error.
    :return: a str value.
    """
    if who is None:
        return default
    try:
        return to(who)
    except:  # noqa
        return default


class ArgParser(ArgumentParser):
    def error(self, message):
        """
        override ArgumentParser.error() to support specific app logging
        Args:
            message: (str) error message
        """
        print('usage error: {}'.format(message), file=sys.stderr)
        print(self.format_usage(), file=sys.stderr)
        # log a warning so that the usage error is logged to syslog
        log.warning('usage error: {}'.format(message))

        # must exist here -- do not return
        exit(1)
