#!/usr/bin/perl
use strict;
use warnings;
use Sys::Hostname;
use Data::Dumper;
use POSIX;
use YAML qw(LoadFile);
use Regexp::IPv6 qw($IPv6_re);
use Net::CIDR;

$ENV{'PATH'} = join ":", qw{
  /etc
  /usr/local/sbin
  /usr/local/bin
  /sbin
  /usr/sbin
  /bin
  /usr/bin
};

my $hostname = hostname;
my $tmp_file = "/etc/quagga/bgpd.conf.new";
my $bgp_conf = "/etc/quagga/bgpd.conf";

main();

sub main {
  unless ( $hostname =~ /^slb\d\d/ ) {
    die "This script is only meant for software load balancers\n";
  }

  create_bgpd_conf();
  if ( check_proc("/var/run/quagga/bgpd.pid") == 0 ) {
    read_bgp_configs();
  }
  rename_file_perms();
}

sub get_ip_version {
  #Check to see if address is ipv4 or ipv6
  my $ip = shift;
  my $ip_version;

  if ( $ip =~ /^(([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\.){3}([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])$/ ) {
    $ip_version = "v4";
  } elsif ( $ip =~ /^$IPv6_re$/ ) {
    $ip_version = "v6";
  } else {
    die "$ip is not valid ipv4 or ipv6";
  }

  return $ip_version;
}

sub create_bgpd_conf {
  my $yml = LoadFile('/etc/keepalived/slb_conf.yml');
  my $my_conf = undef;
  my $yml_vips = undef;

  if ($yml->{config_version} && (($yml->{config_version} eq "2") || ($yml->{config_version} eq "2_ecmp")) ) {
    $my_conf = $yml->{slb_configs}->{$hostname} or die "No config for $hostname found in /etc/keepalived/slb_conf.yml $!";
    $yml_vips = $yml->{vips} or die "No vip config found in /etc/keepalived/slb_conf.yml $!";
  } else {
    $my_conf = $yml->{$hostname}->{slb_config} or die "No config for $hostname found in /etc/keepalived/slb_conf.yml $!";
    $yml_vips = $yml->{$hostname}->{vips} or die "No vip config for $hostname found in /etc/keepalived/slb_conf.yml $!";
  }
  my @keys = keys %$yml_vips;
  my @v4_vip_ips;
  my @v6_vip_ips;
  my $seq = 0;

  # if BGP isn't defined then log a warning and exit with error code 0
  # returning a non-zero will cause salt to raise an error
  if (!defined($my_conf->{bgp_config}->{v4_neighbor})) {
    print STDERR "WARNING: BGP is not configured, ignoring BGP";
    exit 0
  }

  foreach my $vip_name ( @keys ) {
    my $vip_ip = (split '/',$yml_vips->{$vip_name}->{vip_addr})[0];
    my $vip_cidr = (split '/',$yml_vips->{$vip_name}->{vip_addr})[1];
    my $vip_ip_cidr = $yml_vips->{$vip_name}->{vip_addr};
    my @vip_range = Net::CIDR::cidr2range("${vip_ip_cidr}");
    my $vip_network = (split '-',$vip_range[0])[0];
    my $virtual_routes = $yml_vips->{$vip_name}->{virtual_routes};

    my $ip_version = get_ip_version($vip_ip);

    if ( $ip_version eq "v4" ) {
      push @v4_vip_ips, "${vip_network}/${vip_cidr}";
    }
    if ( $ip_version eq "v6" ) {
      push @v6_vip_ips, "${vip_network}/${vip_cidr}";
    }

    if ( $virtual_routes ) {
      foreach my $virt_route ( @{$virtual_routes} ) {
        my $virt_route_ip = (split '/',${virt_route})[0];
        my $ip_version = get_ip_version($virt_route_ip);
        if ( $ip_version eq "v4" ) {
          push @v4_vip_ips, "${virt_route}";
        }
        if ( $ip_version eq "v6" ) {
          push @v6_vip_ips, "${virt_route}";
        }
      }
    }
  }

  my $print_err = "Failed to write to $tmp_file";
  my $bgp_id = (split '/',$my_conf->{address})[0] or die "$print_err $!";
  open( my $bfh,'>',"$tmp_file" ) or die "Failed to open $tmp_file $!";
    print $bfh "!\n" or die "$print_err $!";
    print $bfh "router bgp $my_conf->{bgp_config}->{local_asn}\n" or die "$print_err $!";
    print $bfh " bgp router-id $bgp_id\n" or die "$print_err $!";
    print $bfh " no bgp default ipv4-unicast\n" or die "$print_err $!";
    print $bfh " redistribute kernel\n" or die "$print_err $!";
    print $bfh " redistribute connected\n" or die "$print_err $!";
    foreach my $v4_neighbor ( @{ $my_conf->{bgp_config}->{v4_neighbor} }) {
      print $bfh " neighbor $v4_neighbor remote-as $my_conf->{bgp_config}->{remote_asn}\n" or die "$print_err $!";
      print $bfh " neighbor $v4_neighbor advertisement-interval 2\n" or die "$print_err $!";
      print $bfh " neighbor $v4_neighbor activate\n" or die "$print_err $!";
      print $bfh " neighbor $v4_neighbor route-map lvs out\n" or die "$print_err $!";
      print $bfh " neighbor $v4_neighbor route-map deny-all in\n" or die "$print_err $!";
    }
    foreach my $v6_neighbor ( @{ $my_conf->{bgp_config}->{v6_neighbor} }) {
      print $bfh " neighbor $v6_neighbor remote-as $my_conf->{bgp_config}->{remote_asn}\n" or die "$print_err $!";
      print $bfh " neighbor $v6_neighbor activate\n" or die "$print_err $!";
    }
    print $bfh "!\n" or die "$print_err $!";
    print $bfh " address-family ipv6\n" or die "$print_err $!";
    print $bfh " redistribute kernel\n" or die "$print_err $!";
    print $bfh " redistribute connected\n" or die "$print_err $!";
    foreach my $v6_neighbor ( @{ $my_conf->{bgp_config}->{v6_neighbor} }) {
      print $bfh " neighbor $v6_neighbor activate\n" or die "$print_err $!";
      print $bfh " neighbor $v6_neighbor route-map lvs-v6 out\n" or die "$print_err $!";
      print $bfh " neighbor $v6_neighbor route-map ipv6-deny-all in\n" or die "$print_err $!";
    }
    print $bfh " exit-address-family\n" or die "$print_err $!";
    print $bfh " exit\n" or die "$print_err $!";
    print $bfh "!\n" or die "$print_err $!";
    foreach my $v4_vip ( @v4_vip_ips ) {
      $seq = $seq + 5;
      print $bfh "ip prefix-list lvs-vip seq $seq permit $v4_vip le 32\n" or die "$print_err $!";
    }
    print $bfh "!\n" or die "$print_err $!";
    print $bfh "access-list deny-all deny any\n" or die "$print_err $!";
    print $bfh "!\n" or die "$print_err $!";
    print $bfh "ipv6 access-list deny-all deny any\n" or die "$print_err $!";
    if ( @v6_vip_ips ) {
      print $bfh "!\n" or die "$print_err $!";
      $seq = 0;
      foreach my $v6_vip ( @v6_vip_ips ) {
        $seq = $seq + 5;
        print $bfh "ipv6 prefix-list lvs-vip-v6 seq $seq permit $v6_vip le 128\n" or die "$print_err $!";
      }
    }
    print $bfh "!\n" or die "$print_err $!";
    print $bfh "route-map deny-all deny 5\n" or die "$print_err $!";
    print $bfh " match ip address deny-all\n" or die "$print_err $!";
    print $bfh "!\n" or die "$print_err $!";
    print $bfh "route-map ipv6-deny-all deny 5\n" or die "$print_err $!";
    print $bfh " match ipv6 address deny-all\n" or die "$print_err $!";
    print $bfh "!\n" or die "$print_err $!";
    print $bfh "route-map lvs permit 10\n" or die "$print_err $!";
    print $bfh " match ip address prefix-list lvs-vip\n" or die "$print_err $!";
    print $bfh " set metric 0\n" or die "$print_err $!";
    print $bfh "!\n" or die "$print_err $!";
    print $bfh "route-map lvs-v6 permit 10\n" or die "$print_err $!";
    print $bfh " match ipv6 address prefix-list lvs-vip-v6\n" or die "$print_err $!";
    print $bfh " set metric 0\n" or die "$print_err $!";
    print $bfh "!\n" or die "$print_err $!";
    print $bfh "line vty\n" or die "$print_err $!";
    print $bfh "!\n" or die "$print_err $!";
  close ($bfh) or die "Failed to close bgpd.conf.auto $!";
}

sub read_bgp_configs {
  #You can not graceful Quagga, so if we need to add in prefixes when we add in new vips, we need
  #to do it via vtysh.

  my @v4_prefixs;
  my @v6_prefixs;
  #Read the actual configurations we just generated and see if we are missing pieces
  open ( my $bgpConfig,'<',"$tmp_file" ) or die "Failed to open $tmp_file $!";
    while ( my $line = <$bgpConfig> ) {
      if ( $line =~ /^ip prefix-list/ ) {
        push @v4_prefixs, (split ' ', $line )[6];
      }
      if ( $line =~ /^ipv6 prefix-list/ ) {
        push @v6_prefixs, (split ' ', $line )[6];
      }
    }
  close ($bgpConfig) or die "Failed to close vtysh output $!";

  my %mem_v4_prefixs = ();
  my %mem_v6_prefixs = ();
  #Read the in memory running bgp configs to see what's actually installed
  open ( my $bgpMem, "vtysh -c \"show running-config bgpd\" |" ) or die "Failed to read vtysh output $!";
    while ( my $line = <$bgpMem> ) {
      if ( $line =~ /^ip prefix-list/ ) {
        $mem_v4_prefixs{(split ' ', $line )[6]} = '1';
      }
      if ( $line =~ /^ipv6 prefix-list/ ) {
        $mem_v6_prefixs{(split ' ', $line )[6]} = '1';
      }
    }
  close ($bgpMem) or die "Failed to close vtysh output $!";

  #If ipv4 or ipv6 address not in hash above, add it to live running configs
  foreach my $address ( @v4_prefixs ) {
    if ( ! $mem_v4_prefixs{$address} ) {
      if ( system("vtysh -c \"conf t\" -c \"ip prefix-list lvs-vip permit $address le 32\"") != 0 ) {
        die "Failed to add $address to lvs-vip prefix-list $!";
      }
    }
  }
  foreach my $address ( @v6_prefixs ) {
    if ( ! $mem_v6_prefixs{$address} ) {
      #Address is missing, need to install via vtysh
      if ( system("vtysh -c \"conf t\" -c \"ipv6 prefix-list lvs-vip-v6 permit $address le 128\"") != 0 ) {
        die "Failed to add $address to lvs-vip-v6 prefix-list $!";
      }
    }
  }
  #soft/graceful clear the outbound sessions so our newly added address are advertised out
  system("vtysh -c \"clear ip bgp * soft out\"");
}

sub check_proc {
  my $pid_file = shift;
  my $pid;

  if ( -s $pid_file ) {
    open( my $fh, "<", $pid_file ) or die "could not open $pid_file: $!";
      ( $pid ) = <$fh>;
      chomp $pid;
    close ($fh) or die "Could not close $pid_file: $!";
    my $pid_running = kill 0, $pid;
    if ( $pid_running ) {
      return 0
    } else {
      return 1
    }
  }
  return 1
}

sub rename_file_perms {
  chmod 0644, $tmp_file or die "Failed to change permissions on file $tmp_file $!";
  my $uid = getpwnam "quagga";
  my $gid = getgrnam "quagga";
  chown $uid, $gid, $tmp_file or die "Failed to chown $tmp_file to quagga:quagga $!";
  rename $tmp_file,$bgp_conf or die "Failed to rename $tmp_file to $bgp_conf $!";
}


