#!/usr/bin/env python

import argparse
import logging
import ConfigParser
import mysql.connector
import json
import sys


arg_parser = argparse.ArgumentParser()
arg_parser.add_argument('--dbconf', help='MySQL DB config file', default='/etc/zabbix/zabbix-audit.conf')
arg_parser.add_argument('--section', help='Section in config file to read', default='mysql')
arg_parser.add_argument('--temp', help='Temp file to store last auditid', default='/tmp/zabbix-last-auditid')
arg_parser.add_argument('--log', help='Log file to write auditlog events to', default='/var/log/zabbix-audit/zabbix-audit.log')
arg_parser.add_argument('--cont', type=int, help='auditid in zabbix db to continue from')
args = arg_parser.parse_args()

logging.basicConfig(
    level=logging.INFO,
    format='[%(asctime)s] %(message)s',
    datefmt='%m/%d/%Y %H:%M:%S')
log = logging.getLogger('zaudit')


db_conf = {}
config_parser = ConfigParser.ConfigParser()
config_parser.read(args.dbconf)

if config_parser.has_section(args.section):
    items = config_parser.items(args.section)
    for item in items:
        db_conf[item[0]] = item[1]
else:
    raise Exception('{0} not found in the {1} file'.format(args.section, args.dbconf))


def get_audit_query(last_auditid):
    if args.cont:
        last_auditid = args.cont

    audit_query = """
        SELECT from_unixtime(`a1`.`clock`, '%Y-%m-%dT%T-07:00') as timestamp,
            `u`.`alias`,
            `a1`.`ip`,
            CASE `a1`.`action`
                WHEN 0 THEN 'add'
                WHEN 1 THEN 'update'
                WHEN 2 THEN 'delete'
                WHEN 3 THEN 'login'
                WHEN 4 THEN 'logout'
                WHEN 5 THEN 'enable'
                WHEN 6 THEN 'disable'
            END as `action`,
            CASE `a1`.`resourcetype`
                WHEN 0 THEN 'User'
                WHEN 11 THEN 'User group'
                WHEN 12 THEN 'Application'
                WHEN 13 THEN 'Trigger'
                WHEN 14 THEN 'Host group'
                WHEN 15 THEN 'Item'
                WHEN 16 THEN 'Image'
                WHEN 17 THEN 'Value map'
                WHEN 18 THEN 'IT service'
                WHEN 19 THEN 'Map'
                WHEN 2 THEN 'Configuration of Zabbix'
                WHEN 20 THEN 'Screen'
                WHEN 21 THEN 'Node'
                WHEN 22 THEN 'Scenario'
                WHEN 23 THEN 'Discovery rule'
                WHEN 24 THEN 'Slide show'
                WHEN 25 THEN 'Script'
                WHEN 26 THEN 'Proxy'
                WHEN 27 THEN 'Maintenance'
                WHEN 28 THEN 'Regular expression'
                WHEN 29 THEN 'Macro'
                WHEN 3 THEN 'Media type'
                WHEN 30 THEN 'Template'
                WHEN 31 THEN 'Trigger prototype'
                WHEN 4 THEN 'Host'
                WHEN 5 THEN 'Action'
                WHEN 6 THEN 'Graph'
                WHEN 7 THEN 'Graph element'
            END as `resourcetype`,
            COALESCE(NULLIF(`a1`.`resourcename`, ''),`a1`.`details`) as `name`,
            REPLACE(`a2`.`oldvalue`, '\\n', '') as oldvalue,
            REPLACE(`a2`.`newvalue`, '\\n', '') as newvalue,
            `a1`.`auditid`
        FROM `auditlog` AS `a1`
        LEFT JOIN `auditlog_details` AS `a2`
        ON `a1`.`auditid`=`a2`.`auditid`
        LEFT JOIN `users` AS `u`
        ON `a1`.`userid`=`u`.`userid`
        WHERE `a1`.`auditid` > {last_auditid}
        ORDER BY `a1`.`auditid` ASC;""".format(last_auditid=last_auditid)
    return audit_query


def get_last_auditid(filename):
    last_auditid = 0
    try:
        with open(filename, 'r') as f:
            last_auditid = int(f.read())
    except:
        exit("Could not read auditid file: %s" % filename)
    finally:
        return last_auditid


def save_auditid(filename, data):
    try:
        with open(filename, 'w') as f:
            f.write('{0}'.format(data))
    except:
        log.info("Can't write to file")


def main():
    last_auditid = get_last_auditid(args.temp)

    audit_query = get_audit_query(last_auditid)

    data = []

    try:
        db = mysql.connector.connect(**db_conf)
        cursor = db.cursor()
        cursor.execute(audit_query)
        data = [dict((cursor.description[k][0], str(value)) for k, value in enumerate(row)) for row in cursor.fetchall()]
    except:
        print "Unexpected error: ", sys.exc_info()[0]
        raise
    finally:
        cursor.close()
        db.close()

    if len(data) > 0:
        with open(args.log, 'a') as f:
            for row in data:
                f.write(json.dumps(row))
                f.write('\n')

            last_auditid = data[len(data)-1]['auditid']
            save_auditid(args.temp, last_auditid)
            log.info("Last auditid: {}".format(last_auditid))
    else:
        log.info("No new audit events since last run.  Previous auditid: {}".format(last_auditid))


if __name__ == '__main__':
    main()

